from typing import Dict
from abnumber import Position, Chain
import numpy as np


def get_germline_family_residue_frequency(chain: Chain, imgt_chain: Chain,
                                          germline_family: str) -> Dict[Position, Dict[str, float]]:
    frequencies = GERMLINE_FREQUENCY.get(germline_family, {})
    return {pos: _get_frequency(frequencies, imgt_pos) for pos, imgt_pos in zip(chain.positions, imgt_chain.positions)}


def get_chain_type_residue_frequency(chain: Chain, imgt_chain: Chain) -> Dict[Position, Dict[str, float]]:
    frequencies = GERMLINE_FREQUENCY[chain.chain_type]
    return {pos: _get_frequency(frequencies, imgt_pos) for pos, imgt_pos in zip(chain.positions, imgt_chain.positions)}


def _get_frequency(frequencies: Dict[Position, Dict[str, float]], pos: Position):
    assert pos.scheme == 'imgt', f'Expected imgt scheme, got {pos.scheme}: {pos}'
    return frequencies.get(pos.format(chain_type=False))


def get_oasis_percentile(chain_type: str, oasis_identity: float, min_fraction_subjects: float):
    if np.isnan(oasis_identity):
        return np.nan
    assert chain_type in ['H', 'K', 'L', 'mean'], f'Invalid chain type: {chain_type}'
    assert 0 <= oasis_identity <= 1
    assert 0 < min_fraction_subjects < 1
    subjects_key = int(min_fraction_subjects*100)
    identity_lower_bound = int(np.floor(oasis_identity*100))
    identity_upper_bound = int(np.ceil(oasis_identity*100))
    correction = oasis_identity*100 - np.floor(oasis_identity*100)
    if chain_type == 'mean':
        percentiles = MEAN_PERCENTILES
    elif chain_type == 'H':
        percentiles = HEAVY_PERCENTILES
    else:
        percentiles = LIGHT_PERCENTILES
    lower = percentiles[subjects_key][identity_lower_bound]
    upper = percentiles[subjects_key][identity_upper_bound]
    return lower + (upper - lower) * correction


# Calculated from a subset of OAS (see BioPhi paper)
# Using IMGT numbering
GERMLINE_FREQUENCY = {
  "H": {
    "1": { "D": 0.0044, "E": 0.3186, "H": 0.0036, "Q": 0.6732, "R": 0.0001, "V": 0.0001 },
    "2": { "A": 0.0063, "D": 0.001, "E": 0.0047, "F": 0.0009, "G": 0.005, "I": 0.0236, "K": 0.0001, "L": 0.0734, "M": 0.0099, "N": 0.0001, "P": 0.0008, "Q": 0.0003, "R": 0.0003, "S": 0.0002, "T": 0.0003, "V": 0.8728, "W": 0.0002 },
    "3": { "A": 0.0003, "D": 0.0003, "E": 0.0027, "F": 0.0001, "G": 0.0001, "H": 0.0196, "I": 0.0001, "K": 0.0037, "L": 0.0051, "M": 0.0001, "N": 0.0006, "P": 0.0075, "Q": 0.9134, "R": 0.0092, "S": 0.0005, "T": 0.0357, "V": 0.0003, "W": 0.0001, "Y": 0.0008 },
    "4": { "A": 0.0001, "F": 0.0009, "G": 0.0001, "H": 0.0001, "I": 0.0006, "L": 0.9786, "M": 0.0035, "P": 0.0032, "Q": 0.0007, "R": 0.003, "S": 0.0003, "T": 0.0001, "V": 0.0086, "W": 0.0003 },
    "5": { "A": 0.0035, "D": 0.0002, "E": 0.005, "F": 0.0008, "G": 0.0024, "H": 0.0026, "I": 0.0007, "K": 0.0303, "L": 0.0809, "M": 0.003, "N": 0.0002, "P": 0.0024, "Q": 0.271, "R": 0.0085, "S": 0.0006, "T": 0.0009, "V": 0.5864, "W": 0.0004 },
    "6": { "A": 0.0126, "D": 0.0041, "E": 0.6, "G": 0.0032, "H": 0.0014, "K": 0.0009, "L": 0.0006, "N": 0.0001, "P": 0.0033, "Q": 0.3696, "R": 0.001, "S": 0.0003, "T": 0.0002, "V": 0.0024, "W": 0.0001, "Y": 0.0001 },
    "7": { "A": 0.005, "C": 0.0011, "D": 0.0001, "F": 0.0014, "G": 0.0007, "H": 0.0001, "L": 0.001, "P": 0.0052, "Q": 0.0001, "R": 0.0009, "S": 0.9113, "T": 0.0039, "V": 0.0013, "W": 0.066, "Y": 0.0019 },
    "8": { "A": 0.0031, "C": 0.0016, "D": 0.0007, "E": 0.0018, "G": 0.9836, "L": 0.0001, "R": 0.0022, "S": 0.0007, "V": 0.004, "W": 0.002 },
    "9": { "A": 0.3331, "C": 0.0002, "D": 0.001, "E": 0.0013, "F": 0.0001, "G": 0.3854, "H": 0.0002, "I": 0.0001, "L": 0.0006, "N": 0.0001, "P": 0.2545, "Q": 0.0007, "R": 0.0014, "S": 0.0128, "T": 0.0055, "V": 0.0032 },
    "10": { "A": 0.0123, "C": 0.0006, "D": 0.0068, "E": 0.154, "F": 0.0003, "G": 0.0921, "K": 0.0068, "L": 0.0032, "M": 0.0006, "N": 0.0032, "P": 0.0039, "Q": 0.0091, "R": 0.6707, "S": 0.0172, "T": 0.0094, "V": 0.0071, "W": 0.0019, "Y": 0.0006 },
    "11": { "A": 0.0183, "C": 0.0012, "D": 0.0097, "E": 0.283, "G": 0.6407, "H": 0.0003, "I": 0.0001, "K": 0.0005, "L": 0.0003, "M": 0.0001, "N": 0.001, "P": 0.0006, "Q": 0.0008, "R": 0.006, "S": 0.0015, "T": 0.023, "V": 0.0128, "W": 0.0001, "Y": 0.0001 },
    "12": { "A": 0.0028, "C": 0.0001, "E": 0.0009, "F": 0.0025, "G": 0.0037, "H": 0.0001, "I": 0.0007, "K": 0.0001, "L": 0.6033, "M": 0.0049, "P": 0.0017, "Q": 0.0012, "R": 0.0029, "S": 0.0024, "T": 0.0002, "V": 0.3686, "W": 0.0038 },
    "13": { "A": 0.0054, "C": 0.0001, "D": 0.0002, "E": 0.0024, "F": 0.0019, "G": 0.0035, "H": 0.0001, "I": 0.0149, "K": 0.273, "L": 0.0709, "M": 0.002, "N": 0.0015, "P": 0.0002, "Q": 0.0043, "R": 0.0074, "S": 0.001, "T": 0.0054, "V": 0.6048, "W": 0.0005, "Y": 0.0004 },
    "14": { "A": 0.0006, "D": 0.0002, "E": 0.0084, "G": 0.0004, "H": 0.0026, "I": 0.0002, "K": 0.6407, "L": 0.0009, "M": 0.0019, "N": 0.0055, "P": 0.0032, "Q": 0.303, "R": 0.019, "S": 0.0015, "T": 0.0117, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "15": { "A": 0.0042, "F": 0.0002, "G": 0.0001, "H": 0.0048, "I": 0.0001, "L": 0.0037, "P": 0.9704, "Q": 0.0005, "R": 0.0025, "S": 0.0076, "T": 0.0054, "V": 0.0002, "W": 0.0002, "Y": 0.0001 },
    "16": { "A": 0.0035, "C": 0.0002, "D": 0.0001, "E": 0.0014, "F": 0.0001, "G": 0.6718, "K": 0.0001, "L": 0.0009, "P": 0.002, "Q": 0.0001, "R": 0.0018, "S": 0.2776, "T": 0.0357, "V": 0.0023, "W": 0.0022, "Y": 0.0001 },
    "17": { "A": 0.1773, "C": 0.0002, "D": 0.0039, "E": 0.2517, "F": 0.0001, "G": 0.2855, "H": 0.0006, "K": 0.0024, "L": 0.0002, "M": 0.0006, "N": 0.0001, "P": 0.0021, "Q": 0.0799, "R": 0.1014, "S": 0.0832, "T": 0.006, "V": 0.0024, "W": 0.002, "Y": 0.0003 },
    "18": { "A": 0.0067, "C": 0.0012, "D": 0.0002, "E": 0.0001, "F": 0.0011, "G": 0.0001, "H": 0.0003, "I": 0.0006, "L": 0.0005, "N": 0.001, "P": 0.0101, "Q": 0.0001, "R": 0.0001, "S": 0.6627, "T": 0.3123, "V": 0.0002, "W": 0.0001, "Y": 0.0025 },
    "19": { "A": 0.0014, "C": 0.0001, "E": 0.0002, "F": 0.0003, "G": 0.0016, "H": 0.0002, "I": 0.0007, "L": 0.7245, "M": 0.0048, "P": 0.0031, "Q": 0.0007, "R": 0.0055, "S": 0.0003, "T": 0.0001, "V": 0.2563, "W": 0.0003 },
    "20": { "A": 0.0017, "C": 0.0004, "D": 0.0001, "E": 0.0012, "F": 0.0008, "G": 0.0036, "H": 0.0002, "I": 0.0022, "K": 0.2648, "L": 0.0004, "M": 0.0014, "N": 0.0028, "P": 0.0028, "Q": 0.0046, "R": 0.3736, "S": 0.2889, "T": 0.0491, "V": 0.0003, "W": 0.0001, "Y": 0.001 },
    "21": { "A": 0.0015, "C": 0.0002, "D": 0.0004, "F": 0.0024, "G": 0.0013, "H": 0.0012, "I": 0.0422, "L": 0.6819, "M": 0.0003, "N": 0.0001, "P": 0.0102, "Q": 0.0001, "R": 0.0061, "S": 0.0008, "T": 0.0008, "V": 0.2505 },
    "22": { "A": 0.0086, "C": 0.0014, "D": 0.0002, "F": 0.0015, "G": 0.0001, "H": 0.0008, "I": 0.0032, "K": 0.0002, "L": 0.0006, "M": 0.0003, "N": 0.0016, "P": 0.0201, "R": 0.0004, "S": 0.6542, "T": 0.3024, "V": 0.0002, "Y": 0.0041 },
    "23": { "A": 0.0002, "C": 0.973, "F": 0.004, "G": 0.0073, "L": 0.0007, "P": 0.0001, "R": 0.0041, "S": 0.0035, "V": 0.0006, "W": 0.0053, "Y": 0.0012 },
    "24": { "A": 0.4284, "C": 0.0001, "D": 0.0027, "E": 0.0114, "F": 0.0003, "G": 0.0055, "H": 0.0005, "I": 0.0029, "K": 0.2649, "L": 0.0012, "M": 0.0013, "N": 0.0041, "P": 0.0117, "Q": 0.0088, "R": 0.007, "S": 0.0257, "T": 0.2076, "V": 0.0157, "W": 0.0001, "Y": 0.0001 },
    "25": { "A": 0.5931, "C": 0.0006, "D": 0.0036, "E": 0.0004, "F": 0.03, "G": 0.0388, "H": 0.0002, "I": 0.0099, "L": 0.0028, "M": 0.0002, "N": 0.0001, "P": 0.0047, "Q": 0.0001, "R": 0.0002, "S": 0.0056, "T": 0.016, "V": 0.2934, "W": 0.0001, "Y": 0.0002 },
    "26": { "A": 0.0112, "C": 0.0016, "D": 0.0013, "E": 0.0001, "F": 0.0047, "G": 0.0001, "H": 0.002, "I": 0.0002, "L": 0.0012, "N": 0.0006, "P": 0.0202, "Q": 0.0002, "R": 0.0002, "S": 0.8871, "T": 0.0033, "V": 0.0004, "W": 0.0001, "Y": 0.0653 },
    "27": { "A": 0.0072, "C": 0.0016, "D": 0.0039, "E": 0.0048, "F": 0.0001, "G": 0.9636, "H": 0.0002, "I": 0.0001, "K": 0.0004, "L": 0.0002, "N": 0.0011, "P": 0.0002, "Q": 0.0003, "R": 0.0051, "S": 0.0024, "T": 0.0007, "V": 0.0074, "W": 0.0005, "Y": 0.0002 },
    "28": { "A": 0.0105, "C": 0.0076, "D": 0.0263, "E": 0.0018, "F": 0.3998, "G": 0.2995, "H": 0.0022, "I": 0.0028, "K": 0.0002, "L": 0.012, "M": 0.0001, "N": 0.0018, "P": 0.0005, "Q": 0.0001, "R": 0.002, "S": 0.0112, "T": 0.0008, "V": 0.0106, "W": 0.001, "Y": 0.2092 },
    "29": { "A": 0.01, "C": 0.0004, "D": 0.002, "E": 0.0001, "F": 0.0024, "G": 0.0016, "H": 0.0016, "I": 0.0143, "K": 0.0015, "L": 0.0016, "M": 0.0015, "N": 0.01, "P": 0.0429, "Q": 0.0002, "R": 0.0057, "S": 0.3536, "T": 0.5479, "V": 0.0012, "W": 0.0001, "Y": 0.0017 },
    "30": { "A": 0.0009, "C": 0.005, "D": 0.0002, "F": 0.6752, "G": 0.0007, "H": 0.0003, "I": 0.1805, "L": 0.0732, "M": 0.0038, "N": 0.0004, "P": 0.0016, "R": 0.0007, "S": 0.0074, "T": 0.0023, "V": 0.0461, "W": 0.0001, "Y": 0.0015 },
    "31": { "A": 0.0037, "C": 0.0017, "D": 0.0053, "E": 0.0005, "F": 0.0022, "G": 0.0117, "H": 0.0018, "I": 0.011, "K": 0.0026, "L": 0.0014, "M": 0.0005, "N": 0.0338, "P": 0.0059, "Q": 0.0002, "R": 0.0464, "S": 0.8135, "T": 0.0551, "V": 0.0009, "W": 0.0001, "Y": 0.0015 },
    "32": { "A": 0.0245, "C": 0.0025, "D": 0.0569, "E": 0.0061, "F": 0.058, "G": 0.0943, "H": 0.0068, "I": 0.0256, "K": 0.0072, "L": 0.014, "M": 0.0176, "N": 0.0537, "P": 0.0382, "Q": 0.0004, "R": 0.0281, "S": 0.3363, "T": 0.193, "V": 0.0227, "W": 0.0018, "Y": 0.0122 },
    "33": { "A": 0.0202, "D": 0.0349, "E": 0.0157, "F": 0.0292, "G": 0.0371, "H": 0.0202, "I": 0.0326, "K": 0.0281, "L": 0.0146, "M": 0.0326, "N": 0.0236, "P": 0.0371, "Q": 0.0034, "R": 0.0776, "S": 0.315, "T": 0.1912, "V": 0.0326, "W": 0.0022, "Y": 0.0517 },
    "34": { "A": 0.0049, "C": 0.0014, "D": 0.0064, "E": 0.0005, "F": 0.0034, "G": 0.0159, "H": 0.002, "I": 0.0075, "K": 0.0016, "L": 0.0015, "M": 0.0006, "N": 0.0549, "P": 0.0052, "Q": 0.0002, "R": 0.041, "S": 0.6637, "T": 0.1867, "V": 0.0015, "W": 0.0002, "Y": 0.0011 },
    "35": { "A": 0.0131, "C": 0.0021, "D": 0.0305, "E": 0.0019, "F": 0.0008, "G": 0.0675, "H": 0.0023, "I": 0.0109, "K": 0.0034, "L": 0.0014, "M": 0.0008, "N": 0.0329, "P": 0.0079, "Q": 0.0003, "R": 0.0393, "S": 0.5835, "T": 0.1961, "V": 0.0033, "W": 0.0002, "Y": 0.0018 },
    "36": { "A": 0.0143, "C": 0.0026, "D": 0.0861, "E": 0.012, "F": 0.0013, "G": 0.1758, "H": 0.0052, "I": 0.0068, "K": 0.0062, "L": 0.0018, "M": 0.0006, "N": 0.0718, "P": 0.0038, "Q": 0.0009, "R": 0.0366, "S": 0.521, "T": 0.0445, "V": 0.0037, "W": 0.0002, "Y": 0.005 },
    "37": { "A": 0.0284, "C": 0.0034, "D": 0.0116, "E": 0.0015, "F": 0.0265, "G": 0.001, "H": 0.0309, "I": 0.001, "K": 0.0014, "L": 0.0127, "M": 0.0123, "N": 0.0365, "P": 0.0019, "Q": 0.0015, "R": 0.0014, "S": 0.0366, "T": 0.003, "V": 0.0228, "W": 0.0005, "Y": 0.7652 },
    "38": { "A": 0.2351, "C": 0.0083, "D": 0.0318, "E": 0.0097, "F": 0.0127, "G": 0.1402, "H": 0.0082, "I": 0.0013, "K": 0.0004, "L": 0.0019, "M": 0.0001, "N": 0.0046, "P": 0.0051, "Q": 0.0008, "R": 0.0048, "S": 0.0779, "T": 0.0168, "V": 0.0061, "W": 0.1057, "Y": 0.3285 },
    "39": { "A": 0.0003, "C": 0.0029, "E": 0.0001, "F": 0.0061, "G": 0.0026, "H": 0.0002, "I": 0.2001, "K": 0.0005, "L": 0.0303, "M": 0.4199, "N": 0.0007, "P": 0.0002, "Q": 0.0001, "R": 0.0067, "S": 0.0024, "T": 0.0037, "V": 0.049, "W": 0.2736, "Y": 0.0005 },
    "40": { "A": 0.0129, "C": 0.0028, "D": 0.0062, "E": 0.0003, "F": 0.0006, "G": 0.13, "H": 0.2176, "I": 0.0049, "K": 0.001, "L": 0.0014, "M": 0.0004, "N": 0.119, "P": 0.0102, "Q": 0.0056, "R": 0.0171, "S": 0.4253, "T": 0.038, "V": 0.0026, "Y": 0.0041 },
    "41": { "A": 0.0001, "C": 0.0054, "F": 0.0001, "G": 0.0089, "L": 0.0044, "R": 0.0051, "S": 0.0012, "V": 0.0004, "W": 0.974, "Y": 0.0001 },
    "42": { "A": 0.0045, "C": 0.0005, "D": 0.0005, "E": 0.0003, "F": 0.0132, "G": 0.007, "H": 0.0002, "I": 0.2926, "L": 0.0171, "M": 0.0026, "N": 0.0011, "P": 0.0003, "R": 0.0003, "S": 0.0026, "T": 0.0028, "V": 0.6539, "W": 0.0002, "Y": 0.0002 },
    "43": { "A": 0.0003, "C": 0.0013, "D": 0.0001, "G": 0.0031, "H": 0.0012, "K": 0.0001, "L": 0.0027, "P": 0.003, "Q": 0.0022, "R": 0.9824, "S": 0.003, "T": 0.0001, "V": 0.0001, "W": 0.0002 },
    "44": { "A": 0.0002, "E": 0.0013, "G": 0.0001, "H": 0.006, "K": 0.0049, "L": 0.0043, "M": 0.0001, "N": 0.0001, "P": 0.016, "Q": 0.9578, "R": 0.0079, "S": 0.0002, "T": 0.0009, "W": 0.0001, "Y": 0.0001 },
    "45": { "A": 0.6164, "C": 0.0001, "D": 0.0034, "E": 0.0002, "F": 0.0007, "G": 0.0054, "H": 0.0192, "I": 0.0009, "K": 0.0004, "L": 0.0022, "M": 0.0301, "N": 0.0002, "P": 0.2678, "Q": 0.0003, "R": 0.0021, "S": 0.0283, "T": 0.0111, "V": 0.0108, "Y": 0.0004 },
    "46": { "A": 0.0243, "D": 0.0001, "G": 0.0001, "H": 0.0017, "I": 0.0001, "K": 0.0001, "L": 0.0015, "N": 0.0001, "P": 0.9288, "Q": 0.004, "R": 0.0044, "S": 0.0077, "T": 0.0267, "V": 0.0002 },
    "47": { "A": 0.002, "C": 0.0002, "D": 0.001, "E": 0.0043, "G": 0.976, "H": 0.0001, "K": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0003, "R": 0.0037, "S": 0.008, "T": 0.0003, "V": 0.0025, "W": 0.0012 },
    "48": { "A": 0.0003, "D": 0.0001, "E": 0.0082, "G": 0.0018, "H": 0.0037, "I": 0.0001, "K": 0.6992, "L": 0.0006, "M": 0.0018, "N": 0.003, "P": 0.002, "Q": 0.2491, "R": 0.0212, "S": 0.0004, "T": 0.0084, "V": 0.0001, "W": 0.0001 },
    "49": { "A": 0.0381, "C": 0.0005, "D": 0.0008, "E": 0.0022, "G": 0.9304, "K": 0.0002, "N": 0.0001, "P": 0.0006, "Q": 0.0001, "R": 0.0196, "S": 0.0015, "T": 0.0007, "V": 0.0031, "W": 0.0019 },
    "50": { "A": 0.0007, "C": 0.0001, "F": 0.0024, "G": 0.0004, "H": 0.0003, "I": 0.0011, "L": 0.9612, "M": 0.0038, "P": 0.0135, "Q": 0.0006, "R": 0.0103, "S": 0.0002, "T": 0.0002, "V": 0.0048, "W": 0.0004 },
    "51": { "A": 0.0113, "D": 0.0078, "E": 0.9243, "G": 0.0276, "H": 0.0001, "I": 0.0001, "K": 0.0016, "L": 0.0005, "M": 0.0007, "N": 0.0001, "P": 0.0001, "Q": 0.0065, "R": 0.0003, "S": 0.0003, "T": 0.0002, "V": 0.0177, "W": 0.0006, "Y": 0.0001 },
    "52": { "A": 0.0001, "C": 0.0078, "D": 0.0001, "E": 0.0001, "F": 0.0011, "G": 0.0169, "H": 0.0001, "L": 0.0072, "M": 0.0001, "R": 0.0028, "S": 0.0016, "V": 0.0016, "W": 0.9536, "Y": 0.0068 },
    "53": { "A": 0.0011, "C": 0.0005, "D": 0.0004, "E": 0.0002, "F": 0.0044, "G": 0.0066, "H": 0.0001, "I": 0.2706, "K": 0.0006, "L": 0.0627, "M": 0.2713, "N": 0.0005, "P": 0.0002, "Q": 0.0001, "R": 0.003, "S": 0.0025, "T": 0.0014, "V": 0.3736, "W": 0.0002 },
    "54": { "A": 0.1584, "C": 0.0007, "D": 0.0005, "E": 0.0021, "F": 0.0001, "G": 0.5988, "L": 0.0008, "P": 0.0012, "Q": 0.0001, "R": 0.0012, "S": 0.2282, "T": 0.0024, "V": 0.0033, "W": 0.0018, "Y": 0.0002 },
    "55": { "A": 0.0716, "C": 0.0035, "D": 0.005, "E": 0.0774, "F": 0.0197, "G": 0.1084, "H": 0.0128, "I": 0.0482, "K": 0.0025, "L": 0.0367, "M": 0.0021, "N": 0.0292, "P": 0.0009, "Q": 0.0033, "R": 0.1113, "S": 0.1132, "T": 0.0142, "V": 0.0933, "W": 0.1282, "Y": 0.1185 },
    "56": { "A": 0.0013, "C": 0.0004, "D": 0.0002, "E": 0.0001, "F": 0.0166, "G": 0.0009, "H": 0.0002, "I": 0.8692, "K": 0.0008, "L": 0.0143, "M": 0.033, "N": 0.0028, "P": 0.0002, "Q": 0.0001, "R": 0.0024, "S": 0.0119, "T": 0.0162, "V": 0.029, "W": 0.0001, "Y": 0.0004 },
    "57": { "A": 0.0028, "C": 0.0017, "D": 0.0338, "E": 0.0015, "F": 0.0171, "G": 0.0102, "H": 0.0164, "I": 0.0777, "K": 0.0459, "L": 0.0034, "M": 0.001, "N": 0.1866, "P": 0.0011, "Q": 0.002, "R": 0.0287, "S": 0.2833, "T": 0.0135, "V": 0.0055, "W": 0.0249, "Y": 0.2429 },
    "58": { "A": 0.0552, "C": 0.002, "D": 0.0101, "E": 0.0044, "F": 0.0071, "G": 0.085, "H": 0.0984, "I": 0.003, "K": 0.0025, "L": 0.0037, "M": 0.0005, "N": 0.0121, "P": 0.2222, "Q": 0.0213, "R": 0.0093, "S": 0.15, "T": 0.038, "V": 0.0077, "W": 0.0554, "Y": 0.2121 },
    "59": { "A": 0.0083, "C": 0.0015, "D": 0.1563, "E": 0.0157, "F": 0.0028, "G": 0.0748, "H": 0.0042, "I": 0.0733, "K": 0.0399, "L": 0.004, "M": 0.0035, "N": 0.1131, "P": 0.0014, "Q": 0.0012, "R": 0.0268, "S": 0.3986, "T": 0.0212, "V": 0.0053, "W": 0.0005, "Y": 0.0477 },
    "60": { "A": 0.2877, "C": 0.0007, "D": 0.012, "E": 0.0023, "F": 0.0036, "G": 0.0301, "H": 0.0017, "I": 0.0164, "K": 0.0029, "L": 0.004, "M": 0.0006, "N": 0.0096, "P": 0.0196, "Q": 0.0005, "R": 0.0088, "S": 0.2056, "T": 0.3739, "V": 0.0152, "W": 0.0003, "Y": 0.0046 },
    "61": { "A": 0.0104, "C": 0.0007, "D": 0.517, "E": 0.0281, "F": 0.006, "G": 0.0109, "H": 0.0137, "I": 0.0009, "K": 0.0088, "L": 0.0018, "M": 0.0003, "N": 0.1664, "P": 0.0004, "Q": 0.0033, "R": 0.0047, "S": 0.0167, "T": 0.0043, "V": 0.0028, "W": 0.0001, "Y": 0.2027 },
    "62": { "A": 0.0065, "C": 0.003, "D": 0.0718, "E": 0.0057, "F": 0.0882, "G": 0.4272, "H": 0.0019, "I": 0.0033, "K": 0.0128, "L": 0.0286, "M": 0.0004, "N": 0.0942, "P": 0.0009, "Q": 0.0006, "R": 0.0086, "S": 0.2174, "T": 0.0199, "V": 0.006, "W": 0.0005, "Y": 0.0024 },
    "63": { "A": 0.0086, "C": 0.0023, "D": 0.0496, "E": 0.0065, "F": 0.001, "G": 0.66, "H": 0.0015, "I": 0.0031, "K": 0.0015, "L": 0.0012, "M": 0.0003, "N": 0.0096, "P": 0.0009, "Q": 0.0005, "R": 0.0111, "S": 0.2049, "T": 0.012, "V": 0.0065, "W": 0.0103, "Y": 0.0086 },
    "64": { "A": 0.0169, "C": 0.001, "D": 0.0767, "E": 0.0432, "F": 0.0029, "G": 0.0512, "H": 0.0055, "I": 0.025, "K": 0.014, "L": 0.0022, "M": 0.001, "N": 0.17, "P": 0.0022, "Q": 0.0026, "R": 0.0204, "S": 0.3536, "T": 0.1481, "V": 0.0066, "W": 0.0003, "Y": 0.0566 },
    "65": { "A": 0.082, "D": 0.0008, "E": 0.0054, "F": 0.0004, "G": 0.0015, "H": 0.0007, "I": 0.0981, "K": 0.1312, "L": 0.0022, "M": 0.0019, "N": 0.0103, "P": 0.0193, "Q": 0.0033, "R": 0.0078, "S": 0.0105, "T": 0.6178, "V": 0.0056, "Y": 0.0009 },
    "66": { "A": 0.0084, "C": 0.0012, "D": 0.0463, "E": 0.0141, "F": 0.0153, "G": 0.045, "H": 0.0182, "I": 0.0146, "K": 0.0302, "L": 0.0028, "M": 0.001, "N": 0.3054, "P": 0.0012, "Q": 0.0035, "R": 0.0513, "S": 0.0445, "T": 0.0207, "V": 0.0036, "W": 0.0008, "Y": 0.3721 },
    "67": { "A": 0.0005, "C": 0.003, "D": 0.0112, "E": 0.0006, "F": 0.0103, "G": 0.0001, "H": 0.0098, "I": 0.0011, "K": 0.0004, "L": 0.0021, "N": 0.0044, "P": 0.0006, "Q": 0.0008, "R": 0.0004, "S": 0.0143, "T": 0.001, "V": 0.0008, "W": 0.0001, "Y": 0.9385 },
    "68": { "A": 0.5752, "C": 0.0002, "D": 0.002, "E": 0.0058, "F": 0.0001, "G": 0.0089, "H": 0.0046, "I": 0.0021, "K": 0.0035, "L": 0.002, "M": 0.0005, "N": 0.2511, "P": 0.0063, "Q": 0.0006, "R": 0.0035, "S": 0.0906, "T": 0.0136, "V": 0.0287, "Y": 0.0005 },
    "69": { "A": 0.0419, "C": 0.0001, "D": 0.3274, "E": 0.0128, "F": 0.0002, "G": 0.0068, "H": 0.0028, "I": 0.0003, "K": 0.0033, "L": 0.0024, "M": 0.0001, "N": 0.0028, "P": 0.327, "Q": 0.2398, "R": 0.0045, "S": 0.0043, "T": 0.0143, "V": 0.0078, "W": 0.0001, "Y": 0.0012 },
    "70": { "A": 0.0153, "C": 0.0007, "D": 0.0013, "E": 0.0021, "F": 0.0029, "G": 0.0056, "H": 0.0007, "I": 0.0005, "K": 0.2252, "L": 0.0005, "M": 0.0008, "N": 0.0099, "P": 0.0276, "Q": 0.0019, "R": 0.0081, "S": 0.6873, "T": 0.0062, "V": 0.0005, "Y": 0.0028 },
    "71": { "A": 0.002, "C": 0.0038, "E": 0.0004, "F": 0.2417, "G": 0.0047, "H": 0.0009, "I": 0.0025, "L": 0.344, "M": 0.0024, "P": 0.0028, "Q": 0.0001, "R": 0.0043, "S": 0.0012, "T": 0.0002, "V": 0.3876, "W": 0.0003, "Y": 0.001 },
    "72": { "A": 0.0017, "D": 0.0005, "E": 0.0156, "G": 0.0014, "H": 0.0025, "I": 0.0003, "K": 0.6422, "L": 0.0016, "M": 0.0023, "N": 0.006, "P": 0.0013, "Q": 0.2796, "R": 0.0322, "S": 0.0013, "T": 0.0103, "V": 0.0004, "W": 0.0004, "Y": 0.0002 },
    "73": { "A": 0.0117, "C": 0.0013, "D": 0.0078, "E": 0.2289, "F": 0.0013, "G": 0.1391, "H": 0.0065, "I": 0.0039, "K": 0.2575, "L": 0.013, "M": 0.0195, "N": 0.0312, "P": 0.0091, "Q": 0.0429, "R": 0.1053, "S": 0.0273, "T": 0.0429, "V": 0.0091, "W": 0.0403, "Y": 0.0013 },
    "74": { "A": 0.0021, "C": 0.0043, "D": 0.0132, "E": 0.003, "F": 0.0001, "G": 0.6624, "H": 0.0004, "I": 0.0015, "K": 0.0002, "L": 0.0001, "M": 0.0002, "N": 0.0041, "P": 0.0003, "Q": 0.0001, "R": 0.0083, "S": 0.2876, "T": 0.0091, "V": 0.0029, "W": 0.0001, "Y": 0.0001 },
    "75": { "A": 0.0001, "C": 0.0001, "E": 0.0001, "G": 0.0027, "H": 0.0069, "I": 0.0007, "K": 0.0014, "L": 0.0034, "M": 0.0003, "N": 0.0002, "P": 0.0012, "Q": 0.027, "R": 0.9481, "S": 0.0018, "T": 0.0006, "V": 0.0002, "W": 0.005 },
    "76": { "A": 0.0057, "C": 0.002, "D": 0.0007, "F": 0.3846, "G": 0.0051, "H": 0.0001, "I": 0.0174, "L": 0.0555, "M": 0.0004, "P": 0.0004, "R": 0.0004, "S": 0.0012, "T": 0.0004, "V": 0.5254, "Y": 0.0005 },
    "77": { "A": 0.0098, "D": 0.0004, "F": 0.0009, "G": 0.0003, "H": 0.0005, "I": 0.0119, "K": 0.0007, "L": 0.0007, "M": 0.0005, "N": 0.0043, "P": 0.0043, "Q": 0.0001, "R": 0.0011, "S": 0.0222, "T": 0.9349, "V": 0.0071, "Y": 0.0002 },
    "78": { "A": 0.0004, "C": 0.0002, "F": 0.0098, "G": 0.0002, "H": 0.0001, "I": 0.7683, "K": 0.0005, "L": 0.0186, "M": 0.1662, "N": 0.0011, "P": 0.0001, "Q": 0.0001, "R": 0.0033, "S": 0.005, "T": 0.0044, "V": 0.0213, "W": 0.0001, "Y": 0.0001 },
    "79": { "A": 0.011, "C": 0.0003, "D": 0.0004, "E": 0.0001, "F": 0.0025, "G": 0.0002, "H": 0.0002, "I": 0.0017, "K": 0.0003, "L": 0.0023, "M": 0.0001, "N": 0.0084, "P": 0.0035, "Q": 0.0001, "R": 0.0003, "S": 0.6974, "T": 0.2688, "V": 0.0005, "W": 0.0001, "Y": 0.0018 },
    "80": { "A": 0.1175, "C": 0.0001, "D": 0.0004, "E": 0.0117, "F": 0.0004, "G": 0.0062, "H": 0.0001, "I": 0.0136, "K": 0.0372, "L": 0.0225, "M": 0.0039, "N": 0.0003, "P": 0.0078, "Q": 0.0007, "R": 0.4834, "S": 0.0061, "T": 0.0537, "V": 0.2332, "W": 0.001 },
    "81": { "A": 0.0042, "C": 0.0001, "D": 0.9491, "E": 0.0132, "G": 0.0034, "H": 0.0024, "I": 0.0001, "K": 0.0003, "L": 0.0001, "N": 0.0216, "Q": 0.0002, "R": 0.0001, "S": 0.0007, "T": 0.0004, "V": 0.0015, "Y": 0.0026 },
    "82": { "A": 0.0065, "D": 0.0415, "E": 0.0507, "F": 0.0001, "G": 0.001, "H": 0.0009, "I": 0.0035, "K": 0.0767, "L": 0.0007, "M": 0.0073, "N": 0.3326, "P": 0.0051, "Q": 0.0009, "R": 0.0108, "S": 0.0076, "T": 0.4521, "V": 0.0011, "W": 0.0001, "Y": 0.0006 },
    "83": { "A": 0.157, "C": 0.0005, "D": 0.0015, "E": 0.0001, "F": 0.0038, "G": 0.0016, "H": 0.0003, "I": 0.0004, "L": 0.0011, "N": 0.0003, "P": 0.0078, "Q": 0.0001, "R": 0.0003, "S": 0.8141, "T": 0.0062, "V": 0.0017, "W": 0.0002, "Y": 0.003 },
    "84": { "A": 0.016, "D": 0.0009, "E": 0.0087, "F": 0.0003, "G": 0.0011, "H": 0.0004, "I": 0.092, "K": 0.6505, "L": 0.0017, "M": 0.0033, "N": 0.0122, "P": 0.0007, "Q": 0.0067, "R": 0.0193, "S": 0.0038, "T": 0.1744, "V": 0.0075, "W": 0.0001, "Y": 0.0002 },
    "85": { "A": 0.002, "C": 0.0002, "D": 0.02, "E": 0.0009, "F": 0.0001, "G": 0.0056, "H": 0.0025, "I": 0.0021, "K": 0.0137, "L": 0.0001, "M": 0.0003, "N": 0.6577, "P": 0.0002, "Q": 0.0005, "R": 0.0086, "S": 0.264, "T": 0.0198, "V": 0.0002, "Y": 0.0013 },
    "86": { "A": 0.0049, "D": 0.0002, "E": 0.0028, "F": 0.0005, "G": 0.0001, "H": 0.008, "I": 0.0112, "K": 0.0032, "L": 0.0035, "M": 0.0046, "N": 0.0007, "P": 0.0037, "Q": 0.2991, "R": 0.004, "S": 0.1386, "T": 0.5129, "V": 0.0012, "W": 0.0001, "Y": 0.0004 },
    "87": { "A": 0.2623, "C": 0.0031, "D": 0.0009, "E": 0.0001, "F": 0.2632, "G": 0.0025, "H": 0.0002, "I": 0.0045, "L": 0.3548, "M": 0.0033, "P": 0.0006, "Q": 0.0004, "R": 0.0018, "S": 0.0028, "T": 0.0028, "V": 0.0957, "W": 0.0002, "Y": 0.0008 },
    "88": { "A": 0.0038, "C": 0.0013, "D": 0.0053, "E": 0.0002, "F": 0.0257, "G": 0.0002, "H": 0.0061, "I": 0.0007, "K": 0.0001, "L": 0.001, "N": 0.0018, "P": 0.0014, "Q": 0.0002, "R": 0.0001, "S": 0.2796, "T": 0.0012, "V": 0.0352, "W": 0.0003, "Y": 0.6357 },
    "89": { "A": 0.0001, "F": 0.0005, "H": 0.0001, "I": 0.0016, "K": 0.0002, "L": 0.7389, "M": 0.2441, "P": 0.0014, "Q": 0.0005, "R": 0.0078, "S": 0.0002, "T": 0.0003, "V": 0.0041, "W": 0.0002 },
    "90": { "A": 0.0014, "D": 0.0098, "E": 0.249, "F": 0.0001, "G": 0.0039, "H": 0.0067, "I": 0.0016, "K": 0.2287, "L": 0.0012, "M": 0.0012, "N": 0.0136, "P": 0.0005, "Q": 0.4171, "R": 0.0216, "S": 0.0028, "T": 0.039, "V": 0.001, "W": 0.0003, "Y": 0.0005 },
    "91": { "A": 0.0001, "C": 0.0001, "E": 0.0001, "F": 0.0006, "G": 0.0005, "I": 0.011, "K": 0.0003, "L": 0.5158, "M": 0.4128, "N": 0.0001, "P": 0.0008, "Q": 0.0003, "R": 0.0066, "S": 0.0002, "T": 0.0006, "V": 0.0169, "W": 0.0332 },
    "92": { "A": 0.0025, "C": 0.0006, "D": 0.012, "E": 0.0008, "F": 0.0005, "G": 0.01, "H": 0.0034, "I": 0.0037, "K": 0.007, "L": 0.0004, "M": 0.0006, "N": 0.3772, "P": 0.0007, "Q": 0.0003, "R": 0.0683, "S": 0.4324, "T": 0.0773, "V": 0.0005, "W": 0.0003, "Y": 0.0015 },
    "93": { "A": 0.0049, "C": 0.0008, "D": 0.003, "E": 0.0001, "F": 0.0018, "G": 0.0137, "H": 0.0004, "I": 0.0013, "K": 0.0006, "L": 0.0004, "M": 0.0002, "N": 0.0462, "P": 0.0014, "Q": 0.0001, "R": 0.0449, "S": 0.8728, "T": 0.0054, "V": 0.0003, "W": 0.0003, "Y": 0.0014 },
    "94": { "A": 0.0028, "C": 0.0001, "E": 0.0002, "F": 0.0001, "G": 0.003, "I": 0.0005, "L": 0.6737, "M": 0.0381, "P": 0.0015, "Q": 0.0004, "R": 0.0042, "S": 0.0001, "T": 0.0003, "V": 0.2746, "W": 0.0003 },
    "95": { "A": 0.0026, "D": 0.0348, "E": 0.0022, "G": 0.0086, "H": 0.0002, "I": 0.0032, "K": 0.0783, "L": 0.0002, "M": 0.0004, "N": 0.0019, "P": 0.0011, "Q": 0.0016, "R": 0.5562, "S": 0.0081, "T": 0.3001, "V": 0.0004, "W": 0.0001, "Y": 0.0001 },
    "96": { "A": 0.5993, "C": 0.0003, "D": 0.0126, "E": 0.0005, "F": 0.004, "G": 0.0052, "H": 0.0005, "I": 0.0027, "K": 0.0001, "L": 0.0016, "M": 0.0001, "N": 0.0005, "P": 0.0569, "Q": 0.0001, "R": 0.0007, "S": 0.245, "T": 0.0434, "V": 0.0245, "Y": 0.0021 },
    "97": { "A": 0.2731, "D": 0.1215, "E": 0.5218, "G": 0.0079, "H": 0.0002, "I": 0.0001, "K": 0.0006, "L": 0.0006, "M": 0.0005, "N": 0.0007, "P": 0.0005, "Q": 0.0007, "R": 0.0002, "S": 0.0344, "T": 0.0026, "V": 0.0344, "W": 0.0001, "Y": 0.0003 },
    "98": { "A": 0.0017, "D": 0.9857, "E": 0.0069, "G": 0.002, "H": 0.0004, "N": 0.0008, "S": 0.0001, "V": 0.0006, "Y": 0.0017 },
    "99": { "A": 0.007, "G": 0.0002, "H": 0.0001, "I": 0.0002, "K": 0.001, "L": 0.0001, "M": 0.0044, "N": 0.0002, "P": 0.0021, "Q": 0.0001, "R": 0.001, "S": 0.0097, "T": 0.974, "V": 0.0001 },
    "100": { "A": 0.9791, "C": 0.0001, "D": 0.0015, "G": 0.013, "P": 0.0003, "R": 0.0001, "S": 0.0029, "T": 0.0016, "V": 0.0013 },
    "101": { "A": 0.0022, "C": 0.0001, "D": 0.0002, "E": 0.0008, "F": 0.0046, "G": 0.0037, "H": 0.0002, "I": 0.0306, "K": 0.0007, "L": 0.0471, "M": 0.0433, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0011, "S": 0.0004, "T": 0.0415, "V": 0.8229, "W": 0.0002, "Y": 0.0001 },
    "102": { "C": 0.0007, "D": 0.0038, "E": 0.0002, "F": 0.0052, "H": 0.0011, "I": 0.0001, "L": 0.0002, "N": 0.0004, "S": 0.0004, "V": 0.0001, "Y": 0.9877 },
    "103": { "C": 0.0015, "D": 0.0032, "F": 0.0387, "H": 0.0052, "I": 0.0001, "K": 0.0001, "L": 0.0004, "N": 0.0003, "Q": 0.0001, "R": 0.0001, "S": 0.0025, "T": 0.0001, "V": 0.0002, "W": 0.0003, "Y": 0.9469 },
    "104": { "A": 0.0001, "C": 0.9851, "F": 0.0011, "G": 0.0048, "L": 0.0002, "R": 0.001, "S": 0.0007, "V": 0.0003, "W": 0.0059, "Y": 0.0007 },
    "105": { "A": 0.9052, "C": 0.0005, "D": 0.0002, "E": 0.0015, "F": 0.0003, "G": 0.007, "H": 0.0005, "I": 0.0011, "K": 0.0002, "L": 0.0014, "M": 0.0009, "N": 0.0005, "P": 0.0003, "Q": 0.0002, "R": 0.0009, "S": 0.0063, "T": 0.0411, "V": 0.0314, "W": 0.0003, "Y": 0.0004 },
    "106": { "A": 0.0097, "C": 0.0012, "D": 0.0008, "E": 0.0016, "F": 0.001, "G": 0.0178, "H": 0.0202, "I": 0.0074, "K": 0.1316, "L": 0.0057, "M": 0.0018, "N": 0.0063, "P": 0.0021, "Q": 0.0024, "R": 0.6791, "S": 0.0503, "T": 0.0537, "V": 0.0046, "W": 0.0008, "Y": 0.0018 },
    "107": { "A": 0.0718, "C": 0.0045, "D": 0.2388, "E": 0.086, "F": 0.0105, "G": 0.1919, "H": 0.0338, "I": 0.019, "K": 0.0083, "L": 0.0417, "M": 0.0073, "N": 0.0121, "P": 0.0232, "Q": 0.0215, "R": 0.0498, "S": 0.0557, "T": 0.0279, "V": 0.0724, "W": 0.0087, "Y": 0.015 },
    "108": { "A": 0.0527, "C": 0.006, "D": 0.0424, "E": 0.0328, "F": 0.0239, "G": 0.1558, "H": 0.0235, "I": 0.0355, "K": 0.0243, "L": 0.0821, "M": 0.0133, "N": 0.024, "P": 0.0858, "Q": 0.0294, "R": 0.1221, "S": 0.085, "T": 0.0421, "V": 0.0519, "W": 0.0197, "Y": 0.0477 },
    "109": { "A": 0.059, "C": 0.014, "D": 0.0579, "E": 0.0333, "F": 0.0223, "G": 0.1725, "H": 0.0176, "I": 0.0361, "K": 0.0164, "L": 0.0542, "M": 0.012, "N": 0.024, "P": 0.0534, "Q": 0.0197, "R": 0.0812, "S": 0.0998, "T": 0.0516, "V": 0.0615, "W": 0.0244, "Y": 0.0892 },
    "110": { "A": 0.0646, "C": 0.0148, "D": 0.0673, "E": 0.0289, "F": 0.0249, "G": 0.1545, "H": 0.0133, "I": 0.0381, "K": 0.0114, "L": 0.0467, "M": 0.0123, "N": 0.023, "P": 0.0428, "Q": 0.0187, "R": 0.0574, "S": 0.1218, "T": 0.0523, "V": 0.065, "W": 0.027, "Y": 0.1153 },
    "111": { "A": 0.0692, "C": 0.0201, "D": 0.0648, "E": 0.0211, "F": 0.0296, "G": 0.1486, "H": 0.0107, "I": 0.0389, "K": 0.0085, "L": 0.0443, "M": 0.0137, "N": 0.022, "P": 0.0282, "Q": 0.0152, "R": 0.0444, "S": 0.1409, "T": 0.0548, "V": 0.067, "W": 0.0319, "Y": 0.1262 },
    "111A": { "A": 0.0704, "C": 0.0197, "D": 0.0711, "E": 0.0198, "F": 0.0323, "G": 0.1452, "H": 0.0094, "I": 0.0328, "K": 0.0072, "L": 0.0454, "M": 0.0127, "N": 0.0212, "P": 0.0279, "Q": 0.0126, "R": 0.0389, "S": 0.1509, "T": 0.0584, "V": 0.0721, "W": 0.0407, "Y": 0.1113 },
    "111B": { "A": 0.0576, "C": 0.0192, "D": 0.0631, "E": 0.0187, "F": 0.0376, "G": 0.148, "H": 0.0077, "I": 0.0322, "K": 0.0068, "L": 0.0416, "M": 0.0111, "N": 0.0217, "P": 0.0313, "Q": 0.0116, "R": 0.0384, "S": 0.1745, "T": 0.06, "V": 0.0769, "W": 0.044, "Y": 0.0979 },
    "111C": { "A": 0.0533, "C": 0.0297, "D": 0.0548, "E": 0.0172, "F": 0.0354, "G": 0.1687, "H": 0.0079, "I": 0.03, "K": 0.0076, "L": 0.0427, "M": 0.0097, "N": 0.0181, "P": 0.0322, "Q": 0.0088, "R": 0.0357, "S": 0.1817, "T": 0.0621, "V": 0.0685, "W": 0.0477, "Y": 0.0881 },
    "111D": { "A": 0.0521, "C": 0.0392, "D": 0.0448, "E": 0.0192, "F": 0.03, "G": 0.1637, "H": 0.0068, "I": 0.0291, "K": 0.0078, "L": 0.0399, "M": 0.0084, "N": 0.0161, "P": 0.0457, "Q": 0.0097, "R": 0.0329, "S": 0.1841, "T": 0.0611, "V": 0.0674, "W": 0.0415, "Y": 0.1006 },
    "111E": { "A": 0.0575, "C": 0.059, "D": 0.0374, "E": 0.0163, "F": 0.0292, "G": 0.1501, "H": 0.0091, "I": 0.0314, "K": 0.0078, "L": 0.0429, "M": 0.0069, "N": 0.0176, "P": 0.0437, "Q": 0.0099, "R": 0.0357, "S": 0.1635, "T": 0.0509, "V": 0.0757, "W": 0.0385, "Y": 0.1171 },
    "111F": { "A": 0.0592, "C": 0.0543, "D": 0.0357, "E": 0.0208, "F": 0.0272, "G": 0.1341, "H": 0.009, "I": 0.0299, "K": 0.012, "L": 0.046, "M": 0.0069, "N": 0.0143, "P": 0.0475, "Q": 0.0106, "R": 0.0453, "S": 0.1544, "T": 0.0534, "V": 0.0553, "W": 0.0319, "Y": 0.1522 },
    "111G": { "A": 0.046, "C": 0.055, "D": 0.0386, "E": 0.0229, "F": 0.026, "G": 0.128, "H": 0.0115, "I": 0.0315, "K": 0.0148, "L": 0.0464, "M": 0.0124, "N": 0.0178, "P": 0.0588, "Q": 0.013, "R": 0.0507, "S": 0.1501, "T": 0.0473, "V": 0.0509, "W": 0.0304, "Y": 0.1479 },
    "111H": { "A": 0.0466, "C": 0.0438, "D": 0.0534, "E": 0.0135, "F": 0.0758, "G": 0.1412, "H": 0.0114, "I": 0.0288, "K": 0.0199, "L": 0.0377, "M": 0.0082, "N": 0.0203, "P": 0.0822, "Q": 0.0103, "R": 0.0484, "S": 0.1185, "T": 0.0566, "V": 0.0573, "W": 0.0217, "Y": 0.1046 },
    "111I": { "A": 0.0445, "C": 0.0395, "D": 0.2267, "E": 0.0213, "F": 0.0273, "G": 0.0951, "H": 0.0121, "I": 0.0202, "K": 0.0324, "L": 0.0425, "M": 0.0061, "N": 0.0162, "P": 0.0577, "Q": 0.0081, "R": 0.0729, "S": 0.082, "T": 0.0476, "V": 0.0435, "W": 0.0192, "Y": 0.085 },
    "111J": { "A": 0.0359, "C": 0.0179, "D": 0.0154, "E": 0.0487, "F": 0.0256, "G": 0.0744, "H": 0.0103, "I": 0.0077, "K": 0.0051, "L": 0.0205, "M": 0.0051, "N": 0.0256, "P": 0.0487, "Q": 0.0205, "R": 0.3949, "S": 0.0538, "T": 0.0282, "V": 0.0333, "W": 0.0308, "Y": 0.0974 },
    "112K": { "A": 0.0231, "D": 0.0846, "E": 0.0538, "F": 0.0154, "G": 0.2231, "K": 0.0692, "L": 0.0308, "M": 0.0154, "N": 0.0154, "P": 0.0538, "Q": 0.0231, "R": 0.0615, "S": 0.0923, "T": 0.1077, "V": 0.0538, "W": 0.0077, "Y": 0.0692 },
    "112J": { "A": 0.0296, "C": 0.0132, "D": 0.0214, "E": 0.023, "F": 0.0362, "G": 0.1036, "H": 0.0132, "I": 0.0362, "K": 0.0214, "L": 0.0625, "M": 0.0049, "N": 0.0115, "P": 0.051, "Q": 0.028, "R": 0.0691, "S": 0.0855, "T": 0.023, "V": 0.2385, "W": 0.0378, "Y": 0.0905 },
    "112I": { "A": 0.0534, "C": 0.0212, "D": 0.0296, "E": 0.0244, "F": 0.0283, "G": 0.1068, "H": 0.0186, "I": 0.0244, "K": 0.0148, "L": 0.0617, "M": 0.0913, "N": 0.0341, "P": 0.0643, "Q": 0.0251, "R": 0.063, "S": 0.11, "T": 0.0605, "V": 0.0637, "W": 0.0283, "Y": 0.0765 },
    "112H": { "A": 0.0607, "C": 0.0385, "D": 0.0383, "E": 0.0214, "F": 0.0311, "G": 0.1272, "H": 0.0149, "I": 0.0277, "K": 0.0136, "L": 0.0462, "M": 0.0116, "N": 0.0539, "P": 0.0711, "Q": 0.0177, "R": 0.066, "S": 0.1224, "T": 0.0525, "V": 0.0568, "W": 0.0295, "Y": 0.0989 },
    "112G": { "A": 0.0522, "C": 0.0335, "D": 0.0299, "E": 0.0219, "F": 0.0248, "G": 0.1364, "H": 0.0107, "I": 0.0308, "K": 0.0188, "L": 0.0539, "M": 0.0078, "N": 0.0204, "P": 0.0777, "Q": 0.0227, "R": 0.0549, "S": 0.1288, "T": 0.0497, "V": 0.0597, "W": 0.0317, "Y": 0.1335 },
    "112F": { "A": 0.0551, "C": 0.04, "D": 0.0316, "E": 0.0237, "F": 0.0233, "G": 0.1261, "H": 0.0118, "I": 0.0318, "K": 0.0184, "L": 0.0483, "M": 0.0092, "N": 0.0188, "P": 0.0698, "Q": 0.0148, "R": 0.0597, "S": 0.1342, "T": 0.056, "V": 0.0581, "W": 0.028, "Y": 0.1412 },
    "112E": { "A": 0.0601, "C": 0.0383, "D": 0.0326, "E": 0.0261, "F": 0.0276, "G": 0.142, "H": 0.0124, "I": 0.0305, "K": 0.0185, "L": 0.044, "M": 0.0105, "N": 0.0178, "P": 0.0741, "Q": 0.0168, "R": 0.0593, "S": 0.1355, "T": 0.055, "V": 0.0507, "W": 0.0285, "Y": 0.1196 },
    "112D": { "A": 0.0522, "C": 0.0288, "D": 0.0444, "E": 0.0235, "F": 0.0256, "G": 0.1589, "H": 0.0162, "I": 0.0276, "K": 0.014, "L": 0.0456, "M": 0.0078, "N": 0.0259, "P": 0.0596, "Q": 0.0141, "R": 0.0532, "S": 0.1397, "T": 0.0564, "V": 0.0558, "W": 0.0285, "Y": 0.1223 },
    "112C": { "A": 0.0568, "C": 0.0269, "D": 0.0498, "E": 0.0212, "F": 0.0239, "G": 0.144, "H": 0.0163, "I": 0.0267, "K": 0.012, "L": 0.0434, "M": 0.0069, "N": 0.0277, "P": 0.051, "Q": 0.0118, "R": 0.0484, "S": 0.1387, "T": 0.0551, "V": 0.0537, "W": 0.0292, "Y": 0.1566 },
    "112B": { "A": 0.0532, "C": 0.0252, "D": 0.0486, "E": 0.0203, "F": 0.0226, "G": 0.1382, "H": 0.0162, "I": 0.0234, "K": 0.0113, "L": 0.0409, "M": 0.0073, "N": 0.0247, "P": 0.0443, "Q": 0.0123, "R": 0.0445, "S": 0.1178, "T": 0.05, "V": 0.0481, "W": 0.0301, "Y": 0.221 },
    "112A": { "A": 0.0571, "C": 0.0158, "D": 0.0396, "E": 0.0218, "F": 0.0241, "G": 0.1347, "H": 0.0152, "I": 0.0219, "K": 0.0121, "L": 0.0424, "M": 0.0076, "N": 0.0231, "P": 0.0506, "Q": 0.014, "R": 0.0506, "S": 0.1116, "T": 0.0551, "V": 0.0473, "W": 0.0319, "Y": 0.2233 },
    "112": { "A": 0.053, "C": 0.0123, "D": 0.0432, "E": 0.0263, "F": 0.0244, "G": 0.1285, "H": 0.0227, "I": 0.02, "K": 0.0134, "L": 0.0536, "M": 0.0083, "N": 0.0307, "P": 0.0551, "Q": 0.015, "R": 0.048, "S": 0.1033, "T": 0.0519, "V": 0.04, "W": 0.0326, "Y": 0.218 },
    "113": { "A": 0.0404, "C": 0.0103, "D": 0.0931, "E": 0.0229, "F": 0.0202, "G": 0.1557, "H": 0.021, "I": 0.0105, "K": 0.0098, "L": 0.0292, "M": 0.0051, "N": 0.0684, "P": 0.0284, "Q": 0.0094, "R": 0.044, "S": 0.0752, "T": 0.0354, "V": 0.0244, "W": 0.0471, "Y": 0.2494 },
    "114": { "A": 0.1499, "C": 0.0091, "D": 0.0242, "E": 0.0145, "F": 0.0208, "G": 0.1979, "H": 0.0188, "I": 0.0114, "K": 0.0052, "L": 0.0295, "M": 0.0035, "N": 0.016, "P": 0.0574, "Q": 0.0076, "R": 0.022, "S": 0.0466, "T": 0.0289, "V": 0.0248, "W": 0.1006, "Y": 0.2113 },
    "115": { "A": 0.0128, "C": 0.0053, "D": 0.0066, "E": 0.0063, "F": 0.525, "G": 0.0273, "H": 0.0049, "I": 0.029, "K": 0.0019, "L": 0.0804, "M": 0.1636, "N": 0.0068, "P": 0.0233, "Q": 0.003, "R": 0.0068, "S": 0.0213, "T": 0.0108, "V": 0.0276, "W": 0.009, "Y": 0.0283 },
    "116": { "A": 0.0119, "C": 0.0008, "D": 0.8441, "E": 0.0149, "F": 0.0037, "G": 0.0353, "H": 0.0074, "I": 0.0028, "K": 0.0017, "L": 0.0059, "M": 0.0007, "N": 0.0106, "P": 0.0058, "Q": 0.0145, "R": 0.0057, "S": 0.0099, "T": 0.0055, "V": 0.0087, "W": 0.0021, "Y": 0.0081 },
    "117": { "A": 0.0045, "C": 0.0039, "D": 0.0124, "E": 0.0014, "F": 0.0217, "G": 0.0046, "H": 0.0272, "I": 0.1148, "K": 0.0011, "L": 0.04, "M": 0.0023, "N": 0.0103, "P": 0.1114, "Q": 0.0026, "R": 0.0032, "S": 0.0301, "T": 0.0059, "V": 0.1863, "W": 0.001, "Y": 0.4153 },
    "118": { "A": 0.0002, "C": 0.0015, "E": 0.0001, "F": 0.0005, "G": 0.003, "I": 0.0001, "L": 0.0022, "M": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0017, "S": 0.0005, "T": 0.0002, "V": 0.0009, "W": 0.9886, "Y": 0.0001 },
    "119": { "A": 0.0015, "C": 0.0017, "D": 0.0017, "G": 0.9909, "N": 0.0001, "P": 0.0001, "R": 0.0003, "S": 0.0025, "T": 0.0001, "V": 0.001 },
    "120": { "A": 0.0001, "D": 0.0001, "E": 0.0014, "G": 0.0002, "H": 0.0046, "I": 0.0001, "K": 0.0369, "L": 0.004, "N": 0.0005, "P": 0.0089, "Q": 0.9002, "R": 0.0423, "S": 0.0001, "T": 0.0005 },
    "121": { "A": 0.0003, "C": 0.0001, "D": 0.0001, "E": 0.0008, "G": 0.9962, "R": 0.001, "S": 0.0001, "V": 0.0013, "W": 0.0002 },
    "122": { "A": 0.0145, "D": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0132, "K": 0.0006, "L": 0.0005, "M": 0.0007, "N": 0.0019, "P": 0.0015, "Q": 0.0001, "R": 0.0005, "S": 0.0083, "T": 0.9559, "V": 0.0019 },
    "123": { "A": 0.0025, "E": 0.0002, "F": 0.0002, "G": 0.0001, "H": 0.0003, "I": 0.0006, "K": 0.0018, "L": 0.6723, "M": 0.1187, "N": 0.0001, "P": 0.0077, "Q": 0.0071, "R": 0.006, "S": 0.0017, "T": 0.175, "V": 0.0055, "W": 0.0002 },
    "124": { "A": 0.0021, "D": 0.0004, "F": 0.0008, "G": 0.0022, "I": 0.0048, "L": 0.0028, "S": 0.0001, "V": 0.9867 },
    "125": { "A": 0.0066, "D": 0.0001, "F": 0.0002, "G": 0.0001, "H": 0.0005, "I": 0.015, "K": 0.0001, "L": 0.0003, "M": 0.0001, "N": 0.0016, "P": 0.0014, "R": 0.0002, "S": 0.0105, "T": 0.962, "V": 0.0012 },
    "126": { "A": 0.0025, "D": 0.0004, "F": 0.0007, "G": 0.0017, "I": 0.0028, "L": 0.0004, "R": 0.0006, "S": 0.0003, "V": 0.9904 },
    "127": { "A": 0.0044, "C": 0.0003, "D": 0.0001, "F": 0.0017, "H": 0.0001, "L": 0.0013, "P": 0.0024, "S": 0.9864, "T": 0.0017, "V": 0.0001, "Y": 0.0016 },
    "128": { "A": 0.005, "F": 0.0001, "L": 0.001, "P": 0.0061, "Q": 0.0004, "S": 0.9859, "T": 0.0014 }
  },
  "L": {
    "1": { "H": 0.0072, "K": 0.0002, "N": 0.0185, "Q": 0.6686, "S": 0.3054 },
    "2": { "A": 0.0264, "C": 0.0006, "D": 0.0015, "E": 0.0001, "F": 0.0231, "G": 0.0001, "H": 0.0021, "I": 0.0003, "L": 0.0169, "N": 0.0007, "P": 0.0158, "R": 0.0001, "S": 0.6565, "T": 0.0193, "V": 0.0009, "Y": 0.2356 },
    "3": { "A": 0.3425, "D": 0.0089, "E": 0.2169, "F": 0.0004, "G": 0.0045, "H": 0.0002, "I": 0.0022, "K": 0.0008, "L": 0.0059, "M": 0.0187, "N": 0.0001, "P": 0.002, "Q": 0.0016, "R": 0.0001, "S": 0.0016, "T": 0.0012, "V": 0.3923, "W": 0.0001, "Y": 0.0002 },
    "4": { "A": 0.0003, "F": 0.0013, "G": 0.0002, "I": 0.0001, "L": 0.9463, "M": 0.0031, "P": 0.0043, "Q": 0.0005, "R": 0.0031, "S": 0.0003, "T": 0.0001, "V": 0.0402, "W": 0.0001 },
    "5": { "A": 0.0048, "F": 0.0001, "H": 0.0004, "I": 0.0029, "K": 0.0005, "L": 0.0002, "M": 0.0002, "N": 0.0021, "P": 0.0295, "Q": 0.0001, "R": 0.0004, "S": 0.0054, "T": 0.9533, "V": 0.0001 },
    "6": { "A": 0.0001, "E": 0.0004, "H": 0.0131, "K": 0.0028, "L": 0.0017, "N": 0.0001, "P": 0.0279, "Q": 0.9513, "R": 0.0017, "S": 0.0002, "T": 0.0004, "Y": 0.0001 },
    "7": { "A": 0.0074, "D": 0.0486, "E": 0.0323, "F": 0.0001, "G": 0.0004, "H": 0.0022, "I": 0.0001, "K": 0.0001, "L": 0.0026, "N": 0.0001, "P": 0.8675, "Q": 0.0034, "R": 0.0007, "S": 0.0282, "T": 0.0054, "V": 0.0005, "Y": 0.0005 },
    "8": { "A": 0.2424, "C": 0.0001, "D": 0.0012, "F": 0.0001, "G": 0.0002, "H": 0.023, "I": 0.0001, "L": 0.0093, "N": 0.0001, "P": 0.6517, "Q": 0.0003, "R": 0.0474, "S": 0.0174, "T": 0.0051, "V": 0.0015, "Y": 0.0001 },
    "9": { "A": 0.0568, "C": 0.0003, "D": 0.0005, "E": 0.0001, "F": 0.0007, "G": 0.0001, "H": 0.0003, "L": 0.001, "P": 0.0242, "Q": 0.0003, "S": 0.9102, "T": 0.0028, "V": 0.0006, "W": 0.0001, "Y": 0.0019 },
    "11": { "A": 0.1771, "C": 0.0001, "D": 0.0002, "E": 0.0031, "F": 0.0072, "G": 0.0057, "H": 0.0005, "I": 0.0009, "L": 0.0472, "M": 0.0048, "P": 0.0025, "Q": 0.0001, "R": 0.0003, "S": 0.0028, "T": 0.002, "V": 0.7454, "W": 0.0002 },
    "12": { "A": 0.0077, "C": 0.0006, "D": 0.0001, "F": 0.0014, "H": 0.0002, "I": 0.0001, "L": 0.0006, "N": 0.0001, "P": 0.0117, "Q": 0.0001, "S": 0.9484, "T": 0.0255, "V": 0.0001, "Y": 0.0033 },
    "13": { "A": 0.0866, "C": 0.0003, "D": 0.001, "E": 0.0272, "F": 0.0004, "G": 0.5343, "K": 0.0015, "L": 0.0043, "M": 0.0007, "P": 0.0007, "Q": 0.0004, "R": 0.005, "S": 0.0007, "T": 0.0003, "V": 0.3337, "W": 0.0029 },
    "14": { "A": 0.2923, "C": 0.0007, "D": 0.0017, "F": 0.0021, "G": 0.002, "H": 0.0003, "I": 0.0003, "L": 0.0003, "N": 0.0007, "P": 0.0167, "R": 0.0001, "S": 0.5708, "T": 0.1078, "V": 0.0014, "Y": 0.0027 },
    "15": { "A": 0.0024, "F": 0.001, "H": 0.0025, "I": 0.0001, "L": 0.0844, "M": 0.0004, "P": 0.8957, "Q": 0.0022, "R": 0.0016, "S": 0.0035, "T": 0.0051, "V": 0.0007, "W": 0.0004 },
    "16": { "A": 0.0066, "C": 0.0005, "D": 0.0004, "E": 0.002, "G": 0.9674, "K": 0.0001, "L": 0.0002, "P": 0.0001, "Q": 0.0001, "R": 0.0154, "S": 0.0004, "T": 0.0002, "V": 0.0056, "W": 0.001 },
    "17": { "A": 0.0249, "D": 0.0005, "E": 0.0068, "G": 0.0332, "H": 0.0081, "K": 0.0582, "L": 0.0022, "M": 0.0003, "N": 0.0008, "P": 0.0323, "Q": 0.8207, "R": 0.0054, "S": 0.0035, "T": 0.0026, "V": 0.0003, "W": 0.0002, "Y": 0.0001 },
    "18": { "A": 0.0072, "C": 0.0001, "D": 0.0001, "E": 0.0005, "G": 0.0052, "H": 0.0001, "I": 0.0005, "K": 0.0409, "L": 0.0007, "M": 0.0036, "N": 0.0012, "P": 0.0149, "Q": 0.0012, "R": 0.2087, "S": 0.3607, "T": 0.3529, "V": 0.0002, "W": 0.0013, "Y": 0.0001 },
    "19": { "A": 0.2624, "C": 0.0001, "D": 0.0019, "F": 0.003, "G": 0.0035, "H": 0.0003, "I": 0.2292, "L": 0.0121, "M": 0.0002, "N": 0.0012, "P": 0.0009, "R": 0.0002, "S": 0.0027, "T": 0.0042, "V": 0.4779, "Y": 0.0001 },
    "20": { "A": 0.0057, "C": 0.0001, "D": 0.0002, "E": 0.0003, "F": 0.0001, "G": 0.0009, "H": 0.0005, "I": 0.0053, "K": 0.0217, "L": 0.0003, "M": 0.001, "N": 0.0056, "P": 0.0194, "Q": 0.0005, "R": 0.1837, "S": 0.1117, "T": 0.6419, "V": 0.0006, "W": 0.0003, "Y": 0.0001 },
    "21": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.0041, "H": 0.0005, "I": 0.8834, "L": 0.0825, "M": 0.0032, "N": 0.0025, "P": 0.0024, "R": 0.0009, "S": 0.0056, "T": 0.0076, "V": 0.007 },
    "22": { "A": 0.0093, "C": 0.0003, "D": 0.0001, "F": 0.0016, "H": 0.0007, "I": 0.0013, "K": 0.0003, "L": 0.0006, "M": 0.0001, "N": 0.0017, "P": 0.0438, "R": 0.0004, "S": 0.5869, "T": 0.3506, "V": 0.0002, "Y": 0.0021 },
    "23": { "A": 0.0002, "C": 0.9727, "F": 0.0047, "G": 0.0054, "L": 0.0004, "N": 0.0001, "P": 0.0002, "R": 0.0062, "S": 0.0054, "T": 0.0001, "V": 0.0003, "W": 0.0029, "Y": 0.0015 },
    "24": { "A": 0.0193, "C": 0.0006, "D": 0.0009, "E": 0.0026, "F": 0.0016, "G": 0.1025, "H": 0.0016, "I": 0.0025, "K": 0.0004, "L": 0.0007, "M": 0.0005, "N": 0.0016, "P": 0.0295, "Q": 0.0483, "R": 0.0019, "S": 0.3194, "T": 0.4625, "V": 0.0013, "W": 0.001, "Y": 0.0014 },
    "25": { "A": 0.0093, "C": 0.0006, "D": 0.0004, "E": 0.0026, "F": 0.0007, "G": 0.8972, "H": 0.0001, "I": 0.0001, "K": 0.0001, "L": 0.0379, "M": 0.0001, "N": 0.0001, "P": 0.001, "Q": 0.0001, "R": 0.0188, "S": 0.024, "T": 0.0005, "V": 0.0051, "W": 0.0012, "Y": 0.0001 },
    "26": { "A": 0.0099, "C": 0.001, "D": 0.2061, "E": 0.0058, "F": 0.0004, "G": 0.0094, "H": 0.0045, "I": 0.0051, "K": 0.0018, "L": 0.0007, "M": 0.0001, "N": 0.0808, "P": 0.0092, "Q": 0.0005, "R": 0.0192, "S": 0.3157, "T": 0.3259, "V": 0.0018, "W": 0.0001, "Y": 0.002 },
    "27": { "A": 0.0534, "C": 0.0009, "D": 0.0057, "E": 0.0026, "F": 0.0006, "G": 0.0098, "H": 0.0025, "I": 0.0064, "K": 0.0923, "L": 0.0008, "M": 0.0004, "N": 0.0971, "P": 0.0026, "Q": 0.0024, "R": 0.0239, "S": 0.6482, "T": 0.0427, "V": 0.0054, "W": 0.0002, "Y": 0.0021 },
    "28": { "A": 0.0049, "C": 0.0014, "D": 0.0031, "E": 0.0006, "F": 0.0043, "G": 0.0808, "H": 0.0009, "I": 0.0816, "K": 0.0002, "L": 0.2098, "M": 0.0018, "N": 0.0074, "P": 0.0133, "Q": 0.0001, "R": 0.0155, "S": 0.5584, "T": 0.0067, "V": 0.0056, "W": 0.0018, "Y": 0.0018 },
    "29": { "A": 0.0379, "C": 0.0001, "D": 0.3201, "E": 0.0061, "F": 0.0009, "G": 0.1853, "H": 0.0271, "I": 0.0091, "K": 0.002, "L": 0.001, "M": 0.0001, "N": 0.2561, "P": 0.0532, "Q": 0.0011, "R": 0.0518, "S": 0.0303, "T": 0.0092, "V": 0.0028, "W": 0.0007, "Y": 0.0051 },
    "30": { "A": 0.0054, "C": 0.0003, "D": 0.0016, "E": 0.0001, "F": 0.0072, "G": 0.0059, "H": 0.0006, "I": 0.4083, "K": 0.0008, "L": 0.0204, "M": 0.0016, "N": 0.0141, "P": 0.0009, "R": 0.0015, "S": 0.0317, "T": 0.0073, "V": 0.492, "Y": 0.0003 },
    "31": { "A": 0.0065, "C": 0.0036, "D": 0.0038, "E": 0.0019, "F": 0.0003, "G": 0.8757, "H": 0.0004, "I": 0.0011, "K": 0.0001, "L": 0.0005, "N": 0.0009, "P": 0.0027, "Q": 0.0001, "R": 0.0062, "S": 0.0201, "T": 0.0478, "V": 0.0252, "W": 0.0028, "Y": 0.0003 },
    "32": { "A": 0.2864, "C": 0.005, "D": 0.1508, "F": 0.005, "G": 0.1156, "H": 0.0151, "I": 0.0704, "K": 0.0201, "L": 0.0101, "M": 0.005, "N": 0.0653, "P": 0.0101, "R": 0.0352, "S": 0.1357, "T": 0.0251, "V": 0.0302, "Y": 0.0151 },
    "34": { "A": 0.0479, "D": 0.0396, "E": 0.0021, "F": 0.0104, "G": 0.175, "H": 0.0833, "I": 0.0854, "L": 0.1396, "N": 0.1562, "P": 0.0104, "Q": 0.0021, "R": 0.1146, "S": 0.0437, "T": 0.0146, "V": 0.0125, "Y": 0.0625 },
    "35": { "A": 0.1986, "C": 0.0049, "D": 0.0103, "E": 0.0056, "F": 0.0024, "G": 0.5809, "H": 0.0014, "I": 0.0024, "K": 0.0011, "L": 0.0016, "M": 0.0002, "N": 0.0095, "P": 0.0032, "Q": 0.0004, "R": 0.0137, "S": 0.1315, "T": 0.0223, "V": 0.0068, "W": 0.0007, "Y": 0.0027 },
    "36": { "A": 0.0052, "C": 0.0015, "D": 0.1021, "E": 0.0045, "F": 0.006, "G": 0.1263, "H": 0.0074, "I": 0.004, "K": 0.0552, "L": 0.0013, "M": 0.0005, "N": 0.063, "P": 0.0021, "Q": 0.0017, "R": 0.0173, "S": 0.2569, "T": 0.0255, "V": 0.003, "W": 0.0002, "Y": 0.3163 },
    "37": { "A": 0.0015, "C": 0.0005, "D": 0.0204, "E": 0.0042, "F": 0.0055, "G": 0.0014, "H": 0.0323, "I": 0.0023, "K": 0.2016, "L": 0.0011, "M": 0.0003, "N": 0.4628, "P": 0.0019, "Q": 0.0446, "R": 0.006, "S": 0.0126, "T": 0.0145, "V": 0.0005, "W": 0.0001, "Y": 0.1858 },
    "38": { "A": 0.0273, "C": 0.002, "D": 0.1014, "E": 0.0017, "F": 0.027, "G": 0.0054, "H": 0.0118, "I": 0.0066, "K": 0.0008, "L": 0.0679, "M": 0.0003, "N": 0.0139, "P": 0.0063, "Q": 0.0005, "R": 0.0161, "S": 0.0909, "T": 0.0573, "V": 0.0043, "W": 0.0011, "Y": 0.5574 },
    "39": { "A": 0.2005, "C": 0.0005, "D": 0.0019, "E": 0.0013, "F": 0.0069, "G": 0.0064, "H": 0.0003, "I": 0.0323, "K": 0.0001, "L": 0.0099, "M": 0.0004, "N": 0.0001, "P": 0.0333, "Q": 0.0001, "R": 0.0002, "S": 0.0033, "T": 0.0053, "V": 0.697, "W": 0.0001, "Y": 0.0001 },
    "40": { "A": 0.0263, "C": 0.0932, "D": 0.0025, "E": 0.0003, "F": 0.0055, "G": 0.0017, "H": 0.1826, "I": 0.0008, "K": 0.0007, "L": 0.0011, "N": 0.0779, "P": 0.0209, "Q": 0.021, "R": 0.0075, "S": 0.4304, "T": 0.0093, "V": 0.0004, "W": 0.0004, "Y": 0.1176 },
    "41": { "A": 0.0001, "C": 0.0119, "E": 0.0001, "F": 0.0001, "G": 0.007, "L": 0.006, "M": 0.0001, "Q": 0.0001, "R": 0.0063, "S": 0.0021, "V": 0.0004, "W": 0.9658, "Y": 0.0001 },
    "42": { "A": 0.0001, "C": 0.0022, "D": 0.0045, "E": 0.0001, "F": 0.0386, "H": 0.0224, "I": 0.0005, "L": 0.0043, "N": 0.0015, "P": 0.0008, "Q": 0.0003, "R": 0.0001, "S": 0.0093, "T": 0.0002, "V": 0.0005, "Y": 0.9146 },
    "43": { "A": 0.0001, "E": 0.0007, "H": 0.014, "K": 0.0059, "L": 0.003, "N": 0.0002, "P": 0.0112, "Q": 0.9528, "R": 0.0107, "S": 0.0002, "T": 0.0006, "V": 0.0001, "Y": 0.0003 },
    "44": { "A": 0.0002, "C": 0.0001, "D": 0.001, "E": 0.0014, "F": 0.0002, "G": 0.0001, "H": 0.0361, "I": 0.0001, "K": 0.0077, "L": 0.0057, "M": 0.0001, "N": 0.0004, "P": 0.0128, "Q": 0.9253, "R": 0.0055, "S": 0.0002, "T": 0.0006, "V": 0.0018, "W": 0.0001, "Y": 0.0005 },
    "45": { "A": 0.0006, "C": 0.0003, "D": 0.0007, "E": 0.0012, "F": 0.0138, "G": 0.0005, "H": 0.3055, "I": 0.0034, "K": 0.3117, "L": 0.2333, "M": 0.0007, "N": 0.0049, "P": 0.0202, "Q": 0.0255, "R": 0.0421, "S": 0.0023, "T": 0.0132, "V": 0.0104, "W": 0.0001, "Y": 0.0096 },
    "46": { "A": 0.0071, "E": 0.0002, "H": 0.0006, "L": 0.0012, "P": 0.9581, "Q": 0.0076, "R": 0.0008, "S": 0.0176, "T": 0.0064, "V": 0.0001 },
    "47": { "A": 0.0057, "C": 0.0025, "D": 0.0074, "E": 0.017, "G": 0.9516, "H": 0.0005, "K": 0.0004, "L": 0.0001, "N": 0.0007, "P": 0.0003, "Q": 0.0007, "R": 0.0046, "S": 0.0018, "T": 0.0003, "V": 0.0061, "W": 0.0001 },
    "48": { "A": 0.0087, "D": 0.0001, "E": 0.0044, "G": 0.0008, "H": 0.0044, "I": 0.0009, "K": 0.3405, "L": 0.001, "M": 0.0019, "N": 0.0078, "P": 0.0089, "Q": 0.3335, "R": 0.0102, "S": 0.0287, "T": 0.2477, "V": 0.0003 },
    "49": { "A": 0.8217, "C": 0.0003, "D": 0.0036, "E": 0.0002, "F": 0.0021, "G": 0.0197, "H": 0.0002, "I": 0.0002, "L": 0.0008, "N": 0.0001, "P": 0.0195, "R": 0.0002, "S": 0.114, "T": 0.0057, "V": 0.011, "Y": 0.0008 },
    "50": { "A": 0.0008, "H": 0.0048, "L": 0.0018, "N": 0.0001, "P": 0.9845, "Q": 0.0002, "R": 0.0006, "S": 0.0014, "T": 0.0057 },
    "51": { "A": 0.0022, "C": 0.0001, "D": 0.0002, "E": 0.0035, "F": 0.0004, "G": 0.0034, "H": 0.0015, "I": 0.0077, "K": 0.5595, "L": 0.0106, "M": 0.0022, "N": 0.0173, "P": 0.0009, "Q": 0.0171, "R": 0.0661, "S": 0.0011, "T": 0.0278, "V": 0.2781, "W": 0.0003, "Y": 0.0001 },
    "52": { "A": 0.0094, "C": 0.0003, "D": 0.0003, "E": 0.0001, "F": 0.0067, "G": 0.0004, "H": 0.0023, "I": 0.0046, "K": 0.0002, "L": 0.8565, "M": 0.0042, "N": 0.0003, "P": 0.0113, "Q": 0.0006, "R": 0.0096, "S": 0.0017, "T": 0.0403, "V": 0.0246, "W": 0.0002, "Y": 0.0262 },
    "53": { "A": 0.0011, "C": 0.0002, "D": 0.0003, "E": 0.0001, "F": 0.0036, "G": 0.0028, "H": 0.0011, "I": 0.0415, "K": 0.0006, "L": 0.3425, "M": 0.2762, "N": 0.0001, "P": 0.0043, "Q": 0.0001, "R": 0.0066, "S": 0.0007, "T": 0.0012, "V": 0.3165, "W": 0.0003 },
    "54": { "A": 0.0004, "C": 0.0002, "D": 0.0001, "F": 0.0042, "G": 0.0007, "H": 0.0002, "I": 0.8574, "K": 0.0002, "L": 0.0279, "M": 0.0317, "N": 0.002, "P": 0.0002, "R": 0.0012, "S": 0.0091, "T": 0.0039, "V": 0.0603, "Y": 0.0001 },
    "55": { "A": 0.0004, "C": 0.0037, "D": 0.0078, "E": 0.0007, "F": 0.0297, "G": 0.0002, "H": 0.0117, "I": 0.0004, "K": 0.0185, "L": 0.0013, "M": 0.0002, "N": 0.0033, "P": 0.0003, "Q": 0.0012, "R": 0.0082, "S": 0.0172, "T": 0.0009, "V": 0.0007, "W": 0.0003, "Y": 0.8933 },
    "56": { "A": 0.0119, "C": 0.0015, "D": 0.2574, "E": 0.2195, "F": 0.0014, "G": 0.1469, "H": 0.0037, "I": 0.0017, "K": 0.0512, "L": 0.0156, "M": 0.0008, "N": 0.01, "P": 0.0008, "Q": 0.1003, "R": 0.0527, "S": 0.0727, "T": 0.0043, "V": 0.0085, "W": 0.0004, "Y": 0.0387 },
    "57": { "A": 0.0066, "C": 0.0006, "D": 0.2731, "E": 0.0071, "F": 0.003, "G": 0.0418, "H": 0.0039, "I": 0.003, "K": 0.0572, "L": 0.0013, "M": 0.0001, "N": 0.2578, "P": 0.0006, "Q": 0.0009, "R": 0.0016, "S": 0.0069, "T": 0.0388, "V": 0.292, "Y": 0.0036 },
    "58": { "A": 0.0072, "C": 0.0011, "D": 0.0044, "E": 0.0004, "F": 0.0009, "G": 0.1012, "H": 0.0013, "I": 0.003, "K": 0.0009, "L": 0.0021, "M": 0.0006, "N": 0.0127, "P": 0.005, "R": 0.0208, "S": 0.824, "T": 0.0119, "V": 0.0017, "Y": 0.0008 },
    "59": { "A": 0.0093, "C": 0.0002, "D": 0.8522, "E": 0.0083, "F": 0.0001, "G": 0.0111, "H": 0.0021, "I": 0.0003, "K": 0.0007, "L": 0.0002, "M": 0.0001, "N": 0.0054, "P": 0.0001, "Q": 0.0003, "R": 0.0014, "S": 0.0978, "T": 0.0013, "V": 0.003, "Y": 0.0062 },
    "63": { "A": 0.0064, "C": 0.0033, "D": 0.0031, "E": 0.0009, "F": 0.0007, "G": 0.6696, "H": 0.0002, "I": 0.0018, "K": 0.0003, "L": 0.0035, "N": 0.0003, "P": 0.0032, "Q": 0.0001, "R": 0.0015, "S": 0.3007, "T": 0.0018, "V": 0.0019, "Y": 0.0006 },
    "64": { "A": 0.0037, "C": 0.0008, "D": 0.274, "E": 0.0121, "F": 0.001, "G": 0.0101, "H": 0.0019, "I": 0.0041, "K": 0.0009, "L": 0.0005, "M": 0.001, "N": 0.0244, "P": 0.0008, "Q": 0.0001, "R": 0.0216, "S": 0.6142, "T": 0.022, "V": 0.005, "Y": 0.0016 },
    "65": { "A": 0.003, "C": 0.0007, "D": 0.0267, "E": 0.0014, "F": 0.0025, "G": 0.006, "H": 0.0204, "I": 0.0101, "K": 0.0165, "L": 0.0007, "M": 0.0004, "N": 0.2453, "P": 0.0012, "Q": 0.0007, "R": 0.0168, "S": 0.5729, "T": 0.0632, "V": 0.0025, "W": 0.0001, "Y": 0.0089 },
    "66": { "A": 0.0043, "C": 0.0001, "D": 0.079, "E": 0.0569, "F": 0.0005, "G": 0.005, "H": 0.0107, "I": 0.0049, "K": 0.3139, "L": 0.0087, "M": 0.0013, "N": 0.3188, "P": 0.0041, "Q": 0.1312, "R": 0.0145, "S": 0.0193, "T": 0.0189, "V": 0.0024, "W": 0.0002, "Y": 0.0051 },
    "67": { "A": 0.0001, "C": 0.0001, "E": 0.0002, "G": 0.0023, "H": 0.0002, "I": 0.0002, "K": 0.0408, "L": 0.0102, "M": 0.0003, "N": 0.0007, "P": 0.0038, "Q": 0.0113, "R": 0.9246, "S": 0.0009, "T": 0.001, "V": 0.0002, "W": 0.0033 },
    "68": { "A": 0.0052, "C": 0.0002, "D": 0.0003, "E": 0.0001, "F": 0.0005, "G": 0.0288, "H": 0.0268, "L": 0.0023, "N": 0.0004, "P": 0.9134, "Q": 0.0008, "R": 0.0025, "S": 0.0125, "T": 0.0046, "V": 0.0003, "W": 0.0001, "Y": 0.0011 },
    "69": { "A": 0.0225, "C": 0.0003, "D": 0.0164, "E": 0.0004, "F": 0.0008, "G": 0.0004, "H": 0.0001, "I": 0.0001, "L": 0.0037, "N": 0.0001, "P": 0.0161, "Q": 0.0003, "R": 0.0002, "S": 0.9339, "T": 0.0032, "V": 0.0004, "W": 0.0002, "Y": 0.0008 },
    "70": { "A": 0.0029, "C": 0.0006, "D": 0.0015, "E": 0.006, "G": 0.9449, "H": 0.0001, "K": 0.0003, "L": 0.0005, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.005, "S": 0.0004, "T": 0.0003, "V": 0.0084, "W": 0.0286 },
    "71": { "A": 0.0029, "C": 0.0007, "D": 0.0005, "F": 0.0077, "G": 0.0066, "H": 0.0001, "I": 0.3604, "K": 0.0001, "L": 0.0059, "M": 0.0014, "N": 0.0014, "P": 0.0005, "R": 0.0003, "S": 0.0045, "T": 0.0287, "V": 0.5782, "Y": 0.0001 },
    "72": { "A": 0.0056, "C": 0.0004, "D": 0.0001, "F": 0.0008, "G": 0.0001, "H": 0.0016, "L": 0.0016, "P": 0.7363, "Q": 0.0003, "R": 0.0011, "S": 0.2482, "T": 0.003, "V": 0.0001, "Y": 0.0006 },
    "74": { "A": 0.0393, "C": 0.0004, "D": 0.4545, "E": 0.2388, "F": 0.0004, "G": 0.0135, "H": 0.0057, "I": 0.0017, "K": 0.0042, "L": 0.0004, "M": 0.0001, "N": 0.2043, "P": 0.0009, "Q": 0.001, "R": 0.0011, "S": 0.016, "T": 0.006, "V": 0.0053, "W": 0.0003, "Y": 0.0061 },
    "75": { "A": 0.0001, "C": 0.0008, "G": 0.002, "H": 0.0011, "I": 0.0001, "K": 0.0001, "L": 0.0068, "P": 0.0016, "Q": 0.0009, "R": 0.9853, "S": 0.0009, "T": 0.0001, "V": 0.0001, "W": 0.0001 },
    "76": { "C": 0.0115, "F": 0.9588, "G": 0.0002, "I": 0.0052, "L": 0.0089, "M": 0.0001, "R": 0.0001, "S": 0.0024, "V": 0.011, "Y": 0.0016 },
    "77": { "A": 0.0169, "C": 0.0008, "D": 0.0001, "F": 0.0035, "G": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.0005, "P": 0.0038, "S": 0.9693, "T": 0.0029, "V": 0.0002, "Y": 0.0017 },
    "78": { "A": 0.0228, "C": 0.0073, "D": 0.0021, "E": 0.0001, "F": 0.0001, "G": 0.9593, "P": 0.0001, "R": 0.0008, "S": 0.0018, "T": 0.0001, "V": 0.0053, "W": 0.0001 },
    "79": { "A": 0.0143, "C": 0.0014, "D": 0.0001, "F": 0.003, "G": 0.0001, "H": 0.0001, "L": 0.0003, "P": 0.0031, "S": 0.9711, "T": 0.0027, "V": 0.0001, "Y": 0.0035 },
    "80": { "A": 0.0005, "C": 0.0002, "D": 0.0008, "E": 0.0018, "F": 0.0004, "G": 0.0012, "H": 0.0017, "I": 0.0283, "K": 0.5972, "L": 0.0239, "M": 0.001, "N": 0.1848, "P": 0.0004, "Q": 0.0045, "R": 0.0153, "S": 0.1229, "T": 0.0132, "V": 0.0011, "W": 0.0001, "Y": 0.0008 },
    "81": { "A": 0.0082, "C": 0.0004, "D": 0.9617, "E": 0.0065, "F": 0.0015, "G": 0.0029, "H": 0.0015, "L": 0.0001, "N": 0.0021, "P": 0.0005, "Q": 0.0001, "R": 0.0001, "S": 0.0072, "T": 0.0002, "V": 0.0015, "Y": 0.0057 },
    "82": { "A": 0.3054, "C": 0.0012, "D": 0.0027, "E": 0.0003, "F": 0.0018, "G": 0.0124, "H": 0.0002, "I": 0.0062, "K": 0.0027, "L": 0.003, "M": 0.0008, "N": 0.0018, "P": 0.0035, "Q": 0.0001, "R": 0.0275, "S": 0.5949, "T": 0.0275, "V": 0.0065, "W": 0.0001, "Y": 0.0015 },
    "83": { "A": 0.0148, "C": 0.0007, "D": 0.0001, "F": 0.0023, "G": 0.0001, "H": 0.0001, "I": 0.0003, "L": 0.0319, "N": 0.0001, "P": 0.0028, "R": 0.0009, "S": 0.9394, "T": 0.0029, "V": 0.0012, "Y": 0.0023 },
    "84": { "A": 0.0172, "C": 0.0031, "D": 0.0082, "E": 0.0031, "F": 0.0001, "G": 0.9384, "H": 0.0002, "K": 0.0002, "L": 0.0001, "N": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0021, "S": 0.0196, "T": 0.0009, "V": 0.0048, "W": 0.0014, "Y": 0.0001 },
    "85": { "A": 0.0241, "C": 0.0001, "D": 0.0101, "E": 0.0002, "F": 0.0001, "G": 0.024, "H": 0.0028, "I": 0.0021, "K": 0.0048, "L": 0.0002, "M": 0.0003, "N": 0.5921, "P": 0.0028, "Q": 0.0002, "R": 0.0006, "S": 0.0104, "T": 0.3235, "V": 0.0004, "Y": 0.0015 },
    "86": { "A": 0.0171, "D": 0.0033, "E": 0.0164, "G": 0.0006, "H": 0.0001, "I": 0.0032, "K": 0.0337, "L": 0.0008, "M": 0.0116, "N": 0.0006, "P": 0.0039, "Q": 0.0003, "R": 0.0019, "S": 0.2853, "T": 0.6199, "V": 0.0013, "W": 0.0001, "Y": 0.0002 },
    "87": { "A": 0.9075, "C": 0.0001, "D": 0.0022, "E": 0.0003, "F": 0.0002, "G": 0.0121, "H": 0.0003, "I": 0.0003, "L": 0.0003, "N": 0.0001, "P": 0.0009, "R": 0.0197, "S": 0.0045, "T": 0.0023, "V": 0.0492 },
    "88": { "A": 0.0421, "C": 0.0006, "D": 0.0003, "F": 0.0037, "G": 0.0001, "H": 0.0004, "I": 0.0093, "K": 0.0001, "L": 0.0009, "M": 0.0002, "N": 0.001, "P": 0.0025, "S": 0.6209, "T": 0.2955, "V": 0.0007, "Y": 0.0218 },
    "89": { "F": 0.0011, "H": 0.0001, "I": 0.0003, "L": 0.9851, "M": 0.0025, "P": 0.0026, "Q": 0.0005, "R": 0.0052, "S": 0.0003, "T": 0.0001, "V": 0.0017, "W": 0.0005 },
    "90": { "A": 0.2215, "C": 0.0003, "D": 0.0048, "E": 0.0002, "F": 0.0003, "G": 0.0382, "H": 0.0002, "I": 0.0059, "K": 0.0004, "L": 0.0086, "M": 0.0001, "N": 0.0015, "P": 0.0026, "R": 0.0011, "S": 0.0061, "T": 0.704, "V": 0.0042 },
    "91": { "A": 0.0001, "C": 0.0001, "F": 0.0012, "G": 0.0004, "I": 0.9058, "L": 0.0284, "M": 0.0008, "N": 0.0009, "P": 0.0001, "R": 0.0004, "S": 0.0104, "T": 0.0028, "V": 0.0486 },
    "92": { "A": 0.0111, "C": 0.0008, "D": 0.0006, "F": 0.0009, "G": 0.0034, "H": 0.0002, "I": 0.0024, "K": 0.0002, "L": 0.0004, "M": 0.0001, "N": 0.0044, "P": 0.0027, "R": 0.0058, "S": 0.7563, "T": 0.2094, "V": 0.0003, "W": 0.0001, "Y": 0.0009 },
    "93": { "A": 0.0034, "C": 0.0003, "D": 0.0019, "E": 0.0043, "G": 0.8706, "H": 0.0001, "I": 0.0001, "K": 0.0006, "L": 0.0001, "M": 0.0002, "N": 0.0037, "P": 0.0001, "Q": 0.0001, "R": 0.0858, "S": 0.0179, "T": 0.0009, "V": 0.0064, "W": 0.0036 },
    "94": { "A": 0.0983, "D": 0.0004, "F": 0.0008, "G": 0.0014, "H": 0.0007, "I": 0.0024, "L": 0.6462, "N": 0.0004, "P": 0.0021, "R": 0.0069, "S": 0.0009, "T": 0.1044, "V": 0.1348 },
    "95": { "A": 0.0004, "D": 0.001, "E": 0.0806, "G": 0.0007, "H": 0.0047, "K": 0.0189, "L": 0.0039, "M": 0.0001, "N": 0.0001, "P": 0.0022, "Q": 0.8321, "R": 0.0542, "S": 0.0001, "T": 0.0003, "V": 0.0001, "W": 0.0004 },
    "96": { "A": 0.7058, "C": 0.0002, "D": 0.0035, "E": 0.0009, "F": 0.0018, "G": 0.0027, "H": 0.0003, "I": 0.0011, "L": 0.0006, "M": 0.0001, "N": 0.0004, "P": 0.0434, "R": 0.0002, "S": 0.1472, "T": 0.0705, "V": 0.0209, "Y": 0.0005 },
    "97": { "A": 0.0042, "D": 0.0302, "E": 0.7222, "F": 0.0001, "G": 0.1332, "I": 0.0017, "K": 0.0007, "L": 0.0027, "M": 0.0972, "N": 0.0001, "Q": 0.0006, "R": 0.0009, "S": 0.0001, "T": 0.0009, "V": 0.0045, "W": 0.0004, "Y": 0.0001 },
    "98": { "A": 0.0016, "D": 0.9854, "E": 0.0072, "G": 0.0011, "H": 0.0005, "M": 0.0001, "N": 0.0009, "S": 0.0001, "V": 0.0007, "Y": 0.0023 },
    "99": { "A": 0.0034, "D": 0.0093, "E": 0.9815, "G": 0.0031, "K": 0.0006, "Q": 0.0004, "R": 0.0001, "S": 0.0001, "T": 0.0002, "V": 0.0012, "W": 0.0001, "Y": 0.0001 },
    "100": { "A": 0.961, "C": 0.0008, "D": 0.004, "E": 0.0001, "G": 0.0148, "L": 0.0001, "P": 0.0004, "S": 0.0148, "T": 0.0028, "V": 0.0011, "Y": 0.0001 },
    "101": { "A": 0.0046, "D": 0.929, "E": 0.0333, "F": 0.0005, "G": 0.0019, "H": 0.0059, "I": 0.0009, "K": 0.0002, "L": 0.001, "N": 0.004, "Q": 0.0003, "R": 0.0004, "S": 0.0008, "T": 0.0031, "V": 0.0064, "Y": 0.0076 },
    "102": { "C": 0.0009, "D": 0.0069, "E": 0.0004, "F": 0.0041, "H": 0.0007, "I": 0.0001, "L": 0.0001, "N": 0.0003, "S": 0.0008, "V": 0.0001, "Y": 0.9855 },
    "103": { "A": 0.0001, "C": 0.0032, "D": 0.0053, "E": 0.0001, "F": 0.0226, "H": 0.0077, "I": 0.0003, "L": 0.0003, "N": 0.0004, "Q": 0.0001, "S": 0.005, "T": 0.0001, "V": 0.0004, "W": 0.0002, "Y": 0.9541 },
    "104": { "C": 0.9827, "F": 0.0019, "G": 0.0071, "L": 0.0001, "R": 0.0012, "S": 0.0015, "V": 0.0002, "W": 0.0039, "Y": 0.0013 },
    "105": { "A": 0.1216, "C": 0.1187, "D": 0.0008, "E": 0.0053, "F": 0.0023, "G": 0.0462, "H": 0.0055, "I": 0.0016, "K": 0.0016, "L": 0.0291, "M": 0.0086, "N": 0.0506, "P": 0.0009, "Q": 0.367, "R": 0.0024, "S": 0.206, "T": 0.0065, "V": 0.0087, "W": 0.0004, "Y": 0.0162 },
    "106": { "A": 0.2157, "C": 0.001, "D": 0.0001, "E": 0.0011, "F": 0.0009, "G": 0.0016, "H": 0.0001, "I": 0.0096, "K": 0.0002, "L": 0.0355, "M": 0.0008, "N": 0.0002, "P": 0.0014, "Q": 0.0001, "R": 0.0004, "S": 0.5664, "T": 0.0749, "V": 0.0891, "W": 0.0001, "Y": 0.0009 },
    "107": { "A": 0.0449, "C": 0.0043, "D": 0.0037, "E": 0.0006, "F": 0.0122, "G": 0.0064, "H": 0.0052, "I": 0.0004, "K": 0.0007, "L": 0.0035, "M": 0.0002, "N": 0.0012, "P": 0.0009, "Q": 0.0024, "R": 0.0522, "S": 0.0187, "T": 0.0114, "V": 0.0014, "W": 0.3793, "Y": 0.4507 },
    "108": { "A": 0.1593, "C": 0.0003, "D": 0.5738, "E": 0.0062, "F": 0.0013, "G": 0.0232, "H": 0.0103, "I": 0.0025, "K": 0.0009, "L": 0.0009, "M": 0.0081, "N": 0.0027, "P": 0.0022, "Q": 0.0003, "R": 0.0032, "S": 0.0066, "T": 0.1666, "V": 0.0062, "W": 0.0001, "Y": 0.0253 },
    "109": { "A": 0.0075, "C": 0.0009, "D": 0.1276, "E": 0.0017, "F": 0.0014, "G": 0.1775, "H": 0.0036, "I": 0.0078, "K": 0.0015, "L": 0.0017, "M": 0.0005, "N": 0.0194, "P": 0.0023, "Q": 0.0004, "R": 0.0158, "S": 0.5835, "T": 0.041, "V": 0.0039, "W": 0.0001, "Y": 0.0019 },
    "110": { "A": 0.0035, "C": 0.0006, "D": 0.0062, "E": 0.0005, "F": 0.001, "G": 0.0296, "H": 0.0011, "I": 0.008, "K": 0.0013, "L": 0.0011, "M": 0.0005, "N": 0.0171, "P": 0.0013, "Q": 0.0001, "R": 0.0219, "S": 0.8733, "T": 0.0284, "V": 0.0026, "W": 0.0003, "Y": 0.0017 },
    "111": { "A": 0.0057, "C": 0.0006, "D": 0.0065, "E": 0.0007, "F": 0.0009, "G": 0.1959, "H": 0.003, "I": 0.0025, "K": 0.002, "L": 0.454, "M": 0.0023, "N": 0.019, "P": 0.0086, "Q": 0.0037, "R": 0.0139, "S": 0.2403, "T": 0.0162, "V": 0.0094, "W": 0.0006, "Y": 0.0144 },
    "111A": { "A": 0.0835, "C": 0.0019, "D": 0.2144, "E": 0.0247, "F": 0.0019, "G": 0.0531, "H": 0.0057, "I": 0.0209, "K": 0.0076, "L": 0.0133, "N": 0.0323, "P": 0.0095, "Q": 0.0114, "R": 0.0285, "S": 0.3207, "T": 0.1157, "V": 0.0171, "W": 0.0057, "Y": 0.0323 },
    "111B": { "A": 0.0829, "D": 0.0369, "E": 0.0276, "F": 0.023, "G": 0.4608, "H": 0.0138, "I": 0.0046, "L": 0.0415, "N": 0.0046, "P": 0.0369, "Q": 0.0691, "R": 0.0138, "S": 0.0645, "T": 0.0184, "V": 0.0691, "W": 0.0046, "Y": 0.0276 },
    "112B": { "A": 0.0671, "D": 0.1555, "E": 0.0177, "F": 0.0283, "G": 0.3498, "H": 0.0106, "I": 0.0212, "K": 0.0035, "L": 0.0035, "N": 0.0247, "P": 0.0141, "Q": 0.0071, "R": 0.0707, "S": 0.0671, "T": 0.0883, "V": 0.0565, "Y": 0.0141 },
    "112A": { "A": 0.0489, "C": 0.0033, "D": 0.081, "E": 0.0033, "F": 0.0422, "G": 0.0643, "H": 0.0134, "I": 0.0074, "K": 0.0007, "L": 0.012, "M": 0.0054, "N": 0.095, "P": 0.0167, "Q": 0.0067, "R": 0.0228, "S": 0.3608, "T": 0.1814, "V": 0.0194, "W": 0.0094, "Y": 0.006 },
    "112": { "A": 0.0089, "C": 0.0009, "D": 0.039, "E": 0.0014, "F": 0.0017, "G": 0.1259, "H": 0.0034, "I": 0.0046, "K": 0.0018, "L": 0.4071, "M": 0.0031, "N": 0.0578, "P": 0.0062, "Q": 0.0019, "R": 0.0129, "S": 0.2636, "T": 0.0335, "V": 0.0063, "W": 0.0003, "Y": 0.0198 },
    "113": { "A": 0.0126, "C": 0.0006, "D": 0.0275, "E": 0.0009, "F": 0.0024, "G": 0.1046, "H": 0.011, "I": 0.0063, "K": 0.0022, "L": 0.1982, "M": 0.0015, "N": 0.066, "P": 0.0035, "Q": 0.0013, "R": 0.0166, "S": 0.4596, "T": 0.0659, "V": 0.0043, "W": 0.0003, "Y": 0.0147 },
    "114": { "A": 0.0328, "C": 0.0007, "D": 0.0571, "E": 0.0032, "F": 0.0092, "G": 0.1184, "H": 0.0281, "I": 0.0117, "K": 0.0057, "L": 0.0379, "M": 0.0008, "N": 0.1396, "P": 0.0253, "Q": 0.0044, "R": 0.0202, "S": 0.3366, "T": 0.1303, "V": 0.0068, "W": 0.001, "Y": 0.0303 },
    "115": { "A": 0.0637, "C": 0.0031, "D": 0.0242, "E": 0.0082, "F": 0.032, "G": 0.143, "H": 0.1036, "I": 0.0268, "K": 0.0066, "L": 0.0727, "M": 0.0015, "N": 0.0549, "P": 0.0522, "Q": 0.0133, "R": 0.0317, "S": 0.093, "T": 0.1982, "V": 0.0203, "W": 0.0026, "Y": 0.0484 },
    "116": { "A": 0.0271, "C": 0.0042, "D": 0.0052, "E": 0.0171, "F": 0.0133, "G": 0.073, "H": 0.0082, "I": 0.0042, "K": 0.0088, "L": 0.0697, "M": 0.0046, "N": 0.0045, "P": 0.0312, "Q": 0.0115, "R": 0.058, "S": 0.0224, "T": 0.0121, "V": 0.3199, "W": 0.165, "Y": 0.1401 },
    "117": { "A": 0.0049, "C": 0.0002, "D": 0.0006, "E": 0.0033, "F": 0.0052, "G": 0.0057, "H": 0.0003, "I": 0.0245, "K": 0.0008, "L": 0.0414, "M": 0.0037, "N": 0.0002, "P": 0.0024, "Q": 0.0007, "R": 0.0023, "S": 0.0026, "T": 0.0024, "V": 0.8984, "W": 0.0003, "Y": 0.0002 },
    "118": { "A": 0.0001, "C": 0.0038, "F": 0.9817, "G": 0.0002, "I": 0.0021, "L": 0.0045, "P": 0.0001, "R": 0.0001, "S": 0.001, "V": 0.0055, "Y": 0.0008 },
    "119": { "A": 0.0025, "C": 0.0008, "D": 0.0008, "E": 0.0003, "G": 0.9928, "R": 0.0007, "S": 0.0008, "V": 0.0012 },
    "120": { "A": 0.005, "C": 0.0001, "D": 0.0001, "E": 0.0016, "F": 0.0001, "G": 0.7677, "I": 0.001, "K": 0.0001, "L": 0.0001, "N": 0.0006, "P": 0.0056, "R": 0.0024, "S": 0.0103, "T": 0.2041, "V": 0.0011 },
    "121": { "A": 0.0014, "E": 0.001, "G": 0.9914, "R": 0.0013, "V": 0.0023, "W": 0.0023 },
    "122": { "A": 0.0014, "I": 0.0008, "N": 0.0009, "P": 0.0018, "S": 0.0017, "T": 0.9933 },
    "123": { "A": 0.0004, "D": 0.0002, "E": 0.0056, "F": 0.0001, "G": 0.0002, "H": 0.0013, "I": 0.0003, "K": 0.9376, "L": 0.0002, "M": 0.0014, "N": 0.0054, "P": 0.0001, "Q": 0.0111, "R": 0.0226, "S": 0.0018, "T": 0.0109, "V": 0.0001, "W": 0.0002, "Y": 0.0004 },
    "124": { "A": 0.0003, "D": 0.0001, "F": 0.0008, "G": 0.0005, "I": 0.0021, "L": 0.7462, "M": 0.001, "P": 0.0007, "Q": 0.0002, "R": 0.0015, "S": 0.0001, "V": 0.2464 },
    "125": { "A": 0.0043, "D": 0.0003, "F": 0.0002, "H": 0.0001, "I": 0.0031, "K": 0.0001, "L": 0.0001, "N": 0.0018, "P": 0.0029, "R": 0.0001, "S": 0.0065, "T": 0.9802, "V": 0.0004 },
    "126": { "A": 0.0027, "D": 0.0004, "F": 0.0007, "G": 0.0057, "I": 0.0016, "L": 0.0004, "R": 0.0006, "V": 0.9879 },
    "127": { "I": 0.0016, "L": 0.9949, "P": 0.0006, "V": 0.0028 }
  },
  "K": {
    "1": { "A": 0.032, "D": 0.5967, "E": 0.3652, "V": 0.006 },
    "2": { "A": 0.0007, "C": 0.0001, "D": 0.0002, "F": 0.0038, "G": 0.0004, "H": 0.0003, "I": 0.906, "K": 0.0004, "L": 0.025, "M": 0.006, "N": 0.003, "P": 0.0002, "R": 0.0012, "S": 0.0079, "T": 0.0107, "V": 0.034, "Y": 0.0001 },
    "3": { "A": 0.005, "C": 0.0001, "D": 0.0002, "E": 0.0042, "F": 0.0003, "G": 0.0034, "H": 0.0055, "I": 0.0006, "K": 0.0045, "L": 0.0083, "M": 0.002, "N": 0.0001, "P": 0.004, "Q": 0.3897, "R": 0.0183, "S": 0.0002, "T": 0.0015, "V": 0.5454, "W": 0.0066, "Y": 0.0001 },
    "4": { "C": 0.0001, "F": 0.009, "I": 0.0163, "K": 0.0008, "L": 0.3217, "M": 0.6359, "N": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0031, "S": 0.002, "T": 0.003, "V": 0.0064, "W": 0.0011 },
    "5": { "A": 0.003, "D": 0.0001, "H": 0.0004, "I": 0.0031, "K": 0.0015, "L": 0.0002, "M": 0.0012, "N": 0.0044, "P": 0.0171, "Q": 0.0002, "R": 0.0007, "S": 0.0052, "T": 0.9625, "V": 0.0001 },
    "6": { "A": 0.0001, "D": 0.0001, "E": 0.0016, "H": 0.0146, "K": 0.0071, "L": 0.0024, "M": 0.0001, "N": 0.0002, "P": 0.0201, "Q": 0.9504, "R": 0.0022, "S": 0.0003, "T": 0.0007, "Y": 0.0001 },
    "7": { "A": 0.0053, "C": 0.001, "D": 0.0001, "F": 0.0045, "H": 0.0004, "I": 0.0001, "L": 0.0005, "N": 0.0001, "P": 0.0117, "S": 0.9461, "T": 0.0236, "V": 0.0001, "Y": 0.0062 },
    "8": { "A": 0.0017, "H": 0.0008, "L": 0.0043, "P": 0.9768, "Q": 0.0057, "R": 0.0015, "S": 0.0029, "T": 0.0062 },
    "9": { "A": 0.2311, "C": 0.0011, "D": 0.0887, "E": 0.0015, "F": 0.0024, "G": 0.1329, "H": 0.001, "I": 0.0004, "L": 0.0992, "M": 0.0001, "N": 0.0002, "P": 0.0069, "Q": 0.0001, "R": 0.0015, "S": 0.4223, "T": 0.0021, "V": 0.0063, "Y": 0.0022 },
    "10": { "A": 0.0058, "C": 0.0007, "D": 0.0001, "F": 0.0294, "H": 0.0003, "I": 0.0028, "L": 0.0075, "N": 0.0015, "P": 0.0126, "R": 0.0001, "S": 0.4888, "T": 0.4471, "V": 0.0003, "Y": 0.0028 },
    "11": { "A": 0.0003, "C": 0.0001, "F": 0.0104, "G": 0.0001, "H": 0.0001, "I": 0.0003, "L": 0.9389, "M": 0.0074, "P": 0.0037, "Q": 0.0054, "R": 0.0041, "S": 0.0031, "T": 0.0001, "V": 0.0258, "W": 0.0001 },
    "12": { "A": 0.0858, "C": 0.0005, "D": 0.0005, "F": 0.0014, "G": 0.0002, "H": 0.0013, "L": 0.0012, "P": 0.09, "R": 0.0004, "S": 0.8118, "T": 0.0036, "V": 0.0007, "Y": 0.0025 },
    "13": { "A": 0.429, "C": 0.0001, "D": 0.0002, "E": 0.0014, "F": 0.0029, "G": 0.0024, "I": 0.0006, "L": 0.2605, "M": 0.001, "P": 0.001, "R": 0.0001, "S": 0.0038, "T": 0.0015, "V": 0.2936, "W": 0.0018 },
    "14": { "A": 0.0054, "C": 0.0007, "D": 0.0001, "F": 0.0037, "H": 0.0001, "I": 0.0007, "L": 0.0003, "N": 0.0004, "P": 0.0059, "S": 0.8731, "T": 0.1065, "V": 0.0001, "Y": 0.0029 },
    "15": { "A": 0.0033, "E": 0.0007, "F": 0.0002, "G": 0.0014, "H": 0.0005, "I": 0.0052, "K": 0.0001, "L": 0.1142, "M": 0.0006, "P": 0.446, "Q": 0.0028, "R": 0.0014, "S": 0.0015, "T": 0.0249, "V": 0.3972 },
    "16": { "A": 0.0028, "C": 0.001, "D": 0.0003, "E": 0.0018, "G": 0.9787, "K": 0.0041, "N": 0.0001, "Q": 0.0001, "R": 0.0044, "S": 0.0003, "T": 0.0001, "V": 0.0051, "W": 0.001 },
    "17": { "A": 0.0125, "D": 0.4323, "E": 0.5012, "G": 0.0044, "H": 0.0011, "K": 0.0014, "L": 0.0001, "N": 0.001, "P": 0.0004, "Q": 0.041, "R": 0.0003, "S": 0.0003, "T": 0.0002, "V": 0.0022, "Y": 0.0015 },
    "18": { "A": 0.0008, "C": 0.0001, "E": 0.0002, "G": 0.006, "I": 0.0023, "K": 0.0082, "L": 0.0004, "M": 0.0006, "N": 0.0002, "P": 0.0993, "Q": 0.0011, "R": 0.8547, "S": 0.0172, "T": 0.0083, "V": 0.0001, "W": 0.0004 },
    "19": { "A": 0.5473, "C": 0.0001, "D": 0.0031, "F": 0.0012, "G": 0.0027, "H": 0.0001, "I": 0.0027, "L": 0.0015, "P": 0.0017, "S": 0.0024, "T": 0.0018, "V": 0.4354 },
    "20": { "A": 0.004, "C": 0.0003, "D": 0.0001, "F": 0.0002, "H": 0.0004, "I": 0.0049, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0059, "P": 0.0088, "R": 0.0001, "S": 0.1059, "T": 0.8682, "V": 0.0003, "Y": 0.0005 },
    "21": { "A": 0.0001, "F": 0.0032, "H": 0.0009, "I": 0.609, "L": 0.3675, "M": 0.0017, "N": 0.0012, "P": 0.0034, "R": 0.0026, "S": 0.0031, "T": 0.0036, "V": 0.0036 },
    "22": { "A": 0.0075, "C": 0.0003, "D": 0.0005, "F": 0.0013, "G": 0.0001, "H": 0.0033, "I": 0.001, "K": 0.001, "L": 0.0002, "N": 0.0801, "P": 0.0216, "Q": 0.0001, "R": 0.0003, "S": 0.4601, "T": 0.4204, "V": 0.0001, "Y": 0.0021 },
    "23": { "A": 0.0001, "C": 0.9777, "D": 0.0001, "F": 0.0041, "G": 0.0049, "H": 0.0001, "L": 0.0002, "P": 0.0001, "R": 0.0048, "S": 0.0046, "V": 0.0002, "W": 0.0009, "Y": 0.0022 },
    "24": { "A": 0.0001, "C": 0.0001, "E": 0.0005, "G": 0.0055, "H": 0.0005, "I": 0.0001, "K": 0.098, "L": 0.0023, "M": 0.0019, "N": 0.0013, "P": 0.0014, "Q": 0.0578, "R": 0.8156, "S": 0.0066, "T": 0.0046, "V": 0.0001, "W": 0.0036 },
    "25": { "A": 0.7809, "C": 0.0003, "D": 0.0023, "E": 0.002, "F": 0.0004, "G": 0.001, "H": 0.0001, "I": 0.0001, "L": 0.0002, "M": 0.0064, "P": 0.0053, "R": 0.0001, "S": 0.1891, "T": 0.0063, "V": 0.0046, "Y": 0.0009 },
    "26": { "A": 0.0002, "C": 0.0011, "D": 0.0003, "G": 0.0037, "H": 0.0001, "I": 0.0032, "K": 0.0001, "L": 0.0001, "M": 0.0002, "N": 0.0038, "P": 0.0001, "R": 0.0144, "S": 0.9684, "T": 0.0042, "V": 0.0001 },
    "27": { "A": 0.0001, "D": 0.0005, "E": 0.0063, "G": 0.0001, "H": 0.0099, "K": 0.0055, "L": 0.0025, "N": 0.0001, "P": 0.0089, "Q": 0.9591, "R": 0.0061, "S": 0.0002, "T": 0.0004, "V": 0.0001, "Y": 0.0001 },
    "28": { "A": 0.0029, "C": 0.0034, "D": 0.0643, "E": 0.0005, "F": 0.0004, "G": 0.1276, "H": 0.0009, "I": 0.0047, "K": 0.0003, "L": 0.0003, "M": 0.0002, "N": 0.0127, "P": 0.0009, "Q": 0.0001, "R": 0.0214, "S": 0.7408, "T": 0.0152, "V": 0.0018, "W": 0.0001, "Y": 0.0014 },
    "29": { "A": 0.0018, "C": 0.0001, "D": 0.0007, "F": 0.0044, "G": 0.0028, "H": 0.0003, "I": 0.4386, "L": 0.1114, "M": 0.0023, "N": 0.0007, "P": 0.001, "R": 0.0009, "S": 0.0044, "T": 0.0014, "V": 0.4291 },
    "30": { "A": 0.0032, "C": 0.0005, "D": 0.0029, "E": 0.0009, "F": 0.0115, "G": 0.008, "H": 0.0006, "I": 0.0045, "K": 0.0009, "L": 0.4809, "M": 0.0016, "N": 0.0075, "P": 0.0019, "Q": 0.001, "R": 0.0129, "S": 0.3719, "T": 0.0131, "V": 0.0739, "W": 0.0003, "Y": 0.0017 },
    "31": { "A": 0.0013, "C": 0.0013, "D": 0.0256, "E": 0.0009, "F": 0.0095, "G": 0.0005, "H": 0.4151, "I": 0.0004, "K": 0.0018, "L": 0.002, "M": 0.0001, "N": 0.0096, "P": 0.0038, "Q": 0.0068, "R": 0.0036, "S": 0.0207, "T": 0.0006, "V": 0.0005, "W": 0.0013, "Y": 0.4945 },
    "32": { "A": 0.001, "C": 0.0014, "D": 0.0025, "E": 0.0002, "F": 0.0006, "G": 0.0093, "H": 0.001, "I": 0.0067, "K": 0.0016, "L": 0.0006, "M": 0.0004, "N": 0.019, "P": 0.0013, "Q": 0.0001, "R": 0.0331, "S": 0.8989, "T": 0.0201, "V": 0.0009, "W": 0.0001, "Y": 0.0012 },
    "33": { "A": 0.0157, "C": 0.0019, "D": 0.0386, "E": 0.0003, "F": 0.0094, "G": 0.0014, "H": 0.0016, "I": 0.0008, "L": 0.0023, "N": 0.0007, "P": 0.0298, "Q": 0.0001, "R": 0.0007, "S": 0.8692, "T": 0.0098, "V": 0.0024, "Y": 0.0153 },
    "34": { "A": 0.0028, "D": 0.2282, "E": 0.0016, "F": 0.0003, "G": 0.0022, "H": 0.0056, "I": 0.003, "K": 0.0086, "L": 0.0004, "M": 0.0001, "N": 0.7203, "P": 0.0002, "Q": 0.0004, "R": 0.0009, "S": 0.0087, "T": 0.0119, "V": 0.0011, "Y": 0.0037 },
    "35": { "A": 0.0019, "C": 0.0001, "D": 0.0044, "E": 0.0035, "F": 0.0003, "G": 0.5302, "H": 0.0045, "I": 0.0013, "K": 0.0095, "L": 0.0002, "M": 0.0002, "N": 0.4203, "P": 0.0001, "Q": 0.0003, "R": 0.0041, "S": 0.0092, "T": 0.0044, "V": 0.0033, "W": 0.0001, "Y": 0.0019 },
    "36": { "A": 0.003, "C": 0.0009, "D": 0.0069, "E": 0.0012, "F": 0.0012, "G": 0.0221, "H": 0.0023, "I": 0.0044, "K": 0.0956, "L": 0.0007, "M": 0.0006, "N": 0.0503, "P": 0.0003, "Q": 0.0017, "R": 0.0336, "S": 0.7058, "T": 0.0109, "V": 0.0008, "W": 0.0001, "Y": 0.0577 },
    "37": { "A": 0.0017, "C": 0.0009, "D": 0.0075, "E": 0.0004, "F": 0.0007, "G": 0.006, "H": 0.0043, "I": 0.0079, "K": 0.0071, "L": 0.0004, "M": 0.0005, "N": 0.2719, "P": 0.0012, "Q": 0.0004, "R": 0.0179, "S": 0.5976, "T": 0.0691, "V": 0.0013, "W": 0.0002, "Y": 0.003 },
    "38": { "A": 0.0085, "C": 0.0034, "D": 0.0227, "E": 0.0007, "F": 0.017, "G": 0.001, "H": 0.009, "I": 0.0004, "K": 0.0025, "L": 0.0021, "M": 0.0001, "N": 0.101, "P": 0.0002, "Q": 0.0009, "R": 0.0024, "S": 0.0239, "T": 0.0019, "V": 0.0003, "W": 0.107, "Y": 0.695 },
    "39": { "A": 0.0001, "C": 0.0002, "F": 0.0114, "I": 0.0034, "L": 0.9666, "M": 0.0018, "R": 0.0001, "S": 0.0024, "V": 0.013, "W": 0.0008, "Y": 0.0001 },
    "40": { "A": 0.6369, "C": 0.0003, "D": 0.0655, "E": 0.001, "F": 0.0003, "G": 0.0195, "H": 0.0098, "I": 0.0012, "K": 0.002, "L": 0.0002, "M": 0.0001, "N": 0.2208, "P": 0.0021, "Q": 0.0002, "R": 0.0002, "S": 0.0114, "T": 0.0078, "V": 0.0056, "Y": 0.0152 },
    "41": { "C": 0.006, "G": 0.0045, "L": 0.0053, "R": 0.0032, "S": 0.001, "V": 0.0002, "W": 0.9796 },
    "42": { "C": 0.0015, "D": 0.0034, "E": 0.0001, "F": 0.0407, "H": 0.0028, "I": 0.0003, "L": 0.0038, "N": 0.0012, "Q": 0.0001, "S": 0.0039, "V": 0.0004, "W": 0.0001, "Y": 0.9418 },
    "43": { "E": 0.0005, "F": 0.0002, "H": 0.0095, "K": 0.0039, "L": 0.0771, "M": 0.0004, "N": 0.0001, "P": 0.0036, "Q": 0.8966, "R": 0.0063, "S": 0.0001, "T": 0.0002, "V": 0.0014, "Y": 0.0001 },
    "44": { "A": 0.0001, "E": 0.0011, "F": 0.0001, "H": 0.0139, "K": 0.0055, "L": 0.0025, "M": 0.0001, "N": 0.0001, "P": 0.0038, "Q": 0.9689, "R": 0.0028, "T": 0.0002, "V": 0.0005, "W": 0.0001, "Y": 0.0002 },
    "45": { "A": 0.0001, "E": 0.0023, "G": 0.0002, "H": 0.0003, "I": 0.002, "K": 0.9283, "L": 0.0001, "M": 0.0003, "N": 0.0081, "P": 0.0001, "Q": 0.0081, "R": 0.0416, "S": 0.0006, "T": 0.0076, "V": 0.0001, "Y": 0.0001 },
    "46": { "A": 0.0059, "D": 0.0001, "E": 0.0001, "F": 0.0005, "G": 0.0004, "H": 0.002, "I": 0.0001, "L": 0.0031, "P": 0.97, "Q": 0.0038, "R": 0.0024, "S": 0.0055, "T": 0.0049, "V": 0.001 },
    "47": { "A": 0.0026, "C": 0.0016, "D": 0.0056, "E": 0.0037, "G": 0.9762, "H": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0032, "S": 0.001, "V": 0.0044, "W": 0.0013 },
    "48": { "A": 0.0002, "D": 0.0001, "E": 0.0058, "G": 0.0003, "H": 0.0038, "I": 0.0005, "K": 0.4133, "L": 0.0048, "M": 0.0002, "N": 0.005, "P": 0.0032, "Q": 0.5522, "R": 0.0069, "S": 0.0003, "T": 0.0033 },
    "49": { "A": 0.752, "C": 0.0001, "D": 0.0048, "E": 0.0002, "F": 0.0003, "G": 0.0023, "H": 0.0006, "I": 0.0002, "L": 0.0004, "P": 0.1089, "R": 0.0002, "S": 0.0982, "T": 0.0054, "V": 0.0258, "Y": 0.0005 },
    "50": { "A": 0.0022, "H": 0.0039, "L": 0.0018, "P": 0.9829, "Q": 0.0011, "R": 0.0006, "S": 0.0016, "T": 0.0058 },
    "51": { "A": 0.0001, "D": 0.0007, "E": 0.01, "G": 0.0012, "H": 0.0017, "I": 0.0016, "K": 0.4827, "L": 0.0003, "M": 0.0023, "N": 0.0137, "P": 0.0007, "Q": 0.08, "R": 0.392, "S": 0.0058, "T": 0.0065, "V": 0.0002, "W": 0.0004, "Y": 0.0001 },
    "52": { "A": 0.0004, "C": 0.0002, "F": 0.0048, "G": 0.0001, "H": 0.0027, "I": 0.0044, "L": 0.9139, "M": 0.0005, "N": 0.0001, "P": 0.0076, "Q": 0.0001, "R": 0.0403, "S": 0.0111, "T": 0.0002, "V": 0.0134, "W": 0.0001, "Y": 0.0001 },
    "53": { "F": 0.0017, "H": 0.0009, "I": 0.004, "L": 0.9792, "M": 0.0018, "P": 0.0028, "Q": 0.0005, "R": 0.0053, "S": 0.0001, "V": 0.0035 },
    "54": { "C": 0.0001, "F": 0.0044, "H": 0.0001, "I": 0.955, "L": 0.0145, "M": 0.0065, "N": 0.0025, "P": 0.0001, "R": 0.0003, "S": 0.0071, "T": 0.003, "V": 0.0062, "Y": 0.0001 },
    "55": { "A": 0.0003, "C": 0.0018, "D": 0.0079, "E": 0.0002, "F": 0.0118, "H": 0.0089, "I": 0.0001, "K": 0.0042, "L": 0.0003, "M": 0.0001, "N": 0.0021, "P": 0.0001, "Q": 0.0018, "R": 0.0003, "S": 0.0149, "T": 0.0003, "V": 0.0001, "W": 0.0001, "Y": 0.9444 },
    "56": { "A": 0.2602, "C": 0.0027, "D": 0.2111, "E": 0.0222, "F": 0.001, "G": 0.2362, "H": 0.0014, "I": 0.0002, "K": 0.0843, "L": 0.0575, "M": 0.001, "N": 0.0019, "P": 0.0006, "Q": 0.0031, "R": 0.0042, "S": 0.0095, "T": 0.0076, "V": 0.0047, "W": 0.0844, "Y": 0.0065 },
    "57": { "A": 0.8609, "C": 0.0003, "D": 0.0005, "E": 0.0033, "F": 0.0002, "G": 0.0591, "I": 0.0038, "K": 0.0001, "L": 0.0038, "M": 0.0001, "N": 0.0001, "P": 0.0013, "R": 0.0002, "S": 0.0091, "T": 0.0149, "V": 0.0422 },
    "58": { "A": 0.0047, "D": 0.0094, "F": 0.033, "G": 0.0094, "H": 0.0236, "K": 0.0094, "L": 0.0094, "P": 0.0142, "S": 0.4953, "T": 0.1415, "Y": 0.25 },
    "63": { "A": 0.0571, "C": 0.0057, "D": 0.0343, "E": 0.0286, "G": 0.3829, "H": 0.0057, "K": 0.1543, "L": 0.0057, "P": 0.0114, "R": 0.2629, "S": 0.0114, "T": 0.0229, "V": 0.0171 },
    "64": { "A": 0.559, "C": 0.0035, "E": 0.0903, "F": 0.0035, "G": 0.0521, "H": 0.0104, "I": 0.0139, "L": 0.0104, "N": 0.0139, "P": 0.0069, "Q": 0.0069, "R": 0.0104, "S": 0.125, "T": 0.0243, "V": 0.059, "W": 0.0035, "Y": 0.0069 },
    "65": { "A": 0.0099, "C": 0.0009, "D": 0.0002, "F": 0.0049, "G": 0.0001, "H": 0.0001, "I": 0.0003, "L": 0.0004, "N": 0.0002, "P": 0.0031, "S": 0.9686, "T": 0.0058, "V": 0.0005, "Y": 0.005 },
    "66": { "A": 0.003, "C": 0.0004, "D": 0.0029, "E": 0.0003, "F": 0.0006, "G": 0.0019, "H": 0.0033, "I": 0.0106, "K": 0.0079, "L": 0.0006, "M": 0.0004, "N": 0.2695, "P": 0.0019, "Q": 0.0045, "R": 0.0139, "S": 0.3969, "T": 0.2744, "V": 0.0014, "Y": 0.0057 },
    "67": { "A": 0.0001, "C": 0.0001, "F": 0.0028, "G": 0.0016, "I": 0.0003, "K": 0.0009, "L": 0.4214, "M": 0.0028, "P": 0.0005, "Q": 0.0005, "R": 0.549, "S": 0.0097, "T": 0.0009, "V": 0.0046, "W": 0.0047, "Y": 0.0001 },
    "68": { "A": 0.4264, "C": 0.0004, "D": 0.0267, "E": 0.2297, "F": 0.0194, "G": 0.0027, "H": 0.0046, "I": 0.0017, "K": 0.0036, "L": 0.0012, "M": 0.0001, "N": 0.0003, "P": 0.0031, "Q": 0.2691, "R": 0.0015, "S": 0.0027, "T": 0.0018, "V": 0.0046, "Y": 0.0005 },
    "69": { "A": 0.0118, "C": 0.0009, "D": 0.0009, "E": 0.0001, "F": 0.0012, "G": 0.0042, "H": 0.0003, "I": 0.0032, "K": 0.001, "L": 0.0003, "M": 0.0002, "N": 0.0049, "P": 0.0095, "R": 0.0065, "S": 0.5393, "T": 0.4137, "V": 0.0005, "W": 0.0001, "Y": 0.0014 },
    "70": { "A": 0.0032, "C": 0.002, "D": 0.0051, "E": 0.0021, "G": 0.9739, "N": 0.0002, "P": 0.0001, "R": 0.0034, "S": 0.0015, "T": 0.0001, "V": 0.0052, "W": 0.003 },
    "71": { "A": 0.003, "C": 0.0002, "D": 0.0008, "F": 0.0037, "G": 0.0044, "I": 0.3516, "L": 0.0045, "M": 0.0004, "N": 0.0007, "P": 0.0001, "R": 0.0001, "S": 0.0038, "T": 0.0037, "V": 0.623 },
    "72": { "A": 0.0026, "H": 0.0007, "I": 0.0001, "K": 0.0004, "L": 0.0013, "P": 0.9806, "Q": 0.0032, "R": 0.0006, "S": 0.0055, "T": 0.0048 },
    "74": { "A": 0.2276, "C": 0.0001, "D": 0.3222, "E": 0.0043, "F": 0.0002, "G": 0.0033, "H": 0.0007, "I": 0.0001, "K": 0.0001, "L": 0.0017, "N": 0.0011, "P": 0.0064, "Q": 0.0001, "R": 0.0002, "S": 0.4241, "T": 0.003, "V": 0.0025, "Y": 0.0023 },
    "75": { "C": 0.0001, "E": 0.0001, "G": 0.002, "I": 0.0002, "K": 0.0117, "L": 0.0008, "M": 0.0045, "N": 0.0002, "P": 0.0002, "Q": 0.0004, "R": 0.9692, "S": 0.0073, "T": 0.0021, "V": 0.0001, "W": 0.001 },
    "76": { "C": 0.015, "F": 0.9612, "G": 0.0002, "I": 0.0032, "L": 0.0084, "S": 0.0017, "V": 0.0087, "W": 0.0001, "Y": 0.0013 },
    "77": { "A": 0.0003, "C": 0.0011, "D": 0.0002, "F": 0.0001, "G": 0.0046, "H": 0.0001, "I": 0.005, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0024, "R": 0.0158, "S": 0.9645, "T": 0.0051, "V": 0.0003 },
    "78": { "A": 0.0043, "C": 0.0055, "D": 0.002, "E": 0.0001, "F": 0.0001, "G": 0.9783, "P": 0.0001, "R": 0.0013, "S": 0.0037, "V": 0.0044, "W": 0.0001 },
    "79": { "A": 0.0005, "C": 0.0011, "D": 0.0003, "F": 0.0002, "G": 0.0106, "H": 0.0001, "I": 0.0026, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0019, "R": 0.0146, "S": 0.9638, "T": 0.0033, "V": 0.0002, "W": 0.0001, "Y": 0.0002 },
    "80": { "A": 0.0045, "C": 0.0004, "D": 0.0001, "E": 0.0037, "G": 0.973, "K": 0.0002, "L": 0.0002, "Q": 0.0002, "R": 0.0058, "S": 0.0004, "T": 0.0001, "V": 0.0077, "W": 0.0035 },
    "81": { "A": 0.0184, "C": 0.0046, "G": 0.0092, "I": 0.0046, "N": 0.0046, "P": 0.0092, "S": 0.9217, "T": 0.0184, "Y": 0.0092 },
    "82": { "A": 0.0997, "D": 0.0412, "E": 0.0481, "G": 0.457, "H": 0.0034, "I": 0.0206, "L": 0.0034, "Q": 0.1787, "R": 0.0069, "S": 0.0653, "T": 0.0103, "V": 0.055, "W": 0.0034, "Y": 0.0069 },
    "83": { "A": 0.0147, "C": 0.0025, "D": 0.0004, "E": 0.0001, "F": 0.004, "G": 0.0003, "H": 0.0001, "L": 0.0004, "P": 0.0023, "S": 0.9634, "T": 0.0019, "V": 0.0002, "W": 0.0001, "Y": 0.0093 },
    "84": { "A": 0.0048, "C": 0.0008, "D": 0.001, "E": 0.008, "G": 0.9673, "H": 0.0001, "K": 0.0001, "L": 0.0002, "Q": 0.0002, "R": 0.0036, "S": 0.0004, "T": 0.0001, "V": 0.0078, "W": 0.0056 },
    "85": { "A": 0.0064, "E": 0.0001, "G": 0.0001, "I": 0.002, "K": 0.0066, "L": 0.0001, "N": 0.0003, "P": 0.0021, "Q": 0.0001, "R": 0.0026, "S": 0.0028, "T": 0.9767, "V": 0.0001 },
    "86": { "A": 0.0046, "D": 0.7475, "E": 0.2265, "F": 0.0002, "G": 0.0027, "H": 0.0033, "I": 0.0001, "K": 0.0008, "L": 0.0005, "N": 0.0036, "Q": 0.0014, "R": 0.0002, "S": 0.0005, "T": 0.0004, "V": 0.0026, "Y": 0.0053 },
    "87": { "C": 0.0053, "D": 0.0001, "F": 0.9711, "G": 0.0001, "H": 0.0001, "I": 0.0013, "L": 0.0048, "S": 0.0019, "V": 0.0079, "Y": 0.0074 },
    "88": { "A": 0.0046, "F": 0.0001, "H": 0.0002, "I": 0.0038, "K": 0.0004, "L": 0.0002, "M": 0.0001, "N": 0.0033, "P": 0.0048, "R": 0.0002, "S": 0.0126, "T": 0.9694, "V": 0.0002, "Y": 0.0001 },
    "89": { "C": 0.0004, "F": 0.0547, "H": 0.0011, "I": 0.0061, "L": 0.9281, "M": 0.0005, "P": 0.0022, "Q": 0.0001, "R": 0.005, "S": 0.0003, "T": 0.0001, "V": 0.0012, "Y": 0.0001 },
    "90": { "A": 0.0031, "D": 0.0001, "E": 0.0009, "F": 0.0001, "G": 0.0001, "H": 0.0002, "I": 0.0052, "K": 0.0963, "L": 0.0002, "M": 0.0001, "N": 0.005, "P": 0.0048, "Q": 0.0006, "R": 0.0023, "S": 0.0068, "T": 0.874, "V": 0.0003 },
    "91": { "F": 0.0012, "I": 0.9752, "L": 0.0039, "M": 0.0008, "N": 0.0013, "R": 0.0001, "S": 0.0117, "T": 0.0019, "V": 0.0036 },
    "92": { "A": 0.001, "C": 0.001, "D": 0.0019, "F": 0.0001, "G": 0.0032, "H": 0.0002, "I": 0.0057, "K": 0.0003, "L": 0.0002, "M": 0.0001, "N": 0.0191, "R": 0.0095, "S": 0.9453, "T": 0.0117, "V": 0.0002, "Y": 0.0003 },
    "93": { "A": 0.0004, "C": 0.023, "D": 0.0008, "E": 0.0001, "F": 0.0002, "G": 0.0104, "H": 0.0001, "I": 0.0056, "K": 0.0013, "L": 0.0004, "M": 0.0005, "N": 0.0089, "P": 0.0001, "Q": 0.0001, "R": 0.251, "S": 0.6938, "T": 0.003, "V": 0.0001, "W": 0.0001, "Y": 0.0002 },
    "94": { "A": 0.0002, "E": 0.0001, "F": 0.0001, "G": 0.001, "I": 0.0023, "L": 0.8684, "M": 0.0082, "P": 0.0016, "Q": 0.0011, "R": 0.0111, "T": 0.0001, "V": 0.1056, "W": 0.0001 },
    "95": { "A": 0.0017, "D": 0.0038, "E": 0.3655, "G": 0.0028, "H": 0.0044, "K": 0.0027, "L": 0.0016, "N": 0.0001, "P": 0.0016, "Q": 0.6116, "R": 0.0026, "S": 0.0001, "T": 0.0001, "V": 0.0013, "Y": 0.0001 },
    "96": { "A": 0.1943, "C": 0.0001, "D": 0.0014, "E": 0.0001, "F": 0.0004, "G": 0.0003, "H": 0.0046, "I": 0.0001, "L": 0.0019, "N": 0.0001, "P": 0.6576, "Q": 0.0005, "R": 0.0011, "S": 0.131, "T": 0.005, "V": 0.0008, "Y": 0.0008 },
    "97": { "A": 0.0039, "D": 0.0945, "E": 0.8933, "G": 0.0035, "H": 0.0001, "K": 0.0012, "L": 0.0001, "N": 0.0003, "P": 0.0001, "Q": 0.0008, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0013, "Y": 0.0006 },
    "98": { "A": 0.0025, "D": 0.9811, "E": 0.0069, "F": 0.0001, "G": 0.0013, "H": 0.0005, "I": 0.0001, "K": 0.0001, "M": 0.0001, "N": 0.001, "S": 0.0001, "V": 0.0016, "Y": 0.0044 },
    "99": { "A": 0.0074, "C": 0.0068, "D": 0.0002, "E": 0.0001, "F": 0.6827, "G": 0.0016, "I": 0.0555, "L": 0.0086, "M": 0.0005, "N": 0.0001, "S": 0.0118, "T": 0.0004, "V": 0.2223, "W": 0.0001, "Y": 0.0018 },
    "100": { "A": 0.8761, "C": 0.0003, "D": 0.0001, "E": 0.0037, "G": 0.1072, "L": 0.0001, "P": 0.0004, "Q": 0.0002, "R": 0.0002, "S": 0.0042, "T": 0.0023, "V": 0.0046, "W": 0.0006 },
    "101": { "A": 0.0056, "D": 0.0007, "E": 0.0002, "F": 0.0034, "G": 0.0022, "H": 0.0001, "I": 0.0136, "K": 0.0004, "L": 0.0065, "M": 0.0015, "N": 0.0024, "P": 0.0017, "Q": 0.0001, "R": 0.0002, "S": 0.0044, "T": 0.4211, "V": 0.5347, "Y": 0.0013 },
    "102": { "C": 0.0013, "D": 0.0056, "E": 0.0002, "F": 0.0032, "H": 0.0011, "I": 0.0004, "L": 0.0007, "N": 0.0007, "S": 0.0027, "T": 0.0001, "V": 0.0001, "Y": 0.984 },
    "103": { "C": 0.0046, "D": 0.0044, "E": 0.0001, "F": 0.0184, "H": 0.0043, "I": 0.0003, "L": 0.0007, "N": 0.0006, "S": 0.0063, "T": 0.0007, "V": 0.0003, "W": 0.0002, "Y": 0.959 },
    "104": { "A": 0.0002, "C": 0.9739, "F": 0.0031, "G": 0.0074, "H": 0.0001, "L": 0.0008, "R": 0.0022, "S": 0.0015, "T": 0.0002, "V": 0.0007, "W": 0.0074, "Y": 0.0024 },
    "105": { "A": 0.0001, "C": 0.0002, "E": 0.0009, "F": 0.0002, "G": 0.0001, "H": 0.016, "I": 0.0007, "K": 0.0045, "L": 0.0216, "M": 0.0959, "N": 0.0003, "P": 0.0015, "Q": 0.8475, "R": 0.008, "S": 0.0008, "T": 0.0007, "V": 0.0007, "W": 0.0001, "Y": 0.0004 },
    "106": { "A": 0.0003, "C": 0.0002, "D": 0.0003, "E": 0.0031, "F": 0.0001, "G": 0.0003, "H": 0.0208, "I": 0.0001, "K": 0.0264, "L": 0.0027, "M": 0.0002, "N": 0.0014, "P": 0.0032, "Q": 0.9294, "R": 0.0086, "S": 0.0011, "T": 0.0004, "V": 0.0003, "W": 0.0001, "Y": 0.001 },
    "107": { "A": 0.0786, "C": 0.006, "D": 0.0208, "E": 0.0028, "F": 0.011, "G": 0.0292, "H": 0.0157, "I": 0.0014, "K": 0.0004, "L": 0.021, "M": 0.0002, "N": 0.0035, "P": 0.0014, "Q": 0.001, "R": 0.1266, "S": 0.162, "T": 0.0081, "V": 0.0024, "W": 0.0003, "Y": 0.5074 },
    "108": { "A": 0.0046, "C": 0.0021, "D": 0.0779, "E": 0.0031, "F": 0.0043, "G": 0.1302, "H": 0.0077, "I": 0.0195, "K": 0.0084, "L": 0.0592, "M": 0.0004, "N": 0.2571, "P": 0.0003, "Q": 0.0012, "R": 0.0054, "S": 0.1239, "T": 0.027, "V": 0.0047, "W": 0.0004, "Y": 0.2627 },
    "109": { "A": 0.0029, "C": 0.0009, "D": 0.0121, "E": 0.0059, "F": 0.0009, "G": 0.0092, "H": 0.0266, "I": 0.0059, "K": 0.0059, "L": 0.0017, "M": 0.0004, "N": 0.28, "P": 0.0014, "Q": 0.0684, "R": 0.0142, "S": 0.5286, "T": 0.0299, "V": 0.0015, "W": 0.0002, "Y": 0.0033 },
    "110": { "A": 0.0112, "C": 0.003, "D": 0.0038, "E": 0.0017, "F": 0.0159, "G": 0.0202, "H": 0.0019, "I": 0.001, "K": 0.0004, "L": 0.0644, "M": 0.0001, "N": 0.0054, "P": 0.0028, "Q": 0.0008, "R": 0.006, "S": 0.1109, "T": 0.1006, "V": 0.0022, "W": 0.5756, "Y": 0.0722 },
    "111": { "A": 0.0125, "C": 0.0033, "D": 0.0017, "E": 0.0008, "F": 0.0008, "G": 0.0134, "H": 0.0142, "I": 0.005, "L": 0.0175, "M": 0.0025, "N": 0.0025, "P": 0.7118, "Q": 0.0067, "R": 0.0251, "S": 0.0568, "T": 0.0309, "V": 0.0334, "W": 0.0292, "Y": 0.0317 },
    "111A": { "A": 0.1044, "C": 0.004, "D": 0.012, "E": 0.004, "F": 0.0161, "G": 0.0522, "H": 0.0241, "I": 0.0402, "K": 0.0161, "L": 0.1566, "M": 0.004, "P": 0.2771, "Q": 0.0281, "R": 0.1526, "S": 0.0241, "T": 0.0161, "V": 0.0241, "W": 0.0161, "Y": 0.0281 },
    "112B": { "A": 0.0604, "C": 0.0067, "D": 0.0671, "E": 0.0134, "F": 0.094, "G": 0.1409, "L": 0.0805, "N": 0.0336, "P": 0.0671, "Q": 0.0201, "R": 0.0671, "S": 0.0604, "T": 0.1342, "V": 0.0336, "W": 0.0201, "Y": 0.1007 },
    "112A": { "A": 0.0451, "C": 0.0021, "D": 0.0064, "E": 0.0429, "F": 0.015, "G": 0.0579, "H": 0.015, "I": 0.0107, "K": 0.0107, "L": 0.0579, "M": 0.0086, "N": 0.0665, "P": 0.3412, "Q": 0.0107, "R": 0.0773, "S": 0.0515, "T": 0.0665, "V": 0.03, "Y": 0.0837 },
    "112": { "A": 0.0082, "C": 0.0008, "D": 0.008, "E": 0.0032, "F": 0.003, "G": 0.0398, "H": 0.0328, "I": 0.004, "K": 0.003, "L": 0.0237, "M": 0.0007, "N": 0.0035, "P": 0.7281, "Q": 0.0135, "R": 0.0207, "S": 0.0748, "T": 0.0143, "V": 0.006, "W": 0.0037, "Y": 0.0083 },
    "113": { "A": 0.0149, "C": 0.0028, "D": 0.0018, "E": 0.0005, "F": 0.0154, "G": 0.0055, "H": 0.0044, "I": 0.0025, "K": 0.0005, "L": 0.0533, "M": 0.0003, "N": 0.0048, "P": 0.1835, "Q": 0.0012, "R": 0.0088, "S": 0.1362, "T": 0.1763, "V": 0.0024, "W": 0.3137, "Y": 0.0713 },
    "114": { "A": 0.0254, "C": 0.0019, "D": 0.0035, "E": 0.0015, "F": 0.0379, "G": 0.0062, "H": 0.0073, "I": 0.0074, "K": 0.0014, "L": 0.07, "M": 0.0006, "N": 0.0332, "P": 0.2277, "Q": 0.0117, "R": 0.0135, "S": 0.1491, "T": 0.201, "V": 0.0047, "W": 0.0929, "Y": 0.1032 },
    "115": { "A": 0.0194, "C": 0.0006, "D": 0.0015, "E": 0.0082, "F": 0.0063, "G": 0.0255, "H": 0.004, "I": 0.0053, "K": 0.0042, "L": 0.0745, "M": 0.0181, "N": 0.0034, "P": 0.6118, "Q": 0.0148, "R": 0.0382, "S": 0.0768, "T": 0.0388, "V": 0.007, "W": 0.0247, "Y": 0.017 },
    "116": { "A": 0.0078, "C": 0.0132, "D": 0.0062, "E": 0.0062, "F": 0.0753, "G": 0.0324, "H": 0.0143, "I": 0.0687, "K": 0.0063, "L": 0.2077, "M": 0.0012, "N": 0.0082, "P": 0.0814, "Q": 0.0212, "R": 0.103, "S": 0.0185, "T": 0.0058, "V": 0.0303, "W": 0.1459, "Y": 0.1464 },
    "117": { "A": 0.0208, "C": 0.0003, "D": 0.0011, "E": 0.0006, "F": 0.0025, "G": 0.0049, "H": 0.0017, "I": 0.0029, "K": 0.0009, "L": 0.0077, "M": 0.0009, "N": 0.0026, "P": 0.0094, "Q": 0.0007, "R": 0.0036, "S": 0.0407, "T": 0.8953, "V": 0.0024, "W": 0.0003, "Y": 0.0006 },
    "118": { "A": 0.0001, "C": 0.0069, "D": 0.0001, "F": 0.9787, "G": 0.0004, "H": 0.0001, "I": 0.0006, "L": 0.0044, "N": 0.0001, "P": 0.0005, "Q": 0.0001, "R": 0.0004, "S": 0.0018, "T": 0.0001, "V": 0.0045, "W": 0.0002, "Y": 0.0008 },
    "119": { "A": 0.0019, "C": 0.0006, "D": 0.0016, "G": 0.9914, "H": 0.0001, "I": 0.0001, "L": 0.0001, "P": 0.0002, "R": 0.0011, "S": 0.0017, "T": 0.0001, "V": 0.0009, "W": 0.0002 },
    "120": { "A": 0.0008, "C": 0.0001, "D": 0.0001, "E": 0.0029, "G": 0.2406, "H": 0.0082, "K": 0.0013, "L": 0.0016, "N": 0.0001, "P": 0.1159, "Q": 0.622, "R": 0.0058, "S": 0.0002, "T": 0.0002, "V": 0.0002, "Y": 0.0001 },
    "121": { "A": 0.0005, "E": 0.0031, "G": 0.9895, "R": 0.003, "V": 0.0011, "W": 0.0026 },
    "122": { "A": 0.0022, "D": 0.0001, "I": 0.0012, "K": 0.0002, "N": 0.0008, "P": 0.0018, "R": 0.0001, "S": 0.0024, "T": 0.9911 },
    "123": { "A": 0.0001, "D": 0.0001, "E": 0.0019, "G": 0.0001, "H": 0.0002, "I": 0.0003, "K": 0.8742, "L": 0.0006, "M": 0.0008, "N": 0.0036, "P": 0.0002, "Q": 0.0048, "R": 0.1082, "S": 0.0003, "T": 0.0045, "V": 0.0001, "Y": 0.0001 },
    "124": { "A": 0.0007, "E": 0.0004, "G": 0.0006, "I": 0.0005, "L": 0.3274, "M": 0.0026, "P": 0.0003, "Q": 0.0001, "R": 0.0003, "V": 0.6669, "W": 0.0001 },
    "125": { "A": 0.0012, "D": 0.1218, "E": 0.8664, "G": 0.0029, "H": 0.0006, "I": 0.0001, "K": 0.0017, "N": 0.0006, "Q": 0.0031, "R": 0.0004, "S": 0.0001, "T": 0.0001, "V": 0.0008, "Y": 0.0002 },
    "126": { "A": 0.0001, "D": 0.0001, "F": 0.003, "I": 0.9656, "L": 0.0046, "M": 0.0089, "N": 0.0009, "R": 0.0001, "S": 0.0045, "T": 0.0022, "V": 0.01, "Y": 0.0001 },
    "127": { "E": 0.0024, "K": 0.9938, "N": 0.0008, "Q": 0.003 }
  },
  "IGHV1": {
    "1": { "D": 0.0001, "E": 0.008, "H": 0.0051, "Q": 0.9867 },
    "2": { "A": 0.0105, "C": 0.0001, "D": 0.0018, "E": 0.0034, "F": 0.0029, "G": 0.0059, "I": 0.0044, "K": 0.0001, "L": 0.007, "M": 0.0199, "N": 0.0001, "P": 0.0003, "Q": 0.0002, "R": 0.0003, "S": 0.0004, "T": 0.0003, "V": 0.9424, "W": 0.0002 },
    "3": { "A": 0.0003, "D": 0.0004, "E": 0.0025, "F": 0.0001, "H": 0.0247, "K": 0.0027, "L": 0.0047, "N": 0.0008, "P": 0.0076, "Q": 0.9487, "R": 0.0058, "S": 0.0002, "T": 0.0003, "V": 0.0001, "W": 0.0001, "Y": 0.0008 },
    "4": { "C": 0.0001, "F": 0.0017, "G": 0.0006, "H": 0.0002, "I": 0.0006, "L": 0.9733, "M": 0.0035, "P": 0.0054, "Q": 0.0009, "R": 0.0039, "S": 0.0002, "V": 0.0095, "W": 0.0003 },
    "5": { "A": 0.0052, "C": 0.0001, "D": 0.0001, "E": 0.0062, "F": 0.0003, "G": 0.0057, "I": 0.0023, "K": 0.0002, "L": 0.0128, "M": 0.005, "Q": 0.0008, "R": 0.0002, "S": 0.0003, "T": 0.0005, "V": 0.96, "W": 0.0003 },
    "6": { "A": 0.0003, "D": 0.0001, "E": 0.0066, "G": 0.0002, "H": 0.0045, "K": 0.0019, "L": 0.0011, "P": 0.0073, "Q": 0.9747, "R": 0.0027, "S": 0.0003, "T": 0.0003, "V": 0.0001, "W": 0.0001 },
    "7": { "A": 0.005, "C": 0.0008, "F": 0.0014, "L": 0.0002, "P": 0.0076, "S": 0.9813, "T": 0.0013, "V": 0.0003, "Y": 0.0019 },
    "8": { "A": 0.0047, "C": 0.0001, "D": 0.0001, "E": 0.0041, "G": 0.9836, "L": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0022, "V": 0.0032, "W": 0.0016 },
    "9": { "A": 0.9278, "D": 0.004, "E": 0.0004, "F": 0.0001, "G": 0.0089, "H": 0.0002, "I": 0.0003, "L": 0.0001, "N": 0.0001, "P": 0.0316, "R": 0.0003, "S": 0.0082, "T": 0.0138, "V": 0.0042 },
    "11": { "A": 0.0245, "D": 0.0067, "E": 0.9562, "G": 0.0057, "K": 0.0009, "P": 0.0001, "Q": 0.002, "R": 0.0001, "S": 0.0001, "T": 0.0003, "V": 0.0031, "W": 0.0001, "Y": 0.0002 },
    "12": { "A": 0.007, "E": 0.003, "F": 0.0003, "G": 0.0115, "I": 0.0006, "K": 0.0001, "L": 0.0127, "M": 0.0086, "P": 0.0001, "Q": 0.0001, "R": 0.0004, "S": 0.0001, "T": 0.0002, "V": 0.955, "W": 0.0003 },
    "13": { "A": 0.0007, "D": 0.0001, "E": 0.005, "G": 0.0006, "H": 0.0005, "K": 0.9204, "L": 0.0003, "M": 0.0038, "N": 0.0039, "P": 0.0004, "Q": 0.0175, "R": 0.0242, "S": 0.0004, "T": 0.0219, "V": 0.0003, "W": 0.0002 },
    "14": { "A": 0.0005, "D": 0.0003, "E": 0.0126, "G": 0.0004, "H": 0.0003, "I": 0.0003, "K": 0.9215, "L": 0.0002, "M": 0.0024, "N": 0.0106, "P": 0.0006, "Q": 0.0135, "R": 0.0178, "S": 0.0014, "T": 0.0173, "V": 0.0002, "W": 0.0001, "Y": 0.0001 },
    "15": { "A": 0.0025, "C": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0048, "I": 0.0001, "L": 0.0021, "P": 0.9762, "Q": 0.0001, "R": 0.0011, "S": 0.0062, "T": 0.0065, "V": 0.0003 },
    "16": { "A": 0.0033, "C": 0.0001, "E": 0.0032, "G": 0.9824, "L": 0.0002, "M": 0.0001, "R": 0.0034, "S": 0.0002, "V": 0.004, "W": 0.0031 },
    "17": { "A": 0.6524, "C": 0.0006, "D": 0.004, "E": 0.0001, "F": 0.0004, "G": 0.0009, "H": 0.0001, "L": 0.0002, "P": 0.0051, "S": 0.3141, "T": 0.0189, "V": 0.002, "Y": 0.0011 },
    "18": { "A": 0.0102, "C": 0.0001, "E": 0.0005, "F": 0.0001, "H": 0.0001, "L": 0.0014, "P": 0.0072, "Q": 0.0004, "S": 0.9697, "T": 0.0088, "V": 0.0002, "W": 0.0005, "Y": 0.0007 },
    "19": { "A": 0.0052, "C": 0.0001, "E": 0.0013, "F": 0.0005, "G": 0.008, "I": 0.0009, "L": 0.0098, "M": 0.0053, "R": 0.0002, "S": 0.0001, "T": 0.0002, "V": 0.9679, "W": 0.0002 },
    "20": { "A": 0.0004, "E": 0.0037, "G": 0.0003, "H": 0.0006, "I": 0.0005, "K": 0.8947, "L": 0.0003, "M": 0.0043, "N": 0.0078, "P": 0.0004, "Q": 0.0194, "R": 0.0306, "S": 0.0007, "T": 0.0357, "V": 0.0003, "W": 0.0003, "Y": 0.0001 },
    "21": { "A": 0.0053, "C": 0.0005, "D": 0.0009, "F": 0.0055, "G": 0.0056, "H": 0.0001, "I": 0.0236, "L": 0.0117, "M": 0.0001, "P": 0.0001, "R": 0.0002, "S": 0.0005, "T": 0.0005, "V": 0.9454 },
    "22": { "A": 0.0088, "C": 0.0014, "D": 0.0001, "F": 0.0018, "H": 0.0006, "L": 0.0002, "P": 0.0106, "R": 0.0002, "S": 0.9678, "T": 0.0028, "V": 0.0001, "Y": 0.0055 },
    "23": { "A": 0.0002, "C": 0.9812, "F": 0.0034, "G": 0.006, "L": 0.0003, "R": 0.0032, "S": 0.0026, "V": 0.0003, "W": 0.001, "Y": 0.0016 },
    "24": { "A": 0.0006, "D": 0.0001, "E": 0.0103, "G": 0.0005, "H": 0.0004, "I": 0.0002, "K": 0.894, "L": 0.0003, "M": 0.0046, "N": 0.0052, "P": 0.0007, "Q": 0.0295, "R": 0.0198, "S": 0.0006, "T": 0.0327, "V": 0.0001, "Y": 0.0003 },
    "25": { "A": 0.8718, "C": 0.0001, "D": 0.0045, "E": 0.0007, "F": 0.0005, "G": 0.0036, "H": 0.0003, "I": 0.0019, "L": 0.0009, "P": 0.0093, "R": 0.0001, "S": 0.0107, "T": 0.0363, "V": 0.0591, "Y": 0.0001 },
    "26": { "A": 0.0106, "C": 0.0012, "D": 0.0001, "E": 0.0001, "F": 0.0047, "G": 0.0001, "H": 0.0007, "I": 0.0001, "L": 0.0011, "N": 0.0001, "P": 0.008, "Q": 0.0001, "S": 0.9633, "T": 0.0028, "V": 0.0003, "Y": 0.0065 },
    "27": { "A": 0.0067, "C": 0.0009, "D": 0.0025, "E": 0.0059, "G": 0.9687, "H": 0.0003, "I": 0.0001, "K": 0.0002, "L": 0.0001, "N": 0.0006, "Q": 0.0003, "R": 0.0055, "S": 0.0005, "T": 0.0004, "V": 0.0073, "W": 0.0001 },
    "28": { "A": 0.0039, "C": 0.0044, "D": 0.0157, "E": 0.0006, "F": 0.0219, "G": 0.2914, "H": 0.006, "I": 0.0005, "K": 0.0006, "L": 0.0006, "N": 0.0045, "P": 0.0003, "Q": 0.0001, "R": 0.0025, "S": 0.0176, "T": 0.0009, "V": 0.0048, "W": 0.0003, "Y": 0.6234 },
    "29": { "A": 0.0085, "C": 0.0002, "D": 0.0017, "F": 0.0007, "G": 0.0016, "H": 0.0015, "I": 0.0165, "K": 0.0011, "L": 0.0014, "M": 0.0016, "N": 0.0127, "P": 0.0331, "Q": 0.0001, "R": 0.0041, "S": 0.035, "T": 0.8786, "V": 0.0014, "W": 0.0001, "Y": 0.0004 },
    "30": { "A": 0.0003, "C": 0.0052, "D": 0.0001, "F": 0.9008, "G": 0.0003, "H": 0.0005, "I": 0.0029, "L": 0.0695, "P": 0.0011, "R": 0.0006, "S": 0.0071, "T": 0.0004, "V": 0.0089, "W": 0.0001, "Y": 0.0023 },
    "35": { "A": 0.007, "C": 0.0007, "D": 0.0026, "E": 0.0003, "F": 0.0006, "G": 0.006, "H": 0.0009, "I": 0.0172, "K": 0.0034, "L": 0.0014, "M": 0.001, "N": 0.0213, "P": 0.022, "Q": 0.0001, "R": 0.0195, "S": 0.314, "T": 0.5787, "V": 0.0021, "Y": 0.0011 },
    "36": { "A": 0.018, "C": 0.0016, "D": 0.047, "E": 0.0422, "F": 0.0008, "G": 0.1551, "H": 0.0035, "I": 0.0051, "K": 0.0066, "L": 0.0007, "M": 0.0003, "N": 0.0505, "P": 0.0013, "Q": 0.0016, "R": 0.0257, "S": 0.5939, "T": 0.0399, "V": 0.0027, "Y": 0.0036 },
    "37": { "A": 0.0006, "C": 0.0024, "D": 0.0082, "E": 0.0006, "F": 0.0264, "G": 0.0002, "H": 0.0282, "I": 0.0007, "K": 0.0008, "L": 0.0466, "M": 0.0001, "N": 0.0116, "P": 0.0014, "Q": 0.0021, "R": 0.002, "S": 0.0414, "T": 0.001, "V": 0.0013, "W": 0.0003, "Y": 0.8239 },
    "38": { "A": 0.3264, "C": 0.0017, "D": 0.0838, "E": 0.003, "F": 0.0095, "G": 0.1724, "H": 0.0089, "I": 0.0016, "K": 0.0003, "L": 0.0017, "M": 0.0001, "N": 0.0044, "P": 0.0095, "Q": 0.0008, "R": 0.0011, "S": 0.0642, "T": 0.0381, "V": 0.0082, "W": 0.0014, "Y": 0.2628 },
    "39": { "A": 0.0003, "C": 0.0002, "F": 0.0166, "G": 0.0004, "H": 0.0006, "I": 0.5823, "K": 0.0006, "L": 0.0426, "M": 0.3045, "N": 0.0027, "P": 0.0004, "Q": 0.0003, "R": 0.0023, "S": 0.0054, "T": 0.0079, "V": 0.0321, "W": 0.0001, "Y": 0.0008 },
    "40": { "A": 0.0026, "C": 0.0009, "D": 0.0023, "E": 0.0001, "F": 0.0009, "G": 0.0046, "H": 0.3796, "I": 0.0043, "K": 0.0008, "L": 0.0018, "M": 0.0002, "N": 0.1015, "P": 0.0161, "Q": 0.0184, "R": 0.0149, "S": 0.4192, "T": 0.025, "V": 0.0007, "Y": 0.0059 },
    "41": { "A": 0.0001, "C": 0.005, "F": 0.0002, "G": 0.0105, "L": 0.0037, "P": 0.0001, "Q": 0.0001, "R": 0.0078, "S": 0.0018, "V": 0.0003, "W": 0.9703, "Y": 0.0001 },
    "42": { "A": 0.0053, "C": 0.0001, "E": 0.001, "F": 0.0016, "G": 0.013, "I": 0.0027, "K": 0.0001, "L": 0.0208, "M": 0.0055, "P": 0.0003, "R": 0.0004, "S": 0.0003, "V": 0.9481, "W": 0.0008 },
    "43": { "A": 0.0003, "C": 0.0002, "D": 0.0001, "G": 0.0036, "H": 0.0004, "K": 0.0008, "L": 0.0034, "P": 0.0028, "Q": 0.0082, "R": 0.979, "S": 0.0005, "T": 0.0003, "V": 0.0003 },
    "44": { "A": 0.0001, "E": 0.0011, "G": 0.0002, "H": 0.0037, "K": 0.0047, "L": 0.0043, "M": 0.0001, "P": 0.0144, "Q": 0.9608, "R": 0.009, "S": 0.0002, "T": 0.0011, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "45": { "A": 0.9639, "D": 0.0047, "F": 0.0001, "G": 0.0053, "H": 0.0002, "I": 0.0002, "L": 0.0001, "P": 0.0051, "R": 0.0002, "S": 0.0055, "T": 0.0054, "V": 0.009, "Y": 0.0001 },
    "46": { "A": 0.005, "D": 0.0001, "G": 0.0001, "H": 0.0038, "I": 0.0002, "L": 0.0019, "M": 0.0001, "N": 0.0003, "P": 0.8907, "Q": 0.0004, "R": 0.0159, "S": 0.0036, "T": 0.0777 },
    "47": { "A": 0.0015, "D": 0.0002, "E": 0.0016, "G": 0.9894, "K": 0.0001, "N": 0.0002, "P": 0.0002, "R": 0.0035, "S": 0.0001, "T": 0.0002, "V": 0.0023, "W": 0.0006 },
    "48": { "D": 0.0001, "E": 0.0046, "G": 0.0004, "H": 0.014, "I": 0.0001, "K": 0.0623, "L": 0.0022, "N": 0.0009, "P": 0.0087, "Q": 0.8992, "R": 0.0056, "S": 0.0002, "T": 0.0015, "V": 0.0001 },
    "49": { "A": 0.0033, "C": 0.0001, "D": 0.0002, "E": 0.0017, "G": 0.9072, "H": 0.0001, "I": 0.0001, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0002, "P": 0.0004, "Q": 0.0002, "R": 0.0758, "S": 0.0037, "T": 0.0008, "V": 0.0033, "W": 0.0024 },
    "50": { "A": 0.0012, "C": 0.0001, "F": 0.007, "G": 0.0003, "H": 0.0014, "I": 0.0038, "L": 0.9543, "M": 0.0001, "P": 0.0205, "Q": 0.0002, "R": 0.0084, "S": 0.0003, "T": 0.0003, "V": 0.0021 },
    "51": { "A": 0.0121, "D": 0.006, "E": 0.956, "G": 0.0116, "H": 0.0003, "K": 0.0012, "L": 0.0001, "M": 0.0001, "Q": 0.0076, "R": 0.0006, "T": 0.0001, "V": 0.0028, "W": 0.0011, "Y": 0.0002 },
    "52": { "A": 0.0001, "C": 0.0039, "F": 0.0004, "G": 0.0115, "L": 0.004, "R": 0.003, "S": 0.0008, "V": 0.0011, "W": 0.9725, "Y": 0.0026 },
    "53": { "A": 0.0001, "C": 0.0001, "D": 0.0002, "E": 0.0002, "G": 0.0013, "H": 0.0001, "I": 0.022, "K": 0.0022, "L": 0.0259, "M": 0.908, "N": 0.0002, "Q": 0.0001, "R": 0.0113, "S": 0.0007, "T": 0.0022, "V": 0.0252, "W": 0.0001 },
    "54": { "A": 0.0077, "C": 0.0001, "D": 0.0003, "E": 0.0019, "G": 0.9847, "R": 0.0021, "S": 0.0002, "V": 0.0028, "W": 0.0001 },
    "55": { "A": 0.0026, "C": 0.0031, "D": 0.0006, "E": 0.0014, "F": 0.001, "G": 0.2792, "H": 0.0003, "I": 0.0808, "K": 0.0016, "L": 0.0127, "M": 0.0032, "N": 0.0007, "P": 0.0004, "Q": 0.0009, "R": 0.1238, "S": 0.0038, "T": 0.0019, "V": 0.008, "W": 0.4722, "Y": 0.0019 },
    "56": { "A": 0.0005, "C": 0.0011, "D": 0.0006, "F": 0.0515, "G": 0.0015, "H": 0.0007, "I": 0.7923, "K": 0.0006, "L": 0.013, "M": 0.0805, "N": 0.0039, "P": 0.0002, "R": 0.0017, "S": 0.0121, "T": 0.0075, "V": 0.0313, "W": 0.0001, "Y": 0.0008 },
    "57": { "A": 0.002, "C": 0.0005, "D": 0.0621, "E": 0.0022, "F": 0.0024, "G": 0.0045, "H": 0.0074, "I": 0.2916, "K": 0.0085, "L": 0.0053, "M": 0.0016, "N": 0.397, "P": 0.0006, "Q": 0.0009, "R": 0.0057, "S": 0.1706, "T": 0.0142, "V": 0.02, "Y": 0.0027 },
    "58": { "A": 0.1883, "C": 0.0011, "D": 0.0016, "E": 0.0003, "F": 0.0007, "G": 0.0141, "H": 0.0041, "I": 0.002, "L": 0.0045, "N": 0.0009, "P": 0.7269, "Q": 0.0005, "R": 0.0035, "S": 0.0073, "T": 0.0226, "V": 0.0213, "Y": 0.0002 },
    "59": { "A": 0.0073, "C": 0.0011, "D": 0.0174, "E": 0.054, "F": 0.0082, "G": 0.077, "H": 0.0086, "I": 0.2715, "K": 0.0182, "L": 0.0102, "M": 0.0091, "N": 0.2359, "P": 0.0015, "Q": 0.0019, "R": 0.007, "S": 0.0962, "T": 0.0089, "V": 0.0111, "Y": 0.1548 },
    "62": { "A": 0.0022, "C": 0.0045, "D": 0.0604, "E": 0.0031, "F": 0.2144, "G": 0.0886, "H": 0.0047, "I": 0.006, "K": 0.006, "L": 0.0799, "M": 0.0008, "N": 0.2213, "P": 0.0014, "Q": 0.0006, "R": 0.0122, "S": 0.2585, "T": 0.021, "V": 0.0097, "W": 0.0002, "Y": 0.0045 },
    "63": { "A": 0.0087, "C": 0.0045, "D": 0.0215, "E": 0.0037, "F": 0.0002, "G": 0.9298, "H": 0.0013, "I": 0.0005, "K": 0.0008, "L": 0.0006, "N": 0.0069, "P": 0.0008, "Q": 0.0005, "R": 0.0047, "S": 0.0073, "T": 0.0017, "V": 0.0051, "W": 0.0003, "Y": 0.0011 },
    "64": { "A": 0.0145, "C": 0.0008, "D": 0.0389, "E": 0.0556, "F": 0.0013, "G": 0.1595, "H": 0.006, "I": 0.0718, "K": 0.0192, "L": 0.0025, "M": 0.0017, "N": 0.2835, "P": 0.0022, "Q": 0.0022, "R": 0.0118, "S": 0.085, "T": 0.2275, "V": 0.0101, "W": 0.0001, "Y": 0.0059 },
    "65": { "A": 0.285, "D": 0.0002, "E": 0.0028, "F": 0.0001, "G": 0.0024, "H": 0.0001, "I": 0.0077, "K": 0.0081, "L": 0.0006, "M": 0.0005, "N": 0.0006, "P": 0.0173, "Q": 0.0008, "R": 0.0059, "S": 0.0096, "T": 0.6523, "V": 0.0061 },
    "66": { "A": 0.0041, "C": 0.0002, "D": 0.02, "E": 0.0053, "F": 0.0014, "G": 0.0794, "H": 0.0123, "I": 0.0549, "K": 0.0897, "L": 0.0026, "M": 0.001, "N": 0.5957, "P": 0.0004, "Q": 0.0028, "R": 0.0111, "S": 0.0795, "T": 0.0261, "V": 0.0069, "W": 0.0002, "Y": 0.0064 },
    "67": { "A": 0.0006, "C": 0.0029, "D": 0.0165, "E": 0.001, "F": 0.0163, "H": 0.0089, "I": 0.0017, "K": 0.0006, "L": 0.0033, "M": 0.0001, "N": 0.0061, "P": 0.0012, "Q": 0.0006, "R": 0.0004, "S": 0.0177, "T": 0.0013, "V": 0.0015, "W": 0.0003, "Y": 0.919 },
    "68": { "A": 0.8763, "D": 0.001, "E": 0.0165, "F": 0.0003, "G": 0.0085, "H": 0.0001, "I": 0.0007, "K": 0.0005, "L": 0.0013, "N": 0.0003, "P": 0.0063, "Q": 0.0003, "R": 0.0005, "S": 0.0734, "T": 0.0081, "V": 0.0059, "Y": 0.0002 },
    "69": { "A": 0.0006, "D": 0.0014, "E": 0.022, "G": 0.0007, "H": 0.0075, "I": 0.0001, "K": 0.013, "L": 0.0033, "N": 0.0003, "P": 0.0157, "Q": 0.9205, "R": 0.0121, "S": 0.0005, "T": 0.0009, "V": 0.0006, "W": 0.0005, "Y": 0.0002 },
    "70": { "A": 0.0011, "C": 0.0001, "D": 0.0023, "E": 0.0091, "G": 0.0012, "H": 0.0015, "I": 0.0015, "K": 0.8906, "L": 0.0002, "M": 0.0022, "N": 0.0367, "P": 0.0003, "Q": 0.0065, "R": 0.0294, "S": 0.007, "T": 0.0096, "V": 0.0004, "Y": 0.0003 },
    "71": { "A": 0.0001, "C": 0.0149, "D": 0.0001, "F": 0.8009, "G": 0.0005, "H": 0.0008, "I": 0.0039, "L": 0.1461, "M": 0.0008, "P": 0.0009, "R": 0.0015, "S": 0.0036, "T": 0.0002, "V": 0.0211, "W": 0.0004, "Y": 0.0043 },
    "72": { "A": 0.0003, "D": 0.0002, "E": 0.0075, "F": 0.0001, "G": 0.0003, "H": 0.0073, "K": 0.0121, "L": 0.0045, "M": 0.0002, "N": 0.0003, "P": 0.0035, "Q": 0.945, "R": 0.0169, "S": 0.0002, "T": 0.0007, "V": 0.0003, "W": 0.0004, "Y": 0.0003 },
    "74": { "A": 0.0052, "C": 0.0032, "D": 0.0371, "E": 0.0134, "F": 0.0001, "G": 0.9293, "H": 0.0004, "I": 0.0001, "K": 0.0002, "N": 0.0004, "Q": 0.0002, "R": 0.0015, "S": 0.0042, "T": 0.0003, "V": 0.0039, "W": 0.0001, "Y": 0.0003 },
    "75": { "A": 0.0002, "C": 0.0002, "E": 0.0003, "G": 0.0038, "H": 0.0002, "I": 0.0025, "K": 0.0028, "M": 0.0007, "N": 0.0002, "Q": 0.0003, "R": 0.959, "S": 0.0051, "T": 0.0018, "V": 0.0002, "W": 0.0228 },
    "76": { "A": 0.0039, "C": 0.0005, "D": 0.0016, "E": 0.0001, "F": 0.0062, "G": 0.0114, "H": 0.0001, "I": 0.0154, "L": 0.0262, "M": 0.0004, "P": 0.0002, "R": 0.0004, "S": 0.0004, "T": 0.0002, "V": 0.9331 },
    "77": { "A": 0.008, "D": 0.0001, "F": 0.0007, "G": 0.0001, "H": 0.0006, "I": 0.0096, "K": 0.0024, "L": 0.0006, "M": 0.0012, "N": 0.0039, "P": 0.003, "R": 0.002, "S": 0.0176, "T": 0.9481, "V": 0.0017, "W": 0.0001, "Y": 0.0002 },
    "78": { "A": 0.0003, "D": 0.0001, "F": 0.0074, "G": 0.0003, "I": 0.3898, "K": 0.0006, "L": 0.0321, "M": 0.5448, "N": 0.0002, "P": 0.0001, "Q": 0.0001, "R": 0.0048, "S": 0.0033, "T": 0.002, "V": 0.0139, "W": 0.0002 },
    "79": { "A": 0.0057, "D": 0.0002, "F": 0.0003, "G": 0.0003, "I": 0.0048, "K": 0.0003, "L": 0.0001, "M": 0.0003, "N": 0.0042, "P": 0.0041, "R": 0.0005, "S": 0.0153, "T": 0.9627, "V": 0.0007, "Y": 0.0002 },
    "80": { "A": 0.3183, "C": 0.0001, "D": 0.0013, "E": 0.048, "F": 0.0002, "G": 0.0064, "H": 0.0001, "I": 0.002, "K": 0.0031, "L": 0.0007, "M": 0.0031, "N": 0.0002, "P": 0.0018, "Q": 0.0003, "R": 0.4196, "S": 0.0128, "T": 0.1755, "V": 0.0021, "W": 0.0041 },
    "81": { "A": 0.0047, "C": 0.0001, "D": 0.8971, "E": 0.0125, "F": 0.0001, "G": 0.0034, "H": 0.001, "I": 0.0004, "K": 0.0006, "L": 0.0001, "N": 0.072, "P": 0.0001, "Q": 0.0001, "R": 0.0002, "S": 0.0018, "T": 0.0008, "V": 0.0021, "Y": 0.003 },
    "82": { "A": 0.0088, "D": 0.0036, "E": 0.1876, "G": 0.0017, "H": 0.0002, "I": 0.0046, "K": 0.1166, "L": 0.001, "M": 0.0165, "N": 0.002, "P": 0.0048, "Q": 0.0011, "R": 0.0077, "S": 0.0055, "T": 0.6369, "V": 0.0012 },
    "83": { "A": 0.0175, "C": 0.0006, "D": 0.0005, "F": 0.0048, "G": 0.0004, "H": 0.0004, "I": 0.0003, "K": 0.0002, "L": 0.0017, "N": 0.0003, "P": 0.0077, "S": 0.9564, "T": 0.0051, "V": 0.0007, "W": 0.0001, "Y": 0.0034 },
    "84": { "A": 0.0707, "D": 0.0002, "E": 0.0005, "F": 0.0006, "G": 0.0008, "H": 0.0003, "I": 0.249, "K": 0.0033, "L": 0.0041, "M": 0.0045, "N": 0.0013, "P": 0.0024, "Q": 0.0004, "R": 0.0032, "S": 0.0095, "T": 0.6444, "V": 0.0049 },
    "85": { "A": 0.0054, "C": 0.0008, "D": 0.0576, "E": 0.0017, "F": 0.0002, "G": 0.0146, "H": 0.0008, "I": 0.0025, "K": 0.003, "L": 0.0001, "M": 0.0001, "N": 0.048, "P": 0.0003, "Q": 0.0003, "R": 0.0199, "S": 0.803, "T": 0.0403, "V": 0.0004, "Y": 0.001 },
    "86": { "A": 0.0056, "E": 0.0001, "F": 0.0001, "I": 0.005, "K": 0.0039, "L": 0.0003, "M": 0.0006, "N": 0.0002, "P": 0.0034, "Q": 0.0002, "R": 0.001, "S": 0.0036, "T": 0.9754, "V": 0.0007 },
    "87": { "A": 0.8292, "C": 0.0002, "D": 0.0032, "E": 0.0003, "F": 0.0021, "G": 0.0077, "H": 0.0002, "I": 0.0028, "L": 0.0022, "M": 0.0002, "P": 0.0003, "R": 0.0001, "S": 0.0061, "T": 0.0083, "V": 0.1368, "Y": 0.0002 },
    "88": { "A": 0.0001, "C": 0.0019, "D": 0.0078, "E": 0.0002, "F": 0.0138, "H": 0.0083, "I": 0.0002, "L": 0.0001, "N": 0.0019, "P": 0.0001, "Q": 0.0003, "S": 0.005, "T": 0.0002, "V": 0.0001, "W": 0.0001, "Y": 0.9601 },
    "89": { "F": 0.0003, "H": 0.0001, "I": 0.006, "K": 0.0005, "L": 0.029, "M": 0.9508, "P": 0.0001, "R": 0.007, "S": 0.0002, "T": 0.0008, "V": 0.0053, "W": 0.0001 },
    "90": { "A": 0.0042, "D": 0.0283, "E": 0.9398, "F": 0.0001, "G": 0.0129, "H": 0.0009, "I": 0.0004, "K": 0.0019, "N": 0.0008, "Q": 0.0053, "R": 0.0006, "S": 0.0003, "T": 0.0008, "V": 0.0026, "W": 0.0005, "Y": 0.0006 },
    "91": { "A": 0.0001, "F": 0.0011, "G": 0.0003, "I": 0.0027, "L": 0.9583, "M": 0.0107, "P": 0.0016, "Q": 0.0007, "R": 0.0078, "V": 0.0166, "W": 0.0003 },
    "92": { "A": 0.0008, "C": 0.0009, "D": 0.0014, "E": 0.0004, "F": 0.0004, "G": 0.0115, "H": 0.0018, "I": 0.0024, "K": 0.0038, "L": 0.0001, "M": 0.001, "N": 0.0188, "P": 0.0002, "R": 0.1924, "S": 0.7341, "T": 0.0281, "V": 0.0004, "W": 0.0007, "Y": 0.001 },
    "93": { "A": 0.0007, "C": 0.0006, "D": 0.002, "E": 0.0001, "F": 0.0001, "G": 0.0279, "H": 0.0001, "I": 0.0012, "K": 0.0006, "L": 0.0001, "M": 0.0004, "N": 0.013, "R": 0.1705, "S": 0.7759, "T": 0.0053, "V": 0.0002, "W": 0.0008, "Y": 0.0004 },
    "94": { "F": 0.0001, "I": 0.0002, "L": 0.9846, "M": 0.0039, "P": 0.0015, "Q": 0.0009, "R": 0.0062, "V": 0.0027 },
    "95": { "A": 0.0012, "D": 0.0002, "E": 0.0009, "G": 0.0126, "I": 0.0056, "K": 0.0209, "L": 0.0003, "M": 0.0008, "N": 0.002, "P": 0.0002, "Q": 0.0015, "R": 0.8782, "S": 0.0067, "T": 0.0683, "V": 0.0004 },
    "96": { "A": 0.0098, "C": 0.0011, "D": 0.0005, "E": 0.0001, "F": 0.0142, "G": 0.0002, "H": 0.0006, "I": 0.0006, "K": 0.0001, "L": 0.0015, "N": 0.0004, "P": 0.0127, "Q": 0.0001, "R": 0.0004, "S": 0.9469, "T": 0.0025, "V": 0.0011, "Y": 0.0071 },
    "97": { "A": 0.0058, "D": 0.3906, "E": 0.5883, "G": 0.0075, "H": 0.0005, "K": 0.0009, "N": 0.0025, "Q": 0.0014, "R": 0.0001, "S": 0.0003, "T": 0.0003, "V": 0.0011, "W": 0.0001, "Y": 0.0006 },
    "98": { "A": 0.002, "D": 0.9836, "E": 0.008, "G": 0.0025, "H": 0.0005, "N": 0.0009, "V": 0.0006, "Y": 0.0017 },
    "99": { "A": 0.006, "G": 0.0001, "K": 0.0015, "M": 0.0024, "P": 0.002, "Q": 0.0001, "R": 0.0013, "S": 0.0067, "T": 0.9795, "V": 0.0002 },
    "100": { "A": 0.9867, "C": 0.0001, "D": 0.0015, "G": 0.0055, "P": 0.0004, "S": 0.0025, "T": 0.0014, "V": 0.0017 },
    "101": { "A": 0.0021, "D": 0.0001, "E": 0.0009, "F": 0.0034, "G": 0.0046, "H": 0.0001, "I": 0.0281, "K": 0.0004, "L": 0.0175, "M": 0.0166, "Q": 0.0001, "R": 0.0004, "S": 0.0003, "T": 0.0043, "V": 0.9208, "W": 0.0001 },
    "102": { "C": 0.0008, "D": 0.0038, "E": 0.0001, "F": 0.0072, "H": 0.0008, "I": 0.0002, "L": 0.0003, "N": 0.0006, "S": 0.0006, "V": 0.0002, "W": 0.0001, "Y": 0.9854 },
    "103": { "C": 0.0016, "D": 0.0034, "F": 0.0383, "H": 0.0017, "L": 0.0003, "N": 0.0001, "R": 0.0001, "S": 0.0022, "T": 0.0002, "V": 0.0002, "W": 0.0004, "Y": 0.9515 },
    "104": { "C": 0.9823, "F": 0.0013, "G": 0.0062, "L": 0.0001, "R": 0.0012, "S": 0.0005, "V": 0.0002, "W": 0.0076, "Y": 0.0006 },
    "105": { "A": 0.95, "C": 0.0001, "D": 0.0001, "E": 0.0017, "F": 0.0003, "G": 0.005, "H": 0.0014, "I": 0.0003, "L": 0.0014, "M": 0.0007, "N": 0.0001, "P": 0.0003, "Q": 0.0002, "R": 0.0006, "S": 0.005, "T": 0.0119, "V": 0.02, "W": 0.0003, "Y": 0.0004 },
    "106": { "A": 0.0206, "C": 0.0011, "D": 0.0008, "E": 0.0013, "F": 0.0013, "G": 0.0168, "H": 0.0009, "I": 0.0106, "K": 0.0146, "L": 0.0077, "M": 0.0018, "N": 0.0024, "P": 0.0015, "Q": 0.0014, "R": 0.7772, "S": 0.0569, "T": 0.075, "V": 0.006, "W": 0.0008, "Y": 0.0012 },
    "107": { "A": 0.0886, "C": 0.0043, "D": 0.2467, "E": 0.0998, "F": 0.0099, "G": 0.215, "H": 0.0087, "I": 0.0115, "K": 0.0087, "L": 0.0291, "M": 0.0048, "N": 0.0117, "P": 0.0238, "Q": 0.0107, "R": 0.0328, "S": 0.0544, "T": 0.0287, "V": 0.0904, "W": 0.0077, "Y": 0.0127 },
    "108": { "A": 0.0527, "C": 0.0065, "D": 0.0458, "E": 0.0366, "F": 0.0218, "G": 0.1582, "H": 0.0238, "I": 0.0331, "K": 0.0242, "L": 0.0835, "M": 0.0115, "N": 0.0213, "P": 0.0907, "Q": 0.0285, "R": 0.1151, "S": 0.0825, "T": 0.0426, "V": 0.0541, "W": 0.018, "Y": 0.0495 },
    "109": { "A": 0.06, "C": 0.0158, "D": 0.0608, "E": 0.035, "F": 0.0216, "G": 0.1739, "H": 0.02, "I": 0.0402, "K": 0.0162, "L": 0.0529, "M": 0.0113, "N": 0.0215, "P": 0.0564, "Q": 0.019, "R": 0.0767, "S": 0.0994, "T": 0.0501, "V": 0.0622, "W": 0.02, "Y": 0.087 },
    "110": { "A": 0.0621, "C": 0.0191, "D": 0.0663, "E": 0.0279, "F": 0.0228, "G": 0.1511, "H": 0.0143, "I": 0.0434, "K": 0.0105, "L": 0.0458, "M": 0.0127, "N": 0.0243, "P": 0.046, "Q": 0.0209, "R": 0.0544, "S": 0.1209, "T": 0.0561, "V": 0.0673, "W": 0.0231, "Y": 0.111 },
    "111": { "A": 0.0738, "C": 0.0212, "D": 0.0626, "E": 0.0207, "F": 0.026, "G": 0.1509, "H": 0.0107, "I": 0.036, "K": 0.0082, "L": 0.0439, "M": 0.0123, "N": 0.0228, "P": 0.0308, "Q": 0.0149, "R": 0.0408, "S": 0.1467, "T": 0.0583, "V": 0.0706, "W": 0.0264, "Y": 0.1224 },
    "111A": { "A": 0.0689, "C": 0.0201, "D": 0.0688, "E": 0.0208, "F": 0.0284, "G": 0.1521, "H": 0.0105, "I": 0.0292, "K": 0.0089, "L": 0.0415, "M": 0.0117, "N": 0.0235, "P": 0.0302, "Q": 0.0126, "R": 0.0444, "S": 0.1514, "T": 0.0636, "V": 0.0757, "W": 0.0327, "Y": 0.1052 },
    "111B": { "A": 0.0643, "C": 0.0186, "D": 0.0665, "E": 0.0191, "F": 0.0328, "G": 0.1515, "H": 0.0085, "I": 0.0335, "K": 0.0058, "L": 0.0393, "M": 0.0091, "N": 0.0251, "P": 0.0322, "Q": 0.0137, "R": 0.0379, "S": 0.1746, "T": 0.0593, "V": 0.0773, "W": 0.0386, "Y": 0.0922 },
    "111C": { "A": 0.0598, "C": 0.0343, "D": 0.0557, "E": 0.0166, "F": 0.0292, "G": 0.1718, "H": 0.0078, "I": 0.0273, "K": 0.0078, "L": 0.0419, "M": 0.0103, "N": 0.0199, "P": 0.0399, "Q": 0.0064, "R": 0.0351, "S": 0.1744, "T": 0.0657, "V": 0.0734, "W": 0.0423, "Y": 0.0803 },
    "111D": { "A": 0.0678, "C": 0.036, "D": 0.0456, "E": 0.0184, "F": 0.0318, "G": 0.1493, "H": 0.0071, "I": 0.0272, "K": 0.0075, "L": 0.0406, "M": 0.005, "N": 0.0159, "P": 0.0506, "Q": 0.0088, "R": 0.0402, "S": 0.1723, "T": 0.0627, "V": 0.0728, "W": 0.0364, "Y": 0.1041 },
    "111E": { "A": 0.0776, "C": 0.0706, "D": 0.0398, "E": 0.0149, "F": 0.0239, "G": 0.1493, "H": 0.01, "I": 0.0438, "K": 0.0119, "L": 0.0318, "M": 0.006, "N": 0.009, "P": 0.0478, "Q": 0.0119, "R": 0.0259, "S": 0.1443, "T": 0.0577, "V": 0.0647, "W": 0.0289, "Y": 0.1303 },
    "111F": { "A": 0.0569, "C": 0.0596, "D": 0.0217, "E": 0.0271, "F": 0.0271, "G": 0.1192, "I": 0.0325, "K": 0.019, "L": 0.0379, "M": 0.0054, "N": 0.0108, "P": 0.0515, "Q": 0.0136, "R": 0.0542, "S": 0.1409, "T": 0.0705, "V": 0.0515, "W": 0.0407, "Y": 0.1599 },
    "111G": { "A": 0.027, "C": 0.0541, "D": 0.0631, "E": 0.036, "F": 0.027, "G": 0.1081, "H": 0.018, "I": 0.0631, "K": 0.009, "L": 0.045, "M": 0.009, "N": 0.018, "P": 0.0721, "Q": 0.009, "R": 0.027, "S": 0.1441, "T": 0.036, "V": 0.0631, "W": 0.036, "Y": 0.1351 },
    "112G": { "A": 0.0574, "C": 0.0335, "D": 0.0622, "E": 0.0239, "F": 0.0239, "G": 0.1292, "H": 0.0048, "I": 0.0478, "K": 0.0096, "L": 0.0622, "M": 0.0048, "N": 0.0239, "P": 0.0861, "Q": 0.0144, "R": 0.0622, "S": 0.0957, "T": 0.0335, "V": 0.0957, "W": 0.0144, "Y": 0.1148 },
    "112F": { "A": 0.0493, "C": 0.0411, "D": 0.0378, "E": 0.0247, "F": 0.0329, "G": 0.1447, "H": 0.0132, "I": 0.0559, "K": 0.0148, "L": 0.0461, "M": 0.0099, "N": 0.0197, "P": 0.0707, "Q": 0.0115, "R": 0.0543, "S": 0.1135, "T": 0.0411, "V": 0.0526, "W": 0.0263, "Y": 0.1398 },
    "112E": { "A": 0.073, "C": 0.0349, "D": 0.042, "E": 0.0291, "F": 0.0278, "G": 0.1396, "H": 0.0136, "I": 0.0265, "K": 0.0181, "L": 0.0452, "M": 0.0078, "N": 0.0226, "P": 0.0633, "Q": 0.0149, "R": 0.0562, "S": 0.1325, "T": 0.0679, "V": 0.0537, "W": 0.0226, "Y": 0.1086 },
    "112D": { "A": 0.0593, "C": 0.0322, "D": 0.0457, "E": 0.027, "F": 0.0222, "G": 0.1548, "H": 0.0219, "I": 0.0296, "K": 0.0144, "L": 0.0429, "M": 0.0078, "N": 0.0259, "P": 0.0624, "Q": 0.0118, "R": 0.0572, "S": 0.1401, "T": 0.0633, "V": 0.0469, "W": 0.0279, "Y": 0.1067 },
    "112C": { "A": 0.054, "C": 0.0291, "D": 0.0527, "E": 0.02, "F": 0.0224, "G": 0.1462, "H": 0.0192, "I": 0.0252, "K": 0.0138, "L": 0.0401, "M": 0.0065, "N": 0.0342, "P": 0.0579, "Q": 0.0123, "R": 0.0428, "S": 0.1282, "T": 0.0528, "V": 0.0536, "W": 0.0252, "Y": 0.1636 },
    "112B": { "A": 0.0523, "C": 0.0241, "D": 0.0531, "E": 0.0198, "F": 0.0209, "G": 0.1306, "H": 0.019, "I": 0.022, "K": 0.0105, "L": 0.0375, "M": 0.0075, "N": 0.0258, "P": 0.045, "Q": 0.0122, "R": 0.0404, "S": 0.1208, "T": 0.0523, "V": 0.0414, "W": 0.0232, "Y": 0.2414 },
    "112A": { "A": 0.0548, "C": 0.0178, "D": 0.0395, "E": 0.0207, "F": 0.0209, "G": 0.1301, "H": 0.0138, "I": 0.0243, "K": 0.0126, "L": 0.0369, "M": 0.0069, "N": 0.0216, "P": 0.0501, "Q": 0.0143, "R": 0.0474, "S": 0.1139, "T": 0.0474, "V": 0.0413, "W": 0.0304, "Y": 0.2552 },
    "112": { "A": 0.0508, "C": 0.0137, "D": 0.0417, "E": 0.0259, "F": 0.0203, "G": 0.1148, "H": 0.0235, "I": 0.0206, "K": 0.0149, "L": 0.0516, "M": 0.0086, "N": 0.032, "P": 0.0489, "Q": 0.0145, "R": 0.0436, "S": 0.103, "T": 0.0476, "V": 0.0391, "W": 0.03, "Y": 0.255 },
    "113": { "A": 0.0378, "C": 0.0111, "D": 0.0884, "E": 0.0217, "F": 0.0199, "G": 0.1425, "H": 0.0193, "I": 0.0102, "K": 0.009, "L": 0.0278, "M": 0.0057, "N": 0.0712, "P": 0.029, "Q": 0.0091, "R": 0.0395, "S": 0.0715, "T": 0.0338, "V": 0.0226, "W": 0.0424, "Y": 0.2876 },
    "114": { "A": 0.1367, "C": 0.0077, "D": 0.0261, "E": 0.0177, "F": 0.0179, "G": 0.2311, "H": 0.0171, "I": 0.0098, "K": 0.0045, "L": 0.0273, "M": 0.0023, "N": 0.017, "P": 0.0595, "Q": 0.0075, "R": 0.0214, "S": 0.0456, "T": 0.0267, "V": 0.0233, "W": 0.11, "Y": 0.1909 },
    "115": { "A": 0.0128, "C": 0.0054, "D": 0.0066, "E": 0.0064, "F": 0.4849, "G": 0.0328, "H": 0.0044, "I": 0.0244, "K": 0.0021, "L": 0.0813, "M": 0.2038, "N": 0.0075, "P": 0.0212, "Q": 0.0035, "R": 0.0061, "S": 0.0231, "T": 0.0111, "V": 0.0224, "W": 0.0093, "Y": 0.0307 },
    "116": { "A": 0.0122, "C": 0.0008, "D": 0.8404, "E": 0.0123, "F": 0.004, "G": 0.0388, "H": 0.007, "I": 0.0021, "K": 0.0023, "L": 0.0065, "M": 0.0005, "N": 0.0109, "P": 0.0048, "Q": 0.0172, "R": 0.0062, "S": 0.0089, "T": 0.0053, "V": 0.0097, "W": 0.0021, "Y": 0.008 },
    "117": { "A": 0.0027, "C": 0.0028, "D": 0.0117, "E": 0.002, "F": 0.022, "G": 0.0043, "H": 0.0284, "I": 0.1062, "K": 0.0005, "L": 0.034, "M": 0.0017, "N": 0.008, "P": 0.1276, "Q": 0.0018, "R": 0.0024, "S": 0.0226, "T": 0.0053, "V": 0.2221, "W": 0.0006, "Y": 0.3933 },
    "118": { "A": 0.0003, "C": 0.0012, "F": 0.0004, "G": 0.0019, "I": 0.0001, "L": 0.0015, "M": 0.0001, "P": 0.0001, "R": 0.0008, "S": 0.0003, "V": 0.0008, "W": 0.9924 },
    "119": { "A": 0.0015, "C": 0.0018, "D": 0.0014, "G": 0.9919, "P": 0.0001, "R": 0.0003, "S": 0.0017, "V": 0.0012 },
    "120": { "E": 0.0021, "G": 0.0001, "H": 0.0033, "I": 0.0001, "K": 0.0384, "L": 0.0033, "N": 0.0006, "P": 0.0049, "Q": 0.9159, "R": 0.0305, "T": 0.0007 },
    "121": { "A": 0.0002, "D": 0.0001, "E": 0.0008, "G": 0.9961, "R": 0.0008, "V": 0.0017, "W": 0.0002 },
    "122": { "A": 0.0067, "I": 0.0015, "K": 0.0003, "M": 0.0001, "N": 0.001, "P": 0.0017, "R": 0.0001, "S": 0.0086, "T": 0.9792, "V": 0.0007 },
    "123": { "A": 0.0028, "E": 0.0002, "F": 0.0002, "G": 0.0002, "H": 0.0004, "I": 0.0005, "K": 0.0019, "L": 0.6471, "M": 0.108, "N": 0.0001, "P": 0.0064, "Q": 0.007, "R": 0.0048, "S": 0.002, "T": 0.2128, "V": 0.0053, "W": 0.0003 },
    "124": { "A": 0.0015, "D": 0.0002, "F": 0.0006, "G": 0.0022, "I": 0.0081, "L": 0.004, "S": 0.0002, "V": 0.9832 },
    "125": { "A": 0.0048, "D": 0.0001, "F": 0.0002, "G": 0.0001, "H": 0.0004, "I": 0.0154, "K": 0.0001, "L": 0.0004, "N": 0.002, "P": 0.0009, "R": 0.0002, "S": 0.0094, "T": 0.9647, "V": 0.0011, "Y": 0.0001 },
    "126": { "A": 0.0024, "C": 0.0001, "D": 0.0004, "F": 0.0005, "G": 0.0023, "I": 0.003, "L": 0.0005, "R": 0.0006, "S": 0.0004, "V": 0.9898 },
    "127": { "A": 0.0048, "C": 0.0004, "D": 0.0002, "F": 0.002, "H": 0.0001, "L": 0.0012, "P": 0.0028, "S": 0.9842, "T": 0.0023, "V": 0.0002, "Y": 0.0017 },
    "128": { "A": 0.0046, "L": 0.0012, "P": 0.0039, "Q": 0.0005, "S": 0.9885, "T": 0.0013 }
  },
  "IGHV2": {
    "1": { "H": 0.007, "Q": 0.9927, "R": 0.0002 },
    "2": { "A": 0.0017, "C": 0.0001, "D": 0.0004, "F": 0.0029, "G": 0.0009, "H": 0.0002, "I": 0.6118, "L": 0.0101, "M": 0.0013, "N": 0.0016, "S": 0.002, "T": 0.0042, "V": 0.3627, "Y": 0.0001 },
    "3": { "A": 0.0029, "H": 0.0006, "I": 0.0013, "N": 0.0024, "P": 0.0114, "Q": 0.0003, "R": 0.0001, "S": 0.0044, "T": 0.9761, "V": 0.0003, "Y": 0.0001 },
    "4": { "C": 0.0001, "F": 0.007, "G": 0.0001, "L": 0.9815, "M": 0.0015, "P": 0.0001, "S": 0.0028, "V": 0.0039, "W": 0.0027 },
    "5": { "E": 0.0057, "G": 0.0003, "H": 0.0003, "I": 0.0003, "K": 0.8179, "L": 0.0002, "M": 0.0019, "N": 0.005, "P": 0.0005, "Q": 0.0125, "R": 0.1352, "S": 0.0014, "T": 0.0184, "W": 0.0005 },
    "6": { "A": 0.0142, "D": 0.0054, "E": 0.9713, "G": 0.0025, "K": 0.0008, "L": 0.0001, "Q": 0.0028, "R": 0.0001, "S": 0.0005, "T": 0.0002, "V": 0.0019, "W": 0.0001, "Y": 0.0002 },
    "7": { "A": 0.0038, "C": 0.0004, "D": 0.0002, "F": 0.0012, "H": 0.0003, "L": 0.0002, "P": 0.0053, "Q": 0.0001, "S": 0.983, "T": 0.0031, "V": 0.0001, "Y": 0.0021 },
    "8": { "A": 0.0047, "C": 0.0033, "D": 0.0018, "G": 0.9807, "P": 0.0001, "R": 0.0015, "S": 0.0029, "T": 0.0001, "V": 0.0046, "W": 0.0001 },
    "9": { "A": 0.001, "H": 0.0026, "L": 0.003, "P": 0.9869, "Q": 0.0003, "R": 0.001, "S": 0.0021, "T": 0.003 },
    "11": { "A": 0.1718, "E": 0.0005, "G": 0.0007, "I": 0.0004, "K": 0.0015, "L": 0.001, "M": 0.0017, "N": 0.0002, "P": 0.0098, "Q": 0.0003, "R": 0.0005, "S": 0.0066, "T": 0.6073, "V": 0.1975, "Y": 0.0001 },
    "12": { "A": 0.0006, "E": 0.0004, "F": 0.0003, "G": 0.0001, "I": 0.0002, "L": 0.964, "M": 0.003, "P": 0.0053, "Q": 0.0068, "R": 0.0084, "S": 0.0006, "T": 0.0002, "V": 0.01, "W": 0.0001 },
    "13": { "A": 0.0048, "D": 0.0001, "E": 0.0008, "F": 0.0001, "G": 0.003, "I": 0.0016, "L": 0.0137, "M": 0.005, "P": 0.0001, "R": 0.0002, "S": 0.0002, "V": 0.9699, "W": 0.0002 },
    "14": { "A": 0.0001, "D": 0.0002, "E": 0.0058, "G": 0.0001, "H": 0.0009, "I": 0.0029, "K": 0.9085, "L": 0.0003, "M": 0.0005, "N": 0.0215, "P": 0.0008, "Q": 0.0163, "R": 0.0166, "S": 0.0008, "T": 0.0244, "Y": 0.0003 },
    "15": { "A": 0.0044, "D": 0.0001, "G": 0.0001, "H": 0.0058, "L": 0.0036, "P": 0.976, "R": 0.0025, "S": 0.004, "T": 0.0031, "V": 0.0003 },
    "16": { "A": 0.0046, "H": 0.0003, "I": 0.0014, "K": 0.0042, "L": 0.0002, "M": 0.0004, "N": 0.0002, "P": 0.0101, "Q": 0.0004, "R": 0.0024, "S": 0.0034, "T": 0.9724 },
    "17": { "A": 0.004, "D": 0.002, "E": 0.2088, "G": 0.0004, "H": 0.0051, "K": 0.0019, "L": 0.0008, "P": 0.0091, "Q": 0.7657, "R": 0.0018, "T": 0.0002, "V": 0.0002 },
    "18": { "A": 0.0042, "D": 0.0002, "H": 0.0006, "I": 0.0012, "K": 0.0002, "N": 0.0037, "P": 0.0185, "Q": 0.0001, "S": 0.003, "T": 0.9682 },
    "19": { "A": 0.0001, "C": 0.0001, "F": 0.0018, "H": 0.0017, "I": 0.004, "L": 0.973, "P": 0.0105, "R": 0.004, "S": 0.0002, "T": 0.0003, "V": 0.0042 },
    "20": { "A": 0.0049, "H": 0.0002, "I": 0.0007, "K": 0.0032, "L": 0.0001, "M": 0.0026, "N": 0.0004, "P": 0.0263, "Q": 0.0008, "R": 0.002, "S": 0.0027, "T": 0.9559, "V": 0.0001 },
    "21": { "F": 0.0001, "L": 0.9824, "M": 0.0032, "P": 0.006, "Q": 0.0006, "R": 0.0027, "S": 0.0001, "T": 0.0002, "V": 0.0044, "W": 0.0001 },
    "22": { "A": 0.0031, "H": 0.0005, "I": 0.0013, "K": 0.0001, "L": 0.0001, "N": 0.003, "P": 0.0194, "R": 0.0001, "S": 0.0037, "T": 0.9685 },
    "23": { "A": 0.0002, "C": 0.976, "D": 0.0001, "F": 0.006, "G": 0.004, "L": 0.0002, "R": 0.0029, "S": 0.0075, "T": 0.0001, "V": 0.0001, "W": 0.0003, "Y": 0.0024 },
    "24": { "A": 0.0086, "D": 0.001, "F": 0.0002, "G": 0.0003, "H": 0.0011, "I": 0.0034, "K": 0.0002, "L": 0.0002, "M": 0.0002, "N": 0.0085, "P": 0.0365, "R": 0.0023, "S": 0.0383, "T": 0.8984, "V": 0.0006, "Y": 0.0001 },
    "25": { "A": 0.0025, "C": 0.004, "D": 0.0005, "F": 0.751, "G": 0.0007, "H": 0.0001, "I": 0.0029, "L": 0.0222, "P": 0.0006, "S": 0.0063, "V": 0.2068, "W": 0.0001, "Y": 0.0021 },
    "26": { "A": 0.0079, "C": 0.0005, "D": 0.0003, "F": 0.0016, "G": 0.0001, "H": 0.0005, "L": 0.0004, "P": 0.0365, "Q": 0.0002, "S": 0.9454, "T": 0.0035, "V": 0.0001, "Y": 0.0028 },
    "27": { "A": 0.0071, "C": 0.0006, "D": 0.0015, "E": 0.0032, "G": 0.9739, "Q": 0.0001, "R": 0.0036, "S": 0.0003, "V": 0.0054, "W": 0.004 },
    "28": { "A": 0.0003, "C": 0.0035, "D": 0.0001, "F": 0.9493, "G": 0.0003, "I": 0.0058, "L": 0.0216, "M": 0.0002, "P": 0.0003, "R": 0.0001, "S": 0.0087, "T": 0.0004, "V": 0.0079, "Y": 0.0015 },
    "29": { "A": 0.0075, "F": 0.0002, "H": 0.0002, "L": 0.0026, "P": 0.0218, "Q": 0.0007, "S": 0.9631, "T": 0.0027, "V": 0.0001, "Y": 0.0008 },
    "30": { "A": 0.0003, "C": 0.0002, "D": 0.0001, "F": 0.0139, "H": 0.0019, "I": 0.0096, "L": 0.9344, "M": 0.0004, "P": 0.0227, "R": 0.0066, "S": 0.0007, "T": 0.001, "V": 0.0082, "Y": 0.0001 },
    "31": { "A": 0.0033, "C": 0.0011, "D": 0.0054, "E": 0.0003, "F": 0.0013, "G": 0.0041, "H": 0.0018, "I": 0.0074, "K": 0.0029, "L": 0.0012, "M": 0.0006, "N": 0.0307, "P": 0.0028, "Q": 0.0002, "R": 0.0357, "S": 0.8236, "T": 0.0766, "V": 0.0003, "Y": 0.0008 },
    "34": { "A": 0.0139, "C": 0.0002, "D": 0.0054, "E": 0.0007, "F": 0.0003, "G": 0.002, "H": 0.0048, "I": 0.0032, "K": 0.0021, "L": 0.0006, "M": 0.0003, "N": 0.1837, "P": 0.0192, "Q": 0.0006, "R": 0.0016, "S": 0.0342, "T": 0.7256, "V": 0.0012, "Y": 0.0004 },
    "35": { "A": 0.1816, "C": 0.0015, "D": 0.0092, "E": 0.001, "F": 0.0015, "G": 0.0202, "H": 0.0041, "I": 0.0066, "K": 0.0009, "L": 0.0013, "M": 0.0002, "N": 0.0247, "P": 0.0116, "R": 0.0466, "S": 0.6396, "T": 0.0386, "V": 0.0042, "W": 0.0005, "Y": 0.0062 },
    "36": { "A": 0.0136, "C": 0.0005, "D": 0.0031, "E": 0.0126, "G": 0.7499, "H": 0.0007, "I": 0.0014, "K": 0.0048, "L": 0.0007, "M": 0.0002, "N": 0.0008, "P": 0.0029, "Q": 0.0017, "R": 0.1927, "S": 0.0057, "T": 0.005, "V": 0.0033, "W": 0.0001, "Y": 0.0001 },
    "37": { "A": 0.0073, "D": 0.0003, "E": 0.014, "F": 0.0004, "G": 0.0082, "I": 0.0037, "K": 0.0005, "L": 0.0304, "M": 0.3343, "N": 0.0002, "P": 0.0003, "Q": 0.0006, "R": 0.0036, "S": 0.0005, "T": 0.0034, "V": 0.5915, "W": 0.0007, "Y": 0.0003 },
    "38": { "A": 0.0295, "C": 0.12, "D": 0.0022, "E": 0.0001, "F": 0.0008, "G": 0.7834, "H": 0.0002, "I": 0.0001, "L": 0.0004, "N": 0.0013, "P": 0.0003, "R": 0.0356, "S": 0.0154, "T": 0.0017, "V": 0.0072, "W": 0.0007, "Y": 0.001 },
    "39": { "A": 0.0039, "C": 0.0002, "D": 0.0001, "E": 0.0013, "F": 0.0002, "G": 0.0089, "I": 0.0025, "L": 0.0111, "M": 0.0037, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.9665, "W": 0.0009 },
    "40": { "A": 0.0304, "C": 0.0037, "D": 0.0041, "E": 0.0004, "F": 0.0001, "G": 0.5975, "H": 0.0006, "I": 0.0019, "L": 0.0004, "N": 0.0065, "P": 0.0003, "R": 0.0098, "S": 0.3295, "T": 0.0093, "V": 0.0054, "Y": 0.0001 },
    "41": { "C": 0.0067, "F": 0.0002, "G": 0.0032, "L": 0.0039, "R": 0.0031, "S": 0.001, "V": 0.0002, "W": 0.9816 },
    "42": { "A": 0.0001, "C": 0.0003, "D": 0.0002, "F": 0.0138, "G": 0.0001, "H": 0.0006, "I": 0.9151, "L": 0.0236, "M": 0.0024, "N": 0.0027, "P": 0.0002, "R": 0.0004, "S": 0.007, "T": 0.0075, "V": 0.0257, "Y": 0.0003 },
    "43": { "C": 0.0018, "G": 0.0014, "H": 0.0017, "L": 0.0037, "P": 0.0012, "R": 0.9874, "S": 0.0027, "T": 0.0001 },
    "44": { "A": 0.0001, "E": 0.0005, "H": 0.0079, "K": 0.0026, "L": 0.0022, "P": 0.0125, "Q": 0.9698, "R": 0.0033, "S": 0.0006, "T": 0.0004, "Y": 0.0001 },
    "45": { "A": 0.0062, "C": 0.0001, "D": 0.0001, "F": 0.0004, "H": 0.0055, "L": 0.0022, "P": 0.9586, "R": 0.0011, "S": 0.0169, "T": 0.0083, "V": 0.0004 },
    "46": { "A": 0.0013, "H": 0.0007, "L": 0.0011, "P": 0.9865, "Q": 0.0038, "R": 0.001, "S": 0.0018, "T": 0.0038 },
    "47": { "A": 0.0025, "D": 0.0001, "E": 0.0039, "G": 0.983, "K": 0.0001, "P": 0.0001, "Q": 0.0003, "R": 0.005, "T": 0.0009, "V": 0.0027, "W": 0.0012 },
    "48": { "A": 0.0004, "E": 0.0098, "G": 0.0018, "H": 0.0002, "I": 0.0001, "K": 0.9476, "M": 0.001, "N": 0.0046, "P": 0.0004, "Q": 0.0156, "R": 0.0071, "T": 0.0113, "V": 0.0001 },
    "49": { "A": 0.9757, "D": 0.0033, "E": 0.0001, "G": 0.0029, "P": 0.0074, "S": 0.0041, "T": 0.0049, "V": 0.0013 },
    "50": { "F": 0.0001, "I": 0.0001, "L": 0.9715, "M": 0.005, "P": 0.0169, "Q": 0.0011, "R": 0.0044, "T": 0.0001, "V": 0.0007 },
    "51": { "A": 0.0106, "D": 0.0087, "E": 0.9672, "G": 0.0053, "K": 0.0013, "P": 0.0003, "Q": 0.0044, "R": 0.0002, "S": 0.0001, "V": 0.0018 },
    "52": { "C": 0.0118, "F": 0.0018, "G": 0.0059, "L": 0.0062, "N": 0.0001, "R": 0.0025, "S": 0.0025, "V": 0.0005, "W": 0.9674, "Y": 0.0011 },
    "53": { "A": 0.0006, "F": 0.0031, "H": 0.0011, "I": 0.0041, "L": 0.9787, "P": 0.0034, "R": 0.0052, "S": 0.0002, "V": 0.0034 },
    "54": { "A": 0.9557, "C": 0.0006, "D": 0.0003, "E": 0.004, "G": 0.0163, "I": 0.0002, "P": 0.0016, "Q": 0.0001, "S": 0.009, "T": 0.0068, "V": 0.0051 },
    "55": { "A": 0.0009, "C": 0.0002, "D": 0.0007, "F": 0.0125, "G": 0.0009, "H": 0.196, "I": 0.0173, "K": 0.0001, "L": 0.6341, "M": 0.0016, "N": 0.003, "P": 0.0059, "Q": 0.0016, "R": 0.096, "S": 0.0015, "T": 0.0009, "V": 0.0259, "Y": 0.0008 },
    "56": { "A": 0.0007, "C": 0.0002, "F": 0.0029, "G": 0.0003, "H": 0.0006, "I": 0.9369, "L": 0.0181, "M": 0.0071, "N": 0.0025, "P": 0.0003, "R": 0.0007, "S": 0.0073, "T": 0.0044, "V": 0.0177, "Y": 0.0003 },
    "57": { "A": 0.0019, "C": 0.0034, "D": 0.1541, "E": 0.0013, "F": 0.2096, "G": 0.0005, "H": 0.007, "I": 0.0004, "K": 0.0001, "L": 0.0035, "M": 0.0002, "N": 0.0019, "Q": 0.0004, "R": 0.0002, "S": 0.0108, "T": 0.0002, "V": 0.0017, "W": 0.0008, "Y": 0.602 },
    "58": { "A": 0.0034, "C": 0.0044, "D": 0.0001, "E": 0.0005, "F": 0.0013, "G": 0.0115, "H": 0.0001, "L": 0.0063, "P": 0.0017, "Q": 0.0001, "R": 0.0034, "S": 0.197, "T": 0.0007, "V": 0.0004, "W": 0.7663, "Y": 0.0026 },
    "59": { "A": 0.0041, "C": 0.0001, "D": 0.5605, "E": 0.006, "F": 0.0001, "G": 0.0054, "H": 0.0053, "I": 0.0014, "K": 0.0038, "L": 0.0001, "M": 0.0001, "N": 0.3951, "Q": 0.0004, "R": 0.0005, "S": 0.0068, "T": 0.0054, "V": 0.0013, "Y": 0.0034 },
    "63": { "A": 0.0074, "C": 0.0011, "D": 0.9259, "E": 0.01, "F": 0.0001, "G": 0.0311, "H": 0.0012, "K": 0.0002, "M": 0.0001, "N": 0.0142, "R": 0.0001, "S": 0.001, "T": 0.0001, "V": 0.0025, "Y": 0.0049 },
    "64": { "A": 0.01, "C": 0.0009, "D": 0.7266, "E": 0.2149, "F": 0.0003, "G": 0.0084, "H": 0.0021, "I": 0.0008, "K": 0.0015, "L": 0.0006, "N": 0.0152, "P": 0.0002, "Q": 0.0021, "R": 0.0008, "S": 0.0021, "T": 0.0029, "V": 0.0059, "W": 0.0001, "Y": 0.0047 },
    "65": { "A": 0.0018, "D": 0.0032, "E": 0.0182, "G": 0.0008, "H": 0.0005, "I": 0.0035, "K": 0.9108, "L": 0.0005, "M": 0.001, "N": 0.0117, "P": 0.0004, "Q": 0.0107, "R": 0.0153, "S": 0.0017, "T": 0.0166, "V": 0.0016, "W": 0.0003, "Y": 0.0014 },
    "66": { "A": 0.0065, "C": 0.0015, "D": 0.0015, "E": 0.0009, "F": 0.0359, "G": 0.0023, "H": 0.0164, "I": 0.0012, "K": 0.0016, "L": 0.0076, "M": 0.0001, "N": 0.0018, "P": 0.0044, "Q": 0.0015, "R": 0.5889, "S": 0.1957, "T": 0.0034, "V": 0.0022, "W": 0.0024, "Y": 0.1241 },
    "67": { "A": 0.0006, "C": 0.0015, "D": 0.005, "E": 0.0003, "F": 0.0104, "G": 0.0001, "H": 0.0067, "I": 0.0006, "L": 0.0031, "N": 0.0019, "P": 0.0007, "Q": 0.0002, "R": 0.0004, "S": 0.0095, "T": 0.0002, "V": 0.0003, "Y": 0.9583 },
    "68": { "A": 0.0018, "C": 0.0009, "D": 0.0005, "E": 0.0002, "F": 0.0001, "G": 0.0032, "H": 0.0017, "I": 0.0056, "K": 0.0021, "L": 0.0006, "M": 0.0004, "N": 0.0423, "P": 0.0002, "Q": 0.0006, "R": 0.0356, "S": 0.889, "T": 0.0146, "V": 0.0003, "W": 0.0001, "Y": 0.0002 },
    "69": { "A": 0.0096, "D": 0.0003, "E": 0.0004, "G": 0.0001, "H": 0.0011, "I": 0.0011, "K": 0.0032, "L": 0.002, "M": 0.0001, "N": 0.0004, "P": 0.6301, "Q": 0.0042, "R": 0.0026, "S": 0.0117, "T": 0.3328 },
    "70": { "A": 0.0135, "C": 0.0006, "D": 0.0001, "F": 0.0074, "G": 0.0003, "H": 0.0005, "I": 0.0011, "L": 0.0016, "N": 0.0001, "P": 0.01, "R": 0.0002, "S": 0.951, "T": 0.0091, "V": 0.0003, "Y": 0.0041 },
    "71": { "A": 0.0001, "F": 0.0013, "H": 0.0001, "I": 0.0003, "L": 0.9723, "M": 0.005, "P": 0.008, "Q": 0.0023, "R": 0.0066, "S": 0.0005, "V": 0.0033, "W": 0.0001 },
    "72": { "A": 0.0012, "D": 0.0013, "E": 0.0172, "G": 0.0034, "H": 0.0009, "I": 0.0003, "K": 0.889, "M": 0.0038, "N": 0.0117, "P": 0.0008, "Q": 0.0127, "R": 0.0471, "S": 0.003, "T": 0.0067, "W": 0.0005, "Y": 0.0002 },
    "74": { "A": 0.0023, "C": 0.001, "D": 0.0052, "E": 0.0002, "G": 0.0132, "H": 0.0012, "I": 0.0026, "K": 0.0016, "L": 0.0001, "N": 0.0459, "P": 0.0018, "Q": 0.0001, "R": 0.0154, "S": 0.7281, "T": 0.1805, "V": 0.0004, "Y": 0.0002 },
    "75": { "A": 0.0001, "E": 0.0001, "G": 0.0022, "I": 0.0001, "K": 0.0044, "L": 0.0001, "M": 0.0042, "N": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.9792, "S": 0.0064, "T": 0.0019, "V": 0.0001, "W": 0.0006 },
    "76": { "A": 0.0003, "F": 0.0036, "G": 0.0002, "H": 0.0015, "I": 0.0104, "L": 0.9558, "M": 0.0002, "P": 0.005, "R": 0.0056, "S": 0.0004, "T": 0.0002, "V": 0.0167 },
    "77": { "A": 0.0105, "C": 0.0001, "D": 0.0012, "F": 0.0006, "G": 0.0002, "H": 0.0005, "I": 0.0044, "K": 0.0001, "N": 0.0075, "P": 0.0044, "R": 0.0005, "S": 0.0213, "T": 0.9478, "V": 0.0005, "Y": 0.0003 },
    "78": { "A": 0.0003, "C": 0.0003, "D": 0.0001, "F": 0.0045, "G": 0.0002, "H": 0.0003, "I": 0.9455, "K": 0.0001, "L": 0.0095, "M": 0.0015, "N": 0.0029, "P": 0.0001, "R": 0.0004, "S": 0.0087, "T": 0.0034, "V": 0.0224 },
    "79": { "A": 0.0153, "C": 0.0001, "D": 0.0007, "F": 0.0018, "G": 0.0001, "H": 0.0004, "I": 0.0049, "K": 0.0002, "L": 0.0004, "M": 0.0013, "N": 0.0044, "P": 0.0064, "R": 0.0009, "S": 0.3582, "T": 0.6029, "V": 0.0007, "Y": 0.0014 },
    "80": { "A": 0.0008, "E": 0.004, "G": 0.0006, "H": 0.0004, "K": 0.9507, "M": 0.0017, "N": 0.0063, "P": 0.0002, "Q": 0.0062, "R": 0.0217, "S": 0.0011, "T": 0.0058, "V": 0.0001, "W": 0.0003 },
    "81": { "A": 0.0051, "D": 0.9666, "E": 0.0091, "F": 0.0002, "G": 0.0068, "H": 0.0015, "K": 0.0001, "N": 0.0034, "S": 0.0002, "T": 0.0002, "V": 0.0018, "Y": 0.0049 },
    "82": { "A": 0.0094, "D": 0.0009, "G": 0.0002, "H": 0.0004, "I": 0.0057, "K": 0.0004, "L": 0.0002, "M": 0.0001, "N": 0.008, "P": 0.0118, "R": 0.0003, "S": 0.0126, "T": 0.9489, "V": 0.0009 },
    "83": { "A": 0.0191, "C": 0.0006, "D": 0.0007, "F": 0.0044, "H": 0.0002, "I": 0.0005, "L": 0.0004, "N": 0.0002, "P": 0.011, "R": 0.0001, "S": 0.9507, "T": 0.0067, "V": 0.0005, "Y": 0.005 },
    "84": { "A": 0.0007, "D": 0.0014, "E": 0.0128, "G": 0.0027, "H": 0.0004, "I": 0.0011, "K": 0.9268, "L": 0.0002, "M": 0.0002, "N": 0.0141, "P": 0.0001, "Q": 0.0053, "R": 0.0252, "S": 0.0012, "T": 0.0076, "V": 0.0001, "Y": 0.0001 },
    "85": { "A": 0.0004, "C": 0.0003, "D": 0.0101, "E": 0.0003, "G": 0.003, "H": 0.0047, "I": 0.003, "K": 0.0088, "N": 0.7492, "P": 0.0001, "Q": 0.0002, "R": 0.0037, "S": 0.2064, "T": 0.0086, "V": 0.0001, "Y": 0.0009 },
    "86": { "D": 0.0001, "E": 0.01, "G": 0.0002, "H": 0.0088, "K": 0.0039, "L": 0.0058, "M": 0.0001, "N": 0.0001, "P": 0.0047, "Q": 0.9582, "R": 0.0074, "S": 0.0001, "T": 0.0002, "V": 0.0002, "W": 0.0001 },
    "87": { "A": 0.0029, "E": 0.0012, "F": 0.0004, "G": 0.0046, "I": 0.0007, "L": 0.0062, "M": 0.003, "R": 0.0001, "S": 0.0001, "V": 0.9802, "W": 0.0005 },
    "88": { "A": 0.0051, "C": 0.0004, "D": 0.0014, "E": 0.0001, "F": 0.0105, "G": 0.0048, "I": 0.0072, "L": 0.005, "M": 0.0004, "N": 0.0001, "R": 0.0002, "S": 0.0003, "T": 0.0014, "V": 0.9626, "Y": 0.0006 },
    "89": { "C": 0.0001, "F": 0.0043, "H": 0.001, "I": 0.0038, "L": 0.9809, "M": 0.0002, "P": 0.0018, "Q": 0.0001, "R": 0.0058, "S": 0.0005, "V": 0.0014 },
    "90": { "A": 0.0045, "E": 0.0002, "G": 0.0004, "H": 0.0001, "I": 0.0119, "K": 0.0112, "L": 0.0005, "M": 0.0015, "N": 0.0017, "P": 0.0007, "Q": 0.0008, "R": 0.0127, "S": 0.0097, "T": 0.9425, "V": 0.0015 },
    "91": { "A": 0.0001, "F": 0.0001, "G": 0.0002, "I": 0.0073, "K": 0.0016, "L": 0.0337, "M": 0.9294, "N": 0.0001, "Q": 0.0001, "R": 0.0061, "S": 0.0005, "T": 0.002, "V": 0.0188 },
    "92": { "A": 0.0161, "D": 0.0002, "E": 0.0001, "F": 0.0001, "G": 0.0005, "H": 0.0001, "I": 0.0025, "K": 0.0005, "L": 0.0001, "M": 0.0001, "N": 0.004, "P": 0.0041, "R": 0.0008, "S": 0.0078, "T": 0.9625, "V": 0.0002, "Y": 0.0002 },
    "93": { "A": 0.0004, "D": 0.0235, "E": 0.0004, "F": 0.0002, "G": 0.0009, "H": 0.0035, "I": 0.0009, "K": 0.0088, "N": 0.9465, "Q": 0.0002, "R": 0.0003, "S": 0.0065, "T": 0.0059, "V": 0.0001, "Y": 0.0019 },
    "94": { "A": 0.0008, "F": 0.0002, "G": 0.0003, "I": 0.0084, "K": 0.0008, "L": 0.016, "M": 0.9189, "Q": 0.0001, "R": 0.0052, "S": 0.0003, "T": 0.0023, "V": 0.0464, "W": 0.0001 },
    "95": { "A": 0.0072, "C": 0.0001, "D": 0.9548, "E": 0.0111, "G": 0.0109, "H": 0.0022, "K": 0.0003, "L": 0.0003, "N": 0.005, "Q": 0.0013, "R": 0.0002, "S": 0.0007, "T": 0.0001, "V": 0.0017, "Y": 0.0041 },
    "96": { "A": 0.0042, "D": 0.0001, "E": 0.0001, "F": 0.0006, "G": 0.0001, "H": 0.0022, "L": 0.0029, "P": 0.9752, "Q": 0.0001, "R": 0.0016, "S": 0.0043, "T": 0.0073, "V": 0.0007, "Y": 0.0004 },
    "97": { "A": 0.0256, "C": 0.0001, "D": 0.0012, "E": 0.0102, "F": 0.0005, "G": 0.0076, "I": 0.001, "K": 0.0003, "L": 0.0142, "M": 0.011, "N": 0.0001, "P": 0.0002, "Q": 0.0004, "R": 0.0015, "S": 0.0011, "T": 0.0012, "V": 0.9234, "W": 0.0004 },
    "98": { "A": 0.0015, "D": 0.9831, "E": 0.0083, "G": 0.0018, "H": 0.0007, "K": 0.0001, "N": 0.0014, "V": 0.0008, "Y": 0.0024 },
    "99": { "A": 0.0044, "E": 0.0001, "I": 0.0008, "K": 0.0022, "P": 0.0013, "Q": 0.0001, "R": 0.0012, "S": 0.0049, "T": 0.9849 },
    "100": { "A": 0.933, "C": 0.0003, "D": 0.0017, "G": 0.0596, "P": 0.0003, "S": 0.0032, "T": 0.001, "V": 0.0008 },
    "101": { "A": 0.0024, "I": 0.0017, "K": 0.003, "L": 0.0001, "M": 0.001, "N": 0.0001, "P": 0.0014, "R": 0.0024, "S": 0.0033, "T": 0.9841, "V": 0.0004 },
    "102": { "C": 0.0012, "D": 0.0059, "E": 0.0007, "F": 0.0031, "H": 0.0012, "N": 0.0009, "S": 0.0008, "Y": 0.9859 },
    "103": { "A": 0.0001, "C": 0.0034, "D": 0.0042, "E": 0.0001, "F": 0.0552, "G": 0.0001, "H": 0.0041, "I": 0.0001, "L": 0.0003, "M": 0.0001, "N": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0043, "T": 0.0001, "V": 0.0004, "W": 0.0002, "Y": 0.9268 },
    "104": { "C": 0.9797, "D": 0.0001, "F": 0.0019, "G": 0.0084, "L": 0.0001, "R": 0.0015, "S": 0.0012, "V": 0.0002, "W": 0.0057, "Y": 0.0009 },
    "105": { "A": 0.938, "C": 0.0001, "D": 0.0002, "E": 0.0028, "G": 0.0053, "I": 0.0017, "K": 0.0001, "L": 0.0015, "M": 0.0001, "N": 0.0003, "P": 0.0005, "Q": 0.0001, "R": 0.0001, "S": 0.0058, "T": 0.009, "V": 0.0337, "W": 0.0003, "Y": 0.0002 },
    "106": { "A": 0.0021, "C": 0.0005, "D": 0.0011, "E": 0.0001, "F": 0.0008, "G": 0.0041, "H": 0.5305, "I": 0.0005, "K": 0.002, "L": 0.0193, "M": 0.0004, "N": 0.0042, "P": 0.005, "Q": 0.022, "R": 0.3889, "S": 0.0029, "T": 0.0017, "V": 0.0012, "W": 0.0013, "Y": 0.0113 },
    "107": { "A": 0.0311, "C": 0.0049, "D": 0.0103, "E": 0.0117, "F": 0.0129, "G": 0.0405, "H": 0.0082, "I": 0.1911, "K": 0.0208, "L": 0.0381, "M": 0.0441, "N": 0.0167, "P": 0.0065, "Q": 0.0086, "R": 0.2337, "S": 0.1756, "T": 0.0998, "V": 0.0277, "W": 0.0049, "Y": 0.0127 },
    "108": { "A": 0.0385, "C": 0.0051, "D": 0.0341, "E": 0.024, "F": 0.019, "G": 0.0912, "H": 0.0285, "I": 0.0354, "K": 0.0244, "L": 0.0965, "M": 0.0112, "N": 0.0225, "P": 0.1111, "Q": 0.0408, "R": 0.1821, "S": 0.0732, "T": 0.0507, "V": 0.045, "W": 0.0162, "Y": 0.0504 },
    "109": { "A": 0.0539, "C": 0.0125, "D": 0.0554, "E": 0.034, "F": 0.0248, "G": 0.1555, "H": 0.0178, "I": 0.0386, "K": 0.0156, "L": 0.0594, "M": 0.0164, "N": 0.0229, "P": 0.0578, "Q": 0.018, "R": 0.0806, "S": 0.0922, "T": 0.0509, "V": 0.0676, "W": 0.022, "Y": 0.1041 },
    "110": { "A": 0.0666, "C": 0.0124, "D": 0.0759, "E": 0.0312, "F": 0.0271, "G": 0.1536, "H": 0.0169, "I": 0.0349, "K": 0.0089, "L": 0.0468, "M": 0.01, "N": 0.0218, "P": 0.0388, "Q": 0.0173, "R": 0.0505, "S": 0.1169, "T": 0.0532, "V": 0.0557, "W": 0.0268, "Y": 0.1346 },
    "111": { "A": 0.0721, "C": 0.0231, "D": 0.0705, "E": 0.0188, "F": 0.0317, "G": 0.139, "H": 0.0102, "I": 0.0348, "K": 0.0072, "L": 0.0444, "M": 0.0141, "N": 0.0203, "P": 0.035, "Q": 0.0138, "R": 0.0454, "S": 0.1425, "T": 0.0493, "V": 0.052, "W": 0.031, "Y": 0.1449 },
    "111A": { "A": 0.0597, "C": 0.0165, "D": 0.0849, "E": 0.0154, "F": 0.0375, "G": 0.1519, "H": 0.0073, "I": 0.0279, "K": 0.0047, "L": 0.0449, "M": 0.013, "N": 0.0247, "P": 0.0243, "Q": 0.0095, "R": 0.0334, "S": 0.1661, "T": 0.0507, "V": 0.0591, "W": 0.0458, "Y": 0.1227 },
    "111B": { "A": 0.0505, "C": 0.0231, "D": 0.0741, "E": 0.0162, "F": 0.0444, "G": 0.1534, "H": 0.0082, "I": 0.0325, "K": 0.0067, "L": 0.0409, "M": 0.0076, "N": 0.0248, "P": 0.024, "Q": 0.0119, "R": 0.035, "S": 0.1789, "T": 0.0599, "V": 0.0492, "W": 0.0541, "Y": 0.1046 },
    "111C": { "A": 0.0348, "C": 0.0166, "D": 0.0628, "E": 0.0159, "F": 0.0468, "G": 0.1855, "H": 0.0064, "I": 0.0214, "K": 0.007, "L": 0.0468, "M": 0.0098, "N": 0.0166, "P": 0.0221, "Q": 0.005, "R": 0.0416, "S": 0.2003, "T": 0.0562, "V": 0.0514, "W": 0.065, "Y": 0.088 },
    "111D": { "A": 0.0331, "C": 0.0269, "D": 0.0442, "E": 0.0144, "F": 0.0389, "G": 0.2127, "H": 0.0048, "I": 0.0211, "K": 0.0086, "L": 0.0379, "M": 0.0053, "N": 0.0134, "P": 0.0326, "Q": 0.0091, "R": 0.0192, "S": 0.193, "T": 0.0643, "V": 0.0634, "W": 0.049, "Y": 0.108 },
    "111E": { "A": 0.0465, "C": 0.0203, "D": 0.043, "E": 0.0107, "F": 0.0251, "G": 0.2291, "H": 0.0072, "I": 0.0263, "K": 0.0048, "L": 0.0286, "M": 0.0048, "N": 0.0143, "P": 0.0465, "Q": 0.0084, "R": 0.0286, "S": 0.1647, "T": 0.0656, "V": 0.0549, "W": 0.0573, "Y": 0.1134 },
    "111F": { "A": 0.0439, "C": 0.0345, "D": 0.0408, "E": 0.0251, "F": 0.0251, "G": 0.1348, "H": 0.0157, "I": 0.1442, "K": 0.0157, "L": 0.0282, "M": 0.0031, "N": 0.0282, "P": 0.0345, "Q": 0.0063, "R": 0.0376, "S": 0.1317, "T": 0.0313, "V": 0.0251, "W": 0.0564, "Y": 0.1379 },
    "111G": { "C": 0.0465, "D": 0.0465, "G": 0.1318, "H": 0.0155, "I": 0.0078, "K": 0.0155, "L": 0.031, "M": 0.0078, "N": 0.0388, "P": 0.3953, "Q": 0.0155, "R": 0.0078, "S": 0.1008, "T": 0.0078, "V": 0.0388, "W": 0.0233, "Y": 0.0698 },
    "112G": { "A": 0.0721, "C": 0.0144, "D": 0.0385, "E": 0.0288, "F": 0.0144, "G": 0.1538, "I": 0.0192, "K": 0.024, "L": 0.2212, "M": 0.0048, "N": 0.0048, "P": 0.0481, "Q": 0.0048, "R": 0.0385, "S": 0.101, "T": 0.0529, "V": 0.0385, "W": 0.0144, "Y": 0.1058 },
    "112F": { "A": 0.035, "C": 0.0272, "D": 0.0252, "E": 0.0078, "F": 0.0388, "G": 0.1301, "H": 0.0078, "I": 0.0272, "K": 0.0272, "L": 0.0718, "M": 0.0019, "N": 0.0214, "P": 0.0408, "Q": 0.0252, "R": 0.0524, "S": 0.134, "T": 0.0544, "V": 0.0485, "W": 0.099, "Y": 0.1243 },
    "112E": { "A": 0.0415, "C": 0.0379, "D": 0.0481, "E": 0.0182, "F": 0.0634, "G": 0.1704, "H": 0.016, "I": 0.0306, "K": 0.0124, "L": 0.0459, "M": 0.0044, "N": 0.0138, "P": 0.043, "Q": 0.0066, "R": 0.0357, "S": 0.1457, "T": 0.043, "V": 0.0401, "W": 0.0422, "Y": 0.1413 },
    "112D": { "A": 0.0383, "C": 0.0277, "D": 0.0383, "E": 0.0225, "F": 0.0261, "G": 0.1919, "H": 0.0119, "I": 0.0319, "K": 0.0132, "L": 0.0309, "M": 0.0052, "N": 0.0241, "P": 0.0399, "Q": 0.0119, "R": 0.048, "S": 0.1642, "T": 0.0551, "V": 0.0464, "W": 0.0357, "Y": 0.1368 },
    "112C": { "A": 0.0412, "C": 0.0284, "D": 0.0429, "E": 0.0244, "F": 0.0229, "G": 0.1657, "H": 0.0132, "I": 0.0299, "K": 0.0087, "L": 0.0309, "M": 0.0053, "N": 0.0305, "P": 0.0474, "Q": 0.0108, "R": 0.0429, "S": 0.1585, "T": 0.0462, "V": 0.0492, "W": 0.0392, "Y": 0.162 },
    "112B": { "A": 0.0466, "C": 0.0229, "D": 0.0466, "E": 0.019, "F": 0.0195, "G": 0.1472, "H": 0.0161, "I": 0.0217, "K": 0.0128, "L": 0.0433, "M": 0.0063, "N": 0.0266, "P": 0.0381, "Q": 0.0112, "R": 0.0445, "S": 0.1364, "T": 0.0511, "V": 0.0423, "W": 0.0574, "Y": 0.1904 },
    "112A": { "A": 0.056, "C": 0.0151, "D": 0.0427, "E": 0.0225, "F": 0.0255, "G": 0.1547, "H": 0.0111, "I": 0.0204, "K": 0.0137, "L": 0.041, "M": 0.0069, "N": 0.0342, "P": 0.0511, "Q": 0.012, "R": 0.0506, "S": 0.1196, "T": 0.052, "V": 0.0393, "W": 0.0349, "Y": 0.1968 },
    "112": { "A": 0.0492, "C": 0.0142, "D": 0.0465, "E": 0.0349, "F": 0.0215, "G": 0.1408, "H": 0.0234, "I": 0.0198, "K": 0.0142, "L": 0.0518, "M": 0.0076, "N": 0.0341, "P": 0.0554, "Q": 0.0149, "R": 0.0404, "S": 0.1111, "T": 0.0534, "V": 0.0393, "W": 0.0401, "Y": 0.1874 },
    "113": { "A": 0.035, "C": 0.0078, "D": 0.1043, "E": 0.0271, "F": 0.0232, "G": 0.153, "H": 0.0218, "I": 0.0123, "K": 0.007, "L": 0.0317, "M": 0.0053, "N": 0.0833, "P": 0.031, "Q": 0.0096, "R": 0.0403, "S": 0.0638, "T": 0.0371, "V": 0.0283, "W": 0.0504, "Y": 0.2276 },
    "114": { "A": 0.1489, "C": 0.0105, "D": 0.0252, "E": 0.0142, "F": 0.0212, "G": 0.1511, "H": 0.0217, "I": 0.0119, "K": 0.0044, "L": 0.0319, "M": 0.0029, "N": 0.0152, "P": 0.0467, "Q": 0.0067, "R": 0.0151, "S": 0.0441, "T": 0.027, "V": 0.0274, "W": 0.1283, "Y": 0.2455 },
    "115": { "A": 0.0093, "C": 0.0048, "D": 0.0034, "E": 0.0028, "F": 0.6753, "G": 0.0137, "H": 0.0026, "I": 0.0194, "K": 0.0006, "L": 0.0535, "M": 0.1175, "N": 0.0031, "P": 0.0169, "Q": 0.0006, "R": 0.0023, "S": 0.0091, "T": 0.0042, "V": 0.0209, "W": 0.0078, "Y": 0.0322 },
    "116": { "A": 0.011, "C": 0.0004, "D": 0.893, "E": 0.0103, "F": 0.0019, "G": 0.0218, "H": 0.0065, "I": 0.0021, "K": 0.0007, "L": 0.0037, "M": 0.0006, "N": 0.0074, "P": 0.0036, "Q": 0.0144, "R": 0.0034, "S": 0.0049, "T": 0.0022, "V": 0.0055, "W": 0.0013, "Y": 0.0054 },
    "117": { "A": 0.005, "C": 0.0015, "D": 0.0102, "E": 0.0007, "F": 0.0239, "G": 0.0021, "H": 0.0255, "I": 0.1148, "K": 0.0005, "L": 0.033, "M": 0.0023, "N": 0.0089, "P": 0.1437, "Q": 0.0038, "R": 0.0053, "S": 0.0245, "T": 0.0049, "V": 0.1371, "W": 0.0006, "Y": 0.4516 },
    "118": { "C": 0.0018, "F": 0.0033, "G": 0.0029, "L": 0.0016, "P": 0.0001, "R": 0.0014, "S": 0.0002, "V": 0.0001, "W": 0.9885 },
    "119": { "A": 0.0008, "C": 0.0028, "D": 0.0016, "E": 0.0003, "G": 0.9893, "L": 0.0001, "R": 0.0005, "S": 0.0026, "V": 0.0017 },
    "120": { "A": 0.0001, "E": 0.0007, "G": 0.0001, "H": 0.0044, "K": 0.0289, "L": 0.0033, "N": 0.0002, "P": 0.016, "Q": 0.9082, "R": 0.0376, "T": 0.0002 },
    "121": { "A": 0.0002, "E": 0.0008, "G": 0.996, "R": 0.0007, "V": 0.002, "W": 0.0002 },
    "122": { "A": 0.0179, "C": 0.0001, "D": 0.0001, "F": 0.0004, "H": 0.0004, "I": 0.0354, "K": 0.0008, "L": 0.0022, "M": 0.0019, "N": 0.0016, "P": 0.0012, "R": 0.0002, "S": 0.0076, "T": 0.927, "V": 0.003, "Y": 0.0001 },
    "123": { "A": 0.0017, "E": 0.0003, "F": 0.0007, "H": 0.0006, "I": 0.0005, "K": 0.0007, "L": 0.7169, "M": 0.1192, "P": 0.0111, "Q": 0.005, "R": 0.0068, "S": 0.0007, "T": 0.1274, "V": 0.0083 },
    "124": { "A": 0.0014, "C": 0.0001, "D": 0.0002, "F": 0.0009, "G": 0.0033, "I": 0.0037, "L": 0.0008, "S": 0.0002, "V": 0.9895 },
    "125": { "A": 0.0074, "D": 0.0001, "F": 0.0002, "H": 0.0001, "I": 0.0086, "L": 0.0001, "M": 0.0003, "N": 0.001, "P": 0.0012, "S": 0.0074, "T": 0.9726, "V": 0.0009 },
    "126": { "A": 0.0023, "D": 0.0003, "F": 0.0007, "G": 0.0024, "I": 0.0032, "L": 0.0004, "R": 0.0002, "S": 0.0001, "V": 0.9903 },
    "127": { "A": 0.0047, "C": 0.0005, "D": 0.0003, "F": 0.0015, "L": 0.0004, "P": 0.0022, "S": 0.986, "T": 0.0017, "Y": 0.0025 },
    "128": { "A": 0.0063, "L": 0.0004, "P": 0.0032, "Q": 0.0002, "S": 0.9888, "T": 0.0012 }
  },
  "IGHV3": {
    "1": { "D": 0.0106, "E": 0.7061, "H": 0.0007, "L": 0.0008, "P": 0.0001, "Q": 0.2817, "V": 0.0001 },
    "2": { "A": 0.0069, "D": 0.0009, "E": 0.0078, "F": 0.0003, "G": 0.0064, "I": 0.0009, "K": 0.0001, "L": 0.0066, "M": 0.0065, "P": 0.0002, "Q": 0.0002, "T": 0.0001, "V": 0.9625, "W": 0.0004 },
    "3": { "D": 0.0006, "E": 0.004, "G": 0.0004, "H": 0.0185, "K": 0.0052, "L": 0.0056, "M": 0.0001, "N": 0.0008, "P": 0.0072, "Q": 0.9436, "R": 0.0112, "S": 0.0007, "T": 0.0007, "V": 0.0002, "W": 0.0001, "Y": 0.0009 },
    "4": { "A": 0.0002, "F": 0.0002, "G": 0.0001, "I": 0.0003, "L": 0.9725, "M": 0.0052, "P": 0.0028, "Q": 0.0009, "R": 0.0034, "S": 0.0001, "V": 0.0136, "W": 0.0005 },
    "5": { "A": 0.0047, "C": 0.0001, "E": 0.0051, "F": 0.0022, "G": 0.0031, "I": 0.0003, "K": 0.0001, "L": 0.1759, "M": 0.0045, "Q": 0.0022, "R": 0.0001, "S": 0.0012, "T": 0.0002, "V": 0.7997, "W": 0.0007 },
    "6": { "A": 0.0197, "D": 0.0078, "E": 0.9089, "G": 0.007, "H": 0.0001, "K": 0.0008, "L": 0.0001, "N": 0.0002, "P": 0.0003, "Q": 0.0496, "S": 0.0002, "T": 0.0003, "V": 0.0047, "W": 0.0002 },
    "7": { "A": 0.0063, "C": 0.001, "D": 0.0002, "F": 0.0017, "G": 0.0001, "H": 0.0002, "L": 0.0007, "P": 0.0049, "S": 0.9712, "T": 0.0076, "V": 0.0027, "W": 0.0003, "Y": 0.0029 },
    "8": { "A": 0.0016, "C": 0.0002, "E": 0.0011, "G": 0.9859, "R": 0.0015, "S": 0.0001, "V": 0.0047, "W": 0.0047 },
    "9": { "A": 0.0023, "C": 0.0004, "E": 0.0019, "G": 0.9881, "P": 0.0011, "R": 0.0021, "S": 0.0002, "V": 0.0037 },
    "11": { "A": 0.0114, "C": 0.0036, "D": 0.0218, "E": 0.0009, "F": 0.0001, "G": 0.9387, "H": 0.0007, "I": 0.0001, "K": 0.0004, "L": 0.0004, "N": 0.0021, "P": 0.0001, "Q": 0.0001, "R": 0.0038, "S": 0.0028, "T": 0.002, "V": 0.0106, "W": 0.0001, "Y": 0.0002 },
    "12": { "A": 0.0015, "C": 0.0003, "D": 0.0001, "E": 0.0008, "F": 0.0063, "G": 0.0025, "I": 0.0007, "K": 0.0001, "L": 0.7137, "M": 0.0029, "P": 0.0008, "Q": 0.0004, "R": 0.0016, "S": 0.0056, "T": 0.0001, "V": 0.2548, "W": 0.008 },
    "13": { "A": 0.0084, "C": 0.0004, "D": 0.0005, "E": 0.0018, "F": 0.0029, "G": 0.005, "I": 0.0417, "K": 0.0009, "L": 0.0061, "M": 0.0005, "N": 0.0001, "P": 0.0002, "Q": 0.0003, "R": 0.0002, "S": 0.0013, "T": 0.0006, "V": 0.9282, "Y": 0.0008 },
    "14": { "A": 0.0006, "C": 0.0001, "D": 0.0001, "E": 0.0061, "G": 0.0003, "H": 0.0075, "I": 0.0001, "K": 0.1846, "L": 0.0022, "M": 0.0007, "N": 0.0014, "P": 0.008, "Q": 0.7641, "R": 0.0164, "S": 0.0022, "T": 0.0046, "V": 0.0004, "W": 0.0003, "Y": 0.0001 },
    "15": { "A": 0.0049, "D": 0.0001, "E": 0.0001, "F": 0.0004, "G": 0.0002, "H": 0.0059, "I": 0.0002, "L": 0.0046, "P": 0.9617, "Q": 0.0009, "R": 0.005, "S": 0.0094, "T": 0.0063, "V": 0.0001, "W": 0.0001 },
    "16": { "A": 0.0016, "C": 0.0002, "D": 0.0005, "E": 0.0017, "G": 0.9858, "R": 0.002, "S": 0.0015, "V": 0.0038, "W": 0.003 },
    "17": { "A": 0.0021, "C": 0.0001, "D": 0.0008, "E": 0.0132, "G": 0.6938, "K": 0.0047, "L": 0.0003, "M": 0.0022, "N": 0.0002, "P": 0.0002, "Q": 0.0009, "R": 0.2661, "S": 0.0023, "T": 0.0048, "V": 0.0034, "W": 0.0048 },
    "18": { "A": 0.0071, "C": 0.003, "D": 0.0002, "F": 0.0027, "H": 0.0002, "L": 0.0001, "P": 0.0068, "R": 0.0002, "S": 0.9707, "T": 0.003, "V": 0.0001, "W": 0.0001, "Y": 0.0057 },
    "19": { "A": 0.0002, "C": 0.0002, "F": 0.0005, "G": 0.0002, "H": 0.0001, "I": 0.0005, "L": 0.9716, "M": 0.0052, "P": 0.0028, "Q": 0.0013, "R": 0.0093, "S": 0.0002, "T": 0.0001, "V": 0.0074, "W": 0.0004 },
    "20": { "A": 0.0001, "C": 0.0006, "D": 0.0001, "E": 0.0006, "G": 0.0101, "H": 0.0002, "I": 0.0042, "K": 0.0141, "L": 0.0005, "M": 0.0003, "N": 0.0009, "P": 0.0003, "Q": 0.0007, "R": 0.9258, "S": 0.0308, "T": 0.0103, "V": 0.0002, "W": 0.0001 },
    "21": { "A": 0.0003, "C": 0.0005, "F": 0.0019, "G": 0.0001, "H": 0.0017, "I": 0.008, "L": 0.9555, "P": 0.0125, "R": 0.0105, "S": 0.0006, "T": 0.0008, "V": 0.0075 },
    "22": { "A": 0.0132, "C": 0.0027, "D": 0.0002, "F": 0.0028, "G": 0.0001, "H": 0.001, "I": 0.0001, "L": 0.0011, "P": 0.0229, "R": 0.0003, "S": 0.943, "T": 0.0056, "V": 0.0004, "Y": 0.0064 },
    "23": { "C": 0.9641, "F": 0.0047, "G": 0.0108, "I": 0.0001, "L": 0.0008, "P": 0.0001, "R": 0.0041, "S": 0.0025, "V": 0.001, "W": 0.0109, "Y": 0.0008 },
    "24": { "A": 0.8576, "C": 0.0001, "D": 0.0028, "E": 0.0195, "F": 0.0001, "G": 0.0068, "H": 0.0002, "I": 0.0014, "K": 0.0016, "L": 0.002, "M": 0.0001, "N": 0.0002, "P": 0.0073, "Q": 0.0009, "R": 0.0011, "S": 0.0265, "T": 0.0355, "V": 0.0362, "Y": 0.0001 },
    "25": { "A": 0.926, "C": 0.0004, "D": 0.0055, "E": 0.0005, "F": 0.0003, "G": 0.0171, "H": 0.0001, "I": 0.0005, "L": 0.0003, "M": 0.0002, "P": 0.0063, "R": 0.0004, "S": 0.006, "T": 0.0113, "V": 0.0249, "Y": 0.0001 },
    "26": { "A": 0.0132, "C": 0.0028, "D": 0.0002, "E": 0.0001, "F": 0.005, "H": 0.0007, "L": 0.0012, "N": 0.0002, "P": 0.0183, "Q": 0.0001, "R": 0.0002, "S": 0.9469, "T": 0.004, "V": 0.0007, "Y": 0.0063 },
    "27": { "A": 0.0083, "C": 0.0009, "D": 0.0021, "E": 0.0078, "F": 0.0001, "G": 0.9655, "I": 0.0001, "K": 0.0006, "L": 0.0001, "N": 0.0002, "P": 0.0001, "Q": 0.0007, "R": 0.0055, "S": 0.0005, "T": 0.0007, "V": 0.0063, "W": 0.0007 },
    "28": { "A": 0.0005, "C": 0.0071, "D": 0.0003, "F": 0.9271, "G": 0.0016, "I": 0.0067, "L": 0.0282, "M": 0.0003, "N": 0.0002, "P": 0.0006, "R": 0.0002, "S": 0.013, "T": 0.0003, "V": 0.0106, "Y": 0.0031 },
    "29": { "A": 0.0118, "C": 0.0002, "D": 0.0029, "E": 0.0001, "F": 0.0006, "G": 0.0021, "H": 0.0024, "I": 0.0241, "K": 0.0024, "L": 0.0019, "M": 0.0025, "N": 0.0143, "P": 0.0648, "Q": 0.0001, "R": 0.0056, "S": 0.0448, "T": 0.8163, "V": 0.0025, "W": 0.0001, "Y": 0.0005 },
    "30": { "A": 0.0012, "C": 0.0085, "D": 0.0001, "F": 0.8715, "G": 0.0013, "I": 0.0062, "L": 0.0325, "M": 0.0002, "P": 0.0005, "R": 0.0004, "S": 0.0091, "T": 0.0005, "V": 0.0654, "W": 0.0002, "Y": 0.0025 },
    "35": { "A": 0.0071, "C": 0.0029, "D": 0.0614, "E": 0.0034, "F": 0.0004, "G": 0.0392, "H": 0.0025, "I": 0.0087, "K": 0.0049, "L": 0.001, "M": 0.0006, "N": 0.0266, "P": 0.0021, "Q": 0.0003, "R": 0.0563, "S": 0.7574, "T": 0.0203, "V": 0.0027, "W": 0.0004, "Y": 0.002 },
    "36": { "A": 0.014, "C": 0.0025, "D": 0.1471, "E": 0.0039, "F": 0.0012, "G": 0.028, "H": 0.0073, "I": 0.0088, "K": 0.0075, "L": 0.0028, "M": 0.0008, "N": 0.1027, "P": 0.0035, "Q": 0.0012, "R": 0.0382, "S": 0.5673, "T": 0.052, "V": 0.0044, "W": 0.0004, "Y": 0.0064 },
    "37": { "A": 0.049, "C": 0.0052, "D": 0.0134, "E": 0.0011, "F": 0.0276, "G": 0.0007, "H": 0.0344, "I": 0.0012, "K": 0.0021, "L": 0.0023, "M": 0.0002, "N": 0.0551, "P": 0.0026, "Q": 0.0016, "R": 0.0014, "S": 0.0445, "T": 0.0052, "V": 0.0022, "W": 0.0004, "Y": 0.75 },
    "38": { "A": 0.3461, "C": 0.0036, "D": 0.018, "E": 0.0254, "F": 0.0034, "G": 0.1672, "H": 0.004, "I": 0.0014, "K": 0.0006, "L": 0.0016, "M": 0.0003, "N": 0.0075, "P": 0.0066, "Q": 0.0003, "R": 0.0066, "S": 0.1197, "T": 0.0184, "V": 0.0099, "W": 0.1585, "Y": 0.101 },
    "39": { "A": 0.0001, "C": 0.0001, "F": 0.0035, "G": 0.0004, "H": 0.0002, "I": 0.0294, "K": 0.0009, "L": 0.039, "M": 0.8976, "N": 0.0002, "P": 0.0002, "Q": 0.0002, "R": 0.01, "S": 0.0011, "T": 0.0043, "V": 0.0108, "W": 0.0018, "Y": 0.0001 },
    "40": { "A": 0.0056, "C": 0.0029, "D": 0.011, "E": 0.0006, "F": 0.0009, "G": 0.0131, "H": 0.3492, "I": 0.0049, "K": 0.0019, "L": 0.0022, "M": 0.0007, "N": 0.176, "P": 0.0171, "Q": 0.0046, "R": 0.0175, "S": 0.3384, "T": 0.0445, "V": 0.0013, "W": 0.0001, "Y": 0.0075 },
    "41": { "A": 0.0001, "C": 0.0059, "E": 0.0001, "F": 0.0002, "G": 0.0112, "L": 0.0063, "M": 0.0001, "Q": 0.0001, "R": 0.0045, "S": 0.0014, "T": 0.0001, "V": 0.0007, "W": 0.9692 },
    "42": { "A": 0.0063, "C": 0.0008, "D": 0.0012, "F": 0.0243, "G": 0.0093, "H": 0.0002, "I": 0.0425, "L": 0.0051, "M": 0.0003, "N": 0.0001, "R": 0.0001, "S": 0.0008, "T": 0.0004, "V": 0.9085, "Y": 0.0001 },
    "43": { "A": 0.0003, "C": 0.0024, "G": 0.0044, "H": 0.0016, "I": 0.0001, "L": 0.0028, "P": 0.0043, "Q": 0.0001, "R": 0.9786, "S": 0.0051, "V": 0.0002, "W": 0.0001 },
    "44": { "A": 0.0003, "E": 0.0016, "H": 0.0054, "K": 0.0064, "L": 0.0059, "M": 0.0001, "N": 0.0001, "P": 0.0169, "Q": 0.9522, "R": 0.0098, "S": 0.0002, "T": 0.0009, "W": 0.0001 },
    "45": { "A": 0.9292, "C": 0.0001, "D": 0.0052, "E": 0.0004, "F": 0.0006, "G": 0.0118, "H": 0.0002, "I": 0.0009, "L": 0.0006, "N": 0.0001, "P": 0.0094, "R": 0.0011, "S": 0.0076, "T": 0.0139, "V": 0.0189, "Y": 0.0001 },
    "46": { "A": 0.0042, "E": 0.0001, "F": 0.0001, "H": 0.0008, "I": 0.0002, "K": 0.0002, "L": 0.0012, "P": 0.9578, "Q": 0.0069, "R": 0.002, "S": 0.0141, "T": 0.012, "V": 0.0002 },
    "47": { "A": 0.0018, "C": 0.0004, "D": 0.0015, "E": 0.0026, "G": 0.9855, "P": 0.0001, "R": 0.0033, "S": 0.0004, "T": 0.0005, "V": 0.0028, "W": 0.0009 },
    "48": { "A": 0.0003, "D": 0.0002, "E": 0.01, "G": 0.0021, "H": 0.0002, "I": 0.0001, "K": 0.9407, "L": 0.0001, "M": 0.0029, "N": 0.003, "P": 0.0002, "Q": 0.0146, "R": 0.015, "S": 0.0001, "T": 0.0102, "V": 0.0002, "W": 0.0002 },
    "49": { "A": 0.0019, "C": 0.0002, "D": 0.0002, "E": 0.0014, "G": 0.9854, "L": 0.0001, "R": 0.0037, "S": 0.0005, "T": 0.0001, "V": 0.0036, "W": 0.0029 },
    "50": { "A": 0.0014, "F": 0.0001, "G": 0.0008, "H": 0.0001, "I": 0.0002, "L": 0.9563, "M": 0.0061, "P": 0.0103, "Q": 0.0012, "R": 0.0133, "S": 0.0002, "T": 0.0001, "V": 0.0092, "W": 0.0007 },
    "51": { "A": 0.0115, "C": 0.0001, "D": 0.0107, "E": 0.868, "F": 0.0002, "G": 0.0499, "H": 0.0002, "I": 0.0002, "K": 0.0018, "L": 0.0018, "M": 0.002, "Q": 0.0072, "R": 0.0005, "S": 0.0002, "T": 0.0003, "V": 0.0442, "W": 0.0008, "Y": 0.0004 },
    "52": { "A": 0.0001, "C": 0.0094, "D": 0.0002, "E": 0.0002, "F": 0.001, "G": 0.0267, "H": 0.0003, "L": 0.0099, "Q": 0.0001, "R": 0.0037, "S": 0.0016, "T": 0.0001, "V": 0.0025, "W": 0.9335, "Y": 0.0106 },
    "53": { "A": 0.0026, "C": 0.0009, "D": 0.0007, "E": 0.0005, "F": 0.0075, "G": 0.0147, "I": 0.0211, "L": 0.0181, "M": 0.0017, "N": 0.0001, "P": 0.0001, "R": 0.0003, "S": 0.0002, "T": 0.0002, "V": 0.9306, "W": 0.0006 },
    "54": { "A": 0.3215, "C": 0.001, "D": 0.0012, "E": 0.0028, "F": 0.0002, "G": 0.0911, "H": 0.0001, "K": 0.0001, "L": 0.0027, "P": 0.0029, "Q": 0.0001, "R": 0.0001, "S": 0.5672, "T": 0.0069, "V": 0.0017, "Y": 0.0005 },
    "55": { "A": 0.1611, "C": 0.0033, "D": 0.0044, "E": 0.001, "F": 0.0487, "G": 0.0845, "H": 0.0092, "I": 0.0125, "K": 0.0018, "L": 0.0273, "M": 0.0009, "N": 0.0608, "P": 0.0011, "Q": 0.0015, "R": 0.1105, "S": 0.1085, "T": 0.0167, "V": 0.2419, "W": 0.0002, "Y": 0.104 },
    "56": { "A": 0.0009, "C": 0.0003, "D": 0.0001, "F": 0.0044, "G": 0.0004, "H": 0.0001, "I": 0.9154, "K": 0.0012, "L": 0.0153, "M": 0.0161, "N": 0.0014, "Q": 0.0001, "R": 0.0039, "S": 0.0124, "T": 0.0108, "V": 0.0169, "Y": 0.0001 },
    "57": { "A": 0.0046, "C": 0.0021, "D": 0.0077, "E": 0.0019, "F": 0.0021, "G": 0.0223, "H": 0.0036, "I": 0.0045, "K": 0.1075, "L": 0.0031, "M": 0.0011, "N": 0.0594, "P": 0.0016, "Q": 0.0026, "R": 0.0715, "S": 0.5555, "T": 0.0222, "V": 0.0015, "W": 0.0696, "Y": 0.0557 },
    "58": { "A": 0.0125, "C": 0.0029, "D": 0.0146, "E": 0.0099, "F": 0.0111, "G": 0.1868, "H": 0.0101, "I": 0.0056, "K": 0.0054, "L": 0.0029, "M": 0.0005, "N": 0.0218, "P": 0.0117, "Q": 0.0486, "R": 0.0168, "S": 0.337, "T": 0.0261, "V": 0.0047, "W": 0.0652, "Y": 0.2057 },
    "59": { "A": 0.0129, "C": 0.0011, "D": 0.3572, "E": 0.0091, "F": 0.0007, "G": 0.0736, "H": 0.0028, "I": 0.0038, "K": 0.0808, "L": 0.0011, "M": 0.0004, "N": 0.0795, "P": 0.0009, "Q": 0.0017, "R": 0.0166, "S": 0.3364, "T": 0.0134, "V": 0.0038, "W": 0.0007, "Y": 0.0035 },
    "60": { "A": 0.3541, "D": 0.0136, "E": 0.0013, "F": 0.0009, "G": 0.0298, "H": 0.0022, "I": 0.0197, "K": 0.0048, "L": 0.0035, "M": 0.0013, "N": 0.0145, "P": 0.0281, "Q": 0.0004, "R": 0.0136, "S": 0.0566, "T": 0.434, "V": 0.0193, "Y": 0.0022 },
    "61": { "A": 0.0107, "C": 0.0013, "D": 0.5029, "E": 0.0336, "F": 0.0076, "G": 0.013, "H": 0.017, "I": 0.0004, "K": 0.0094, "L": 0.0013, "N": 0.1657, "Q": 0.0036, "R": 0.0072, "S": 0.0157, "T": 0.0058, "V": 0.0013, "Y": 0.2033 },
    "62": { "A": 0.0109, "C": 0.0022, "D": 0.0069, "E": 0.0089, "G": 0.7377, "H": 0.0006, "I": 0.0014, "K": 0.0003, "L": 0.0004, "M": 0.0003, "N": 0.0042, "P": 0.0004, "Q": 0.0001, "R": 0.007, "S": 0.2084, "T": 0.0037, "V": 0.0055, "W": 0.0004, "Y": 0.0006 },
    "63": { "A": 0.0097, "C": 0.0029, "D": 0.0247, "E": 0.0077, "F": 0.0015, "G": 0.3717, "H": 0.003, "I": 0.0081, "K": 0.0036, "L": 0.0013, "M": 0.0011, "N": 0.0222, "P": 0.0005, "Q": 0.0008, "R": 0.0226, "S": 0.4568, "T": 0.033, "V": 0.0085, "W": 0.0008, "Y": 0.0195 },
    "64": { "A": 0.0237, "C": 0.0008, "D": 0.0303, "E": 0.0606, "F": 0.0047, "G": 0.0157, "H": 0.008, "I": 0.014, "K": 0.0162, "L": 0.0023, "M": 0.0013, "N": 0.2176, "P": 0.0029, "Q": 0.0035, "R": 0.0186, "S": 0.2902, "T": 0.1835, "V": 0.0062, "W": 0.0008, "Y": 0.0991 },
    "65": { "A": 0.01, "D": 0.0017, "E": 0.0129, "F": 0.0005, "G": 0.001, "H": 0.0005, "I": 0.2314, "K": 0.269, "L": 0.0042, "M": 0.0041, "N": 0.0038, "P": 0.0061, "Q": 0.0074, "R": 0.0132, "S": 0.0062, "T": 0.4212, "V": 0.0061, "Y": 0.0006 },
    "66": { "A": 0.015, "C": 0.0028, "D": 0.0677, "E": 0.0294, "F": 0.0257, "G": 0.0547, "H": 0.0214, "I": 0.0034, "K": 0.0057, "L": 0.004, "M": 0.0005, "N": 0.0341, "P": 0.0004, "Q": 0.0033, "R": 0.0038, "S": 0.0352, "T": 0.0104, "V": 0.0032, "W": 0.0021, "Y": 0.6774 },
    "67": { "A": 0.0002, "C": 0.002, "D": 0.0098, "E": 0.0009, "F": 0.0065, "G": 0.0002, "H": 0.0104, "I": 0.0003, "K": 0.0001, "L": 0.001, "N": 0.0035, "P": 0.0002, "Q": 0.0007, "S": 0.0068, "T": 0.0002, "V": 0.0007, "Y": 0.9564 },
    "68": { "A": 0.864, "C": 0.0002, "D": 0.0012, "E": 0.0053, "G": 0.0163, "H": 0.0001, "I": 0.0022, "K": 0.0005, "L": 0.0038, "M": 0.001, "N": 0.0002, "P": 0.0103, "Q": 0.0003, "R": 0.0014, "S": 0.0088, "T": 0.0115, "V": 0.0726, "W": 0.0001 },
    "69": { "A": 0.0974, "C": 0.0002, "D": 0.8477, "E": 0.0179, "G": 0.0172, "H": 0.0014, "I": 0.0001, "K": 0.0006, "L": 0.0001, "N": 0.007, "P": 0.0017, "Q": 0.0004, "R": 0.0004, "S": 0.0011, "T": 0.0012, "V": 0.0023, "Y": 0.0032 },
    "70": { "A": 0.019, "C": 0.0009, "D": 0.0006, "F": 0.0049, "G": 0.0002, "H": 0.0005, "L": 0.0008, "N": 0.0001, "P": 0.0538, "R": 0.0003, "S": 0.911, "T": 0.0035, "V": 0.0005, "Y": 0.0036 },
    "71": { "A": 0.0048, "C": 0.0002, "E": 0.0013, "F": 0.0006, "G": 0.0115, "I": 0.0007, "L": 0.0122, "M": 0.0048, "P": 0.0001, "R": 0.0002, "S": 0.0002, "T": 0.0001, "V": 0.9625, "W": 0.0008 },
    "72": { "A": 0.0005, "D": 0.0002, "E": 0.0116, "G": 0.001, "H": 0.0002, "I": 0.0002, "K": 0.9237, "L": 0.0002, "M": 0.0024, "N": 0.0047, "P": 0.0001, "Q": 0.0098, "R": 0.039, "S": 0.0007, "T": 0.0047, "V": 0.0004, "W": 0.0003 },
    "74": { "A": 0.0016, "C": 0.006, "D": 0.006, "E": 0.0001, "F": 0.0001, "G": 0.9782, "H": 0.0001, "I": 0.0001, "N": 0.0001, "R": 0.0014, "S": 0.0025, "T": 0.0001, "V": 0.0035, "Y": 0.0001 },
    "75": { "C": 0.0001, "E": 0.0001, "G": 0.0015, "H": 0.0005, "I": 0.0005, "K": 0.0004, "L": 0.0034, "P": 0.0006, "Q": 0.0019, "R": 0.9894, "S": 0.0009, "T": 0.0003, "W": 0.0004 },
    "76": { "A": 0.0002, "C": 0.0043, "F": 0.9683, "G": 0.0002, "I": 0.0033, "L": 0.0117, "S": 0.0023, "T": 0.0001, "V": 0.0089, "Y": 0.0006 },
    "77": { "A": 0.008, "D": 0.0002, "F": 0.0008, "G": 0.0003, "H": 0.0002, "I": 0.0131, "K": 0.0002, "L": 0.0004, "M": 0.0001, "N": 0.0031, "P": 0.0032, "R": 0.0005, "S": 0.0165, "T": 0.9519, "V": 0.0014, "Y": 0.0002 },
    "78": { "A": 0.0004, "C": 0.0001, "F": 0.0016, "G": 0.0002, "H": 0.0002, "I": 0.9418, "L": 0.0061, "M": 0.004, "N": 0.002, "P": 0.0001, "R": 0.0002, "S": 0.0082, "T": 0.0059, "V": 0.0292 },
    "79": { "A": 0.0115, "C": 0.0004, "D": 0.0002, "F": 0.004, "L": 0.0008, "N": 0.0001, "P": 0.0035, "S": 0.9721, "T": 0.0044, "V": 0.0002, "Y": 0.0028 },
    "80": { "A": 0.0001, "E": 0.0004, "G": 0.0031, "I": 0.0017, "K": 0.005, "L": 0.0002, "M": 0.0002, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.9826, "S": 0.0044, "T": 0.0015, "W": 0.0003 },
    "81": { "A": 0.0021, "D": 0.972, "E": 0.0141, "G": 0.0024, "H": 0.0031, "K": 0.0002, "N": 0.0021, "Q": 0.0001, "S": 0.0001, "T": 0.0002, "V": 0.0012, "Y": 0.0021 },
    "82": { "A": 0.0007, "D": 0.1004, "E": 0.0022, "F": 0.0003, "G": 0.0013, "H": 0.0017, "I": 0.0036, "K": 0.0097, "L": 0.0002, "N": 0.8584, "P": 0.0002, "Q": 0.0004, "R": 0.0005, "S": 0.0104, "T": 0.0076, "V": 0.0004, "Y": 0.0019 },
    "83": { "A": 0.3772, "C": 0.0003, "D": 0.0037, "E": 0.0002, "F": 0.0041, "G": 0.004, "H": 0.0001, "I": 0.0005, "L": 0.0014, "M": 0.0002, "N": 0.0003, "P": 0.0047, "R": 0.0008, "S": 0.5875, "T": 0.0092, "V": 0.0033, "Y": 0.0024 },
    "84": { "A": 0.001, "D": 0.0017, "E": 0.0152, "G": 0.0016, "H": 0.0006, "I": 0.0005, "K": 0.9171, "L": 0.0002, "M": 0.0025, "N": 0.0141, "P": 0.0002, "Q": 0.0112, "R": 0.0282, "S": 0.0004, "T": 0.0048, "V": 0.0002, "W": 0.0002, "Y": 0.0004 },
    "85": { "A": 0.0006, "D": 0.0142, "E": 0.0011, "G": 0.0016, "H": 0.0034, "I": 0.0015, "K": 0.0201, "M": 0.0001, "N": 0.9059, "P": 0.0001, "Q": 0.0005, "R": 0.0032, "S": 0.0404, "T": 0.0061, "V": 0.0001, "Y": 0.0012 },
    "86": { "A": 0.0084, "D": 0.0001, "E": 0.0007, "F": 0.0014, "G": 0.0001, "H": 0.0002, "I": 0.0236, "K": 0.0025, "L": 0.0047, "M": 0.0123, "N": 0.0005, "P": 0.0029, "Q": 0.0001, "R": 0.0028, "S": 0.3472, "T": 0.5901, "V": 0.0022, "Y": 0.0003 },
    "87": { "A": 0.0314, "C": 0.0001, "D": 0.0002, "E": 0.0002, "F": 0.0017, "G": 0.0007, "H": 0.0001, "I": 0.0018, "K": 0.0001, "L": 0.8914, "M": 0.0089, "P": 0.0015, "Q": 0.001, "R": 0.0052, "S": 0.0004, "T": 0.0006, "V": 0.0544, "W": 0.0003 },
    "88": { "A": 0.0002, "C": 0.0019, "D": 0.009, "E": 0.0003, "F": 0.0491, "H": 0.0105, "I": 0.0005, "K": 0.0001, "L": 0.0011, "N": 0.0033, "Q": 0.0001, "R": 0.0001, "S": 0.0171, "T": 0.0007, "V": 0.001, "W": 0.0007, "Y": 0.9043 },
    "89": { "C": 0.0001, "F": 0.0004, "I": 0.0006, "L": 0.9846, "M": 0.0038, "P": 0.0012, "Q": 0.0005, "R": 0.0045, "S": 0.0001, "V": 0.0038, "W": 0.0003 },
    "90": { "A": 0.0002, "D": 0.0019, "E": 0.0197, "G": 0.0003, "H": 0.0106, "K": 0.0078, "L": 0.0025, "N": 0.0002, "P": 0.0007, "Q": 0.9516, "R": 0.0033, "S": 0.0001, "T": 0.0002, "V": 0.0003, "Y": 0.0005 },
    "91": { "A": 0.0002, "F": 0.0001, "G": 0.0002, "I": 0.0102, "K": 0.0006, "L": 0.0152, "M": 0.9604, "R": 0.0055, "S": 0.0002, "T": 0.0012, "V": 0.0058, "W": 0.0002 },
    "92": { "A": 0.0005, "C": 0.0001, "D": 0.028, "E": 0.0009, "F": 0.0001, "G": 0.0056, "H": 0.0063, "I": 0.0019, "K": 0.008, "L": 0.0001, "M": 0.0001, "N": 0.8986, "Q": 0.0003, "R": 0.0015, "S": 0.0338, "T": 0.0119, "V": 0.0004, "Y": 0.0018 },
    "93": { "A": 0.0011, "C": 0.0004, "D": 0.003, "E": 0.0001, "F": 0.0002, "G": 0.0146, "H": 0.0001, "I": 0.0021, "K": 0.0004, "L": 0.0002, "M": 0.0003, "N": 0.0208, "P": 0.0001, "R": 0.0127, "S": 0.9354, "T": 0.0074, "V": 0.0002, "W": 0.0002, "Y": 0.0007 },
    "94": { "A": 0.0001, "I": 0.0001, "L": 0.981, "M": 0.0033, "P": 0.0021, "Q": 0.0007, "R": 0.0075, "T": 0.0001, "V": 0.005, "W": 0.0001 },
    "95": { "A": 0.0004, "C": 0.0001, "D": 0.0001, "E": 0.0025, "G": 0.0125, "I": 0.0021, "K": 0.088, "L": 0.0002, "M": 0.0004, "N": 0.0008, "P": 0.0001, "Q": 0.0013, "R": 0.8644, "S": 0.0094, "T": 0.0174, "V": 0.0003, "W": 0.0001 },
    "96": { "A": 0.7471, "C": 0.0001, "D": 0.0295, "E": 0.0014, "F": 0.0004, "G": 0.0129, "H": 0.0006, "I": 0.0066, "K": 0.0004, "L": 0.0024, "M": 0.0002, "N": 0.001, "P": 0.0266, "R": 0.0012, "S": 0.0101, "T": 0.0982, "V": 0.061, "Y": 0.0004 },
    "97": { "A": 0.0031, "D": 0.0482, "E": 0.9326, "G": 0.0126, "H": 0.0001, "K": 0.001, "M": 0.0001, "N": 0.0003, "Q": 0.0004, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0009, "Y": 0.0001 },
    "98": { "A": 0.0014, "D": 0.9859, "E": 0.0069, "G": 0.0013, "H": 0.0003, "N": 0.0012, "R": 0.0001, "S": 0.0001, "V": 0.0006, "Y": 0.0021 },
    "99": { "A": 0.0059, "E": 0.0001, "G": 0.0001, "I": 0.0001, "K": 0.0011, "L": 0.0001, "M": 0.0078, "N": 0.0001, "P": 0.001, "R": 0.0009, "S": 0.0093, "T": 0.9733, "V": 0.0001 },
    "100": { "A": 0.9795, "C": 0.0001, "D": 0.001, "G": 0.0134, "P": 0.0003, "R": 0.0002, "S": 0.0026, "T": 0.0014, "V": 0.0014 },
    "101": { "A": 0.0025, "C": 0.0001, "D": 0.0005, "E": 0.0007, "F": 0.0064, "G": 0.0028, "H": 0.0002, "I": 0.0328, "K": 0.0005, "L": 0.0936, "M": 0.0116, "P": 0.0002, "Q": 0.0002, "R": 0.0013, "S": 0.0003, "T": 0.0059, "V": 0.8402, "W": 0.0003, "Y": 0.0001 },
    "102": { "C": 0.0007, "D": 0.0033, "E": 0.0002, "F": 0.0043, "H": 0.0016, "I": 0.0002, "L": 0.0001, "N": 0.0005, "S": 0.0006, "Y": 0.9885 },
    "103": { "A": 0.0001, "C": 0.0014, "D": 0.0032, "F": 0.033, "H": 0.0084, "I": 0.0001, "L": 0.0005, "N": 0.0005, "Q": 0.0002, "R": 0.0001, "S": 0.0023, "T": 0.0004, "V": 0.0002, "W": 0.0005, "Y": 0.9491 },
    "104": { "A": 0.0001, "C": 0.9873, "D": 0.0001, "F": 0.0014, "G": 0.0041, "I": 0.0001, "L": 0.0003, "R": 0.0008, "S": 0.0009, "T": 0.0002, "V": 0.0004, "W": 0.0038, "Y": 0.0006 },
    "105": { "A": 0.8387, "C": 0.0008, "D": 0.0005, "E": 0.0013, "F": 0.0002, "G": 0.009, "H": 0.0002, "I": 0.0032, "K": 0.0006, "L": 0.0019, "M": 0.0011, "N": 0.0011, "P": 0.0005, "Q": 0.0004, "R": 0.0011, "S": 0.0093, "T": 0.0824, "V": 0.0473, "W": 0.0003, "Y": 0.0003 },
    "106": { "A": 0.0079, "C": 0.0011, "D": 0.002, "E": 0.0021, "F": 0.0011, "G": 0.016, "H": 0.0013, "I": 0.0053, "K": 0.3064, "L": 0.0039, "M": 0.0017, "N": 0.0116, "P": 0.0034, "Q": 0.0031, "R": 0.5197, "S": 0.0379, "T": 0.0678, "V": 0.0039, "W": 0.0013, "Y": 0.0024 },
    "107": { "A": 0.0728, "C": 0.0045, "D": 0.355, "E": 0.1003, "F": 0.009, "G": 0.1645, "H": 0.0099, "I": 0.0099, "K": 0.0076, "L": 0.0284, "M": 0.0049, "N": 0.0126, "P": 0.0214, "Q": 0.011, "R": 0.0274, "S": 0.0484, "T": 0.021, "V": 0.0678, "W": 0.0094, "Y": 0.0143 },
    "108": { "A": 0.0521, "C": 0.0071, "D": 0.0391, "E": 0.0322, "F": 0.0256, "G": 0.1554, "H": 0.0249, "I": 0.041, "K": 0.0238, "L": 0.0931, "M": 0.0169, "N": 0.0223, "P": 0.0798, "Q": 0.0305, "R": 0.1175, "S": 0.0847, "T": 0.039, "V": 0.0465, "W": 0.0203, "Y": 0.0482 },
    "109": { "A": 0.0582, "C": 0.0151, "D": 0.0593, "E": 0.0353, "F": 0.0211, "G": 0.1767, "H": 0.0183, "I": 0.0322, "K": 0.0155, "L": 0.051, "M": 0.0109, "N": 0.0234, "P": 0.05, "Q": 0.0218, "R": 0.078, "S": 0.1046, "T": 0.0514, "V": 0.0609, "W": 0.0263, "Y": 0.0899 },
    "110": { "A": 0.0626, "C": 0.0141, "D": 0.0678, "E": 0.0285, "F": 0.0272, "G": 0.1555, "H": 0.0133, "I": 0.0367, "K": 0.0098, "L": 0.0456, "M": 0.0113, "N": 0.0228, "P": 0.0441, "Q": 0.0194, "R": 0.0536, "S": 0.1217, "T": 0.0551, "V": 0.067, "W": 0.0294, "Y": 0.1146 },
    "111": { "A": 0.0651, "C": 0.0185, "D": 0.0654, "E": 0.0214, "F": 0.03, "G": 0.1476, "H": 0.0088, "I": 0.0368, "K": 0.0091, "L": 0.0458, "M": 0.0121, "N": 0.0233, "P": 0.0246, "Q": 0.0169, "R": 0.046, "S": 0.1481, "T": 0.0552, "V": 0.0657, "W": 0.0332, "Y": 0.1264 },
    "111A": { "A": 0.0593, "C": 0.0197, "D": 0.0717, "E": 0.0185, "F": 0.0315, "G": 0.1454, "H": 0.008, "I": 0.0354, "K": 0.0071, "L": 0.0472, "M": 0.0123, "N": 0.0203, "P": 0.0307, "Q": 0.013, "R": 0.0337, "S": 0.152, "T": 0.0574, "V": 0.0709, "W": 0.0449, "Y": 0.1212 },
    "111B": { "A": 0.0522, "C": 0.0205, "D": 0.0727, "E": 0.0186, "F": 0.0391, "G": 0.1416, "H": 0.0065, "I": 0.0308, "K": 0.0052, "L": 0.0452, "M": 0.0109, "N": 0.0238, "P": 0.0321, "Q": 0.0129, "R": 0.0333, "S": 0.1786, "T": 0.0534, "V": 0.0694, "W": 0.0468, "Y": 0.1062 },
    "111C": { "A": 0.044, "C": 0.0329, "D": 0.0599, "E": 0.017, "F": 0.037, "G": 0.169, "H": 0.0092, "I": 0.0287, "K": 0.0084, "L": 0.0365, "M": 0.01, "N": 0.0178, "P": 0.0278, "Q": 0.0097, "R": 0.032, "S": 0.1857, "T": 0.061, "V": 0.0702, "W": 0.0512, "Y": 0.0919 },
    "111D": { "A": 0.0448, "C": 0.0353, "D": 0.0467, "E": 0.0227, "F": 0.0441, "G": 0.1551, "H": 0.0076, "I": 0.0271, "K": 0.0076, "L": 0.046, "M": 0.0095, "N": 0.0145, "P": 0.0448, "Q": 0.0082, "R": 0.0202, "S": 0.1854, "T": 0.0605, "V": 0.0643, "W": 0.0429, "Y": 0.1129 },
    "111E": { "A": 0.0577, "C": 0.0692, "D": 0.028, "E": 0.0231, "F": 0.0247, "G": 0.1631, "H": 0.0082, "I": 0.0313, "K": 0.0033, "L": 0.0445, "M": 0.0049, "N": 0.0066, "P": 0.0379, "Q": 0.0082, "R": 0.0313, "S": 0.1351, "T": 0.056, "V": 0.0923, "W": 0.0478, "Y": 0.1269 },
    "111F": { "A": 0.05, "C": 0.0556, "D": 0.0333, "E": 0.0389, "F": 0.0167, "G": 0.1722, "H": 0.0056, "I": 0.0278, "K": 0.0056, "L": 0.05, "P": 0.05, "R": 0.0222, "S": 0.1333, "T": 0.0611, "V": 0.0611, "W": 0.0333, "Y": 0.1833 },
    "112F": { "A": 0.0841, "C": 0.0405, "D": 0.0374, "E": 0.0125, "F": 0.028, "G": 0.1308, "H": 0.0093, "I": 0.0343, "K": 0.0187, "L": 0.0592, "M": 0.0062, "N": 0.0249, "P": 0.0841, "Q": 0.0093, "R": 0.0592, "S": 0.1059, "T": 0.0343, "V": 0.0592, "W": 0.0312, "Y": 0.1308 },
    "112E": { "A": 0.0554, "C": 0.0386, "D": 0.0277, "E": 0.0297, "F": 0.0188, "G": 0.1525, "H": 0.0119, "I": 0.0198, "K": 0.0178, "L": 0.0515, "M": 0.0109, "N": 0.0168, "P": 0.0822, "Q": 0.0119, "R": 0.0545, "S": 0.1178, "T": 0.0495, "V": 0.0515, "W": 0.0356, "Y": 0.1455 },
    "112D": { "A": 0.0547, "C": 0.0308, "D": 0.0441, "E": 0.0259, "F": 0.0263, "G": 0.1591, "H": 0.0134, "I": 0.0316, "K": 0.0097, "L": 0.0421, "M": 0.0093, "N": 0.0267, "P": 0.0591, "Q": 0.0126, "R": 0.0482, "S": 0.1429, "T": 0.051, "V": 0.0623, "W": 0.0255, "Y": 0.1247 },
    "112C": { "A": 0.051, "C": 0.0326, "D": 0.0536, "E": 0.0159, "F": 0.0232, "G": 0.1416, "H": 0.0155, "I": 0.0262, "K": 0.0123, "L": 0.0449, "M": 0.007, "N": 0.0258, "P": 0.051, "Q": 0.0085, "R": 0.0399, "S": 0.144, "T": 0.053, "V": 0.0564, "W": 0.0336, "Y": 0.1639 },
    "112B": { "A": 0.0496, "C": 0.0247, "D": 0.0535, "E": 0.0232, "F": 0.0251, "G": 0.1314, "H": 0.0157, "I": 0.0256, "K": 0.0092, "L": 0.0409, "M": 0.0066, "N": 0.0239, "P": 0.0397, "Q": 0.0134, "R": 0.0328, "S": 0.1207, "T": 0.0483, "V": 0.0504, "W": 0.0302, "Y": 0.2352 },
    "112A": { "A": 0.0591, "C": 0.0147, "D": 0.0386, "E": 0.0232, "F": 0.0246, "G": 0.1253, "H": 0.0136, "I": 0.0231, "K": 0.0092, "L": 0.0433, "M": 0.0075, "N": 0.0274, "P": 0.047, "Q": 0.0143, "R": 0.0451, "S": 0.117, "T": 0.0547, "V": 0.0574, "W": 0.0347, "Y": 0.2201 },
    "112": { "A": 0.0595, "C": 0.0156, "D": 0.0441, "E": 0.0246, "F": 0.0247, "G": 0.1288, "H": 0.0209, "I": 0.0208, "K": 0.0106, "L": 0.053, "M": 0.0066, "N": 0.03, "P": 0.0544, "Q": 0.0153, "R": 0.0419, "S": 0.1113, "T": 0.0483, "V": 0.0461, "W": 0.0356, "Y": 0.208 },
    "113": { "A": 0.0424, "C": 0.0105, "D": 0.0962, "E": 0.0263, "F": 0.0208, "G": 0.1659, "H": 0.0212, "I": 0.0105, "K": 0.0086, "L": 0.0253, "M": 0.0048, "N": 0.0525, "P": 0.0296, "Q": 0.0096, "R": 0.0387, "S": 0.0821, "T": 0.0393, "V": 0.0245, "W": 0.0486, "Y": 0.2426 },
    "114": { "A": 0.1514, "C": 0.0109, "D": 0.0284, "E": 0.0172, "F": 0.0204, "G": 0.2072, "H": 0.0194, "I": 0.0136, "K": 0.0051, "L": 0.0332, "M": 0.0038, "N": 0.0175, "P": 0.0636, "Q": 0.0088, "R": 0.0189, "S": 0.0529, "T": 0.0323, "V": 0.027, "W": 0.0631, "Y": 0.2051 },
    "115": { "A": 0.0152, "C": 0.0067, "D": 0.0089, "E": 0.0081, "F": 0.4837, "G": 0.0339, "H": 0.0063, "I": 0.0328, "K": 0.0025, "L": 0.0893, "M": 0.1585, "N": 0.0093, "P": 0.0281, "Q": 0.0035, "R": 0.009, "S": 0.0263, "T": 0.0133, "V": 0.0286, "W": 0.0095, "Y": 0.0263 },
    "116": { "A": 0.0139, "C": 0.0017, "D": 0.8228, "E": 0.0174, "F": 0.0043, "G": 0.0436, "H": 0.0073, "I": 0.0036, "K": 0.002, "L": 0.0059, "M": 0.0009, "N": 0.011, "P": 0.0074, "Q": 0.0145, "R": 0.0051, "S": 0.0113, "T": 0.0068, "V": 0.0092, "W": 0.0029, "Y": 0.0086 },
    "117": { "A": 0.0053, "C": 0.0061, "D": 0.0149, "E": 0.0019, "F": 0.0216, "G": 0.0068, "H": 0.0303, "I": 0.1114, "K": 0.0015, "L": 0.0373, "M": 0.0023, "N": 0.0134, "P": 0.064, "Q": 0.0029, "R": 0.0037, "S": 0.0322, "T": 0.0061, "V": 0.1771, "W": 0.0017, "Y": 0.4595 },
    "118": { "A": 0.0003, "C": 0.0015, "D": 0.0001, "E": 0.0002, "F": 0.0007, "G": 0.0041, "I": 0.0003, "K": 0.0001, "L": 0.0027, "M": 0.0002, "N": 0.0001, "P": 0.0004, "Q": 0.0001, "R": 0.002, "S": 0.0008, "T": 0.0004, "V": 0.0021, "W": 0.9835, "Y": 0.0003 },
    "119": { "A": 0.0011, "C": 0.0014, "D": 0.0012, "E": 0.0001, "F": 0.0001, "G": 0.9915, "L": 0.0002, "N": 0.0002, "P": 0.0001, "R": 0.0002, "S": 0.0026, "T": 0.0002, "V": 0.0011 },
    "120": { "D": 0.0001, "E": 0.001, "G": 0.0003, "H": 0.0058, "K": 0.0314, "L": 0.0049, "N": 0.0006, "P": 0.0077, "Q": 0.9061, "R": 0.0416, "S": 0.0001, "T": 0.0004 },
    "121": { "A": 0.0003, "E": 0.0007, "G": 0.9967, "R": 0.0007, "V": 0.0014, "W": 0.0001 },
    "122": { "A": 0.0186, "D": 0.0001, "E": 0.0001, "F": 0.0001, "I": 0.0089, "K": 0.0003, "L": 0.0002, "M": 0.0004, "N": 0.0021, "P": 0.0015, "R": 0.0002, "S": 0.0073, "T": 0.9573, "V": 0.0029 },
    "123": { "A": 0.0021, "E": 0.0002, "F": 0.0001, "G": 0.0002, "H": 0.0002, "I": 0.0007, "K": 0.0012, "L": 0.6834, "M": 0.1155, "P": 0.0059, "Q": 0.0084, "R": 0.0054, "S": 0.0012, "T": 0.1695, "V": 0.0058, "W": 0.0001 },
    "124": { "A": 0.0018, "C": 0.0001, "D": 0.0003, "F": 0.0005, "G": 0.0016, "I": 0.0026, "L": 0.0019, "V": 0.9911 },
    "125": { "A": 0.0061, "F": 0.0002, "H": 0.0005, "I": 0.014, "L": 0.0003, "M": 0.0001, "N": 0.0019, "P": 0.001, "R": 0.0002, "S": 0.0092, "T": 0.9653, "V": 0.0009 },
    "126": { "A": 0.0028, "D": 0.0007, "F": 0.0006, "G": 0.0015, "I": 0.0026, "L": 0.0004, "R": 0.0007, "S": 0.0001, "V": 0.9906 },
    "127": { "A": 0.0049, "C": 0.0002, "F": 0.0016, "H": 0.0001, "L": 0.0009, "P": 0.0022, "S": 0.9875, "T": 0.0012, "V": 0.0001, "Y": 0.0011 },
    "128": { "A": 0.0061, "F": 0.0001, "L": 0.0009, "P": 0.0051, "Q": 0.0002, "S": 0.9863, "T": 0.0013 }
  },
  "IGHV4": {
    "1": { "H": 0.0056, "Q": 0.9938, "R": 0.0004, "V": 0.0001 },
    "2": { "A": 0.0036, "D": 0.0002, "E": 0.0016, "F": 0.0006, "G": 0.0033, "I": 0.0012, "K": 0.0001, "L": 0.2155, "M": 0.0056, "P": 0.0019, "Q": 0.0007, "R": 0.0004, "S": 0.0001, "V": 0.7649, "W": 0.0002 },
    "3": { "A": 0.0001, "D": 0.0001, "E": 0.0029, "F": 0.0001, "H": 0.0226, "K": 0.0024, "L": 0.006, "N": 0.0005, "P": 0.0071, "Q": 0.9469, "R": 0.0102, "S": 0.0001, "T": 0.0001, "V": 0.0002, "W": 0.0001, "Y": 0.0004 },
    "4": { "A": 0.0001, "F": 0.0006, "H": 0.0001, "I": 0.0008, "L": 0.987, "M": 0.0026, "P": 0.0024, "Q": 0.0006, "R": 0.0013, "S": 0.0001, "V": 0.0042, "W": 0.0001 },
    "5": { "A": 0.0002, "D": 0.0002, "E": 0.0048, "G": 0.0003, "H": 0.0084, "I": 0.0001, "K": 0.0029, "L": 0.0042, "M": 0.0001, "N": 0.0001, "P": 0.0073, "Q": 0.9349, "R": 0.0104, "T": 0.0003, "V": 0.0254, "W": 0.0002, "Y": 0.0002 },
    "6": { "A": 0.0114, "D": 0.0036, "E": 0.7453, "G": 0.0037, "H": 0.0008, "K": 0.0009, "L": 0.0007, "M": 0.0001, "P": 0.0035, "Q": 0.2269, "R": 0.0009, "S": 0.0002, "T": 0.0001, "V": 0.0019, "W": 0.0001 },
    "7": { "A": 0.0033, "C": 0.0022, "F": 0.0001, "G": 0.0013, "L": 0.0023, "P": 0.0039, "Q": 0.0001, "R": 0.0025, "S": 0.7858, "T": 0.001, "V": 0.0003, "W": 0.197, "Y": 0.0001 },
    "8": { "A": 0.0043, "C": 0.0054, "D": 0.0019, "E": 0.0001, "F": 0.0001, "G": 0.9788, "L": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.0038, "S": 0.0014, "T": 0.0001, "V": 0.0039 },
    "9": { "A": 0.2069, "D": 0.0001, "E": 0.0009, "G": 0.0149, "H": 0.0002, "I": 0.0001, "L": 0.0009, "P": 0.7492, "Q": 0.0016, "R": 0.0005, "S": 0.0199, "T": 0.0046, "V": 0.0003 },
    "11": { "A": 0.0031, "C": 0.0001, "D": 0.0008, "E": 0.009, "G": 0.9654, "H": 0.0001, "L": 0.0001, "N": 0.0002, "Q": 0.0004, "R": 0.0164, "S": 0.0009, "T": 0.0005, "V": 0.0028, "W": 0.0001 },
    "12": { "A": 0.0003, "F": 0.0004, "G": 0.0001, "H": 0.0002, "I": 0.0006, "K": 0.0001, "L": 0.9657, "M": 0.0028, "P": 0.0051, "Q": 0.0026, "R": 0.0066, "S": 0.0003, "T": 0.0001, "V": 0.0148, "W": 0.0003 },
    "13": { "A": 0.0067, "C": 0.0002, "E": 0.0012, "F": 0.0029, "G": 0.0038, "I": 0.0022, "K": 0.0071, "L": 0.2058, "M": 0.0028, "Q": 0.0001, "R": 0.0003, "S": 0.001, "T": 0.0003, "V": 0.7641, "W": 0.0015 },
    "14": { "A": 0.0006, "D": 0.0002, "E": 0.0069, "F": 0.0001, "G": 0.0002, "H": 0.0004, "I": 0.0001, "K": 0.9084, "L": 0.0005, "M": 0.0024, "N": 0.0057, "P": 0.0007, "Q": 0.0278, "R": 0.0262, "S": 0.0004, "T": 0.0189, "V": 0.0003, "W": 0.0001, "Y": 0.0001 },
    "15": { "A": 0.0069, "D": 0.0001, "F": 0.0002, "H": 0.0028, "L": 0.0029, "P": 0.9714, "Q": 0.0002, "R": 0.0007, "S": 0.0098, "T": 0.0048, "V": 0.0002, "Y": 0.0001 },
    "16": { "A": 0.0066, "E": 0.0002, "F": 0.0002, "G": 0.022, "H": 0.0001, "L": 0.0039, "P": 0.0054, "Q": 0.0002, "R": 0.0001, "S": 0.959, "T": 0.0009, "W": 0.0013 },
    "17": { "A": 0.0176, "D": 0.0065, "E": 0.73, "G": 0.0542, "H": 0.0007, "K": 0.0017, "L": 0.0002, "P": 0.0023, "Q": 0.1799, "R": 0.0038, "S": 0.0009, "T": 0.0003, "V": 0.0014, "W": 0.0002, "Y": 0.0001 },
    "18": { "A": 0.0043, "D": 0.0002, "G": 0.0001, "H": 0.0007, "I": 0.002, "K": 0.0001, "L": 0.0001, "N": 0.0032, "P": 0.0144, "R": 0.0001, "S": 0.0251, "T": 0.9493, "V": 0.0001, "Y": 0.0002 },
    "19": { "A": 0.0001, "F": 0.0001, "H": 0.0001, "L": 0.9839, "M": 0.0029, "P": 0.0039, "Q": 0.0006, "R": 0.0036, "S": 0.0001, "T": 0.0001, "V": 0.0046, "W": 0.0001 },
    "20": { "A": 0.0052, "C": 0.0009, "D": 0.0001, "E": 0.0001, "F": 0.0024, "G": 0.0001, "H": 0.0001, "K": 0.0062, "L": 0.0005, "P": 0.0044, "Q": 0.0001, "R": 0.015, "S": 0.9577, "T": 0.0036, "V": 0.0003, "W": 0.0001, "Y": 0.0031 },
    "21": { "A": 0.0001, "C": 0.0001, "F": 0.0014, "H": 0.0014, "I": 0.008, "L": 0.9634, "P": 0.0139, "R": 0.005, "S": 0.0005, "T": 0.0005, "V": 0.0054, "Y": 0.0001 },
    "22": { "A": 0.0046, "D": 0.0001, "F": 0.0002, "G": 0.0001, "H": 0.0004, "I": 0.01, "K": 0.0007, "L": 0.0003, "M": 0.0012, "N": 0.005, "P": 0.0231, "R": 0.0011, "S": 0.0359, "T": 0.9167, "V": 0.0004, "Y": 0.0001 },
    "23": { "A": 0.0003, "C": 0.9777, "F": 0.0037, "G": 0.0047, "H": 0.0001, "L": 0.0002, "P": 0.0001, "R": 0.0056, "S": 0.0049, "V": 0.0003, "W": 0.0011, "Y": 0.0013 },
    "24": { "A": 0.3011, "C": 0.0002, "D": 0.0058, "E": 0.0012, "F": 0.0007, "G": 0.0094, "H": 0.0004, "I": 0.0079, "K": 0.0071, "L": 0.0012, "M": 0.0001, "N": 0.0093, "P": 0.0236, "Q": 0.0004, "R": 0.0026, "S": 0.0534, "T": 0.5658, "V": 0.0095, "Y": 0.0001 },
    "25": { "A": 0.0249, "C": 0.0007, "D": 0.0017, "E": 0.0001, "F": 0.0068, "G": 0.0101, "I": 0.0066, "L": 0.0054, "M": 0.0013, "P": 0.0002, "R": 0.0002, "S": 0.0007, "T": 0.0007, "V": 0.9404, "W": 0.0001, "Y": 0.0001 },
    "26": { "A": 0.0107, "C": 0.0013, "D": 0.0034, "E": 0.0002, "F": 0.0046, "G": 0.0002, "H": 0.0051, "I": 0.0002, "K": 0.0002, "L": 0.001, "N": 0.0017, "P": 0.0355, "Q": 0.0002, "R": 0.0004, "S": 0.7424, "T": 0.0034, "V": 0.0003, "Y": 0.1891 },
    "27": { "A": 0.0089, "C": 0.003, "D": 0.0098, "E": 0.0006, "F": 0.0002, "G": 0.9478, "H": 0.001, "I": 0.0001, "K": 0.0001, "L": 0.0003, "N": 0.003, "P": 0.0003, "Q": 0.0001, "R": 0.0056, "S": 0.0072, "T": 0.0009, "V": 0.0103, "W": 0.0003, "Y": 0.0003 },
    "28": { "A": 0.0409, "C": 0.0103, "D": 0.0586, "E": 0.0051, "F": 0.0135, "G": 0.7824, "H": 0.0013, "I": 0.0004, "L": 0.0009, "M": 0.0001, "N": 0.0008, "P": 0.0007, "Q": 0.0003, "R": 0.0047, "S": 0.0061, "T": 0.003, "V": 0.0162, "W": 0.0031, "Y": 0.0514 },
    "29": { "A": 0.01, "C": 0.001, "D": 0.0004, "E": 0.0001, "F": 0.0066, "G": 0.0003, "H": 0.0006, "I": 0.0012, "L": 0.0013, "M": 0.0001, "N": 0.0007, "P": 0.0312, "Q": 0.0001, "R": 0.0006, "S": 0.9249, "T": 0.0159, "V": 0.0003, "Y": 0.0045 },
    "30": { "A": 0.0026, "C": 0.0016, "D": 0.0002, "F": 0.2228, "G": 0.0004, "H": 0.0004, "I": 0.644, "K": 0.0001, "L": 0.0381, "M": 0.0133, "N": 0.0012, "P": 0.0007, "R": 0.0006, "S": 0.007, "T": 0.0074, "V": 0.059, "Y": 0.0005 },
    "31": { "A": 0.0026, "C": 0.0021, "D": 0.0061, "E": 0.0005, "F": 0.0017, "G": 0.0153, "H": 0.0012, "I": 0.0121, "K": 0.0033, "L": 0.0014, "M": 0.0005, "N": 0.0412, "P": 0.0027, "Q": 0.0003, "R": 0.052, "S": 0.8014, "T": 0.0526, "V": 0.0006, "W": 0.0001, "Y": 0.0023 },
    "34": { "A": 0.0017, "C": 0.0018, "D": 0.0063, "E": 0.0002, "F": 0.0006, "G": 0.023, "H": 0.0017, "I": 0.008, "K": 0.0011, "L": 0.0012, "M": 0.0003, "N": 0.027, "P": 0.0014, "Q": 0.0002, "R": 0.0546, "S": 0.8311, "T": 0.037, "V": 0.0019, "W": 0.0001, "Y": 0.0009 },
    "35": { "A": 0.0089, "C": 0.0027, "D": 0.0148, "E": 0.0027, "F": 0.0014, "G": 0.1937, "H": 0.0017, "I": 0.0071, "K": 0.0023, "L": 0.0013, "M": 0.0007, "N": 0.0328, "P": 0.004, "Q": 0.0003, "R": 0.0397, "S": 0.6384, "T": 0.0414, "V": 0.004, "W": 0.0003, "Y": 0.0018 },
    "36": { "A": 0.0163, "C": 0.0035, "D": 0.0667, "E": 0.0028, "F": 0.0025, "G": 0.3476, "H": 0.0055, "I": 0.0068, "K": 0.0028, "L": 0.0026, "M": 0.0004, "N": 0.052, "P": 0.0072, "Q": 0.0004, "R": 0.0249, "S": 0.4022, "T": 0.0433, "V": 0.0055, "W": 0.0002, "Y": 0.0069 },
    "37": { "A": 0.002, "C": 0.0026, "D": 0.0179, "E": 0.0004, "F": 0.0278, "G": 0.0007, "H": 0.0428, "I": 0.0011, "K": 0.0021, "L": 0.0021, "N": 0.0508, "P": 0.0015, "Q": 0.0018, "R": 0.0012, "S": 0.039, "T": 0.0038, "V": 0.0011, "W": 0.0006, "Y": 0.8007 },
    "38": { "A": 0.0032, "C": 0.0052, "D": 0.0044, "E": 0.0008, "F": 0.0343, "G": 0.0027, "H": 0.0193, "I": 0.0004, "K": 0.0006, "L": 0.0028, "N": 0.0039, "P": 0.0013, "Q": 0.0016, "R": 0.0012, "S": 0.0478, "T": 0.0012, "V": 0.0004, "W": 0.0516, "Y": 0.8172 },
    "39": { "C": 0.0094, "F": 0.0005, "G": 0.0067, "I": 0.0059, "L": 0.0086, "M": 0.0103, "R": 0.0054, "S": 0.0019, "T": 0.0001, "V": 0.001, "W": 0.9489, "Y": 0.0012 },
    "40": { "A": 0.0217, "C": 0.0064, "D": 0.0043, "F": 0.0001, "G": 0.2628, "H": 0.0045, "I": 0.0052, "K": 0.0002, "L": 0.0006, "M": 0.0002, "N": 0.0392, "P": 0.0011, "Q": 0.0002, "R": 0.0216, "S": 0.5731, "T": 0.0539, "V": 0.0042, "Y": 0.0006 },
    "41": { "A": 0.0001, "C": 0.0053, "F": 0.0003, "G": 0.0046, "L": 0.0036, "R": 0.0041, "S": 0.0009, "V": 0.0002, "W": 0.9805, "Y": 0.0002 },
    "42": { "A": 0.0006, "C": 0.0002, "D": 0.0003, "F": 0.0107, "G": 0.0005, "H": 0.0006, "I": 0.8452, "L": 0.0317, "M": 0.0024, "N": 0.0026, "P": 0.0004, "R": 0.0005, "S": 0.0056, "T": 0.0078, "V": 0.0908, "Y": 0.0002 },
    "43": { "A": 0.0001, "C": 0.0014, "G": 0.0018, "H": 0.0009, "L": 0.0018, "M": 0.0001, "P": 0.0024, "Q": 0.0003, "R": 0.9882, "S": 0.0019, "T": 0.0001, "V": 0.0001, "W": 0.0009 },
    "44": { "A": 0.0002, "E": 0.0013, "G": 0.0001, "H": 0.0087, "K": 0.0038, "L": 0.004, "N": 0.0002, "P": 0.0183, "Q": 0.9548, "R": 0.0073, "S": 0.0004, "T": 0.0007, "V": 0.0001, "W": 0.0001 },
    "45": { "A": 0.0249, "C": 0.0001, "D": 0.0004, "F": 0.002, "H": 0.0769, "I": 0.0001, "K": 0.0002, "L": 0.004, "M": 0.0014, "N": 0.0006, "P": 0.8046, "Q": 0.0009, "R": 0.0044, "S": 0.0602, "T": 0.0165, "V": 0.0014, "Y": 0.0013 },
    "46": { "A": 0.0811, "D": 0.0004, "E": 0.0002, "H": 0.0006, "L": 0.0012, "P": 0.906, "Q": 0.0033, "R": 0.001, "S": 0.0022, "T": 0.0031, "V": 0.0007 },
    "47": { "A": 0.0027, "C": 0.0001, "D": 0.0019, "E": 0.0093, "G": 0.974, "H": 0.0001, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0007, "R": 0.0048, "S": 0.0002, "T": 0.0005, "V": 0.0036, "W": 0.0013 },
    "48": { "A": 0.0005, "D": 0.0001, "E": 0.0109, "G": 0.0023, "H": 0.0003, "I": 0.0001, "K": 0.9186, "L": 0.0003, "M": 0.0034, "N": 0.0033, "P": 0.0005, "Q": 0.0223, "R": 0.0208, "S": 0.0005, "T": 0.0156, "V": 0.0003, "W": 0.0001 },
    "49": { "A": 0.0077, "C": 0.0005, "D": 0.0019, "E": 0.0058, "G": 0.968, "H": 0.0002, "K": 0.0006, "L": 0.0001, "N": 0.0004, "P": 0.0007, "Q": 0.0004, "R": 0.007, "S": 0.001, "T": 0.0012, "V": 0.003, "W": 0.0014 },
    "50": { "A": 0.0003, "F": 0.0004, "G": 0.0002, "I": 0.0002, "K": 0.0001, "L": 0.9698, "M": 0.0048, "P": 0.0109, "Q": 0.0009, "R": 0.01, "S": 0.0001, "T": 0.0002, "V": 0.0017, "W": 0.0003 },
    "51": { "A": 0.0135, "D": 0.0087, "E": 0.9479, "G": 0.0146, "H": 0.0001, "K": 0.0022, "P": 0.0001, "Q": 0.0082, "R": 0.0002, "S": 0.0003, "T": 0.0002, "V": 0.0036, "W": 0.0002, "Y": 0.0001 },
    "52": { "A": 0.0001, "C": 0.0059, "E": 0.0001, "F": 0.0021, "G": 0.0108, "L": 0.0054, "R": 0.0033, "S": 0.0019, "T": 0.0001, "V": 0.0011, "W": 0.9644, "Y": 0.0047 },
    "53": { "A": 0.0001, "C": 0.0004, "D": 0.0001, "F": 0.005, "G": 0.0005, "H": 0.0004, "I": 0.9214, "L": 0.0252, "M": 0.0125, "N": 0.0019, "P": 0.0001, "R": 0.0005, "S": 0.0088, "T": 0.0034, "V": 0.0194, "W": 0.0001, "Y": 0.0001 },
    "54": { "A": 0.0279, "C": 0.0004, "D": 0.0003, "E": 0.0012, "G": 0.9579, "L": 0.0002, "P": 0.0001, "R": 0.0017, "S": 0.0009, "T": 0.0005, "V": 0.0041, "W": 0.0048 },
    "55": { "A": 0.0036, "C": 0.008, "D": 0.0119, "E": 0.2516, "F": 0.0111, "G": 0.0098, "H": 0.0152, "I": 0.0026, "K": 0.0027, "L": 0.0058, "M": 0.0003, "N": 0.0214, "P": 0.001, "Q": 0.0096, "R": 0.0794, "S": 0.2266, "T": 0.0233, "V": 0.0016, "W": 0.0009, "Y": 0.3139 },
    "56": { "A": 0.0031, "C": 0.0004, "D": 0.0002, "E": 0.0001, "F": 0.0134, "G": 0.0014, "H": 0.0006, "I": 0.8559, "K": 0.0004, "L": 0.0215, "M": 0.0195, "N": 0.0031, "P": 0.0004, "Q": 0.0001, "R": 0.0013, "S": 0.013, "T": 0.0113, "V": 0.0537, "W": 0.0002, "Y": 0.0003 },
    "57": { "A": 0.0018, "C": 0.0037, "D": 0.0149, "E": 0.0009, "F": 0.0344, "G": 0.0013, "H": 0.0515, "I": 0.0025, "K": 0.0041, "L": 0.004, "M": 0.0003, "N": 0.2021, "P": 0.0025, "Q": 0.003, "R": 0.0051, "S": 0.0483, "T": 0.0057, "V": 0.0009, "W": 0.0011, "Y": 0.6119 },
    "58": { "A": 0.0058, "C": 0.0019, "D": 0.0217, "E": 0.0034, "F": 0.0137, "G": 0.0033, "H": 0.3194, "I": 0.0031, "K": 0.0028, "L": 0.0035, "M": 0.0002, "N": 0.0179, "P": 0.0144, "Q": 0.0105, "R": 0.0078, "S": 0.034, "T": 0.0582, "V": 0.0016, "W": 0.0003, "Y": 0.4763 },
    "59": { "A": 0.0037, "C": 0.0032, "D": 0.0082, "E": 0.0006, "F": 0.0017, "G": 0.0272, "H": 0.0011, "I": 0.0086, "K": 0.0017, "L": 0.002, "M": 0.0007, "N": 0.0243, "P": 0.0027, "Q": 0.0003, "R": 0.0409, "S": 0.8072, "T": 0.0622, "V": 0.0022, "W": 0.0008, "Y": 0.0008 },
    "62": { "A": 0.0164, "C": 0.0055, "D": 0.0437, "F": 0.0164, "G": 0.4317, "K": 0.0055, "L": 0.0109, "N": 0.0328, "P": 0.0055, "R": 0.0164, "S": 0.306, "T": 0.1038, "V": 0.0055 },
    "63": { "A": 0.0051, "C": 0.0004, "D": 0.0016, "E": 0.0092, "F": 0.0001, "G": 0.9578, "H": 0.0002, "I": 0.0003, "K": 0.0006, "L": 0.0007, "N": 0.0003, "P": 0.0003, "Q": 0.0006, "R": 0.0061, "S": 0.0015, "T": 0.0012, "V": 0.0065, "W": 0.0073, "Y": 0.0001 },
    "64": { "A": 0.0118, "C": 0.0016, "D": 0.0112, "E": 0.0018, "F": 0.0023, "G": 0.0342, "H": 0.0035, "I": 0.0133, "K": 0.0073, "L": 0.0018, "M": 0.0011, "N": 0.0538, "P": 0.0034, "Q": 0.0009, "R": 0.0431, "S": 0.7353, "T": 0.0639, "V": 0.0044, "W": 0.0004, "Y": 0.0047 },
    "65": { "A": 0.0235, "C": 0.0001, "D": 0.0007, "E": 0.0002, "F": 0.0007, "G": 0.0014, "H": 0.0014, "I": 0.0103, "K": 0.0009, "L": 0.002, "M": 0.0005, "N": 0.0074, "P": 0.0301, "R": 0.0023, "S": 0.0203, "T": 0.8944, "V": 0.003, "Y": 0.0008 },
    "66": { "A": 0.0033, "C": 0.0011, "D": 0.0223, "E": 0.0021, "F": 0.0148, "G": 0.0009, "H": 0.0249, "I": 0.0044, "K": 0.0289, "L": 0.002, "M": 0.0009, "N": 0.4736, "P": 0.0031, "Q": 0.0037, "R": 0.0048, "S": 0.0266, "T": 0.0184, "V": 0.0017, "W": 0.0006, "Y": 0.3618 },
    "67": { "A": 0.0009, "C": 0.0043, "D": 0.0116, "E": 0.0007, "F": 0.0092, "G": 0.0002, "H": 0.0153, "I": 0.0015, "K": 0.0007, "L": 0.0021, "N": 0.0061, "P": 0.0018, "Q": 0.002, "R": 0.0012, "S": 0.025, "T": 0.0022, "V": 0.0007, "W": 0.0003, "Y": 0.9141 },
    "68": { "A": 0.0004, "C": 0.0001, "D": 0.0047, "E": 0.0002, "F": 0.0001, "G": 0.0004, "H": 0.0187, "I": 0.0035, "K": 0.0141, "L": 0.0003, "M": 0.0003, "N": 0.891, "P": 0.0014, "Q": 0.0018, "R": 0.0022, "S": 0.0308, "T": 0.0282, "V": 0.0002, "Y": 0.0015 },
    "69": { "A": 0.0067, "D": 0.0002, "E": 0.0001, "F": 0.0003, "G": 0.0002, "H": 0.0012, "K": 0.0002, "L": 0.0034, "M": 0.0003, "N": 0.0001, "P": 0.9607, "Q": 0.0027, "R": 0.002, "S": 0.0147, "T": 0.0059, "V": 0.0006, "W": 0.0003, "Y": 0.0003 },
    "70": { "A": 0.0219, "C": 0.0007, "D": 0.0003, "F": 0.0035, "G": 0.0001, "H": 0.0004, "I": 0.0001, "L": 0.0004, "N": 0.0001, "P": 0.0218, "S": 0.9388, "T": 0.0062, "V": 0.0004, "Y": 0.0052 },
    "71": { "A": 0.0001, "C": 0.0001, "F": 0.0119, "H": 0.0041, "I": 0.005, "L": 0.9465, "M": 0.0002, "P": 0.0139, "Q": 0.0001, "R": 0.0122, "S": 0.0008, "T": 0.0004, "V": 0.0045, "Y": 0.0002 },
    "72": { "A": 0.0041, "D": 0.0012, "E": 0.0283, "G": 0.0038, "H": 0.001, "I": 0.0003, "K": 0.8615, "L": 0.0007, "M": 0.0048, "N": 0.0149, "P": 0.0006, "Q": 0.0211, "R": 0.0427, "S": 0.0032, "T": 0.0102, "V": 0.0005, "W": 0.001, "Y": 0.0001 },
    "74": { "A": 0.0004, "C": 0.0024, "D": 0.0011, "E": 0.0001, "F": 0.0001, "G": 0.0326, "H": 0.0006, "I": 0.0055, "K": 0.0005, "L": 0.0007, "M": 0.0009, "N": 0.0084, "P": 0.0003, "R": 0.0272, "S": 0.9085, "T": 0.0093, "V": 0.0012, "W": 0.0001 },
    "75": { "A": 0.0001, "C": 0.0001, "E": 0.0001, "G": 0.0036, "H": 0.0003, "I": 0.0001, "L": 0.0063, "P": 0.0047, "Q": 0.0037, "R": 0.9804, "S": 0.0001, "T": 0.0002, "V": 0.0002, "W": 0.0001 },
    "76": { "A": 0.0163, "C": 0.0004, "D": 0.0012, "F": 0.0073, "G": 0.0083, "H": 0.0002, "I": 0.0192, "L": 0.0408, "M": 0.0005, "P": 0.0007, "R": 0.0003, "S": 0.0011, "T": 0.0009, "V": 0.9025, "Y": 0.0001 },
    "77": { "A": 0.0176, "C": 0.0001, "D": 0.0012, "F": 0.0014, "G": 0.0004, "H": 0.0006, "I": 0.0126, "K": 0.0006, "L": 0.001, "M": 0.0006, "N": 0.0059, "P": 0.0091, "Q": 0.0001, "R": 0.0011, "S": 0.0433, "T": 0.9017, "V": 0.0022, "Y": 0.0004 },
    "78": { "A": 0.0004, "C": 0.0001, "F": 0.0032, "G": 0.0002, "I": 0.8334, "K": 0.0018, "L": 0.0246, "M": 0.104, "N": 0.0004, "P": 0.0002, "Q": 0.0004, "R": 0.0064, "S": 0.0013, "T": 0.0059, "V": 0.0172 },
    "79": { "A": 0.0129, "C": 0.0002, "D": 0.0002, "E": 0.0002, "F": 0.0035, "G": 0.0001, "I": 0.0002, "L": 0.0089, "P": 0.0052, "Q": 0.0001, "R": 0.0001, "S": 0.9609, "T": 0.0051, "V": 0.0005, "W": 0.0003, "Y": 0.0016 },
    "80": { "A": 0.0189, "C": 0.0002, "D": 0.0002, "E": 0.006, "F": 0.002, "G": 0.0119, "H": 0.0001, "I": 0.0499, "K": 0.0018, "L": 0.076, "M": 0.0117, "N": 0.0001, "P": 0.001, "Q": 0.0014, "R": 0.0071, "S": 0.0016, "T": 0.0031, "V": 0.8068, "W": 0.0001, "Y": 0.0001 },
    "81": { "A": 0.0091, "C": 0.0002, "D": 0.9526, "E": 0.0162, "F": 0.0001, "G": 0.007, "H": 0.0027, "I": 0.0001, "K": 0.0004, "L": 0.0001, "N": 0.0044, "P": 0.0001, "Q": 0.0004, "R": 0.0003, "S": 0.0008, "T": 0.0008, "V": 0.0019, "Y": 0.003 },
    "82": { "A": 0.0137, "D": 0.0001, "E": 0.002, "G": 0.0011, "I": 0.0016, "K": 0.0496, "L": 0.0017, "M": 0.0155, "N": 0.0019, "P": 0.0147, "Q": 0.0011, "R": 0.0289, "S": 0.008, "T": 0.8582, "V": 0.0014, "W": 0.0004 },
    "83": { "A": 0.0215, "C": 0.0007, "D": 0.0003, "F": 0.0031, "G": 0.0001, "H": 0.0003, "I": 0.0002, "L": 0.0009, "N": 0.0002, "P": 0.0141, "R": 0.0001, "S": 0.9479, "T": 0.0056, "V": 0.0009, "W": 0.0002, "Y": 0.0039 },
    "84": { "A": 0.0011, "D": 0.0007, "E": 0.0116, "G": 0.0013, "H": 0.0005, "I": 0.0005, "K": 0.8963, "L": 0.0002, "M": 0.0029, "N": 0.0275, "P": 0.0004, "Q": 0.013, "R": 0.0279, "S": 0.0027, "T": 0.0126, "V": 0.0002, "W": 0.0002, "Y": 0.0002 },
    "85": { "A": 0.0011, "C": 0.0001, "D": 0.0091, "E": 0.0009, "F": 0.0003, "G": 0.0027, "H": 0.0054, "I": 0.0043, "K": 0.0233, "L": 0.0002, "M": 0.0006, "N": 0.8986, "P": 0.0002, "Q": 0.001, "R": 0.0053, "S": 0.0271, "T": 0.0178, "V": 0.0004, "W": 0.0001, "Y": 0.0016 },
    "86": { "A": 0.0006, "D": 0.0008, "E": 0.0117, "F": 0.0002, "G": 0.0003, "H": 0.0338, "I": 0.0001, "K": 0.0055, "L": 0.0061, "M": 0.0001, "N": 0.0009, "P": 0.0075, "Q": 0.9176, "R": 0.0113, "S": 0.0006, "T": 0.0008, "V": 0.0006, "W": 0.0004, "Y": 0.0011 },
    "87": { "A": 0.0004, "C": 0.0126, "F": 0.9089, "G": 0.0004, "I": 0.0129, "L": 0.0271, "M": 0.0008, "N": 0.0001, "P": 0.0004, "R": 0.0003, "S": 0.0051, "T": 0.0003, "V": 0.0278, "W": 0.0001, "Y": 0.0026 },
    "88": { "A": 0.0127, "C": 0.0014, "D": 0.0001, "F": 0.0131, "H": 0.0003, "I": 0.0001, "L": 0.001, "P": 0.0067, "R": 0.0001, "S": 0.9535, "T": 0.0041, "V": 0.0003, "Y": 0.0065 },
    "89": { "A": 0.0001, "C": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0002, "L": 0.9687, "M": 0.0043, "P": 0.0036, "Q": 0.0014, "R": 0.0174, "S": 0.0001, "V": 0.0032, "W": 0.0005 },
    "90": { "A": 0.0008, "D": 0.0034, "E": 0.0151, "F": 0.0001, "G": 0.0036, "H": 0.003, "I": 0.0047, "K": 0.7827, "L": 0.0006, "M": 0.0041, "N": 0.0521, "P": 0.0004, "Q": 0.0111, "R": 0.0884, "S": 0.0094, "T": 0.0186, "V": 0.0005, "W": 0.001, "Y": 0.0006 },
    "91": { "A": 0.0002, "C": 0.0001, "E": 0.0001, "F": 0.0009, "G": 0.0008, "H": 0.0001, "I": 0.0013, "K": 0.0001, "L": 0.9345, "M": 0.0153, "P": 0.0021, "Q": 0.0007, "R": 0.0092, "S": 0.0005, "T": 0.0002, "V": 0.0325, "W": 0.0013 },
    "92": { "A": 0.0056, "C": 0.0011, "D": 0.0053, "E": 0.0011, "F": 0.0015, "G": 0.0152, "H": 0.002, "I": 0.0082, "K": 0.0092, "L": 0.0006, "M": 0.0015, "N": 0.0641, "P": 0.0019, "Q": 0.0003, "R": 0.0427, "S": 0.7157, "T": 0.1178, "V": 0.0013, "W": 0.0003, "Y": 0.0047 },
    "93": { "A": 0.0168, "C": 0.0017, "D": 0.0004, "F": 0.0063, "G": 0.0002, "H": 0.0003, "L": 0.0011, "N": 0.0001, "P": 0.0071, "Q": 0.0001, "R": 0.0001, "S": 0.9589, "T": 0.0026, "V": 0.0006, "W": 0.0001, "Y": 0.0036 },
    "94": { "A": 0.0102, "C": 0.0002, "E": 0.0006, "F": 0.0001, "G": 0.0129, "I": 0.0005, "L": 0.016, "M": 0.0112, "P": 0.0002, "R": 0.0008, "S": 0.0001, "T": 0.0008, "V": 0.9457, "W": 0.0007 },
    "95": { "A": 0.0082, "D": 0.0005, "F": 0.0001, "G": 0.0005, "H": 0.0002, "I": 0.0041, "K": 0.0002, "L": 0.0001, "M": 0.0002, "N": 0.0024, "P": 0.0053, "R": 0.0008, "S": 0.0114, "T": 0.9656, "V": 0.0001, "Y": 0.0001 },
    "96": { "A": 0.9709, "D": 0.0038, "E": 0.0002, "G": 0.0026, "I": 0.0001, "N": 0.0001, "P": 0.0041, "R": 0.0002, "S": 0.0033, "T": 0.0072, "V": 0.0075 },
    "97": { "A": 0.9748, "D": 0.0004, "E": 0.0033, "G": 0.0014, "I": 0.0007, "K": 0.0001, "M": 0.0003, "P": 0.0017, "R": 0.0002, "S": 0.0054, "T": 0.0089, "V": 0.0027 },
    "98": { "A": 0.0021, "C": 0.0001, "D": 0.9837, "E": 0.0067, "G": 0.0028, "H": 0.0004, "N": 0.001, "Q": 0.0001, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0009, "Y": 0.0019 },
    "99": { "A": 0.009, "E": 0.0001, "G": 0.0002, "K": 0.0012, "M": 0.0031, "N": 0.0001, "P": 0.0035, "Q": 0.0001, "R": 0.0013, "S": 0.0082, "T": 0.973, "W": 0.0001 },
    "100": { "A": 0.981, "C": 0.0001, "D": 0.0017, "G": 0.0096, "H": 0.0001, "P": 0.0003, "S": 0.0043, "T": 0.0019, "V": 0.001 },
    "101": { "A": 0.0027, "D": 0.0002, "E": 0.0008, "F": 0.0045, "G": 0.0057, "H": 0.0002, "I": 0.0311, "K": 0.0006, "L": 0.0201, "M": 0.0192, "N": 0.0003, "Q": 0.0001, "R": 0.0011, "S": 0.0003, "T": 0.0064, "V": 0.9063, "W": 0.0003, "Y": 0.0001 },
    "102": { "C": 0.001, "D": 0.0047, "E": 0.0007, "F": 0.0049, "G": 0.0001, "H": 0.0018, "L": 0.0002, "N": 0.0004, "Q": 0.0001, "S": 0.0006, "Y": 0.9855 },
    "103": { "C": 0.002, "D": 0.0041, "F": 0.0543, "H": 0.0053, "L": 0.0005, "N": 0.0003, "Q": 0.0001, "R": 0.0001, "S": 0.0031, "T": 0.0002, "V": 0.0007, "W": 0.0002, "Y": 0.929 },
    "104": { "A": 0.0001, "C": 0.9805, "D": 0.0001, "F": 0.0014, "G": 0.0057, "L": 0.0003, "R": 0.0012, "S": 0.0007, "V": 0.0004, "W": 0.009, "Y": 0.0007 },
    "105": { "A": 0.9446, "C": 0.0003, "D": 0.0003, "E": 0.0013, "F": 0.0002, "G": 0.0065, "H": 0.0002, "I": 0.0003, "K": 0.0001, "L": 0.0007, "M": 0.0008, "N": 0.0001, "P": 0.0004, "Q": 0.0002, "R": 0.0009, "S": 0.0054, "T": 0.0105, "V": 0.0262, "W": 0.0006, "Y": 0.0002 },
    "106": { "A": 0.0073, "C": 0.0014, "D": 0.0004, "E": 0.0012, "F": 0.0007, "G": 0.0268, "H": 0.0009, "I": 0.0059, "K": 0.0152, "L": 0.0035, "M": 0.0017, "N": 0.0032, "P": 0.0012, "Q": 0.0013, "R": 0.836, "S": 0.0609, "T": 0.0274, "V": 0.0039, "W": 0.0003, "Y": 0.001 },
    "107": { "A": 0.0704, "C": 0.0044, "D": 0.1209, "E": 0.079, "F": 0.0138, "G": 0.229, "H": 0.0738, "I": 0.0176, "K": 0.0103, "L": 0.0585, "M": 0.0068, "N": 0.0121, "P": 0.0207, "Q": 0.0354, "R": 0.0629, "S": 0.0476, "T": 0.0284, "V": 0.0752, "W": 0.0122, "Y": 0.0208 },
    "108": { "A": 0.0532, "C": 0.0061, "D": 0.0518, "E": 0.0347, "F": 0.0233, "G": 0.1536, "H": 0.0231, "I": 0.0292, "K": 0.0251, "L": 0.0681, "M": 0.01, "N": 0.0274, "P": 0.0815, "Q": 0.0252, "R": 0.1285, "S": 0.0839, "T": 0.0456, "V": 0.0595, "W": 0.0193, "Y": 0.051 },
    "109": { "A": 0.0604, "C": 0.0135, "D": 0.0544, "E": 0.0267, "F": 0.0255, "G": 0.167, "H": 0.0168, "I": 0.0364, "K": 0.0175, "L": 0.0529, "M": 0.012, "N": 0.0274, "P": 0.053, "Q": 0.018, "R": 0.0903, "S": 0.0983, "T": 0.0525, "V": 0.0598, "W": 0.0263, "Y": 0.0916 },
    "110": { "A": 0.0658, "C": 0.0149, "D": 0.0668, "E": 0.0273, "F": 0.0263, "G": 0.1523, "H": 0.0145, "I": 0.0378, "K": 0.0101, "L": 0.0496, "M": 0.013, "N": 0.0261, "P": 0.0334, "Q": 0.0181, "R": 0.0592, "S": 0.1203, "T": 0.053, "V": 0.0691, "W": 0.0279, "Y": 0.1145 },
    "111": { "A": 0.0654, "C": 0.0213, "D": 0.0646, "E": 0.02, "F": 0.0302, "G": 0.1432, "H": 0.011, "I": 0.0378, "K": 0.0085, "L": 0.0451, "M": 0.013, "N": 0.0235, "P": 0.0286, "Q": 0.0135, "R": 0.0501, "S": 0.1417, "T": 0.0551, "V": 0.0687, "W": 0.0353, "Y": 0.1232 },
    "111A": { "A": 0.0671, "C": 0.019, "D": 0.0701, "E": 0.0216, "F": 0.0383, "G": 0.1487, "H": 0.0117, "I": 0.03, "K": 0.0079, "L": 0.0424, "M": 0.0126, "N": 0.0212, "P": 0.026, "Q": 0.0118, "R": 0.043, "S": 0.1569, "T": 0.0553, "V": 0.0686, "W": 0.0395, "Y": 0.1083 },
    "111B": { "A": 0.0532, "C": 0.0159, "D": 0.0651, "E": 0.0261, "F": 0.0435, "G": 0.1436, "H": 0.0071, "I": 0.0332, "K": 0.0057, "L": 0.0414, "M": 0.0108, "N": 0.021, "P": 0.0337, "Q": 0.0097, "R": 0.0404, "S": 0.1724, "T": 0.0588, "V": 0.0757, "W": 0.0494, "Y": 0.0933 },
    "111C": { "A": 0.0508, "C": 0.0332, "D": 0.0512, "E": 0.014, "F": 0.0357, "G": 0.1679, "H": 0.0083, "I": 0.0287, "K": 0.0103, "L": 0.0441, "M": 0.0066, "N": 0.0188, "P": 0.0262, "Q": 0.0078, "R": 0.0406, "S": 0.1836, "T": 0.0629, "V": 0.0704, "W": 0.0489, "Y": 0.09 },
    "111D": { "A": 0.0472, "C": 0.0418, "D": 0.0439, "E": 0.0184, "F": 0.0251, "G": 0.165, "H": 0.0096, "I": 0.0272, "K": 0.0075, "L": 0.043, "M": 0.01, "N": 0.0117, "P": 0.0439, "Q": 0.0067, "R": 0.0301, "S": 0.1876, "T": 0.0627, "V": 0.0681, "W": 0.0468, "Y": 0.104 },
    "111E": { "A": 0.0508, "C": 0.0579, "D": 0.0366, "E": 0.0071, "F": 0.0335, "G": 0.1321, "H": 0.0071, "I": 0.0193, "K": 0.0081, "L": 0.0437, "M": 0.0081, "N": 0.0315, "P": 0.0366, "Q": 0.0081, "R": 0.0335, "S": 0.1931, "T": 0.06, "V": 0.0701, "W": 0.0366, "Y": 0.126 },
    "111F": { "A": 0.0536, "C": 0.0595, "D": 0.0565, "E": 0.0149, "F": 0.0327, "G": 0.122, "H": 0.0149, "I": 0.0298, "K": 0.0089, "L": 0.0268, "M": 0.006, "N": 0.0238, "P": 0.0476, "Q": 0.0149, "R": 0.0536, "S": 0.1548, "T": 0.0565, "V": 0.0417, "W": 0.0327, "Y": 0.1488 },
    "111G": { "A": 0.0522, "C": 0.0609, "D": 0.0261, "E": 0.0261, "F": 0.0435, "G": 0.1478, "H": 0.0087, "I": 0.0522, "K": 0.0087, "L": 0.0435, "M": 0.0174, "N": 0.0261, "P": 0.0696, "Q": 0.0087, "R": 0.0696, "S": 0.1043, "T": 0.0435, "V": 0.0435, "W": 0.0261, "Y": 0.1217 },
    "112G": { "A": 0.0417, "C": 0.0312, "D": 0.0417, "E": 0.0104, "F": 0.0417, "G": 0.1458, "H": 0.0104, "I": 0.0208, "K": 0.0104, "L": 0.0365, "N": 0.0208, "P": 0.0885, "Q": 0.0208, "R": 0.0625, "S": 0.151, "T": 0.0521, "V": 0.0469, "W": 0.0312, "Y": 0.1354 },
    "112F": { "A": 0.0504, "C": 0.0521, "D": 0.0303, "E": 0.0134, "F": 0.0202, "G": 0.1412, "H": 0.0118, "I": 0.0185, "K": 0.0235, "L": 0.0521, "M": 0.0067, "N": 0.0202, "P": 0.0487, "Q": 0.0185, "R": 0.0571, "S": 0.1563, "T": 0.0588, "V": 0.0555, "W": 0.0218, "Y": 0.1429 },
    "112E": { "A": 0.0514, "C": 0.0354, "D": 0.0399, "E": 0.0206, "F": 0.0244, "G": 0.1466, "H": 0.0174, "I": 0.0264, "K": 0.0161, "L": 0.0508, "M": 0.0122, "N": 0.018, "P": 0.0688, "Q": 0.0154, "R": 0.0669, "S": 0.1395, "T": 0.0476, "V": 0.045, "W": 0.027, "Y": 0.1305 },
    "112D": { "A": 0.0509, "C": 0.0287, "D": 0.0398, "E": 0.0223, "F": 0.027, "G": 0.1652, "H": 0.0186, "I": 0.0284, "K": 0.0156, "L": 0.0381, "M": 0.0083, "N": 0.022, "P": 0.0515, "Q": 0.015, "R": 0.0604, "S": 0.1369, "T": 0.0559, "V": 0.0576, "W": 0.0339, "Y": 0.1241 },
    "112C": { "A": 0.0592, "C": 0.0273, "D": 0.0403, "E": 0.0216, "F": 0.0238, "G": 0.1453, "H": 0.0127, "I": 0.0243, "K": 0.0125, "L": 0.0441, "M": 0.0057, "N": 0.0279, "P": 0.0522, "Q": 0.0119, "R": 0.0556, "S": 0.1442, "T": 0.0599, "V": 0.0541, "W": 0.0317, "Y": 0.1458 },
    "112B": { "A": 0.0516, "C": 0.0294, "D": 0.0454, "E": 0.0218, "F": 0.0233, "G": 0.1439, "H": 0.018, "I": 0.025, "K": 0.0142, "L": 0.0411, "M": 0.0068, "N": 0.0219, "P": 0.0487, "Q": 0.0142, "R": 0.0531, "S": 0.1142, "T": 0.0546, "V": 0.044, "W": 0.0284, "Y": 0.2006 },
    "112A": { "A": 0.0542, "C": 0.0145, "D": 0.0402, "E": 0.0202, "F": 0.0231, "G": 0.1439, "H": 0.0191, "I": 0.0206, "K": 0.0146, "L": 0.0446, "M": 0.0087, "N": 0.0237, "P": 0.0556, "Q": 0.0153, "R": 0.0656, "S": 0.1097, "T": 0.055, "V": 0.0408, "W": 0.0308, "Y": 0.2 },
    "112": { "A": 0.0511, "C": 0.0118, "D": 0.0409, "E": 0.0273, "F": 0.0267, "G": 0.1311, "H": 0.0254, "I": 0.0169, "K": 0.0183, "L": 0.0558, "M": 0.0088, "N": 0.0303, "P": 0.0569, "Q": 0.0153, "R": 0.0593, "S": 0.101, "T": 0.0533, "V": 0.0388, "W": 0.0309, "Y": 0.2001 },
    "113": { "A": 0.0414, "C": 0.009, "D": 0.0824, "E": 0.0213, "F": 0.0223, "G": 0.1582, "H": 0.0238, "I": 0.0104, "K": 0.0122, "L": 0.0291, "M": 0.005, "N": 0.0852, "P": 0.0289, "Q": 0.0095, "R": 0.0526, "S": 0.0713, "T": 0.0299, "V": 0.0242, "W": 0.0508, "Y": 0.2325 },
    "114": { "A": 0.1391, "C": 0.0079, "D": 0.021, "E": 0.0134, "F": 0.0211, "G": 0.1723, "H": 0.0191, "I": 0.0107, "K": 0.007, "L": 0.0253, "M": 0.0056, "N": 0.0156, "P": 0.05, "Q": 0.0072, "R": 0.026, "S": 0.0455, "T": 0.0285, "V": 0.0273, "W": 0.1335, "Y": 0.224 },
    "115": { "A": 0.01, "C": 0.0034, "D": 0.0047, "E": 0.0046, "F": 0.5806, "G": 0.0198, "H": 0.0052, "I": 0.0376, "K": 0.0013, "L": 0.0763, "M": 0.1328, "N": 0.0048, "P": 0.0177, "Q": 0.0027, "R": 0.0056, "S": 0.0169, "T": 0.0099, "V": 0.0309, "W": 0.0085, "Y": 0.0268 },
    "116": { "A": 0.0125, "C": 0.0005, "D": 0.8465, "E": 0.0155, "F": 0.0032, "G": 0.0302, "H": 0.0093, "I": 0.0031, "K": 0.002, "L": 0.0056, "M": 0.0006, "N": 0.0126, "P": 0.0054, "Q": 0.0141, "R": 0.0054, "S": 0.0116, "T": 0.0057, "V": 0.008, "W": 0.0013, "Y": 0.007 },
    "117": { "A": 0.0055, "C": 0.0027, "D": 0.0125, "E": 0.0019, "F": 0.0244, "G": 0.0031, "H": 0.0254, "I": 0.1087, "K": 0.0016, "L": 0.0498, "M": 0.0023, "N": 0.0108, "P": 0.152, "Q": 0.0027, "R": 0.0035, "S": 0.0424, "T": 0.0073, "V": 0.1617, "W": 0.0007, "Y": 0.3807 },
    "118": { "C": 0.0016, "F": 0.0003, "G": 0.0027, "L": 0.0024, "M": 0.0002, "P": 0.0001, "Q": 0.0001, "R": 0.0014, "S": 0.0006, "T": 0.0001, "V": 0.0014, "W": 0.9889, "Y": 0.0001 },
    "119": { "A": 0.0014, "C": 0.0018, "D": 0.0021, "G": 0.9877, "N": 0.0002, "R": 0.0003, "S": 0.0053, "V": 0.001 },
    "120": { "A": 0.0001, "E": 0.0014, "G": 0.0001, "H": 0.0059, "K": 0.0371, "L": 0.0046, "N": 0.0008, "P": 0.0154, "Q": 0.8753, "R": 0.0584, "S": 0.0002, "T": 0.0005 },
    "121": { "A": 0.0003, "D": 0.0002, "E": 0.0009, "G": 0.9961, "L": 0.0001, "R": 0.0011, "S": 0.0001, "V": 0.001, "W": 0.0002 },
    "122": { "A": 0.0206, "D": 0.0002, "F": 0.0006, "G": 0.0001, "H": 0.0003, "I": 0.0359, "K": 0.0021, "L": 0.0017, "M": 0.0021, "N": 0.0028, "P": 0.0011, "Q": 0.0001, "R": 0.0012, "S": 0.0115, "T": 0.9163, "V": 0.0033 },
    "123": { "A": 0.0028, "E": 0.0003, "F": 0.0004, "G": 0.0001, "H": 0.0004, "I": 0.0008, "K": 0.0019, "L": 0.699, "M": 0.1123, "N": 0.0001, "P": 0.0093, "Q": 0.0075, "R": 0.0068, "S": 0.0023, "T": 0.1482, "V": 0.0071, "W": 0.0008 },
    "124": { "A": 0.0033, "D": 0.0003, "F": 0.0008, "G": 0.0024, "I": 0.0049, "L": 0.0017, "S": 0.0003, "V": 0.9863 },
    "125": { "A": 0.0104, "F": 0.0006, "G": 0.0001, "H": 0.0006, "I": 0.0209, "K": 0.0001, "L": 0.0005, "N": 0.0012, "P": 0.0014, "R": 0.0002, "S": 0.0134, "T": 0.9489, "V": 0.0015, "Y": 0.0001 },
    "126": { "A": 0.0031, "C": 0.0001, "D": 0.0008, "F": 0.0006, "G": 0.002, "I": 0.0034, "L": 0.0002, "R": 0.0006, "S": 0.0002, "V": 0.989 },
    "127": { "A": 0.0051, "C": 0.0003, "F": 0.0019, "L": 0.0012, "N": 0.0001, "P": 0.0036, "S": 0.9839, "T": 0.0021, "V": 0.0002, "Y": 0.0016 },
    "128": { "A": 0.0061, "F": 0.0001, "L": 0.0012, "P": 0.0069, "Q": 0.0003, "S": 0.9837, "T": 0.0019 }
  },
  "IGHV5": {
    "1": { "D": 0.0151, "E": 0.9826, "H": 0.0007, "Q": 0.0008, "V": 0.0008 },
    "2": { "A": 0.0053, "D": 0.0011, "E": 0.0056, "F": 0.0003, "G": 0.0032, "I": 0.0004, "L": 0.0066, "M": 0.0031, "Q": 0.0008, "V": 0.9734, "W": 0.0001 },
    "3": { "E": 0.0014, "H": 0.0108, "K": 0.0023, "L": 0.0052, "P": 0.0085, "Q": 0.962, "R": 0.0076, "T": 0.0001, "V": 0.0016, "Y": 0.0002 },
    "4": { "A": 0.0001, "F": 0.0004, "L": 0.9852, "M": 0.0021, "P": 0.0038, "Q": 0.0005, "R": 0.0025, "S": 0.0001, "T": 0.0001, "V": 0.0051 },
    "5": { "A": 0.0092, "D": 0.0001, "E": 0.0103, "F": 0.0004, "G": 0.0029, "I": 0.0006, "L": 0.0103, "M": 0.002, "P": 0.0002, "Q": 0.0011, "R": 0.0001, "S": 0.0003, "T": 0.0002, "V": 0.9616, "W": 0.0006 },
    "6": { "A": 0.0003, "E": 0.0188, "H": 0.0027, "K": 0.0016, "L": 0.0015, "P": 0.0129, "Q": 0.9594, "R": 0.0017, "S": 0.0008, "T": 0.0004 },
    "7": { "A": 0.0038, "C": 0.0006, "F": 0.0013, "G": 0.0007, "L": 0.0003, "P": 0.0089, "R": 0.0001, "S": 0.9795, "T": 0.0026, "V": 0.0001, "Y": 0.0021 },
    "8": { "A": 0.0067, "C": 0.0002, "D": 0.0002, "E": 0.0032, "G": 0.9816, "K": 0.0001, "Q": 0.0001, "R": 0.0035, "S": 0.0003, "T": 0.0001, "V": 0.0037, "W": 0.0002 },
    "9": { "A": 0.9534, "D": 0.0003, "E": 0.0035, "G": 0.0052, "I": 0.0003, "K": 0.0001, "L": 0.0002, "P": 0.0144, "Q": 0.0001, "R": 0.0004, "S": 0.004, "T": 0.0118, "V": 0.0062 },
    "11": { "A": 0.0309, "D": 0.0045, "E": 0.9478, "G": 0.0054, "K": 0.0007, "P": 0.0001, "Q": 0.0037, "R": 0.0004, "S": 0.0001, "T": 0.0004, "V": 0.0056 },
    "12": { "A": 0.0093, "D": 0.0014, "E": 0.0025, "F": 0.0003, "G": 0.009, "I": 0.002, "K": 0.0005, "L": 0.0115, "M": 0.0087, "R": 0.0004, "S": 0.0002, "T": 0.0003, "V": 0.9532, "W": 0.0006 },
    "13": { "A": 0.0001, "D": 0.0002, "E": 0.0025, "G": 0.0007, "H": 0.0012, "I": 0.0028, "K": 0.9189, "M": 0.0001, "N": 0.0209, "P": 0.0003, "Q": 0.0137, "R": 0.0166, "S": 0.0005, "T": 0.021, "V": 0.0003, "Y": 0.0001 },
    "14": { "A": 0.0007, "D": 0.0006, "E": 0.0123, "G": 0.0009, "H": 0.0002, "I": 0.0001, "K": 0.9213, "L": 0.0001, "M": 0.0021, "N": 0.0044, "P": 0.0005, "Q": 0.026, "R": 0.0098, "S": 0.0008, "T": 0.0199, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "15": { "A": 0.0062, "C": 0.0001, "F": 0.0001, "G": 0.0008, "H": 0.0109, "I": 0.0001, "L": 0.0037, "M": 0.0001, "P": 0.9476, "Q": 0.0018, "R": 0.0047, "S": 0.0155, "T": 0.0078, "V": 0.0002, "W": 0.0001 },
    "16": { "A": 0.0018, "C": 0.0001, "E": 0.0033, "G": 0.9862, "R": 0.0027, "S": 0.0001, "T": 0.0001, "V": 0.0037, "W": 0.0019 },
    "17": { "A": 0.0328, "C": 0.0001, "D": 0.0134, "E": 0.9331, "G": 0.0074, "H": 0.0004, "K": 0.0018, "L": 0.0001, "M": 0.0001, "N": 0.0004, "Q": 0.0055, "R": 0.001, "S": 0.0005, "T": 0.0001, "V": 0.003, "W": 0.0002, "Y": 0.0001 },
    "18": { "A": 0.0107, "C": 0.0011, "D": 0.0002, "F": 0.0021, "G": 0.0001, "H": 0.0003, "L": 0.0003, "P": 0.0112, "R": 0.0003, "S": 0.9662, "T": 0.0035, "Y": 0.0037 },
    "19": { "A": 0.0016, "C": 0.0001, "E": 0.0003, "F": 0.0002, "H": 0.0004, "I": 0.0002, "L": 0.97, "M": 0.0036, "P": 0.0048, "Q": 0.0011, "R": 0.0136, "S": 0.0001, "T": 0.0002, "V": 0.0035, "W": 0.0001 },
    "20": { "A": 0.0001, "D": 0.0005, "E": 0.0029, "G": 0.0009, "H": 0.0003, "I": 0.0007, "K": 0.7502, "L": 0.0001, "M": 0.0042, "N": 0.0055, "P": 0.0005, "Q": 0.0161, "R": 0.1785, "S": 0.0019, "T": 0.0367, "V": 0.0001, "W": 0.0006 },
    "21": { "F": 0.0028, "G": 0.0001, "H": 0.0006, "I": 0.9506, "K": 0.0001, "L": 0.0216, "M": 0.0019, "N": 0.0028, "P": 0.0004, "R": 0.0005, "S": 0.0072, "T": 0.0079, "V": 0.0032, "Y": 0.0001 },
    "22": { "A": 0.0075, "C": 0.0021, "D": 0.0002, "F": 0.0024, "H": 0.0021, "L": 0.0007, "P": 0.0163, "R": 0.0001, "S": 0.9587, "T": 0.0028, "V": 0.0002, "Y": 0.0069 },
    "23": { "A": 0.0001, "C": 0.9658, "F": 0.0022, "G": 0.013, "L": 0.0006, "R": 0.0063, "S": 0.0028, "V": 0.0009, "W": 0.0079, "Y": 0.0005 },
    "24": { "A": 0.0011, "D": 0.0001, "E": 0.0186, "G": 0.0003, "H": 0.0009, "I": 0.0001, "K": 0.8413, "L": 0.001, "M": 0.005, "N": 0.0041, "P": 0.0008, "Q": 0.0616, "R": 0.016, "S": 0.0029, "T": 0.0445, "V": 0.0004, "W": 0.0004, "Y": 0.0006 },
    "25": { "A": 0.065, "C": 0.0048, "D": 0.0043, "E": 0.0001, "F": 0.0003, "G": 0.8748, "H": 0.0003, "I": 0.0019, "L": 0.0007, "N": 0.0005, "P": 0.0011, "R": 0.0022, "S": 0.0068, "T": 0.0143, "V": 0.0222, "W": 0.0001, "Y": 0.0006 },
    "26": { "A": 0.0074, "C": 0.0021, "D": 0.0004, "F": 0.0112, "G": 0.0001, "H": 0.0001, "I": 0.0007, "L": 0.0025, "N": 0.0001, "P": 0.0073, "Q": 0.0001, "R": 0.0002, "S": 0.9563, "T": 0.0032, "V": 0.0005, "W": 0.0001, "Y": 0.0077 },
    "27": { "A": 0.0056, "C": 0.0003, "D": 0.0018, "E": 0.0067, "G": 0.973, "K": 0.0003, "L": 0.0002, "P": 0.0007, "Q": 0.0002, "R": 0.0042, "S": 0.0002, "T": 0.0007, "V": 0.0058, "W": 0.0001 },
    "28": { "A": 0.0005, "C": 0.0032, "D": 0.0122, "E": 0.0006, "F": 0.0133, "G": 0.0003, "H": 0.0063, "I": 0.0004, "K": 0.0003, "L": 0.0011, "N": 0.0067, "P": 0.0002, "Q": 0.0003, "S": 0.0262, "T": 0.001, "V": 0.0004, "Y": 0.9271 },
    "29": { "A": 0.0018, "C": 0.0014, "D": 0.0099, "E": 0.0004, "F": 0.0009, "G": 0.0063, "H": 0.0016, "I": 0.0203, "K": 0.0072, "L": 0.0018, "M": 0.002, "N": 0.0322, "P": 0.0035, "Q": 0.0004, "R": 0.0504, "S": 0.7985, "T": 0.059, "V": 0.0011, "W": 0.0001, "Y": 0.0012 },
    "30": { "A": 0.0002, "C": 0.0083, "D": 0.0001, "F": 0.9562, "G": 0.0002, "H": 0.0001, "I": 0.0025, "L": 0.0153, "M": 0.0001, "P": 0.0001, "R": 0.0002, "S": 0.006, "T": 0.0001, "V": 0.0082, "W": 0.0002, "Y": 0.0022 },
    "35": { "A": 0.0256, "C": 0.0003, "D": 0.0046, "E": 0.0009, "F": 0.0009, "G": 0.0053, "H": 0.0017, "I": 0.0108, "K": 0.0017, "L": 0.0011, "M": 0.0009, "N": 0.0213, "P": 0.0325, "Q": 0.0002, "R": 0.0032, "S": 0.0578, "T": 0.8282, "V": 0.0024, "Y": 0.0006 },
    "36": { "A": 0.0042, "C": 0.0021, "D": 0.0187, "E": 0.0008, "F": 0.0001, "G": 0.0114, "H": 0.0039, "I": 0.0087, "K": 0.0098, "L": 0.0013, "M": 0.0006, "N": 0.0951, "P": 0.0019, "Q": 0.0007, "R": 0.0252, "S": 0.7379, "T": 0.0736, "V": 0.0007, "W": 0.0001, "Y": 0.0032 },
    "37": { "A": 0.0006, "C": 0.0013, "D": 0.0072, "E": 0.0004, "F": 0.0194, "H": 0.0278, "I": 0.0001, "K": 0.0011, "L": 0.0008, "N": 0.0134, "P": 0.0009, "Q": 0.0022, "R": 0.0007, "S": 0.0344, "T": 0.0014, "V": 0.0003, "W": 0.0002, "Y": 0.8876 },
    "38": { "A": 0.0004, "C": 0.011, "D": 0.0003, "E": 0.0001, "F": 0.0011, "G": 0.0106, "H": 0.0002, "K": 0.0004, "L": 0.004, "N": 0.0002, "P": 0.0001, "R": 0.0062, "S": 0.0047, "V": 0.0001, "W": 0.9545, "Y": 0.006 },
    "39": { "A": 0.0001, "C": 0.0004, "D": 0.0001, "F": 0.0048, "G": 0.0002, "H": 0.0014, "I": 0.9121, "K": 0.0001, "L": 0.0306, "M": 0.0021, "N": 0.0084, "P": 0.0007, "R": 0.0011, "S": 0.0096, "T": 0.0138, "V": 0.0135, "W": 0.0001, "Y": 0.0006 },
    "40": { "A": 0.0754, "C": 0.0027, "D": 0.004, "E": 0.0001, "F": 0.0003, "G": 0.7408, "H": 0.0022, "I": 0.0017, "L": 0.0007, "M": 0.0001, "N": 0.0085, "P": 0.0016, "R": 0.0043, "S": 0.1382, "T": 0.0107, "V": 0.0081, "Y": 0.0005 },
    "41": { "A": 0.0002, "C": 0.0045, "G": 0.0111, "L": 0.0041, "P": 0.0001, "R": 0.0104, "S": 0.0016, "T": 0.0001, "V": 0.0008, "W": 0.967 },
    "42": { "A": 0.0053, "C": 0.0005, "D": 0.0001, "E": 0.0008, "F": 0.0004, "G": 0.0168, "I": 0.0005, "L": 0.0106, "M": 0.0021, "R": 0.0004, "S": 0.0001, "V": 0.9614, "W": 0.0007 },
    "43": { "A": 0.0007, "C": 0.0031, "G": 0.0068, "H": 0.0015, "L": 0.0034, "P": 0.004, "R": 0.975, "S": 0.0049, "T": 0.0001, "V": 0.0003 },
    "44": { "A": 0.0002, "E": 0.0012, "H": 0.007, "K": 0.0058, "L": 0.0028, "P": 0.0134, "Q": 0.9617, "R": 0.0066, "S": 0.0002, "T": 0.0009, "W": 0.0001 },
    "45": { "A": 0.001, "C": 0.0004, "D": 0.0002, "E": 0.0007, "F": 0.0001, "G": 0.0007, "I": 0.0103, "K": 0.0115, "L": 0.0276, "M": 0.8879, "N": 0.0001, "P": 0.0006, "Q": 0.0003, "R": 0.0169, "S": 0.0008, "T": 0.0142, "V": 0.0262, "W": 0.0006 },
    "46": { "A": 0.0049, "F": 0.0001, "G": 0.0001, "H": 0.0085, "L": 0.0017, "P": 0.9708, "Q": 0.0001, "R": 0.0012, "S": 0.0074, "T": 0.0049 },
    "47": { "A": 0.0006, "C": 0.0001, "D": 0.0014, "E": 0.004, "G": 0.9866, "P": 0.0001, "R": 0.0027, "S": 0.0001, "V": 0.0024, "W": 0.002 },
    "48": { "A": 0.0002, "D": 0.0001, "E": 0.0093, "G": 0.0023, "H": 0.0005, "I": 0.0013, "K": 0.9396, "M": 0.0001, "N": 0.0127, "P": 0.0001, "Q": 0.0134, "R": 0.0131, "S": 0.0007, "T": 0.0064, "Y": 0.0002 },
    "49": { "A": 0.0023, "C": 0.0032, "D": 0.0029, "E": 0.0001, "G": 0.9841, "H": 0.0001, "K": 0.0001, "R": 0.0018, "S": 0.0016, "V": 0.0035, "W": 0.0001 },
    "50": { "A": 0.0001, "F": 0.0002, "I": 0.0001, "K": 0.0001, "L": 0.9711, "M": 0.0048, "P": 0.0107, "Q": 0.0008, "R": 0.0103, "S": 0.0001, "T": 0.0001, "V": 0.001, "W": 0.0004 },
    "51": { "A": 0.0131, "D": 0.0077, "E": 0.9549, "G": 0.0152, "K": 0.0017, "Q": 0.0043, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0022, "W": 0.0002 },
    "52": { "C": 0.0086, "E": 0.0002, "F": 0.0011, "G": 0.0147, "L": 0.0076, "M": 0.0001, "P": 0.0001, "R": 0.0021, "S": 0.0015, "V": 0.0017, "W": 0.9575, "Y": 0.0046 },
    "53": { "E": 0.0002, "G": 0.0013, "I": 0.0099, "K": 0.0018, "L": 0.0238, "M": 0.9167, "N": 0.0001, "Q": 0.0003, "R": 0.0134, "S": 0.0008, "T": 0.0027, "V": 0.0284, "W": 0.0004 },
    "54": { "A": 0.0152, "C": 0.0004, "D": 0.0001, "E": 0.0012, "G": 0.9703, "M": 0.0001, "P": 0.0001, "R": 0.0022, "T": 0.0005, "V": 0.0053, "W": 0.0045 },
    "55": { "A": 0.0024, "C": 0.001, "D": 0.0005, "E": 0.0005, "F": 0.0113, "G": 0.0074, "H": 0.0008, "I": 0.7194, "K": 0.006, "L": 0.0174, "M": 0.0135, "N": 0.0053, "P": 0.0005, "R": 0.1485, "S": 0.0151, "T": 0.0104, "V": 0.0371, "W": 0.0018, "Y": 0.0012 },
    "56": { "A": 0.0001, "C": 0.0001, "F": 0.0058, "G": 0.0007, "H": 0.0005, "I": 0.9268, "K": 0.0001, "L": 0.0138, "M": 0.0059, "N": 0.0027, "P": 0.0001, "Q": 0.0013, "R": 0.0008, "S": 0.0128, "T": 0.0089, "V": 0.0193, "Y": 0.0002 },
    "57": { "A": 0.0033, "C": 0.0038, "D": 0.1592, "E": 0.0031, "F": 0.023, "G": 0.0025, "H": 0.0152, "I": 0.0005, "K": 0.0007, "L": 0.0026, "N": 0.0105, "P": 0.0005, "Q": 0.001, "R": 0.001, "S": 0.0158, "T": 0.0004, "V": 0.0018, "W": 0.003, "Y": 0.7522 },
    "58": { "A": 0.005, "C": 0.0006, "D": 0.0001, "F": 0.0012, "G": 0.002, "H": 0.0025, "I": 0.0002, "L": 0.0051, "P": 0.9713, "Q": 0.0001, "R": 0.0014, "S": 0.0029, "T": 0.0056, "V": 0.0017, "Y": 0.0003 },
    "59": { "A": 0.0157, "C": 0.0071, "D": 0.0268, "E": 0.0028, "F": 0.0012, "G": 0.7426, "H": 0.0027, "I": 0.0027, "K": 0.0009, "L": 0.0015, "M": 0.0003, "N": 0.0024, "P": 0.0018, "Q": 0.0005, "R": 0.0172, "S": 0.1544, "T": 0.0064, "V": 0.0101, "W": 0.0009, "Y": 0.0021 },
    "62": { "A": 0.0105, "C": 0.0003, "D": 0.9381, "E": 0.0099, "G": 0.0071, "H": 0.003, "I": 0.0007, "K": 0.0007, "L": 0.0009, "N": 0.0107, "Q": 0.0004, "R": 0.0001, "S": 0.0016, "T": 0.0025, "V": 0.0019, "W": 0.0001, "Y": 0.0115 },
    "63": { "A": 0.0161, "C": 0.0013, "D": 0.0008, "E": 0.0001, "F": 0.0055, "G": 0.0008, "H": 0.0005, "L": 0.001, "N": 0.0001, "P": 0.0086, "R": 0.0003, "S": 0.9547, "T": 0.0051, "V": 0.0004, "W": 0.0002, "Y": 0.0045 },
    "64": { "A": 0.0102, "C": 0.0007, "D": 0.7526, "E": 0.0381, "F": 0.0034, "G": 0.0051, "H": 0.0064, "I": 0.0003, "K": 0.004, "L": 0.0009, "M": 0.0001, "N": 0.0113, "Q": 0.0024, "R": 0.0025, "S": 0.0048, "T": 0.0032, "V": 0.0027, "Y": 0.1512 },
    "65": { "A": 0.0162, "C": 0.0001, "D": 0.0006, "F": 0.0015, "G": 0.001, "H": 0.0006, "I": 0.0215, "K": 0.0003, "L": 0.0009, "M": 0.0007, "N": 0.0051, "P": 0.009, "R": 0.0025, "S": 0.0122, "T": 0.9183, "V": 0.009, "W": 0.0001, "Y": 0.0004 },
    "66": { "A": 0.0018, "C": 0.0001, "D": 0.0035, "E": 0.0024, "F": 0.0002, "G": 0.0058, "H": 0.0023, "I": 0.0123, "K": 0.0348, "L": 0.0013, "M": 0.0013, "N": 0.1401, "P": 0.0007, "Q": 0.0017, "R": 0.7526, "S": 0.0146, "T": 0.0207, "V": 0.0018, "W": 0.0001, "Y": 0.0018 },
    "67": { "A": 0.0003, "C": 0.0021, "D": 0.0113, "E": 0.0001, "F": 0.0034, "G": 0.0001, "H": 0.0033, "I": 0.0014, "K": 0.0003, "L": 0.0014, "N": 0.0036, "P": 0.0003, "Q": 0.0002, "R": 0.0002, "S": 0.008, "T": 0.0003, "V": 0.0007, "W": 0.0001, "Y": 0.963 },
    "68": { "A": 0.0005, "C": 0.0019, "D": 0.0005, "F": 0.0001, "G": 0.0138, "H": 0.0006, "I": 0.0037, "K": 0.0005, "L": 0.0004, "M": 0.0003, "N": 0.0234, "P": 0.0003, "R": 0.0206, "S": 0.9279, "T": 0.0052, "V": 0.0001, "W": 0.0002 },
    "69": { "A": 0.0024, "D": 0.0013, "E": 0.0003, "H": 0.0001, "L": 0.0028, "P": 0.9799, "Q": 0.0041, "R": 0.0021, "S": 0.0025, "T": 0.0043 },
    "70": { "A": 0.0253, "C": 0.0009, "D": 0.0002, "F": 0.0013, "G": 0.0001, "H": 0.0001, "K": 0.0011, "L": 0.0003, "P": 0.0095, "R": 0.0002, "S": 0.953, "T": 0.0041, "V": 0.0002, "W": 0.0001, "Y": 0.0035 },
    "71": { "A": 0.0002, "C": 0.0122, "F": 0.9415, "G": 0.0004, "H": 0.0001, "I": 0.0028, "L": 0.0156, "M": 0.0001, "P": 0.0001, "R": 0.0002, "S": 0.0065, "V": 0.0176, "W": 0.0001, "Y": 0.0026 },
    "72": { "A": 0.0005, "D": 0.0005, "E": 0.0282, "G": 0.0003, "H": 0.0089, "K": 0.0081, "L": 0.0011, "N": 0.0001, "P": 0.0046, "Q": 0.9278, "R": 0.0186, "T": 0.0004, "Y": 0.0008 },
    "74": { "A": 0.0021, "C": 0.0041, "D": 0.0053, "E": 0.0001, "G": 0.9805, "R": 0.002, "S": 0.0018, "V": 0.0039, "W": 0.0001 },
    "75": { "D": 0.0002, "E": 0.0011, "G": 0.0003, "H": 0.174, "K": 0.008, "L": 0.0044, "M": 0.0001, "N": 0.0034, "P": 0.0041, "Q": 0.786, "R": 0.0167, "T": 0.0008, "Y": 0.0006 },
    "76": { "A": 0.0047, "C": 0.0003, "D": 0.001, "F": 0.0055, "G": 0.0083, "I": 0.0064, "L": 0.0035, "R": 0.0001, "S": 0.0001, "T": 0.0003, "V": 0.9698 },
    "77": { "A": 0.0057, "D": 0.0001, "E": 0.0001, "F": 0.0002, "G": 0.0006, "H": 0.0004, "I": 0.01, "L": 0.0006, "M": 0.0001, "N": 0.0023, "P": 0.0075, "Q": 0.0001, "R": 0.0009, "S": 0.0197, "T": 0.9513, "V": 0.0004 },
    "78": { "A": 0.0003, "C": 0.0003, "F": 0.0176, "H": 0.0002, "I": 0.9141, "L": 0.0144, "M": 0.0241, "N": 0.0022, "P": 0.0001, "R": 0.0006, "S": 0.0133, "T": 0.0038, "V": 0.0089 },
    "79": { "A": 0.013, "C": 0.0001, "E": 0.0001, "F": 0.0002, "L": 0.0017, "M": 0.0013, "P": 0.0057, "Q": 0.0001, "S": 0.9729, "T": 0.0042, "V": 0.0001, "Y": 0.0003 },
    "80": { "A": 0.8982, "D": 0.0027, "E": 0.0001, "F": 0.0007, "G": 0.0035, "H": 0.0001, "I": 0.0013, "K": 0.0013, "L": 0.001, "N": 0.0001, "P": 0.0009, "R": 0.0001, "S": 0.0037, "T": 0.0073, "V": 0.079 },
    "81": { "A": 0.0049, "C": 0.0004, "D": 0.9766, "E": 0.0057, "G": 0.0024, "H": 0.0022, "K": 0.0001, "N": 0.0029, "R": 0.0001, "S": 0.0003, "T": 0.0002, "V": 0.0015, "Y": 0.0026 },
    "82": { "A": 0.0005, "D": 0.0003, "E": 0.0089, "G": 0.001, "H": 0.0003, "I": 0.0004, "K": 0.9246, "L": 0.0004, "M": 0.0031, "N": 0.0101, "P": 0.0003, "Q": 0.0062, "R": 0.0255, "S": 0.0012, "T": 0.0148, "V": 0.0014, "W": 0.0008, "Y": 0.0003 },
    "83": { "A": 0.0111, "C": 0.0004, "D": 0.0005, "E": 0.0002, "F": 0.0019, "G": 0.0002, "L": 0.0002, "P": 0.005, "S": 0.9739, "T": 0.003, "V": 0.0001, "Y": 0.0033 },
    "84": { "A": 0.0015, "C": 0.0001, "D": 0.0002, "F": 0.003, "G": 0.0004, "H": 0.0001, "I": 0.9036, "K": 0.0007, "L": 0.0121, "M": 0.0027, "N": 0.0056, "P": 0.0006, "Q": 0.0001, "R": 0.0011, "S": 0.0197, "T": 0.0337, "V": 0.0145, "Y": 0.0001 },
    "85": { "A": 0.0057, "C": 0.0006, "D": 0.0075, "E": 0.0003, "F": 0.0003, "G": 0.0074, "H": 0.0008, "I": 0.0023, "K": 0.0024, "L": 0.0001, "M": 0.0002, "N": 0.0752, "P": 0.0004, "Q": 0.0001, "R": 0.017, "S": 0.8442, "T": 0.0345, "V": 0.0002, "Y": 0.0008 },
    "86": { "A": 0.0065, "G": 0.0001, "H": 0.0003, "I": 0.005, "K": 0.0001, "M": 0.0004, "N": 0.0022, "P": 0.0049, "Q": 0.0001, "R": 0.0001, "S": 0.0047, "T": 0.9734, "V": 0.0021 },
    "87": { "A": 0.9619, "D": 0.002, "F": 0.0003, "G": 0.0011, "H": 0.0001, "I": 0.0008, "L": 0.0005, "P": 0.0013, "S": 0.0045, "T": 0.0091, "V": 0.0177, "Y": 0.0007 },
    "88": { "C": 0.0021, "D": 0.0054, "F": 0.0205, "H": 0.0042, "I": 0.0001, "L": 0.0006, "M": 0.0001, "N": 0.0007, "P": 0.0001, "Q": 0.0001, "S": 0.0052, "V": 0.0003, "Y": 0.9604 },
    "89": { "C": 0.0001, "F": 0.0004, "I": 0.0004, "L": 0.9785, "M": 0.0027, "P": 0.0012, "Q": 0.0009, "R": 0.009, "S": 0.0001, "V": 0.0064, "W": 0.0003, "Y": 0.0001 },
    "90": { "A": 0.0001, "D": 0.0001, "E": 0.0076, "G": 0.0001, "H": 0.0201, "K": 0.0056, "L": 0.0012, "M": 0.0005, "N": 0.0002, "P": 0.0017, "Q": 0.9546, "R": 0.0061, "S": 0.0006, "T": 0.0001, "Y": 0.0012 },
    "91": { "C": 0.0022, "F": 0.0014, "G": 0.0069, "L": 0.0104, "M": 0.0001, "R": 0.0021, "S": 0.0006, "V": 0.0005, "W": 0.9754, "Y": 0.0002 },
    "92": { "A": 0.0021, "C": 0.0007, "D": 0.0063, "E": 0.0005, "F": 0.0001, "G": 0.0155, "H": 0.0014, "I": 0.0029, "K": 0.0012, "L": 0.0002, "M": 0.0001, "N": 0.0278, "P": 0.0004, "Q": 0.0001, "R": 0.0214, "S": 0.8992, "T": 0.0192, "V": 0.0004, "W": 0.0001, "Y": 0.0003 },
    "93": { "A": 0.0011, "C": 0.0002, "D": 0.0027, "F": 0.0004, "G": 0.0091, "H": 0.0002, "I": 0.0016, "K": 0.0002, "L": 0.0001, "M": 0.0001, "N": 0.007, "P": 0.0001, "R": 0.0129, "S": 0.9565, "T": 0.0074, "V": 0.0003 },
    "94": { "G": 0.0001, "I": 0.0001, "L": 0.9843, "M": 0.0045, "P": 0.0012, "Q": 0.0008, "R": 0.0057, "V": 0.0031 },
    "95": { "A": 0.0003, "D": 0.0001, "E": 0.0146, "G": 0.0012, "K": 0.934, "L": 0.0001, "M": 0.0012, "N": 0.0027, "Q": 0.0131, "R": 0.0257, "S": 0.0005, "T": 0.0058, "V": 0.0001, "W": 0.0004 },
    "96": { "A": 0.9703, "D": 0.0029, "E": 0.0001, "G": 0.001, "P": 0.0066, "S": 0.0043, "T": 0.0101, "V": 0.0045 },
    "97": { "A": 0.0124, "E": 0.0001, "G": 0.0001, "L": 0.0009, "P": 0.0033, "R": 0.0001, "S": 0.9794, "T": 0.0022, "V": 0.0001, "W": 0.001, "Y": 0.0001 },
    "98": { "A": 0.0017, "C": 0.0001, "D": 0.9891, "E": 0.0037, "G": 0.0016, "H": 0.0001, "N": 0.001, "V": 0.0004, "Y": 0.002 },
    "99": { "A": 0.0032, "D": 0.0001, "G": 0.0004, "I": 0.0023, "N": 0.0035, "P": 0.0029, "R": 0.0001, "S": 0.0381, "T": 0.949, "V": 0.0001 },
    "100": { "A": 0.974, "C": 0.0001, "D": 0.0018, "G": 0.0178, "P": 0.0007, "R": 0.0001, "S": 0.0028, "T": 0.0018, "V": 0.0011 },
    "101": { "A": 0.0001, "C": 0.0001, "F": 0.0016, "G": 0.0001, "I": 0.0895, "K": 0.0034, "L": 0.0057, "M": 0.861, "N": 0.0006, "P": 0.0001, "R": 0.0048, "S": 0.0003, "T": 0.0172, "V": 0.0154 },
    "102": { "C": 0.0007, "D": 0.003, "E": 0.0002, "F": 0.0045, "H": 0.001, "I": 0.0001, "N": 0.0003, "S": 0.0005, "V": 0.0002, "Y": 0.9893 },
    "103": { "C": 0.0012, "D": 0.0034, "E": 0.0001, "F": 0.0328, "H": 0.0011, "I": 0.0001, "L": 0.0002, "N": 0.0002, "S": 0.0013, "V": 0.0002, "Y": 0.9592 },
    "104": { "C": 0.9852, "F": 0.0014, "G": 0.004, "L": 0.0004, "R": 0.0007, "S": 0.0008, "W": 0.0061, "Y": 0.0013 },
    "105": { "A": 0.9489, "C": 0.0002, "E": 0.001, "G": 0.0071, "H": 0.0004, "I": 0.0006, "L": 0.0002, "M": 0.0007, "P": 0.0005, "R": 0.0002, "S": 0.0036, "T": 0.0064, "V": 0.0295, "W": 0.0003, "Y": 0.0002 },
    "106": { "A": 0.0046, "C": 0.0013, "D": 0.0001, "E": 0.0005, "F": 0.0006, "G": 0.0106, "H": 0.0004, "I": 0.0125, "K": 0.0122, "L": 0.005, "M": 0.0027, "N": 0.0011, "P": 0.0002, "Q": 0.0004, "R": 0.8711, "S": 0.0416, "T": 0.0284, "V": 0.0056, "W": 0.0006, "Y": 0.0005 },
    "107": { "A": 0.0261, "C": 0.0049, "D": 0.0092, "E": 0.0154, "F": 0.0111, "G": 0.0721, "H": 0.1835, "I": 0.014, "K": 0.0085, "L": 0.1588, "M": 0.0051, "N": 0.0075, "P": 0.0745, "Q": 0.1109, "R": 0.1686, "S": 0.0537, "T": 0.0327, "V": 0.0222, "W": 0.0076, "Y": 0.0133 },
    "108": { "A": 0.0574, "C": 0.0053, "D": 0.052, "E": 0.0441, "F": 0.0221, "G": 0.1839, "H": 0.0201, "I": 0.0336, "K": 0.0237, "L": 0.0644, "M": 0.0129, "N": 0.0273, "P": 0.0733, "Q": 0.0226, "R": 0.0958, "S": 0.0937, "T": 0.0426, "V": 0.0554, "W": 0.0174, "Y": 0.0525 },
    "109": { "A": 0.0604, "C": 0.0176, "D": 0.0705, "E": 0.0331, "F": 0.0194, "G": 0.187, "H": 0.0174, "I": 0.0349, "K": 0.0152, "L": 0.0423, "M": 0.0121, "N": 0.0244, "P": 0.0467, "Q": 0.0234, "R": 0.0678, "S": 0.1101, "T": 0.0493, "V": 0.0559, "W": 0.0189, "Y": 0.0933 },
    "110": { "A": 0.068, "C": 0.0185, "D": 0.0672, "E": 0.0277, "F": 0.0235, "G": 0.1616, "H": 0.0135, "I": 0.032, "K": 0.0079, "L": 0.0435, "M": 0.0109, "N": 0.0257, "P": 0.0377, "Q": 0.0155, "R": 0.0476, "S": 0.1446, "T": 0.0527, "V": 0.063, "W": 0.0208, "Y": 0.1181 },
    "111": { "A": 0.0797, "C": 0.0216, "D": 0.0645, "E": 0.0216, "F": 0.0225, "G": 0.1723, "H": 0.0093, "I": 0.0256, "K": 0.0069, "L": 0.0337, "M": 0.0114, "N": 0.0245, "P": 0.0277, "Q": 0.0159, "R": 0.0398, "S": 0.1591, "T": 0.0599, "V": 0.0565, "W": 0.0235, "Y": 0.1241 },
    "111A": { "A": 0.0721, "C": 0.0226, "D": 0.0766, "E": 0.0244, "F": 0.0185, "G": 0.1603, "H": 0.0098, "I": 0.029, "K": 0.0119, "L": 0.0373, "M": 0.0097, "N": 0.024, "P": 0.0259, "Q": 0.0122, "R": 0.0319, "S": 0.1747, "T": 0.0574, "V": 0.0535, "W": 0.0382, "Y": 0.1101 },
    "111B": { "A": 0.0548, "C": 0.0237, "D": 0.0721, "E": 0.0143, "F": 0.0246, "G": 0.1655, "H": 0.01, "I": 0.0271, "K": 0.0088, "L": 0.0342, "M": 0.0087, "N": 0.024, "P": 0.0313, "Q": 0.0075, "R": 0.0342, "S": 0.2073, "T": 0.0605, "V": 0.0575, "W": 0.03, "Y": 0.1039 },
    "111C": { "A": 0.0496, "C": 0.0384, "D": 0.0501, "E": 0.0178, "F": 0.0335, "G": 0.1696, "H": 0.01, "I": 0.0255, "K": 0.0115, "L": 0.0387, "M": 0.0054, "N": 0.0195, "P": 0.0309, "Q": 0.0097, "R": 0.0378, "S": 0.2025, "T": 0.0771, "V": 0.0539, "W": 0.0286, "Y": 0.0899 },
    "111D": { "A": 0.0486, "C": 0.0479, "D": 0.0461, "E": 0.0202, "F": 0.0265, "G": 0.2025, "H": 0.005, "I": 0.0259, "K": 0.0095, "L": 0.0303, "M": 0.0063, "N": 0.0233, "P": 0.0297, "Q": 0.0101, "R": 0.0341, "S": 0.1748, "T": 0.0587, "V": 0.0467, "W": 0.0391, "Y": 0.1148 },
    "111E": { "A": 0.062, "C": 0.0652, "D": 0.0445, "E": 0.0095, "F": 0.0127, "G": 0.1463, "H": 0.0064, "I": 0.0238, "K": 0.0064, "L": 0.0413, "M": 0.0095, "N": 0.0159, "P": 0.0572, "Q": 0.0143, "R": 0.0366, "S": 0.1749, "T": 0.0588, "V": 0.0429, "W": 0.0334, "Y": 0.1383 },
    "111F": { "A": 0.0524, "C": 0.0655, "D": 0.0393, "E": 0.0131, "F": 0.0306, "G": 0.1747, "H": 0.0131, "I": 0.048, "K": 0.0044, "L": 0.0655, "N": 0.0087, "P": 0.0262, "Q": 0.0044, "R": 0.0568, "S": 0.1572, "T": 0.0699, "V": 0.0218, "W": 0.0262, "Y": 0.1223 },
    "112G": { "A": 0.0238, "C": 0.0873, "D": 0.0397, "E": 0.0079, "F": 0.0238, "G": 0.1508, "I": 0.0079, "K": 0.0476, "L": 0.0556, "M": 0.0079, "N": 0.0079, "P": 0.0635, "Q": 0.0317, "R": 0.0397, "S": 0.119, "T": 0.0794, "V": 0.0476, "W": 0.0159, "Y": 0.1429 },
    "112F": { "A": 0.0508, "C": 0.0722, "D": 0.0481, "E": 0.0374, "F": 0.0241, "G": 0.1337, "H": 0.0107, "I": 0.0267, "K": 0.016, "L": 0.0428, "M": 0.008, "N": 0.0214, "P": 0.0829, "Q": 0.0214, "R": 0.0348, "S": 0.1658, "T": 0.0241, "V": 0.0481, "W": 0.0187, "Y": 0.1123 },
    "112E": { "A": 0.0569, "C": 0.0474, "D": 0.0284, "E": 0.0227, "F": 0.0199, "G": 0.1564, "H": 0.0123, "I": 0.0256, "K": 0.0171, "L": 0.0398, "M": 0.0066, "N": 0.018, "P": 0.0882, "Q": 0.0199, "R": 0.055, "S": 0.163, "T": 0.0483, "V": 0.0389, "W": 0.0199, "Y": 0.1156 },
    "112D": { "A": 0.0479, "C": 0.037, "D": 0.0431, "E": 0.0248, "F": 0.0257, "G": 0.154, "H": 0.0161, "I": 0.0196, "K": 0.0126, "L": 0.0348, "M": 0.0074, "N": 0.0274, "P": 0.0601, "Q": 0.0144, "R": 0.0435, "S": 0.1549, "T": 0.0735, "V": 0.0413, "W": 0.0252, "Y": 0.1366 },
    "112C": { "A": 0.0504, "C": 0.0409, "D": 0.0599, "E": 0.0173, "F": 0.0204, "G": 0.1524, "H": 0.0171, "I": 0.0167, "K": 0.0154, "L": 0.0409, "M": 0.0103, "N": 0.0352, "P": 0.052, "Q": 0.0134, "R": 0.0518, "S": 0.1403, "T": 0.0522, "V": 0.0366, "W": 0.0226, "Y": 0.1543 },
    "112B": { "A": 0.0464, "C": 0.0328, "D": 0.0591, "E": 0.0244, "F": 0.0197, "G": 0.1462, "H": 0.0184, "I": 0.0191, "K": 0.009, "L": 0.0366, "M": 0.0068, "N": 0.0343, "P": 0.0365, "Q": 0.011, "R": 0.0363, "S": 0.1303, "T": 0.0515, "V": 0.0402, "W": 0.0234, "Y": 0.2181 },
    "112A": { "A": 0.0602, "C": 0.018, "D": 0.0396, "E": 0.0266, "F": 0.0209, "G": 0.1483, "H": 0.0144, "I": 0.0199, "K": 0.01, "L": 0.0374, "M": 0.006, "N": 0.0299, "P": 0.0526, "Q": 0.0143, "R": 0.0393, "S": 0.1273, "T": 0.056, "V": 0.0393, "W": 0.0288, "Y": 0.2112 },
    "112": { "A": 0.0459, "C": 0.0112, "D": 0.0504, "E": 0.0283, "F": 0.0273, "G": 0.1282, "H": 0.0254, "I": 0.0225, "K": 0.0126, "L": 0.052, "M": 0.0087, "N": 0.0321, "P": 0.0612, "Q": 0.0173, "R": 0.0409, "S": 0.0996, "T": 0.0498, "V": 0.0415, "W": 0.0338, "Y": 0.2113 },
    "113": { "A": 0.0377, "C": 0.0115, "D": 0.125, "E": 0.0242, "F": 0.0168, "G": 0.1463, "H": 0.0189, "I": 0.0086, "K": 0.007, "L": 0.0264, "M": 0.0041, "N": 0.081, "P": 0.0227, "Q": 0.0106, "R": 0.0339, "S": 0.0819, "T": 0.0327, "V": 0.0216, "W": 0.0549, "Y": 0.2342 },
    "114": { "A": 0.1803, "C": 0.0056, "D": 0.0234, "E": 0.0197, "F": 0.0184, "G": 0.1855, "H": 0.0194, "I": 0.009, "K": 0.0037, "L": 0.021, "M": 0.0033, "N": 0.0183, "P": 0.0648, "Q": 0.0067, "R": 0.0154, "S": 0.0452, "T": 0.0264, "V": 0.0152, "W": 0.1075, "Y": 0.2111 },
    "115": { "A": 0.0128, "C": 0.0056, "D": 0.0063, "E": 0.0052, "F": 0.5603, "G": 0.0272, "H": 0.0029, "I": 0.0311, "K": 0.0015, "L": 0.0637, "M": 0.1584, "N": 0.0057, "P": 0.0247, "Q": 0.0014, "R": 0.0035, "S": 0.0189, "T": 0.0085, "V": 0.0296, "W": 0.0085, "Y": 0.024 },
    "116": { "A": 0.0093, "C": 0.0002, "D": 0.8766, "E": 0.0142, "F": 0.002, "G": 0.0277, "H": 0.0057, "I": 0.0029, "K": 0.0007, "L": 0.0038, "M": 0.0006, "N": 0.0091, "P": 0.0036, "Q": 0.0132, "R": 0.0036, "S": 0.0059, "T": 0.0038, "V": 0.0077, "W": 0.0011, "Y": 0.0082 },
    "117": { "A": 0.0029, "C": 0.0024, "D": 0.0087, "E": 0.0007, "F": 0.025, "G": 0.0017, "H": 0.0226, "I": 0.1535, "K": 0.0006, "L": 0.049, "M": 0.0022, "N": 0.0079, "P": 0.1136, "Q": 0.0014, "R": 0.0021, "S": 0.022, "T": 0.0047, "V": 0.1773, "W": 0.0008, "Y": 0.4009 },
    "118": { "C": 0.0011, "F": 0.0003, "G": 0.0026, "L": 0.0012, "R": 0.001, "V": 0.0003, "W": 0.9934 },
    "119": { "A": 0.002, "C": 0.0015, "D": 0.0007, "G": 0.9913, "R": 0.0001, "S": 0.0034, "V": 0.0008 },
    "120": { "A": 0.0001, "D": 0.0001, "E": 0.001, "G": 0.0003, "H": 0.0045, "K": 0.0356, "L": 0.0026, "N": 0.0006, "P": 0.0049, "Q": 0.9065, "R": 0.0432, "S": 0.0004, "T": 0.0002 },
    "121": { "A": 0.0001, "D": 0.0001, "E": 0.0006, "G": 0.9967, "R": 0.0009, "V": 0.0014, "W": 0.0002 },
    "122": { "A": 0.0067, "I": 0.0016, "K": 0.0005, "N": 0.0007, "P": 0.0011, "R": 0.0002, "S": 0.0051, "T": 0.9836, "V": 0.0003 },
    "123": { "A": 0.0023, "D": 0.0001, "E": 0.0003, "F": 0.0003, "H": 0.0003, "I": 0.0004, "K": 0.0028, "L": 0.6362, "M": 0.1568, "N": 0.0002, "P": 0.0117, "Q": 0.0104, "R": 0.0063, "S": 0.0026, "T": 0.1635, "V": 0.0055, "W": 0.0004 },
    "124": { "A": 0.001, "D": 0.0003, "F": 0.0008, "G": 0.0016, "I": 0.0039, "L": 0.0056, "V": 0.9867 },
    "125": { "A": 0.004, "F": 0.0002, "H": 0.0002, "I": 0.0131, "L": 0.0017, "M": 0.0006, "N": 0.0031, "P": 0.0006, "R": 0.0002, "S": 0.0105, "T": 0.9652, "V": 0.0005 },
    "126": { "A": 0.0022, "D": 0.0007, "F": 0.0008, "G": 0.0009, "I": 0.0017, "L": 0.001, "R": 0.0002, "V": 0.9924 },
    "127": { "A": 0.0049, "C": 0.0002, "F": 0.0015, "L": 0.0004, "P": 0.002, "S": 0.9866, "T": 0.0025, "Y": 0.0018 },
    "128": { "A": 0.0049, "F": 0.0001, "L": 0.0003, "P": 0.0053, "S": 0.9878, "T": 0.0016 }
  },
  "IGHV6": {
    "1": { "H": 0.0094, "Q": 0.9905, "V": 0.0001 },
    "2": { "A": 0.0054, "D": 0.0007, "E": 0.0032, "F": 0.0005, "G": 0.0024, "I": 0.0124, "L": 0.0145, "M": 0.0008, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0002, "T": 0.0005, "V": 0.959 },
    "3": { "D": 0.0003, "E": 0.0049, "H": 0.0165, "K": 0.0022, "L": 0.005, "N": 0.0004, "P": 0.0075, "Q": 0.95, "R": 0.0088, "S": 0.0011, "T": 0.0005, "V": 0.0004, "Y": 0.0023 },
    "4": { "F": 0.0011, "H": 0.0002, "I": 0.0002, "L": 0.9851, "M": 0.0032, "P": 0.004, "Q": 0.0008, "R": 0.0026, "S": 0.0003, "T": 0.0001, "V": 0.0021, "W": 0.0001 },
    "5": { "A": 0.0001, "D": 0.0001, "E": 0.0015, "F": 0.0001, "G": 0.0012, "H": 0.012, "K": 0.0034, "L": 0.0044, "M": 0.001, "N": 0.0001, "P": 0.0135, "Q": 0.9477, "R": 0.0072, "S": 0.0001, "T": 0.0005, "V": 0.0063, "W": 0.0002, "Y": 0.0003 },
    "6": { "A": 0.0005, "E": 0.0152, "H": 0.0031, "K": 0.0015, "L": 0.0051, "P": 0.0124, "Q": 0.9586, "R": 0.0018, "S": 0.0006, "T": 0.0001, "V": 0.0009 },
    "7": { "A": 0.0055, "F": 0.0004, "G": 0.0001, "H": 0.0001, "L": 0.0014, "P": 0.0073, "Q": 0.0001, "S": 0.9819, "T": 0.0023, "V": 0.0001, "W": 0.0001, "Y": 0.0004 },
    "8": { "A": 0.0036, "C": 0.002, "D": 0.0013, "G": 0.9821, "H": 0.0001, "P": 0.0002, "R": 0.0028, "S": 0.0031, "V": 0.0046 },
    "9": { "A": 0.002, "G": 0.0002, "H": 0.0004, "L": 0.005, "P": 0.9791, "Q": 0.0047, "R": 0.0019, "S": 0.0029, "T": 0.0035 },
    "11": { "A": 0.0021, "C": 0.0001, "D": 0.0002, "E": 0.0029, "G": 0.9788, "H": 0.0001, "L": 0.0002, "Q": 0.0006, "R": 0.0119, "S": 0.0005, "V": 0.0024, "W": 0.0001 },
    "12": { "H": 0.0002, "L": 0.9781, "M": 0.0025, "P": 0.0046, "Q": 0.0022, "R": 0.0074, "S": 0.0001, "T": 0.0001, "V": 0.0046 },
    "13": { "A": 0.0086, "C": 0.0001, "D": 0.0002, "E": 0.0014, "F": 0.0003, "G": 0.0077, "I": 0.0009, "K": 0.0005, "L": 0.0182, "M": 0.0031, "Q": 0.0001, "R": 0.0002, "S": 0.0002, "V": 0.9579, "W": 0.0003 },
    "14": { "A": 0.0004, "E": 0.0119, "G": 0.0007, "H": 0.0003, "I": 0.0002, "K": 0.9188, "L": 0.0001, "M": 0.0016, "N": 0.0064, "P": 0.0004, "Q": 0.0211, "R": 0.0172, "S": 0.0007, "T": 0.0202 },
    "15": { "A": 0.0054, "H": 0.0095, "L": 0.0038, "P": 0.9654, "Q": 0.0002, "R": 0.0019, "S": 0.007, "T": 0.0066 },
    "16": { "A": 0.0053, "C": 0.0003, "D": 0.0001, "E": 0.0002, "G": 0.0005, "H": 0.0001, "L": 0.0016, "P": 0.0076, "Q": 0.0003, "R": 0.0003, "S": 0.9806, "T": 0.0014, "V": 0.0001, "W": 0.0011, "Y": 0.0002 },
    "17": { "A": 0.0004, "D": 0.0002, "E": 0.008, "G": 0.0002, "H": 0.0075, "K": 0.0032, "L": 0.0013, "N": 0.0001, "P": 0.0164, "Q": 0.9598, "R": 0.0019, "S": 0.0001, "T": 0.0006, "Y": 0.0001 },
    "18": { "A": 0.0024, "D": 0.0002, "H": 0.0009, "I": 0.0029, "L": 0.0005, "N": 0.0033, "P": 0.0218, "R": 0.0002, "S": 0.0059, "T": 0.9615 },
    "19": { "C": 0.0001, "F": 0.0018, "G": 0.0001, "H": 0.002, "I": 0.0058, "L": 0.9595, "P": 0.0205, "R": 0.0065, "S": 0.0009, "T": 0.0005, "V": 0.0022 },
    "20": { "A": 0.0096, "E": 0.0001, "F": 0.001, "H": 0.0006, "I": 0.0001, "K": 0.0003, "L": 0.0025, "N": 0.0001, "P": 0.0407, "Q": 0.001, "R": 0.0006, "S": 0.9376, "T": 0.0046, "V": 0.0002, "Y": 0.0009 },
    "21": { "A": 0.0003, "C": 0.0001, "F": 0.0012, "H": 0.0024, "I": 0.0051, "L": 0.9529, "P": 0.0183, "R": 0.0044, "S": 0.0006, "T": 0.0009, "V": 0.0137 },
    "22": { "A": 0.0042, "D": 0.0013, "F": 0.0002, "H": 0.0009, "I": 0.0028, "K": 0.0002, "L": 0.0002, "M": 0.0001, "N": 0.0032, "P": 0.0311, "R": 0.0003, "S": 0.0067, "T": 0.9484, "V": 0.0001, "W": 0.0001 },
    "23": { "C": 0.9631, "D": 0.0002, "F": 0.0113, "G": 0.0045, "H": 0.0001, "L": 0.0012, "P": 0.0001, "R": 0.0052, "S": 0.0079, "V": 0.0011, "W": 0.0035, "Y": 0.0015 },
    "24": { "A": 0.8841, "C": 0.0004, "D": 0.0167, "F": 0.0009, "G": 0.0088, "H": 0.0005, "I": 0.0004, "K": 0.0004, "L": 0.0004, "N": 0.0019, "P": 0.0164, "R": 0.0011, "S": 0.0283, "T": 0.0057, "V": 0.0333, "Y": 0.0004 },
    "25": { "A": 0.0003, "C": 0.0001, "D": 0.0002, "F": 0.0028, "G": 0.0003, "H": 0.0001, "I": 0.9423, "L": 0.0277, "M": 0.001, "N": 0.0012, "P": 0.0005, "R": 0.0004, "S": 0.0044, "T": 0.0061, "V": 0.0124, "Y": 0.0001 },
    "26": { "A": 0.0077, "C": 0.0008, "D": 0.0003, "F": 0.0039, "H": 0.0009, "I": 0.0002, "L": 0.0007, "P": 0.0532, "R": 0.0002, "S": 0.9227, "T": 0.0043, "V": 0.0001, "Y": 0.0048 },
    "27": { "A": 0.0069, "C": 0.0002, "D": 0.0003, "E": 0.003, "F": 0.0001, "G": 0.9677, "P": 0.0001, "R": 0.0071, "S": 0.0007, "V": 0.0084, "W": 0.0054 },
    "28": { "A": 0.0563, "C": 0.0001, "D": 0.9054, "E": 0.0081, "F": 0.0001, "G": 0.0044, "H": 0.0073, "K": 0.0001, "L": 0.0002, "N": 0.003, "P": 0.0002, "Q": 0.0002, "R": 0.0001, "S": 0.0013, "T": 0.0011, "V": 0.0038, "Y": 0.0082 },
    "29": { "A": 0.0002, "C": 0.0024, "D": 0.0008, "F": 0.0006, "G": 0.007, "H": 0.0009, "I": 0.0056, "K": 0.0003, "L": 0.0011, "M": 0.0005, "N": 0.0066, "P": 0.0008, "Q": 0.0002, "R": 0.0696, "S": 0.8846, "T": 0.0182, "V": 0.0003, "Y": 0.0001 },
    "30": { "A": 0.0138, "C": 0.0012, "D": 0.0023, "F": 0.0138, "G": 0.0082, "I": 0.0093, "L": 0.0133, "P": 0.0006, "R": 0.0002, "S": 0.0011, "T": 0.0004, "V": 0.9355, "Y": 0.0002 },
    "31": { "A": 0.0169, "C": 0.001, "D": 0.0002, "F": 0.0066, "G": 0.0001, "H": 0.002, "I": 0.0001, "K": 0.0001, "L": 0.0036, "P": 0.0771, "R": 0.0003, "S": 0.8787, "T": 0.0087, "V": 0.0004, "Y": 0.0042 },
    "34": { "A": 0.0049, "C": 0.0021, "D": 0.0027, "E": 0.0001, "F": 0.0002, "G": 0.0139, "H": 0.0018, "I": 0.0052, "K": 0.0007, "L": 0.0004, "M": 0.0001, "N": 0.025, "P": 0.001, "R": 0.0538, "S": 0.8543, "T": 0.0321, "V": 0.0012, "W": 0.0001, "Y": 0.0001 },
    "35": { "A": 0.002, "C": 0.0002, "D": 0.0288, "E": 0.0007, "F": 0.0002, "G": 0.0035, "H": 0.0441, "I": 0.0057, "K": 0.022, "L": 0.0001, "N": 0.8066, "P": 0.0018, "Q": 0.0018, "R": 0.0025, "S": 0.0299, "T": 0.0408, "V": 0.0006, "Y": 0.0087 },
    "36": { "A": 0.0017, "C": 0.0016, "D": 0.0052, "E": 0.0007, "F": 0.0018, "G": 0.0223, "H": 0.0024, "I": 0.0153, "K": 0.0009, "L": 0.0022, "M": 0.0006, "N": 0.0272, "P": 0.0032, "Q": 0.0004, "R": 0.056, "S": 0.8139, "T": 0.042, "V": 0.002, "W": 0.0001, "Y": 0.0005 },
    "37": { "A": 0.8581, "C": 0.0006, "D": 0.0063, "E": 0.0006, "F": 0.0013, "G": 0.0146, "H": 0.0005, "I": 0.0055, "L": 0.0017, "N": 0.0008, "P": 0.022, "R": 0.0002, "S": 0.0245, "T": 0.0122, "V": 0.0498, "Y": 0.001 },
    "38": { "A": 0.8845, "C": 0.0008, "D": 0.0041, "E": 0.0004, "F": 0.0004, "G": 0.0124, "H": 0.0004, "I": 0.0027, "L": 0.0012, "N": 0.0002, "P": 0.0147, "R": 0.0003, "S": 0.0247, "T": 0.0423, "V": 0.0105, "W": 0.0002, "Y": 0.0002 },
    "39": { "C": 0.0095, "F": 0.0002, "G": 0.0047, "I": 0.0004, "L": 0.0051, "R": 0.0051, "S": 0.0032, "V": 0.0002, "W": 0.9712, "Y": 0.0003 },
    "40": { "A": 0.0012, "C": 0.0001, "D": 0.0069, "E": 0.0002, "G": 0.0009, "H": 0.0455, "I": 0.0044, "K": 0.004, "L": 0.0002, "M": 0.0001, "N": 0.8768, "P": 0.001, "Q": 0.0003, "R": 0.0006, "S": 0.02, "T": 0.0335, "V": 0.0004, "Y": 0.0037 },
    "41": { "C": 0.0071, "F": 0.0001, "G": 0.0063, "L": 0.0052, "R": 0.0055, "S": 0.0023, "T": 0.0001, "V": 0.0005, "W": 0.9726 },
    "42": { "A": 0.0005, "C": 0.0002, "D": 0.0002, "F": 0.0071, "G": 0.0001, "H": 0.0012, "I": 0.9153, "K": 0.0001, "L": 0.0285, "M": 0.0015, "N": 0.0047, "P": 0.0006, "R": 0.0008, "S": 0.0079, "T": 0.0119, "V": 0.0187, "Y": 0.0004 },
    "43": { "G": 0.0034, "K": 0.0017, "L": 0.0002, "M": 0.003, "P": 0.0001, "Q": 0.0004, "R": 0.9809, "S": 0.0074, "T": 0.0016, "V": 0.0001, "W": 0.0011 },
    "44": { "A": 0.0005, "E": 0.0016, "H": 0.0079, "K": 0.0041, "L": 0.0049, "M": 0.0001, "P": 0.0147, "Q": 0.961, "R": 0.0033, "S": 0.0003, "T": 0.0012, "W": 0.0001 },
    "45": { "A": 0.0092, "C": 0.0012, "D": 0.0002, "F": 0.0031, "H": 0.0001, "L": 0.0001, "M": 0.0002, "P": 0.0065, "R": 0.0001, "S": 0.9716, "T": 0.0022, "V": 0.0008, "Y": 0.0046 },
    "46": { "A": 0.0008, "H": 0.0013, "L": 0.0009, "P": 0.9827, "Q": 0.006, "R": 0.0009, "S": 0.0021, "T": 0.0053 },
    "47": { "A": 0.0125, "C": 0.0001, "E": 0.0001, "F": 0.0003, "G": 0.0006, "I": 0.0002, "L": 0.005, "M": 0.0001, "P": 0.0055, "Q": 0.0004, "R": 0.0008, "S": 0.9663, "T": 0.0053, "V": 0.0001, "W": 0.0022, "Y": 0.0002 },
    "48": { "A": 0.0002, "D": 0.0001, "E": 0.0005, "G": 0.022, "I": 0.0037, "K": 0.0082, "L": 0.0004, "M": 0.0004, "N": 0.0012, "P": 0.0002, "Q": 0.0009, "R": 0.946, "S": 0.0124, "T": 0.0031, "V": 0.0002 },
    "49": { "A": 0.004, "C": 0.0047, "D": 0.0022, "E": 0.0001, "G": 0.9802, "H": 0.0001, "L": 0.0001, "R": 0.0023, "S": 0.0024, "V": 0.0038, "W": 0.0001 },
    "50": { "A": 0.0002, "C": 0.0001, "F": 0.0057, "H": 0.0006, "I": 0.0043, "L": 0.9755, "P": 0.0051, "R": 0.0064, "S": 0.0002, "T": 0.0001, "V": 0.0016 },
    "51": { "A": 0.0128, "D": 0.005, "E": 0.9632, "G": 0.01, "K": 0.0008, "L": 0.0003, "P": 0.0002, "Q": 0.0032, "R": 0.0004, "S": 0.0004, "V": 0.002, "W": 0.0014, "Y": 0.0001 },
    "52": { "C": 0.0055, "F": 0.0006, "G": 0.0093, "L": 0.0048, "R": 0.0025, "S": 0.0009, "V": 0.0011, "W": 0.9746, "Y": 0.0007 },
    "53": { "A": 0.0018, "G": 0.0005, "I": 0.0002, "L": 0.9751, "M": 0.0053, "P": 0.005, "Q": 0.001, "R": 0.0069, "S": 0.0002, "V": 0.0033, "W": 0.0004 },
    "54": { "A": 0.0054, "C": 0.0001, "D": 0.0001, "E": 0.0016, "G": 0.9853, "I": 0.0001, "R": 0.0025, "S": 0.0005, "T": 0.0003, "V": 0.0037, "W": 0.0001 },
    "55": { "A": 0.0002, "E": 0.0003, "G": 0.0047, "I": 0.0002, "K": 0.0113, "L": 0.0001, "M": 0.0054, "N": 0.0002, "Q": 0.0003, "R": 0.9685, "S": 0.0053, "T": 0.0016, "W": 0.0017 },
    "56": { "A": 0.0163, "D": 0.0001, "G": 0.001, "H": 0.0002, "I": 0.0086, "K": 0.0062, "L": 0.0001, "N": 0.0008, "P": 0.0058, "Q": 0.0004, "R": 0.0017, "S": 0.0023, "T": 0.9557, "V": 0.0006 },
    "57": { "A": 0.0002, "C": 0.0029, "D": 0.0062, "E": 0.0002, "F": 0.0179, "G": 0.0002, "H": 0.0043, "I": 0.0007, "K": 0.0009, "L": 0.0013, "M": 0.0001, "N": 0.005, "Q": 0.0036, "R": 0.0002, "S": 0.0098, "T": 0.0004, "V": 0.0004, "W": 0.0009, "Y": 0.9445 },
    "58": { "A": 0.0002, "C": 0.0035, "D": 0.0097, "E": 0.0004, "F": 0.0134, "H": 0.0161, "I": 0.0001, "K": 0.0001, "L": 0.0056, "N": 0.0035, "P": 0.0005, "Q": 0.0014, "R": 0.0032, "S": 0.0129, "T": 0.0004, "V": 0.0004, "W": 0.0001, "Y": 0.9282 },
    "59": { "A": 0.0004, "C": 0.0001, "D": 0.0001, "E": 0.0006, "G": 0.0097, "I": 0.0001, "K": 0.0135, "L": 0.0003, "M": 0.0093, "N": 0.0032, "P": 0.0001, "Q": 0.0003, "R": 0.9359, "S": 0.0133, "T": 0.0087, "V": 0.0002, "W": 0.0028, "Y": 0.0013 },
    "60": { "A": 0.0181, "C": 0.0018, "D": 0.0003, "F": 0.0031, "G": 0.0001, "H": 0.0002, "L": 0.0007, "M": 0.0001, "N": 0.0001, "P": 0.0027, "R": 0.0014, "S": 0.9588, "T": 0.0064, "V": 0.0002, "Y": 0.0058 },
    "62": { "A": 0.0003, "D": 0.0005, "E": 0.0082, "G": 0.0019, "H": 0.0004, "I": 0.0002, "K": 0.9052, "M": 0.0028, "N": 0.0159, "P": 0.0002, "Q": 0.0174, "R": 0.0307, "S": 0.0044, "T": 0.0093, "V": 0.0002, "W": 0.0006, "Y": 0.0016 },
    "63": { "A": 0.0001, "C": 0.0081, "E": 0.0002, "F": 0.0031, "G": 0.011, "K": 0.0001, "L": 0.0128, "M": 0.0006, "N": 0.0003, "P": 0.0001, "R": 0.003, "S": 0.0021, "V": 0.001, "W": 0.9552, "Y": 0.0019 },
    "64": { "A": 0.0004, "C": 0.0014, "D": 0.0071, "E": 0.0012, "F": 0.0325, "G": 0.0018, "H": 0.0159, "I": 0.0047, "K": 0.0032, "L": 0.0052, "M": 0.0001, "N": 0.0178, "P": 0.0001, "Q": 0.0026, "R": 0.0046, "S": 0.0239, "T": 0.002, "V": 0.0018, "W": 0.001, "Y": 0.8727 },
    "65": { "A": 0.0004, "C": 0.0002, "D": 0.0075, "E": 0.0029, "F": 0.0073, "G": 0.0004, "H": 0.0166, "I": 0.0115, "K": 0.0119, "L": 0.0009, "M": 0.0002, "N": 0.8314, "P": 0.0008, "Q": 0.0012, "R": 0.0035, "S": 0.0366, "T": 0.0232, "V": 0.0013, "Y": 0.0422 },
    "66": { "A": 0.0057, "C": 0.0002, "D": 0.8925, "E": 0.0623, "F": 0.0001, "G": 0.0072, "H": 0.0062, "K": 0.0002, "L": 0.0009, "N": 0.009, "Q": 0.0022, "R": 0.0003, "S": 0.0007, "T": 0.0006, "V": 0.0045, "Y": 0.0071 },
    "67": { "A": 0.0003, "C": 0.0023, "D": 0.0086, "E": 0.0005, "F": 0.0086, "H": 0.002, "I": 0.0001, "L": 0.0019, "N": 0.0021, "P": 0.0001, "S": 0.0073, "V": 0.0004, "Y": 0.9655 },
    "68": { "A": 0.952, "C": 0.0002, "D": 0.0007, "E": 0.0084, "G": 0.0089, "K": 0.0001, "L": 0.0002, "N": 0.0002, "P": 0.0062, "Q": 0.0001, "S": 0.0128, "T": 0.0043, "V": 0.0057 },
    "69": { "A": 0.0171, "C": 0.0002, "D": 0.0064, "E": 0.0297, "F": 0.0027, "G": 0.0127, "H": 0.0002, "I": 0.0185, "K": 0.0009, "L": 0.0447, "M": 0.0022, "N": 0.0001, "P": 0.0157, "Q": 0.001, "R": 0.0011, "S": 0.0032, "T": 0.0086, "V": 0.8348 },
    "70": { "A": 0.0157, "C": 0.0004, "D": 0.0006, "F": 0.0121, "G": 0.0001, "H": 0.0004, "I": 0.0002, "L": 0.0009, "N": 0.0001, "P": 0.0041, "Q": 0.0001, "S": 0.9553, "T": 0.0052, "V": 0.0007, "Y": 0.0038 },
    "71": { "A": 0.0026, "C": 0.0003, "E": 0.0017, "F": 0.0012, "G": 0.0078, "I": 0.0002, "L": 0.0275, "M": 0.017, "R": 0.0005, "V": 0.9404, "W": 0.0006 },
    "72": { "A": 0.0012, "D": 0.0007, "E": 0.017, "G": 0.0013, "H": 0.0005, "I": 0.002, "K": 0.8973, "L": 0.0001, "M": 0.0005, "N": 0.0079, "P": 0.0001, "Q": 0.0124, "R": 0.0506, "S": 0.0017, "T": 0.0063, "V": 0.0001, "Y": 0.0003 },
    "74": { "A": 0.0006, "C": 0.0009, "D": 0.0006, "G": 0.0742, "H": 0.0006, "I": 0.0022, "K": 0.0001, "L": 0.0001, "M": 0.0003, "N": 0.0021, "R": 0.0129, "S": 0.9009, "T": 0.0039, "V": 0.0004 },
    "75": { "A": 0.0001, "C": 0.0002, "G": 0.0008, "H": 0.0001, "L": 0.0039, "P": 0.0009, "Q": 0.0022, "R": 0.9908, "S": 0.0003, "T": 0.0001, "V": 0.0002, "W": 0.0001 },
    "76": { "A": 0.0013, "E": 0.0001, "F": 0.0001, "G": 0.0001, "I": 0.9324, "K": 0.0015, "L": 0.0188, "M": 0.0211, "N": 0.0001, "P": 0.0001, "R": 0.0048, "S": 0.0005, "T": 0.0033, "V": 0.0155, "Y": 0.0001 },
    "77": { "A": 0.0124, "D": 0.0013, "F": 0.0018, "G": 0.0008, "H": 0.0007, "I": 0.031, "K": 0.0017, "L": 0.0016, "M": 0.0008, "N": 0.0138, "P": 0.0073, "Q": 0.0004, "R": 0.0024, "S": 0.0344, "T": 0.8839, "V": 0.0056 },
    "78": { "A": 0.0002, "C": 0.0002, "F": 0.011, "G": 0.0001, "H": 0.0002, "I": 0.9279, "K": 0.0001, "L": 0.0093, "M": 0.0021, "N": 0.0042, "R": 0.0004, "S": 0.0115, "T": 0.0034, "V": 0.0292 },
    "79": { "A": 0.0012, "D": 0.0145, "E": 0.0011, "F": 0.0002, "G": 0.0003, "H": 0.0068, "I": 0.0087, "K": 0.0179, "L": 0.0002, "M": 0.0001, "N": 0.8796, "P": 0.0003, "Q": 0.0022, "R": 0.003, "S": 0.0309, "T": 0.0289, "V": 0.0007, "Y": 0.0034 },
    "80": { "A": 0.0475, "E": 0.0004, "F": 0.0001, "G": 0.0007, "H": 0.0007, "I": 0.0001, "L": 0.0028, "P": 0.906, "Q": 0.0031, "R": 0.0012, "S": 0.0219, "T": 0.0111, "V": 0.0042, "Y": 0.0001 },
    "81": { "A": 0.0055, "D": 0.9735, "E": 0.0075, "G": 0.0027, "H": 0.0012, "I": 0.0002, "K": 0.0003, "N": 0.0021, "R": 0.0001, "S": 0.0003, "T": 0.0002, "V": 0.0024, "Y": 0.0038 },
    "82": { "A": 0.0054, "E": 0.0002, "H": 0.0002, "I": 0.0055, "K": 0.0037, "L": 0.0001, "M": 0.0001, "N": 0.0004, "P": 0.0099, "Q": 0.0006, "R": 0.001, "S": 0.006, "T": 0.9663, "V": 0.0005 },
    "83": { "A": 0.0136, "C": 0.0008, "D": 0.0008, "E": 0.0003, "F": 0.0033, "G": 0.0003, "H": 0.0001, "I": 0.0016, "L": 0.0001, "N": 0.0001, "P": 0.0078, "R": 0.0001, "S": 0.962, "T": 0.0061, "V": 0.0001, "W": 0.0001, "Y": 0.0025 },
    "84": { "A": 0.0007, "D": 0.0008, "E": 0.0084, "G": 0.0006, "H": 0.0003, "I": 0.0001, "K": 0.9339, "M": 0.0025, "N": 0.0113, "P": 0.0005, "Q": 0.0103, "R": 0.0223, "S": 0.0007, "T": 0.0065, "V": 0.0008, "Y": 0.0001 },
    "85": { "D": 0.0026, "H": 0.0023, "I": 0.0021, "K": 0.0031, "L": 0.0001, "N": 0.9769, "P": 0.0001, "S": 0.0057, "T": 0.0062, "Y": 0.0007 },
    "86": { "E": 0.0055, "H": 0.0244, "K": 0.003, "L": 0.0061, "P": 0.0039, "Q": 0.9473, "R": 0.0091, "S": 0.0002, "T": 0.0001, "Y": 0.0002 },
    "87": { "C": 0.0088, "F": 0.9327, "G": 0.0001, "I": 0.0105, "L": 0.0184, "M": 0.0001, "S": 0.0019, "V": 0.0252, "Y": 0.0019 },
    "88": { "A": 0.0078, "C": 0.0006, "D": 0.0001, "F": 0.0052, "I": 0.0004, "L": 0.0001, "N": 0.0001, "P": 0.0036, "S": 0.9738, "T": 0.0053, "V": 0.0005, "Y": 0.0023 },
    "89": { "F": 0.0002, "H": 0.0001, "I": 0.0001, "L": 0.9822, "M": 0.0037, "P": 0.0018, "Q": 0.0008, "R": 0.0097, "V": 0.0008, "W": 0.0001 },
    "90": { "D": 0.0003, "E": 0.0043, "H": 0.0517, "K": 0.0034, "L": 0.0069, "N": 0.0009, "P": 0.0015, "Q": 0.914, "R": 0.0141, "S": 0.0002, "V": 0.0007, "W": 0.0005, "Y": 0.0012 },
    "91": { "C": 0.0001, "F": 0.0006, "I": 0.0004, "L": 0.9756, "M": 0.007, "P": 0.0016, "Q": 0.0003, "R": 0.0051, "V": 0.0088, "W": 0.0002 },
    "92": { "C": 0.0001, "D": 0.0186, "E": 0.001, "G": 0.0017, "H": 0.0069, "I": 0.0022, "K": 0.0248, "L": 0.0003, "M": 0.0007, "N": 0.8969, "Q": 0.0006, "R": 0.0043, "S": 0.0221, "T": 0.0159, "V": 0.0002, "Y": 0.0033 },
    "93": { "A": 0.0129, "C": 0.0008, "D": 0.0001, "F": 0.0054, "G": 0.0002, "H": 0.0009, "L": 0.0005, "P": 0.0022, "Q": 0.0004, "S": 0.9712, "T": 0.0015, "V": 0.0003, "W": 0.0001, "Y": 0.0032 },
    "94": { "A": 0.0044, "E": 0.0006, "G": 0.0085, "L": 0.0087, "M": 0.0076, "R": 0.0002, "S": 0.0002, "V": 0.9692, "W": 0.0003 },
    "95": { "A": 0.0054, "D": 0.0001, "G": 0.0003, "H": 0.0001, "I": 0.0065, "K": 0.0002, "M": 0.0003, "N": 0.0033, "P": 0.0028, "R": 0.0002, "S": 0.0105, "T": 0.9685, "V": 0.0016 },
    "96": { "A": 0.006, "D": 0.0001, "F": 0.0004, "H": 0.0025, "L": 0.0033, "P": 0.9767, "Q": 0.0001, "R": 0.0006, "S": 0.0046, "T": 0.0051, "V": 0.0003 },
    "97": { "A": 0.0042, "D": 0.0458, "E": 0.9412, "G": 0.0054, "K": 0.0009, "N": 0.0002, "P": 0.0001, "Q": 0.001, "R": 0.0001, "T": 0.0001, "V": 0.0007, "W": 0.0001 },
    "98": { "A": 0.0013, "D": 0.9837, "E": 0.0093, "G": 0.0015, "H": 0.0003, "N": 0.0009, "S": 0.0001, "T": 0.0001, "V": 0.0007, "Y": 0.002 },
    "99": { "A": 0.0078, "G": 0.0001, "K": 0.0019, "M": 0.0019, "N": 0.0001, "P": 0.0011, "Q": 0.0001, "R": 0.0006, "S": 0.01, "T": 0.9762 },
    "100": { "A": 0.9829, "D": 0.002, "G": 0.0095, "P": 0.0001, "S": 0.0018, "T": 0.0018, "V": 0.0018 },
    "101": { "A": 0.0011, "D": 0.0001, "E": 0.0004, "F": 0.0004, "G": 0.0038, "I": 0.0178, "K": 0.0001, "L": 0.021, "M": 0.0118, "R": 0.0001, "T": 0.004, "V": 0.9389, "W": 0.0002 },
    "102": { "C": 0.0004, "D": 0.0037, "E": 0.0002, "F": 0.0012, "H": 0.0014, "I": 0.0001, "N": 0.0004, "S": 0.0007, "Y": 0.9915 },
    "103": { "C": 0.0016, "D": 0.0032, "F": 0.0313, "G": 0.0001, "H": 0.0007, "L": 0.0002, "N": 0.0003, "S": 0.0022, "T": 0.0001, "V": 0.0005, "W": 0.0004, "Y": 0.9594 },
    "104": { "C": 0.9874, "F": 0.0013, "G": 0.0045, "L": 0.0001, "R": 0.001, "S": 0.0005, "V": 0.0001, "W": 0.0038, "Y": 0.0012 },
    "105": { "A": 0.9242, "D": 0.0001, "E": 0.0011, "G": 0.004, "I": 0.0002, "L": 0.0004, "N": 0.0001, "P": 0.0003, "Q": 0.0004, "S": 0.0101, "T": 0.0212, "V": 0.0375, "W": 0.0001 },
    "106": { "A": 0.003, "C": 0.0006, "D": 0.0004, "E": 0.0005, "F": 0.0007, "G": 0.0157, "H": 0.0008, "I": 0.0023, "K": 0.0143, "L": 0.0017, "M": 0.0022, "N": 0.0011, "P": 0.0003, "Q": 0.0003, "R": 0.9137, "S": 0.0315, "T": 0.0062, "V": 0.0022, "W": 0.0006, "Y": 0.0021 },
    "107": { "A": 0.0891, "C": 0.0025, "D": 0.2676, "E": 0.1703, "F": 0.007, "G": 0.2353, "H": 0.0056, "I": 0.0075, "K": 0.0027, "L": 0.0128, "M": 0.0036, "N": 0.0113, "P": 0.0009, "Q": 0.0123, "R": 0.0127, "S": 0.0558, "T": 0.0234, "V": 0.0625, "W": 0.009, "Y": 0.008 },
    "108": { "A": 0.0612, "C": 0.0029, "D": 0.0372, "E": 0.0321, "F": 0.0183, "G": 0.1514, "H": 0.019, "I": 0.0286, "K": 0.0273, "L": 0.0786, "M": 0.0101, "N": 0.0283, "P": 0.081, "Q": 0.035, "R": 0.1116, "S": 0.0897, "T": 0.0522, "V": 0.064, "W": 0.0293, "Y": 0.0422 },
    "109": { "A": 0.0715, "C": 0.0114, "D": 0.0493, "E": 0.029, "F": 0.0251, "G": 0.1772, "H": 0.0203, "I": 0.0349, "K": 0.014, "L": 0.0563, "M": 0.0099, "N": 0.0261, "P": 0.0518, "Q": 0.0245, "R": 0.066, "S": 0.1131, "T": 0.053, "V": 0.0519, "W": 0.0361, "Y": 0.0786 },
    "110": { "A": 0.0799, "C": 0.0093, "D": 0.0573, "E": 0.0259, "F": 0.0152, "G": 0.1676, "H": 0.0126, "I": 0.0425, "K": 0.0109, "L": 0.0529, "M": 0.0114, "N": 0.0279, "P": 0.0396, "Q": 0.0164, "R": 0.0464, "S": 0.1368, "T": 0.0497, "V": 0.0642, "W": 0.0357, "Y": 0.0978 },
    "111": { "A": 0.0863, "C": 0.0154, "D": 0.0617, "E": 0.0224, "F": 0.0194, "G": 0.1546, "H": 0.0114, "I": 0.0395, "K": 0.0107, "L": 0.0418, "M": 0.0086, "N": 0.0243, "P": 0.0307, "Q": 0.0128, "R": 0.0417, "S": 0.1664, "T": 0.0498, "V": 0.0624, "W": 0.0401, "Y": 0.1004 },
    "111A": { "A": 0.0875, "C": 0.0114, "D": 0.0519, "E": 0.02, "F": 0.0358, "G": 0.1717, "H": 0.0079, "I": 0.0299, "K": 0.0083, "L": 0.0382, "M": 0.0093, "N": 0.0222, "P": 0.0253, "Q": 0.0146, "R": 0.0343, "S": 0.1794, "T": 0.0483, "V": 0.0575, "W": 0.0449, "Y": 0.1015 },
    "111B": { "A": 0.0863, "C": 0.0101, "D": 0.0597, "E": 0.0158, "F": 0.0273, "G": 0.1809, "H": 0.0055, "I": 0.0265, "K": 0.0103, "L": 0.0392, "M": 0.0082, "N": 0.0142, "P": 0.0462, "Q": 0.0119, "R": 0.038, "S": 0.1538, "T": 0.0595, "V": 0.0595, "W": 0.0691, "Y": 0.0781 },
    "111C": { "A": 0.0643, "C": 0.023, "D": 0.0477, "E": 0.0195, "F": 0.0234, "G": 0.1523, "H": 0.0058, "I": 0.04, "K": 0.0099, "L": 0.0397, "M": 0.0042, "N": 0.0211, "P": 0.0461, "Q": 0.0096, "R": 0.0339, "S": 0.1843, "T": 0.0672, "V": 0.0586, "W": 0.0666, "Y": 0.0829 },
    "111D": { "A": 0.0567, "C": 0.0374, "D": 0.048, "E": 0.0118, "F": 0.0218, "G": 0.1771, "H": 0.0069, "I": 0.0156, "K": 0.0062, "L": 0.0368, "M": 0.0131, "N": 0.0125, "P": 0.0711, "Q": 0.0094, "R": 0.0399, "S": 0.2014, "T": 0.0405, "V": 0.0892, "W": 0.0305, "Y": 0.0742 },
    "111E": { "A": 0.057, "C": 0.0266, "D": 0.0165, "E": 0.0329, "F": 0.0696, "G": 0.1241, "H": 0.0076, "I": 0.0418, "K": 0.0063, "L": 0.1152, "M": 0.0127, "N": 0.0089, "P": 0.1139, "Q": 0.0025, "R": 0.019, "S": 0.1304, "T": 0.0684, "V": 0.0367, "W": 0.0291, "Y": 0.081 },
    "111F": { "A": 0.265, "C": 0.0164, "D": 0.0082, "E": 0.0137, "F": 0.0137, "G": 0.1393, "H": 0.0109, "I": 0.0082, "K": 0.0055, "L": 0.0519, "M": 0.0027, "N": 0.0055, "P": 0.041, "Q": 0.0027, "R": 0.0109, "S": 0.2377, "T": 0.0546, "V": 0.041, "W": 0.0219, "Y": 0.0492 },
    "111G": { "A": 0.0442, "C": 0.0177, "D": 0.0354, "F": 0.0088, "G": 0.0619, "H": 0.0088, "K": 0.0177, "L": 0.0177, "N": 0.0354, "P": 0.0708, "Q": 0.0088, "R": 0.0354, "S": 0.1239, "T": 0.1239, "V": 0.0619, "W": 0.2566, "Y": 0.0708 },
    "112G": { "A": 0.067, "C": 0.0402, "D": 0.0045, "E": 0.0357, "G": 0.4509, "I": 0.0089, "K": 0.0223, "L": 0.0268, "M": 0.0089, "N": 0.0268, "P": 0.0402, "Q": 0.0045, "R": 0.0223, "S": 0.1205, "T": 0.0312, "V": 0.0223, "W": 0.0134, "Y": 0.0536 },
    "112F": { "A": 0.0617, "C": 0.0417, "D": 0.0181, "E": 0.0145, "F": 0.0181, "G": 0.1234, "H": 0.0127, "I": 0.0054, "K": 0.0073, "L": 0.1089, "M": 0.0036, "N": 0.0073, "P": 0.167, "Q": 0.0145, "R": 0.0309, "S": 0.1851, "T": 0.0417, "V": 0.02, "W": 0.0181, "Y": 0.0998 },
    "112E": { "A": 0.0431, "C": 0.0245, "D": 0.0363, "E": 0.0203, "F": 0.027, "G": 0.1133, "H": 0.0101, "I": 0.0237, "K": 0.0786, "L": 0.038, "M": 0.0034, "N": 0.0144, "P": 0.0642, "Q": 0.011, "R": 0.0744, "S": 0.1369, "T": 0.0549, "V": 0.1065, "W": 0.0211, "Y": 0.0981 },
    "112D": { "A": 0.0519, "C": 0.0197, "D": 0.0376, "E": 0.0309, "F": 0.0175, "G": 0.1647, "H": 0.0358, "I": 0.026, "K": 0.0215, "L": 0.0407, "M": 0.0036, "N": 0.0237, "P": 0.0698, "Q": 0.0148, "R": 0.0595, "S": 0.12, "T": 0.0636, "V": 0.0425, "W": 0.0286, "Y": 0.1276 },
    "112C": { "A": 0.0532, "C": 0.0199, "D": 0.057, "E": 0.018, "F": 0.034, "G": 0.1394, "H": 0.0145, "I": 0.0245, "K": 0.0119, "L": 0.0361, "M": 0.009, "N": 0.0271, "P": 0.0601, "Q": 0.0095, "R": 0.048, "S": 0.1209, "T": 0.047, "V": 0.0489, "W": 0.0356, "Y": 0.1855 },
    "112B": { "A": 0.0581, "C": 0.016, "D": 0.0422, "E": 0.0182, "F": 0.021, "G": 0.1261, "H": 0.0208, "I": 0.0204, "K": 0.0127, "L": 0.0417, "M": 0.0065, "N": 0.0214, "P": 0.0433, "Q": 0.008, "R": 0.0404, "S": 0.098, "T": 0.0464, "V": 0.0472, "W": 0.0421, "Y": 0.2696 },
    "112A": { "A": 0.0575, "C": 0.0194, "D": 0.0418, "E": 0.0228, "F": 0.0193, "G": 0.1339, "H": 0.0116, "I": 0.0137, "K": 0.0127, "L": 0.0371, "M": 0.0103, "N": 0.0244, "P": 0.0439, "Q": 0.0111, "R": 0.0456, "S": 0.1026, "T": 0.0507, "V": 0.0444, "W": 0.0295, "Y": 0.2677 },
    "112": { "A": 0.051, "C": 0.0084, "D": 0.0392, "E": 0.0248, "F": 0.0273, "G": 0.1111, "H": 0.0233, "I": 0.0206, "K": 0.0117, "L": 0.0556, "M": 0.0058, "N": 0.0324, "P": 0.0473, "Q": 0.0159, "R": 0.0438, "S": 0.0966, "T": 0.0467, "V": 0.0413, "W": 0.0377, "Y": 0.2595 },
    "113": { "A": 0.0425, "C": 0.0054, "D": 0.0914, "E": 0.0167, "F": 0.0194, "G": 0.1562, "H": 0.0198, "I": 0.0071, "K": 0.0141, "L": 0.0278, "M": 0.0038, "N": 0.0776, "P": 0.0162, "Q": 0.0096, "R": 0.0459, "S": 0.087, "T": 0.0377, "V": 0.0175, "W": 0.0602, "Y": 0.2441 },
    "114": { "A": 0.1619, "C": 0.0067, "D": 0.018, "E": 0.0135, "F": 0.0198, "G": 0.2317, "H": 0.0204, "I": 0.013, "K": 0.0072, "L": 0.0242, "M": 0.0021, "N": 0.0174, "P": 0.039, "Q": 0.0055, "R": 0.0213, "S": 0.0451, "T": 0.0314, "V": 0.0227, "W": 0.1125, "Y": 0.1865 },
    "115": { "A": 0.0055, "C": 0.0051, "D": 0.0011, "E": 0.0017, "F": 0.595, "G": 0.0083, "H": 0.004, "I": 0.0293, "K": 0.0008, "L": 0.0972, "M": 0.1748, "N": 0.0026, "P": 0.0087, "Q": 0.0005, "R": 0.0011, "S": 0.0052, "T": 0.0057, "V": 0.0227, "W": 0.0071, "Y": 0.0237 },
    "116": { "A": 0.0057, "C": 0.0002, "D": 0.9023, "E": 0.0124, "F": 0.0014, "G": 0.0173, "H": 0.0045, "I": 0.0027, "K": 0.0014, "L": 0.0043, "M": 0.0001, "N": 0.008, "P": 0.0021, "Q": 0.0071, "R": 0.0048, "S": 0.0055, "T": 0.0067, "V": 0.0045, "W": 0.006, "Y": 0.0029 },
    "117": { "A": 0.0044, "C": 0.0064, "D": 0.0085, "E": 0.002, "F": 0.0235, "G": 0.0023, "H": 0.0134, "I": 0.1256, "K": 0.0004, "L": 0.0432, "M": 0.0015, "N": 0.008, "P": 0.1221, "Q": 0.0008, "R": 0.0019, "S": 0.0378, "T": 0.0086, "V": 0.1985, "W": 0.0013, "Y": 0.3899 },
    "118": { "A": 0.0001, "C": 0.0009, "F": 0.0003, "G": 0.0019, "K": 0.0001, "L": 0.0015, "M": 0.0002, "R": 0.0011, "V": 0.0001, "W": 0.9937 },
    "119": { "A": 0.0044, "C": 0.0019, "D": 0.0048, "F": 0.0001, "G": 0.984, "P": 0.0002, "R": 0.0001, "S": 0.0031, "T": 0.0001, "V": 0.0011 },
    "120": { "E": 0.0022, "G": 0.0002, "H": 0.004, "K": 0.0287, "L": 0.005, "N": 0.0001, "P": 0.0057, "Q": 0.9069, "R": 0.0466, "S": 0.0001, "T": 0.0001, "V": 0.0001 },
    "121": { "A": 0.0001, "D": 0.0001, "E": 0.0006, "G": 0.9967, "R": 0.0007, "V": 0.0014, "W": 0.0002 },
    "122": { "A": 0.0113, "G": 0.0001, "H": 0.0011, "I": 0.0241, "K": 0.0006, "L": 0.0003, "M": 0.0002, "N": 0.002, "P": 0.0011, "R": 0.0003, "S": 0.0102, "T": 0.9458, "V": 0.0028 },
    "123": { "A": 0.0011, "E": 0.0002, "H": 0.0004, "I": 0.0013, "K": 0.001, "L": 0.6485, "M": 0.1291, "P": 0.0094, "Q": 0.0097, "R": 0.004, "S": 0.003, "T": 0.1864, "V": 0.0057 },
    "124": { "A": 0.0008, "D": 0.0002, "F": 0.0009, "G": 0.0028, "I": 0.0027, "L": 0.0004, "V": 0.9922 },
    "125": { "A": 0.0037, "F": 0.0002, "H": 0.0001, "I": 0.0115, "L": 0.0004, "M": 0.0001, "N": 0.0034, "P": 0.0007, "R": 0.0003, "S": 0.0052, "T": 0.9738, "V": 0.0003 },
    "126": { "A": 0.0015, "D": 0.001, "F": 0.0005, "G": 0.0015, "I": 0.0012, "L": 0.0002, "R": 0.0003, "V": 0.9935 },
    "127": { "A": 0.0029, "C": 0.0004, "D": 0.0001, "F": 0.0009, "L": 0.0004, "P": 0.0025, "S": 0.9895, "T": 0.0015, "Y": 0.0018 },
    "128": { "A": 0.0028, "L": 0.0003, "P": 0.0069, "Q": 0.0001, "S": 0.9875, "T": 0.0022 }
  },
  "IGHV7": {
    "1": { "H": 0.0057, "Q": 0.9943 },
    "2": { "A": 0.0133, "D": 0.0012, "E": 0.0045, "F": 0.0009, "G": 0.0076, "I": 0.0004, "K": 0.0003, "L": 0.0068, "M": 0.0043, "P": 0.0003, "Q": 0.0002, "S": 0.0003, "V": 0.9595, "W": 0.0005 },
    "3": { "A": 0.0002, "D": 0.0003, "E": 0.0023, "H": 0.0144, "K": 0.0024, "L": 0.0117, "N": 0.0004, "P": 0.0088, "Q": 0.9534, "R": 0.0053, "T": 0.0003, "V": 0.0002, "W": 0.0003, "Y": 0.0002 },
    "4": { "A": 0.0002, "C": 0.0001, "E": 0.0001, "F": 0.0002, "H": 0.0001, "I": 0.0001, "L": 0.9809, "M": 0.0025, "P": 0.0052, "Q": 0.0011, "R": 0.0038, "S": 0.0004, "T": 0.0002, "V": 0.0044, "W": 0.0008 },
    "5": { "A": 0.0064, "D": 0.0015, "E": 0.0069, "F": 0.0004, "G": 0.0053, "H": 0.0001, "I": 0.0005, "K": 0.0001, "L": 0.0082, "M": 0.0038, "P": 0.0003, "Q": 0.0022, "R": 0.0003, "S": 0.0004, "T": 0.0003, "V": 0.9634, "W": 0.0001 },
    "6": { "A": 0.0001, "D": 0.0002, "E": 0.0072, "H": 0.0222, "I": 0.0001, "K": 0.0028, "L": 0.002, "N": 0.0002, "P": 0.0088, "Q": 0.9537, "R": 0.0022, "S": 0.0002, "T": 0.0003, "Y": 0.0001 },
    "7": { "A": 0.006, "C": 0.0006, "E": 0.0001, "F": 0.0037, "H": 0.0001, "L": 0.0006, "P": 0.0136, "S": 0.9704, "T": 0.0026, "V": 0.0001, "W": 0.0001, "Y": 0.0021 },
    "8": { "A": 0.0047, "D": 0.0004, "E": 0.0169, "G": 0.9657, "L": 0.0002, "P": 0.0017, "Q": 0.0002, "R": 0.0025, "S": 0.0002, "V": 0.0047, "W": 0.0029 },
    "9": { "A": 0.0437, "C": 0.0006, "D": 0.0003, "F": 0.003, "G": 0.0014, "H": 0.0018, "I": 0.0001, "L": 0.0006, "N": 0.0002, "P": 0.0105, "Q": 0.0001, "S": 0.9292, "T": 0.0034, "V": 0.0009, "W": 0.0002, "Y": 0.004 },
    "11": { "A": 0.0187, "D": 0.008, "E": 0.9569, "G": 0.0085, "K": 0.0012, "P": 0.0001, "Q": 0.0039, "R": 0.0002, "S": 0.0005, "T": 0.0002, "V": 0.0016, "W": 0.0004 },
    "12": { "A": 0.0004, "C": 0.0005, "E": 0.0004, "F": 0.0117, "G": 0.0006, "H": 0.0001, "I": 0.0007, "L": 0.9209, "M": 0.0031, "Q": 0.0001, "S": 0.0038, "V": 0.0527, "W": 0.0049, "Y": 0.0001 },
    "13": { "A": 0.0009, "D": 0.0004, "E": 0.0115, "G": 0.0004, "H": 0.0004, "K": 0.9194, "L": 0.0001, "M": 0.0036, "N": 0.0046, "P": 0.0001, "Q": 0.013, "R": 0.0255, "S": 0.0005, "T": 0.019, "V": 0.0006, "W": 0.0002 },
    "14": { "A": 0.0004, "C": 0.0001, "D": 0.0018, "E": 0.015, "G": 0.0004, "H": 0.0018, "I": 0.0006, "K": 0.9135, "L": 0.0001, "M": 0.0039, "N": 0.0145, "P": 0.0014, "Q": 0.0125, "R": 0.0147, "S": 0.002, "T": 0.0172, "V": 0.0002, "Y": 0.0001 },
    "15": { "A": 0.0029, "C": 0.0001, "H": 0.0056, "L": 0.0027, "P": 0.976, "Q": 0.0003, "R": 0.0017, "S": 0.0053, "T": 0.0055 },
    "16": { "A": 0.0033, "C": 0.0002, "D": 0.0002, "E": 0.0029, "G": 0.9842, "P": 0.0001, "Q": 0.0001, "R": 0.0016, "S": 0.0002, "T": 0.0001, "V": 0.0037, "W": 0.0035 },
    "17": { "A": 0.9654, "C": 0.0002, "D": 0.0068, "E": 0.0004, "G": 0.0014, "H": 0.0002, "L": 0.0002, "N": 0.0001, "P": 0.004, "R": 0.0002, "S": 0.0135, "T": 0.006, "V": 0.0018 },
    "18": { "A": 0.0132, "E": 0.0001, "H": 0.0005, "I": 0.0003, "K": 0.0001, "L": 0.0018, "P": 0.0041, "Q": 0.0005, "S": 0.9776, "T": 0.001, "V": 0.0002, "Y": 0.0007 },
    "19": { "A": 0.0047, "C": 0.0002, "D": 0.0001, "E": 0.0011, "F": 0.0002, "G": 0.0067, "I": 0.0002, "L": 0.0053, "M": 0.0012, "P": 0.0001, "R": 0.0003, "S": 0.0002, "V": 0.9795, "W": 0.0004 },
    "20": { "A": 0.0001, "C": 0.0001, "D": 0.0002, "E": 0.0026, "G": 0.0006, "H": 0.0006, "I": 0.0017, "K": 0.8745, "L": 0.0001, "M": 0.0051, "N": 0.0131, "P": 0.0005, "Q": 0.0196, "R": 0.0301, "S": 0.0012, "T": 0.0479, "V": 0.0017, "W": 0.0004 },
    "21": { "A": 0.0068, "C": 0.0004, "D": 0.0017, "E": 0.0001, "F": 0.0041, "G": 0.0073, "H": 0.0002, "I": 0.0514, "L": 0.023, "M": 0.0002, "N": 0.0001, "P": 0.0002, "R": 0.0006, "S": 0.0003, "T": 0.0005, "V": 0.9031, "Y": 0.0001 },
    "22": { "A": 0.0076, "C": 0.0011, "D": 0.0002, "F": 0.0021, "H": 0.0002, "K": 0.0001, "L": 0.0002, "P": 0.0041, "S": 0.976, "T": 0.0026, "Y": 0.0058 },
    "23": { "A": 0.0003, "C": 0.9824, "E": 0.0001, "F": 0.0044, "G": 0.0047, "H": 0.0001, "I": 0.0001, "L": 0.0005, "R": 0.0021, "S": 0.0025, "V": 0.0007, "W": 0.0006, "Y": 0.0016 },
    "24": { "A": 0.0004, "E": 0.0087, "G": 0.0001, "H": 0.0007, "I": 0.0001, "K": 0.9114, "L": 0.0006, "M": 0.0038, "N": 0.0063, "P": 0.0004, "Q": 0.0251, "R": 0.0148, "S": 0.0005, "T": 0.0266, "V": 0.0001, "W": 0.0002, "Y": 0.0003 },
    "25": { "A": 0.9324, "D": 0.0051, "E": 0.0002, "F": 0.0001, "G": 0.0053, "H": 0.0003, "I": 0.0004, "L": 0.0002, "M": 0.0001, "N": 0.0001, "P": 0.0069, "Q": 0.0001, "S": 0.0108, "T": 0.0323, "V": 0.0058, "Y": 0.0001 },
    "26": { "A": 0.0084, "C": 0.0007, "D": 0.0002, "F": 0.0046, "H": 0.0003, "I": 0.0001, "K": 0.0001, "L": 0.0006, "N": 0.0001, "P": 0.0065, "S": 0.97, "T": 0.0014, "V": 0.0003, "W": 0.0001, "Y": 0.0067 },
    "27": { "A": 0.0053, "C": 0.0005, "D": 0.0007, "E": 0.0041, "G": 0.9781, "K": 0.0003, "P": 0.0003, "Q": 0.0001, "R": 0.005, "S": 0.0001, "T": 0.0001, "V": 0.0053, "W": 0.0002 },
    "28": { "A": 0.0002, "C": 0.0016, "D": 0.0076, "E": 0.0005, "F": 0.0058, "G": 0.0059, "H": 0.0034, "K": 0.0003, "L": 0.0003, "N": 0.0035, "P": 0.0006, "Q": 0.0001, "S": 0.0215, "T": 0.0003, "Y": 0.9485 },
    "29": { "A": 0.0075, "C": 0.0001, "D": 0.0025, "F": 0.0011, "G": 0.0018, "H": 0.0014, "I": 0.024, "K": 0.0008, "L": 0.001, "M": 0.0012, "N": 0.0216, "P": 0.022, "R": 0.0027, "S": 0.0486, "T": 0.8627, "V": 0.0002, "Y": 0.001 },
    "30": { "A": 0.0002, "C": 0.0041, "D": 0.0004, "F": 0.9465, "I": 0.0055, "L": 0.0307, "P": 0.0001, "R": 0.0002, "S": 0.0052, "T": 0.0006, "V": 0.0049, "Y": 0.0018 },
    "35": { "A": 0.013, "C": 0.0002, "D": 0.0014, "E": 0.0002, "F": 0.0008, "G": 0.0019, "H": 0.0026, "I": 0.0162, "K": 0.0042, "L": 0.0019, "M": 0.001, "N": 0.0245, "P": 0.0222, "R": 0.0075, "S": 0.0538, "T": 0.8469, "V": 0.0013, "Y": 0.0003 },
    "36": { "A": 0.0075, "C": 0.001, "D": 0.0308, "E": 0.0029, "F": 0.0015, "G": 0.0125, "H": 0.0053, "I": 0.0042, "K": 0.0138, "L": 0.0009, "M": 0.0008, "N": 0.0704, "P": 0.0019, "Q": 0.0017, "R": 0.0351, "S": 0.7296, "T": 0.0759, "V": 0.0005, "W": 0.0004, "Y": 0.0032 },
    "37": { "A": 0.0006, "C": 0.0024, "D": 0.0056, "E": 0.0006, "F": 0.0259, "H": 0.0269, "I": 0.001, "K": 0.0011, "L": 0.0044, "M": 0.0001, "N": 0.0137, "P": 0.0007, "Q": 0.0016, "R": 0.0009, "S": 0.0253, "T": 0.0007, "V": 0.0005, "W": 0.0011, "Y": 0.8867 },
    "38": { "A": 0.8348, "D": 0.0062, "E": 0.0006, "F": 0.0006, "G": 0.0484, "H": 0.0012, "I": 0.0006, "K": 0.0002, "L": 0.0011, "M": 0.0001, "N": 0.0042, "P": 0.0291, "Q": 0.0002, "R": 0.0002, "S": 0.0293, "T": 0.0213, "V": 0.0166, "W": 0.0004, "Y": 0.005 },
    "39": { "A": 0.0003, "C": 0.0002, "E": 0.0001, "F": 0.0028, "G": 0.0003, "H": 0.0001, "I": 0.1286, "K": 0.0008, "L": 0.0593, "M": 0.7741, "N": 0.0003, "P": 0.0001, "Q": 0.0001, "R": 0.0072, "S": 0.0005, "T": 0.0035, "V": 0.0217, "W": 0.0001 },
    "40": { "A": 0.0008, "C": 0.0005, "D": 0.0038, "E": 0.0004, "G": 0.0005, "H": 0.0424, "I": 0.0073, "K": 0.0083, "L": 0.0005, "M": 0.0002, "N": 0.876, "P": 0.0008, "Q": 0.0012, "R": 0.0006, "S": 0.0257, "T": 0.0277, "V": 0.0012, "W": 0.0001, "Y": 0.0021 },
    "41": { "A": 0.0002, "C": 0.0075, "F": 0.0002, "G": 0.0067, "L": 0.0065, "P": 0.0001, "Q": 0.0003, "R": 0.0042, "S": 0.0024, "V": 0.0005, "W": 0.9714 },
    "42": { "A": 0.0032, "D": 0.0001, "E": 0.001, "F": 0.0008, "G": 0.0104, "I": 0.001, "L": 0.0259, "M": 0.0054, "R": 0.0003, "S": 0.0001, "T": 0.0001, "V": 0.9505, "W": 0.0011 },
    "43": { "A": 0.0002, "C": 0.0003, "G": 0.0026, "H": 0.0002, "I": 0.0001, "K": 0.0001, "L": 0.0036, "P": 0.0038, "Q": 0.0029, "R": 0.9856, "S": 0.0003, "T": 0.0001, "V": 0.0001, "W": 0.0003 },
    "44": { "A": 0.0001, "D": 0.0001, "E": 0.0017, "H": 0.0051, "K": 0.005, "L": 0.0026, "N": 0.0001, "P": 0.0124, "Q": 0.9672, "R": 0.0053, "S": 0.0001, "T": 0.0005 },
    "45": { "A": 0.9736, "C": 0.0003, "D": 0.0057, "E": 0.0001, "G": 0.0019, "H": 0.0001, "I": 0.0001, "M": 0.0001, "P": 0.0031, "S": 0.0056, "T": 0.0033, "V": 0.0059, "Y": 0.0001 },
    "46": { "A": 0.0024, "H": 0.0063, "L": 0.0018, "P": 0.977, "Q": 0.0004, "R": 0.0009, "S": 0.0033, "T": 0.0079, "V": 0.0001 },
    "47": { "A": 0.0025, "C": 0.0002, "D": 0.0002, "E": 0.0018, "G": 0.9886, "R": 0.0028, "S": 0.0001, "T": 0.0002, "V": 0.0035, "W": 0.0002 },
    "48": { "D": 0.0003, "E": 0.0035, "G": 0.0004, "H": 0.0173, "K": 0.0078, "L": 0.0008, "N": 0.0002, "P": 0.0064, "Q": 0.9512, "R": 0.0118, "S": 0.0001, "T": 0.0002, "V": 0.0001 },
    "49": { "A": 0.0011, "C": 0.0002, "D": 0.0007, "E": 0.001, "G": 0.9804, "P": 0.0007, "Q": 0.0001, "R": 0.0079, "S": 0.0003, "V": 0.004, "W": 0.0036 },
    "50": { "A": 0.0005, "F": 0.0087, "G": 0.0001, "H": 0.0006, "I": 0.0043, "L": 0.9595, "P": 0.0166, "R": 0.0069, "S": 0.0003, "T": 0.0001, "V": 0.0024 },
    "51": { "A": 0.01, "D": 0.0055, "E": 0.9622, "G": 0.0087, "H": 0.0008, "K": 0.0011, "L": 0.0001, "M": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0084, "R": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.0021, "W": 0.0005 },
    "52": { "C": 0.004, "E": 0.0003, "F": 0.0006, "G": 0.0094, "H": 0.0002, "L": 0.0068, "P": 0.0001, "R": 0.0042, "S": 0.001, "V": 0.0009, "W": 0.9698, "Y": 0.0027 },
    "53": { "A": 0.0001, "D": 0.0001, "E": 0.0001, "F": 0.0001, "G": 0.0004, "I": 0.0111, "K": 0.0014, "L": 0.0275, "M": 0.9301, "N": 0.0001, "P": 0.0001, "Q": 0.0002, "R": 0.0096, "S": 0.0004, "T": 0.0018, "V": 0.017, "W": 0.0002 },
    "54": { "A": 0.004, "C": 0.0002, "D": 0.0004, "E": 0.0011, "G": 0.9883, "R": 0.0023, "S": 0.0002, "V": 0.0035, "W": 0.0001 },
    "55": { "A": 0.0002, "C": 0.0061, "D": 0.0001, "E": 0.0001, "F": 0.0013, "G": 0.0128, "L": 0.0065, "M": 0.0004, "N": 0.0009, "R": 0.0099, "S": 0.0032, "T": 0.0003, "V": 0.0009, "W": 0.954, "Y": 0.0034 },
    "56": { "A": 0.0005, "C": 0.0004, "D": 0.0001, "F": 0.0062, "G": 0.0008, "H": 0.0003, "I": 0.9375, "K": 0.0001, "L": 0.0096, "M": 0.0054, "N": 0.0042, "P": 0.0003, "R": 0.0007, "S": 0.0141, "T": 0.0047, "V": 0.0147, "Y": 0.0004 },
    "57": { "A": 0.0041, "D": 0.0236, "E": 0.0002, "F": 0.0003, "G": 0.0006, "H": 0.0123, "I": 0.0024, "K": 0.0124, "N": 0.9125, "P": 0.0005, "Q": 0.0011, "R": 0.0008, "S": 0.0143, "T": 0.0121, "V": 0.0006, "Y": 0.0023 },
    "58": { "A": 0.0086, "D": 0.0002, "H": 0.0014, "I": 0.0089, "K": 0.0005, "L": 0.0006, "M": 0.0006, "N": 0.0059, "P": 0.0305, "R": 0.0004, "S": 0.0083, "T": 0.9335, "V": 0.0005, "Y": 0.0002 },
    "59": { "A": 0.0026, "D": 0.0287, "E": 0.0027, "F": 0.0009, "G": 0.0046, "H": 0.0137, "I": 0.0116, "K": 0.0382, "L": 0.0005, "M": 0.0008, "N": 0.847, "P": 0.0015, "Q": 0.0041, "R": 0.0062, "S": 0.0142, "T": 0.0115, "V": 0.0008, "Y": 0.0104 },
    "62": { "A": 0.0086, "C": 0.0001, "D": 0.0003, "E": 0.0002, "F": 0.0006, "G": 0.0004, "H": 0.0006, "I": 0.0035, "K": 0.0002, "L": 0.0008, "M": 0.0006, "N": 0.0096, "P": 0.0255, "Q": 0.0001, "R": 0.0006, "S": 0.0163, "T": 0.9315, "V": 0.0005, "W": 0.0001 },
    "63": { "A": 0.007, "C": 0.0008, "D": 0.0014, "E": 0.01, "F": 0.0001, "G": 0.9581, "H": 0.0003, "K": 0.0002, "L": 0.0005, "N": 0.0002, "P": 0.0002, "Q": 0.0004, "R": 0.0056, "S": 0.0004, "T": 0.0004, "V": 0.0069, "W": 0.0076 },
    "64": { "A": 0.001, "C": 0.0001, "D": 0.021, "E": 0.0026, "F": 0.0015, "G": 0.0012, "H": 0.012, "I": 0.0088, "K": 0.0335, "L": 0.0002, "M": 0.0001, "N": 0.8672, "P": 0.0001, "Q": 0.0007, "R": 0.0052, "S": 0.0205, "T": 0.0197, "V": 0.0023, "Y": 0.0023 },
    "65": { "A": 0.0079, "E": 0.0001, "G": 0.0003, "H": 0.0006, "K": 0.0001, "L": 0.0014, "N": 0.0001, "P": 0.9596, "Q": 0.0084, "R": 0.0053, "S": 0.0083, "T": 0.0078, "V": 0.0001 },
    "66": { "A": 0.0126, "D": 0.0003, "E": 0.0005, "F": 0.0006, "G": 0.0006, "H": 0.0001, "I": 0.003, "K": 0.0047, "L": 0.0012, "M": 0.0157, "N": 0.0008, "P": 0.0063, "Q": 0.0003, "R": 0.01, "S": 0.0096, "T": 0.9304, "V": 0.002, "W": 0.0002, "Y": 0.0013 },
    "67": { "A": 0.0003, "C": 0.0022, "D": 0.0103, "E": 0.0016, "F": 0.0118, "H": 0.0016, "I": 0.0003, "K": 0.0001, "L": 0.0009, "N": 0.0033, "P": 0.0001, "Q": 0.0004, "R": 0.0002, "S": 0.007, "V": 0.0015, "Y": 0.9585 },
    "68": { "A": 0.9528, "C": 0.0001, "D": 0.0098, "E": 0.0005, "F": 0.0001, "G": 0.0088, "I": 0.0001, "L": 0.001, "N": 0.0002, "P": 0.0011, "R": 0.0002, "S": 0.0057, "T": 0.0058, "V": 0.0138, "Y": 0.0001 },
    "69": { "A": 0.0001, "D": 0.0001, "E": 0.0042, "G": 0.0007, "H": 0.0089, "K": 0.0134, "L": 0.0022, "M": 0.0001, "N": 0.0002, "P": 0.0182, "Q": 0.9384, "R": 0.0125, "S": 0.0002, "T": 0.0004, "V": 0.0002, "W": 0.0002, "Y": 0.0001 },
    "70": { "A": 0.0261, "C": 0.0093, "D": 0.0388, "E": 0.0007, "G": 0.9089, "I": 0.0002, "K": 0.0003, "N": 0.0029, "Q": 0.0002, "R": 0.0018, "S": 0.0029, "T": 0.0002, "V": 0.0074, "Y": 0.0004 },
    "71": { "C": 0.0116, "D": 0.0001, "F": 0.9421, "G": 0.0003, "I": 0.0014, "L": 0.0268, "R": 0.0001, "S": 0.003, "T": 0.0002, "V": 0.0133, "Y": 0.0011 },
    "72": { "A": 0.0144, "E": 0.0006, "F": 0.0001, "G": 0.0014, "H": 0.0003, "I": 0.007, "K": 0.0071, "L": 0.0003, "M": 0.0004, "N": 0.0003, "P": 0.0052, "Q": 0.0006, "R": 0.0243, "S": 0.011, "T": 0.9269, "V": 0.0002 },
    "74": { "A": 0.0015, "C": 0.0003, "D": 0.0013, "E": 0.0029, "G": 0.9813, "K": 0.0001, "L": 0.0001, "P": 0.0001, "R": 0.0043, "T": 0.0002, "V": 0.0076, "W": 0.0005 },
    "75": { "A": 0.0002, "C": 0.0004, "G": 0.0021, "H": 0.0018, "K": 0.0001, "L": 0.0081, "M": 0.0002, "N": 0.0001, "P": 0.0012, "Q": 0.0032, "R": 0.9802, "S": 0.0005, "T": 0.0002, "W": 0.0018 },
    "76": { "A": 0.0001, "C": 0.0073, "F": 0.9436, "G": 0.0002, "H": 0.0003, "I": 0.005, "L": 0.0217, "S": 0.0022, "T": 0.0002, "V": 0.0147, "W": 0.0002, "Y": 0.0045 },
    "77": { "A": 0.0102, "C": 0.0012, "D": 0.0033, "E": 0.0002, "F": 0.007, "G": 0.0103, "H": 0.0002, "I": 0.0041, "L": 0.0022, "M": 0.0002, "N": 0.0001, "P": 0.0001, "R": 0.0002, "S": 0.0007, "T": 0.0002, "V": 0.9595, "W": 0.0001, "Y": 0.0002 },
    "78": { "C": 0.0101, "F": 0.9486, "G": 0.0002, "I": 0.0036, "L": 0.0203, "M": 0.0002, "P": 0.0002, "S": 0.0043, "V": 0.0113, "W": 0.0001, "Y": 0.0011 },
    "79": { "A": 0.0148, "C": 0.0005, "D": 0.0001, "F": 0.0014, "H": 0.0002, "L": 0.001, "P": 0.0047, "S": 0.9704, "T": 0.0035, "V": 0.0003, "W": 0.0004, "Y": 0.0028 },
    "80": { "A": 0.0002, "C": 0.0005, "F": 0.0121, "G": 0.0004, "I": 0.0005, "L": 0.9382, "M": 0.0046, "Q": 0.0003, "R": 0.0007, "S": 0.0087, "T": 0.0003, "V": 0.0267, "W": 0.007 },
    "81": { "A": 0.0057, "C": 0.0002, "D": 0.9726, "E": 0.0075, "F": 0.0001, "G": 0.0038, "H": 0.0007, "N": 0.0029, "Q": 0.0006, "R": 0.0005, "S": 0.0002, "T": 0.0001, "V": 0.0014, "Y": 0.0038 },
    "82": { "A": 0.0096, "D": 0.0002, "F": 0.0001, "G": 0.0001, "H": 0.0004, "I": 0.0064, "K": 0.0013, "L": 0.0002, "M": 0.0008, "N": 0.0042, "P": 0.0094, "R": 0.0008, "S": 0.0119, "T": 0.9545, "V": 0.0002 },
    "83": { "A": 0.016, "C": 0.0011, "F": 0.0031, "G": 0.0001, "H": 0.0001, "L": 0.001, "P": 0.0085, "Q": 0.0001, "S": 0.9624, "T": 0.0031, "V": 0.0006, "W": 0.0002, "Y": 0.0037 },
    "84": { "A": 0.0252, "C": 0.0006, "D": 0.0035, "E": 0.0005, "F": 0.0044, "G": 0.0123, "H": 0.0002, "I": 0.0061, "L": 0.0073, "M": 0.0005, "P": 0.0002, "R": 0.0003, "S": 0.0018, "T": 0.0006, "V": 0.9367 },
    "85": { "A": 0.003, "C": 0.0006, "D": 0.003, "F": 0.0005, "G": 0.0071, "H": 0.0001, "I": 0.0023, "K": 0.0015, "L": 0.0002, "N": 0.0467, "P": 0.0004, "R": 0.0204, "S": 0.8761, "T": 0.0375, "Y": 0.0007 },
    "86": { "A": 0.0039, "E": 0.0016, "G": 0.0001, "I": 0.0002, "K": 0.0027, "L": 0.0001, "M": 0.0024, "N": 0.0001, "P": 0.0033, "Q": 0.0003, "R": 0.0012, "S": 0.0029, "T": 0.9812, "V": 0.0002 },
    "87": { "A": 0.953, "D": 0.0005, "E": 0.0046, "G": 0.0018, "H": 0.0003, "I": 0.0002, "N": 0.0001, "P": 0.0015, "S": 0.0101, "T": 0.0234, "V": 0.0046 },
    "88": { "C": 0.0011, "D": 0.0068, "E": 0.0005, "F": 0.0197, "G": 0.0001, "H": 0.0102, "I": 0.0003, "K": 0.0002, "L": 0.0021, "N": 0.002, "P": 0.0001, "R": 0.0003, "S": 0.0041, "T": 0.0001, "V": 0.0006, "Y": 0.9517 },
    "89": { "C": 0.0001, "F": 0.0001, "I": 0.0004, "L": 0.9798, "M": 0.0053, "P": 0.0019, "Q": 0.0006, "R": 0.0055, "S": 0.0003, "T": 0.0001, "V": 0.0057, "W": 0.0001 },
    "90": { "A": 0.0003, "D": 0.0018, "E": 0.0289, "G": 0.0004, "H": 0.0151, "K": 0.0068, "L": 0.0019, "N": 0.0004, "P": 0.0013, "Q": 0.9359, "R": 0.0053, "T": 0.0001, "V": 0.0006, "W": 0.0003, "Y": 0.0008 },
    "91": { "C": 0.0006, "F": 0.0032, "I": 0.9715, "K": 0.0001, "L": 0.0026, "M": 0.0042, "N": 0.0012, "R": 0.0001, "S": 0.0095, "T": 0.0022, "V": 0.0046, "W": 0.0001 },
    "92": { "A": 0.0015, "C": 0.2254, "D": 0.004, "E": 0.0002, "F": 0.0016, "G": 0.0045, "H": 0.0015, "I": 0.0041, "K": 0.0025, "L": 0.0013, "M": 0.0005, "N": 0.0474, "P": 0.0009, "R": 0.0118, "S": 0.6568, "T": 0.025, "V": 0.0009, "W": 0.0001, "Y": 0.0101 },
    "93": { "A": 0.0025, "C": 0.0008, "D": 0.0082, "G": 0.0369, "H": 0.0002, "I": 0.0018, "K": 0.0003, "L": 0.0001, "N": 0.0252, "P": 0.0003, "R": 0.0094, "S": 0.9067, "T": 0.007, "V": 0.0001, "Y": 0.0005 },
    "94": { "I": 0.004, "L": 0.984, "M": 0.0004, "P": 0.0012, "Q": 0.0007, "R": 0.0072, "V": 0.0026 },
    "95": { "A": 0.0002, "D": 0.0004, "E": 0.026, "G": 0.0006, "H": 0.0003, "I": 0.0001, "K": 0.8985, "L": 0.0003, "M": 0.0049, "N": 0.0051, "P": 0.0002, "Q": 0.023, "R": 0.0277, "S": 0.0006, "T": 0.0117, "V": 0.0003, "W": 0.0001, "Y": 0.0001 },
    "96": { "A": 0.8881, "C": 0.0001, "D": 0.0097, "E": 0.0009, "F": 0.0003, "G": 0.0018, "I": 0.0014, "L": 0.0005, "M": 0.0001, "N": 0.0001, "P": 0.0333, "R": 0.0004, "S": 0.0212, "T": 0.0342, "V": 0.0077, "Y": 0.0003 },
    "97": { "A": 0.0056, "D": 0.0401, "E": 0.9409, "G": 0.007, "K": 0.0011, "N": 0.0029, "Q": 0.0007, "R": 0.0002, "S": 0.0004, "V": 0.0007, "W": 0.0002, "Y": 0.0002 },
    "98": { "A": 0.0017, "C": 0.0001, "D": 0.9847, "E": 0.0053, "G": 0.0028, "H": 0.0005, "N": 0.0009, "T": 0.0002, "V": 0.0014, "Y": 0.0025 },
    "99": { "A": 0.008, "G": 0.0001, "H": 0.0001, "I": 0.0017, "K": 0.0002, "L": 0.0002, "M": 0.0017, "N": 0.0022, "P": 0.0016, "R": 0.0001, "S": 0.0215, "T": 0.9621, "V": 0.0005, "Y": 0.0002 },
    "100": { "A": 0.9772, "C": 0.0002, "D": 0.0021, "E": 0.0002, "G": 0.0136, "P": 0.0006, "S": 0.0026, "T": 0.001, "V": 0.0024 },
    "101": { "A": 0.0007, "C": 0.0003, "E": 0.0006, "F": 0.0029, "G": 0.0059, "I": 0.0312, "K": 0.0006, "L": 0.0151, "M": 0.026, "P": 0.0001, "Q": 0.0001, "R": 0.0006, "S": 0.0001, "T": 0.0079, "V": 0.9076, "W": 0.0002, "Y": 0.0001 },
    "102": { "C": 0.0012, "D": 0.0069, "E": 0.0005, "F": 0.0038, "H": 0.0011, "I": 0.0001, "L": 0.0003, "N": 0.0003, "Q": 0.0001, "R": 0.0001, "S": 0.0008, "W": 0.0001, "Y": 0.9848 },
    "103": { "C": 0.0026, "D": 0.0041, "F": 0.0327, "H": 0.0034, "I": 0.0001, "L": 0.0002, "N": 0.0004, "R": 0.0001, "S": 0.0028, "T": 0.0002, "V": 0.0002, "Y": 0.9534 },
    "104": { "A": 0.0002, "C": 0.983, "D": 0.0001, "F": 0.0009, "G": 0.0044, "L": 0.0006, "M": 0.0002, "R": 0.001, "S": 0.0005, "T": 0.0002, "V": 0.0006, "W": 0.0079, "Y": 0.0005 },
    "105": { "A": 0.953, "D": 0.0001, "E": 0.0016, "F": 0.0001, "G": 0.0059, "I": 0.0002, "L": 0.0009, "M": 0.0007, "N": 0.0002, "P": 0.0003, "R": 0.0003, "S": 0.0048, "T": 0.0148, "V": 0.0162, "W": 0.0008, "Y": 0.0001 },
    "106": { "A": 0.0049, "C": 0.0004, "D": 0.0001, "E": 0.0028, "F": 0.0006, "G": 0.0116, "H": 0.0004, "I": 0.0091, "K": 0.0122, "L": 0.0111, "M": 0.0011, "N": 0.0008, "P": 0.0019, "Q": 0.0017, "R": 0.8751, "S": 0.0384, "T": 0.0225, "V": 0.0041, "W": 0.0004, "Y": 0.0008 },
    "107": { "A": 0.0756, "C": 0.0022, "D": 0.2784, "E": 0.1261, "F": 0.0065, "G": 0.2207, "H": 0.0069, "I": 0.0091, "K": 0.0095, "L": 0.0333, "M": 0.0039, "N": 0.0116, "P": 0.0199, "Q": 0.0099, "R": 0.0201, "S": 0.047, "T": 0.0211, "V": 0.081, "W": 0.0069, "Y": 0.0101 },
    "108": { "A": 0.0442, "C": 0.0051, "D": 0.0454, "E": 0.0338, "F": 0.029, "G": 0.1692, "H": 0.0204, "I": 0.0391, "K": 0.0284, "L": 0.0902, "M": 0.0136, "N": 0.0201, "P": 0.0862, "Q": 0.0268, "R": 0.1057, "S": 0.0824, "T": 0.0399, "V": 0.0482, "W": 0.0175, "Y": 0.055 },
    "109": { "A": 0.0642, "C": 0.0154, "D": 0.0576, "E": 0.0422, "F": 0.0214, "G": 0.172, "H": 0.0225, "I": 0.0283, "K": 0.0153, "L": 0.0505, "M": 0.0101, "N": 0.0194, "P": 0.0732, "Q": 0.0192, "R": 0.0873, "S": 0.1001, "T": 0.0441, "V": 0.0516, "W": 0.0235, "Y": 0.0819 },
    "110": { "A": 0.0836, "C": 0.0208, "D": 0.0721, "E": 0.0266, "F": 0.0282, "G": 0.1242, "H": 0.009, "I": 0.0426, "K": 0.0107, "L": 0.0637, "M": 0.0121, "N": 0.0211, "P": 0.0472, "Q": 0.0181, "R": 0.0509, "S": 0.1284, "T": 0.0501, "V": 0.0609, "W": 0.0278, "Y": 0.1017 },
    "111": { "A": 0.0972, "C": 0.0198, "D": 0.0635, "E": 0.0193, "F": 0.0291, "G": 0.1408, "H": 0.0161, "I": 0.0344, "K": 0.0115, "L": 0.0382, "M": 0.0122, "N": 0.0221, "P": 0.0292, "Q": 0.0182, "R": 0.045, "S": 0.1247, "T": 0.0505, "V": 0.0705, "W": 0.0355, "Y": 0.1221 },
    "111A": { "A": 0.0657, "C": 0.0221, "D": 0.072, "E": 0.0152, "F": 0.0314, "G": 0.1664, "H": 0.0124, "I": 0.0289, "K": 0.0068, "L": 0.0431, "M": 0.0112, "N": 0.017, "P": 0.031, "Q": 0.0151, "R": 0.0408, "S": 0.1657, "T": 0.0478, "V": 0.0669, "W": 0.0436, "Y": 0.0967 },
    "111B": { "A": 0.0637, "C": 0.0144, "D": 0.0679, "E": 0.0141, "F": 0.0373, "G": 0.1454, "H": 0.005, "I": 0.0298, "K": 0.0094, "L": 0.0407, "M": 0.0115, "N": 0.0128, "P": 0.0308, "Q": 0.0298, "R": 0.0256, "S": 0.1663, "T": 0.0767, "V": 0.0663, "W": 0.0611, "Y": 0.0916 },
    "111C": { "A": 0.062, "C": 0.0246, "D": 0.0444, "E": 0.0273, "F": 0.0541, "G": 0.1464, "H": 0.0057, "I": 0.0475, "K": 0.0062, "L": 0.0484, "M": 0.0097, "N": 0.0163, "P": 0.0259, "Q": 0.0048, "R": 0.0273, "S": 0.1587, "T": 0.0593, "V": 0.0571, "W": 0.0699, "Y": 0.1046 },
    "111D": { "A": 0.0644, "C": 0.0186, "D": 0.0331, "E": 0.0076, "F": 0.0398, "G": 0.1508, "H": 0.0059, "I": 0.0203, "K": 0.0025, "L": 0.0559, "M": 0.0008, "N": 0.0153, "P": 0.05, "Q": 0.0059, "R": 0.0458, "S": 0.1602, "T": 0.0669, "V": 0.0627, "W": 0.0771, "Y": 0.1161 },
    "111E": { "A": 0.0932, "C": 0.0342, "D": 0.0437, "E": 0.019, "F": 0.0361, "G": 0.1369, "H": 0.0019, "I": 0.0304, "L": 0.0285, "M": 0.0133, "P": 0.0684, "Q": 0.0038, "R": 0.0304, "S": 0.1806, "T": 0.0894, "V": 0.0608, "W": 0.0437, "Y": 0.0856 },
    "111F": { "A": 0.0455, "D": 0.017, "E": 0.017, "F": 0.0284, "G": 0.2727, "I": 0.0057, "L": 0.0114, "M": 0.0057, "P": 0.0398, "Q": 0.0114, "R": 0.0455, "S": 0.1818, "T": 0.0455, "V": 0.0455, "W": 0.0398, "Y": 0.1875 },
    "112G": { "A": 0.0263, "C": 0.0263, "D": 0.2368, "E": 0.0526, "F": 0.0088, "G": 0.1316, "I": 0.0175, "K": 0.0088, "L": 0.0351, "N": 0.0088, "P": 0.0175, "Q": 0.0088, "R": 0.0263, "S": 0.1053, "T": 0.114, "V": 0.0263, "W": 0.1053, "Y": 0.0439 },
    "112F": { "A": 0.0472, "C": 0.0678, "D": 0.0354, "E": 0.0059, "F": 0.0265, "G": 0.1239, "H": 0.0059, "I": 0.0413, "K": 0.0088, "L": 0.056, "M": 0.0059, "N": 0.0236, "P": 0.0383, "Q": 0.0088, "R": 0.0855, "S": 0.1121, "T": 0.0442, "V": 0.0737, "W": 0.0265, "Y": 0.1622 },
    "112E": { "A": 0.0247, "C": 0.0306, "D": 0.0283, "E": 0.0377, "F": 0.0188, "G": 0.1555, "H": 0.0047, "I": 0.0165, "K": 0.0188, "L": 0.0577, "M": 0.0118, "N": 0.0377, "P": 0.0777, "Q": 0.0165, "R": 0.0495, "S": 0.1048, "T": 0.0565, "V": 0.0766, "W": 0.0624, "Y": 0.1131 },
    "112D": { "A": 0.0429, "C": 0.0175, "D": 0.0616, "E": 0.0223, "F": 0.0193, "G": 0.1425, "H": 0.0254, "I": 0.0217, "K": 0.0133, "L": 0.0713, "M": 0.0085, "N": 0.0211, "P": 0.0936, "Q": 0.0175, "R": 0.0344, "S": 0.1087, "T": 0.0525, "V": 0.0477, "W": 0.0447, "Y": 0.1335 },
    "112C": { "A": 0.0453, "C": 0.0321, "D": 0.0546, "E": 0.0206, "F": 0.0203, "G": 0.1199, "H": 0.019, "I": 0.0598, "K": 0.0058, "L": 0.0559, "M": 0.0106, "N": 0.0357, "P": 0.0411, "Q": 0.0167, "R": 0.0428, "S": 0.1392, "T": 0.0463, "V": 0.0428, "W": 0.0286, "Y": 0.163 },
    "112B": { "A": 0.0496, "C": 0.0307, "D": 0.0557, "E": 0.0275, "F": 0.0223, "G": 0.1237, "H": 0.0174, "I": 0.033, "K": 0.0109, "L": 0.0252, "M": 0.0042, "N": 0.0158, "P": 0.038, "Q": 0.0113, "R": 0.062, "S": 0.1048, "T": 0.0414, "V": 0.0357, "W": 0.0189, "Y": 0.272 },
    "112A": { "A": 0.058, "C": 0.0122, "D": 0.0477, "E": 0.0183, "F": 0.0151, "G": 0.1322, "H": 0.0117, "I": 0.0179, "K": 0.0127, "L": 0.0364, "M": 0.0072, "N": 0.0179, "P": 0.0438, "Q": 0.0275, "R": 0.0469, "S": 0.0989, "T": 0.0353, "V": 0.0573, "W": 0.0265, "Y": 0.2765 },
    "112": { "A": 0.0506, "C": 0.0103, "D": 0.0436, "E": 0.0265, "F": 0.0151, "G": 0.1145, "H": 0.0199, "I": 0.0186, "K": 0.0071, "L": 0.0426, "M": 0.0082, "N": 0.0319, "P": 0.0489, "Q": 0.013, "R": 0.0405, "S": 0.0986, "T": 0.0505, "V": 0.0493, "W": 0.0329, "Y": 0.2774 },
    "113": { "A": 0.0351, "C": 0.0126, "D": 0.0808, "E": 0.0296, "F": 0.0203, "G": 0.1117, "H": 0.0183, "I": 0.0207, "K": 0.0089, "L": 0.0204, "M": 0.0032, "N": 0.0993, "P": 0.0401, "Q": 0.0109, "R": 0.0494, "S": 0.0719, "T": 0.0321, "V": 0.0239, "W": 0.0381, "Y": 0.2726 },
    "114": { "A": 0.111, "C": 0.0104, "D": 0.033, "E": 0.0193, "F": 0.0147, "G": 0.2426, "H": 0.0125, "I": 0.0091, "K": 0.0085, "L": 0.0285, "M": 0.003, "N": 0.0218, "P": 0.0437, "Q": 0.0088, "R": 0.0344, "S": 0.0471, "T": 0.0275, "V": 0.0235, "W": 0.1367, "Y": 0.1641 },
    "115": { "A": 0.0126, "C": 0.0058, "D": 0.0065, "E": 0.009, "F": 0.4655, "G": 0.0298, "H": 0.0041, "I": 0.0238, "K": 0.0016, "L": 0.088, "M": 0.2044, "N": 0.0104, "P": 0.0296, "Q": 0.0043, "R": 0.0041, "S": 0.0272, "T": 0.0102, "V": 0.0228, "W": 0.0121, "Y": 0.0279 },
    "116": { "A": 0.0103, "C": 0.0011, "D": 0.8335, "E": 0.0167, "F": 0.0029, "G": 0.0433, "H": 0.0084, "I": 0.0008, "K": 0.002, "L": 0.0077, "M": 0.0001, "N": 0.0127, "P": 0.0076, "Q": 0.0148, "R": 0.0061, "S": 0.0102, "T": 0.0042, "V": 0.0084, "W": 0.0015, "Y": 0.0076 },
    "117": { "A": 0.0059, "C": 0.0036, "D": 0.0096, "E": 0.0006, "F": 0.0169, "G": 0.0025, "H": 0.0277, "I": 0.0782, "K": 0.0004, "L": 0.033, "M": 0.0006, "N": 0.0101, "P": 0.1707, "Q": 0.0021, "R": 0.002, "S": 0.0379, "T": 0.0034, "V": 0.2217, "W": 0.0005, "Y": 0.3727 },
    "118": { "A": 0.0001, "C": 0.0017, "F": 0.0005, "G": 0.0029, "L": 0.0012, "R": 0.002, "S": 0.0002, "V": 0.0003, "W": 0.9909, "Y": 0.0004 },
    "119": { "A": 0.0018, "C": 0.0018, "D": 0.0006, "G": 0.993, "H": 0.0001, "L": 0.0001, "P": 0.0001, "R": 0.0003, "S": 0.0015, "V": 0.0009 },
    "120": { "A": 0.0002, "D": 0.0002, "E": 0.0022, "G": 0.0002, "H": 0.0039, "I": 0.0001, "K": 0.0411, "L": 0.0026, "N": 0.0005, "P": 0.0029, "Q": 0.9193, "R": 0.0268, "T": 0.0002 },
    "121": { "A": 0.0001, "C": 0.0001, "E": 0.0006, "F": 0.0001, "G": 0.9962, "Q": 0.0001, "R": 0.0011, "S": 0.0001, "V": 0.0014, "W": 0.0003 },
    "122": { "A": 0.0064, "C": 0.0001, "H": 0.0001, "I": 0.0007, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0005, "P": 0.0022, "Q": 0.0001, "R": 0.0002, "S": 0.0092, "T": 0.9801 },
    "123": { "A": 0.0024, "D": 0.0001, "F": 0.0014, "I": 0.0012, "K": 0.0034, "L": 0.6751, "M": 0.0829, "P": 0.0057, "Q": 0.0082, "R": 0.0053, "S": 0.0007, "T": 0.2107, "V": 0.0029 },
    "124": { "A": 0.0007, "D": 0.0003, "F": 0.0006, "G": 0.0029, "I": 0.0074, "L": 0.0083, "M": 0.0001, "P": 0.0002, "V": 0.9795 },
    "125": { "A": 0.0047, "F": 0.0003, "H": 0.0004, "I": 0.0099, "K": 0.0002, "L": 0.0001, "M": 0.0001, "N": 0.001, "P": 0.0012, "R": 0.0006, "S": 0.0103, "T": 0.9712, "V": 0.0002 },
    "126": { "A": 0.0023, "D": 0.0004, "F": 0.0011, "G": 0.0017, "I": 0.003, "L": 0.0003, "R": 0.0004, "S": 0.0001, "T": 0.0001, "V": 0.9907 },
    "127": { "A": 0.0028, "C": 0.0003, "D": 0.0001, "F": 0.0013, "L": 0.0006, "P": 0.0022, "S": 0.9899, "T": 0.0014, "V": 0.0001, "Y": 0.0015 },
    "128": { "A": 0.0065, "L": 0.0005, "P": 0.0046, "Q": 0.0001, "S": 0.9853, "T": 0.0029 }
  },
  "IGKV1": {
    "1": { "A": 0.1077, "D": 0.8873, "E": 0.0051 },
    "2": { "A": 0.0001, "F": 0.0038, "G": 0.0001, "H": 0.0001, "I": 0.9539, "L": 0.013, "M": 0.0025, "N": 0.0019, "P": 0.0001, "R": 0.0007, "S": 0.0077, "T": 0.0086, "V": 0.0077 },
    "3": { "A": 0.0001, "C": 0.0001, "D": 0.0003, "E": 0.0039, "G": 0.0001, "H": 0.0142, "K": 0.0083, "L": 0.0029, "M": 0.0001, "N": 0.0002, "P": 0.0092, "Q": 0.8785, "R": 0.0808, "T": 0.0006, "V": 0.0003, "W": 0.0004 },
    "4": { "A": 0.0001, "C": 0.0001, "E": 0.0001, "F": 0.0038, "I": 0.0275, "K": 0.0013, "L": 0.1472, "M": 0.8014, "N": 0.0001, "P": 0.0001, "Q": 0.0002, "R": 0.0033, "S": 0.0015, "T": 0.0053, "V": 0.0066, "W": 0.0013, "Y": 0.0001 },
    "5": { "A": 0.0033, "C": 0.0001, "D": 0.0001, "H": 0.0006, "I": 0.004, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0062, "P": 0.0156, "Q": 0.0001, "R": 0.0001, "S": 0.0051, "T": 0.9644, "V": 0.0001 },
    "6": { "A": 0.0001, "D": 0.0001, "E": 0.0029, "H": 0.0118, "K": 0.0081, "L": 0.0017, "M": 0.0003, "N": 0.0001, "P": 0.0202, "Q": 0.9511, "R": 0.0023, "S": 0.0003, "T": 0.0008, "V": 0.0001, "W": 0.0001 },
    "7": { "A": 0.0063, "C": 0.001, "D": 0.0003, "F": 0.0057, "H": 0.0007, "I": 0.0001, "L": 0.0005, "P": 0.0139, "S": 0.962, "T": 0.0033, "V": 0.0001, "Y": 0.0062 },
    "8": { "A": 0.0012, "H": 0.0016, "L": 0.0033, "P": 0.9844, "Q": 0.002, "R": 0.001, "S": 0.0025, "T": 0.004 },
    "9": { "A": 0.0048, "C": 0.0006, "D": 0.0001, "F": 0.0029, "H": 0.0001, "I": 0.0002, "L": 0.0006, "M": 0.0001, "P": 0.0092, "R": 0.0003, "S": 0.9743, "T": 0.0029, "V": 0.0001, "Y": 0.0038 },
    "10": { "A": 0.0087, "C": 0.0005, "F": 0.1039, "I": 0.0014, "K": 0.0001, "L": 0.0019, "N": 0.0019, "P": 0.0083, "R": 0.0002, "S": 0.4733, "T": 0.3968, "V": 0.0012, "Y": 0.0017 },
    "11": { "A": 0.0005, "C": 0.0005, "E": 0.0002, "F": 0.0544, "G": 0.0009, "H": 0.0001, "I": 0.0005, "L": 0.8162, "M": 0.0083, "P": 0.0025, "Q": 0.0014, "R": 0.0038, "S": 0.0007, "T": 0.0001, "V": 0.1101, "W": 0.0001, "Y": 0.0001 },
    "12": { "A": 0.0054, "C": 0.0004, "F": 0.0011, "H": 0.0001, "I": 0.0001, "L": 0.0003, "N": 0.0001, "P": 0.0093, "S": 0.9777, "T": 0.0024, "V": 0.0001, "Y": 0.0031 },
    "13": { "A": 0.9824, "D": 0.0003, "E": 0.0028, "G": 0.0007, "L": 0.0001, "P": 0.0031, "S": 0.0039, "T": 0.0034, "V": 0.0032 },
    "14": { "A": 0.0055, "C": 0.0005, "D": 0.0002, "F": 0.0064, "G": 0.0003, "H": 0.0001, "L": 0.0003, "N": 0.0001, "P": 0.0046, "S": 0.977, "T": 0.0015, "V": 0.0001, "W": 0.0001, "Y": 0.0034 },
    "15": { "A": 0.0055, "C": 0.0001, "D": 0.0001, "E": 0.0012, "F": 0.0001, "G": 0.0043, "I": 0.0129, "K": 0.0002, "L": 0.0072, "M": 0.0004, "P": 0.0013, "R": 0.0005, "S": 0.0004, "T": 0.0765, "V": 0.8894 },
    "16": { "A": 0.0036, "C": 0.0001, "D": 0.0003, "E": 0.0029, "G": 0.9817, "L": 0.0002, "R": 0.0051, "S": 0.0002, "V": 0.0059 },
    "17": { "A": 0.0102, "D": 0.965, "E": 0.0085, "G": 0.0034, "H": 0.0027, "N": 0.0034, "S": 0.0002, "T": 0.0002, "V": 0.0017, "Y": 0.0046 },
    "18": { "E": 0.0003, "G": 0.0032, "I": 0.0025, "K": 0.0043, "M": 0.0001, "N": 0.0003, "P": 0.0004, "Q": 0.0004, "R": 0.9615, "S": 0.0163, "T": 0.0107, "V": 0.0001 },
    "19": { "A": 0.0046, "C": 0.0002, "D": 0.0018, "F": 0.0031, "G": 0.0042, "I": 0.0071, "L": 0.0038, "P": 0.0001, "R": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.9747, "Y": 0.0001 },
    "20": { "A": 0.0031, "C": 0.0003, "D": 0.0002, "F": 0.0001, "H": 0.0005, "I": 0.0085, "K": 0.0003, "L": 0.0002, "M": 0.0001, "N": 0.0072, "P": 0.0102, "R": 0.0003, "S": 0.0087, "T": 0.9599, "V": 0.0005, "Y": 0.0001 },
    "21": { "A": 0.0001, "C": 0.0001, "F": 0.0039, "H": 0.0001, "I": 0.9611, "K": 0.0001, "L": 0.0135, "M": 0.0036, "N": 0.0021, "R": 0.0002, "S": 0.0059, "T": 0.0051, "V": 0.0042, "Y": 0.0001 },
    "22": { "A": 0.0039, "H": 0.0005, "I": 0.0016, "K": 0.0001, "L": 0.0001, "N": 0.0023, "P": 0.0199, "R": 0.0004, "S": 0.0095, "T": 0.9615, "V": 0.0002, "Y": 0.0001 },
    "23": { "A": 0.0001, "C": 0.9757, "F": 0.0048, "G": 0.0065, "H": 0.0001, "I": 0.0001, "L": 0.0002, "N": 0.0001, "R": 0.004, "S": 0.0047, "T": 0.0002, "V": 0.0003, "W": 0.0016, "Y": 0.0018 },
    "24": { "C": 0.0001, "G": 0.001, "H": 0.0001, "L": 0.004, "M": 0.0001, "P": 0.0023, "Q": 0.0038, "R": 0.9826, "S": 0.0003, "T": 0.0001, "W": 0.0057 },
    "25": { "A": 0.9619, "D": 0.0038, "E": 0.0025, "F": 0.0001, "G": 0.001, "H": 0.0001, "P": 0.0035, "Q": 0.0001, "S": 0.0064, "T": 0.0066, "V": 0.0139, "W": 0.0001 },
    "26": { "A": 0.0001, "C": 0.0009, "D": 0.0001, "G": 0.0039, "H": 0.0004, "I": 0.0032, "K": 0.0002, "L": 0.0002, "M": 0.0003, "N": 0.0037, "P": 0.0005, "Q": 0.0001, "R": 0.0167, "S": 0.9659, "T": 0.0038, "V": 0.0001 },
    "27": { "A": 0.0001, "D": 0.0006, "E": 0.0069, "F": 0.0001, "H": 0.0102, "I": 0.0001, "K": 0.0051, "L": 0.0027, "N": 0.0001, "P": 0.0089, "Q": 0.9587, "R": 0.0057, "S": 0.0001, "T": 0.0004, "V": 0.0001, "W": 0.0001, "Y": 0.0002 },
    "28": { "A": 0.0072, "C": 0.0059, "D": 0.0252, "E": 0.0003, "F": 0.0004, "G": 0.4796, "H": 0.0012, "I": 0.003, "K": 0.0002, "L": 0.0001, "M": 0.0002, "N": 0.0142, "P": 0.0011, "R": 0.0147, "S": 0.4248, "T": 0.0141, "V": 0.0068, "Y": 0.0012 },
    "29": { "A": 0.0003, "C": 0.0001, "D": 0.0001, "F": 0.0028, "G": 0.0001, "H": 0.0001, "I": 0.9505, "L": 0.0099, "M": 0.0041, "N": 0.0009, "P": 0.0001, "R": 0.0002, "S": 0.0081, "T": 0.0026, "V": 0.0201 },
    "36": { "A": 0.0041, "C": 0.0007, "D": 0.0078, "E": 0.001, "F": 0.0007, "G": 0.0248, "H": 0.0011, "I": 0.0051, "K": 0.0023, "L": 0.0007, "M": 0.0003, "N": 0.0278, "P": 0.0008, "R": 0.0799, "S": 0.8257, "T": 0.0126, "V": 0.0013, "W": 0.0001, "Y": 0.0033 },
    "37": { "A": 0.0016, "C": 0.0009, "D": 0.0075, "E": 0.0004, "F": 0.0009, "G": 0.0063, "H": 0.004, "I": 0.0102, "K": 0.0107, "L": 0.0005, "M": 0.0007, "N": 0.2409, "P": 0.0016, "Q": 0.0005, "R": 0.0243, "S": 0.6481, "T": 0.0362, "V": 0.0018, "W": 0.0003, "Y": 0.0027 },
    "38": { "A": 0.0016, "C": 0.0065, "D": 0.0653, "E": 0.0021, "F": 0.0146, "G": 0.0027, "H": 0.0052, "I": 0.0001, "K": 0.0004, "L": 0.0042, "M": 0.0001, "N": 0.0034, "P": 0.0001, "Q": 0.0005, "R": 0.0064, "S": 0.0283, "T": 0.0001, "V": 0.0004, "W": 0.4905, "Y": 0.3675 },
    "39": { "A": 0.0001, "F": 0.0085, "G": 0.0001, "I": 0.0018, "L": 0.9716, "M": 0.001, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.002, "V": 0.0132, "W": 0.0014, "Y": 0.0002 },
    "40": { "A": 0.8239, "C": 0.0008, "D": 0.0111, "E": 0.0001, "F": 0.0001, "G": 0.0645, "H": 0.0021, "I": 0.0006, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0713, "P": 0.0011, "Q": 0.0001, "R": 0.0001, "S": 0.0078, "T": 0.0057, "V": 0.0097, "Y": 0.0005 },
    "41": { "C": 0.0063, "E": 0.0001, "G": 0.002, "L": 0.0072, "M": 0.0001, "R": 0.0023, "S": 0.0007, "V": 0.0003, "W": 0.9811 },
    "42": { "C": 0.0016, "D": 0.0025, "F": 0.0561, "H": 0.0037, "I": 0.0005, "L": 0.001, "N": 0.0008, "R": 0.0001, "S": 0.0032, "V": 0.0007, "W": 0.0002, "Y": 0.9296 },
    "43": { "D": 0.0001, "E": 0.0007, "G": 0.0001, "H": 0.0073, "K": 0.0043, "L": 0.0012, "M": 0.0001, "P": 0.0038, "Q": 0.968, "R": 0.0141, "S": 0.0001, "T": 0.0001, "Y": 0.0001 },
    "44": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "E": 0.0016, "F": 0.0001, "G": 0.0001, "H": 0.0122, "I": 0.0001, "K": 0.0062, "L": 0.0025, "M": 0.0003, "N": 0.0003, "P": 0.0046, "Q": 0.9677, "R": 0.0026, "S": 0.0001, "T": 0.0003, "V": 0.001, "W": 0.0001, "Y": 0.0001 },
    "45": { "A": 0.0001, "D": 0.0003, "E": 0.0036, "H": 0.0001, "I": 0.0023, "K": 0.9549, "L": 0.0001, "M": 0.0005, "N": 0.0067, "P": 0.0002, "Q": 0.0089, "R": 0.0152, "S": 0.0007, "T": 0.0062, "V": 0.0003 },
    "46": { "A": 0.0058, "D": 0.0001, "E": 0.0002, "G": 0.0003, "H": 0.0003, "I": 0.0001, "K": 0.0001, "L": 0.0027, "P": 0.9708, "Q": 0.0055, "R": 0.001, "S": 0.0068, "T": 0.0061, "V": 0.0003 },
    "47": { "A": 0.0011, "C": 0.0002, "D": 0.0003, "E": 0.0049, "G": 0.9821, "K": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0033, "S": 0.0001, "V": 0.0051, "W": 0.0025 },
    "48": { "A": 0.0005, "D": 0.0002, "E": 0.0089, "G": 0.0007, "H": 0.0002, "I": 0.001, "K": 0.9439, "L": 0.0001, "M": 0.0006, "N": 0.0133, "Q": 0.009, "R": 0.0127, "S": 0.0011, "T": 0.0079 },
    "49": { "A": 0.87, "D": 0.0089, "E": 0.0001, "F": 0.0002, "G": 0.0032, "H": 0.0001, "I": 0.0008, "K": 0.0001, "L": 0.0004, "M": 0.0001, "P": 0.0093, "R": 0.0003, "S": 0.003, "T": 0.0032, "V": 0.1004 },
    "50": { "A": 0.0008, "H": 0.0053, "L": 0.0017, "N": 0.0001, "P": 0.9828, "Q": 0.0003, "R": 0.0007, "S": 0.0018, "T": 0.0064, "V": 0.0001 },
    "51": { "A": 0.0004, "D": 0.0016, "E": 0.0072, "G": 0.0001, "H": 0.0012, "I": 0.0012, "K": 0.9263, "L": 0.0001, "M": 0.0006, "N": 0.0307, "P": 0.0001, "Q": 0.0078, "R": 0.0111, "S": 0.001, "T": 0.01, "V": 0.0003, "Y": 0.0004 },
    "52": { "A": 0.0012, "C": 0.0002, "F": 0.0066, "G": 0.0003, "H": 0.0023, "I": 0.0058, "L": 0.8604, "N": 0.0001, "P": 0.0062, "Q": 0.0001, "R": 0.0483, "S": 0.0438, "T": 0.0004, "V": 0.0242, "W": 0.0001, "Y": 0.0001 },
    "53": { "F": 0.0002, "G": 0.0001, "I": 0.0014, "L": 0.9879, "M": 0.0034, "P": 0.0011, "Q": 0.0008, "R": 0.0034, "T": 0.0001, "V": 0.0016, "W": 0.0001 },
    "54": { "C": 0.0003, "D": 0.0001, "F": 0.0027, "H": 0.0002, "I": 0.9471, "K": 0.0001, "L": 0.0245, "M": 0.0036, "N": 0.0026, "P": 0.0002, "R": 0.0005, "S": 0.0063, "T": 0.0027, "V": 0.009, "Y": 0.0001 },
    "55": { "A": 0.0003, "C": 0.0012, "D": 0.0064, "E": 0.0003, "F": 0.0113, "H": 0.0094, "I": 0.0001, "K": 0.0003, "L": 0.0003, "N": 0.0024, "P": 0.0001, "Q": 0.0013, "R": 0.0001, "S": 0.0175, "T": 0.0003, "V": 0.0001, "W": 0.0005, "Y": 0.948 },
    "56": { "A": 0.5174, "C": 0.0003, "D": 0.0847, "E": 0.0105, "F": 0.0005, "G": 0.0269, "H": 0.0008, "I": 0.0002, "K": 0.2991, "L": 0.0012, "M": 0.0012, "N": 0.0024, "P": 0.0013, "Q": 0.0077, "R": 0.008, "S": 0.0212, "T": 0.0107, "V": 0.0048, "W": 0.0001, "Y": 0.001 },
    "57": { "A": 0.9587, "D": 0.0007, "E": 0.0027, "G": 0.0021, "I": 0.0007, "L": 0.0003, "M": 0.0001, "P": 0.0016, "R": 0.0001, "S": 0.0111, "T": 0.0152, "V": 0.0068 },
    "65": { "A": 0.0079, "C": 0.0007, "D": 0.0003, "F": 0.0033, "G": 0.0006, "I": 0.0002, "K": 0.0001, "L": 0.0003, "N": 0.0001, "P": 0.0033, "R": 0.0001, "S": 0.9729, "T": 0.0039, "V": 0.0002, "W": 0.0001, "Y": 0.0062 },
    "66": { "A": 0.0033, "C": 0.0007, "D": 0.0016, "E": 0.0001, "F": 0.0003, "G": 0.004, "H": 0.0009, "I": 0.0122, "K": 0.0022, "L": 0.0008, "M": 0.0007, "N": 0.0355, "P": 0.0027, "Q": 0.0003, "R": 0.0269, "S": 0.5983, "T": 0.3056, "V": 0.0029, "Y": 0.0011 },
    "67": { "A": 0.0001, "C": 0.0002, "F": 0.0055, "G": 0.0001, "I": 0.0005, "L": 0.9697, "M": 0.0017, "P": 0.0001, "Q": 0.0001, "R": 0.0007, "S": 0.0032, "T": 0.0001, "V": 0.013, "W": 0.0051, "Y": 0.0001 },
    "68": { "A": 0.0042, "C": 0.0001, "D": 0.0031, "E": 0.3988, "F": 0.0002, "G": 0.0033, "H": 0.0098, "I": 0.0001, "K": 0.0071, "L": 0.0014, "M": 0.0001, "N": 0.0006, "P": 0.0032, "Q": 0.5627, "R": 0.0035, "S": 0.0003, "T": 0.0002, "V": 0.0007, "Y": 0.0007 },
    "69": { "A": 0.0023, "C": 0.0008, "D": 0.0016, "E": 0.0003, "F": 0.0003, "G": 0.0112, "H": 0.0001, "I": 0.0043, "K": 0.0006, "L": 0.0008, "M": 0.0002, "N": 0.0102, "P": 0.0021, "R": 0.0154, "S": 0.9304, "T": 0.0184, "V": 0.0005, "W": 0.0001, "Y": 0.0005 },
    "70": { "A": 0.0025, "C": 0.0001, "D": 0.0007, "E": 0.004, "G": 0.9796, "H": 0.0001, "K": 0.0001, "P": 0.0002, "R": 0.0034, "S": 0.0003, "T": 0.0001, "V": 0.0045, "W": 0.0047 },
    "71": { "A": 0.0051, "C": 0.0001, "D": 0.0018, "F": 0.0049, "G": 0.0062, "I": 0.003, "L": 0.0011, "R": 0.0001, "S": 0.0001, "T": 0.0002, "V": 0.9774 },
    "72": { "A": 0.0015, "H": 0.0001, "I": 0.0002, "K": 0.0001, "L": 0.002, "P": 0.9815, "Q": 0.0033, "R": 0.0006, "S": 0.0051, "T": 0.0056, "V": 0.0001 },
    "74": { "A": 0.0148, "C": 0.0004, "D": 0.0003, "E": 0.0007, "F": 0.0007, "G": 0.0002, "I": 0.0001, "K": 0.0001, "L": 0.0037, "P": 0.009, "Q": 0.0001, "R": 0.0001, "S": 0.9645, "T": 0.0047, "V": 0.0001, "W": 0.0001, "Y": 0.0004 },
    "75": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "E": 0.0001, "G": 0.002, "H": 0.0003, "I": 0.0002, "K": 0.0435, "L": 0.001, "M": 0.0045, "N": 0.001, "P": 0.0003, "Q": 0.0008, "R": 0.9348, "S": 0.0076, "T": 0.0025, "W": 0.0013 },
    "76": { "A": 0.0001, "C": 0.0135, "D": 0.0001, "F": 0.9656, "G": 0.0001, "I": 0.0025, "L": 0.0061, "P": 0.0001, "S": 0.0023, "V": 0.0081, "Y": 0.0016 },
    "77": { "A": 0.0004, "C": 0.0009, "F": 0.0001, "G": 0.0056, "H": 0.0002, "I": 0.0053, "L": 0.0003, "M": 0.0001, "N": 0.0027, "P": 0.0001, "R": 0.0134, "S": 0.9669, "T": 0.0032, "V": 0.0007, "Y": 0.0002 },
    "78": { "A": 0.0031, "C": 0.0061, "D": 0.002, "F": 0.0002, "G": 0.9787, "P": 0.0001, "R": 0.0018, "S": 0.004, "V": 0.004 },
    "79": { "A": 0.0005, "C": 0.0014, "D": 0.0007, "F": 0.0001, "G": 0.0112, "H": 0.0001, "I": 0.0029, "L": 0.0001, "M": 0.0001, "N": 0.003, "R": 0.0173, "S": 0.9577, "T": 0.0046, "V": 0.0001, "W": 0.0001, "Y": 0.0002 },
    "80": { "A": 0.0053, "C": 0.0005, "D": 0.0001, "E": 0.0036, "F": 0.0001, "G": 0.9744, "I": 0.0001, "K": 0.0006, "L": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0002, "R": 0.0076, "S": 0.0008, "T": 0.0002, "V": 0.0059, "W": 0.0002, "Y": 0.0001 },
    "83": { "A": 0.0141, "C": 0.0018, "D": 0.0007, "F": 0.0064, "G": 0.0003, "H": 0.0002, "I": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0023, "S": 0.9598, "T": 0.0028, "V": 0.0003, "W": 0.0001, "Y": 0.0109 },
    "84": { "A": 0.0047, "C": 0.0005, "D": 0.0007, "E": 0.0097, "G": 0.9649, "L": 0.0001, "Q": 0.0001, "R": 0.0049, "S": 0.0005, "T": 0.0001, "V": 0.0082, "W": 0.0057 },
    "85": { "A": 0.0062, "E": 0.0001, "I": 0.003, "K": 0.0069, "L": 0.0001, "P": 0.002, "Q": 0.0001, "R": 0.0017, "S": 0.0023, "T": 0.9777 },
    "86": { "A": 0.0046, "D": 0.4322, "E": 0.5417, "F": 0.0004, "G": 0.0025, "H": 0.0025, "K": 0.002, "L": 0.0003, "N": 0.0025, "P": 0.0002, "Q": 0.0039, "R": 0.0003, "S": 0.0006, "T": 0.0004, "V": 0.0029, "Y": 0.0031 },
    "87": { "A": 0.0001, "C": 0.0045, "D": 0.0003, "F": 0.9602, "G": 0.0001, "H": 0.0001, "I": 0.0021, "L": 0.0038, "N": 0.0001, "S": 0.002, "V": 0.0081, "Y": 0.0187 },
    "88": { "A": 0.0056, "D": 0.0001, "F": 0.0002, "H": 0.0002, "I": 0.0041, "N": 0.0035, "P": 0.0048, "S": 0.0121, "T": 0.9692, "Y": 0.0001 },
    "89": { "F": 0.0029, "H": 0.0013, "I": 0.0064, "L": 0.9802, "P": 0.002, "R": 0.0058, "S": 0.0002, "T": 0.0001, "V": 0.0012 },
    "90": { "A": 0.0046, "D": 0.0001, "E": 0.0001, "F": 0.0001, "H": 0.0001, "I": 0.0046, "K": 0.0007, "L": 0.0001, "M": 0.0001, "N": 0.0033, "P": 0.0044, "R": 0.0005, "S": 0.0064, "T": 0.9751, "V": 0.0001 },
    "91": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.0016, "I": 0.9732, "K": 0.0001, "L": 0.0049, "M": 0.0007, "N": 0.0008, "P": 0.0001, "R": 0.0001, "S": 0.0115, "T": 0.0022, "V": 0.0045 },
    "92": { "A": 0.0015, "C": 0.001, "D": 0.0031, "F": 0.0002, "G": 0.0034, "H": 0.0002, "I": 0.006, "K": 0.0002, "L": 0.0002, "M": 0.0001, "N": 0.0176, "Q": 0.0001, "R": 0.0104, "S": 0.9407, "T": 0.0148, "V": 0.0001, "Y": 0.0003 },
    "93": { "A": 0.0002, "C": 0.0782, "D": 0.0021, "F": 0.0004, "G": 0.0158, "H": 0.0001, "I": 0.0061, "K": 0.0005, "L": 0.0004, "M": 0.0006, "N": 0.0105, "P": 0.0002, "R": 0.0217, "S": 0.8592, "T": 0.0034, "V": 0.0001, "W": 0.0003, "Y": 0.0003 },
    "94": { "A": 0.0001, "F": 0.0001, "H": 0.0001, "K": 0.0001, "L": 0.982, "M": 0.0028, "P": 0.0014, "Q": 0.0012, "R": 0.0098, "V": 0.0024 },
    "95": { "D": 0.0003, "E": 0.0038, "H": 0.0076, "K": 0.0022, "L": 0.0021, "P": 0.002, "Q": 0.9779, "R": 0.0041, "S": 0.0001, "Y": 0.0001 },
    "96": { "A": 0.0077, "C": 0.0001, "D": 0.0001, "F": 0.0002, "H": 0.0053, "L": 0.0025, "P": 0.8932, "Q": 0.0006, "R": 0.0016, "S": 0.0818, "T": 0.0061, "V": 0.0003, "Y": 0.0006 },
    "97": { "A": 0.0028, "C": 0.0002, "D": 0.4004, "E": 0.5845, "G": 0.0042, "H": 0.0006, "K": 0.0008, "L": 0.0001, "N": 0.001, "P": 0.0001, "Q": 0.0006, "S": 0.0003, "T": 0.0001, "V": 0.0016, "Y": 0.0029 },
    "98": { "A": 0.0026, "D": 0.9806, "E": 0.0065, "G": 0.0011, "H": 0.0005, "I": 0.0001, "N": 0.0012, "V": 0.0015, "Y": 0.0059 },
    "99": { "A": 0.0007, "C": 0.0071, "D": 0.0001, "F": 0.8523, "G": 0.0008, "I": 0.0039, "L": 0.0094, "N": 0.0001, "P": 0.0001, "R": 0.0001, "S": 0.014, "T": 0.0003, "V": 0.1093, "W": 0.0001, "Y": 0.0016 },
    "100": { "A": 0.9773, "C": 0.0002, "D": 0.0002, "E": 0.0036, "G": 0.0081, "H": 0.0001, "L": 0.0001, "P": 0.0005, "Q": 0.0001, "R": 0.0001, "S": 0.0042, "T": 0.0033, "V": 0.0023 },
    "101": { "A": 0.0085, "D": 0.0003, "F": 0.0001, "H": 0.0002, "I": 0.0076, "K": 0.0001, "L": 0.0004, "M": 0.0002, "N": 0.0051, "P": 0.0041, "R": 0.0003, "S": 0.0105, "T": 0.9589, "V": 0.0037, "Y": 0.0001 },
    "102": { "A": 0.0002, "C": 0.0014, "D": 0.0054, "E": 0.0005, "F": 0.0027, "H": 0.0008, "I": 0.0003, "L": 0.0007, "N": 0.0009, "Q": 0.0001, "S": 0.0019, "Y": 0.9852 },
    "103": { "A": 0.0001, "C": 0.0049, "D": 0.0051, "E": 0.0001, "F": 0.0185, "G": 0.0001, "H": 0.0053, "I": 0.0003, "L": 0.0007, "N": 0.0007, "P": 0.0001, "S": 0.0089, "T": 0.0007, "V": 0.0002, "W": 0.0005, "Y": 0.9539 },
    "104": { "A": 0.0002, "C": 0.9667, "D": 0.0002, "F": 0.0037, "G": 0.0156, "L": 0.0006, "Q": 0.0001, "R": 0.0027, "S": 0.0018, "V": 0.0006, "W": 0.0052, "Y": 0.0026 },
    "105": { "C": 0.0001, "E": 0.001, "G": 0.0001, "H": 0.0088, "I": 0.0003, "K": 0.0048, "L": 0.0724, "M": 0.0002, "N": 0.0005, "P": 0.0015, "Q": 0.9029, "R": 0.0066, "S": 0.0004, "T": 0.0002, "Y": 0.0003 },
    "106": { "A": 0.0001, "C": 0.0001, "D": 0.0002, "E": 0.0037, "F": 0.0002, "G": 0.0005, "H": 0.0246, "I": 0.0001, "K": 0.0855, "L": 0.0023, "M": 0.0001, "N": 0.0025, "P": 0.0034, "Q": 0.8563, "R": 0.0169, "S": 0.0014, "T": 0.0007, "V": 0.0004, "W": 0.0002, "Y": 0.0008 },
    "107": { "A": 0.098, "C": 0.0051, "D": 0.0402, "E": 0.0032, "F": 0.0049, "G": 0.0046, "H": 0.0417, "I": 0.0028, "K": 0.0006, "L": 0.0937, "M": 0.0003, "N": 0.0036, "P": 0.0024, "Q": 0.0021, "R": 0.008, "S": 0.0584, "T": 0.0192, "V": 0.0045, "W": 0.0002, "Y": 0.6063 },
    "108": { "A": 0.0013, "C": 0.0005, "D": 0.0416, "E": 0.0035, "F": 0.0043, "G": 0.0031, "H": 0.0149, "I": 0.0031, "K": 0.0178, "L": 0.0018, "M": 0.0003, "N": 0.6957, "P": 0.0002, "Q": 0.0018, "R": 0.0021, "S": 0.0199, "T": 0.0061, "V": 0.0013, "W": 0.0001, "Y": 0.1807 },
    "109": { "A": 0.0033, "C": 0.0011, "D": 0.0071, "E": 0.0006, "F": 0.0019, "G": 0.0117, "H": 0.0013, "I": 0.0067, "K": 0.0021, "L": 0.0015, "M": 0.0004, "N": 0.06, "P": 0.0007, "Q": 0.0002, "R": 0.0173, "S": 0.8369, "T": 0.0438, "V": 0.0014, "W": 0.0001, "Y": 0.0017 },
    "110": { "A": 0.1342, "C": 0.0026, "D": 0.0211, "F": 0.0868, "H": 0.0026, "K": 0.0053, "L": 0.0026, "N": 0.0395, "S": 0.0053, "T": 0.0316, "V": 0.0053, "W": 0.0079, "Y": 0.6553 },
    "113": { "A": 0.0807, "C": 0.0037, "D": 0.0083, "E": 0.0012, "F": 0.0686, "G": 0.0008, "H": 0.0054, "I": 0.0046, "K": 0.0004, "L": 0.0087, "N": 0.0166, "P": 0.0907, "Q": 0.0004, "R": 0.0037, "S": 0.0628, "T": 0.0715, "V": 0.005, "W": 0.0021, "Y": 0.5645 },
    "114": { "A": 0.0767, "C": 0.0024, "D": 0.0062, "E": 0.0012, "F": 0.1041, "G": 0.0043, "H": 0.0043, "I": 0.0044, "K": 0.0007, "L": 0.0208, "M": 0.0005, "N": 0.0078, "P": 0.0892, "Q": 0.0013, "R": 0.0093, "S": 0.1519, "T": 0.0581, "V": 0.0052, "W": 0.0019, "Y": 0.4495 },
    "115": { "A": 0.0188, "C": 0.0009, "D": 0.0026, "E": 0.0051, "F": 0.0135, "G": 0.0162, "H": 0.003, "I": 0.0033, "K": 0.0024, "L": 0.0576, "M": 0.0104, "N": 0.0054, "P": 0.5882, "Q": 0.0119, "R": 0.0227, "S": 0.1351, "T": 0.0187, "V": 0.0043, "W": 0.0069, "Y": 0.0729 },
    "116": { "A": 0.0076, "C": 0.0111, "D": 0.0048, "E": 0.0055, "F": 0.0744, "G": 0.0324, "H": 0.0145, "I": 0.0509, "K": 0.0053, "L": 0.1851, "M": 0.0009, "N": 0.0064, "P": 0.0688, "Q": 0.0184, "R": 0.1255, "S": 0.032, "T": 0.0033, "V": 0.0238, "W": 0.1938, "Y": 0.1356 },
    "117": { "A": 0.0246, "C": 0.0003, "D": 0.0017, "E": 0.0005, "F": 0.0025, "G": 0.0042, "H": 0.0012, "I": 0.0027, "K": 0.0011, "L": 0.0048, "M": 0.0008, "N": 0.0023, "P": 0.0105, "Q": 0.001, "R": 0.0032, "S": 0.0376, "T": 0.8968, "V": 0.0016, "W": 0.0002, "Y": 0.0022 },
    "118": { "A": 0.0001, "C": 0.0076, "F": 0.976, "G": 0.0003, "H": 0.0001, "I": 0.0008, "K": 0.0001, "L": 0.005, "N": 0.0001, "P": 0.0008, "R": 0.0002, "S": 0.0022, "T": 0.0001, "V": 0.0058, "W": 0.0001, "Y": 0.0009 },
    "119": { "A": 0.0018, "C": 0.0008, "D": 0.002, "G": 0.9909, "H": 0.0003, "L": 0.0001, "N": 0.0001, "P": 0.0003, "R": 0.0011, "S": 0.0016, "V": 0.001, "W": 0.0002 },
    "120": { "A": 0.0012, "D": 0.0001, "E": 0.0031, "G": 0.2151, "H": 0.0089, "K": 0.0011, "L": 0.0012, "N": 0.0001, "P": 0.1129, "Q": 0.6483, "R": 0.0071, "S": 0.0005, "T": 0.0001, "V": 0.0003 },
    "121": { "A": 0.0003, "E": 0.003, "G": 0.9887, "R": 0.0033, "V": 0.0018, "W": 0.0029 },
    "122": { "A": 0.0022, "D": 0.0003, "I": 0.001, "K": 0.0001, "N": 0.0008, "P": 0.0016, "S": 0.0027, "T": 0.9913 },
    "123": { "A": 0.0003, "D": 0.0001, "E": 0.0023, "G": 0.0003, "H": 0.0003, "I": 0.0005, "K": 0.9004, "L": 0.0012, "M": 0.001, "N": 0.0037, "P": 0.0001, "Q": 0.0048, "R": 0.079, "S": 0.0002, "T": 0.0057, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "124": { "A": 0.0005, "E": 0.0003, "F": 0.0001, "G": 0.0004, "I": 0.0007, "L": 0.2699, "M": 0.003, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "V": 0.7248 },
    "125": { "A": 0.0021, "D": 0.1214, "E": 0.8653, "G": 0.003, "H": 0.0007, "K": 0.002, "N": 0.0005, "Q": 0.0031, "R": 0.0005, "S": 0.0001, "T": 0.0001, "V": 0.001, "W": 0.0001, "Y": 0.0002 },
    "126": { "C": 0.0001, "F": 0.0031, "I": 0.9537, "L": 0.0064, "M": 0.0122, "N": 0.0006, "Q": 0.0001, "R": 0.0002, "S": 0.0052, "T": 0.0024, "V": 0.0157, "Y": 0.0003 },
    "127": { "E": 0.0031, "K": 0.9932, "N": 0.0008, "Q": 0.0029, "S": 0.0001, "T": 0.0001 }
  },
  "IGKV1D": {
    "1": { "A": 0.0677, "D": 0.8884, "E": 0.0062, "I": 0.0001, "N": 0.0001, "V": 0.0376 },
    "2": { "C": 0.0001, "D": 0.0001, "F": 0.0035, "G": 0.0002, "H": 0.0001, "I": 0.9473, "K": 0.0001, "L": 0.0176, "M": 0.003, "N": 0.0024, "Q": 0.0001, "R": 0.0005, "S": 0.0083, "T": 0.007, "V": 0.0097 },
    "3": { "A": 0.0001, "C": 0.0004, "D": 0.0002, "E": 0.0047, "G": 0.0019, "H": 0.0146, "K": 0.011, "L": 0.0032, "N": 0.0005, "P": 0.0083, "Q": 0.903, "R": 0.0098, "S": 0.0004, "T": 0.0004, "V": 0.0001, "W": 0.0413, "Y": 0.0003 },
    "4": { "C": 0.0001, "D": 0.0001, "F": 0.0014, "G": 0.0001, "H": 0.0001, "I": 0.0226, "K": 0.0013, "L": 0.0986, "M": 0.8552, "N": 0.0002, "P": 0.0005, "Q": 0.0001, "R": 0.005, "S": 0.0007, "T": 0.0058, "V": 0.0083, "W": 0.0002 },
    "5": { "A": 0.0037, "D": 0.0002, "F": 0.0001, "H": 0.0005, "I": 0.0036, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0067, "P": 0.0171, "R": 0.0005, "S": 0.0064, "T": 0.9608, "V": 0.0001, "Y": 0.0001 },
    "6": { "A": 0.0001, "E": 0.0024, "H": 0.0148, "K": 0.0101, "L": 0.002, "M": 0.0001, "P": 0.0202, "Q": 0.9465, "R": 0.0025, "S": 0.0005, "T": 0.0005, "V": 0.0001, "W": 0.0001, "Y": 0.0002 },
    "7": { "A": 0.0057, "C": 0.0014, "D": 0.0002, "E": 0.0001, "F": 0.0032, "G": 0.0001, "H": 0.0005, "I": 0.0002, "L": 0.0001, "P": 0.0131, "Q": 0.0001, "S": 0.9647, "T": 0.0036, "V": 0.0002, "W": 0.0001, "Y": 0.0066 },
    "8": { "A": 0.0014, "H": 0.0003, "I": 0.0001, "K": 0.0001, "L": 0.003, "N": 0.0001, "P": 0.9803, "Q": 0.0051, "R": 0.0012, "S": 0.0024, "T": 0.0061 },
    "9": { "A": 0.0054, "C": 0.0005, "D": 0.0001, "E": 0.0001, "F": 0.0063, "G": 0.0001, "I": 0.0002, "L": 0.0006, "N": 0.0001, "P": 0.0098, "R": 0.0001, "S": 0.9689, "T": 0.0035, "V": 0.0003, "W": 0.0001, "Y": 0.004 },
    "10": { "A": 0.0045, "C": 0.0007, "D": 0.0001, "F": 0.005, "G": 0.0001, "H": 0.0001, "I": 0.0005, "L": 0.0434, "N": 0.0001, "P": 0.0094, "Q": 0.0001, "S": 0.9284, "T": 0.0039, "V": 0.0001, "Y": 0.0037 },
    "11": { "F": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0003, "L": 0.9831, "M": 0.0049, "P": 0.0036, "Q": 0.0006, "R": 0.003, "S": 0.0001, "V": 0.0041, "W": 0.0001 },
    "12": { "A": 0.005, "C": 0.0005, "F": 0.0017, "G": 0.0001, "L": 0.0002, "P": 0.0068, "S": 0.9814, "T": 0.0016, "V": 0.0001, "Y": 0.0026 },
    "13": { "A": 0.9831, "D": 0.0003, "E": 0.0023, "G": 0.0009, "I": 0.0001, "L": 0.0001, "P": 0.0021, "Q": 0.0001, "R": 0.0001, "S": 0.0036, "T": 0.0031, "V": 0.0042 },
    "14": { "A": 0.0065, "C": 0.0012, "D": 0.0001, "F": 0.0046, "I": 0.0001, "L": 0.0001, "N": 0.0002, "P": 0.0043, "R": 0.0001, "S": 0.9777, "T": 0.0017, "V": 0.0002, "Y": 0.0031 },
    "15": { "A": 0.0057, "C": 0.0001, "D": 0.0001, "E": 0.0021, "F": 0.0004, "G": 0.0036, "I": 0.0104, "K": 0.0003, "L": 0.0077, "M": 0.0003, "P": 0.0007, "Q": 0.0001, "R": 0.0003, "S": 0.0001, "T": 0.0439, "V": 0.9243, "W": 0.0001 },
    "16": { "A": 0.0043, "C": 0.0003, "D": 0.0003, "E": 0.0022, "G": 0.9814, "L": 0.0001, "P": 0.0001, "R": 0.0055, "S": 0.0002, "T": 0.0002, "V": 0.0053, "W": 0.0002 },
    "17": { "A": 0.01, "D": 0.97, "E": 0.0063, "F": 0.0001, "G": 0.003, "H": 0.002, "N": 0.0026, "P": 0.0001, "Q": 0.0002, "R": 0.0001, "S": 0.0004, "T": 0.0001, "V": 0.0019, "Y": 0.0034 },
    "18": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "G": 0.0028, "I": 0.0029, "K": 0.0028, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.9699, "S": 0.0152, "T": 0.0058 },
    "19": { "A": 0.0047, "C": 0.0001, "D": 0.0022, "F": 0.0036, "G": 0.0038, "H": 0.0001, "I": 0.0074, "L": 0.0034, "S": 0.0001, "T": 0.0001, "V": 0.9746 },
    "20": { "A": 0.0035, "C": 0.0007, "D": 0.0001, "F": 0.0002, "H": 0.0005, "I": 0.0063, "K": 0.0002, "L": 0.0001, "N": 0.0052, "P": 0.0093, "R": 0.0004, "S": 0.008, "T": 0.965, "V": 0.0005 },
    "21": { "D": 0.0001, "F": 0.0037, "G": 0.0001, "H": 0.0001, "I": 0.9683, "L": 0.0109, "M": 0.0013, "N": 0.0025, "P": 0.0001, "R": 0.0002, "S": 0.0043, "T": 0.0035, "V": 0.005 },
    "22": { "A": 0.0038, "C": 0.0001, "F": 0.0002, "H": 0.0009, "I": 0.0021, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0047, "P": 0.0157, "R": 0.0013, "S": 0.0514, "T": 0.9195, "V": 0.0001, "Y": 0.0001 },
    "23": { "A": 0.0001, "C": 0.9755, "F": 0.0053, "G": 0.005, "I": 0.0001, "L": 0.0003, "P": 0.0001, "R": 0.005, "S": 0.0054, "W": 0.0009, "Y": 0.0024 },
    "24": { "C": 0.0001, "E": 0.0001, "G": 0.0005, "H": 0.003, "K": 0.0007, "L": 0.0052, "P": 0.0041, "Q": 0.3267, "R": 0.6536, "S": 0.0001, "T": 0.0001, "W": 0.0057 },
    "25": { "A": 0.9259, "D": 0.0003, "E": 0.0058, "F": 0.0001, "G": 0.0014, "I": 0.0007, "K": 0.0003, "L": 0.0027, "M": 0.0401, "P": 0.0041, "Q": 0.0002, "R": 0.0003, "S": 0.0056, "T": 0.0087, "V": 0.0038, "W": 0.0001 },
    "26": { "C": 0.001, "D": 0.0002, "E": 0.0001, "F": 0.0001, "G": 0.0033, "H": 0.0001, "I": 0.0034, "K": 0.0002, "L": 0.0001, "M": 0.0001, "N": 0.0055, "P": 0.0001, "R": 0.0132, "S": 0.9669, "T": 0.0053, "V": 0.0003, "Y": 0.0001 },
    "27": { "A": 0.0001, "D": 0.0004, "E": 0.0043, "G": 0.0001, "H": 0.0138, "I": 0.0001, "K": 0.0073, "L": 0.002, "N": 0.0003, "P": 0.0057, "Q": 0.959, "R": 0.0057, "S": 0.0002, "T": 0.0005, "V": 0.0001, "Y": 0.0003 },
    "28": { "A": 0.0053, "C": 0.0022, "D": 0.3284, "E": 0.0017, "F": 0.0003, "G": 0.1096, "H": 0.0031, "I": 0.003, "K": 0.0003, "L": 0.0002, "M": 0.0001, "N": 0.0099, "P": 0.001, "Q": 0.0001, "R": 0.0097, "S": 0.5079, "T": 0.0109, "V": 0.0024, "Y": 0.0039 },
    "29": { "A": 0.0003, "F": 0.002, "G": 0.0001, "H": 0.0002, "I": 0.9631, "L": 0.0062, "M": 0.0041, "N": 0.0014, "P": 0.0001, "R": 0.0005, "S": 0.0094, "T": 0.0037, "V": 0.0089 },
    "36": { "A": 0.0037, "C": 0.001, "D": 0.0057, "E": 0.0003, "F": 0.0009, "G": 0.0147, "H": 0.0012, "I": 0.0059, "K": 0.0038, "L": 0.001, "M": 0.0001, "N": 0.0243, "P": 0.0002, "Q": 0.0001, "R": 0.0259, "S": 0.8903, "T": 0.0174, "V": 0.0007, "Y": 0.0028 },
    "37": { "A": 0.0008, "C": 0.0007, "D": 0.0056, "E": 0.0004, "F": 0.0011, "G": 0.0048, "H": 0.005, "I": 0.0086, "K": 0.0184, "L": 0.0003, "M": 0.0005, "N": 0.3164, "P": 0.0007, "Q": 0.0005, "R": 0.0157, "S": 0.5904, "T": 0.0254, "V": 0.0019, "Y": 0.003 },
    "38": { "A": 0.0488, "C": 0.0039, "D": 0.0058, "E": 0.0004, "F": 0.0166, "G": 0.001, "H": 0.011, "I": 0.0003, "K": 0.0009, "L": 0.0012, "N": 0.0087, "P": 0.0001, "Q": 0.0007, "R": 0.0017, "S": 0.0184, "T": 0.001, "V": 0.0008, "W": 0.0107, "Y": 0.868 },
    "39": { "C": 0.0003, "F": 0.0086, "I": 0.0026, "L": 0.973, "M": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0022, "V": 0.013, "W": 0.0001 },
    "40": { "A": 0.1129, "D": 0.005, "E": 0.0003, "G": 0.0012, "H": 0.0161, "I": 0.0031, "K": 0.0067, "L": 0.0002, "N": 0.8281, "P": 0.0003, "Q": 0.0003, "R": 0.0001, "S": 0.0103, "T": 0.0114, "V": 0.001, "Y": 0.0029 },
    "41": { "C": 0.0052, "G": 0.0034, "L": 0.0059, "M": 0.0001, "Q": 0.0002, "R": 0.0037, "S": 0.0015, "V": 0.0001, "W": 0.9797, "Y": 0.0001 },
    "42": { "A": 0.0001, "C": 0.001, "D": 0.0036, "E": 0.0002, "F": 0.0106, "H": 0.0035, "I": 0.0003, "L": 0.0005, "N": 0.0012, "S": 0.0035, "V": 0.0001, "W": 0.0001, "Y": 0.9752 },
    "43": { "C": 0.0001, "E": 0.0009, "H": 0.0079, "K": 0.0052, "L": 0.0015, "P": 0.0029, "Q": 0.975, "R": 0.0058, "S": 0.0001, "T": 0.0001, "V": 0.0003, "W": 0.0001, "Y": 0.0002 },
    "44": { "A": 0.0003, "E": 0.0015, "F": 0.0001, "H": 0.0162, "I": 0.0001, "K": 0.0046, "L": 0.0021, "M": 0.0001, "N": 0.0002, "P": 0.0028, "Q": 0.9677, "R": 0.003, "T": 0.0001, "V": 0.0009, "W": 0.0001, "Y": 0.0004 },
    "45": { "A": 0.0002, "E": 0.0031, "G": 0.0002, "H": 0.0003, "I": 0.0019, "K": 0.9587, "L": 0.0001, "M": 0.0004, "N": 0.0064, "P": 0.0001, "Q": 0.0083, "R": 0.0147, "S": 0.0004, "T": 0.0052, "V": 0.0002, "Y": 0.0001 },
    "46": { "A": 0.0057, "E": 0.0007, "F": 0.0001, "G": 0.0003, "H": 0.0001, "I": 0.0002, "K": 0.0001, "L": 0.0031, "P": 0.9687, "Q": 0.0061, "R": 0.0008, "S": 0.0064, "T": 0.007, "V": 0.0009 },
    "47": { "A": 0.0055, "C": 0.0001, "D": 0.0004, "E": 0.0123, "G": 0.9697, "L": 0.0001, "P": 0.0001, "Q": 0.0003, "R": 0.0043, "T": 0.0001, "V": 0.0049, "W": 0.0022 },
    "48": { "A": 0.0002, "D": 0.0001, "E": 0.0094, "G": 0.0004, "H": 0.0002, "I": 0.0009, "K": 0.9494, "L": 0.0003, "M": 0.0005, "N": 0.0102, "P": 0.0001, "Q": 0.011, "R": 0.0102, "S": 0.0007, "T": 0.0063, "V": 0.0001 },
    "49": { "A": 0.9601, "D": 0.0098, "E": 0.0003, "F": 0.0003, "G": 0.0038, "I": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0116, "S": 0.0047, "T": 0.0041, "V": 0.005 },
    "50": { "A": 0.0011, "H": 0.0055, "I": 0.0001, "L": 0.0016, "N": 0.0001, "P": 0.9813, "Q": 0.0002, "R": 0.0007, "S": 0.0016, "T": 0.0079 },
    "51": { "A": 0.0005, "D": 0.0023, "E": 0.0481, "G": 0.0004, "H": 0.002, "I": 0.0006, "K": 0.8879, "L": 0.0001, "M": 0.0011, "N": 0.0231, "P": 0.0001, "Q": 0.0127, "R": 0.0102, "S": 0.0009, "T": 0.0094, "V": 0.0003, "W": 0.0001, "Y": 0.0001 },
    "52": { "A": 0.0005, "C": 0.0003, "D": 0.0001, "F": 0.0049, "G": 0.0003, "H": 0.0039, "I": 0.0054, "L": 0.9383, "M": 0.0001, "N": 0.0001, "P": 0.0068, "Q": 0.0001, "R": 0.0099, "S": 0.0104, "T": 0.0001, "V": 0.0188, "Y": 0.0001 },
    "53": { "F": 0.0043, "H": 0.0001, "I": 0.0005, "L": 0.9827, "M": 0.0034, "P": 0.0018, "Q": 0.001, "R": 0.0039, "S": 0.0001, "V": 0.0023 },
    "54": { "C": 0.0001, "D": 0.0002, "F": 0.003, "H": 0.0001, "I": 0.9689, "K": 0.0001, "L": 0.0066, "M": 0.0029, "N": 0.0031, "R": 0.0003, "S": 0.0071, "T": 0.0026, "V": 0.0049, "Y": 0.0001 },
    "55": { "A": 0.0003, "C": 0.0014, "D": 0.0069, "E": 0.0002, "F": 0.0136, "H": 0.0081, "K": 0.0001, "L": 0.0007, "N": 0.0023, "Q": 0.0002, "R": 0.0002, "S": 0.0129, "T": 0.0005, "V": 0.0002, "W": 0.0002, "Y": 0.9523 },
    "56": { "A": 0.5651, "C": 0.0003, "D": 0.3753, "E": 0.0059, "F": 0.0013, "G": 0.0203, "H": 0.0022, "I": 0.0001, "K": 0.0003, "L": 0.0007, "N": 0.0008, "P": 0.0009, "Q": 0.0001, "R": 0.0007, "S": 0.0089, "T": 0.0061, "V": 0.0045, "Y": 0.0066 },
    "57": { "A": 0.9609, "C": 0.0001, "D": 0.0003, "E": 0.0037, "G": 0.0037, "I": 0.0002, "K": 0.0001, "N": 0.0001, "P": 0.0007, "R": 0.0001, "S": 0.0092, "T": 0.0104, "V": 0.0106 },
    "65": { "A": 0.0085, "C": 0.0009, "D": 0.0001, "E": 0.0001, "F": 0.0051, "G": 0.0001, "H": 0.0001, "I": 0.0006, "K": 0.0001, "L": 0.0006, "N": 0.0002, "P": 0.0035, "S": 0.9666, "T": 0.0073, "V": 0.0005, "Y": 0.0059 },
    "66": { "A": 0.0007, "C": 0.0006, "D": 0.003, "E": 0.0003, "F": 0.0005, "G": 0.002, "H": 0.0041, "I": 0.0072, "K": 0.0095, "L": 0.0007, "M": 0.0003, "N": 0.316, "P": 0.0005, "Q": 0.0006, "R": 0.0115, "S": 0.5649, "T": 0.0752, "V": 0.0005, "Y": 0.0019 },
    "67": { "C": 0.0004, "E": 0.0001, "F": 0.0071, "G": 0.0002, "I": 0.0001, "L": 0.9669, "M": 0.0027, "P": 0.0003, "R": 0.0006, "S": 0.0037, "T": 0.0001, "V": 0.01, "W": 0.0078, "Y": 0.0001 },
    "68": { "A": 0.003, "D": 0.0038, "E": 0.3642, "F": 0.0002, "G": 0.0018, "H": 0.0079, "I": 0.0002, "K": 0.0085, "L": 0.0016, "N": 0.0006, "P": 0.0033, "Q": 0.5988, "R": 0.0034, "S": 0.0003, "T": 0.0003, "V": 0.0018, "Y": 0.0002 },
    "69": { "A": 0.0075, "C": 0.0004, "D": 0.0007, "E": 0.0009, "F": 0.0001, "G": 0.0076, "H": 0.0003, "I": 0.0058, "K": 0.0046, "L": 0.0003, "M": 0.0005, "N": 0.0054, "P": 0.0039, "Q": 0.0002, "R": 0.0153, "S": 0.6305, "T": 0.3148, "V": 0.0012, "Y": 0.0001 },
    "70": { "A": 0.0037, "C": 0.0003, "D": 0.0009, "E": 0.003, "G": 0.9748, "K": 0.0003, "L": 0.0001, "P": 0.0003, "R": 0.0047, "V": 0.0073, "W": 0.0046 },
    "71": { "A": 0.0052, "C": 0.0003, "D": 0.0013, "E": 0.0001, "F": 0.0035, "G": 0.0068, "I": 0.0031, "L": 0.0014, "P": 0.0001, "S": 0.0001, "V": 0.978 },
    "72": { "A": 0.0014, "D": 0.0001, "E": 0.0002, "H": 0.0003, "I": 0.0001, "K": 0.0002, "L": 0.0014, "P": 0.9787, "Q": 0.0038, "R": 0.0006, "S": 0.007, "T": 0.0061, "V": 0.0001 },
    "74": { "A": 0.0142, "C": 0.0001, "D": 0.0001, "E": 0.0005, "F": 0.0003, "G": 0.0003, "I": 0.0001, "K": 0.0001, "L": 0.0033, "P": 0.0092, "Q": 0.0001, "R": 0.0001, "S": 0.9671, "T": 0.0041, "V": 0.0001, "W": 0.0001, "Y": 0.0002 },
    "75": { "A": 0.0001, "E": 0.0001, "F": 0.0001, "G": 0.0026, "I": 0.0001, "K": 0.0045, "L": 0.0001, "M": 0.0068, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.9729, "S": 0.0082, "T": 0.0028, "W": 0.0014, "Y": 0.0001 },
    "76": { "C": 0.0129, "D": 0.0001, "F": 0.9671, "G": 0.0002, "I": 0.002, "L": 0.0076, "M": 0.0001, "S": 0.0014, "V": 0.0075, "Y": 0.001 },
    "77": { "A": 0.0003, "C": 0.0013, "D": 0.0001, "G": 0.0045, "H": 0.0001, "I": 0.0049, "K": 0.0001, "L": 0.0002, "M": 0.0001, "N": 0.0035, "R": 0.0148, "S": 0.9631, "T": 0.0063, "V": 0.0004, "Y": 0.0002 },
    "78": { "A": 0.0056, "C": 0.0072, "D": 0.0017, "E": 0.0008, "F": 0.0001, "G": 0.971, "R": 0.0029, "S": 0.0041, "T": 0.0001, "V": 0.0066 },
    "79": { "A": 0.0004, "C": 0.001, "D": 0.0012, "E": 0.0001, "F": 0.0002, "G": 0.0134, "H": 0.0003, "I": 0.0034, "K": 0.0001, "L": 0.0001, "M": 0.0004, "N": 0.0029, "Q": 0.0001, "R": 0.0157, "S": 0.9565, "T": 0.0037, "V": 0.0002, "W": 0.0001, "Y": 0.0001 },
    "80": { "A": 0.0048, "C": 0.0005, "D": 0.0003, "E": 0.0022, "G": 0.9727, "I": 0.0001, "K": 0.0004, "L": 0.0002, "P": 0.0001, "Q": 0.0012, "R": 0.01, "S": 0.0009, "T": 0.0001, "V": 0.0064, "W": 0.0002 },
    "83": { "A": 0.017, "C": 0.0029, "D": 0.0005, "E": 0.0001, "F": 0.0059, "G": 0.0003, "H": 0.0003, "L": 0.0001, "N": 0.0001, "P": 0.0018, "S": 0.9559, "T": 0.0023, "V": 0.0003, "Y": 0.0125 },
    "84": { "A": 0.0045, "C": 0.0003, "D": 0.0002, "E": 0.0091, "G": 0.9633, "K": 0.0001, "L": 0.0002, "M": 0.0001, "Q": 0.0004, "R": 0.0043, "S": 0.0003, "T": 0.0001, "V": 0.0099, "W": 0.0071, "Y": 0.0001 },
    "85": { "A": 0.0066, "E": 0.0004, "H": 0.0001, "I": 0.0023, "K": 0.0071, "L": 0.0001, "N": 0.0001, "P": 0.0018, "Q": 0.0001, "R": 0.0034, "S": 0.0022, "T": 0.9759 },
    "86": { "A": 0.0056, "C": 0.0002, "D": 0.9286, "E": 0.0257, "F": 0.0005, "G": 0.0028, "H": 0.0153, "I": 0.0002, "K": 0.0005, "L": 0.0005, "N": 0.0058, "P": 0.0001, "Q": 0.001, "R": 0.0005, "S": 0.0008, "T": 0.0012, "V": 0.003, "Y": 0.0077 },
    "87": { "A": 0.0001, "C": 0.0052, "D": 0.0002, "F": 0.9674, "G": 0.0002, "I": 0.0018, "L": 0.0054, "N": 0.0001, "R": 0.0001, "S": 0.0017, "T": 0.0001, "V": 0.0092, "Y": 0.0085 },
    "88": { "A": 0.006, "D": 0.0001, "E": 0.0001, "F": 0.0001, "H": 0.0001, "I": 0.0051, "K": 0.0001, "L": 0.0002, "M": 0.0001, "N": 0.0039, "P": 0.0046, "R": 0.0004, "S": 0.0146, "T": 0.9633, "V": 0.0012 },
    "89": { "C": 0.0014, "F": 0.3059, "G": 0.0001, "H": 0.0016, "I": 0.0073, "L": 0.6725, "M": 0.0001, "P": 0.002, "R": 0.0037, "S": 0.0012, "T": 0.0001, "V": 0.0034, "Y": 0.0007 },
    "90": { "A": 0.0046, "E": 0.0001, "F": 0.0003, "G": 0.0001, "H": 0.0001, "I": 0.0067, "K": 0.0003, "L": 0.0001, "N": 0.0063, "P": 0.0062, "R": 0.0003, "S": 0.009, "T": 0.9652, "V": 0.0005, "Y": 0.0001 },
    "91": { "F": 0.0009, "G": 0.0001, "H": 0.0001, "I": 0.979, "L": 0.0029, "M": 0.0009, "N": 0.0009, "R": 0.0003, "S": 0.0096, "T": 0.0021, "V": 0.0034 },
    "92": { "A": 0.0013, "C": 0.001, "D": 0.0024, "E": 0.0001, "G": 0.0033, "H": 0.0005, "I": 0.0066, "K": 0.0007, "L": 0.0008, "M": 0.0001, "N": 0.0163, "P": 0.0002, "Q": 0.0001, "R": 0.01, "S": 0.9376, "T": 0.0186, "V": 0.0001, "Y": 0.0004 },
    "93": { "A": 0.0004, "C": 0.0416, "D": 0.0018, "F": 0.0004, "G": 0.0141, "H": 0.0003, "I": 0.0053, "K": 0.0002, "L": 0.0001, "M": 0.0003, "N": 0.0127, "P": 0.0001, "Q": 0.0001, "R": 0.0184, "S": 0.9004, "T": 0.0031, "V": 0.0001, "W": 0.0001, "Y": 0.0005 },
    "94": { "A": 0.0001, "C": 0.0001, "F": 0.0002, "G": 0.0001, "I": 0.0001, "L": 0.9732, "M": 0.0047, "P": 0.0024, "Q": 0.001, "R": 0.0141, "T": 0.0003, "V": 0.0037, "Y": 0.0001 },
    "95": { "C": 0.0001, "D": 0.0001, "E": 0.0047, "H": 0.0079, "K": 0.0033, "L": 0.0027, "P": 0.0041, "Q": 0.9724, "R": 0.0045, "S": 0.0001, "T": 0.0001, "V": 0.0001 },
    "96": { "A": 0.0077, "C": 0.0001, "D": 0.0003, "E": 0.0002, "F": 0.0006, "G": 0.0002, "H": 0.0067, "K": 0.0001, "L": 0.0037, "N": 0.0002, "P": 0.9211, "Q": 0.0012, "R": 0.0026, "S": 0.0485, "T": 0.0056, "V": 0.0009, "Y": 0.0005 },
    "97": { "A": 0.0046, "D": 0.0131, "E": 0.9743, "G": 0.0039, "K": 0.0014, "L": 0.0002, "Q": 0.0008, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0012, "Y": 0.0003 },
    "98": { "A": 0.0032, "D": 0.976, "E": 0.0095, "F": 0.0001, "G": 0.0017, "H": 0.0005, "I": 0.0001, "K": 0.0001, "N": 0.0016, "R": 0.0001, "V": 0.0021, "Y": 0.005 },
    "99": { "A": 0.001, "C": 0.0053, "D": 0.0003, "F": 0.6404, "G": 0.0007, "H": 0.0001, "I": 0.2946, "L": 0.0123, "M": 0.0016, "N": 0.0008, "R": 0.0001, "S": 0.0106, "T": 0.0027, "V": 0.0276, "Y": 0.0019 },
    "100": { "A": 0.9663, "C": 0.0001, "D": 0.0001, "E": 0.0044, "G": 0.0153, "K": 0.0001, "L": 0.0001, "P": 0.0005, "Q": 0.0005, "S": 0.0061, "T": 0.0033, "V": 0.0033 },
    "101": { "A": 0.0081, "D": 0.0001, "E": 0.0001, "H": 0.0001, "I": 0.0063, "K": 0.0013, "L": 0.0007, "M": 0.0005, "N": 0.0037, "P": 0.0047, "Q": 0.0005, "R": 0.0003, "S": 0.0064, "T": 0.9636, "V": 0.0035, "Y": 0.0002 },
    "102": { "A": 0.0001, "C": 0.0024, "D": 0.0066, "E": 0.0003, "F": 0.0041, "H": 0.0014, "I": 0.0004, "L": 0.0003, "N": 0.0006, "S": 0.0031, "T": 0.0003, "V": 0.0002, "Y": 0.9802 },
    "103": { "A": 0.0001, "C": 0.0053, "D": 0.0051, "E": 0.0001, "F": 0.0192, "G": 0.0001, "H": 0.0046, "I": 0.0004, "K": 0.0001, "L": 0.0006, "N": 0.001, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0062, "T": 0.0005, "V": 0.0003, "W": 0.0002, "Y": 0.9561 },
    "104": { "A": 0.0002, "C": 0.9712, "F": 0.0033, "G": 0.0066, "H": 0.0002, "L": 0.001, "R": 0.0023, "S": 0.002, "T": 0.0002, "V": 0.0009, "W": 0.0094, "Y": 0.0028 },
    "105": { "E": 0.0014, "H": 0.0073, "K": 0.0061, "L": 0.0059, "M": 0.0003, "N": 0.0007, "P": 0.002, "Q": 0.9707, "R": 0.0049, "S": 0.0003, "T": 0.0001, "V": 0.0003, "Y": 0.0001 },
    "106": { "C": 0.0004, "D": 0.0005, "E": 0.0033, "F": 0.0001, "G": 0.0006, "H": 0.0181, "K": 0.0045, "L": 0.0033, "N": 0.0009, "P": 0.0039, "Q": 0.9548, "R": 0.0075, "S": 0.0004, "T": 0.0005, "V": 0.0003, "W": 0.0001, "Y": 0.0009 },
    "107": { "A": 0.0022, "C": 0.0074, "D": 0.0226, "E": 0.0053, "F": 0.0455, "G": 0.0078, "H": 0.007, "I": 0.0018, "K": 0.0005, "L": 0.0028, "M": 0.0003, "N": 0.0066, "P": 0.0003, "Q": 0.0005, "R": 0.0083, "S": 0.5173, "T": 0.0127, "V": 0.0024, "W": 0.0004, "Y": 0.3481 },
    "108": { "A": 0.0054, "C": 0.0007, "D": 0.3023, "E": 0.0093, "F": 0.0057, "G": 0.0052, "H": 0.0113, "I": 0.0009, "K": 0.0045, "L": 0.0018, "M": 0.0001, "N": 0.0607, "P": 0.0001, "Q": 0.0013, "R": 0.0007, "S": 0.0064, "T": 0.0011, "V": 0.0014, "W": 0.0003, "Y": 0.581 },
    "109": { "A": 0.002, "C": 0.0009, "D": 0.0142, "E": 0.0008, "F": 0.0005, "G": 0.0071, "H": 0.0037, "I": 0.0059, "K": 0.0064, "L": 0.001, "M": 0.0007, "N": 0.2949, "P": 0.0001, "Q": 0.0005, "R": 0.013, "S": 0.6217, "T": 0.0204, "V": 0.0026, "W": 0.0001, "Y": 0.0034 },
    "110": { "A": 0.0051, "D": 0.0187, "F": 0.0187, "G": 0.0034, "H": 0.0068, "I": 0.0085, "L": 0.4242, "M": 0.0017, "N": 0.0017, "P": 0.0051, "R": 0.017, "S": 0.0068, "T": 0.356, "V": 0.017, "W": 0.0034, "Y": 0.1056 },
    "113": { "A": 0.0046, "C": 0.0006, "D": 0.0014, "E": 0.0003, "F": 0.0124, "G": 0.0049, "H": 0.0064, "I": 0.0055, "K": 0.0014, "L": 0.2699, "M": 0.0012, "N": 0.0067, "P": 0.1467, "Q": 0.0009, "R": 0.0139, "S": 0.0124, "T": 0.4605, "V": 0.0104, "W": 0.0009, "Y": 0.0391 },
    "114": { "A": 0.0069, "C": 0.0008, "D": 0.0027, "E": 0.0007, "F": 0.0522, "G": 0.0036, "H": 0.0082, "I": 0.0083, "K": 0.0011, "L": 0.2068, "M": 0.0011, "N": 0.037, "P": 0.1821, "Q": 0.0034, "R": 0.0138, "S": 0.0605, "T": 0.3587, "V": 0.0095, "W": 0.0005, "Y": 0.0418 },
    "115": { "A": 0.018, "C": 0.0007, "D": 0.0024, "E": 0.0057, "F": 0.0061, "G": 0.0163, "H": 0.0043, "I": 0.007, "K": 0.0025, "L": 0.111, "M": 0.0189, "N": 0.0048, "P": 0.6427, "Q": 0.0156, "R": 0.0374, "S": 0.0381, "T": 0.0506, "V": 0.0084, "W": 0.003, "Y": 0.0066 },
    "116": { "A": 0.0039, "C": 0.0148, "D": 0.0066, "E": 0.0036, "F": 0.0909, "G": 0.0218, "H": 0.0158, "I": 0.0778, "K": 0.0041, "L": 0.2344, "M": 0.0007, "N": 0.0083, "P": 0.0948, "Q": 0.0193, "R": 0.108, "S": 0.0139, "T": 0.0049, "V": 0.0298, "W": 0.0998, "Y": 0.1469 },
    "117": { "A": 0.0209, "C": 0.0005, "D": 0.0009, "E": 0.0006, "F": 0.0028, "G": 0.0054, "H": 0.0027, "I": 0.0029, "K": 0.0005, "L": 0.0099, "M": 0.0005, "N": 0.0031, "P": 0.0113, "Q": 0.0006, "R": 0.0041, "S": 0.0412, "T": 0.8887, "V": 0.0029, "W": 0.0001, "Y": 0.0006 },
    "118": { "A": 0.0002, "C": 0.0073, "D": 0.0001, "F": 0.9772, "G": 0.0002, "H": 0.0001, "I": 0.0005, "K": 0.0001, "L": 0.0043, "N": 0.0001, "P": 0.0006, "Q": 0.0001, "R": 0.0002, "S": 0.0015, "T": 0.0001, "V": 0.0062, "W": 0.0002, "Y": 0.001 },
    "119": { "A": 0.0019, "C": 0.0011, "D": 0.0019, "E": 0.0001, "G": 0.99, "H": 0.0002, "I": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.001, "S": 0.002, "T": 0.0001, "V": 0.0014 },
    "120": { "A": 0.001, "E": 0.0035, "G": 0.2734, "H": 0.0123, "K": 0.0012, "L": 0.0016, "N": 0.0001, "P": 0.142, "Q": 0.5566, "R": 0.0074, "S": 0.0002, "T": 0.0004, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "121": { "A": 0.0005, "E": 0.0039, "G": 0.9866, "K": 0.0001, "L": 0.0001, "P": 0.0001, "R": 0.0039, "S": 0.0001, "T": 0.0001, "V": 0.002, "W": 0.0028 },
    "122": { "A": 0.0024, "C": 0.0001, "I": 0.0012, "K": 0.0005, "N": 0.001, "P": 0.0018, "R": 0.0003, "S": 0.0016, "T": 0.991 },
    "123": { "A": 0.0001, "D": 0.0001, "E": 0.0019, "G": 0.0001, "H": 0.0003, "I": 0.0009, "K": 0.8643, "L": 0.001, "M": 0.0005, "N": 0.0046, "P": 0.0001, "Q": 0.007, "R": 0.1142, "S": 0.0002, "T": 0.0043, "V": 0.0003, "W": 0.0003, "Y": 0.0001 },
    "124": { "A": 0.0005, "E": 0.0007, "G": 0.0007, "I": 0.0014, "L": 0.3377, "M": 0.0033, "P": 0.0002, "Q": 0.0001, "R": 0.0001, "V": 0.6552 },
    "125": { "A": 0.0015, "D": 0.1536, "E": 0.8301, "F": 0.0001, "G": 0.003, "H": 0.0012, "I": 0.0001, "K": 0.0018, "L": 0.0001, "M": 0.0001, "N": 0.0019, "Q": 0.0042, "R": 0.0007, "S": 0.0001, "T": 0.0003, "V": 0.001, "Y": 0.0003 },
    "126": { "A": 0.0001, "F": 0.0025, "I": 0.9633, "L": 0.0075, "M": 0.0061, "N": 0.0009, "Q": 0.0001, "R": 0.0001, "S": 0.0047, "T": 0.0027, "V": 0.0122 },
    "127": { "E": 0.003, "K": 0.9934, "N": 0.0007, "Q": 0.0029, "T": 0.0001 }
  },
  "IGKV2": {
    "1": { "D": 0.9633, "E": 0.0367 },
    "2": { "A": 0.0048, "C": 0.0001, "D": 0.0012, "F": 0.0029, "G": 0.0034, "H": 0.0003, "I": 0.6991, "K": 0.0001, "L": 0.0215, "M": 0.0135, "N": 0.0019, "P": 0.0001, "R": 0.0011, "S": 0.0134, "T": 0.0078, "V": 0.2286, "Y": 0.0001 },
    "3": { "A": 0.008, "C": 0.0001, "D": 0.0001, "E": 0.0046, "F": 0.0007, "G": 0.0102, "I": 0.0005, "K": 0.0001, "L": 0.0081, "M": 0.0022, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0001, "V": 0.9646, "W": 0.0003 },
    "4": { "C": 0.0001, "D": 0.0001, "E": 0.0001, "F": 0.0004, "G": 0.0001, "I": 0.0141, "K": 0.0009, "L": 0.0514, "M": 0.9093, "P": 0.0002, "Q": 0.0001, "R": 0.0087, "S": 0.0001, "T": 0.0025, "V": 0.0119, "W": 0.0001 },
    "5": { "A": 0.0038, "C": 0.0001, "D": 0.0002, "F": 0.0003, "G": 0.0001, "H": 0.0002, "I": 0.0046, "K": 0.0001, "L": 0.0003, "M": 0.0001, "N": 0.0055, "P": 0.0168, "R": 0.0002, "S": 0.0074, "T": 0.9602, "V": 0.0001, "Y": 0.0001 },
    "6": { "A": 0.0001, "E": 0.0026, "G": 0.0001, "H": 0.0116, "K": 0.0062, "L": 0.0026, "M": 0.0002, "N": 0.0002, "P": 0.0146, "Q": 0.9588, "R": 0.0022, "S": 0.0004, "T": 0.0004, "W": 0.0001 },
    "7": { "A": 0.0029, "C": 0.0011, "D": 0.0001, "F": 0.0057, "H": 0.0002, "I": 0.0006, "L": 0.0005, "N": 0.0003, "P": 0.0084, "R": 0.0001, "S": 0.9038, "T": 0.0698, "V": 0.0003, "Y": 0.0062 },
    "8": { "A": 0.0027, "E": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0005, "I": 0.0001, "K": 0.0001, "L": 0.009, "M": 0.0001, "P": 0.9587, "Q": 0.0111, "R": 0.0036, "S": 0.0041, "T": 0.0097, "V": 0.0001 },
    "9": { "A": 0.0005, "D": 0.0009, "F": 0.003, "G": 0.0001, "H": 0.003, "I": 0.0034, "L": 0.9573, "M": 0.0001, "P": 0.0167, "Q": 0.0001, "R": 0.007, "S": 0.0004, "T": 0.0002, "V": 0.0072, "Y": 0.0001 },
    "10": { "A": 0.0052, "C": 0.0016, "D": 0.0001, "E": 0.0001, "F": 0.0072, "H": 0.0009, "I": 0.0001, "L": 0.0013, "P": 0.02, "S": 0.9513, "T": 0.0042, "V": 0.0003, "Y": 0.0077 },
    "11": { "A": 0.0003, "F": 0.0003, "G": 0.0001, "H": 0.0001, "I": 0.0005, "K": 0.0001, "L": 0.9483, "M": 0.0073, "P": 0.0038, "Q": 0.002, "R": 0.0036, "S": 0.0317, "T": 0.0002, "V": 0.0018 },
    "12": { "A": 0.0086, "D": 0.0001, "F": 0.0003, "H": 0.0132, "I": 0.0002, "L": 0.0114, "N": 0.0001, "P": 0.8961, "Q": 0.0001, "R": 0.003, "S": 0.0583, "T": 0.0075, "V": 0.0008, "Y": 0.0003 },
    "13": { "A": 0.0059, "C": 0.0001, "D": 0.0009, "F": 0.0036, "G": 0.003, "I": 0.0016, "L": 0.0022, "P": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.9824 },
    "14": { "A": 0.004, "D": 0.0001, "G": 0.0001, "H": 0.0003, "I": 0.0057, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0032, "P": 0.0109, "R": 0.0003, "S": 0.0097, "T": 0.9654, "Y": 0.0001 },
    "15": { "A": 0.0016, "F": 0.0007, "G": 0.0001, "H": 0.004, "I": 0.0013, "L": 0.2588, "M": 0.0001, "P": 0.7198, "Q": 0.0002, "R": 0.003, "S": 0.0034, "T": 0.0049, "V": 0.0022, "Y": 0.0001 },
    "16": { "A": 0.002, "C": 0.0004, "D": 0.0002, "E": 0.0013, "G": 0.9882, "R": 0.0028, "S": 0.0001, "V": 0.0047, "W": 0.0001 },
    "17": { "A": 0.0133, "D": 0.0047, "E": 0.6677, "G": 0.0036, "H": 0.002, "K": 0.0023, "L": 0.0006, "N": 0.0001, "P": 0.0027, "Q": 0.2991, "R": 0.0016, "S": 0.0001, "T": 0.0001, "V": 0.0022, "Y": 0.0001 },
    "18": { "A": 0.005, "G": 0.0001, "L": 0.0025, "P": 0.9666, "Q": 0.0061, "R": 0.0026, "S": 0.0101, "T": 0.0069, "W": 0.0001 },
    "19": { "A": 0.9808, "D": 0.0034, "G": 0.0008, "H": 0.0001, "P": 0.0042, "R": 0.0001, "S": 0.0057, "T": 0.0028, "V": 0.0022 },
    "20": { "A": 0.0068, "C": 0.0006, "D": 0.0001, "F": 0.0021, "H": 0.0003, "L": 0.0001, "P": 0.0071, "S": 0.9751, "T": 0.003, "V": 0.0001, "Y": 0.0047 },
    "21": { "F": 0.003, "H": 0.0001, "I": 0.9595, "L": 0.0143, "M": 0.0036, "N": 0.0019, "P": 0.0001, "R": 0.0002, "S": 0.0051, "T": 0.0097, "V": 0.0026 },
    "22": { "A": 0.0104, "C": 0.0005, "D": 0.0003, "F": 0.0017, "H": 0.0005, "I": 0.0001, "K": 0.0001, "L": 0.0005, "N": 0.0001, "P": 0.0354, "Q": 0.0001, "R": 0.0001, "S": 0.9397, "T": 0.0053, "V": 0.0001, "Y": 0.0053 },
    "23": { "A": 0.0003, "C": 0.9792, "D": 0.0001, "F": 0.0059, "G": 0.0034, "L": 0.0001, "R": 0.0032, "S": 0.0057, "V": 0.0002, "W": 0.0003, "Y": 0.0016 },
    "24": { "A": 0.0001, "C": 0.0001, "E": 0.0004, "F": 0.0001, "G": 0.0025, "H": 0.0001, "I": 0.0003, "K": 0.0429, "L": 0.0001, "M": 0.0044, "N": 0.0018, "P": 0.0003, "Q": 0.0009, "R": 0.9219, "S": 0.0117, "T": 0.0094, "V": 0.0007, "W": 0.0023 },
    "25": { "A": 0.0099, "C": 0.0009, "D": 0.0001, "F": 0.0021, "H": 0.0001, "L": 0.0003, "P": 0.011, "R": 0.0001, "S": 0.9701, "T": 0.0022, "V": 0.0001, "W": 0.0001, "Y": 0.0031 },
    "26": { "A": 0.0004, "C": 0.001, "D": 0.0008, "G": 0.0039, "H": 0.0001, "I": 0.0035, "K": 0.0001, "L": 0.0003, "M": 0.0002, "N": 0.0057, "P": 0.0004, "Q": 0.0001, "R": 0.0185, "S": 0.9575, "T": 0.0073, "V": 0.0001, "Y": 0.0001 },
    "27": { "A": 0.0004, "D": 0.0001, "E": 0.0084, "H": 0.0119, "I": 0.0001, "K": 0.0047, "L": 0.0013, "N": 0.0001, "P": 0.0103, "Q": 0.9563, "R": 0.0058, "S": 0.0003, "T": 0.0002, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "28": { "A": 0.0003, "C": 0.0032, "D": 0.0003, "F": 0.0001, "G": 0.0091, "H": 0.0003, "I": 0.0042, "K": 0.0001, "L": 0.0001, "N": 0.0057, "P": 0.0001, "R": 0.0235, "S": 0.948, "T": 0.0052, "V": 0.0001 },
    "29": { "C": 0.0001, "F": 0.003, "H": 0.0019, "I": 0.0055, "L": 0.9692, "P": 0.0102, "R": 0.0074, "S": 0.0003, "V": 0.0025 },
    "30": { "A": 0.0032, "D": 0.0001, "E": 0.0024, "F": 0.0008, "G": 0.0013, "H": 0.0006, "I": 0.0027, "K": 0.0001, "L": 0.7268, "M": 0.0044, "P": 0.0028, "Q": 0.0021, "R": 0.0056, "S": 0.0005, "T": 0.0004, "V": 0.2461, "W": 0.0001 },
    "31": { "A": 0.0009, "C": 0.0004, "D": 0.0051, "E": 0.0008, "F": 0.0038, "G": 0.0005, "H": 0.763, "I": 0.0001, "K": 0.0011, "L": 0.0033, "N": 0.0109, "P": 0.0053, "Q": 0.0095, "R": 0.0044, "S": 0.0109, "T": 0.0005, "V": 0.0003, "W": 0.0003, "Y": 0.1789 },
    "32": { "A": 0.0005, "C": 0.0007, "D": 0.0013, "E": 0.0002, "F": 0.0007, "G": 0.0093, "H": 0.001, "I": 0.0052, "K": 0.0014, "L": 0.0008, "M": 0.0003, "N": 0.0091, "P": 0.0011, "Q": 0.0004, "R": 0.0284, "S": 0.9221, "T": 0.0158, "V": 0.0009, "W": 0.0001, "Y": 0.0007 },
    "34": { "A": 0.0032, "D": 0.2945, "E": 0.0018, "G": 0.0026, "H": 0.0042, "I": 0.003, "K": 0.0035, "L": 0.0002, "M": 0.0001, "N": 0.6617, "P": 0.0001, "R": 0.0006, "S": 0.0078, "T": 0.0123, "V": 0.0013, "Y": 0.003 },
    "35": { "A": 0.0029, "C": 0.0003, "D": 0.0003, "E": 0.0057, "G": 0.9769, "K": 0.0005, "L": 0.0001, "N": 0.0002, "P": 0.0001, "Q": 0.0001, "R": 0.0063, "S": 0.0002, "T": 0.0004, "V": 0.0057, "W": 0.0002 },
    "36": { "A": 0.0001, "C": 0.0009, "D": 0.0107, "E": 0.0006, "F": 0.0064, "G": 0.0005, "H": 0.0144, "I": 0.002, "K": 0.0412, "L": 0.0009, "M": 0.0001, "N": 0.2571, "P": 0.0001, "Q": 0.0018, "R": 0.0022, "S": 0.0106, "T": 0.0039, "V": 0.0007, "Y": 0.6459 },
    "37": { "A": 0.0016, "D": 0.0034, "E": 0.0001, "G": 0.0001, "H": 0.0101, "I": 0.0113, "K": 0.0056, "L": 0.0008, "M": 0.0003, "N": 0.6492, "P": 0.0022, "Q": 0.0006, "R": 0.0003, "S": 0.0073, "T": 0.3014, "V": 0.0005, "Y": 0.0052 },
    "38": { "A": 0.0002, "C": 0.0049, "D": 0.0079, "E": 0.0001, "F": 0.0146, "G": 0.0001, "H": 0.0109, "I": 0.0004, "K": 0.0001, "L": 0.0014, "N": 0.0029, "P": 0.0003, "Q": 0.0005, "R": 0.0005, "S": 0.0134, "T": 0.0003, "V": 0.0008, "W": 0.0004, "Y": 0.9403 },
    "39": { "A": 0.0001, "C": 0.0003, "F": 0.0138, "I": 0.0001, "L": 0.9596, "M": 0.0027, "R": 0.0001, "S": 0.003, "V": 0.0158, "W": 0.0046 },
    "40": { "A": 0.0095, "C": 0.001, "D": 0.6564, "E": 0.0124, "F": 0.0019, "G": 0.0024, "H": 0.0136, "I": 0.0026, "K": 0.0014, "L": 0.0003, "M": 0.0003, "N": 0.2057, "P": 0.0002, "Q": 0.0008, "R": 0.0009, "S": 0.0419, "T": 0.0063, "V": 0.0026, "Y": 0.0397 },
    "41": { "C": 0.0057, "G": 0.0042, "I": 0.0001, "L": 0.0048, "R": 0.0025, "S": 0.0013, "V": 0.0001, "W": 0.9813, "Y": 0.0001 },
    "42": { "C": 0.0016, "D": 0.0023, "E": 0.0001, "F": 0.2268, "G": 0.0001, "H": 0.0018, "I": 0.0016, "L": 0.0356, "N": 0.0007, "P": 0.0001, "R": 0.0003, "S": 0.0044, "V": 0.0022, "Y": 0.7225 },
    "43": { "E": 0.0002, "F": 0.0011, "G": 0.0001, "H": 0.0253, "I": 0.001, "K": 0.001, "L": 0.713, "M": 0.0026, "N": 0.0001, "P": 0.0025, "Q": 0.2375, "R": 0.0041, "S": 0.0001, "T": 0.0001, "V": 0.0108, "W": 0.0001, "Y": 0.0003 },
    "44": { "E": 0.0007, "H": 0.0052, "K": 0.0059, "L": 0.0023, "N": 0.0001, "P": 0.0032, "Q": 0.9805, "R": 0.0018, "S": 0.0002, "T": 0.0001, "V": 0.0001, "Y": 0.0001 },
    "45": { "C": 0.0001, "E": 0.0012, "G": 0.0012, "H": 0.0001, "K": 0.7086, "L": 0.0001, "M": 0.0017, "N": 0.0036, "P": 0.0001, "Q": 0.0043, "R": 0.2712, "S": 0.0025, "T": 0.005, "W": 0.0004 },
    "46": { "A": 0.003, "E": 0.0001, "G": 0.0001, "H": 0.0005, "L": 0.0012, "P": 0.9769, "Q": 0.0072, "R": 0.001, "S": 0.0036, "T": 0.0061, "V": 0.0001, "Y": 0.0001 },
    "47": { "A": 0.0012, "C": 0.0016, "D": 0.0009, "E": 0.0014, "G": 0.9836, "R": 0.0018, "S": 0.0008, "T": 0.0001, "V": 0.0062, "W": 0.0022, "Y": 0.0001 },
    "48": { "A": 0.0001, "E": 0.0009, "H": 0.0075, "K": 0.008, "L": 0.0015, "M": 0.0001, "N": 0.0001, "P": 0.0045, "Q": 0.9731, "R": 0.0038, "S": 0.0001, "T": 0.0003, "V": 0.0001 },
    "49": { "A": 0.0138, "C": 0.0008, "D": 0.0003, "F": 0.0022, "H": 0.0001, "I": 0.0001, "L": 0.0006, "P": 0.0417, "S": 0.9339, "T": 0.0023, "V": 0.0001, "Y": 0.0043 },
    "50": { "A": 0.0011, "H": 0.0001, "K": 0.0001, "L": 0.0014, "P": 0.9823, "Q": 0.0092, "R": 0.001, "S": 0.0015, "T": 0.0032 },
    "51": { "A": 0.0001, "E": 0.0013, "G": 0.0005, "H": 0.0092, "I": 0.0001, "K": 0.0036, "L": 0.0011, "M": 0.0007, "N": 0.0003, "P": 0.0048, "Q": 0.7086, "R": 0.2659, "S": 0.003, "T": 0.0004, "W": 0.0003 },
    "52": { "A": 0.0002, "C": 0.0008, "F": 0.0018, "G": 0.0003, "H": 0.0026, "I": 0.0038, "L": 0.7438, "P": 0.0049, "R": 0.2335, "S": 0.0016, "T": 0.0001, "V": 0.0064, "W": 0.0001, "Y": 0.0001 },
    "53": { "F": 0.0001, "I": 0.0016, "L": 0.9892, "M": 0.0024, "P": 0.0013, "Q": 0.0005, "R": 0.0035, "T": 0.0001, "V": 0.0013 },
    "54": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.0084, "G": 0.0002, "H": 0.0001, "I": 0.9506, "L": 0.0147, "M": 0.0068, "N": 0.0015, "Q": 0.0001, "R": 0.0002, "S": 0.0079, "T": 0.0028, "V": 0.0062, "Y": 0.0002 },
    "55": { "C": 0.0024, "D": 0.0107, "E": 0.0002, "F": 0.0075, "H": 0.0063, "I": 0.0001, "K": 0.0002, "L": 0.0003, "N": 0.0018, "P": 0.0001, "Q": 0.0005, "R": 0.0001, "S": 0.011, "T": 0.0001, "W": 0.0001, "Y": 0.9587 },
    "56": { "A": 0.0005, "C": 0.0003, "D": 0.0013, "E": 0.0398, "F": 0.0098, "G": 0.0013, "H": 0.0003, "I": 0.0002, "K": 0.2332, "L": 0.6595, "M": 0.0068, "N": 0.0033, "P": 0.0001, "Q": 0.0094, "R": 0.0092, "S": 0.0034, "T": 0.0026, "V": 0.0099, "W": 0.0085, "Y": 0.0005 },
    "57": { "A": 0.0171, "C": 0.0038, "D": 0.0016, "E": 0.0004, "F": 0.0028, "G": 0.6633, "H": 0.0001, "I": 0.0341, "K": 0.0001, "L": 0.002, "M": 0.0003, "N": 0.0007, "R": 0.0009, "S": 0.004, "T": 0.0026, "V": 0.2661, "Y": 0.0001 },
    "65": { "A": 0.0093, "C": 0.0011, "D": 0.0003, "F": 0.0055, "G": 0.0001, "H": 0.0001, "L": 0.0002, "N": 0.0001, "P": 0.0015, "R": 0.0001, "S": 0.9762, "T": 0.0022, "V": 0.0001, "W": 0.0001, "Y": 0.0032 },
    "66": { "A": 0.0009, "C": 0.0003, "D": 0.01, "E": 0.0011, "F": 0.0006, "G": 0.0009, "H": 0.0131, "I": 0.0085, "K": 0.0253, "L": 0.0009, "M": 0.0002, "N": 0.8522, "P": 0.0001, "Q": 0.0012, "R": 0.0057, "S": 0.0479, "T": 0.0235, "V": 0.0005, "W": 0.0001, "Y": 0.0069 },
    "67": { "C": 0.0001, "F": 0.0001, "G": 0.0017, "H": 0.0002, "L": 0.0059, "P": 0.0013, "Q": 0.002, "R": 0.9873, "S": 0.0002, "V": 0.0001, "W": 0.0012 },
    "68": { "A": 0.6765, "C": 0.0013, "D": 0.2277, "E": 0.0053, "F": 0.0646, "G": 0.004, "H": 0.0004, "I": 0.0003, "K": 0.0002, "L": 0.0027, "M": 0.0001, "N": 0.0006, "P": 0.0022, "R": 0.0001, "S": 0.0071, "T": 0.0015, "V": 0.0035, "Y": 0.0018 },
    "69": { "A": 0.0234, "C": 0.0036, "D": 0.0004, "F": 0.0027, "G": 0.0004, "H": 0.0006, "I": 0.0001, "L": 0.0002, "N": 0.0001, "P": 0.0103, "S": 0.943, "T": 0.0035, "V": 0.0005, "W": 0.0002, "Y": 0.0109 },
    "70": { "A": 0.0019, "C": 0.0001, "D": 0.0001, "E": 0.002, "G": 0.9791, "L": 0.0001, "P": 0.0002, "R": 0.0041, "S": 0.0002, "T": 0.0001, "V": 0.0059, "W": 0.0061 },
    "71": { "A": 0.004, "C": 0.0001, "D": 0.0009, "F": 0.0048, "G": 0.0117, "I": 0.0028, "L": 0.0012, "M": 0.0001, "N": 0.0001, "S": 0.0001, "V": 0.974 },
    "72": { "A": 0.0033, "F": 0.0001, "H": 0.0033, "I": 0.0001, "K": 0.0001, "L": 0.0013, "N": 0.0001, "P": 0.9796, "Q": 0.0018, "R": 0.0007, "S": 0.0043, "T": 0.0053, "V": 0.0001 },
    "74": { "A": 0.0053, "C": 0.0001, "D": 0.9661, "E": 0.0116, "F": 0.0001, "G": 0.004, "H": 0.0016, "K": 0.0001, "N": 0.0024, "P": 0.0001, "S": 0.0006, "T": 0.0002, "V": 0.002, "Y": 0.0058 },
    "75": { "C": 0.0001, "D": 0.0001, "E": 0.0001, "G": 0.002, "I": 0.0013, "K": 0.0028, "L": 0.0001, "M": 0.0034, "R": 0.9788, "S": 0.009, "T": 0.0012, "W": 0.0011 },
    "76": { "C": 0.0168, "F": 0.9515, "G": 0.0001, "I": 0.0059, "L": 0.0096, "P": 0.0001, "R": 0.0001, "S": 0.0013, "V": 0.0127, "W": 0.0003, "Y": 0.0016 },
    "77": { "A": 0.0007, "C": 0.0012, "D": 0.0001, "F": 0.0001, "G": 0.0038, "H": 0.0001, "I": 0.0039, "L": 0.0001, "M": 0.0002, "N": 0.0024, "R": 0.0131, "S": 0.9701, "T": 0.0036, "V": 0.0005 },
    "78": { "A": 0.0059, "C": 0.0054, "D": 0.002, "E": 0.0001, "F": 0.0001, "G": 0.9776, "I": 0.0001, "R": 0.0013, "S": 0.0036, "T": 0.0001, "V": 0.0038, "W": 0.0001, "Y": 0.0001 },
    "79": { "A": 0.0003, "C": 0.0006, "D": 0.0001, "F": 0.0001, "G": 0.0059, "H": 0.0001, "I": 0.0038, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0011, "P": 0.0001, "R": 0.0164, "S": 0.9684, "T": 0.0025, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "80": { "A": 0.0049, "C": 0.0003, "D": 0.0003, "E": 0.0042, "F": 0.0001, "G": 0.9786, "I": 0.0001, "L": 0.0001, "N": 0.0001, "Q": 0.0001, "R": 0.0042, "S": 0.0001, "T": 0.0001, "V": 0.005, "W": 0.0018 },
    "83": { "A": 0.0497, "C": 0.0001, "E": 0.0006, "F": 0.0001, "G": 0.0002, "I": 0.0001, "L": 0.002, "P": 0.0015, "R": 0.0001, "S": 0.942, "T": 0.0034, "V": 0.0001, "Y": 0.0002 },
    "84": { "A": 0.003, "C": 0.0038, "D": 0.0057, "E": 0.0007, "F": 0.0001, "G": 0.9783, "H": 0.0005, "M": 0.0001, "P": 0.0001, "R": 0.001, "S": 0.0018, "T": 0.0002, "V": 0.0044, "W": 0.0003 },
    "85": { "A": 0.0051, "E": 0.0001, "F": 0.0001, "G": 0.0001, "I": 0.002, "K": 0.0044, "M": 0.0001, "N": 0.0017, "P": 0.0021, "R": 0.0022, "S": 0.0044, "T": 0.9778, "V": 0.0002 },
    "86": { "A": 0.0026, "C": 0.0001, "D": 0.9556, "E": 0.0184, "F": 0.0001, "G": 0.0025, "H": 0.0029, "I": 0.0002, "K": 0.0003, "L": 0.0004, "N": 0.007, "Q": 0.0004, "R": 0.0001, "S": 0.0001, "T": 0.0003, "V": 0.0013, "Y": 0.0077 },
    "87": { "C": 0.0059, "D": 0.0001, "F": 0.9734, "G": 0.0002, "I": 0.0009, "L": 0.0049, "S": 0.0018, "V": 0.011, "W": 0.0001, "Y": 0.0017 },
    "88": { "A": 0.0027, "F": 0.0001, "G": 0.0001, "I": 0.0048, "K": 0.0025, "M": 0.0004, "N": 0.0001, "P": 0.0028, "Q": 0.0001, "R": 0.0008, "S": 0.002, "T": 0.9836, "V": 0.0001 },
    "89": { "E": 0.0001, "F": 0.0001, "I": 0.0002, "L": 0.9854, "M": 0.0063, "P": 0.0018, "Q": 0.0007, "R": 0.004, "S": 0.0001, "T": 0.0002, "V": 0.0009, "W": 0.0001 },
    "90": { "A": 0.0001, "D": 0.0003, "E": 0.008, "G": 0.0003, "H": 0.0009, "I": 0.0028, "K": 0.9279, "L": 0.0001, "M": 0.0006, "N": 0.0133, "Q": 0.0061, "R": 0.0202, "S": 0.0003, "T": 0.0185, "V": 0.0003, "W": 0.0003, "Y": 0.0001 },
    "91": { "F": 0.0011, "G": 0.0001, "I": 0.9709, "L": 0.0036, "M": 0.0006, "N": 0.0011, "R": 0.0003, "S": 0.018, "T": 0.0013, "V": 0.0027 },
    "92": { "A": 0.0001, "C": 0.0004, "D": 0.0002, "G": 0.0022, "H": 0.0001, "I": 0.0042, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0063, "R": 0.0068, "S": 0.9721, "T": 0.0069 },
    "93": { "A": 0.0004, "E": 0.0003, "G": 0.0046, "I": 0.0041, "K": 0.0057, "L": 0.0004, "M": 0.0013, "N": 0.0002, "P": 0.0001, "Q": 0.0001, "R": 0.9687, "S": 0.0112, "T": 0.002, "W": 0.0008 },
    "94": { "A": 0.0009, "D": 0.0002, "E": 0.0007, "F": 0.0001, "G": 0.009, "I": 0.0003, "L": 0.0035, "M": 0.0018, "V": 0.9834, "W": 0.0002 },
    "95": { "A": 0.0023, "D": 0.0055, "E": 0.9831, "G": 0.0023, "H": 0.0001, "K": 0.0014, "M": 0.0001, "N": 0.0001, "Q": 0.0032, "R": 0.0003, "S": 0.0001, "T": 0.0001, "V": 0.0013, "W": 0.0001 },
    "96": { "A": 0.9747, "D": 0.0059, "E": 0.0006, "G": 0.0011, "H": 0.0001, "L": 0.0005, "P": 0.0067, "S": 0.0036, "T": 0.0042, "V": 0.0024 },
    "97": { "A": 0.0031, "D": 0.0198, "E": 0.9679, "G": 0.0052, "H": 0.0001, "K": 0.0007, "Q": 0.0008, "R": 0.0005, "V": 0.0016, "W": 0.0001, "Y": 0.0002 },
    "98": { "A": 0.0006, "D": 0.9856, "E": 0.0044, "G": 0.001, "H": 0.0007, "I": 0.0001, "K": 0.0001, "M": 0.0003, "N": 0.0016, "P": 0.0001, "V": 0.0005, "Y": 0.0049 },
    "99": { "A": 0.0051, "C": 0.0001, "D": 0.0006, "F": 0.0069, "G": 0.0038, "I": 0.0075, "L": 0.0021, "N": 0.0001, "S": 0.0002, "T": 0.0002, "V": 0.9733, "W": 0.0001 },
    "100": { "A": 0.0068, "C": 0.0005, "E": 0.0013, "F": 0.0001, "G": 0.9743, "K": 0.0001, "L": 0.0002, "P": 0.0001, "R": 0.0024, "S": 0.0001, "V": 0.0084, "W": 0.0059 },
    "101": { "A": 0.0015, "D": 0.0017, "F": 0.0083, "G": 0.003, "I": 0.0368, "L": 0.0093, "M": 0.0005, "N": 0.0003, "R": 0.0001, "S": 0.0002, "T": 0.0054, "V": 0.9327, "Y": 0.0001 },
    "102": { "C": 0.0009, "D": 0.004, "E": 0.0002, "F": 0.0033, "H": 0.0005, "I": 0.0005, "K": 0.0001, "L": 0.0005, "N": 0.0007, "S": 0.003, "Y": 0.9865 },
    "103": { "A": 0.0002, "C": 0.0044, "D": 0.0028, "F": 0.0133, "H": 0.0019, "I": 0.0001, "L": 0.0006, "N": 0.0004, "P": 0.0001, "S": 0.005, "T": 0.0005, "V": 0.0001, "Y": 0.9706 },
    "104": { "A": 0.0009, "C": 0.984, "D": 0.0001, "F": 0.0028, "G": 0.0054, "L": 0.0008, "N": 0.0001, "R": 0.002, "S": 0.0009, "T": 0.0001, "V": 0.0002, "W": 0.0008, "Y": 0.002 },
    "105": { "A": 0.0001, "C": 0.0007, "E": 0.0001, "F": 0.0007, "G": 0.0003, "H": 0.0005, "I": 0.0046, "K": 0.0053, "L": 0.0069, "M": 0.9355, "N": 0.0003, "P": 0.0001, "Q": 0.0011, "R": 0.0342, "S": 0.0014, "T": 0.004, "V": 0.0038, "W": 0.0004 },
    "106": { "A": 0.0007, "C": 0.0001, "E": 0.0036, "G": 0.0003, "H": 0.0051, "I": 0.0003, "K": 0.0039, "L": 0.0013, "P": 0.0024, "Q": 0.9764, "R": 0.0047, "T": 0.0001, "V": 0.0009, "W": 0.0001 },
    "107": { "A": 0.6652, "C": 0.0031, "D": 0.0134, "E": 0.001, "F": 0.0005, "G": 0.2621, "H": 0.0017, "I": 0.0016, "K": 0.0002, "L": 0.0008, "N": 0.0018, "P": 0.0052, "Q": 0.0003, "R": 0.002, "S": 0.0192, "T": 0.0126, "V": 0.0067, "W": 0.0001, "Y": 0.0026 },
    "108": { "A": 0.0057, "D": 0.0001, "E": 0.0016, "F": 0.0005, "G": 0.0009, "H": 0.0006, "I": 0.044, "K": 0.0084, "L": 0.6608, "M": 0.0008, "N": 0.0005, "P": 0.0018, "Q": 0.0041, "R": 0.0158, "S": 0.0079, "T": 0.239, "V": 0.0066, "Y": 0.0009 },
    "109": { "A": 0.0003, "C": 0.0001, "D": 0.0032, "E": 0.0078, "F": 0.0006, "G": 0.0006, "H": 0.2614, "I": 0.0002, "K": 0.0063, "L": 0.0023, "N": 0.0029, "P": 0.0055, "Q": 0.6904, "R": 0.0079, "S": 0.002, "T": 0.0015, "V": 0.0004, "W": 0.0002, "Y": 0.0064 },
    "110": { "A": 0.002, "C": 0.0059, "D": 0.002, "F": 0.0494, "G": 0.0079, "I": 0.004, "K": 0.002, "L": 0.0198, "N": 0.002, "P": 0.004, "Q": 0.004, "R": 0.0059, "S": 0.0099, "T": 0.3557, "W": 0.5237, "Y": 0.002 },
    "113": { "A": 0.0028, "C": 0.0014, "D": 0.0011, "E": 0.0004, "F": 0.021, "G": 0.006, "H": 0.0018, "I": 0.0043, "K": 0.0007, "L": 0.0274, "N": 0.0046, "P": 0.161, "Q": 0.0032, "R": 0.0043, "S": 0.0135, "T": 0.4634, "V": 0.0011, "W": 0.2818, "Y": 0.0004 },
    "114": { "A": 0.0072, "C": 0.002, "D": 0.001, "E": 0.002, "F": 0.023, "G": 0.0043, "H": 0.0265, "I": 0.0099, "K": 0.0021, "L": 0.0373, "M": 0.0007, "N": 0.0044, "P": 0.1718, "Q": 0.1017, "R": 0.0129, "S": 0.0147, "T": 0.4438, "V": 0.0025, "W": 0.1305, "Y": 0.0019 },
    "115": { "A": 0.016, "C": 0.0005, "D": 0.0007, "E": 0.0048, "F": 0.0044, "G": 0.0189, "H": 0.0068, "I": 0.0062, "K": 0.0037, "L": 0.0618, "M": 0.0159, "N": 0.0005, "P": 0.6545, "Q": 0.0193, "R": 0.036, "S": 0.0362, "T": 0.0931, "V": 0.0064, "W": 0.0128, "Y": 0.0015 },
    "116": { "A": 0.0055, "C": 0.0118, "D": 0.0051, "E": 0.0024, "F": 0.0833, "G": 0.0284, "H": 0.0233, "I": 0.0697, "K": 0.0073, "L": 0.1565, "M": 0.0011, "N": 0.0085, "P": 0.1167, "Q": 0.0223, "R": 0.1096, "S": 0.0104, "T": 0.009, "V": 0.0238, "W": 0.1449, "Y": 0.1604 },
    "117": { "A": 0.0238, "C": 0.0005, "D": 0.0011, "E": 0.0007, "F": 0.0026, "G": 0.0068, "H": 0.0022, "I": 0.0028, "K": 0.001, "L": 0.0069, "M": 0.0011, "N": 0.0018, "P": 0.0102, "Q": 0.0007, "R": 0.0042, "S": 0.0368, "T": 0.8929, "V": 0.0036, "Y": 0.0003 },
    "118": { "A": 0.0001, "C": 0.0062, "D": 0.0002, "E": 0.0001, "F": 0.9769, "G": 0.0009, "H": 0.0004, "I": 0.0016, "K": 0.0001, "L": 0.0051, "N": 0.0002, "P": 0.0007, "Q": 0.0001, "R": 0.0005, "S": 0.0016, "T": 0.0002, "V": 0.0044, "Y": 0.0007 },
    "119": { "A": 0.001, "C": 0.0003, "D": 0.0015, "E": 0.0001, "G": 0.993, "I": 0.0001, "L": 0.0001, "P": 0.0001, "R": 0.0011, "S": 0.0016, "T": 0.0001, "V": 0.0011 },
    "120": { "A": 0.0005, "D": 0.0001, "E": 0.0018, "G": 0.1993, "H": 0.008, "K": 0.0016, "L": 0.0011, "N": 0.0002, "P": 0.1279, "Q": 0.6551, "R": 0.004, "S": 0.0003, "V": 0.0001 },
    "121": { "A": 0.0006, "E": 0.003, "G": 0.99, "Q": 0.0001, "R": 0.0027, "V": 0.0011, "W": 0.0024 },
    "122": { "A": 0.0018, "H": 0.0001, "I": 0.0011, "K": 0.0003, "N": 0.0007, "P": 0.0018, "S": 0.0007, "T": 0.9936 },
    "123": { "A": 0.0001, "D": 0.0001, "E": 0.0011, "G": 0.0002, "H": 0.0001, "I": 0.0001, "K": 0.8665, "L": 0.0006, "M": 0.0003, "N": 0.0031, "P": 0.0003, "Q": 0.004, "R": 0.118, "S": 0.0003, "T": 0.0049, "W": 0.0001 },
    "124": { "A": 0.0007, "D": 0.0001, "E": 0.0006, "F": 0.0001, "G": 0.0003, "I": 0.0002, "L": 0.3745, "M": 0.0032, "P": 0.0003, "Q": 0.0002, "R": 0.0001, "V": 0.6198 },
    "125": { "A": 0.0009, "D": 0.1304, "E": 0.8612, "G": 0.0018, "H": 0.0006, "K": 0.0013, "N": 0.0001, "Q": 0.0024, "R": 0.0001, "V": 0.0007, "Y": 0.0003 },
    "126": { "F": 0.0015, "H": 0.0001, "I": 0.9814, "L": 0.0037, "M": 0.0026, "N": 0.0007, "Q": 0.0001, "S": 0.0032, "T": 0.0015, "V": 0.0051, "Y": 0.0001 },
    "127": { "E": 0.0032, "K": 0.9929, "N": 0.0005, "Q": 0.0034, "T": 0.0001 }
  },
  "IGKV2D": {
    "1": { "D": 0.9434, "E": 0.0566 },
    "2": { "A": 0.0013, "C": 0.0001, "D": 0.0005, "E": 0.0001, "F": 0.0028, "G": 0.0017, "H": 0.0001, "I": 0.8535, "L": 0.0202, "M": 0.0089, "N": 0.002, "P": 0.0003, "R": 0.0007, "S": 0.0087, "T": 0.0069, "V": 0.092, "Y": 0.0001 },
    "3": { "A": 0.005, "C": 0.0001, "D": 0.0001, "E": 0.0034, "F": 0.0003, "G": 0.0056, "I": 0.0008, "L": 0.0098, "M": 0.0029, "P": 0.0001, "R": 0.0002, "S": 0.0001, "V": 0.9714, "W": 0.0004 },
    "4": { "F": 0.0003, "I": 0.0173, "K": 0.0007, "L": 0.0422, "M": 0.9259, "P": 0.0003, "Q": 0.0001, "R": 0.0046, "S": 0.0003, "T": 0.0023, "V": 0.0059, "W": 0.0001 },
    "5": { "A": 0.0027, "C": 0.0001, "D": 0.0001, "F": 0.0001, "H": 0.0005, "I": 0.0053, "K": 0.0003, "L": 0.0003, "M": 0.0001, "N": 0.0053, "P": 0.0122, "R": 0.0002, "S": 0.006, "T": 0.9667, "V": 0.0001 },
    "6": { "A": 0.0002, "E": 0.0027, "G": 0.0001, "H": 0.0098, "K": 0.0085, "L": 0.0016, "N": 0.0001, "P": 0.013, "Q": 0.9608, "R": 0.0021, "S": 0.0003, "T": 0.0005, "V": 0.0001, "Y": 0.0001 },
    "7": { "A": 0.0047, "C": 0.0005, "D": 0.0001, "E": 0.0001, "F": 0.0005, "H": 0.0009, "I": 0.0035, "K": 0.0001, "L": 0.0001, "N": 0.0042, "P": 0.0289, "Q": 0.0001, "R": 0.0003, "S": 0.0903, "T": 0.8647, "V": 0.0001, "Y": 0.0008 },
    "8": { "A": 0.0029, "H": 0.001, "K": 0.0002, "L": 0.0089, "N": 0.0001, "P": 0.958, "Q": 0.013, "R": 0.0031, "S": 0.003, "T": 0.0096 },
    "9": { "A": 0.0028, "C": 0.0001, "D": 0.0003, "E": 0.0001, "F": 0.0017, "G": 0.0007, "H": 0.0038, "I": 0.0046, "L": 0.951, "P": 0.0214, "Q": 0.0001, "R": 0.0067, "S": 0.0008, "T": 0.0008, "V": 0.0053, "Y": 0.0001 },
    "10": { "A": 0.0073, "C": 0.0013, "E": 0.0001, "F": 0.0032, "G": 0.0001, "H": 0.0003, "I": 0.0004, "L": 0.0005, "P": 0.0197, "S": 0.9563, "T": 0.0063, "W": 0.0001, "Y": 0.0044 },
    "11": { "A": 0.0001, "C": 0.0001, "F": 0.0007, "H": 0.0001, "I": 0.0001, "L": 0.9836, "M": 0.0059, "P": 0.0033, "Q": 0.0012, "R": 0.0031, "S": 0.0001, "T": 0.0001, "V": 0.0016 },
    "12": { "A": 0.0034, "C": 0.0011, "D": 0.0001, "F": 0.0027, "H": 0.0055, "L": 0.0038, "N": 0.0001, "P": 0.3197, "Q": 0.0001, "R": 0.0018, "S": 0.6548, "T": 0.003, "V": 0.0003, "Y": 0.0037 },
    "13": { "A": 0.0055, "C": 0.0001, "D": 0.0004, "F": 0.0043, "G": 0.0028, "I": 0.0248, "L": 0.0016, "M": 0.0003, "N": 0.0001, "S": 0.0004, "T": 0.0001, "V": 0.9596 },
    "14": { "A": 0.0059, "D": 0.0001, "H": 0.0003, "I": 0.0054, "L": 0.0003, "M": 0.0001, "N": 0.0029, "P": 0.0111, "R": 0.0003, "S": 0.0051, "T": 0.9677, "V": 0.0008, "Y": 0.0001 },
    "15": { "A": 0.0018, "C": 0.0001, "F": 0.0007, "H": 0.0038, "I": 0.0002, "L": 0.0794, "M": 0.0001, "P": 0.8986, "Q": 0.0002, "R": 0.0023, "S": 0.0039, "T": 0.0089, "V": 0.0001, "Y": 0.0001 },
    "16": { "A": 0.004, "C": 0.0003, "E": 0.0023, "G": 0.9847, "L": 0.0001, "P": 0.0001, "R": 0.0032, "S": 0.0001, "T": 0.0001, "V": 0.005, "W": 0.0002 },
    "17": { "A": 0.0049, "D": 0.0036, "E": 0.2703, "G": 0.0021, "H": 0.0027, "K": 0.0032, "L": 0.0005, "M": 0.0001, "N": 0.0001, "P": 0.005, "Q": 0.704, "R": 0.0029, "S": 0.0001, "T": 0.0001, "V": 0.0003, "W": 0.0001 },
    "18": { "A": 0.0063, "E": 0.0001, "H": 0.0002, "K": 0.0001, "L": 0.0021, "P": 0.9447, "Q": 0.0304, "R": 0.0019, "S": 0.0076, "T": 0.0064 },
    "19": { "A": 0.9808, "C": 0.0001, "D": 0.0048, "E": 0.0001, "G": 0.0004, "H": 0.0001, "N": 0.0001, "P": 0.0037, "S": 0.0055, "T": 0.0023, "V": 0.0021, "Y": 0.0001 },
    "20": { "A": 0.006, "C": 0.0004, "D": 0.0001, "F": 0.0016, "H": 0.0002, "L": 0.0001, "N": 0.0001, "P": 0.008, "R": 0.0001, "S": 0.9762, "T": 0.0032, "Y": 0.004 },
    "21": { "A": 0.0001, "C": 0.0001, "F": 0.0033, "H": 0.0003, "I": 0.945, "L": 0.0129, "M": 0.0226, "N": 0.0023, "P": 0.0001, "R": 0.0003, "S": 0.0043, "T": 0.0073, "V": 0.0016 },
    "22": { "A": 0.0101, "C": 0.0003, "D": 0.0002, "F": 0.0017, "H": 0.0009, "L": 0.0005, "N": 0.0001, "P": 0.0299, "S": 0.9471, "T": 0.005, "W": 0.0001, "Y": 0.0042 },
    "23": { "A": 0.0004, "C": 0.9799, "F": 0.0042, "G": 0.0031, "L": 0.0002, "R": 0.0038, "S": 0.0056, "V": 0.0003, "W": 0.0005, "Y": 0.002 },
    "24": { "A": 0.0002, "C": 0.0001, "E": 0.0016, "G": 0.0005, "H": 0.0002, "I": 0.0002, "K": 0.5994, "L": 0.0001, "M": 0.0027, "N": 0.0074, "Q": 0.0095, "R": 0.3542, "S": 0.0036, "T": 0.0194, "V": 0.0001, "W": 0.0009, "Y": 0.0001 },
    "25": { "A": 0.0065, "C": 0.0008, "D": 0.0001, "F": 0.0023, "H": 0.0001, "L": 0.0001, "P": 0.0112, "Q": 0.0001, "R": 0.0001, "S": 0.9718, "T": 0.0028, "V": 0.0002, "Y": 0.0041 },
    "26": { "A": 0.0001, "C": 0.001, "D": 0.0008, "G": 0.004, "I": 0.0037, "L": 0.0003, "M": 0.0004, "N": 0.005, "P": 0.0006, "R": 0.0158, "S": 0.9586, "T": 0.0095, "V": 0.0001 },
    "27": { "A": 0.0001, "C": 0.0003, "D": 0.0008, "E": 0.0077, "H": 0.0091, "K": 0.0053, "L": 0.0014, "M": 0.0001, "N": 0.0001, "P": 0.0079, "Q": 0.9595, "R": 0.0072, "S": 0.0001, "T": 0.0004 },
    "28": { "A": 0.0003, "C": 0.0027, "D": 0.0001, "F": 0.0001, "G": 0.0049, "H": 0.0001, "I": 0.0043, "K": 0.0003, "L": 0.0005, "N": 0.0033, "P": 0.0002, "R": 0.0219, "S": 0.9535, "T": 0.0076, "V": 0.0002 },
    "29": { "C": 0.0001, "F": 0.0022, "H": 0.0029, "I": 0.0058, "L": 0.9693, "N": 0.0001, "P": 0.0083, "Q": 0.0001, "R": 0.0088, "S": 0.0005, "T": 0.0002, "V": 0.0017 },
    "30": { "A": 0.0003, "C": 0.0001, "D": 0.0001, "E": 0.0003, "F": 0.0105, "G": 0.0007, "H": 0.0003, "I": 0.0015, "K": 0.0012, "L": 0.8839, "M": 0.0055, "N": 0.0001, "P": 0.0022, "Q": 0.002, "R": 0.0048, "S": 0.0015, "T": 0.0003, "V": 0.0829, "W": 0.0017, "Y": 0.0001 },
    "31": { "A": 0.0059, "C": 0.0001, "D": 0.2394, "E": 0.0045, "F": 0.0028, "G": 0.0011, "H": 0.6266, "I": 0.0001, "K": 0.001, "L": 0.0015, "N": 0.0123, "P": 0.0056, "Q": 0.0077, "R": 0.004, "S": 0.0032, "T": 0.0003, "V": 0.0011, "W": 0.0002, "Y": 0.0826 },
    "32": { "A": 0.0007, "C": 0.0014, "D": 0.0015, "E": 0.0001, "F": 0.0001, "G": 0.0126, "H": 0.0009, "I": 0.0041, "K": 0.0004, "L": 0.0005, "M": 0.0001, "N": 0.0044, "P": 0.0014, "R": 0.0236, "S": 0.9306, "T": 0.0169, "V": 0.0001, "W": 0.0001, "Y": 0.0005 },
    "33": { "A": 0.0122, "C": 0.0003, "D": 0.9316, "E": 0.0072, "F": 0.0005, "G": 0.0157, "H": 0.0077, "N": 0.0088, "Q": 0.0011, "R": 0.0003, "S": 0.0011, "V": 0.0077, "Y": 0.0059 },
    "34": { "A": 0.0099, "D": 0.9469, "E": 0.0045, "F": 0.0001, "G": 0.0053, "H": 0.0032, "I": 0.0002, "N": 0.0157, "P": 0.0002, "R": 0.0004, "S": 0.0016, "T": 0.0007, "V": 0.0037, "Y": 0.0076 },
    "35": { "A": 0.0036, "C": 0.0001, "D": 0.0005, "E": 0.0056, "G": 0.9745, "I": 0.0001, "K": 0.0002, "L": 0.0001, "P": 0.0001, "Q": 0.0002, "R": 0.0063, "S": 0.001, "T": 0.0002, "V": 0.0072, "W": 0.0002 },
    "36": { "A": 0.0004, "D": 0.0035, "E": 0.0032, "F": 0.0001, "G": 0.0003, "H": 0.0053, "I": 0.0006, "K": 0.6028, "L": 0.0001, "M": 0.004, "N": 0.3136, "Q": 0.0068, "R": 0.0181, "S": 0.0016, "T": 0.0123, "V": 0.0003, "W": 0.0001, "Y": 0.0269 },
    "37": { "A": 0.0041, "D": 0.0001, "F": 0.0001, "H": 0.0007, "I": 0.0061, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0089, "P": 0.009, "R": 0.0004, "S": 0.0066, "T": 0.9632, "Y": 0.0004 },
    "38": { "A": 0.0001, "C": 0.0025, "D": 0.0038, "F": 0.0079, "H": 0.0101, "I": 0.0001, "L": 0.0018, "N": 0.0042, "P": 0.0005, "Q": 0.0002, "R": 0.0027, "S": 0.0116, "T": 0.0005, "Y": 0.9541 },
    "39": { "A": 0.0001, "C": 0.0007, "F": 0.0151, "H": 0.0001, "I": 0.0008, "L": 0.9634, "M": 0.0018, "R": 0.0002, "S": 0.0031, "V": 0.0103, "W": 0.0042, "Y": 0.0001 },
    "40": { "A": 0.0024, "C": 0.0042, "D": 0.2534, "E": 0.0016, "F": 0.0086, "G": 0.001, "H": 0.0079, "I": 0.001, "K": 0.0008, "L": 0.0005, "N": 0.0741, "Q": 0.0009, "R": 0.0001, "S": 0.015, "T": 0.001, "V": 0.0009, "W": 0.0001, "Y": 0.6265 },
    "41": { "C": 0.006, "G": 0.0051, "L": 0.0053, "P": 0.0001, "R": 0.0029, "S": 0.0012, "T": 0.0001, "V": 0.0003, "W": 0.9789, "Y": 0.0001 },
    "42": { "C": 0.0016, "D": 0.0023, "F": 0.1107, "H": 0.0029, "I": 0.0007, "L": 0.002, "N": 0.0012, "Q": 0.0001, "S": 0.0036, "T": 0.0001, "V": 0.0007, "Y": 0.8741 },
    "43": { "A": 0.0001, "C": 0.0005, "D": 0.0001, "F": 0.0017, "G": 0.0001, "H": 0.0042, "I": 0.0001, "K": 0.0003, "L": 0.8975, "M": 0.004, "P": 0.0022, "Q": 0.0738, "R": 0.0044, "S": 0.0003, "T": 0.0001, "V": 0.0107 },
    "44": { "E": 0.001, "H": 0.0068, "K": 0.0046, "L": 0.001, "N": 0.0001, "P": 0.0042, "Q": 0.979, "R": 0.0031, "T": 0.0002, "V": 0.0001, "W": 0.0001 },
    "45": { "A": 0.0001, "E": 0.0014, "G": 0.0003, "H": 0.0003, "K": 0.8843, "M": 0.001, "N": 0.0031, "P": 0.0001, "Q": 0.0054, "R": 0.0959, "S": 0.0008, "T": 0.0073, "W": 0.0001 },
    "46": { "A": 0.036, "D": 0.0001, "E": 0.0002, "G": 0.0001, "H": 0.0001, "K": 0.0001, "L": 0.0015, "P": 0.9405, "Q": 0.0068, "R": 0.001, "S": 0.0072, "T": 0.0064, "V": 0.0001 },
    "47": { "A": 0.0012, "C": 0.0026, "D": 0.0016, "E": 0.0009, "F": 0.0001, "G": 0.9615, "K": 0.0001, "M": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0241, "S": 0.0011, "V": 0.0058, "W": 0.0008 },
    "48": { "E": 0.0008, "H": 0.006, "K": 0.0051, "L": 0.0012, "P": 0.0276, "Q": 0.9555, "R": 0.0033, "T": 0.0004 },
    "49": { "A": 0.0065, "C": 0.0008, "D": 0.0001, "F": 0.0011, "G": 0.0003, "H": 0.0031, "I": 0.0001, "L": 0.0005, "P": 0.5154, "Q": 0.0001, "R": 0.0007, "S": 0.4416, "T": 0.0053, "V": 0.0231, "W": 0.0001, "Y": 0.0013 },
    "50": { "A": 0.0011, "H": 0.0003, "K": 0.0001, "L": 0.0022, "P": 0.9601, "Q": 0.0091, "R": 0.0012, "S": 0.0234, "T": 0.0025, "Y": 0.0001 },
    "51": { "C": 0.0001, "D": 0.0003, "E": 0.0014, "G": 0.0001, "H": 0.02, "K": 0.004, "L": 0.0016, "M": 0.0001, "N": 0.0001, "P": 0.0074, "Q": 0.8611, "R": 0.0795, "S": 0.0011, "T": 0.0231, "W": 0.0002 },
    "52": { "A": 0.0003, "C": 0.0003, "D": 0.0001, "F": 0.0021, "G": 0.0001, "H": 0.0024, "I": 0.0048, "L": 0.8885, "M": 0.0001, "P": 0.0089, "R": 0.0832, "S": 0.0003, "V": 0.0089, "Y": 0.0001 },
    "53": { "F": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.9898, "M": 0.0018, "P": 0.0015, "Q": 0.0011, "R": 0.0035, "T": 0.0002, "V": 0.0016, "W": 0.0001 },
    "54": { "C": 0.0003, "D": 0.0001, "F": 0.0052, "G": 0.0001, "I": 0.9614, "L": 0.0111, "M": 0.0076, "N": 0.0017, "P": 0.0002, "R": 0.0003, "S": 0.0048, "T": 0.0027, "V": 0.0044, "Y": 0.0001 },
    "55": { "A": 0.0003, "C": 0.0018, "D": 0.007, "E": 0.0001, "F": 0.0061, "G": 0.0001, "H": 0.0142, "I": 0.0001, "K": 0.0003, "L": 0.0003, "N": 0.0014, "P": 0.0001, "Q": 0.0007, "R": 0.0006, "S": 0.0123, "T": 0.0001, "V": 0.0001, "W": 0.0002, "Y": 0.9542 },
    "56": { "A": 0.0084, "D": 0.0086, "E": 0.6408, "F": 0.0001, "G": 0.0072, "H": 0.0003, "I": 0.0001, "K": 0.0741, "L": 0.0018, "M": 0.0016, "N": 0.0008, "P": 0.0048, "Q": 0.0021, "R": 0.0018, "S": 0.0053, "T": 0.24, "V": 0.0018, "W": 0.0002, "Y": 0.0002 },
    "57": { "A": 0.01, "C": 0.0001, "D": 0.0023, "E": 0.0001, "F": 0.0047, "G": 0.0126, "H": 0.0005, "I": 0.007, "L": 0.2498, "M": 0.0001, "N": 0.0005, "P": 0.0016, "R": 0.0043, "S": 0.001, "T": 0.0004, "V": 0.7051 },
    "65": { "A": 0.0085, "C": 0.0007, "D": 0.0001, "F": 0.0059, "H": 0.0005, "I": 0.0003, "L": 0.0001, "N": 0.0003, "P": 0.002, "S": 0.9738, "T": 0.0018, "V": 0.0006, "Y": 0.0055 },
    "66": { "A": 0.0005, "C": 0.0014, "D": 0.0039, "E": 0.0001, "F": 0.0017, "G": 0.0003, "H": 0.0077, "I": 0.0023, "K": 0.0197, "L": 0.0001, "M": 0.0001, "N": 0.6962, "P": 0.0005, "Q": 0.0005, "R": 0.0034, "S": 0.0141, "T": 0.0095, "W": 0.0001, "Y": 0.2379 },
    "67": { "C": 0.0001, "E": 0.0001, "G": 0.001, "H": 0.0003, "L": 0.0052, "M": 0.0001, "P": 0.002, "Q": 0.0018, "R": 0.9132, "S": 0.0003, "T": 0.0001, "W": 0.0759 },
    "68": { "A": 0.2544, "C": 0.0139, "D": 0.0736, "E": 0.0013, "F": 0.6239, "G": 0.0036, "H": 0.0001, "I": 0.0016, "L": 0.0083, "M": 0.0003, "P": 0.0008, "R": 0.0005, "S": 0.0031, "T": 0.0008, "V": 0.0087, "W": 0.0002, "Y": 0.0048 },
    "69": { "A": 0.0218, "C": 0.0025, "D": 0.0004, "E": 0.0001, "F": 0.0025, "G": 0.0003, "H": 0.0002, "L": 0.0003, "P": 0.0098, "R": 0.0003, "S": 0.9495, "T": 0.0064, "V": 0.0003, "W": 0.0001, "Y": 0.0057 },
    "70": { "A": 0.0021, "C": 0.0003, "E": 0.0012, "G": 0.984, "I": 0.0001, "K": 0.0001, "L": 0.0001, "R": 0.0048, "V": 0.0065, "W": 0.0008 },
    "71": { "A": 0.0027, "C": 0.0001, "D": 0.0003, "E": 0.0006, "F": 0.0017, "G": 0.0117, "I": 0.0007, "L": 0.0027, "M": 0.0014, "P": 0.0001, "R": 0.0001, "S": 0.0001, "V": 0.9778, "W": 0.0001 },
    "72": { "A": 0.0066, "K": 0.0001, "L": 0.0018, "P": 0.9671, "Q": 0.0039, "R": 0.0011, "S": 0.0109, "T": 0.0085 },
    "74": { "A": 0.006, "D": 0.9533, "E": 0.023, "G": 0.0038, "H": 0.0014, "I": 0.0003, "K": 0.0009, "L": 0.0001, "N": 0.0035, "P": 0.0007, "Q": 0.0003, "R": 0.0001, "S": 0.0005, "V": 0.0023, "Y": 0.0038 },
    "75": { "C": 0.0002, "G": 0.0023, "I": 0.0014, "K": 0.006, "L": 0.0001, "M": 0.0059, "N": 0.0001, "P": 0.0001, "R": 0.9757, "S": 0.0057, "T": 0.0012, "V": 0.0001, "W": 0.0012 },
    "76": { "A": 0.0001, "C": 0.0179, "F": 0.9534, "G": 0.0008, "H": 0.0001, "I": 0.0033, "L": 0.0087, "M": 0.0001, "S": 0.0014, "V": 0.0129, "W": 0.0001, "Y": 0.0015 },
    "77": { "A": 0.0001, "C": 0.0008, "D": 0.0001, "G": 0.0036, "H": 0.0001, "I": 0.0048, "K": 0.0001, "L": 0.0002, "M": 0.0001, "N": 0.0018, "R": 0.0145, "S": 0.9686, "T": 0.0048, "V": 0.0003, "Y": 0.0001 },
    "78": { "A": 0.0042, "C": 0.0048, "D": 0.0027, "E": 0.0001, "G": 0.9799, "H": 0.0001, "L": 0.0001, "R": 0.0007, "S": 0.0034, "V": 0.0041, "W": 0.0001 },
    "79": { "A": 0.0001, "C": 0.0007, "D": 0.0001, "G": 0.0093, "I": 0.0019, "K": 0.0001, "L": 0.0001, "N": 0.001, "Q": 0.0001, "R": 0.0134, "S": 0.9695, "T": 0.0027, "V": 0.0002, "W": 0.0001, "Y": 0.0008 },
    "80": { "A": 0.0023, "C": 0.0004, "D": 0.0001, "E": 0.0029, "G": 0.9797, "K": 0.0003, "R": 0.0028, "S": 0.0002, "V": 0.0062, "W": 0.0051 },
    "83": { "A": 0.007, "C": 0.0003, "E": 0.0003, "F": 0.0001, "G": 0.0001, "K": 0.0001, "L": 0.0021, "P": 0.001, "Q": 0.0001, "S": 0.9866, "T": 0.0019, "V": 0.0001, "W": 0.0001, "Y": 0.0003 },
    "84": { "A": 0.0032, "C": 0.001, "D": 0.0034, "E": 0.0048, "F": 0.0001, "G": 0.9755, "L": 0.0001, "M": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0022, "S": 0.0005, "V": 0.0066, "W": 0.0022, "Y": 0.0001 },
    "85": { "A": 0.0041, "D": 0.0003, "I": 0.0013, "K": 0.0049, "L": 0.0001, "N": 0.0018, "P": 0.0014, "Q": 0.0001, "R": 0.0017, "S": 0.0051, "T": 0.9793 },
    "86": { "A": 0.0023, "D": 0.9595, "E": 0.0177, "F": 0.0001, "G": 0.002, "H": 0.0032, "K": 0.0001, "L": 0.0001, "N": 0.0068, "P": 0.0001, "Q": 0.0003, "S": 0.0003, "T": 0.0003, "V": 0.001, "Y": 0.0061 },
    "87": { "C": 0.0063, "D": 0.0001, "F": 0.975, "I": 0.0012, "L": 0.0044, "S": 0.0015, "V": 0.0102, "Y": 0.0014 },
    "88": { "A": 0.0033, "E": 0.0001, "H": 0.0001, "I": 0.0034, "K": 0.0025, "L": 0.0002, "M": 0.0001, "N": 0.0001, "P": 0.005, "Q": 0.0001, "R": 0.0006, "S": 0.0057, "T": 0.9787, "V": 0.0001, "Y": 0.0001 },
    "89": { "E": 0.0001, "F": 0.0001, "I": 0.0001, "L": 0.9847, "M": 0.0036, "P": 0.0026, "Q": 0.0014, "R": 0.0063, "T": 0.0001, "V": 0.0009 },
    "90": { "A": 0.0002, "D": 0.0001, "E": 0.008, "G": 0.0002, "H": 0.001, "I": 0.0018, "K": 0.9443, "L": 0.0001, "M": 0.0008, "N": 0.0115, "Q": 0.005, "R": 0.0162, "S": 0.0001, "T": 0.0106, "Y": 0.0001 },
    "91": { "F": 0.0012, "H": 0.0001, "I": 0.9741, "L": 0.0025, "M": 0.0012, "N": 0.0008, "R": 0.0001, "S": 0.0153, "T": 0.0012, "V": 0.0035 },
    "92": { "A": 0.0016, "C": 0.0007, "D": 0.0005, "G": 0.0021, "I": 0.0028, "K": 0.0005, "L": 0.0001, "M": 0.0001, "N": 0.0035, "Q": 0.0001, "R": 0.007, "S": 0.9776, "T": 0.0032, "V": 0.0003 },
    "93": { "A": 0.0004, "E": 0.0001, "G": 0.0017, "H": 0.0001, "I": 0.0001, "K": 0.0033, "L": 0.0047, "M": 0.0013, "N": 0.0016, "P": 0.001, "Q": 0.0014, "R": 0.9785, "S": 0.0044, "T": 0.0001, "V": 0.0001, "W": 0.0013 },
    "94": { "A": 0.001, "E": 0.001, "F": 0.0001, "G": 0.0065, "I": 0.0001, "L": 0.007, "M": 0.0039, "P": 0.0001, "R": 0.0002, "V": 0.9799, "W": 0.0003 },
    "95": { "A": 0.0032, "D": 0.0048, "E": 0.9832, "G": 0.0029, "K": 0.0012, "L": 0.0001, "Q": 0.002, "R": 0.0008, "S": 0.0003, "V": 0.0014 },
    "96": { "A": 0.9718, "D": 0.0062, "E": 0.0004, "G": 0.0014, "I": 0.0001, "L": 0.0006, "N": 0.0001, "P": 0.0059, "S": 0.0055, "T": 0.0051, "V": 0.0028 },
    "97": { "A": 0.004, "D": 0.0185, "E": 0.9681, "G": 0.0053, "K": 0.001, "Q": 0.0006, "R": 0.0005, "T": 0.0002, "V": 0.0017, "W": 0.0001, "Y": 0.0001 },
    "98": { "A": 0.0015, "C": 0.0001, "D": 0.9829, "E": 0.0038, "F": 0.0002, "G": 0.0014, "H": 0.0007, "M": 0.0006, "N": 0.0012, "S": 0.0002, "V": 0.0016, "Y": 0.0057 },
    "99": { "A": 0.0068, "C": 0.0001, "D": 0.0007, "F": 0.0289, "G": 0.0048, "I": 0.0083, "L": 0.0014, "R": 0.0001, "S": 0.0003, "V": 0.9483, "Y": 0.0003 },
    "100": { "A": 0.0038, "C": 0.0003, "D": 0.0002, "E": 0.001, "G": 0.9814, "L": 0.0001, "Q": 0.0001, "R": 0.0023, "V": 0.0063, "W": 0.0044 },
    "101": { "A": 0.0019, "D": 0.0008, "F": 0.0068, "G": 0.0035, "I": 0.0304, "L": 0.014, "M": 0.0008, "N": 0.0003, "P": 0.0001, "R": 0.0002, "S": 0.002, "T": 0.0053, "V": 0.9336, "Y": 0.0003 },
    "102": { "A": 0.0001, "C": 0.0012, "D": 0.0035, "E": 0.0001, "F": 0.0033, "H": 0.0008, "I": 0.0009, "L": 0.0005, "N": 0.0004, "S": 0.004, "V": 0.0001, "Y": 0.9851 },
    "103": { "A": 0.0001, "C": 0.0042, "D": 0.0034, "E": 0.0001, "F": 0.0074, "G": 0.0001, "H": 0.0009, "I": 0.0001, "L": 0.0007, "N": 0.0001, "P": 0.0001, "S": 0.0039, "T": 0.001, "W": 0.0001, "Y": 0.9779 },
    "104": { "A": 0.0011, "C": 0.9835, "F": 0.0023, "G": 0.004, "H": 0.0003, "L": 0.0009, "P": 0.0001, "R": 0.0018, "S": 0.0014, "V": 0.0001, "W": 0.0018, "Y": 0.0027 },
    "105": { "A": 0.0005, "C": 0.0009, "E": 0.0003, "F": 0.0014, "G": 0.0008, "H": 0.0005, "I": 0.0069, "K": 0.0053, "L": 0.0105, "M": 0.9248, "N": 0.0001, "P": 0.0001, "Q": 0.0014, "R": 0.0379, "S": 0.0017, "T": 0.0029, "V": 0.0035, "W": 0.0005, "Y": 0.0001 },
    "106": { "A": 0.0004, "D": 0.0003, "E": 0.0023, "G": 0.0003, "H": 0.0085, "I": 0.0001, "K": 0.005, "L": 0.0012, "M": 0.0001, "N": 0.0001, "P": 0.0025, "Q": 0.9747, "R": 0.004, "S": 0.0002, "T": 0.0001, "V": 0.0001, "Y": 0.0001 },
    "107": { "A": 0.0072, "C": 0.0037, "D": 0.0234, "E": 0.0005, "F": 0.0001, "G": 0.0752, "H": 0.0018, "I": 0.0005, "K": 0.0008, "L": 0.0003, "M": 0.0001, "N": 0.0102, "P": 0.0007, "Q": 0.0001, "R": 0.2643, "S": 0.5916, "T": 0.0161, "V": 0.0018, "W": 0.0001, "Y": 0.0016 },
    "108": { "A": 0.0229, "D": 0.0004, "E": 0.0012, "F": 0.0008, "G": 0.0012, "H": 0.0003, "I": 0.7938, "K": 0.0154, "L": 0.036, "M": 0.0059, "N": 0.001, "Q": 0.0002, "R": 0.0173, "S": 0.0018, "T": 0.0837, "V": 0.0165, "W": 0.0001, "Y": 0.0015 },
    "109": { "A": 0.002, "D": 0.0066, "E": 0.2614, "F": 0.0002, "G": 0.0024, "H": 0.0884, "I": 0.0001, "K": 0.0092, "L": 0.0023, "M": 0.0001, "N": 0.0021, "P": 0.0044, "Q": 0.6122, "R": 0.0047, "S": 0.001, "T": 0.0011, "V": 0.0002, "W": 0.0001, "Y": 0.0016 },
    "110": { "C": 0.0048, "D": 0.0336, "E": 0.0072, "F": 0.1631, "G": 0.0072, "H": 0.0024, "I": 0.0144, "L": 0.5947, "N": 0.0024, "P": 0.0096, "R": 0.0216, "T": 0.0024, "V": 0.0072, "W": 0.1295 },
    "113": { "A": 0.0016, "C": 0.0063, "D": 0.036, "E": 0.0012, "F": 0.1512, "G": 0.0031, "H": 0.0031, "I": 0.0043, "L": 0.4814, "M": 0.0004, "P": 0.1582, "Q": 0.002, "R": 0.0149, "S": 0.0086, "T": 0.0059, "V": 0.0063, "W": 0.1144, "Y": 0.0012 },
    "114": { "A": 0.0023, "C": 0.0036, "D": 0.0105, "E": 0.0104, "F": 0.2102, "G": 0.0026, "H": 0.0154, "I": 0.0068, "K": 0.0006, "L": 0.4228, "M": 0.0001, "N": 0.0008, "P": 0.1715, "Q": 0.0587, "R": 0.0132, "S": 0.0123, "T": 0.0052, "V": 0.0083, "W": 0.0389, "Y": 0.0057 },
    "115": { "A": 0.0148, "C": 0.0005, "D": 0.0025, "E": 0.0086, "F": 0.0117, "G": 0.0132, "H": 0.0053, "I": 0.0044, "K": 0.0017, "L": 0.079, "M": 0.0125, "N": 0.0003, "P": 0.7429, "Q": 0.017, "R": 0.0242, "S": 0.0409, "T": 0.0084, "V": 0.0045, "W": 0.0071, "Y": 0.0005 },
    "116": { "A": 0.0059, "C": 0.0135, "D": 0.0076, "E": 0.0049, "F": 0.0672, "G": 0.0173, "H": 0.0124, "I": 0.0969, "K": 0.0051, "L": 0.2482, "M": 0.0014, "N": 0.0045, "P": 0.0769, "Q": 0.0112, "R": 0.08, "S": 0.0152, "T": 0.002, "V": 0.0222, "W": 0.1347, "Y": 0.1728 },
    "117": { "A": 0.0178, "C": 0.0005, "D": 0.0012, "E": 0.0003, "F": 0.0017, "G": 0.0038, "H": 0.0024, "I": 0.003, "K": 0.0008, "L": 0.0065, "M": 0.0004, "N": 0.0008, "P": 0.0173, "Q": 0.0006, "R": 0.0035, "S": 0.0355, "T": 0.9009, "V": 0.0027, "W": 0.0001, "Y": 0.0001 },
    "118": { "A": 0.0003, "C": 0.0062, "D": 0.0005, "E": 0.0001, "F": 0.971, "G": 0.0005, "H": 0.0003, "I": 0.0016, "K": 0.0001, "L": 0.007, "M": 0.0001, "N": 0.0003, "P": 0.0018, "Q": 0.0005, "R": 0.0008, "S": 0.0021, "T": 0.0003, "V": 0.005, "W": 0.0001, "Y": 0.001 },
    "119": { "A": 0.0018, "C": 0.0006, "D": 0.0016, "G": 0.9906, "H": 0.0001, "I": 0.0001, "L": 0.0002, "P": 0.0008, "Q": 0.0001, "R": 0.0008, "S": 0.0021, "T": 0.0001, "V": 0.0009, "W": 0.0003, "Y": 0.0001 },
    "120": { "A": 0.0009, "C": 0.0002, "E": 0.0025, "G": 0.2767, "H": 0.0065, "K": 0.0011, "L": 0.0006, "P": 0.0999, "Q": 0.6064, "R": 0.0042, "S": 0.0004, "T": 0.0001, "V": 0.0005, "Y": 0.0001 },
    "121": { "A": 0.0003, "E": 0.0026, "G": 0.9899, "R": 0.0032, "V": 0.0014, "W": 0.0027 },
    "122": { "A": 0.0016, "D": 0.0001, "I": 0.001, "K": 0.0005, "N": 0.0007, "P": 0.0015, "R": 0.0001, "S": 0.0012, "T": 0.9934, "V": 0.0001 },
    "123": { "E": 0.0009, "G": 0.0007, "I": 0.0001, "K": 0.8415, "L": 0.0008, "M": 0.0006, "N": 0.0012, "P": 0.0004, "Q": 0.0031, "R": 0.148, "T": 0.0026, "W": 0.0001 },
    "124": { "A": 0.0002, "E": 0.0003, "G": 0.0005, "I": 0.0001, "L": 0.4046, "M": 0.0031, "P": 0.0003, "Q": 0.0001, "R": 0.0006, "T": 0.0001, "V": 0.5899 },
    "125": { "A": 0.0007, "D": 0.1132, "E": 0.881, "G": 0.0019, "H": 0.0001, "K": 0.0012, "L": 0.0001, "N": 0.0001, "Q": 0.0012, "R": 0.0001, "V": 0.0002, "Y": 0.0001 },
    "126": { "D": 0.0001, "F": 0.0008, "I": 0.984, "L": 0.0025, "M": 0.0044, "N": 0.0002, "S": 0.0034, "T": 0.0018, "V": 0.0027 },
    "127": { "E": 0.0033, "K": 0.9939, "N": 0.0004, "Q": 0.0024 }
  },
  "IGKV3": {
    "1": { "D": 0.0597, "E": 0.9403, "L": 0.0001 },
    "2": { "A": 0.0004, "C": 0.0001, "D": 0.0001, "F": 0.005, "G": 0.0003, "H": 0.0006, "I": 0.8962, "K": 0.0001, "L": 0.0433, "M": 0.0098, "N": 0.0036, "P": 0.0005, "R": 0.0009, "S": 0.0088, "T": 0.0126, "V": 0.0172, "Y": 0.0004 },
    "3": { "A": 0.0093, "E": 0.0044, "F": 0.0007, "G": 0.0055, "I": 0.0012, "K": 0.0001, "L": 0.0121, "M": 0.0035, "P": 0.0002, "Q": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.9624, "W": 0.0004 },
    "4": { "C": 0.0005, "F": 0.0276, "G": 0.0001, "I": 0.0001, "L": 0.9549, "M": 0.0043, "Q": 0.0001, "S": 0.0055, "T": 0.0001, "V": 0.0028, "W": 0.004, "Y": 0.0001 },
    "5": { "A": 0.0024, "H": 0.0001, "I": 0.0019, "K": 0.0027, "L": 0.0001, "M": 0.0021, "N": 0.0004, "P": 0.0219, "Q": 0.0004, "R": 0.0017, "S": 0.0043, "T": 0.9621 },
    "6": { "A": 0.0001, "E": 0.0007, "H": 0.0174, "K": 0.0048, "L": 0.0034, "M": 0.0002, "N": 0.0001, "P": 0.0237, "Q": 0.9471, "R": 0.0019, "S": 0.0003, "T": 0.0004, "Y": 0.0001 },
    "7": { "A": 0.005, "C": 0.0007, "D": 0.0001, "F": 0.0034, "H": 0.0003, "I": 0.0003, "L": 0.0007, "N": 0.0001, "P": 0.0093, "Q": 0.0001, "S": 0.9713, "T": 0.0023, "V": 0.0001, "W": 0.0001, "Y": 0.0063 },
    "8": { "A": 0.002, "D": 0.0001, "G": 0.0001, "H": 0.0004, "I": 0.0001, "L": 0.0046, "P": 0.9745, "Q": 0.0066, "R": 0.0018, "S": 0.0026, "T": 0.0074, "V": 0.0001 },
    "9": { "A": 0.4658, "C": 0.0046, "D": 0.0099, "F": 0.0001, "G": 0.495, "L": 0.0002, "P": 0.0021, "R": 0.0039, "S": 0.004, "T": 0.0013, "V": 0.0129, "Y": 0.0001 },
    "10": { "A": 0.0023, "C": 0.0001, "F": 0.0002, "G": 0.0001, "H": 0.0005, "I": 0.0053, "L": 0.0003, "N": 0.0036, "P": 0.0179, "R": 0.0001, "S": 0.005, "T": 0.9643, "V": 0.0001, "Y": 0.0001 },
    "11": { "A": 0.0006, "C": 0.0001, "I": 0.0001, "L": 0.9823, "M": 0.0046, "P": 0.0029, "Q": 0.0016, "R": 0.0046, "S": 0.0001, "T": 0.0001, "V": 0.003, "W": 0.0001 },
    "12": { "A": 0.006, "C": 0.0007, "F": 0.0027, "H": 0.0001, "L": 0.0002, "P": 0.0094, "S": 0.9757, "T": 0.0013, "Y": 0.0038 },
    "13": { "A": 0.0001, "C": 0.0005, "F": 0.0112, "L": 0.9624, "M": 0.0009, "P": 0.0001, "S": 0.0073, "V": 0.0109, "W": 0.0066 },
    "14": { "A": 0.004, "C": 0.0005, "F": 0.0029, "H": 0.0001, "L": 0.0003, "P": 0.006, "R": 0.0001, "S": 0.9806, "T": 0.0023, "V": 0.0001, "Y": 0.0032 },
    "15": { "A": 0.0035, "E": 0.0001, "G": 0.0001, "H": 0.0009, "L": 0.0053, "P": 0.9723, "Q": 0.008, "R": 0.0013, "S": 0.0023, "T": 0.0059, "V": 0.0003 },
    "16": { "A": 0.0035, "C": 0.0001, "E": 0.0029, "G": 0.9811, "L": 0.0001, "Q": 0.0001, "R": 0.0038, "S": 0.0001, "V": 0.005, "W": 0.0034 },
    "17": { "A": 0.016, "D": 0.0312, "E": 0.9385, "G": 0.0042, "H": 0.0002, "K": 0.0025, "L": 0.0001, "N": 0.0001, "Q": 0.0046, "R": 0.0002, "S": 0.0002, "T": 0.0003, "V": 0.0019, "Y": 0.0002 },
    "18": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "E": 0.0001, "G": 0.0104, "I": 0.0035, "K": 0.0025, "L": 0.0002, "M": 0.0002, "N": 0.0001, "P": 0.0003, "Q": 0.0001, "R": 0.948, "S": 0.0245, "T": 0.0095, "V": 0.0003, "W": 0.0001 },
    "19": { "A": 0.9697, "D": 0.0042, "F": 0.0001, "G": 0.0023, "H": 0.0001, "I": 0.0001, "L": 0.0001, "P": 0.004, "S": 0.0042, "T": 0.0035, "V": 0.0117 },
    "20": { "A": 0.0035, "D": 0.0001, "H": 0.0005, "I": 0.0032, "K": 0.0001, "L": 0.0002, "N": 0.0063, "P": 0.0097, "R": 0.0001, "S": 0.0049, "T": 0.9707, "V": 0.0006, "Y": 0.0001 },
    "21": { "A": 0.0001, "C": 0.0001, "F": 0.0036, "H": 0.002, "I": 0.0079, "L": 0.9681, "M": 0.0001, "P": 0.0084, "Q": 0.0001, "R": 0.0071, "S": 0.0004, "T": 0.0002, "V": 0.0021 },
    "22": { "A": 0.0129, "C": 0.0008, "D": 0.0002, "F": 0.0022, "G": 0.0001, "H": 0.0006, "L": 0.0005, "P": 0.0339, "R": 0.0001, "S": 0.9382, "T": 0.0067, "V": 0.0001, "Y": 0.0036 },
    "23": { "A": 0.0001, "C": 0.9781, "F": 0.0042, "G": 0.0054, "P": 0.0001, "R": 0.0042, "S": 0.0057, "V": 0.0003, "W": 0.0005, "Y": 0.0013 },
    "24": { "C": 0.0001, "E": 0.0001, "G": 0.0032, "I": 0.0001, "K": 0.0041, "L": 0.0001, "M": 0.0043, "P": 0.0001, "Q": 0.0004, "R": 0.9651, "S": 0.0166, "T": 0.0038, "V": 0.0001, "W": 0.002 },
    "25": { "A": 0.9701, "D": 0.004, "E": 0.0001, "G": 0.0014, "H": 0.0002, "L": 0.0001, "P": 0.004, "R": 0.0001, "S": 0.0066, "T": 0.0098, "V": 0.0034, "Y": 0.0001 },
    "26": { "A": 0.0003, "C": 0.001, "D": 0.0003, "G": 0.0046, "I": 0.004, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0025, "R": 0.0137, "S": 0.9705, "T": 0.0028 },
    "27": { "A": 0.0004, "C": 0.0001, "D": 0.0002, "E": 0.0073, "G": 0.0003, "H": 0.0089, "K": 0.0054, "L": 0.0026, "M": 0.0001, "N": 0.0002, "P": 0.0115, "Q": 0.9577, "R": 0.0051, "S": 0.0002, "T": 0.0001, "V": 0.0001, "Y": 0.0001 },
    "28": { "A": 0.0008, "C": 0.0033, "D": 0.0005, "F": 0.0005, "G": 0.0044, "H": 0.0003, "I": 0.0066, "K": 0.0004, "L": 0.0006, "M": 0.0003, "N": 0.0101, "P": 0.0009, "Q": 0.0001, "R": 0.0326, "S": 0.917, "T": 0.0195, "V": 0.0004, "W": 0.0001, "Y": 0.0014 },
    "29": { "A": 0.0039, "C": 0.0001, "D": 0.0015, "F": 0.0079, "G": 0.0041, "H": 0.0002, "I": 0.0402, "L": 0.0115, "M": 0.0003, "N": 0.0001, "P": 0.0002, "R": 0.0002, "S": 0.0005, "T": 0.0001, "V": 0.9291, "Y": 0.0001 },
    "30": { "A": 0.0059, "C": 0.0013, "D": 0.0065, "E": 0.0003, "F": 0.0012, "G": 0.0131, "H": 0.0007, "I": 0.0068, "K": 0.0023, "L": 0.0013, "M": 0.0005, "N": 0.016, "P": 0.0036, "Q": 0.0003, "R": 0.0269, "S": 0.879, "T": 0.0298, "V": 0.0017, "W": 0.0003, "Y": 0.0027 },
    "36": { "A": 0.0027, "C": 0.001, "D": 0.0048, "E": 0.0001, "F": 0.0015, "G": 0.0232, "H": 0.0013, "I": 0.0055, "K": 0.0015, "L": 0.0008, "M": 0.0001, "N": 0.0266, "P": 0.0001, "Q": 0.0001, "R": 0.0204, "S": 0.8966, "T": 0.0101, "V": 0.0007, "W": 0.0001, "Y": 0.0027 },
    "37": { "A": 0.0025, "C": 0.0015, "D": 0.0052, "E": 0.0005, "F": 0.0011, "G": 0.0095, "H": 0.0017, "I": 0.0084, "K": 0.0044, "L": 0.0004, "M": 0.0007, "N": 0.0585, "P": 0.0011, "Q": 0.0003, "R": 0.026, "S": 0.8449, "T": 0.03, "V": 0.0013, "W": 0.0002, "Y": 0.0019 },
    "38": { "A": 0.0025, "C": 0.0021, "D": 0.0082, "E": 0.0002, "F": 0.0227, "G": 0.0001, "H": 0.0097, "I": 0.0003, "K": 0.0012, "L": 0.0018, "N": 0.0074, "P": 0.0003, "Q": 0.0015, "R": 0.0011, "S": 0.0231, "T": 0.0018, "V": 0.0003, "W": 0.001, "Y": 0.9148 },
    "39": { "C": 0.0001, "E": 0.0001, "F": 0.0121, "I": 0.0058, "L": 0.9654, "M": 0.0003, "Q": 0.0001, "R": 0.0001, "S": 0.0029, "V": 0.0132 },
    "40": { "A": 0.9559, "C": 0.0003, "D": 0.0076, "E": 0.0001, "F": 0.0002, "G": 0.0087, "H": 0.0003, "I": 0.001, "L": 0.0003, "N": 0.0001, "P": 0.0037, "S": 0.0074, "T": 0.0062, "V": 0.0082, "Y": 0.0002 },
    "41": { "A": 0.0001, "C": 0.0091, "F": 0.0001, "G": 0.006, "L": 0.0039, "R": 0.0037, "S": 0.0006, "V": 0.0003, "W": 0.9761, "Y": 0.0001 },
    "42": { "C": 0.0016, "D": 0.0034, "E": 0.0001, "F": 0.007, "H": 0.0025, "I": 0.0001, "K": 0.0001, "L": 0.0002, "N": 0.0005, "P": 0.0001, "S": 0.0055, "T": 0.0002, "V": 0.0002, "Y": 0.9787 },
    "43": { "E": 0.0005, "H": 0.01, "I": 0.0001, "K": 0.005, "L": 0.0015, "M": 0.0003, "P": 0.0042, "Q": 0.9731, "R": 0.0046, "S": 0.0001, "T": 0.0003, "V": 0.0004 },
    "44": { "A": 0.0003, "E": 0.0003, "F": 0.0001, "G": 0.0001, "H": 0.0172, "K": 0.006, "L": 0.0025, "M": 0.0002, "N": 0.0003, "P": 0.0048, "Q": 0.965, "R": 0.0029, "T": 0.0001, "V": 0.0002, "Y": 0.0001 },
    "45": { "A": 0.0001, "D": 0.0001, "E": 0.0022, "G": 0.0003, "H": 0.0005, "I": 0.0022, "K": 0.935, "L": 0.0001, "M": 0.0001, "N": 0.0152, "P": 0.0001, "Q": 0.0091, "R": 0.0229, "S": 0.0006, "T": 0.0113 },
    "46": { "A": 0.0034, "C": 0.0001, "D": 0.0001, "F": 0.0013, "G": 0.0004, "H": 0.0048, "I": 0.0002, "L": 0.004, "N": 0.0002, "P": 0.9702, "Q": 0.0001, "R": 0.0044, "S": 0.005, "T": 0.0042, "V": 0.0015, "Y": 0.0001 },
    "47": { "A": 0.0017, "C": 0.0039, "D": 0.0041, "E": 0.0001, "G": 0.9819, "H": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.0013, "S": 0.0025, "V": 0.0042, "W": 0.0001 },
    "48": { "E": 0.0011, "H": 0.0055, "K": 0.0051, "L": 0.0021, "M": 0.0001, "P": 0.0058, "Q": 0.9766, "R": 0.0034, "S": 0.0001, "T": 0.0002 },
    "49": { "A": 0.9558, "D": 0.0041, "E": 0.0004, "G": 0.0031, "L": 0.0001, "P": 0.0166, "R": 0.0001, "S": 0.0091, "T": 0.0082, "V": 0.0025 },
    "50": { "A": 0.0007, "H": 0.0048, "I": 0.0001, "L": 0.0019, "N": 0.0001, "P": 0.986, "Q": 0.0001, "R": 0.0008, "S": 0.0013, "T": 0.0041, "V": 0.0001, "Y": 0.0001 },
    "51": { "A": 0.0001, "E": 0.0003, "F": 0.0001, "G": 0.0024, "I": 0.0005, "K": 0.0046, "L": 0.0005, "M": 0.0044, "N": 0.0003, "P": 0.0001, "Q": 0.0003, "R": 0.9704, "S": 0.012, "T": 0.0032, "V": 0.0002, "W": 0.0008, "Y": 0.0001 },
    "52": { "C": 0.0002, "D": 0.0001, "F": 0.0019, "H": 0.0032, "I": 0.0038, "L": 0.9659, "P": 0.0103, "R": 0.011, "S": 0.0001, "T": 0.0001, "V": 0.0034, "Y": 0.0001 },
    "53": { "C": 0.0002, "F": 0.0015, "G": 0.0001, "H": 0.0018, "I": 0.0048, "L": 0.9745, "P": 0.0036, "Q": 0.0001, "R": 0.0074, "S": 0.0002, "T": 0.0001, "V": 0.0058 },
    "54": { "C": 0.0001, "F": 0.0035, "G": 0.0001, "H": 0.0001, "I": 0.9563, "L": 0.0101, "M": 0.0079, "N": 0.0029, "P": 0.0001, "Q": 0.0001, "R": 0.0002, "S": 0.0088, "T": 0.004, "V": 0.0058 },
    "55": { "A": 0.0002, "C": 0.0025, "D": 0.0102, "E": 0.0002, "F": 0.0146, "H": 0.0101, "I": 0.0001, "K": 0.0002, "L": 0.0005, "N": 0.0026, "P": 0.0001, "Q": 0.0005, "R": 0.0001, "S": 0.0147, "T": 0.0003, "V": 0.0005, "W": 0.0002, "Y": 0.9421 },
    "56": { "A": 0.0166, "C": 0.0034, "D": 0.4645, "E": 0.0099, "F": 0.0003, "G": 0.4838, "H": 0.0035, "I": 0.0003, "K": 0.0001, "L": 0.0001, "N": 0.0023, "P": 0.0001, "Q": 0.0001, "R": 0.0025, "S": 0.0038, "T": 0.0007, "V": 0.0044, "W": 0.0001, "Y": 0.0037 },
    "57": { "A": 0.9491, "D": 0.0003, "E": 0.003, "F": 0.0001, "G": 0.0025, "I": 0.0009, "K": 0.0001, "L": 0.0003, "M": 0.0001, "P": 0.0017, "Q": 0.0001, "R": 0.0002, "S": 0.011, "T": 0.0227, "V": 0.0079 },
    "65": { "A": 0.0122, "C": 0.0007, "F": 0.006, "G": 0.0002, "I": 0.0005, "L": 0.0007, "N": 0.0002, "P": 0.0036, "S": 0.9665, "T": 0.0036, "V": 0.0004, "W": 0.0001, "Y": 0.0053 },
    "66": { "A": 0.0011, "C": 0.0003, "D": 0.0032, "E": 0.0001, "F": 0.0009, "G": 0.0033, "H": 0.0044, "I": 0.0087, "K": 0.0134, "L": 0.0009, "M": 0.0002, "N": 0.4564, "P": 0.0007, "Q": 0.0007, "R": 0.0123, "S": 0.4595, "T": 0.0294, "V": 0.0006, "W": 0.0001, "Y": 0.0037 },
    "67": { "E": 0.0002, "G": 0.0045, "I": 0.0001, "K": 0.0029, "L": 0.0003, "M": 0.0062, "P": 0.0001, "Q": 0.0002, "R": 0.9704, "S": 0.0118, "T": 0.002, "W": 0.0013 },
    "68": { "A": 0.9734, "D": 0.0059, "E": 0.0002, "F": 0.0002, "G": 0.0019, "H": 0.0001, "I": 0.0001, "L": 0.0002, "N": 0.0001, "P": 0.0042, "Q": 0.0001, "S": 0.0054, "T": 0.0037, "V": 0.0044, "Y": 0.0003 },
    "69": { "A": 0.0153, "D": 0.0001, "G": 0.0006, "H": 0.0003, "I": 0.0037, "K": 0.0011, "L": 0.0001, "M": 0.0001, "N": 0.0059, "P": 0.0144, "Q": 0.0001, "R": 0.0007, "S": 0.0089, "T": 0.9486, "V": 0.0001, "Y": 0.0002 },
    "70": { "A": 0.004, "C": 0.0059, "D": 0.0108, "E": 0.0001, "G": 0.9702, "H": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.0012, "S": 0.0024, "T": 0.0001, "V": 0.0048, "W": 0.0001, "Y": 0.0001 },
    "71": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.0013, "G": 0.0002, "H": 0.0003, "I": 0.9409, "L": 0.0115, "M": 0.0009, "N": 0.0014, "P": 0.0003, "R": 0.0001, "S": 0.0113, "T": 0.0099, "V": 0.0217 },
    "72": { "A": 0.0028, "G": 0.0001, "H": 0.0003, "I": 0.0001, "K": 0.0012, "L": 0.0012, "P": 0.9817, "Q": 0.0031, "R": 0.001, "S": 0.0047, "T": 0.0038 },
    "74": { "A": 0.4636, "D": 0.5079, "E": 0.0046, "F": 0.0001, "G": 0.0052, "H": 0.0007, "K": 0.0001, "L": 0.0001, "N": 0.0009, "P": 0.0034, "R": 0.0003, "S": 0.0031, "T": 0.0017, "V": 0.0044, "Y": 0.0038 },
    "75": { "C": 0.0001, "F": 0.0002, "G": 0.0026, "I": 0.0001, "K": 0.0029, "L": 0.0003, "M": 0.0053, "N": 0.0001, "P": 0.0001, "R": 0.9769, "S": 0.0089, "T": 0.0013, "V": 0.0001, "W": 0.001 },
    "76": { "C": 0.0172, "F": 0.9574, "G": 0.0004, "I": 0.003, "L": 0.0094, "M": 0.0001, "R": 0.0001, "S": 0.0015, "V": 0.0096, "W": 0.0003, "Y": 0.0009 },
    "77": { "A": 0.0003, "C": 0.0009, "D": 0.0003, "E": 0.0001, "G": 0.0049, "H": 0.0001, "I": 0.0054, "L": 0.0001, "M": 0.0001, "N": 0.0018, "P": 0.0001, "R": 0.0168, "S": 0.9634, "T": 0.0054, "V": 0.0003, "Y": 0.0001 },
    "78": { "A": 0.0053, "C": 0.0054, "D": 0.0018, "G": 0.9789, "L": 0.0001, "P": 0.0002, "R": 0.0009, "S": 0.003, "T": 0.0001, "V": 0.0042, "W": 0.0001 },
    "79": { "A": 0.0003, "C": 0.0005, "D": 0.0001, "F": 0.0004, "G": 0.0105, "H": 0.0001, "I": 0.0027, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0019, "P": 0.0001, "R": 0.0123, "S": 0.9683, "T": 0.0024, "V": 0.0001, "Y": 0.0003 },
    "80": { "A": 0.0046, "C": 0.0001, "D": 0.0001, "E": 0.0038, "F": 0.0001, "G": 0.9702, "H": 0.0002, "K": 0.0001, "L": 0.0003, "Q": 0.0001, "R": 0.0036, "S": 0.0003, "V": 0.01, "W": 0.0066, "Y": 0.0001 },
    "83": { "A": 0.0095, "C": 0.003, "D": 0.001, "E": 0.0001, "F": 0.0034, "G": 0.0003, "H": 0.0001, "P": 0.0009, "S": 0.9677, "T": 0.0018, "V": 0.0002, "W": 0.0002, "Y": 0.0117 },
    "84": { "A": 0.0047, "C": 0.0005, "D": 0.0004, "E": 0.0092, "F": 0.0001, "G": 0.9655, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0033, "S": 0.0003, "T": 0.0001, "V": 0.0094, "W": 0.0059 },
    "85": { "A": 0.0064, "E": 0.0002, "I": 0.0021, "K": 0.0067, "L": 0.0001, "M": 0.0001, "N": 0.0002, "P": 0.0023, "Q": 0.0001, "R": 0.0019, "S": 0.0023, "T": 0.9776 },
    "86": { "A": 0.0033, "C": 0.0001, "D": 0.97, "E": 0.0093, "F": 0.0001, "G": 0.0036, "H": 0.0018, "K": 0.0002, "L": 0.0002, "N": 0.0025, "Q": 0.0001, "T": 0.0001, "V": 0.0029, "Y": 0.0059 },
    "87": { "C": 0.0046, "D": 0.0002, "F": 0.9763, "G": 0.0001, "I": 0.0017, "L": 0.0058, "R": 0.0001, "S": 0.0013, "T": 0.0001, "V": 0.008, "Y": 0.0019 },
    "88": { "A": 0.0046, "D": 0.0001, "F": 0.0001, "G": 0.0001, "I": 0.0056, "K": 0.0002, "L": 0.0003, "N": 0.0036, "P": 0.0042, "R": 0.0004, "S": 0.019, "T": 0.9614, "V": 0.0001, "Y": 0.0001 },
    "89": { "F": 0.0026, "H": 0.0015, "I": 0.0064, "L": 0.9793, "P": 0.0023, "Q": 0.0001, "R": 0.0063, "S": 0.0004, "T": 0.0002, "V": 0.0009 },
    "90": { "A": 0.0035, "D": 0.0001, "F": 0.0001, "I": 0.0058, "K": 0.0001, "L": 0.0004, "N": 0.0036, "P": 0.0047, "R": 0.0001, "S": 0.009, "T": 0.972, "V": 0.0006, "Y": 0.0001 },
    "91": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.0011, "G": 0.0001, "I": 0.974, "L": 0.0041, "M": 0.0006, "N": 0.0011, "R": 0.0001, "S": 0.0133, "T": 0.0015, "V": 0.0038 },
    "92": { "A": 0.0009, "C": 0.0013, "D": 0.002, "F": 0.0001, "G": 0.0032, "H": 0.0002, "I": 0.006, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0103, "P": 0.0001, "R": 0.0105, "S": 0.9554, "T": 0.0095, "Y": 0.0001 },
    "93": { "A": 0.0003, "C": 0.0009, "D": 0.0002, "E": 0.0001, "G": 0.0094, "H": 0.0002, "I": 0.0064, "K": 0.0011, "L": 0.0003, "M": 0.0002, "N": 0.0044, "R": 0.5076, "S": 0.4654, "T": 0.003, "V": 0.0001, "Y": 0.0002 },
    "94": { "E": 0.0001, "G": 0.0001, "I": 0.003, "L": 0.9732, "M": 0.0014, "P": 0.0011, "Q": 0.0009, "R": 0.0152, "V": 0.0049 },
    "95": { "A": 0.0052, "D": 0.0122, "E": 0.9622, "G": 0.0086, "H": 0.0002, "K": 0.0017, "L": 0.0001, "Q": 0.0051, "R": 0.0002, "S": 0.0001, "V": 0.004, "W": 0.0001, "Y": 0.0002 },
    "96": { "A": 0.0091, "D": 0.0004, "E": 0.0001, "F": 0.0001, "H": 0.0066, "K": 0.0001, "L": 0.0019, "N": 0.0003, "P": 0.966, "Q": 0.0005, "R": 0.0017, "S": 0.0075, "T": 0.0056, "V": 0.0002, "Y": 0.0001 },
    "97": { "A": 0.0029, "D": 0.0101, "E": 0.9793, "G": 0.0036, "H": 0.0001, "K": 0.0015, "L": 0.0001, "Q": 0.0011, "R": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.0011 },
    "98": { "A": 0.0026, "D": 0.9809, "E": 0.0091, "G": 0.0013, "H": 0.0002, "N": 0.0011, "S": 0.0001, "V": 0.0015, "Y": 0.0034 },
    "99": { "A": 0.0004, "C": 0.0107, "D": 0.0003, "F": 0.9429, "G": 0.0003, "I": 0.002, "L": 0.0081, "M": 0.0001, "R": 0.0001, "S": 0.0152, "V": 0.0165, "Y": 0.0034 },
    "100": { "A": 0.9779, "E": 0.004, "G": 0.0054, "I": 0.0001, "P": 0.0004, "S": 0.0036, "T": 0.0025, "V": 0.006, "Y": 0.0001 },
    "101": { "A": 0.0019, "C": 0.0001, "D": 0.0004, "E": 0.0007, "F": 0.0046, "G": 0.005, "I": 0.0076, "K": 0.0001, "L": 0.0106, "M": 0.0042, "N": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0003, "T": 0.0032, "V": 0.9608, "W": 0.0001 },
    "102": { "A": 0.0001, "C": 0.0011, "D": 0.0052, "E": 0.0002, "F": 0.0032, "H": 0.0005, "K": 0.0001, "L": 0.0001, "N": 0.0007, "S": 0.0013, "V": 0.0001, "W": 0.0001, "Y": 0.9875 },
    "103": { "C": 0.0038, "D": 0.0046, "F": 0.0156, "H": 0.0047, "I": 0.0001, "L": 0.0003, "N": 0.0009, "R": 0.0001, "S": 0.005, "T": 0.0001, "V": 0.0002, "W": 0.0002, "Y": 0.9645 },
    "104": { "A": 0.0001, "C": 0.9763, "F": 0.0023, "G": 0.005, "L": 0.0003, "N": 0.0001, "P": 0.0001, "R": 0.0019, "S": 0.0007, "V": 0.0003, "W": 0.0106, "Y": 0.0023 },
    "105": { "A": 0.0001, "E": 0.0015, "F": 0.0001, "G": 0.0001, "H": 0.0162, "K": 0.0031, "L": 0.0044, "M": 0.0001, "N": 0.0002, "P": 0.0017, "Q": 0.9691, "R": 0.0029, "S": 0.0004, "Y": 0.0001 },
    "106": { "A": 0.0001, "C": 0.0002, "D": 0.0001, "E": 0.0031, "F": 0.0003, "G": 0.0006, "H": 0.0296, "I": 0.0002, "K": 0.006, "L": 0.0055, "M": 0.0001, "N": 0.0006, "P": 0.0021, "Q": 0.9399, "R": 0.0086, "S": 0.0005, "T": 0.0001, "V": 0.0006, "W": 0.0001, "Y": 0.0017 },
    "107": { "A": 0.0007, "C": 0.0059, "D": 0.0132, "E": 0.0027, "F": 0.007, "G": 0.0042, "H": 0.0076, "K": 0.0003, "L": 0.0035, "M": 0.0001, "N": 0.0015, "P": 0.0025, "Q": 0.001, "R": 0.4573, "S": 0.0067, "T": 0.0008, "V": 0.0007, "W": 0.0005, "Y": 0.4838 },
    "108": { "A": 0.0089, "C": 0.0046, "D": 0.0176, "E": 0.002, "F": 0.0015, "G": 0.4797, "H": 0.0029, "I": 0.0027, "K": 0.0016, "L": 0.0014, "M": 0.0001, "N": 0.0119, "P": 0.0003, "Q": 0.0005, "R": 0.0093, "S": 0.4336, "T": 0.0078, "V": 0.0088, "W": 0.0004, "Y": 0.0044 },
    "109": { "A": 0.0042, "C": 0.0004, "D": 0.0154, "E": 0.0014, "F": 0.0008, "G": 0.0122, "H": 0.004, "I": 0.0074, "K": 0.0062, "L": 0.002, "M": 0.0004, "N": 0.44, "P": 0.0023, "Q": 0.0003, "R": 0.0162, "S": 0.4561, "T": 0.0247, "V": 0.0015, "W": 0.0007, "Y": 0.0038 },
    "110": { "A": 0.0039, "C": 0.0079, "E": 0.0039, "G": 0.0439, "L": 0.0033, "P": 0.002, "Q": 0.0007, "R": 0.0066, "S": 0.2182, "T": 0.0026, "V": 0.0007, "W": 0.7058, "Y": 0.0007 },
    "112": { "D": 0.0141, "E": 0.007, "G": 0.0423, "H": 0.0141, "I": 0.007, "K": 0.007, "L": 0.0282, "P": 0.7676, "Q": 0.0141, "R": 0.0211, "S": 0.0352, "T": 0.0282, "V": 0.007, "Y": 0.007 },
    "113": { "A": 0.0046, "C": 0.0039, "D": 0.0004, "E": 0.0009, "F": 0.0004, "G": 0.0074, "H": 0.0042, "I": 0.0009, "K": 0.0002, "L": 0.0072, "N": 0.0004, "P": 0.2464, "Q": 0.0021, "R": 0.007, "S": 0.3377, "T": 0.0042, "V": 0.0005, "W": 0.3715, "Y": 0.0005 },
    "114": { "A": 0.0076, "C": 0.0019, "D": 0.0029, "E": 0.0012, "F": 0.0043, "G": 0.0104, "H": 0.0036, "I": 0.0019, "K": 0.0013, "L": 0.0328, "M": 0.0003, "N": 0.06, "P": 0.3326, "Q": 0.0029, "R": 0.0183, "S": 0.33, "T": 0.0127, "V": 0.0014, "W": 0.1719, "Y": 0.0021 },
    "115": { "A": 0.0233, "C": 0.0005, "D": 0.0007, "E": 0.0144, "F": 0.0054, "G": 0.0503, "H": 0.0032, "I": 0.0083, "K": 0.0082, "L": 0.0756, "M": 0.0256, "N": 0.005, "P": 0.5236, "Q": 0.0172, "R": 0.051, "S": 0.1052, "T": 0.0216, "V": 0.0101, "W": 0.0499, "Y": 0.0009 },
    "116": { "A": 0.0139, "C": 0.0123, "D": 0.0065, "E": 0.0095, "F": 0.0815, "G": 0.0368, "H": 0.0105, "I": 0.0871, "K": 0.008, "L": 0.2299, "M": 0.0013, "N": 0.0111, "P": 0.0697, "Q": 0.0168, "R": 0.0774, "S": 0.0194, "T": 0.0056, "V": 0.0432, "W": 0.1236, "Y": 0.1359 },
    "117": { "A": 0.0178, "C": 0.0002, "D": 0.0009, "E": 0.0004, "F": 0.0023, "G": 0.005, "H": 0.0008, "I": 0.0039, "K": 0.0011, "L": 0.008, "M": 0.0005, "N": 0.004, "P": 0.0089, "Q": 0.0007, "R": 0.0034, "S": 0.0364, "T": 0.9022, "V": 0.0028, "W": 0.0003, "Y": 0.0003 },
    "118": { "C": 0.0072, "D": 0.0001, "F": 0.9793, "G": 0.0003, "I": 0.0005, "L": 0.0046, "N": 0.0001, "P": 0.0001, "R": 0.0003, "S": 0.0014, "T": 0.0001, "V": 0.0052, "W": 0.0006, "Y": 0.0005 },
    "119": { "A": 0.0013, "C": 0.0007, "D": 0.0015, "E": 0.0001, "G": 0.9934, "H": 0.0001, "I": 0.0001, "P": 0.0001, "R": 0.0007, "S": 0.0014, "T": 0.0001, "V": 0.0005, "W": 0.0001 },
    "120": { "A": 0.001, "C": 0.0001, "D": 0.0001, "E": 0.0023, "G": 0.2801, "H": 0.0078, "K": 0.0008, "L": 0.0011, "N": 0.0001, "P": 0.1171, "Q": 0.5837, "R": 0.005, "S": 0.0003, "T": 0.0002, "V": 0.0005, "W": 0.0001 },
    "121": { "A": 0.0006, "E": 0.0031, "G": 0.9884, "R": 0.0038, "T": 0.0001, "V": 0.0016, "W": 0.0024 },
    "122": { "A": 0.0044, "H": 0.0001, "I": 0.0009, "K": 0.0003, "N": 0.0003, "P": 0.0011, "S": 0.0021, "T": 0.9907, "V": 0.0001 },
    "123": { "E": 0.002, "G": 0.0001, "I": 0.0001, "K": 0.8499, "L": 0.001, "M": 0.0009, "N": 0.0036, "P": 0.0001, "Q": 0.0042, "R": 0.1334, "S": 0.0001, "T": 0.0044, "V": 0.0001, "W": 0.0001 },
    "124": { "A": 0.0005, "E": 0.0003, "G": 0.0004, "I": 0.0009, "L": 0.3359, "M": 0.0031, "P": 0.0001, "Q": 0.0003, "R": 0.0005, "V": 0.6579, "W": 0.0001 },
    "125": { "A": 0.001, "D": 0.1221, "E": 0.8689, "F": 0.0001, "G": 0.0027, "H": 0.0007, "K": 0.0015, "L": 0.0001, "N": 0.0003, "Q": 0.0016, "R": 0.0003, "T": 0.0001, "V": 0.0007, "Y": 0.0001 },
    "126": { "A": 0.0001, "C": 0.0001, "F": 0.003, "I": 0.9706, "K": 0.0001, "L": 0.0046, "M": 0.0051, "N": 0.0016, "R": 0.0001, "S": 0.0044, "T": 0.0023, "V": 0.008 },
    "127": { "E": 0.0018, "K": 0.9948, "N": 0.0006, "Q": 0.0028, "S": 0.0001 }
  },
  "IGKV3D": {
    "1": { "D": 0.0347, "E": 0.9653 },
    "2": { "A": 0.0007, "F": 0.0005, "G": 0.0001, "H": 0.0001, "I": 0.9058, "K": 0.0039, "L": 0.0356, "M": 0.0063, "N": 0.0006, "P": 0.0003, "R": 0.0067, "S": 0.0007, "T": 0.0154, "V": 0.0232 },
    "3": { "A": 0.0087, "D": 0.0005, "E": 0.0045, "F": 0.0009, "G": 0.0043, "I": 0.0012, "L": 0.0183, "M": 0.0038, "P": 0.0001, "Q": 0.0002, "R": 0.0002, "S": 0.0001, "T": 0.0001, "V": 0.9569, "W": 0.0002 },
    "4": { "A": 0.0001, "C": 0.0001, "F": 0.0024, "I": 0.0194, "K": 0.001, "L": 0.1082, "M": 0.857, "N": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0019, "S": 0.0005, "T": 0.0035, "V": 0.0054, "W": 0.0003 },
    "5": { "A": 0.0029, "H": 0.0001, "I": 0.0003, "K": 0.0045, "L": 0.0002, "M": 0.0046, "N": 0.0004, "P": 0.0208, "Q": 0.0006, "R": 0.0023, "S": 0.0034, "T": 0.9595, "V": 0.0001, "W": 0.0004 },
    "6": { "D": 0.0001, "E": 0.0004, "H": 0.0182, "K": 0.006, "L": 0.003, "M": 0.0001, "N": 0.0002, "P": 0.0222, "Q": 0.947, "R": 0.0019, "S": 0.0001, "T": 0.0007, "V": 0.0001 },
    "7": { "A": 0.005, "C": 0.0011, "D": 0.0003, "F": 0.0041, "H": 0.0001, "I": 0.0002, "K": 0.0001, "L": 0.0005, "M": 0.0001, "N": 0.0001, "P": 0.0132, "S": 0.9657, "T": 0.0043, "V": 0.0001, "Y": 0.0053 },
    "8": { "A": 0.0016, "G": 0.0001, "H": 0.0004, "I": 0.0001, "K": 0.0001, "L": 0.0057, "P": 0.9727, "Q": 0.0067, "R": 0.0021, "S": 0.0033, "T": 0.0073 },
    "9": { "A": 0.9449, "D": 0.0096, "E": 0.0002, "G": 0.0136, "I": 0.0001, "L": 0.0002, "P": 0.0058, "R": 0.0001, "S": 0.0061, "T": 0.0035, "V": 0.0159, "Y": 0.0001 },
    "10": { "A": 0.0045, "F": 0.0001, "H": 0.0003, "I": 0.0067, "K": 0.0001, "L": 0.0002, "M": 0.0004, "N": 0.0031, "P": 0.0127, "S": 0.0077, "T": 0.9641, "V": 0.0001 },
    "11": { "A": 0.0001, "C": 0.0001, "E": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.9778, "M": 0.0053, "P": 0.0037, "Q": 0.0017, "R": 0.0063, "S": 0.0001, "T": 0.0001, "V": 0.0042, "W": 0.0001 },
    "12": { "A": 0.0051, "C": 0.0005, "F": 0.0016, "H": 0.0002, "L": 0.0001, "P": 0.0078, "S": 0.9806, "T": 0.0014, "Y": 0.0028 },
    "13": { "A": 0.0079, "E": 0.0017, "F": 0.001, "G": 0.0049, "I": 0.0001, "K": 0.0001, "L": 0.0758, "M": 0.0027, "P": 0.0001, "R": 0.0001, "S": 0.0007, "V": 0.9043, "W": 0.0007 },
    "14": { "A": 0.0054, "C": 0.0011, "D": 0.0001, "F": 0.0047, "H": 0.0001, "L": 0.0001, "P": 0.0057, "S": 0.9775, "T": 0.0024, "V": 0.0001, "Y": 0.003 },
    "15": { "A": 0.0028, "F": 0.0001, "H": 0.0007, "K": 0.0001, "L": 0.0058, "N": 0.0001, "P": 0.9708, "Q": 0.0081, "R": 0.0013, "S": 0.0031, "T": 0.0069, "V": 0.0003 },
    "16": { "A": 0.0027, "C": 0.0001, "E": 0.0011, "G": 0.985, "R": 0.0039, "S": 0.0001, "V": 0.0048, "W": 0.0024 },
    "17": { "A": 0.0174, "C": 0.0001, "D": 0.0266, "E": 0.9388, "G": 0.0067, "H": 0.0001, "K": 0.003, "L": 0.0001, "N": 0.0006, "Q": 0.0042, "R": 0.0003, "S": 0.0001, "T": 0.0004, "V": 0.0017, "Y": 0.0001 },
    "18": { "A": 0.0005, "C": 0.0001, "E": 0.0002, "G": 0.0141, "I": 0.0032, "K": 0.0029, "L": 0.0004, "N": 0.0002, "P": 0.0003, "Q": 0.0003, "R": 0.9372, "S": 0.0283, "T": 0.011, "V": 0.0011, "W": 0.0003 },
    "19": { "A": 0.9497, "D": 0.004, "E": 0.0001, "F": 0.0001, "G": 0.0027, "I": 0.0003, "L": 0.0004, "P": 0.0035, "S": 0.0041, "T": 0.0035, "V": 0.0316 },
    "20": { "A": 0.0048, "H": 0.0003, "I": 0.0056, "K": 0.0001, "M": 0.0001, "N": 0.0051, "P": 0.0105, "R": 0.0001, "S": 0.0049, "T": 0.968, "V": 0.0005, "Y": 0.0001 },
    "21": { "A": 0.0001, "F": 0.0039, "H": 0.0029, "I": 0.0097, "L": 0.9639, "M": 0.0001, "P": 0.0098, "R": 0.0054, "S": 0.0005, "T": 0.0002, "V": 0.0036 },
    "22": { "A": 0.0164, "C": 0.0009, "D": 0.0003, "F": 0.003, "H": 0.0007, "I": 0.0001, "L": 0.0012, "N": 0.0001, "P": 0.0348, "R": 0.0001, "S": 0.9294, "T": 0.0088, "V": 0.0001, "Y": 0.0042 },
    "23": { "A": 0.0002, "C": 0.9782, "F": 0.0035, "G": 0.0069, "H": 0.0001, "L": 0.0001, "P": 0.0001, "R": 0.0031, "S": 0.0046, "T": 0.0001, "V": 0.0003, "W": 0.0009, "Y": 0.0021 },
    "24": { "A": 0.0002, "C": 0.0001, "E": 0.0012, "G": 0.0356, "I": 0.0001, "K": 0.0037, "L": 0.0002, "M": 0.0037, "P": 0.0002, "Q": 0.0003, "R": 0.9325, "S": 0.0138, "T": 0.0047, "V": 0.0003, "W": 0.0035 },
    "25": { "A": 0.9728, "D": 0.0041, "E": 0.0002, "G": 0.0009, "H": 0.0002, "P": 0.0053, "S": 0.0068, "T": 0.0068, "V": 0.0029, "Y": 0.0001 },
    "26": { "A": 0.0001, "C": 0.0013, "D": 0.0001, "G": 0.0039, "H": 0.0002, "I": 0.0031, "K": 0.0003, "M": 0.0001, "N": 0.0032, "P": 0.0001, "Q": 0.0001, "R": 0.015, "S": 0.9697, "T": 0.0028, "V": 0.0001 },
    "27": { "A": 0.0001, "D": 0.0006, "E": 0.009, "G": 0.0002, "H": 0.0087, "K": 0.0045, "L": 0.0031, "M": 0.0001, "N": 0.0002, "P": 0.0106, "Q": 0.955, "R": 0.0069, "S": 0.0001, "T": 0.0005, "Y": 0.0003 },
    "28": { "A": 0.0016, "C": 0.0038, "D": 0.0018, "E": 0.0001, "F": 0.0007, "G": 0.023, "H": 0.001, "I": 0.0053, "K": 0.0002, "L": 0.0003, "M": 0.0005, "N": 0.0146, "P": 0.0006, "Q": 0.0001, "R": 0.0333, "S": 0.8994, "T": 0.0126, "V": 0.0002, "Y": 0.0011 },
    "29": { "A": 0.0044, "C": 0.0002, "D": 0.0011, "F": 0.0056, "G": 0.0074, "I": 0.0502, "L": 0.0109, "M": 0.0003, "N": 0.0001, "P": 0.0001, "R": 0.0001, "S": 0.0009, "T": 0.0002, "V": 0.9183, "Y": 0.0001 },
    "30": { "A": 0.0029, "C": 0.0044, "D": 0.0044, "F": 0.0015, "G": 0.016, "H": 0.0015, "I": 0.0044, "K": 0.0015, "L": 0.0029, "M": 0.0044, "N": 0.0291, "P": 0.0044, "R": 0.0262, "S": 0.8474, "T": 0.0363, "V": 0.0015, "W": 0.0102, "Y": 0.0015 },
    "36": { "A": 0.0052, "C": 0.0015, "D": 0.0063, "E": 0.0003, "F": 0.0011, "G": 0.0322, "H": 0.0019, "I": 0.0065, "K": 0.002, "L": 0.0005, "M": 0.0005, "N": 0.0352, "P": 0.0003, "R": 0.0261, "S": 0.8655, "T": 0.0114, "V": 0.0005, "W": 0.0003, "Y": 0.0029 },
    "37": { "A": 0.0026, "C": 0.0012, "D": 0.0048, "E": 0.0003, "F": 0.0011, "G": 0.0093, "H": 0.0007, "I": 0.0095, "K": 0.0017, "L": 0.0003, "N": 0.0353, "P": 0.0005, "Q": 0.0002, "R": 0.0224, "S": 0.8763, "T": 0.0306, "V": 0.001, "W": 0.0001, "Y": 0.002 },
    "38": { "A": 0.0004, "C": 0.0005, "D": 0.014, "E": 0.0005, "F": 0.0034, "G": 0.0003, "H": 0.0174, "I": 0.0015, "K": 0.0172, "L": 0.0005, "M": 0.0002, "N": 0.8478, "P": 0.0004, "Q": 0.0009, "R": 0.0017, "S": 0.0148, "T": 0.0109, "V": 0.0003, "W": 0.0001, "Y": 0.0673 },
    "39": { "A": 0.0001, "C": 0.0002, "E": 0.0001, "F": 0.016, "G": 0.0001, "I": 0.0031, "L": 0.9553, "M": 0.0008, "Q": 0.0001, "R": 0.0001, "S": 0.0033, "V": 0.0204, "W": 0.0003, "Y": 0.0002 },
    "40": { "A": 0.9629, "D": 0.0069, "E": 0.0002, "F": 0.0002, "G": 0.0034, "H": 0.0001, "L": 0.0004, "M": 0.0001, "N": 0.0001, "P": 0.0037, "R": 0.0001, "S": 0.0088, "T": 0.0045, "V": 0.0085, "Y": 0.0001 },
    "41": { "C": 0.008, "E": 0.0001, "G": 0.0071, "L": 0.0039, "R": 0.0035, "S": 0.0013, "T": 0.0001, "V": 0.0001, "W": 0.9758 },
    "42": { "C": 0.0013, "D": 0.0029, "E": 0.0001, "F": 0.006, "H": 0.0025, "I": 0.0001, "L": 0.0008, "N": 0.001, "P": 0.0001, "Q": 0.0001, "S": 0.0056, "T": 0.0001, "Y": 0.9796 },
    "43": { "A": 0.0001, "D": 0.0001, "E": 0.0005, "F": 0.0001, "H": 0.0069, "K": 0.0047, "L": 0.0015, "N": 0.0001, "P": 0.0047, "Q": 0.9774, "R": 0.0037, "T": 0.0001, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "44": { "A": 0.0002, "C": 0.0001, "E": 0.0007, "F": 0.0001, "H": 0.0183, "K": 0.0062, "L": 0.0031, "M": 0.0001, "N": 0.0001, "P": 0.0031, "Q": 0.9639, "R": 0.0033, "S": 0.0001, "T": 0.0003, "V": 0.0003, "W": 0.0001, "Y": 0.0003 },
    "45": { "A": 0.0002, "D": 0.0001, "E": 0.0019, "G": 0.0001, "H": 0.0006, "I": 0.0032, "K": 0.9332, "L": 0.0001, "M": 0.0003, "N": 0.013, "P": 0.0001, "Q": 0.0111, "R": 0.0227, "S": 0.0003, "T": 0.013, "V": 0.0001, "W": 0.0001 },
    "46": { "A": 0.0041, "C": 0.0001, "D": 0.0003, "F": 0.001, "G": 0.0011, "H": 0.0052, "L": 0.0036, "N": 0.0001, "P": 0.9708, "Q": 0.0003, "R": 0.0038, "S": 0.0037, "T": 0.0048, "V": 0.0011 },
    "47": { "A": 0.0025, "C": 0.006, "D": 0.0033, "E": 0.0002, "G": 0.9805, "H": 0.0002, "L": 0.0001, "P": 0.0001, "R": 0.0015, "S": 0.002, "T": 0.0001, "V": 0.0036, "W": 0.0001 },
    "48": { "E": 0.0013, "H": 0.0054, "K": 0.006, "L": 0.0349, "N": 0.0001, "P": 0.0056, "Q": 0.9406, "R": 0.0057, "T": 0.0003, "V": 0.0001, "Y": 0.0001 },
    "49": { "A": 0.9474, "D": 0.0038, "E": 0.0004, "G": 0.0049, "H": 0.0001, "I": 0.0001, "L": 0.0002, "P": 0.0194, "R": 0.0002, "S": 0.0104, "T": 0.0087, "V": 0.0041, "Y": 0.0001 },
    "50": { "A": 0.0015, "H": 0.0047, "L": 0.002, "P": 0.9862, "R": 0.0008, "S": 0.0017, "T": 0.0031 },
    "51": { "A": 0.0003, "F": 0.0001, "G": 0.0026, "I": 0.0004, "K": 0.0059, "L": 0.0004, "M": 0.0053, "N": 0.0001, "P": 0.0002, "Q": 0.0005, "R": 0.9662, "S": 0.0138, "T": 0.0033, "V": 0.0001, "W": 0.001 },
    "52": { "C": 0.0001, "F": 0.0023, "H": 0.004, "I": 0.0041, "L": 0.962, "M": 0.0001, "P": 0.0125, "Q": 0.0001, "R": 0.0114, "S": 0.0001, "T": 0.0001, "V": 0.0032, "W": 0.0001 },
    "53": { "F": 0.0017, "H": 0.0012, "I": 0.0043, "L": 0.9724, "M": 0.0001, "P": 0.0055, "R": 0.0089, "S": 0.0003, "T": 0.0001, "V": 0.0056 },
    "54": { "A": 0.0001, "C": 0.0001, "F": 0.0031, "G": 0.0001, "H": 0.0001, "I": 0.9508, "L": 0.0141, "M": 0.0087, "N": 0.0029, "R": 0.0004, "S": 0.0082, "T": 0.0041, "V": 0.0072, "Y": 0.0001 },
    "55": { "A": 0.0001, "C": 0.0021, "D": 0.0084, "E": 0.0003, "F": 0.0154, "H": 0.0102, "I": 0.0001, "K": 0.0001, "L": 0.0007, "N": 0.0019, "P": 0.0003, "Q": 0.0003, "R": 0.0004, "S": 0.0172, "T": 0.0001, "V": 0.0001, "W": 0.0001, "Y": 0.9422 },
    "56": { "A": 0.0148, "C": 0.0064, "D": 0.0848, "E": 0.0038, "F": 0.0005, "G": 0.8658, "H": 0.0014, "I": 0.0001, "L": 0.0002, "N": 0.0011, "R": 0.0057, "S": 0.0061, "T": 0.0008, "V": 0.0073, "W": 0.0001, "Y": 0.0011 },
    "57": { "A": 0.9606, "E": 0.0035, "G": 0.003, "I": 0.0011, "K": 0.0002, "L": 0.0001, "P": 0.0019, "R": 0.0002, "S": 0.0099, "T": 0.013, "V": 0.0065 },
    "65": { "A": 0.0146, "C": 0.0007, "D": 0.0002, "F": 0.0058, "G": 0.0001, "I": 0.0003, "L": 0.0003, "N": 0.0001, "P": 0.0045, "R": 0.0001, "S": 0.9619, "T": 0.0057, "V": 0.0007, "W": 0.0003, "Y": 0.0048 },
    "66": { "A": 0.0081, "C": 0.0002, "D": 0.0007, "E": 0.0001, "F": 0.0007, "G": 0.0007, "H": 0.0003, "I": 0.0143, "K": 0.0021, "L": 0.0005, "M": 0.0009, "N": 0.025, "P": 0.0076, "Q": 0.0001, "R": 0.0019, "S": 0.0451, "T": 0.8899, "V": 0.0015, "Y": 0.0005 },
    "67": { "A": 0.0001, "G": 0.0035, "I": 0.0001, "K": 0.0033, "L": 0.0003, "M": 0.0058, "N": 0.0001, "P": 0.0001, "Q": 0.0002, "R": 0.9714, "S": 0.0108, "T": 0.0026, "V": 0.0001, "W": 0.0015 },
    "68": { "A": 0.9685, "D": 0.0079, "E": 0.0001, "F": 0.0003, "G": 0.0019, "H": 0.0002, "I": 0.0002, "L": 0.0002, "M": 0.0001, "P": 0.0048, "Q": 0.0001, "R": 0.0001, "S": 0.0055, "T": 0.0045, "V": 0.0057 },
    "69": { "A": 0.0136, "D": 0.0003, "F": 0.0001, "G": 0.0003, "H": 0.0003, "I": 0.0033, "K": 0.0004, "M": 0.0004, "N": 0.0057, "P": 0.0133, "Q": 0.0001, "R": 0.0003, "S": 0.0107, "T": 0.9509, "V": 0.0002 },
    "70": { "A": 0.0047, "C": 0.0039, "D": 0.0159, "E": 0.0006, "F": 0.0001, "G": 0.9596, "H": 0.0001, "I": 0.0001, "N": 0.0013, "P": 0.0001, "Q": 0.0001, "R": 0.0013, "S": 0.0058, "T": 0.0005, "V": 0.0058, "Y": 0.0002 },
    "71": { "A": 0.0002, "C": 0.0002, "D": 0.0001, "F": 0.0052, "G": 0.0006, "H": 0.0001, "I": 0.9176, "L": 0.0079, "M": 0.0017, "N": 0.0021, "P": 0.0001, "R": 0.0005, "S": 0.0138, "T": 0.0097, "V": 0.0403, "Y": 0.0001 },
    "72": { "A": 0.0044, "H": 0.0001, "I": 0.0001, "K": 0.0006, "L": 0.0007, "P": 0.98, "Q": 0.0042, "R": 0.0007, "S": 0.0054, "T": 0.0037, "V": 0.0001 },
    "74": { "A": 0.9101, "C": 0.0001, "D": 0.0534, "E": 0.0006, "F": 0.0001, "G": 0.0085, "H": 0.0003, "I": 0.0005, "L": 0.0004, "N": 0.0003, "P": 0.0081, "R": 0.0001, "S": 0.0052, "T": 0.0041, "V": 0.0078, "W": 0.0001, "Y": 0.0005 },
    "75": { "A": 0.0001, "E": 0.0002, "G": 0.0038, "I": 0.0001, "K": 0.0037, "M": 0.0055, "N": 0.0003, "P": 0.0001, "Q": 0.0002, "R": 0.9718, "S": 0.0105, "T": 0.0029, "V": 0.0001, "W": 0.0007 },
    "76": { "A": 0.0001, "C": 0.0169, "F": 0.957, "G": 0.0003, "I": 0.0037, "L": 0.0097, "M": 0.0001, "S": 0.0021, "V": 0.009, "W": 0.0001, "Y": 0.0011 },
    "77": { "A": 0.0002, "C": 0.0013, "D": 0.0001, "F": 0.0003, "G": 0.0038, "H": 0.0001, "I": 0.0057, "L": 0.0002, "M": 0.0003, "N": 0.0015, "P": 0.0001, "R": 0.0202, "S": 0.9595, "T": 0.0061, "V": 0.0005, "W": 0.0001, "Y": 0.0001 },
    "78": { "A": 0.0062, "C": 0.0057, "D": 0.002, "E": 0.0001, "F": 0.0001, "G": 0.9755, "H": 0.0001, "M": 0.0001, "R": 0.0011, "S": 0.0046, "V": 0.0045, "W": 0.0001, "Y": 0.0001 },
    "79": { "A": 0.0007, "C": 0.0009, "D": 0.0003, "F": 0.0001, "G": 0.0122, "H": 0.0001, "I": 0.0019, "L": 0.0001, "N": 0.0017, "R": 0.0154, "S": 0.9619, "T": 0.0041, "V": 0.0002, "W": 0.0002, "Y": 0.0002 },
    "80": { "A": 0.0042, "C": 0.0005, "D": 0.0001, "E": 0.0045, "G": 0.9669, "K": 0.0002, "L": 0.0004, "Q": 0.0002, "R": 0.0049, "S": 0.0007, "V": 0.0088, "W": 0.0087 },
    "83": { "A": 0.0087, "C": 0.0046, "D": 0.0008, "F": 0.0048, "G": 0.0003, "H": 0.0002, "I": 0.0001, "L": 0.0001, "N": 0.0002, "P": 0.0091, "S": 0.9547, "T": 0.0015, "V": 0.0001, "W": 0.0005, "Y": 0.0144 },
    "84": { "A": 0.0049, "C": 0.0005, "D": 0.0008, "E": 0.0093, "G": 0.9648, "K": 0.0001, "L": 0.0001, "M": 0.0001, "P": 0.0001, "Q": 0.0002, "R": 0.004, "S": 0.0002, "T": 0.0001, "V": 0.0076, "W": 0.0073 },
    "85": { "A": 0.0047, "E": 0.0001, "G": 0.0001, "I": 0.0029, "K": 0.011, "L": 0.0001, "N": 0.0002, "P": 0.0023, "Q": 0.0002, "R": 0.0041, "S": 0.0029, "T": 0.9714, "V": 0.0001 },
    "86": { "A": 0.0057, "D": 0.0926, "E": 0.889, "G": 0.0031, "H": 0.0003, "K": 0.002, "L": 0.0002, "N": 0.0008, "Q": 0.0023, "R": 0.0002, "S": 0.0001, "V": 0.0025, "Y": 0.0011 },
    "87": { "C": 0.0059, "D": 0.0001, "F": 0.9768, "I": 0.0005, "L": 0.0056, "S": 0.0018, "T": 0.0001, "V": 0.0057, "Y": 0.0035 },
    "88": { "A": 0.0041, "F": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0049, "K": 0.0002, "L": 0.0001, "M": 0.0001, "N": 0.0044, "P": 0.0071, "R": 0.0001, "S": 0.0103, "T": 0.9685, "V": 0.0001 },
    "89": { "F": 0.0025, "H": 0.0014, "I": 0.0073, "L": 0.9804, "M": 0.0001, "N": 0.0001, "P": 0.0021, "R": 0.0053, "S": 0.0001, "T": 0.0001, "V": 0.0007, "Y": 0.0001 },
    "90": { "A": 0.0031, "D": 0.0001, "F": 0.0001, "H": 0.0001, "I": 0.0063, "K": 0.0003, "L": 0.0003, "N": 0.0033, "P": 0.0052, "R": 0.0001, "S": 0.0082, "T": 0.9722, "V": 0.0005, "Y": 0.0001 },
    "91": { "F": 0.0015, "G": 0.0001, "H": 0.0001, "I": 0.9738, "K": 0.0001, "L": 0.0048, "M": 0.0006, "N": 0.0013, "R": 0.0003, "S": 0.0121, "T": 0.0022, "V": 0.0031 },
    "92": { "A": 0.0009, "C": 0.0008, "D": 0.0017, "E": 0.0001, "F": 0.0001, "G": 0.0037, "H": 0.0001, "I": 0.0071, "K": 0.0002, "L": 0.0003, "N": 0.0104, "P": 0.0001, "R": 0.0077, "S": 0.9541, "T": 0.0123, "V": 0.0001, "Y": 0.0002 },
    "93": { "A": 0.0006, "C": 0.0019, "D": 0.0008, "G": 0.0116, "H": 0.0001, "I": 0.0092, "K": 0.0003, "L": 0.0003, "M": 0.0002, "N": 0.008, "P": 0.0001, "R": 0.0513, "S": 0.9121, "T": 0.0031, "V": 0.0003, "Y": 0.0001 },
    "94": { "F": 0.0001, "I": 0.0016, "L": 0.9414, "M": 0.0298, "P": 0.0017, "Q": 0.0008, "R": 0.0178, "S": 0.0001, "T": 0.0003, "V": 0.0064, "W": 0.0001 },
    "95": { "A": 0.0005, "D": 0.0003, "E": 0.0631, "F": 0.0001, "G": 0.0005, "H": 0.0057, "K": 0.0029, "L": 0.0031, "N": 0.0001, "P": 0.003, "Q": 0.9167, "R": 0.0037, "T": 0.0001, "V": 0.0004 },
    "96": { "A": 0.0072, "C": 0.0003, "D": 0.0002, "E": 0.0002, "F": 0.0023, "H": 0.0003, "L": 0.0006, "P": 0.0682, "Q": 0.0001, "R": 0.0002, "S": 0.9123, "T": 0.0021, "V": 0.0001, "Y": 0.006 },
    "97": { "A": 0.0041, "D": 0.0095, "E": 0.9796, "G": 0.0025, "K": 0.001, "N": 0.0001, "P": 0.0001, "Q": 0.0011, "R": 0.0001, "S": 0.0003, "V": 0.0015, "Y": 0.0002 },
    "98": { "A": 0.0028, "C": 0.0001, "D": 0.9801, "E": 0.0082, "G": 0.0017, "H": 0.0004, "I": 0.0001, "K": 0.0001, "N": 0.0007, "P": 0.0001, "R": 0.0001, "S": 0.0002, "V": 0.0012, "Y": 0.0044 },
    "99": { "A": 0.0012, "C": 0.0116, "D": 0.0001, "F": 0.937, "G": 0.0003, "I": 0.0027, "L": 0.0095, "R": 0.0001, "S": 0.0167, "T": 0.0001, "V": 0.0177, "W": 0.0001, "Y": 0.0029 },
    "100": { "A": 0.9758, "C": 0.0003, "D": 0.0002, "E": 0.0032, "G": 0.0075, "L": 0.0001, "P": 0.0006, "Q": 0.0001, "R": 0.0001, "S": 0.0041, "T": 0.0024, "V": 0.0057, "W": 0.0001 },
    "101": { "A": 0.0028, "D": 0.0015, "E": 0.0004, "F": 0.0086, "G": 0.0051, "I": 0.0245, "L": 0.0128, "M": 0.0006, "N": 0.0001, "P": 0.0001, "S": 0.0011, "T": 0.0051, "V": 0.9372, "Y": 0.0001 },
    "102": { "A": 0.0001, "C": 0.0009, "D": 0.0093, "E": 0.0006, "F": 0.0037, "G": 0.0001, "H": 0.0009, "I": 0.0003, "K": 0.0001, "L": 0.0002, "N": 0.0009, "S": 0.0037, "V": 0.0002, "Y": 0.9792 },
    "103": { "A": 0.0001, "C": 0.0065, "D": 0.0056, "E": 0.0001, "F": 0.021, "G": 0.0001, "H": 0.0051, "I": 0.0005, "L": 0.0002, "N": 0.0005, "P": 0.0001, "R": 0.0001, "S": 0.0083, "T": 0.0006, "V": 0.0004, "W": 0.0002, "Y": 0.9507 },
    "104": { "C": 0.9762, "F": 0.0018, "G": 0.0055, "H": 0.0001, "I": 0.0001, "L": 0.0001, "N": 0.0001, "R": 0.0019, "S": 0.0014, "T": 0.0003, "V": 0.0006, "W": 0.0093, "Y": 0.0026 },
    "105": { "E": 0.0011, "H": 0.0147, "K": 0.0039, "L": 0.0046, "N": 0.0003, "P": 0.0021, "Q": 0.9693, "R": 0.0027, "S": 0.0005, "V": 0.0003, "W": 0.0002, "Y": 0.0003 },
    "106": { "A": 0.0003, "C": 0.0001, "E": 0.0024, "F": 0.0004, "G": 0.0001, "H": 0.0238, "K": 0.0078, "L": 0.0022, "M": 0.0001, "N": 0.0004, "P": 0.0027, "Q": 0.9534, "R": 0.0047, "S": 0.001, "T": 0.0001, "V": 0.0002, "Y": 0.0005 },
    "107": { "A": 0.0002, "C": 0.0035, "D": 0.0262, "E": 0.0035, "F": 0.0047, "G": 0.0012, "H": 0.0072, "I": 0.0002, "K": 0.0002, "L": 0.0009, "M": 0.0002, "N": 0.0041, "P": 0.0003, "Q": 0.0005, "R": 0.0201, "S": 0.0043, "T": 0.0002, "V": 0.0005, "W": 0.0002, "Y": 0.9219 },
    "108": { "A": 0.0013, "C": 0.0004, "D": 0.0249, "E": 0.0032, "F": 0.0007, "G": 0.04, "H": 0.0129, "I": 0.0038, "K": 0.0202, "L": 0.0008, "M": 0.0005, "N": 0.8381, "P": 0.0001, "Q": 0.0009, "R": 0.0024, "S": 0.029, "T": 0.005, "V": 0.0011, "W": 0.0003, "Y": 0.0141 },
    "109": { "A": 0.0018, "C": 0.0001, "D": 0.0289, "E": 0.0043, "F": 0.0007, "G": 0.0026, "H": 0.0087, "I": 0.0048, "K": 0.0175, "L": 0.0016, "M": 0.0006, "N": 0.8447, "P": 0.0004, "Q": 0.0011, "R": 0.0053, "S": 0.0536, "T": 0.0156, "V": 0.0019, "W": 0.0001, "Y": 0.0057 },
    "110": { "C": 0.0038, "G": 0.0334, "K": 0.0011, "L": 0.0033, "M": 0.0005, "N": 0.0005, "R": 0.0082, "S": 0.0142, "V": 0.0011, "W": 0.9338 },
    "112": { "A": 0.0046, "G": 0.023, "H": 0.0691, "K": 0.0046, "L": 0.0138, "M": 0.0046, "N": 0.0046, "P": 0.8065, "Q": 0.0046, "R": 0.023, "S": 0.0092, "T": 0.0184, "V": 0.0138 },
    "113": { "A": 0.0015, "C": 0.0054, "D": 0.0006, "E": 0.0001, "F": 0.0006, "G": 0.0085, "H": 0.0044, "I": 0.0006, "K": 0.0004, "L": 0.0062, "M": 0.0002, "N": 0.0004, "P": 0.2051, "Q": 0.0005, "R": 0.0084, "S": 0.016, "T": 0.0026, "W": 0.7382, "Y": 0.0002 },
    "114": { "A": 0.0057, "C": 0.0048, "D": 0.0039, "E": 0.002, "F": 0.0031, "G": 0.0089, "H": 0.0052, "I": 0.0009, "K": 0.0011, "L": 0.0225, "M": 0.0001, "N": 0.0646, "P": 0.4888, "Q": 0.0015, "R": 0.0134, "S": 0.0378, "T": 0.0091, "V": 0.0015, "W": 0.3226, "Y": 0.0026 },
    "115": { "A": 0.0208, "C": 0.0005, "D": 0.0013, "E": 0.0116, "F": 0.0042, "G": 0.0279, "H": 0.003, "I": 0.0037, "K": 0.0066, "L": 0.0638, "M": 0.0216, "N": 0.0025, "P": 0.6419, "Q": 0.0119, "R": 0.0413, "S": 0.0428, "T": 0.0177, "V": 0.0066, "W": 0.0696, "Y": 0.0008 },
    "116": { "A": 0.0065, "C": 0.0115, "D": 0.0099, "E": 0.0091, "F": 0.0491, "G": 0.049, "H": 0.0087, "I": 0.0734, "K": 0.0095, "L": 0.1975, "M": 0.0018, "N": 0.0085, "P": 0.0623, "Q": 0.0222, "R": 0.1093, "S": 0.0125, "T": 0.0047, "V": 0.0261, "W": 0.1559, "Y": 0.1726 },
    "117": { "A": 0.0197, "C": 0.0002, "D": 0.0007, "E": 0.0009, "F": 0.0016, "G": 0.0054, "H": 0.0013, "I": 0.002, "K": 0.0013, "L": 0.0092, "M": 0.0012, "N": 0.0019, "P": 0.0053, "Q": 0.0001, "R": 0.0035, "S": 0.0374, "T": 0.9046, "V": 0.003, "W": 0.0001, "Y": 0.0005 },
    "118": { "A": 0.0001, "C": 0.0069, "D": 0.0001, "F": 0.9781, "G": 0.0005, "H": 0.0001, "I": 0.0005, "L": 0.0048, "P": 0.0003, "R": 0.0003, "S": 0.0019, "T": 0.0001, "V": 0.0053, "W": 0.0003, "Y": 0.0008 },
    "119": { "A": 0.0021, "C": 0.0007, "D": 0.0013, "E": 0.0001, "G": 0.9925, "I": 0.0001, "L": 0.0001, "P": 0.0001, "R": 0.0007, "S": 0.0016, "T": 0.0001, "V": 0.0007, "W": 0.0001 },
    "120": { "A": 0.0005, "D": 0.0001, "E": 0.0032, "F": 0.0001, "G": 0.2198, "H": 0.0073, "K": 0.0015, "L": 0.0015, "M": 0.0001, "N": 0.0001, "P": 0.0777, "Q": 0.6815, "R": 0.0064, "S": 0.0002, "T": 0.0001, "V": 0.0001, "Y": 0.0001 },
    "121": { "A": 0.0004, "E": 0.0028, "F": 0.0001, "G": 0.99, "L": 0.0001, "R": 0.0033, "S": 0.0001, "V": 0.001, "W": 0.0023 },
    "122": { "A": 0.0023, "D": 0.0001, "G": 0.0001, "I": 0.0012, "K": 0.0003, "N": 0.001, "P": 0.0013, "R": 0.0002, "S": 0.0025, "T": 0.991 },
    "123": { "D": 0.0001, "E": 0.0024, "H": 0.0002, "I": 0.0003, "K": 0.8821, "L": 0.0005, "M": 0.0018, "N": 0.0033, "P": 0.0001, "Q": 0.0049, "R": 0.0995, "S": 0.0003, "T": 0.0047, "V": 0.0001 },
    "124": { "A": 0.0007, "E": 0.0005, "G": 0.0007, "I": 0.0008, "L": 0.3403, "M": 0.0025, "P": 0.0003, "Q": 0.0002, "R": 0.0006, "S": 0.0001, "V": 0.6534 },
    "125": { "A": 0.0014, "D": 0.0853, "E": 0.9039, "G": 0.0031, "H": 0.0003, "K": 0.0015, "L": 0.0001, "N": 0.0005, "Q": 0.0031, "R": 0.0001, "V": 0.0007, "Y": 0.0001 },
    "126": { "A": 0.0001, "F": 0.0025, "G": 0.0001, "I": 0.969, "L": 0.0043, "M": 0.0073, "N": 0.0007, "P": 0.0001, "S": 0.0041, "T": 0.0019, "V": 0.01, "Y": 0.0001 },
    "127": { "E": 0.0029, "K": 0.9945, "N": 0.0004, "Q": 0.0023 }
  },
  "IGKV4": {
    "1": { "D": 0.9912, "E": 0.0088 },
    "2": { "C": 0.0001, "F": 0.0061, "G": 0.0001, "H": 0.0002, "I": 0.9469, "L": 0.0197, "M": 0.0019, "N": 0.0015, "P": 0.0001, "R": 0.0004, "S": 0.0075, "T": 0.0063, "V": 0.0091 },
    "3": { "A": 0.0055, "C": 0.0001, "E": 0.0026, "F": 0.0007, "G": 0.0038, "I": 0.001, "K": 0.0001, "L": 0.0138, "M": 0.0039, "Q": 0.0001, "R": 0.0001, "T": 0.0001, "V": 0.9681, "W": 0.0002 },
    "4": { "A": 0.0001, "D": 0.0001, "F": 0.0006, "H": 0.0001, "I": 0.0305, "K": 0.0007, "L": 0.042, "M": 0.9072, "N": 0.0001, "P": 0.0001, "R": 0.0027, "S": 0.0002, "T": 0.0031, "V": 0.0122, "W": 0.0002 },
    "5": { "A": 0.0033, "C": 0.0001, "D": 0.0002, "F": 0.0001, "H": 0.0005, "I": 0.0048, "K": 0.0001, "L": 0.0003, "M": 0.0001, "N": 0.0093, "P": 0.0144, "R": 0.0001, "S": 0.0063, "T": 0.9602, "V": 0.0002, "Y": 0.0001 },
    "6": { "E": 0.0023, "H": 0.0159, "K": 0.0077, "L": 0.0019, "M": 0.0001, "N": 0.0002, "P": 0.0193, "Q": 0.9491, "R": 0.002, "S": 0.0001, "T": 0.001, "V": 0.0001, "Y": 0.0002 },
    "7": { "A": 0.009, "C": 0.001, "D": 0.0003, "F": 0.0034, "H": 0.0003, "L": 0.0006, "N": 0.0001, "P": 0.0146, "Q": 0.0001, "S": 0.9578, "T": 0.0039, "V": 0.0003, "Y": 0.0086 },
    "8": { "A": 0.0026, "H": 0.0012, "I": 0.0001, "L": 0.005, "N": 0.0001, "P": 0.9685, "Q": 0.0084, "R": 0.0021, "S": 0.0042, "T": 0.0078, "V": 0.0001 },
    "9": { "A": 0.0345, "C": 0.0001, "D": 0.9225, "E": 0.0134, "F": 0.0002, "G": 0.0029, "H": 0.0079, "I": 0.0001, "K": 0.0001, "L": 0.0003, "N": 0.0019, "P": 0.0005, "Q": 0.0003, "R": 0.0002, "S": 0.0012, "T": 0.0006, "V": 0.0059, "Y": 0.0073 },
    "10": { "A": 0.0072, "C": 0.0018, "D": 0.0001, "F": 0.0062, "G": 0.0001, "H": 0.0006, "K": 0.0001, "L": 0.0012, "P": 0.0257, "R": 0.0001, "S": 0.9425, "T": 0.0038, "V": 0.0004, "Y": 0.01 },
    "11": { "A": 0.0001, "F": 0.0003, "H": 0.0001, "I": 0.0004, "L": 0.9758, "M": 0.0073, "P": 0.0068, "Q": 0.0017, "R": 0.0049, "S": 0.0003, "T": 0.0004, "V": 0.0017, "W": 0.0002 },
    "12": { "A": 0.9313, "C": 0.0001, "D": 0.006, "E": 0.0001, "F": 0.0001, "G": 0.0024, "H": 0.0005, "I": 0.0003, "L": 0.0006, "P": 0.018, "R": 0.0002, "S": 0.0199, "T": 0.0135, "V": 0.007 },
    "13": { "A": 0.0145, "C": 0.0001, "D": 0.0004, "E": 0.0033, "F": 0.001, "G": 0.013, "I": 0.0001, "K": 0.0001, "L": 0.0221, "M": 0.0021, "P": 0.0003, "Q": 0.0001, "R": 0.0003, "S": 0.0005, "T": 0.0001, "V": 0.9411, "W": 0.0008 },
    "14": { "A": 0.0086, "C": 0.0018, "F": 0.0034, "H": 0.0002, "L": 0.0004, "P": 0.0083, "S": 0.9698, "T": 0.0022, "V": 0.0001, "Y": 0.0052 },
    "15": { "A": 0.0002, "E": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0003, "L": 0.9746, "M": 0.005, "P": 0.0075, "Q": 0.0019, "R": 0.0067, "S": 0.0005, "T": 0.0001, "V": 0.0028, "W": 0.0001 },
    "16": { "A": 0.0044, "C": 0.0084, "D": 0.0019, "F": 0.0001, "G": 0.9696, "H": 0.0001, "L": 0.0001, "R": 0.0056, "S": 0.0017, "T": 0.0001, "V": 0.0079, "W": 0.0001, "Y": 0.0001 },
    "17": { "A": 0.0246, "D": 0.0092, "E": 0.9464, "G": 0.0097, "K": 0.0014, "L": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0023, "R": 0.0003, "S": 0.0001, "T": 0.0001, "V": 0.005, "W": 0.0001, "Y": 0.0003 },
    "18": { "C": 0.0001, "E": 0.0002, "G": 0.0066, "H": 0.0001, "I": 0.0002, "K": 0.0039, "L": 0.0004, "M": 0.0054, "N": 0.0001, "P": 0.0003, "Q": 0.0005, "R": 0.9609, "S": 0.0108, "T": 0.0072, "V": 0.0001, "W": 0.0031 },
    "19": { "A": 0.9722, "D": 0.0067, "E": 0.0001, "G": 0.0016, "I": 0.0001, "L": 0.0001, "P": 0.0028, "R": 0.0001, "S": 0.0052, "T": 0.0049, "V": 0.0061 },
    "20": { "A": 0.0039, "D": 0.0001, "E": 0.0001, "H": 0.0013, "I": 0.0023, "M": 0.0001, "N": 0.0069, "P": 0.0143, "R": 0.0001, "S": 0.0046, "T": 0.9662, "Y": 0.0001 },
    "21": { "A": 0.0001, "F": 0.0035, "G": 0.0001, "H": 0.0004, "I": 0.9538, "L": 0.0148, "M": 0.0024, "N": 0.0037, "P": 0.0003, "R": 0.0003, "S": 0.0066, "T": 0.0089, "V": 0.005, "Y": 0.0001 },
    "22": { "A": 0.0004, "D": 0.0039, "G": 0.0001, "H": 0.0305, "I": 0.0025, "K": 0.0103, "L": 0.0001, "M": 0.0001, "N": 0.9135, "P": 0.0003, "Q": 0.0007, "R": 0.0012, "S": 0.0101, "T": 0.023, "Y": 0.0031 },
    "23": { "A": 0.0001, "C": 0.97, "D": 0.0001, "F": 0.0048, "G": 0.006, "H": 0.0005, "I": 0.0001, "L": 0.0002, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0076, "S": 0.007, "V": 0.0004, "W": 0.0005, "Y": 0.0023 },
    "24": { "E": 0.0026, "G": 0.0001, "H": 0.0006, "I": 0.0001, "K": 0.9368, "L": 0.0001, "M": 0.0014, "N": 0.0097, "P": 0.0005, "Q": 0.0117, "R": 0.0166, "S": 0.0006, "T": 0.0188, "V": 0.0001, "W": 0.0003 },
    "25": { "A": 0.012, "C": 0.0022, "D": 0.0001, "F": 0.0017, "H": 0.0004, "I": 0.0001, "L": 0.0001, "P": 0.0104, "S": 0.9625, "T": 0.0032, "V": 0.0003, "W": 0.0001, "Y": 0.007 },
    "26": { "A": 0.0002, "C": 0.0012, "D": 0.0006, "F": 0.0001, "G": 0.003, "H": 0.0003, "I": 0.0045, "L": 0.0002, "N": 0.0037, "R": 0.0169, "S": 0.9644, "T": 0.0049, "W": 0.0001 },
    "27": { "A": 0.0001, "D": 0.0004, "E": 0.0028, "G": 0.0001, "H": 0.0113, "K": 0.0086, "L": 0.003, "M": 0.0001, "N": 0.0003, "P": 0.0148, "Q": 0.9448, "R": 0.0119, "S": 0.0006, "T": 0.0008, "V": 0.0001, "W": 0.0002 },
    "28": { "A": 0.0003, "C": 0.0024, "D": 0.001, "G": 0.0056, "H": 0.0003, "I": 0.0052, "K": 0.0006, "L": 0.0003, "M": 0.0001, "N": 0.0206, "P": 0.0011, "R": 0.0246, "S": 0.9161, "T": 0.0212, "V": 0.0004, "Y": 0.0001 },
    "29": { "A": 0.0057, "C": 0.0006, "D": 0.002, "F": 0.0056, "G": 0.0079, "I": 0.0322, "K": 0.0001, "L": 0.037, "M": 0.0001, "N": 0.0002, "P": 0.0005, "R": 0.0005, "S": 0.0007, "T": 0.0002, "V": 0.9067, "Y": 0.0001 },
    "30": { "A": 0.0003, "C": 0.0006, "F": 0.038, "I": 0.0028, "L": 0.9403, "M": 0.0003, "P": 0.0001, "Q": 0.0002, "R": 0.0006, "S": 0.0075, "T": 0.0001, "V": 0.0077, "W": 0.0003, "Y": 0.0012 },
    "31": { "A": 0.0012, "C": 0.0024, "D": 0.0101, "E": 0.0006, "F": 0.0137, "G": 0.0001, "H": 0.0171, "I": 0.0001, "K": 0.0028, "L": 0.0017, "M": 0.0001, "N": 0.0077, "P": 0.0004, "Q": 0.0008, "R": 0.0008, "S": 0.0305, "T": 0.0004, "V": 0.0006, "W": 0.0018, "Y": 0.9071 },
    "32": { "A": 0.0016, "C": 0.0015, "D": 0.003, "E": 0.0001, "F": 0.0003, "G": 0.0059, "H": 0.0018, "I": 0.0058, "K": 0.0012, "L": 0.0006, "M": 0.0002, "N": 0.0258, "P": 0.001, "Q": 0.0001, "R": 0.0385, "S": 0.8922, "T": 0.019, "V": 0.0005, "Y": 0.0011 },
    "33": { "A": 0.0151, "C": 0.0012, "D": 0.0007, "E": 0.0001, "F": 0.0087, "G": 0.001, "H": 0.0012, "I": 0.0006, "L": 0.0019, "N": 0.0003, "P": 0.0271, "Q": 0.0001, "R": 0.0006, "S": 0.9167, "T": 0.0096, "V": 0.0017, "Y": 0.0134 },
    "34": { "A": 0.0007, "C": 0.0001, "D": 0.0171, "E": 0.0008, "F": 0.0001, "G": 0.0007, "H": 0.0071, "I": 0.0019, "K": 0.0139, "L": 0.0001, "M": 0.0002, "N": 0.9279, "P": 0.0001, "Q": 0.0007, "R": 0.0012, "S": 0.0099, "T": 0.0132, "V": 0.0001, "Y": 0.0041 },
    "35": { "A": 0.0001, "C": 0.0001, "D": 0.0072, "E": 0.0005, "F": 0.0003, "G": 0.0016, "H": 0.009, "I": 0.0026, "K": 0.0196, "L": 0.0001, "M": 0.0003, "N": 0.929, "P": 0.0001, "Q": 0.0004, "R": 0.0018, "S": 0.0167, "T": 0.0073, "V": 0.0001, "W": 0.0001, "Y": 0.0032 },
    "36": { "A": 0.0008, "D": 0.0003, "E": 0.0061, "F": 0.0001, "G": 0.0006, "H": 0.0004, "I": 0.0006, "K": 0.916, "L": 0.0017, "M": 0.0037, "N": 0.0196, "P": 0.0001, "Q": 0.0167, "R": 0.0223, "S": 0.0003, "T": 0.0093, "V": 0.0001, "W": 0.0004, "Y": 0.001 },
    "37": { "A": 0.0003, "D": 0.0106, "E": 0.0004, "F": 0.0001, "G": 0.0002, "H": 0.0143, "I": 0.0021, "K": 0.0068, "L": 0.0001, "M": 0.0001, "N": 0.9373, "Q": 0.0002, "R": 0.0003, "S": 0.0092, "T": 0.0124, "V": 0.0001, "Y": 0.0055 },
    "38": { "A": 0.0013, "C": 0.0072, "D": 0.0077, "E": 0.0001, "F": 0.0137, "H": 0.0095, "K": 0.0003, "L": 0.0008, "N": 0.0041, "P": 0.0005, "Q": 0.0008, "R": 0.0006, "S": 0.0131, "T": 0.0003, "V": 0.0001, "W": 0.0002, "Y": 0.9396 },
    "39": { "A": 0.0001, "C": 0.0001, "F": 0.011, "G": 0.0001, "I": 0.0045, "K": 0.0001, "L": 0.9687, "M": 0.0005, "R": 0.0001, "S": 0.0019, "V": 0.0128, "Y": 0.0001 },
    "40": { "A": 0.9296, "C": 0.0001, "D": 0.0052, "E": 0.0002, "F": 0.0001, "G": 0.0196, "H": 0.0002, "I": 0.0003, "L": 0.0003, "M": 0.0001, "N": 0.0017, "P": 0.0048, "R": 0.0001, "S": 0.0154, "T": 0.0159, "V": 0.0064, "Y": 0.0001 },
    "41": { "A": 0.0001, "C": 0.0069, "F": 0.0001, "G": 0.0068, "L": 0.0052, "M": 0.0001, "P": 0.0001, "R": 0.0034, "S": 0.0006, "V": 0.0002, "W": 0.9765, "Y": 0.0001 },
    "42": { "C": 0.0006, "D": 0.0053, "E": 0.0002, "F": 0.0138, "G": 0.0001, "H": 0.0031, "I": 0.0001, "L": 0.0006, "N": 0.0017, "P": 0.0001, "Q": 0.0001, "S": 0.0041, "V": 0.0002, "W": 0.0001, "Y": 0.9699 },
    "43": { "E": 0.0004, "G": 0.0001, "H": 0.0068, "K": 0.005, "L": 0.0019, "P": 0.005, "Q": 0.9757, "R": 0.0047, "T": 0.0003, "V": 0.0001, "Y": 0.0001 },
    "44": { "A": 0.0001, "D": 0.0001, "E": 0.0006, "F": 0.0001, "G": 0.0001, "H": 0.0148, "K": 0.0083, "L": 0.002, "M": 0.0001, "N": 0.0001, "P": 0.0039, "Q": 0.9651, "R": 0.0035, "T": 0.0001, "V": 0.0008, "W": 0.0001, "Y": 0.0002 },
    "45": { "A": 0.0001, "D": 0.0001, "E": 0.0014, "G": 0.0002, "H": 0.0001, "I": 0.0009, "K": 0.9607, "L": 0.0001, "M": 0.0002, "N": 0.0071, "Q": 0.0069, "R": 0.0165, "S": 0.0003, "T": 0.0052, "Y": 0.0001 },
    "46": { "A": 0.0109, "E": 0.0001, "F": 0.0001, "G": 0.0003, "H": 0.0003, "I": 0.0001, "L": 0.0041, "M": 0.0001, "P": 0.9636, "Q": 0.0052, "R": 0.0014, "S": 0.0077, "T": 0.0041, "V": 0.0021 },
    "47": { "A": 0.0017, "C": 0.0001, "D": 0.0001, "E": 0.0023, "G": 0.98, "H": 0.0001, "L": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0092, "S": 0.0003, "V": 0.0056, "W": 0.0002 },
    "48": { "A": 0.0001, "E": 0.0006, "H": 0.0102, "K": 0.0048, "L": 0.0019, "N": 0.0002, "P": 0.0057, "Q": 0.9715, "R": 0.0048, "S": 0.0001, "T": 0.0001 },
    "49": { "A": 0.004, "G": 0.0001, "H": 0.0048, "I": 0.0001, "L": 0.0012, "P": 0.969, "Q": 0.0004, "R": 0.0014, "S": 0.0126, "T": 0.0061, "V": 0.0001, "Y": 0.0001 },
    "50": { "A": 0.0008, "H": 0.004, "I": 0.0001, "L": 0.0021, "N": 0.0001, "P": 0.9866, "Q": 0.0001, "R": 0.001, "S": 0.001, "T": 0.0042 },
    "51": { "D": 0.0001, "E": 0.0035, "G": 0.0001, "H": 0.0008, "I": 0.0002, "K": 0.9431, "L": 0.0001, "M": 0.0021, "N": 0.0131, "P": 0.0001, "Q": 0.0064, "R": 0.0216, "S": 0.0008, "T": 0.0077, "V": 0.0001, "Y": 0.0001 },
    "52": { "A": 0.0005, "C": 0.0001, "E": 0.0001, "F": 0.0006, "G": 0.0001, "H": 0.0001, "I": 0.0007, "K": 0.0001, "L": 0.9658, "M": 0.0052, "P": 0.0021, "Q": 0.0012, "R": 0.0053, "S": 0.0003, "V": 0.0173, "W": 0.0006 },
    "53": { "C": 0.0003, "F": 0.0029, "H": 0.0022, "I": 0.0075, "L": 0.9696, "N": 0.0001, "P": 0.0059, "R": 0.0077, "S": 0.0005, "T": 0.0001, "V": 0.0032, "Y": 0.0001 },
    "54": { "A": 0.0005, "C": 0.0001, "F": 0.0117, "G": 0.0001, "I": 0.9513, "L": 0.0117, "M": 0.0107, "N": 0.0008, "R": 0.0002, "S": 0.005, "T": 0.0016, "V": 0.0063 },
    "55": { "A": 0.0008, "C": 0.0043, "D": 0.0073, "F": 0.0049, "G": 0.0001, "H": 0.0082, "I": 0.0001, "K": 0.0007, "L": 0.0003, "N": 0.0028, "P": 0.0001, "Q": 0.0002, "R": 0.0006, "S": 0.0313, "T": 0.0015, "Y": 0.9368 },
    "56": { "A": 0.0003, "C": 0.0118, "D": 0.0001, "E": 0.0002, "G": 0.0123, "H": 0.0001, "L": 0.007, "P": 0.0001, "Q": 0.0001, "R": 0.0043, "S": 0.0014, "V": 0.0003, "W": 0.9619, "Y": 0.0001 },
    "57": { "A": 0.9727, "D": 0.0004, "E": 0.006, "G": 0.0032, "K": 0.0001, "P": 0.0014, "Q": 0.0001, "R": 0.0001, "S": 0.0084, "T": 0.0057, "V": 0.0019 },
    "65": { "A": 0.0126, "C": 0.0009, "D": 0.0003, "F": 0.0036, "G": 0.0001, "H": 0.0001, "L": 0.0003, "P": 0.0029, "R": 0.0001, "S": 0.9714, "T": 0.0041, "V": 0.0001, "Y": 0.0035 },
    "66": { "A": 0.0079, "C": 0.0002, "D": 0.0002, "E": 0.0001, "F": 0.0003, "G": 0.0003, "H": 0.0003, "I": 0.0086, "K": 0.0005, "L": 0.0006, "M": 0.0008, "N": 0.0086, "P": 0.0033, "R": 0.0006, "S": 0.0122, "T": 0.9528, "V": 0.0024, "Y": 0.0003 },
    "67": { "C": 0.0001, "G": 0.0009, "L": 0.0044, "P": 0.0024, "Q": 0.002, "R": 0.9887, "S": 0.0001, "T": 0.0001, "W": 0.0014 },
    "68": { "A": 0.016, "D": 0.0142, "E": 0.9395, "G": 0.0075, "H": 0.0014, "I": 0.0003, "K": 0.0057, "L": 0.0001, "M": 0.0001, "N": 0.0006, "P": 0.0001, "Q": 0.0104, "R": 0.0007, "T": 0.0002, "V": 0.003, "Y": 0.0002 },
    "69": { "A": 0.028, "C": 0.0032, "D": 0.0003, "E": 0.0001, "F": 0.0079, "G": 0.0006, "H": 0.0001, "I": 0.0001, "K": 0.0001, "L": 0.0004, "N": 0.0001, "P": 0.0097, "R": 0.0001, "S": 0.9348, "T": 0.0066, "V": 0.0005, "Y": 0.0073 },
    "70": { "A": 0.0015, "C": 0.0002, "D": 0.0002, "E": 0.0017, "G": 0.9798, "L": 0.0001, "P": 0.0001, "R": 0.0037, "S": 0.0001, "V": 0.0084, "W": 0.0043 },
    "71": { "A": 0.0024, "C": 0.0002, "D": 0.0003, "F": 0.0061, "G": 0.003, "I": 0.0022, "L": 0.0017, "R": 0.0001, "S": 0.0001, "V": 0.9839 },
    "72": { "A": 0.0029, "D": 0.0001, "G": 0.0001, "H": 0.0035, "L": 0.0013, "P": 0.9838, "R": 0.0004, "S": 0.0044, "T": 0.0035 },
    "74": { "A": 0.0083, "C": 0.0001, "D": 0.9542, "E": 0.0162, "G": 0.0047, "H": 0.0014, "N": 0.0027, "P": 0.0001, "S": 0.0008, "T": 0.0002, "V": 0.0026, "Y": 0.0086 },
    "75": { "C": 0.0001, "G": 0.001, "H": 0.0001, "L": 0.0059, "P": 0.0008, "Q": 0.0016, "R": 0.9899, "S": 0.0005, "W": 0.0001 },
    "76": { "C": 0.0165, "F": 0.9571, "G": 0.0007, "I": 0.0045, "L": 0.0063, "R": 0.0001, "S": 0.0015, "V": 0.0114, "W": 0.0001, "Y": 0.0018 },
    "77": { "A": 0.0001, "C": 0.0008, "D": 0.0006, "F": 0.0001, "G": 0.0031, "H": 0.0003, "I": 0.0064, "K": 0.0001, "L": 0.0002, "M": 0.0001, "N": 0.003, "R": 0.0134, "S": 0.9655, "T": 0.0059, "V": 0.0003, "Y": 0.0001 },
    "78": { "A": 0.0048, "C": 0.005, "D": 0.0017, "E": 0.0001, "F": 0.0001, "G": 0.9804, "H": 0.0001, "R": 0.001, "S": 0.0032, "T": 0.0001, "V": 0.0035, "Y": 0.0001 },
    "79": { "A": 0.0006, "C": 0.001, "D": 0.0001, "F": 0.0001, "G": 0.0163, "H": 0.0001, "I": 0.0022, "M": 0.0001, "N": 0.0011, "R": 0.0116, "S": 0.9651, "T": 0.0017, "V": 0.0001 },
    "80": { "A": 0.0023, "C": 0.0003, "D": 0.0002, "E": 0.0032, "F": 0.0001, "G": 0.9798, "L": 0.0001, "P": 0.0001, "R": 0.0029, "V": 0.0059, "W": 0.0051 },
    "83": { "A": 0.0064, "C": 0.0021, "D": 0.0003, "E": 0.0001, "F": 0.0013, "H": 0.0001, "L": 0.0001, "P": 0.0019, "S": 0.9821, "T": 0.0012, "V": 0.0001, "Y": 0.0043 },
    "84": { "A": 0.007, "C": 0.0001, "D": 0.0003, "E": 0.0075, "G": 0.9685, "K": 0.0001, "L": 0.0002, "M": 0.0001, "P": 0.0001, "Q": 0.0003, "R": 0.0029, "S": 0.0003, "T": 0.0002, "V": 0.0086, "W": 0.0039 },
    "85": { "A": 0.0062, "E": 0.0003, "G": 0.0001, "I": 0.0014, "K": 0.0089, "N": 0.0001, "P": 0.0019, "Q": 0.0001, "R": 0.002, "S": 0.0023, "T": 0.9766 },
    "86": { "A": 0.0032, "C": 0.0001, "D": 0.9664, "E": 0.0117, "G": 0.0014, "H": 0.0035, "K": 0.0001, "N": 0.0057, "Q": 0.0003, "R": 0.0003, "S": 0.0002, "V": 0.0017, "Y": 0.0053 },
    "87": { "A": 0.0001, "C": 0.0072, "F": 0.9759, "G": 0.0001, "H": 0.0001, "I": 0.0006, "L": 0.0038, "S": 0.0019, "V": 0.0084, "Y": 0.0019 },
    "88": { "A": 0.0049, "D": 0.0001, "I": 0.0023, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0034, "P": 0.0064, "R": 0.0001, "S": 0.0101, "T": 0.9723, "V": 0.0001 },
    "89": { "C": 0.0001, "F": 0.0017, "H": 0.0012, "I": 0.0039, "L": 0.9826, "P": 0.0026, "Q": 0.0001, "R": 0.0066, "S": 0.0001, "T": 0.0001, "V": 0.0009 },
    "90": { "A": 0.0048, "F": 0.0002, "H": 0.0001, "I": 0.0021, "K": 0.0002, "N": 0.003, "P": 0.0081, "R": 0.0001, "S": 0.0086, "T": 0.9725, "V": 0.0002 },
    "91": { "A": 0.0001, "C": 0.0001, "F": 0.0023, "I": 0.9713, "K": 0.0001, "L": 0.0046, "M": 0.0008, "N": 0.0014, "P": 0.0001, "R": 0.0001, "S": 0.0138, "T": 0.0016, "V": 0.0039 },
    "92": { "A": 0.0012, "C": 0.001, "D": 0.0032, "F": 0.0002, "G": 0.003, "H": 0.0001, "I": 0.0047, "K": 0.0001, "L": 0.0003, "M": 0.0001, "N": 0.0186, "P": 0.0001, "R": 0.0092, "S": 0.9465, "T": 0.0115, "V": 0.0001, "Y": 0.0001 },
    "93": { "A": 0.0004, "C": 0.0007, "D": 0.001, "G": 0.0095, "H": 0.0004, "I": 0.005, "K": 0.0001, "L": 0.0002, "N": 0.0125, "P": 0.0002, "R": 0.0125, "S": 0.9537, "T": 0.0036, "V": 0.0002 },
    "94": { "F": 0.0001, "G": 0.0001, "L": 0.9776, "M": 0.0053, "P": 0.0014, "Q": 0.0011, "R": 0.01, "S": 0.0001, "V": 0.0043 },
    "95": { "E": 0.0042, "H": 0.0059, "K": 0.0037, "L": 0.0017, "P": 0.0012, "Q": 0.9788, "R": 0.0043, "S": 0.0001, "T": 0.0001 },
    "96": { "A": 0.968, "D": 0.005, "E": 0.0002, "G": 0.0003, "I": 0.0001, "K": 0.0001, "L": 0.0001, "M": 0.0001, "P": 0.0136, "S": 0.0049, "T": 0.005, "V": 0.0023, "Y": 0.0001 },
    "97": { "A": 0.0041, "D": 0.0072, "E": 0.9837, "G": 0.0021, "K": 0.0008, "L": 0.0001, "P": 0.0001, "Q": 0.0006, "S": 0.0004, "T": 0.0001, "V": 0.0009, "Y": 0.0001 },
    "98": { "A": 0.0023, "C": 0.0001, "D": 0.9863, "E": 0.0022, "G": 0.0013, "H": 0.0005, "N": 0.0008, "Q": 0.0001, "R": 0.0001, "S": 0.0003, "V": 0.0012, "Y": 0.0049 },
    "99": { "A": 0.0071, "D": 0.0001, "E": 0.0016, "F": 0.0021, "G": 0.01, "I": 0.0001, "L": 0.0043, "M": 0.0014, "Q": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.9729, "W": 0.0001 },
    "100": { "A": 0.9854, "E": 0.0023, "G": 0.0028, "P": 0.0003, "R": 0.0001, "S": 0.0057, "T": 0.0019, "V": 0.0015 },
    "101": { "A": 0.0023, "C": 0.0001, "D": 0.001, "E": 0.0003, "F": 0.0064, "G": 0.0021, "I": 0.0145, "L": 0.0096, "M": 0.0005, "P": 0.0001, "R": 0.0001, "S": 0.0006, "T": 0.0037, "V": 0.9588 },
    "102": { "C": 0.0011, "D": 0.0039, "E": 0.0003, "F": 0.0013, "H": 0.0009, "N": 0.0004, "S": 0.003, "Y": 0.989 },
    "103": { "C": 0.0039, "D": 0.0041, "F": 0.0186, "H": 0.0037, "L": 0.0001, "N": 0.0001, "S": 0.0046, "V": 0.0001, "W": 0.0003, "Y": 0.9645 },
    "104": { "C": 0.9798, "F": 0.0034, "G": 0.0041, "H": 0.0001, "L": 0.0001, "R": 0.0021, "S": 0.0011, "V": 0.0001, "W": 0.0077, "Y": 0.0015 },
    "105": { "A": 0.0001, "E": 0.0006, "F": 0.0001, "G": 0.0001, "H": 0.0226, "K": 0.0025, "L": 0.0048, "M": 0.0003, "P": 0.0003, "Q": 0.9657, "R": 0.0026, "S": 0.0001, "V": 0.0001, "Y": 0.0002 },
    "106": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "E": 0.0028, "G": 0.0003, "H": 0.008, "K": 0.0092, "L": 0.0004, "N": 0.0001, "P": 0.0018, "Q": 0.9738, "R": 0.0028, "S": 0.0004, "Y": 0.0001 },
    "107": { "A": 0.0006, "C": 0.0071, "D": 0.0305, "E": 0.0047, "F": 0.0088, "G": 0.0012, "H": 0.0096, "I": 0.0001, "K": 0.0003, "L": 0.0003, "N": 0.0024, "P": 0.0001, "Q": 0.0003, "R": 0.0013, "S": 0.0035, "T": 0.0008, "V": 0.0008, "W": 0.0002, "Y": 0.9275 },
    "108": { "A": 0.0014, "C": 0.0047, "D": 0.0352, "E": 0.0013, "F": 0.0192, "G": 0.0008, "H": 0.0095, "I": 0.0008, "L": 0.0053, "N": 0.0042, "P": 0.0001, "Q": 0.0008, "R": 0.0006, "S": 0.009, "T": 0.0002, "V": 0.0028, "W": 0.0006, "Y": 0.9036 },
    "109": { "A": 0.0054, "C": 0.0015, "D": 0.0111, "E": 0.0021, "F": 0.001, "G": 0.0181, "H": 0.0025, "I": 0.0056, "K": 0.002, "L": 0.0011, "M": 0.0002, "N": 0.027, "P": 0.0002, "Q": 0.0002, "R": 0.0197, "S": 0.8541, "T": 0.0432, "V": 0.0017, "W": 0.0004, "Y": 0.0028 },
    "110": { "A": 0.0123, "M": 0.0041, "N": 0.0288, "P": 0.0247, "R": 0.0041, "S": 0.0494, "T": 0.8642, "W": 0.0082, "Y": 0.0041 },
    "113": { "A": 0.0125, "D": 0.0022, "F": 0.0013, "G": 0.0013, "H": 0.0027, "I": 0.0049, "L": 0.0022, "N": 0.0169, "P": 0.111, "Q": 0.0004, "R": 0.0067, "S": 0.0241, "T": 0.8088, "V": 0.0013, "W": 0.0027, "Y": 0.0009 },
    "114": { "A": 0.0153, "C": 0.0002, "D": 0.0022, "E": 0.0009, "F": 0.0051, "G": 0.0042, "H": 0.0026, "I": 0.0235, "K": 0.0008, "L": 0.0264, "M": 0.0008, "N": 0.0126, "P": 0.1367, "Q": 0.0012, "R": 0.0092, "S": 0.0993, "T": 0.6512, "V": 0.0056, "W": 0.0008, "Y": 0.0017 },
    "115": { "A": 0.0159, "C": 0.0002, "D": 0.0003, "E": 0.003, "F": 0.0022, "G": 0.0094, "H": 0.0028, "I": 0.0055, "K": 0.0012, "L": 0.0449, "M": 0.0117, "N": 0.0009, "P": 0.7591, "Q": 0.0112, "R": 0.0202, "S": 0.0346, "T": 0.0686, "V": 0.0042, "W": 0.0032, "Y": 0.001 },
    "116": { "A": 0.003, "C": 0.0169, "D": 0.0051, "E": 0.0032, "F": 0.0636, "G": 0.0173, "H": 0.0177, "I": 0.0465, "K": 0.0055, "L": 0.2103, "M": 0.0012, "N": 0.0059, "P": 0.1196, "Q": 0.0389, "R": 0.1019, "S": 0.0134, "T": 0.0047, "V": 0.0232, "W": 0.1323, "Y": 0.1697 },
    "117": { "A": 0.0228, "C": 0.0005, "D": 0.0009, "E": 0.001, "F": 0.0023, "G": 0.004, "H": 0.0019, "I": 0.0021, "K": 0.0005, "L": 0.007, "M": 0.0007, "N": 0.0019, "P": 0.0044, "Q": 0.0004, "R": 0.003, "S": 0.0396, "T": 0.9044, "V": 0.0019, "W": 0.0002, "Y": 0.0006 },
    "118": { "A": 0.0001, "C": 0.0041, "F": 0.9861, "G": 0.0001, "L": 0.0028, "M": 0.0001, "P": 0.0001, "S": 0.0012, "T": 0.0001, "V": 0.0043, "W": 0.0001, "Y": 0.0008 },
    "119": { "A": 0.0015, "C": 0.0007, "D": 0.0014, "G": 0.9933, "L": 0.0001, "P": 0.0001, "R": 0.0006, "S": 0.0016, "T": 0.0001, "V": 0.0004, "W": 0.0001 },
    "120": { "A": 0.0006, "C": 0.0001, "E": 0.0019, "G": 0.2308, "H": 0.0089, "K": 0.0006, "L": 0.0014, "P": 0.1013, "Q": 0.6481, "R": 0.0056, "S": 0.0003, "T": 0.0003, "V": 0.0002, "Y": 0.0001 },
    "121": { "A": 0.0003, "E": 0.0033, "G": 0.9889, "L": 0.0001, "R": 0.0037, "V": 0.0008, "W": 0.003 },
    "122": { "A": 0.0017, "H": 0.0001, "I": 0.0012, "K": 0.0003, "N": 0.0008, "P": 0.0008, "S": 0.0023, "T": 0.9928 },
    "123": { "A": 0.0002, "D": 0.0001, "E": 0.0014, "G": 0.0001, "H": 0.0002, "I": 0.0001, "K": 0.9094, "L": 0.0003, "M": 0.0006, "N": 0.0025, "P": 0.0001, "Q": 0.0037, "R": 0.0766, "S": 0.0003, "T": 0.0043, "V": 0.0001, "Y": 0.0001 },
    "124": { "A": 0.001, "E": 0.0006, "F": 0.0001, "G": 0.0002, "I": 0.0006, "K": 0.0001, "L": 0.3294, "M": 0.0029, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0001, "V": 0.6646, "W": 0.0001 },
    "125": { "A": 0.0028, "D": 0.1006, "E": 0.8855, "G": 0.0028, "H": 0.0008, "I": 0.0001, "K": 0.0014, "N": 0.0007, "Q": 0.0036, "R": 0.0003, "S": 0.0001, "T": 0.0001, "V": 0.001, "Y": 0.0001 },
    "126": { "A": 0.0001, "F": 0.0019, "I": 0.9759, "L": 0.0057, "M": 0.0016, "N": 0.0007, "S": 0.0028, "T": 0.0018, "V": 0.0095 },
    "127": { "E": 0.0019, "K": 0.9961, "N": 0.0006, "Q": 0.0014 }
  },
  "IGKV5": {
    "1": { "D": 0.0419, "E": 0.9581 },
    "2": { "A": 0.0021, "E": 0.0001, "I": 0.0001, "K": 0.0046, "L": 0.0001, "M": 0.003, "P": 0.0241, "Q": 0.0007, "R": 0.0012, "S": 0.0037, "T": 0.9602, "V": 0.0001 },
    "3": { "A": 0.0024, "D": 0.0001, "G": 0.0002, "H": 0.0002, "I": 0.0034, "K": 0.0056, "L": 0.0004, "N": 0.0009, "P": 0.0224, "Q": 0.0006, "R": 0.0015, "S": 0.0036, "T": 0.9588 },
    "4": { "A": 0.0001, "F": 0.0029, "H": 0.0034, "I": 0.0051, "L": 0.9452, "P": 0.0255, "Q": 0.0001, "R": 0.0087, "S": 0.0004, "T": 0.0008, "V": 0.0079 },
    "5": { "A": 0.0015, "K": 0.006, "L": 0.0002, "M": 0.0032, "N": 0.0001, "P": 0.0246, "Q": 0.0009, "R": 0.0026, "S": 0.003, "T": 0.958 },
    "6": { "E": 0.0006, "H": 0.0131, "K": 0.0051, "L": 0.0013, "N": 0.0003, "P": 0.021, "Q": 0.9556, "R": 0.0027, "S": 0.0002, "T": 0.0002 },
    "7": { "A": 0.0084, "C": 0.0011, "D": 0.0002, "F": 0.003, "H": 0.0006, "L": 0.0008, "P": 0.0134, "S": 0.9625, "T": 0.003, "V": 0.0002, "Y": 0.007 },
    "8": { "A": 0.001, "E": 0.0001, "H": 0.0002, "I": 0.0001, "L": 0.003, "P": 0.9786, "Q": 0.0078, "R": 0.0017, "S": 0.0016, "T": 0.006 },
    "9": { "A": 0.9823, "D": 0.0002, "E": 0.0037, "G": 0.0002, "P": 0.0039, "Q": 0.0001, "S": 0.0035, "T": 0.0039, "V": 0.0021 },
    "10": { "C": 0.0025, "F": 0.9785, "I": 0.0009, "L": 0.0099, "M": 0.0001, "S": 0.0041, "T": 0.0001, "V": 0.0032, "Y": 0.0008 },
    "11": { "E": 0.0001, "F": 0.0001, "H": 0.0001, "I": 0.0151, "K": 0.001, "L": 0.0121, "M": 0.9619, "N": 0.0001, "Q": 0.0002, "R": 0.0024, "S": 0.0002, "T": 0.003, "V": 0.0038 },
    "12": { "A": 0.0029, "E": 0.0001, "F": 0.0004, "K": 0.0001, "L": 0.0018, "P": 0.0043, "Q": 0.0001, "S": 0.9889, "T": 0.0009, "V": 0.0001, "Y": 0.0005 },
    "13": { "A": 0.9842, "E": 0.0034, "G": 0.0007, "H": 0.0001, "P": 0.0047, "Q": 0.0001, "S": 0.0036, "T": 0.0017, "V": 0.0015 },
    "14": { "A": 0.0044, "G": 0.0001, "H": 0.0002, "I": 0.0012, "N": 0.0035, "P": 0.0078, "R": 0.0001, "S": 0.0018, "T": 0.9809, "Y": 0.0001 },
    "15": { "A": 0.0015, "H": 0.0005, "L": 0.0018, "P": 0.9763, "Q": 0.009, "R": 0.0005, "S": 0.0022, "T": 0.0082 },
    "16": { "A": 0.0029, "C": 0.0001, "D": 0.0003, "E": 0.0016, "G": 0.9805, "I": 0.0001, "K": 0.0002, "L": 0.0001, "R": 0.0111, "S": 0.0001, "V": 0.0029, "Y": 0.0001 },
    "17": { "A": 0.0086, "D": 0.9706, "E": 0.0053, "G": 0.0018, "H": 0.0028, "I": 0.0001, "K": 0.0001, "N": 0.0047, "Q": 0.0001, "S": 0.0004, "T": 0.0002, "V": 0.0014, "Y": 0.0038 },
    "18": { "D": 0.0001, "E": 0.0018, "H": 0.0007, "I": 0.0012, "K": 0.9581, "N": 0.0171, "Q": 0.0067, "R": 0.0037, "S": 0.0002, "T": 0.0104 },
    "19": { "A": 0.0093, "D": 0.0044, "F": 0.0036, "G": 0.006, "H": 0.0002, "I": 0.0025, "L": 0.0028, "N": 0.0001, "P": 0.0001, "R": 0.0002, "S": 0.0005, "T": 0.0002, "V": 0.97, "Y": 0.0002 },
    "20": { "D": 0.0015, "F": 0.0004, "H": 0.005, "I": 0.0032, "K": 0.0058, "L": 0.0001, "N": 0.9597, "P": 0.0004, "Q": 0.0001, "S": 0.0055, "T": 0.0175, "Y": 0.001 },
    "21": { "F": 0.0017, "H": 0.0005, "I": 0.971, "K": 0.0001, "L": 0.0108, "M": 0.0013, "N": 0.0032, "R": 0.0005, "S": 0.0053, "T": 0.0034, "V": 0.0023 },
    "22": { "A": 0.0077, "C": 0.0004, "D": 0.0001, "F": 0.0015, "H": 0.0009, "L": 0.0001, "P": 0.023, "S": 0.9496, "T": 0.0113, "V": 0.0001, "Y": 0.0053 },
    "23": { "C": 0.9801, "F": 0.0095, "G": 0.002, "I": 0.0002, "L": 0.0003, "R": 0.0013, "S": 0.0029, "V": 0.0002, "W": 0.0004, "Y": 0.0031 },
    "24": { "D": 0.0001, "E": 0.001, "H": 0.0002, "I": 0.0013, "K": 0.9658, "L": 0.0001, "N": 0.0102, "P": 0.0001, "Q": 0.0045, "R": 0.0108, "S": 0.0005, "T": 0.0054, "Y": 0.0001 },
    "25": { "A": 0.9772, "D": 0.0071, "G": 0.0019, "H": 0.0001, "I": 0.0001, "N": 0.0001, "P": 0.0023, "S": 0.0037, "T": 0.0028, "V": 0.0047 },
    "26": { "C": 0.0009, "G": 0.0012, "H": 0.0001, "I": 0.0033, "K": 0.0001, "L": 0.0001, "N": 0.0031, "R": 0.0146, "S": 0.9738, "T": 0.0029, "Y": 0.0001 },
    "27": { "A": 0.0001, "D": 0.0001, "E": 0.0019, "H": 0.0079, "K": 0.0076, "L": 0.0006, "N": 0.0002, "P": 0.0033, "Q": 0.9692, "R": 0.0088, "T": 0.0003, "V": 0.0001 },
    "28": { "A": 0.0072, "D": 0.9666, "E": 0.0071, "G": 0.0039, "H": 0.0014, "K": 0.0001, "N": 0.0051, "Q": 0.0002, "S": 0.0008, "T": 0.0002, "V": 0.0027, "Y": 0.0047 },
    "29": { "D": 0.0001, "F": 0.0011, "H": 0.0001, "I": 0.9682, "L": 0.0049, "M": 0.0035, "N": 0.0023, "S": 0.0124, "T": 0.005, "V": 0.0025 },
    "36": { "A": 0.0053, "D": 0.9453, "E": 0.039, "G": 0.002, "H": 0.0023, "K": 0.0001, "N": 0.0017, "P": 0.0001, "Q": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0008, "Y": 0.0032 },
    "37": { "A": 0.0047, "C": 0.0001, "D": 0.9723, "E": 0.0084, "G": 0.0022, "H": 0.0015, "N": 0.0043, "Q": 0.0001, "S": 0.0002, "V": 0.0015, "Y": 0.0047 },
    "38": { "A": 0.0037, "C": 0.0002, "D": 0.968, "E": 0.0143, "G": 0.0012, "H": 0.0012, "K": 0.0002, "N": 0.0022, "Q": 0.0001, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0032, "Y": 0.0054 },
    "39": { "D": 0.0001, "E": 0.0002, "G": 0.0003, "I": 0.0153, "K": 0.0019, "L": 0.0178, "M": 0.938, "N": 0.0001, "R": 0.0034, "S": 0.0001, "T": 0.0018, "V": 0.0211 },
    "40": { "C": 0.0002, "D": 0.004, "H": 0.0055, "I": 0.0018, "K": 0.0076, "N": 0.9707, "P": 0.0001, "R": 0.0001, "S": 0.0044, "T": 0.0041, "Y": 0.0014 },
    "41": { "C": 0.0084, "G": 0.0042, "L": 0.0054, "R": 0.003, "S": 0.0009, "V": 0.0002, "W": 0.9778, "Y": 0.0001 },
    "42": { "C": 0.0033, "D": 0.0034, "E": 0.0002, "F": 0.0042, "H": 0.0014, "K": 0.0001, "L": 0.0001, "N": 0.0019, "Q": 0.0002, "S": 0.0036, "W": 0.0003, "Y": 0.9813 },
    "43": { "E": 0.0003, "H": 0.0038, "I": 0.0001, "K": 0.006, "L": 0.0004, "N": 0.0001, "P": 0.0042, "Q": 0.981, "R": 0.0038, "Y": 0.0002 },
    "44": { "E": 0.001, "H": 0.0064, "K": 0.0045, "L": 0.0007, "M": 0.0001, "P": 0.0031, "Q": 0.9814, "R": 0.0025, "S": 0.0001, "T": 0.0002, "Y": 0.0001 },
    "45": { "D": 0.0001, "E": 0.0013, "G": 0.0001, "H": 0.0002, "I": 0.0007, "K": 0.9846, "N": 0.004, "Q": 0.0035, "R": 0.0033, "T": 0.0022 },
    "46": { "A": 0.0043, "K": 0.0002, "L": 0.0036, "P": 0.9796, "Q": 0.004, "R": 0.0008, "S": 0.0024, "T": 0.0048, "V": 0.0003 },
    "47": { "A": 0.0007, "C": 0.0001, "D": 0.0002, "E": 0.0011, "G": 0.9901, "L": 0.0002, "R": 0.0029, "V": 0.0048 },
    "48": { "A": 0.0061, "D": 0.006, "E": 0.9694, "G": 0.0036, "H": 0.0001, "K": 0.0065, "Q": 0.0062, "S": 0.0001, "T": 0.0003, "V": 0.0015, "Y": 0.0001 },
    "49": { "A": 0.958, "D": 0.0047, "E": 0.0002, "F": 0.0001, "G": 0.0113, "I": 0.0001, "N": 0.0002, "P": 0.0067, "S": 0.0052, "T": 0.0033, "V": 0.0103 },
    "50": { "A": 0.9574, "D": 0.004, "E": 0.0001, "G": 0.0023, "I": 0.0008, "N": 0.0001, "P": 0.0175, "S": 0.0055, "T": 0.0049, "V": 0.0074, "Y": 0.0001 },
    "51": { "A": 0.0006, "F": 0.0026, "G": 0.0002, "I": 0.9575, "L": 0.0062, "M": 0.0082, "N": 0.0018, "P": 0.0004, "R": 0.0004, "S": 0.0049, "T": 0.0026, "V": 0.0147 },
    "52": { "C": 0.0042, "F": 0.9699, "G": 0.0001, "H": 0.0004, "I": 0.0034, "L": 0.012, "M": 0.0002, "R": 0.0001, "S": 0.0019, "V": 0.0063, "Y": 0.0015 },
    "53": { "F": 0.0036, "I": 0.9622, "K": 0.0001, "L": 0.0163, "M": 0.0067, "N": 0.0023, "S": 0.0044, "T": 0.0023, "V": 0.0021 },
    "54": { "C": 0.0001, "D": 0.0001, "F": 0.0023, "G": 0.0001, "I": 0.9721, "L": 0.0053, "M": 0.0043, "N": 0.0024, "R": 0.0001, "S": 0.0046, "T": 0.0012, "V": 0.0074 },
    "55": { "E": 0.0012, "G": 0.0001, "H": 0.0084, "I": 0.0001, "K": 0.0063, "L": 0.0056, "N": 0.0001, "P": 0.0053, "Q": 0.9661, "R": 0.0063, "S": 0.0001, "V": 0.0001, "Y": 0.0002 },
    "56": { "A": 0.0077, "D": 0.0153, "E": 0.9649, "G": 0.0048, "K": 0.0032, "L": 0.0001, "Q": 0.0019, "R": 0.0001, "S": 0.0001, "V": 0.0017, "Y": 0.0003 },
    "57": { "A": 0.9607, "C": 0.0002, "D": 0.0035, "E": 0.0001, "F": 0.0001, "G": 0.0037, "H": 0.0001, "K": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0065, "S": 0.0093, "T": 0.0109, "V": 0.0044, "Y": 0.0001 },
    "65": { "A": 0.006, "G": 0.0005, "H": 0.0001, "I": 0.0073, "K": 0.0004, "N": 0.004, "P": 0.0062, "S": 0.0116, "T": 0.9639 },
    "66": { "A": 0.0064, "C": 0.0001, "D": 0.0002, "G": 0.0002, "I": 0.0016, "K": 0.0008, "N": 0.0059, "P": 0.0062, "Q": 0.0001, "R": 0.0003, "S": 0.0108, "T": 0.9669, "V": 0.0006 },
    "67": { "A": 0.0001, "C": 0.0001, "F": 0.003, "G": 0.0002, "H": 0.0034, "I": 0.0062, "L": 0.9542, "N": 0.0001, "P": 0.0051, "Q": 0.0002, "R": 0.0227, "S": 0.0009, "T": 0.0003, "V": 0.0036 },
    "68": { "A": 0.0054, "C": 0.0003, "D": 0.0034, "E": 0.0003, "F": 0.0032, "G": 0.0179, "H": 0.0001, "I": 0.0028, "L": 0.0039, "M": 0.0006, "P": 0.0001, "R": 0.0006, "S": 0.0003, "V": 0.9613 },
    "69": { "A": 0.0015, "C": 0.0001, "D": 0.0001, "H": 0.0053, "L": 0.001, "N": 0.0006, "P": 0.9773, "Q": 0.0003, "R": 0.0026, "S": 0.002, "T": 0.0092 },
    "70": { "A": 0.0013, "C": 0.0004, "D": 0.0002, "E": 0.0027, "G": 0.9806, "L": 0.0001, "R": 0.0074, "S": 0.0001, "V": 0.007, "W": 0.0002 },
    "71": { "A": 0.0004, "C": 0.0001, "F": 0.0048, "G": 0.0003, "I": 0.9553, "K": 0.0001, "L": 0.0108, "M": 0.0008, "N": 0.0033, "S": 0.0093, "T": 0.0038, "V": 0.011 },
    "72": { "A": 0.0035, "E": 0.0001, "G": 0.0001, "H": 0.0002, "K": 0.0002, "L": 0.0012, "N": 0.0001, "P": 0.7029, "Q": 0.0043, "R": 0.001, "S": 0.2833, "T": 0.003, "Y": 0.0002 },
    "74": { "A": 0.0016, "D": 0.0002, "F": 0.0002, "H": 0.0039, "I": 0.0001, "L": 0.0017, "N": 0.0005, "P": 0.9787, "R": 0.0032, "S": 0.0018, "T": 0.0079, "V": 0.0001, "Y": 0.0001 },
    "75": { "C": 0.0001, "G": 0.0012, "L": 0.0046, "P": 0.0008, "Q": 0.0019, "R": 0.9907, "S": 0.0005, "V": 0.0001, "W": 0.0001 },
    "76": { "C": 0.012, "F": 0.9624, "G": 0.0002, "I": 0.0045, "L": 0.0067, "M": 0.0001, "R": 0.0001, "S": 0.0025, "V": 0.0099, "Y": 0.0016 },
    "77": { "A": 0.0003, "C": 0.0007, "D": 0.0002, "G": 0.0018, "I": 0.0029, "L": 0.0001, "N": 0.001, "P": 0.0002, "R": 0.0095, "S": 0.9816, "T": 0.0018 },
    "78": { "A": 0.0018, "C": 0.006, "D": 0.0024, "F": 0.0003, "G": 0.981, "R": 0.001, "S": 0.0039, "V": 0.0036 },
    "79": { "A": 0.0001, "C": 0.0013, "G": 0.0036, "I": 0.0028, "L": 0.0003, "N": 0.0005, "P": 0.0001, "R": 0.01, "S": 0.9807, "T": 0.0006, "V": 0.0001 },
    "80": { "A": 0.0014, "C": 0.0004, "D": 0.0009, "E": 0.0042, "G": 0.9744, "K": 0.0001, "L": 0.0002, "R": 0.004, "S": 0.0002, "V": 0.006, "W": 0.0082 },
    "83": { "A": 0.0001, "C": 0.0025, "D": 0.0108, "E": 0.0004, "F": 0.0104, "H": 0.0011, "L": 0.0002, "N": 0.0017, "R": 0.0002, "S": 0.0053, "V": 0.0001, "Y": 0.9674 },
    "84": { "A": 0.0011, "E": 0.0004, "G": 0.99, "I": 0.0002, "K": 0.0005, "L": 0.0001, "R": 0.0035, "S": 0.0001, "V": 0.0041, "W": 0.0001 },
    "85": { "A": 0.0063, "G": 0.0001, "I": 0.0008, "K": 0.0033, "M": 0.0003, "P": 0.0023, "R": 0.0078, "S": 0.0035, "T": 0.9757 },
    "86": { "A": 0.0012, "D": 0.9776, "E": 0.0105, "G": 0.0006, "H": 0.0035, "K": 0.0001, "N": 0.0013, "Q": 0.0001, "S": 0.0005, "T": 0.0007, "V": 0.0005, "Y": 0.0035 },
    "87": { "C": 0.0038, "F": 0.9762, "G": 0.0004, "I": 0.002, "L": 0.0046, "S": 0.0012, "V": 0.0101, "W": 0.0001, "Y": 0.0015 },
    "88": { "A": 0.0023, "D": 0.0001, "F": 0.0001, "G": 0.0001, "I": 0.0023, "L": 0.0001, "M": 0.0001, "N": 0.0042, "P": 0.0035, "S": 0.0035, "T": 0.9837, "V": 0.0001 },
    "89": { "F": 0.0011, "H": 0.0011, "I": 0.0044, "L": 0.9889, "P": 0.0017, "R": 0.0026, "V": 0.0003 },
    "90": { "A": 0.0011, "I": 0.0027, "K": 0.0016, "L": 0.0001, "P": 0.002, "R": 0.0005, "S": 0.0025, "T": 0.9895 },
    "91": { "D": 0.0001, "F": 0.0012, "I": 0.97, "K": 0.0002, "L": 0.0031, "M": 0.0089, "N": 0.0032, "R": 0.0002, "S": 0.0074, "T": 0.0036, "V": 0.0023 },
    "92": { "A": 0.0001, "D": 0.0042, "E": 0.0003, "H": 0.0069, "I": 0.0014, "K": 0.0105, "M": 0.0001, "N": 0.9606, "Q": 0.0011, "S": 0.0094, "T": 0.0038, "Y": 0.0015 },
    "93": { "A": 0.0001, "D": 0.0076, "F": 0.0001, "G": 0.0004, "H": 0.0033, "I": 0.0046, "K": 0.0041, "L": 0.0002, "M": 0.0001, "N": 0.9613, "P": 0.0001, "R": 0.0001, "S": 0.0125, "T": 0.0045, "Y": 0.0011 },
    "94": { "A": 0.0001, "F": 0.0003, "I": 0.9443, "K": 0.002, "L": 0.0108, "M": 0.0096, "Q": 0.0001, "R": 0.0239, "S": 0.0004, "T": 0.0018, "V": 0.0067 },
    "95": { "A": 0.005, "C": 0.0001, "D": 0.0085, "E": 0.9583, "G": 0.0029, "H": 0.0003, "K": 0.004, "L": 0.0001, "N": 0.0001, "Q": 0.0164, "R": 0.0004, "S": 0.0003, "T": 0.0001, "V": 0.0033, "Y": 0.0002 },
    "96": { "A": 0.0153, "C": 0.0005, "D": 0.0001, "F": 0.0022, "H": 0.0001, "L": 0.0003, "P": 0.0034, "S": 0.9738, "T": 0.0016, "W": 0.0001, "Y": 0.0027 },
    "97": { "A": 0.0027, "D": 0.0141, "E": 0.9771, "G": 0.0018, "K": 0.0009, "L": 0.0002, "Q": 0.0011, "R": 0.0001, "V": 0.002, "Y": 0.0001 },
    "98": { "A": 0.0077, "D": 0.9784, "E": 0.0047, "G": 0.0016, "H": 0.0005, "K": 0.0001, "N": 0.0018, "S": 0.0001, "V": 0.0015, "Y": 0.0036 },
    "99": { "A": 0.9619, "D": 0.011, "F": 0.0003, "G": 0.0028, "P": 0.0005, "S": 0.0046, "T": 0.0023, "V": 0.0166 },
    "100": { "A": 0.9868, "D": 0.0001, "E": 0.0032, "G": 0.0005, "N": 0.0001, "P": 0.0007, "S": 0.0044, "T": 0.0029, "V": 0.0013 },
    "101": { "A": 0.0002, "C": 0.0027, "D": 0.0067, "E": 0.0003, "F": 0.0087, "H": 0.003, "I": 0.0001, "L": 0.0002, "N": 0.0033, "S": 0.019, "T": 0.0001, "V": 0.0006, "Y": 0.9552 },
    "102": { "C": 0.004, "D": 0.0037, "F": 0.0024, "G": 0.0001, "H": 0.0012, "L": 0.0001, "N": 0.0059, "S": 0.0038, "T": 0.0001, "V": 0.0001, "Y": 0.9786 },
    "103": { "A": 0.0001, "C": 0.0108, "D": 0.0001, "F": 0.9647, "G": 0.0002, "H": 0.0001, "I": 0.0009, "L": 0.003, "S": 0.0042, "V": 0.0075, "Y": 0.0085 },
    "104": { "C": 0.9799, "F": 0.0037, "G": 0.0033, "P": 0.0001, "R": 0.0013, "S": 0.002, "V": 0.0004, "W": 0.0063, "Y": 0.0029 },
    "105": { "E": 0.0001, "G": 0.0001, "H": 0.0003, "I": 0.0018, "K": 0.0003, "L": 0.9593, "P": 0.0039, "Q": 0.0044, "R": 0.0291, "V": 0.0003, "Y": 0.0004 },
    "106": { "E": 0.0087, "G": 0.0004, "H": 0.0052, "I": 0.0002, "K": 0.0054, "L": 0.002, "M": 0.0003, "N": 0.0006, "P": 0.0043, "Q": 0.9683, "R": 0.0036, "S": 0.0004, "T": 0.0003, "W": 0.0005 },
    "107": { "C": 0.0001, "D": 0.001, "E": 0.0001, "F": 0.0001, "G": 0.0007, "H": 0.9602, "K": 0.0001, "L": 0.001, "M": 0.0004, "N": 0.0037, "P": 0.0031, "Q": 0.0193, "R": 0.0065, "S": 0.0005, "T": 0.0001, "V": 0.0001, "Y": 0.0032 },
    "108": { "A": 0.0055, "C": 0.0002, "D": 0.957, "E": 0.0134, "F": 0.0001, "G": 0.0072, "H": 0.0009, "I": 0.0003, "K": 0.0001, "N": 0.0036, "P": 0.0004, "Q": 0.0002, "S": 0.0006, "T": 0.0001, "V": 0.0029, "Y": 0.0076 },
    "109": { "A": 0.0002, "D": 0.0071, "E": 0.0011, "H": 0.0041, "I": 0.0014, "K": 0.0081, "L": 0.0001, "N": 0.9588, "P": 0.0003, "Q": 0.0001, "S": 0.0072, "T": 0.0077, "Y": 0.0037 },
    "110": { "C": 0.0181, "F": 0.9065, "G": 0.0015, "I": 0.003, "L": 0.0075, "N": 0.0015, "S": 0.0015, "V": 0.0452, "Y": 0.0151 },
    "112": { "F": 0.0097, "G": 0.0097, "H": 0.0874, "L": 0.0583, "P": 0.8058, "R": 0.0097, "S": 0.0097, "T": 0.0097 },
    "113": { "A": 0.0006, "C": 0.0206, "F": 0.7798, "G": 0.0008, "H": 0.0019, "I": 0.0006, "L": 0.0303, "P": 0.1475, "Q": 0.0022, "R": 0.0011, "S": 0.0061, "T": 0.0011, "V": 0.0058, "Y": 0.0017 },
    "114": { "A": 0.0008, "C": 0.0076, "D": 0.0007, "E": 0.001, "F": 0.5306, "G": 0.0041, "H": 0.0022, "I": 0.0031, "K": 0.0011, "L": 0.0641, "M": 0.0002, "N": 0.0619, "P": 0.2943, "Q": 0.0012, "R": 0.0079, "S": 0.0074, "T": 0.0037, "V": 0.0039, "W": 0.0005, "Y": 0.0034 },
    "115": { "A": 0.0162, "C": 0.0019, "D": 0.0062, "E": 0.0069, "F": 0.0522, "G": 0.0122, "H": 0.0041, "I": 0.0014, "K": 0.0045, "L": 0.1501, "M": 0.0207, "N": 0.0357, "P": 0.5706, "Q": 0.0173, "R": 0.0295, "S": 0.0326, "T": 0.0277, "V": 0.0057, "W": 0.0031, "Y": 0.0012 },
    "116": { "A": 0.0039, "C": 0.0229, "D": 0.0264, "E": 0.0136, "F": 0.068, "G": 0.0441, "H": 0.026, "I": 0.0116, "K": 0.0089, "L": 0.237, "M": 0.0005, "N": 0.0135, "P": 0.0619, "Q": 0.0093, "R": 0.0753, "S": 0.0095, "T": 0.0036, "V": 0.0178, "W": 0.1592, "Y": 0.1872 },
    "117": { "A": 0.0232, "C": 0.0001, "D": 0.0017, "E": 0.0017, "F": 0.006, "G": 0.0103, "H": 0.0011, "I": 0.0014, "K": 0.0017, "L": 0.0192, "M": 0.0015, "N": 0.0096, "P": 0.0578, "Q": 0.0012, "R": 0.0083, "S": 0.042, "T": 0.8114, "V": 0.0014, "W": 0.0001, "Y": 0.0001 },
    "118": { "A": 0.0002, "C": 0.0074, "D": 0.0007, "E": 0.0012, "F": 0.9366, "G": 0.004, "H": 0.0006, "I": 0.0025, "L": 0.0248, "N": 0.0011, "P": 0.0046, "Q": 0.001, "R": 0.0027, "S": 0.0041, "T": 0.0009, "V": 0.0071, "W": 0.0004, "Y": 0.0002 },
    "119": { "A": 0.0012, "C": 0.0017, "D": 0.0036, "E": 0.0002, "G": 0.9739, "H": 0.0001, "L": 0.0014, "N": 0.0001, "P": 0.0052, "Q": 0.0004, "R": 0.0084, "S": 0.0026, "T": 0.0004, "V": 0.0005, "W": 0.0003 },
    "120": { "A": 0.0002, "C": 0.0002, "D": 0.0001, "E": 0.0016, "F": 0.0003, "G": 0.1482, "H": 0.0036, "K": 0.0012, "L": 0.0017, "M": 0.0001, "P": 0.0545, "Q": 0.7815, "R": 0.005, "S": 0.0004, "T": 0.0005, "V": 0.0007, "Y": 0.0002 },
    "121": { "A": 0.0011, "E": 0.0037, "G": 0.9882, "K": 0.0001, "L": 0.0004, "R": 0.0027, "S": 0.0001, "T": 0.0001, "V": 0.0012, "W": 0.0025 },
    "122": { "A": 0.0019, "I": 0.0006, "K": 0.0001, "L": 0.0001, "N": 0.0006, "P": 0.0014, "R": 0.0001, "S": 0.0013, "T": 0.9939 },
    "123": { "A": 0.0001, "E": 0.0017, "H": 0.0001, "I": 0.0001, "K": 0.9641, "L": 0.0006, "M": 0.0002, "N": 0.0013, "Q": 0.0028, "R": 0.0264, "T": 0.0026 },
    "124": { "A": 0.0003, "E": 0.0007, "G": 0.001, "L": 0.4107, "M": 0.004, "P": 0.0005, "Q": 0.0003, "R": 0.0008, "V": 0.5818 },
    "125": { "A": 0.0003, "D": 0.0566, "E": 0.9368, "G": 0.0022, "K": 0.0017, "N": 0.0002, "Q": 0.0012, "R": 0.0003, "V": 0.0006 },
    "126": { "F": 0.0008, "I": 0.9841, "L": 0.0017, "M": 0.0009, "N": 0.0001, "S": 0.0103, "T": 0.0005, "V": 0.0017 },
    "127": { "E": 0.001, "K": 0.9972, "N": 0.0004, "Q": 0.0014 }
  },
  "IGKV6": {
    "1": { "D": 0.0417, "E": 0.9583 },
    "2": { "A": 0.0011, "F": 0.0035, "H": 0.0001, "I": 0.9263, "K": 0.0001, "L": 0.0227, "M": 0.0071, "N": 0.0012, "P": 0.0001, "R": 0.0004, "S": 0.0067, "T": 0.0066, "V": 0.0241 },
    "3": { "A": 0.0044, "C": 0.0001, "D": 0.0001, "E": 0.003, "F": 0.0017, "G": 0.0021, "I": 0.0001, "L": 0.012, "M": 0.0044, "Q": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.9715, "W": 0.0001 },
    "4": { "C": 0.0001, "F": 0.0002, "G": 0.0001, "H": 0.0001, "I": 0.0002, "L": 0.982, "M": 0.008, "P": 0.0025, "Q": 0.0013, "R": 0.0033, "S": 0.0001, "T": 0.0001, "V": 0.0018, "W": 0.0001, "Y": 0.0001 },
    "5": { "A": 0.0017, "D": 0.0001, "F": 0.0001, "H": 0.0004, "I": 0.0046, "K": 0.0001, "L": 0.0001, "M": 0.0004, "N": 0.005, "P": 0.013, "R": 0.0002, "S": 0.0046, "T": 0.9696, "Y": 0.0001 },
    "6": { "A": 0.0001, "E": 0.0022, "F": 0.0001, "G": 0.0001, "H": 0.0128, "K": 0.0073, "L": 0.0028, "M": 0.0001, "N": 0.0001, "P": 0.0205, "Q": 0.9496, "R": 0.003, "S": 0.0002, "T": 0.0009, "W": 0.0001, "Y": 0.0002 },
    "7": { "A": 0.0074, "C": 0.0015, "D": 0.0001, "E": 0.0001, "F": 0.0047, "H": 0.0001, "L": 0.0007, "P": 0.0087, "S": 0.9664, "T": 0.0031, "V": 0.0002, "Y": 0.0072 },
    "8": { "A": 0.0026, "H": 0.0004, "L": 0.0041, "P": 0.9777, "Q": 0.0056, "R": 0.0011, "S": 0.0036, "T": 0.005 },
    "9": { "A": 0.0253, "C": 0.0001, "D": 0.9313, "E": 0.0194, "G": 0.0026, "H": 0.0023, "K": 0.0001, "L": 0.0001, "N": 0.0058, "P": 0.0002, "Q": 0.0016, "S": 0.0009, "T": 0.0002, "V": 0.003, "Y": 0.0069 },
    "10": { "A": 0.0001, "C": 0.0055, "F": 0.9617, "H": 0.0001, "I": 0.0022, "L": 0.0158, "M": 0.0001, "N": 0.0002, "P": 0.0002, "S": 0.0064, "V": 0.0064, "Y": 0.0013 },
    "11": { "E": 0.0007, "H": 0.008, "K": 0.0048, "L": 0.0084, "N": 0.0001, "P": 0.0158, "Q": 0.9591, "R": 0.002, "S": 0.0001, "T": 0.0005, "V": 0.0002, "Y": 0.0001 },
    "12": { "A": 0.0056, "C": 0.0006, "D": 0.0001, "F": 0.0011, "H": 0.0001, "L": 0.0001, "P": 0.0076, "S": 0.9806, "T": 0.0012, "Y": 0.003 },
    "13": { "A": 0.0077, "E": 0.0012, "F": 0.0001, "G": 0.0064, "L": 0.006, "M": 0.0014, "P": 0.0001, "S": 0.0001, "V": 0.9762, "W": 0.0007 },
    "14": { "A": 0.0134, "C": 0.0001, "H": 0.0001, "I": 0.0048, "K": 0.0001, "L": 0.0002, "M": 0.0001, "N": 0.0029, "P": 0.0107, "Q": 0.0001, "R": 0.0001, "S": 0.0081, "T": 0.9591, "V": 0.0001, "Y": 0.0001 },
    "15": { "A": 0.002, "H": 0.0007, "I": 0.0001, "L": 0.0082, "P": 0.968, "Q": 0.0092, "R": 0.0026, "S": 0.0027, "T": 0.0062, "V": 0.0002 },
    "16": { "A": 0.0003, "E": 0.0077, "G": 0.0018, "H": 0.0034, "I": 0.0002, "K": 0.9301, "L": 0.0001, "M": 0.0028, "N": 0.0126, "P": 0.0001, "Q": 0.0174, "R": 0.0058, "S": 0.0002, "T": 0.0174, "V": 0.0001, "W": 0.0001 },
    "17": { "A": 0.0158, "D": 0.0126, "E": 0.9523, "G": 0.0096, "H": 0.0001, "K": 0.0018, "N": 0.0003, "P": 0.0001, "Q": 0.0041, "R": 0.0002, "S": 0.0004, "T": 0.0001, "V": 0.0024, "W": 0.0001, "Y": 0.0001 },
    "18": { "A": 0.0001, "D": 0.0001, "E": 0.0037, "H": 0.0003, "I": 0.0015, "K": 0.9432, "M": 0.0001, "N": 0.0117, "P": 0.0004, "Q": 0.0092, "R": 0.008, "S": 0.0032, "T": 0.0185, "Y": 0.0001 },
    "19": { "A": 0.0047, "C": 0.0001, "D": 0.0022, "F": 0.0043, "G": 0.0045, "I": 0.0053, "L": 0.0039, "M": 0.0001, "P": 0.0001, "R": 0.0001, "S": 0.0002, "V": 0.9745, "Y": 0.0001 },
    "20": { "A": 0.0015, "H": 0.0004, "I": 0.0049, "L": 0.0002, "N": 0.0067, "P": 0.0098, "R": 0.0001, "S": 0.0026, "T": 0.9738 },
    "21": { "F": 0.0019, "H": 0.0002, "I": 0.9752, "L": 0.008, "M": 0.0009, "N": 0.0029, "P": 0.0001, "R": 0.0001, "S": 0.004, "T": 0.005, "V": 0.0016 },
    "22": { "A": 0.0011, "D": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0005, "I": 0.002, "K": 0.0001, "L": 0.0004, "N": 0.0045, "P": 0.0188, "R": 0.0001, "S": 0.0044, "T": 0.968 },
    "23": { "A": 0.0001, "C": 0.9794, "F": 0.0055, "G": 0.0029, "L": 0.0002, "P": 0.0001, "R": 0.0033, "S": 0.004, "V": 0.0002, "W": 0.0009, "Y": 0.0034 },
    "24": { "C": 0.0001, "G": 0.0008, "L": 0.0055, "M": 0.0001, "P": 0.002, "Q": 0.0034, "R": 0.9852, "S": 0.0001, "W": 0.0028 },
    "25": { "A": 0.9762, "D": 0.0072, "G": 0.0011, "P": 0.0026, "S": 0.0053, "T": 0.0047, "V": 0.0028, "Y": 0.0001 },
    "26": { "A": 0.0001, "C": 0.0013, "G": 0.0018, "I": 0.0018, "K": 0.0001, "M": 0.0002, "N": 0.0018, "R": 0.0185, "S": 0.9728, "T": 0.0015, "V": 0.0001, "W": 0.0001 },
    "27": { "A": 0.0002, "D": 0.0001, "E": 0.0108, "H": 0.0128, "K": 0.0055, "L": 0.0009, "N": 0.0003, "P": 0.0073, "Q": 0.9558, "R": 0.0048, "S": 0.0001, "T": 0.0004, "Y": 0.0007 },
    "28": { "A": 0.0009, "C": 0.0028, "D": 0.0034, "F": 0.0002, "G": 0.0082, "H": 0.0007, "I": 0.0042, "K": 0.0004, "L": 0.0004, "N": 0.035, "P": 0.0005, "Q": 0.0001, "R": 0.0204, "S": 0.9099, "T": 0.0125, "V": 0.0001, "Y": 0.0003 },
    "29": { "A": 0.0001, "D": 0.0001, "F": 0.004, "G": 0.0004, "H": 0.0001, "I": 0.9536, "L": 0.008, "M": 0.0034, "N": 0.0012, "P": 0.0001, "R": 0.0002, "S": 0.008, "T": 0.0022, "V": 0.0186 },
    "36": { "A": 0.006, "C": 0.0129, "D": 0.0077, "E": 0.0008, "F": 0.0001, "G": 0.9484, "I": 0.0001, "N": 0.001, "P": 0.0001, "R": 0.0054, "S": 0.011, "V": 0.0053, "W": 0.0001, "Y": 0.0009 },
    "37": { "A": 0.0001, "C": 0.0014, "D": 0.0055, "E": 0.0003, "F": 0.0001, "G": 0.015, "H": 0.0009, "I": 0.0091, "K": 0.0006, "L": 0.0007, "M": 0.0002, "N": 0.0213, "Q": 0.0001, "R": 0.0201, "S": 0.8862, "T": 0.0332, "V": 0.003, "Y": 0.002 },
    "38": { "A": 0.0039, "C": 0.0018, "D": 0.0079, "F": 0.0004, "G": 0.0141, "H": 0.0023, "I": 0.0047, "K": 0.0026, "M": 0.0001, "N": 0.0365, "P": 0.0004, "Q": 0.0001, "R": 0.0269, "S": 0.8872, "T": 0.0093, "V": 0.0004, "W": 0.0001, "Y": 0.0013 },
    "39": { "C": 0.0004, "F": 0.0062, "G": 0.0001, "I": 0.0043, "L": 0.9774, "R": 0.0001, "S": 0.0016, "V": 0.0099, "W": 0.0001, "Y": 0.0001 },
    "40": { "A": 0.0001, "D": 0.0011, "F": 0.0001, "H": 0.9601, "K": 0.0004, "L": 0.0015, "N": 0.0114, "P": 0.0086, "Q": 0.0119, "R": 0.0015, "S": 0.0006, "T": 0.0007, "Y": 0.0021 },
    "41": { "C": 0.0056, "G": 0.0043, "K": 0.0001, "L": 0.0055, "M": 0.0001, "P": 0.0001, "R": 0.0034, "S": 0.0013, "V": 0.0003, "W": 0.9792, "Y": 0.0001 },
    "42": { "A": 0.0002, "C": 0.0018, "D": 0.0037, "E": 0.0001, "F": 0.0064, "H": 0.0017, "K": 0.0001, "L": 0.0004, "N": 0.0012, "P": 0.0001, "S": 0.0049, "Y": 0.9796 },
    "43": { "E": 0.0004, "H": 0.005, "K": 0.0042, "L": 0.0013, "N": 0.0002, "P": 0.0039, "Q": 0.9823, "R": 0.002, "S": 0.0002, "T": 0.0004, "Y": 0.0001 },
    "44": { "E": 0.002, "H": 0.0096, "I": 0.0001, "K": 0.0063, "L": 0.0015, "M": 0.0037, "N": 0.0001, "P": 0.0044, "Q": 0.9692, "R": 0.0025, "T": 0.0001, "V": 0.0004, "Y": 0.0001 },
    "45": { "D": 0.0001, "E": 0.0019, "H": 0.0005, "I": 0.0012, "K": 0.9577, "N": 0.0153, "P": 0.0001, "Q": 0.0086, "R": 0.0046, "T": 0.0099, "Y": 0.0001 },
    "46": { "A": 0.0082, "E": 0.0001, "G": 0.0036, "H": 0.0002, "I": 0.0001, "K": 0.0001, "L": 0.0026, "P": 0.9683, "Q": 0.0055, "R": 0.0007, "S": 0.0044, "T": 0.0062, "V": 0.0001 },
    "47": { "A": 0.0085, "C": 0.0001, "D": 0.903, "E": 0.0251, "F": 0.0001, "G": 0.0348, "H": 0.0067, "K": 0.0003, "L": 0.0004, "N": 0.0129, "P": 0.0013, "Q": 0.0004, "R": 0.0002, "S": 0.0001, "T": 0.0001, "V": 0.0018, "Y": 0.004 },
    "48": { "A": 0.0001, "E": 0.0013, "H": 0.0074, "K": 0.0056, "L": 0.0007, "M": 0.0001, "P": 0.0061, "Q": 0.9777, "R": 0.0009, "T": 0.0001 },
    "49": { "A": 0.0128, "C": 0.001, "D": 0.0001, "F": 0.0018, "H": 0.0001, "P": 0.0137, "S": 0.9608, "T": 0.0037, "Y": 0.0059 },
    "50": { "A": 0.0006, "H": 0.0004, "L": 0.0007, "P": 0.985, "Q": 0.0076, "R": 0.0005, "S": 0.0012, "T": 0.0039 },
    "51": { "D": 0.0002, "E": 0.0026, "G": 0.0001, "H": 0.0003, "I": 0.0001, "K": 0.9366, "L": 0.0001, "M": 0.0008, "N": 0.0155, "P": 0.0001, "Q": 0.009, "R": 0.0215, "S": 0.0015, "T": 0.0116 },
    "52": { "A": 0.0006, "C": 0.0001, "D": 0.0001, "F": 0.0018, "G": 0.0001, "H": 0.0034, "I": 0.0096, "L": 0.9543, "P": 0.0101, "R": 0.0098, "S": 0.0003, "T": 0.0001, "V": 0.0098 },
    "53": { "A": 0.0001, "F": 0.0012, "H": 0.0021, "I": 0.0047, "L": 0.9723, "P": 0.0062, "R": 0.0093, "S": 0.0002, "T": 0.0001, "V": 0.0038 },
    "54": { "A": 0.0001, "F": 0.0041, "H": 0.0003, "I": 0.9649, "L": 0.0057, "M": 0.0025, "N": 0.0034, "P": 0.0001, "R": 0.0002, "S": 0.009, "T": 0.0028, "V": 0.0069 },
    "55": { "D": 0.0011, "E": 0.0042, "G": 0.0002, "H": 0.0002, "I": 0.0003, "K": 0.9307, "L": 0.0002, "M": 0.0085, "N": 0.0102, "P": 0.0001, "Q": 0.0149, "R": 0.0167, "S": 0.0007, "T": 0.0103, "V": 0.0001, "W": 0.0003, "Y": 0.0012 },
    "56": { "A": 0.0004, "C": 0.0019, "D": 0.0169, "E": 0.0007, "F": 0.0204, "G": 0.0001, "H": 0.0051, "I": 0.0001, "K": 0.0002, "L": 0.0004, "N": 0.003, "P": 0.0001, "Q": 0.0007, "S": 0.0158, "T": 0.0001, "V": 0.0008, "W": 0.0002, "Y": 0.9331 },
    "57": { "A": 0.9668, "C": 0.0001, "D": 0.0039, "E": 0.0001, "G": 0.0023, "H": 0.0001, "I": 0.0001, "K": 0.0001, "P": 0.0006, "R": 0.0001, "S": 0.0077, "T": 0.0073, "V": 0.0107, "Y": 0.0001 },
    "65": { "A": 0.0062, "C": 0.0012, "D": 0.0001, "F": 0.0023, "H": 0.0001, "I": 0.0001, "K": 0.0001, "L": 0.0004, "N": 0.0001, "P": 0.0018, "S": 0.9778, "T": 0.0026, "V": 0.0001, "Y": 0.0069 },
    "66": { "E": 0.0037, "H": 0.0113, "K": 0.0075, "L": 0.0016, "N": 0.0004, "P": 0.009, "Q": 0.9568, "R": 0.0087, "S": 0.0001, "T": 0.0009 },
    "67": { "A": 0.0174, "C": 0.0018, "D": 0.0008, "F": 0.0032, "H": 0.0004, "L": 0.0004, "P": 0.0065, "R": 0.0002, "S": 0.9546, "T": 0.0053, "V": 0.0002, "Y": 0.0091 },
    "68": { "A": 0.0012, "C": 0.0172, "D": 0.0004, "E": 0.0001, "F": 0.7651, "G": 0.0039, "H": 0.0001, "I": 0.1488, "K": 0.0001, "L": 0.0191, "M": 0.0025, "N": 0.0009, "P": 0.0002, "R": 0.0011, "S": 0.0107, "T": 0.0013, "V": 0.0239, "W": 0.0004, "Y": 0.0029 },
    "69": { "A": 0.0248, "D": 0.0001, "E": 0.0004, "G": 0.0001, "H": 0.0001, "I": 0.0002, "K": 0.0001, "L": 0.002, "P": 0.0142, "Q": 0.0003, "R": 0.0001, "S": 0.9504, "T": 0.0068, "Y": 0.0005 },
    "70": { "A": 0.002, "C": 0.0002, "E": 0.0013, "G": 0.9839, "K": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.0035, "S": 0.0001, "V": 0.0061, "W": 0.0028 },
    "71": { "A": 0.0069, "C": 0.0006, "D": 0.0012, "F": 0.0048, "G": 0.0121, "I": 0.0021, "L": 0.0018, "P": 0.0001, "R": 0.0001, "S": 0.0001, "V": 0.9701, "Y": 0.0001 },
    "72": { "A": 0.0012, "C": 0.0001, "H": 0.0026, "L": 0.0012, "N": 0.0001, "P": 0.9873, "R": 0.0005, "S": 0.0034, "T": 0.0036 },
    "74": { "A": 0.0161, "E": 0.0004, "G": 0.0001, "L": 0.0009, "P": 0.0088, "R": 0.0001, "S": 0.9704, "T": 0.0016, "W": 0.0017, "Y": 0.0001 },
    "75": { "E": 0.0001, "G": 0.0067, "I": 0.0001, "K": 0.0035, "L": 0.0002, "M": 0.0078, "Q": 0.0001, "R": 0.9696, "S": 0.0077, "T": 0.0016, "V": 0.0004, "W": 0.0022 },
    "76": { "C": 0.0103, "F": 0.968, "G": 0.0001, "I": 0.0011, "L": 0.0077, "M": 0.0001, "P": 0.0001, "S": 0.0009, "V": 0.0111, "Y": 0.0006 },
    "77": { "A": 0.0001, "C": 0.0005, "G": 0.0031, "H": 0.0001, "I": 0.0035, "K": 0.0002, "M": 0.0001, "N": 0.0027, "P": 0.0001, "R": 0.0199, "S": 0.9649, "T": 0.0044, "V": 0.0003, "Y": 0.0001 },
    "78": { "A": 0.0019, "C": 0.0077, "D": 0.0019, "G": 0.9803, "L": 0.0001, "R": 0.001, "S": 0.0023, "V": 0.0047 },
    "79": { "C": 0.0011, "D": 0.0001, "G": 0.0088, "H": 0.0001, "I": 0.0034, "K": 0.0013, "M": 0.0001, "N": 0.0014, "Q": 0.0001, "R": 0.014, "S": 0.9675, "T": 0.0019, "W": 0.0001, "Y": 0.0001 },
    "80": { "A": 0.0012, "C": 0.0004, "E": 0.0014, "G": 0.9845, "K": 0.0001, "L": 0.0002, "Q": 0.0003, "R": 0.0061, "S": 0.0002, "T": 0.0001, "V": 0.0053, "W": 0.0001 },
    "83": { "A": 0.015, "C": 0.0018, "D": 0.0001, "E": 0.0026, "F": 0.0023, "G": 0.0001, "H": 0.0005, "N": 0.0001, "P": 0.0021, "S": 0.9668, "T": 0.0018, "Y": 0.0066 },
    "84": { "A": 0.0014, "C": 0.0002, "D": 0.0001, "E": 0.0043, "F": 0.0001, "G": 0.9805, "K": 0.0001, "R": 0.0025, "S": 0.0001, "T": 0.0001, "V": 0.0053, "W": 0.0053 },
    "85": { "A": 0.0042, "E": 0.0001, "G": 0.0001, "I": 0.0012, "K": 0.0061, "L": 0.0001, "N": 0.0002, "P": 0.0012, "R": 0.0037, "S": 0.0009, "T": 0.9822, "V": 0.0001 },
    "86": { "A": 0.0025, "C": 0.0001, "D": 0.9562, "E": 0.0212, "G": 0.0017, "H": 0.0026, "K": 0.0001, "L": 0.001, "N": 0.0031, "P": 0.0001, "Q": 0.0002, "S": 0.0001, "T": 0.0017, "V": 0.0038, "Y": 0.0057 },
    "87": { "C": 0.0054, "D": 0.0001, "F": 0.9778, "G": 0.0001, "I": 0.0009, "L": 0.0045, "S": 0.0013, "V": 0.0059, "Y": 0.0039 },
    "88": { "A": 0.0054, "D": 0.0001, "F": 0.0003, "H": 0.0001, "I": 0.0029, "K": 0.0001, "L": 0.0004, "M": 0.0001, "N": 0.0073, "P": 0.0074, "R": 0.0004, "S": 0.0046, "T": 0.9711 },
    "89": { "F": 0.003, "G": 0.0001, "H": 0.0008, "I": 0.0069, "K": 0.0001, "L": 0.9793, "P": 0.0023, "R": 0.0065, "S": 0.0003, "T": 0.0001, "V": 0.0007 },
    "90": { "A": 0.0021, "H": 0.0001, "I": 0.0038, "K": 0.0002, "L": 0.0001, "N": 0.0067, "P": 0.0031, "R": 0.0001, "S": 0.0053, "T": 0.9785, "V": 0.0001 },
    "91": { "F": 0.0004, "I": 0.9845, "K": 0.0001, "L": 0.002, "M": 0.0012, "N": 0.0015, "R": 0.0001, "S": 0.0079, "T": 0.0014, "V": 0.0009 },
    "92": { "D": 0.0195, "E": 0.0005, "G": 0.0001, "H": 0.0067, "I": 0.002, "K": 0.0137, "L": 0.0001, "M": 0.0001, "N": 0.9277, "P": 0.0002, "Q": 0.0004, "R": 0.0009, "S": 0.0219, "T": 0.0038, "V": 0.0001, "Y": 0.0025 },
    "93": { "A": 0.0013, "C": 0.0031, "D": 0.0012, "G": 0.0387, "H": 0.0001, "I": 0.0063, "K": 0.0001, "L": 0.0004, "M": 0.0002, "N": 0.0057, "P": 0.0003, "R": 0.0193, "S": 0.9206, "T": 0.0025, "V": 0.0002 },
    "94": { "A": 0.0001, "G": 0.0002, "H": 0.0001, "L": 0.9776, "M": 0.0036, "P": 0.002, "Q": 0.0018, "R": 0.0103, "V": 0.0042, "W": 0.0001 },
    "95": { "A": 0.0031, "D": 0.004, "E": 0.9844, "G": 0.0022, "I": 0.0001, "K": 0.0012, "Q": 0.0036, "R": 0.0001, "T": 0.0001, "V": 0.0011, "Y": 0.0001 },
    "96": { "A": 0.9463, "C": 0.0001, "D": 0.0045, "E": 0.0001, "F": 0.0004, "G": 0.0047, "I": 0.0001, "L": 0.0001, "P": 0.0137, "R": 0.0003, "S": 0.0045, "T": 0.0074, "V": 0.0177 },
    "97": { "A": 0.0019, "D": 0.0063, "E": 0.9869, "G": 0.0027, "I": 0.0001, "K": 0.0006, "Q": 0.0004, "S": 0.0001, "T": 0.0001, "V": 0.0009, "Y": 0.0001 },
    "98": { "A": 0.0013, "D": 0.986, "E": 0.0058, "G": 0.0011, "H": 0.0005, "M": 0.0001, "N": 0.0008, "S": 0.0001, "V": 0.0009, "Y": 0.0034 },
    "99": { "A": 0.9728, "C": 0.0001, "D": 0.0032, "E": 0.0002, "F": 0.0006, "G": 0.0012, "I": 0.0012, "L": 0.0001, "P": 0.0003, "S": 0.0061, "T": 0.0018, "V": 0.0126 },
    "100": { "A": 0.988, "E": 0.0039, "G": 0.0017, "P": 0.0004, "Q": 0.0001, "S": 0.0031, "T": 0.0015, "V": 0.0013 },
    "101": { "A": 0.022, "E": 0.0012, "G": 0.0002, "I": 0.0014, "K": 0.0026, "L": 0.0001, "M": 0.0027, "N": 0.0001, "P": 0.0013, "R": 0.0006, "S": 0.0027, "T": 0.9532, "V": 0.0116, "W": 0.0002 },
    "102": { "C": 0.0008, "D": 0.0036, "E": 0.0001, "F": 0.0004, "H": 0.0009, "I": 0.0001, "N": 0.0004, "Q": 0.0001, "S": 0.0013, "Y": 0.9923 },
    "103": { "C": 0.0027, "D": 0.0039, "F": 0.0195, "G": 0.0001, "H": 0.0011, "L": 0.0001, "N": 0.0006, "S": 0.0015, "T": 0.0001, "V": 0.0004, "Y": 0.97 },
    "104": { "A": 0.0001, "C": 0.9818, "F": 0.0031, "G": 0.0029, "I": 0.0001, "L": 0.0002, "P": 0.0001, "R": 0.0019, "S": 0.0009, "V": 0.0001, "W": 0.0069, "Y": 0.0018 },
    "105": { "C": 0.0005, "D": 0.0003, "E": 0.0001, "G": 0.0001, "H": 0.9236, "I": 0.0001, "K": 0.0003, "L": 0.0072, "N": 0.0023, "P": 0.0015, "Q": 0.0518, "R": 0.0034, "V": 0.0001, "W": 0.0001, "Y": 0.0085 },
    "106": { "A": 0.0001, "E": 0.0007, "G": 0.0001, "H": 0.0191, "K": 0.0028, "L": 0.0004, "N": 0.0001, "P": 0.0023, "Q": 0.9698, "R": 0.003, "S": 0.0006, "T": 0.0007, "V": 0.0001, "Y": 0.0002 },
    "107": { "C": 0.0037, "D": 0.0001, "E": 0.0008, "F": 0.0001, "G": 0.0071, "H": 0.0001, "I": 0.005, "K": 0.0002, "L": 0.0003, "N": 0.0026, "Q": 0.0006, "R": 0.017, "S": 0.9422, "T": 0.019, "V": 0.0001, "W": 0.0001, "Y": 0.0009 },
    "108": { "A": 0.0083, "C": 0.0024, "D": 0.0093, "E": 0.0006, "F": 0.0007, "G": 0.0125, "H": 0.0028, "I": 0.0021, "K": 0.0038, "L": 0.0002, "M": 0.0002, "N": 0.0135, "P": 0.0003, "R": 0.0183, "S": 0.9089, "T": 0.0089, "V": 0.0007, "W": 0.0004, "Y": 0.0058 },
    "109": { "A": 0.0027, "C": 0.0005, "D": 0.0052, "E": 0.0006, "F": 0.0009, "G": 0.0083, "H": 0.0007, "I": 0.0055, "K": 0.0007, "L": 0.0008, "M": 0.0001, "N": 0.0182, "Q": 0.0002, "R": 0.0113, "S": 0.9195, "T": 0.0213, "V": 0.0027, "Y": 0.0011 },
    "113": { "C": 0.0022, "F": 0.0513, "H": 0.0156, "L": 0.7879, "P": 0.1161, "Q": 0.0022, "S": 0.0089, "T": 0.0045, "V": 0.0045, "W": 0.0022, "Y": 0.0045 },
    "114": { "A": 0.0007, "C": 0.0009, "D": 0.0001, "E": 0.0001, "F": 0.072, "G": 0.0025, "H": 0.0019, "I": 0.0037, "K": 0.0006, "L": 0.8056, "M": 0.0007, "N": 0.0002, "P": 0.026, "Q": 0.0021, "R": 0.0028, "S": 0.0649, "T": 0.0016, "V": 0.0083, "W": 0.0002, "Y": 0.005 },
    "115": { "A": 0.0069, "C": 0.0004, "D": 0.0004, "E": 0.0022, "F": 0.0077, "G": 0.0057, "H": 0.0039, "I": 0.0001, "K": 0.001, "L": 0.0549, "M": 0.0024, "P": 0.8567, "Q": 0.0088, "R": 0.0144, "S": 0.0236, "T": 0.0071, "V": 0.0011, "W": 0.001, "Y": 0.0015 },
    "116": { "A": 0.0029, "C": 0.005, "D": 0.0072, "E": 0.011, "F": 0.0547, "G": 0.0452, "H": 0.0676, "I": 0.0514, "K": 0.0064, "L": 0.1523, "M": 0.0005, "N": 0.0053, "P": 0.0525, "Q": 0.0687, "R": 0.0756, "S": 0.0106, "T": 0.0018, "V": 0.015, "W": 0.2376, "Y": 0.1288 },
    "117": { "A": 0.0066, "C": 0.0001, "E": 0.0004, "F": 0.0001, "G": 0.0026, "H": 0.0014, "I": 0.0009, "K": 0.001, "L": 0.0023, "M": 0.0007, "N": 0.0009, "P": 0.0069, "Q": 0.0003, "R": 0.001, "S": 0.035, "T": 0.9391, "V": 0.0007, "W": 0.0001 },
    "118": { "A": 0.0001, "C": 0.006, "D": 0.0004, "F": 0.9828, "G": 0.0003, "H": 0.0001, "I": 0.0005, "K": 0.0008, "L": 0.0034, "P": 0.0001, "Q": 0.0003, "R": 0.0001, "S": 0.0009, "V": 0.0039, "W": 0.0001, "Y": 0.0003 },
    "119": { "A": 0.0006, "C": 0.0007, "D": 0.0012, "G": 0.9936, "H": 0.0002, "L": 0.0001, "P": 0.0002, "R": 0.0007, "S": 0.0018, "V": 0.0009 },
    "120": { "A": 0.0009, "E": 0.0014, "G": 0.1876, "H": 0.0045, "K": 0.0012, "L": 0.0002, "P": 0.0615, "Q": 0.7371, "R": 0.0053, "S": 0.0002, "V": 0.0001, "Y": 0.0001 },
    "121": { "A": 0.0004, "C": 0.0001, "E": 0.0036, "G": 0.9899, "R": 0.0027, "V": 0.001, "W": 0.0022 },
    "122": { "A": 0.0028, "I": 0.0009, "K": 0.0003, "N": 0.0007, "P": 0.0007, "S": 0.0036, "T": 0.991 },
    "123": { "E": 0.0006, "K": 0.9031, "L": 0.0007, "M": 0.0001, "N": 0.0009, "P": 0.0001, "Q": 0.0035, "R": 0.0888, "T": 0.0021 },
    "124": { "A": 0.0004, "C": 0.0001, "E": 0.0004, "F": 0.0001, "G": 0.0004, "I": 0.001, "L": 0.3258, "M": 0.0025, "P": 0.0002, "Q": 0.0002, "R": 0.0005, "V": 0.6684 },
    "125": { "A": 0.0011, "D": 0.0618, "E": 0.9313, "G": 0.003, "K": 0.0013, "N": 0.0004, "Q": 0.0003, "R": 0.0002, "V": 0.0007 },
    "126": { "D": 0.0001, "F": 0.0009, "I": 0.9825, "L": 0.0021, "M": 0.0019, "N": 0.0008, "S": 0.0034, "T": 0.0028, "V": 0.0055 },
    "127": { "E": 0.001, "K": 0.9946, "N": 0.0005, "Q": 0.0039 }
  },
  "IGKV6D": {
    "1": { "D": 0.0352, "E": 0.9648 },
    "2": { "F": 0.0031, "H": 0.0003, "I": 0.9442, "K": 0.0001, "L": 0.021, "M": 0.007, "N": 0.0026, "R": 0.0002, "S": 0.0051, "T": 0.0047, "V": 0.0117 },
    "3": { "A": 0.0091, "C": 0.0001, "D": 0.0001, "E": 0.0016, "F": 0.0016, "G": 0.0027, "I": 0.0009, "L": 0.0118, "M": 0.0021, "P": 0.0001, "R": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.9695 },
    "4": { "F": 0.0002, "I": 0.0005, "L": 0.9797, "M": 0.0094, "P": 0.0028, "Q": 0.0009, "R": 0.0028, "T": 0.0003, "V": 0.0034, "W": 0.0001 },
    "5": { "A": 0.0021, "D": 0.0003, "E": 0.0001, "H": 0.0001, "I": 0.0039, "K": 0.0002, "L": 0.0001, "N": 0.0046, "P": 0.0151, "Q": 0.0001, "R": 0.0001, "S": 0.0039, "T": 0.9689, "V": 0.0001, "Y": 0.0002 },
    "6": { "E": 0.0014, "G": 0.0001, "H": 0.0134, "K": 0.0056, "L": 0.002, "N": 0.0001, "P": 0.0184, "Q": 0.9555, "R": 0.0021, "S": 0.0004, "T": 0.0006, "V": 0.0001, "Y": 0.0001 },
    "7": { "A": 0.007, "C": 0.0014, "D": 0.0002, "F": 0.0049, "G": 0.0002, "H": 0.0004, "L": 0.0006, "P": 0.0067, "S": 0.9677, "T": 0.0029, "V": 0.0001, "Y": 0.0078 },
    "8": { "A": 0.0016, "E": 0.0001, "H": 0.0004, "I": 0.0001, "L": 0.0044, "P": 0.9783, "Q": 0.0056, "R": 0.0018, "S": 0.0018, "T": 0.0059 },
    "9": { "A": 0.024, "D": 0.9429, "E": 0.0067, "F": 0.0001, "G": 0.0038, "H": 0.004, "N": 0.004, "P": 0.0003, "Q": 0.0018, "S": 0.0007, "T": 0.0002, "V": 0.0043, "Y": 0.007 },
    "10": { "C": 0.005, "F": 0.9684, "I": 0.0005, "L": 0.0124, "M": 0.0001, "P": 0.0002, "S": 0.006, "V": 0.0063, "Y": 0.0012 },
    "11": { "C": 0.0001, "E": 0.0002, "H": 0.0084, "K": 0.0027, "L": 0.0033, "M": 0.0001, "P": 0.0151, "Q": 0.9669, "R": 0.0025, "S": 0.0003, "T": 0.0003, "Y": 0.0001 },
    "12": { "A": 0.0035, "C": 0.0002, "F": 0.0006, "P": 0.0036, "S": 0.9867, "T": 0.0015, "V": 0.0001, "Y": 0.0037 },
    "13": { "A": 0.0056, "C": 0.0001, "E": 0.0013, "F": 0.0002, "G": 0.0076, "L": 0.0036, "M": 0.001, "R": 0.0002, "S": 0.0002, "V": 0.9801, "W": 0.0001 },
    "14": { "A": 0.0081, "C": 0.0001, "H": 0.0004, "I": 0.001, "L": 0.0002, "M": 0.0001, "N": 0.0039, "P": 0.0105, "R": 0.0001, "S": 0.0047, "T": 0.9708 },
    "15": { "A": 0.0017, "H": 0.0005, "L": 0.0034, "P": 0.9742, "Q": 0.0084, "R": 0.0014, "S": 0.0023, "T": 0.0079, "V": 0.0001 },
    "16": { "A": 0.0003, "E": 0.0044, "G": 0.0025, "H": 0.0003, "I": 0.0001, "K": 0.9335, "L": 0.0001, "M": 0.0028, "N": 0.0061, "P": 0.0003, "Q": 0.0181, "R": 0.0141, "S": 0.0004, "T": 0.0167, "V": 0.0002, "W": 0.0001 },
    "17": { "A": 0.0137, "D": 0.0097, "E": 0.9569, "G": 0.0048, "H": 0.0003, "K": 0.0027, "L": 0.0001, "P": 0.0003, "Q": 0.0087, "S": 0.0001, "T": 0.0003, "V": 0.0021, "W": 0.0001, "Y": 0.0001 },
    "18": { "D": 0.0002, "E": 0.0067, "G": 0.0001, "H": 0.0001, "I": 0.0004, "K": 0.945, "M": 0.0001, "N": 0.0111, "P": 0.0002, "Q": 0.0093, "R": 0.0102, "S": 0.0001, "T": 0.0163, "V": 0.0001 },
    "19": { "A": 0.0035, "C": 0.0002, "D": 0.0015, "F": 0.0038, "G": 0.006, "I": 0.0023, "L": 0.0018, "P": 0.0001, "S": 0.0005, "V": 0.9801, "Y": 0.0001 },
    "20": { "A": 0.0015, "F": 0.0001, "H": 0.0005, "I": 0.0061, "N": 0.0058, "P": 0.0083, "R": 0.0001, "S": 0.0019, "T": 0.9757 },
    "21": { "F": 0.002, "H": 0.0001, "I": 0.9699, "L": 0.016, "M": 0.0004, "N": 0.0021, "P": 0.0002, "R": 0.0003, "S": 0.0032, "T": 0.0047, "V": 0.0011 },
    "22": { "A": 0.0013, "H": 0.0002, "I": 0.0017, "L": 0.0001, "N": 0.0038, "P": 0.018, "S": 0.0022, "T": 0.9726 },
    "23": { "C": 0.9779, "F": 0.0078, "G": 0.0035, "L": 0.0001, "R": 0.0025, "S": 0.0037, "T": 0.0001, "V": 0.0003, "W": 0.0007, "Y": 0.0033 },
    "24": { "G": 0.0015, "L": 0.0063, "M": 0.0001, "P": 0.0018, "Q": 0.0022, "R": 0.9844, "S": 0.0002, "T": 0.0001, "W": 0.0033 },
    "25": { "A": 0.9785, "D": 0.0069, "F": 0.0001, "G": 0.0014, "H": 0.0002, "I": 0.0001, "L": 0.0001, "P": 0.003, "S": 0.0031, "T": 0.0042, "V": 0.0022, "Y": 0.0002 },
    "26": { "C": 0.0007, "G": 0.0068, "I": 0.0028, "M": 0.0004, "N": 0.0022, "Q": 0.0001, "R": 0.0146, "S": 0.969, "T": 0.0032, "V": 0.0001 },
    "27": { "E": 0.0044, "H": 0.0124, "K": 0.0054, "L": 0.0014, "P": 0.0067, "Q": 0.965, "R": 0.0047, "T": 0.0001 },
    "28": { "A": 0.0027, "C": 0.0026, "D": 0.0004, "G": 0.0047, "H": 0.0004, "I": 0.0031, "L": 0.0001, "N": 0.013, "P": 0.0001, "R": 0.0183, "S": 0.9475, "T": 0.0069, "V": 0.0001, "W": 0.0001 },
    "29": { "F": 0.0006, "H": 0.0001, "I": 0.9742, "L": 0.0069, "M": 0.0027, "N": 0.0014, "R": 0.0001, "S": 0.0067, "T": 0.0026, "V": 0.0047 },
    "36": { "A": 0.0017, "C": 0.012, "D": 0.0034, "E": 0.0002, "F": 0.0001, "G": 0.9659, "N": 0.0026, "P": 0.0001, "R": 0.0036, "S": 0.006, "V": 0.0041, "W": 0.0002, "Y": 0.0001 },
    "37": { "A": 0.0001, "C": 0.0013, "D": 0.0015, "F": 0.0001, "G": 0.0065, "H": 0.0003, "I": 0.0146, "K": 0.0002, "L": 0.0004, "M": 0.0001, "N": 0.0193, "P": 0.0003, "Q": 0.002, "R": 0.0135, "S": 0.9072, "T": 0.0316, "V": 0.0001, "Y": 0.0007 },
    "38": { "A": 0.007, "C": 0.0029, "D": 0.0014, "G": 0.0059, "H": 0.0005, "I": 0.0063, "K": 0.0004, "L": 0.0005, "M": 0.0001, "N": 0.0275, "P": 0.0001, "R": 0.0125, "S": 0.927, "T": 0.0071, "V": 0.0001, "W": 0.0001, "Y": 0.0005 },
    "39": { "A": 0.0001, "C": 0.0001, "F": 0.0067, "G": 0.0001, "I": 0.0034, "L": 0.9798, "S": 0.0011, "V": 0.0087 },
    "40": { "A": 0.0002, "D": 0.0009, "H": 0.9679, "K": 0.0002, "L": 0.0011, "N": 0.0098, "P": 0.0088, "Q": 0.0045, "R": 0.001, "S": 0.0022, "T": 0.0006, "V": 0.0001, "Y": 0.0027 },
    "41": { "C": 0.0051, "E": 0.0001, "F": 0.0002, "G": 0.0039, "L": 0.0058, "N": 0.0003, "Q": 0.0001, "R": 0.0032, "S": 0.0009, "T": 0.0001, "V": 0.0001, "W": 0.9802 },
    "42": { "C": 0.0011, "D": 0.0033, "F": 0.008, "H": 0.0014, "L": 0.0001, "N": 0.0012, "S": 0.0038, "T": 0.0001, "Y": 0.981 },
    "43": { "E": 0.0001, "H": 0.005, "K": 0.004, "L": 0.0009, "N": 0.0001, "P": 0.0033, "Q": 0.9852, "R": 0.0014 },
    "44": { "E": 0.0006, "H": 0.0094, "K": 0.0047, "L": 0.0007, "N": 0.0001, "P": 0.0031, "Q": 0.9801, "R": 0.0012, "T": 0.0001 },
    "45": { "A": 0.0001, "E": 0.0011, "G": 0.0002, "H": 0.0003, "I": 0.0013, "K": 0.9614, "L": 0.0001, "N": 0.0126, "P": 0.0001, "Q": 0.0082, "R": 0.005, "S": 0.0003, "T": 0.0093 },
    "46": { "A": 0.008, "H": 0.0002, "L": 0.0005, "P": 0.9777, "Q": 0.005, "R": 0.0006, "S": 0.0023, "T": 0.0054, "V": 0.0001 },
    "47": { "A": 0.008, "C": 0.0001, "D": 0.9232, "E": 0.0232, "G": 0.0198, "H": 0.0046, "K": 0.0003, "N": 0.0144, "Q": 0.0002, "S": 0.0004, "V": 0.0017, "Y": 0.004 },
    "48": { "D": 0.0001, "E": 0.0004, "H": 0.0077, "K": 0.0043, "L": 0.0011, "P": 0.0084, "Q": 0.976, "R": 0.0017, "S": 0.0001, "T": 0.0002 },
    "49": { "A": 0.0143, "C": 0.0002, "D": 0.0001, "E": 0.0001, "F": 0.0012, "N": 0.0001, "P": 0.0066, "S": 0.9672, "T": 0.004, "V": 0.0001, "Y": 0.0061 },
    "50": { "A": 0.001, "H": 0.0004, "L": 0.0007, "P": 0.986, "Q": 0.0066, "R": 0.0007, "S": 0.0012, "T": 0.0033 },
    "51": { "E": 0.001, "I": 0.0004, "K": 0.9608, "M": 0.0009, "N": 0.0099, "Q": 0.0089, "R": 0.0094, "S": 0.0001, "T": 0.0086 },
    "52": { "C": 0.0001, "F": 0.0005, "H": 0.0042, "I": 0.0038, "L": 0.9701, "N": 0.0001, "P": 0.0089, "R": 0.0098, "S": 0.0001, "T": 0.0002, "V": 0.0021 },
    "53": { "F": 0.0006, "H": 0.0032, "I": 0.0036, "L": 0.976, "P": 0.005, "R": 0.0089, "S": 0.0001, "V": 0.0025 },
    "54": { "D": 0.0001, "F": 0.0006, "H": 0.0001, "I": 0.972, "L": 0.007, "M": 0.0026, "N": 0.0033, "P": 0.0002, "R": 0.0003, "S": 0.0091, "T": 0.0025, "V": 0.0022 },
    "55": { "A": 0.0001, "E": 0.0055, "H": 0.0001, "I": 0.0002, "K": 0.9495, "M": 0.0013, "N": 0.0114, "P": 0.0001, "Q": 0.0154, "R": 0.0097, "S": 0.0001, "T": 0.0063, "W": 0.0002 },
    "56": { "A": 0.0004, "C": 0.0016, "D": 0.0132, "F": 0.019, "G": 0.0002, "H": 0.0034, "L": 0.0009, "N": 0.0013, "Q": 0.0001, "S": 0.0099, "V": 0.0007, "Y": 0.9493 },
    "57": { "A": 0.9761, "D": 0.0026, "E": 0.0001, "G": 0.0058, "P": 0.0003, "S": 0.0065, "T": 0.0065, "V": 0.0021 },
    "65": { "A": 0.0101, "C": 0.0016, "D": 0.0004, "F": 0.0015, "L": 0.0001, "P": 0.002, "S": 0.9766, "T": 0.0027, "Y": 0.005 },
    "66": { "E": 0.0054, "G": 0.0001, "H": 0.0084, "K": 0.008, "L": 0.0015, "M": 0.0001, "N": 0.001, "P": 0.0074, "Q": 0.9627, "R": 0.0045, "T": 0.001 },
    "67": { "A": 0.017, "C": 0.0015, "D": 0.0009, "F": 0.0021, "G": 0.0001, "H": 0.0006, "L": 0.0002, "P": 0.0046, "R": 0.0001, "S": 0.9603, "T": 0.0045, "V": 0.0003, "Y": 0.0078 },
    "68": { "C": 0.0033, "D": 0.0001, "G": 0.0009, "H": 0.0003, "I": 0.924, "K": 0.0004, "L": 0.0064, "M": 0.012, "N": 0.0062, "P": 0.0001, "R": 0.0026, "S": 0.0336, "T": 0.004, "V": 0.006, "Y": 0.0001 },
    "69": { "A": 0.0276, "C": 0.0002, "E": 0.0002, "F": 0.0001, "L": 0.0013, "P": 0.0114, "Q": 0.0002, "R": 0.0002, "S": 0.9518, "T": 0.0068, "Y": 0.0001 },
    "70": { "A": 0.002, "C": 0.0002, "E": 0.0014, "G": 0.9843, "M": 0.0001, "Q": 0.0001, "R": 0.0034, "S": 0.0003, "V": 0.0059, "W": 0.0022 },
    "71": { "A": 0.0018, "C": 0.0003, "D": 0.0005, "F": 0.0025, "G": 0.008, "I": 0.002, "L": 0.0013, "V": 0.9836 },
    "72": { "A": 0.0012, "H": 0.0027, "L": 0.0021, "P": 0.9895, "R": 0.0004, "S": 0.0013, "T": 0.0029 },
    "74": { "A": 0.016, "E": 0.0005, "L": 0.0012, "P": 0.0096, "S": 0.9697, "T": 0.0021, "V": 0.0002, "W": 0.0005, "Y": 0.0001 },
    "75": { "E": 0.0004, "G": 0.008, "I": 0.0002, "K": 0.0031, "L": 0.0002, "M": 0.0071, "P": 0.0001, "Q": 0.0002, "R": 0.9701, "S": 0.0072, "T": 0.002, "W": 0.0013 },
    "76": { "A": 0.0001, "C": 0.0064, "F": 0.9727, "I": 0.0011, "L": 0.0084, "S": 0.0018, "V": 0.0084, "Y": 0.0011 },
    "77": { "C": 0.0009, "G": 0.0035, "I": 0.0033, "L": 0.0004, "N": 0.0005, "R": 0.0138, "S": 0.9746, "T": 0.0028, "V": 0.0001 },
    "78": { "A": 0.0018, "C": 0.0069, "D": 0.0036, "G": 0.9808, "H": 0.0001, "R": 0.0007, "S": 0.0021, "V": 0.0038 },
    "79": { "A": 0.0001, "C": 0.0007, "G": 0.0033, "I": 0.0019, "M": 0.0001, "N": 0.0003, "R": 0.0104, "S": 0.9822, "T": 0.0009 },
    "80": { "A": 0.0016, "C": 0.0001, "E": 0.0018, "G": 0.9899, "P": 0.0001, "R": 0.0033, "S": 0.0002, "V": 0.003 },
    "83": { "A": 0.0105, "C": 0.0007, "F": 0.002, "G": 0.0005, "P": 0.0011, "S": 0.9769, "T": 0.001, "V": 0.0021, "Y": 0.0051 },
    "84": { "A": 0.0016, "C": 0.0001, "E": 0.0049, "G": 0.9804, "L": 0.0001, "M": 0.0001, "R": 0.0021, "S": 0.0003, "V": 0.0042, "W": 0.0062 },
    "85": { "A": 0.0053, "G": 0.0003, "I": 0.0019, "K": 0.0062, "N": 0.0002, "P": 0.0011, "R": 0.0014, "S": 0.0007, "T": 0.9828 },
    "86": { "A": 0.004, "D": 0.9658, "E": 0.0146, "F": 0.0001, "G": 0.0013, "H": 0.003, "K": 0.0001, "N": 0.0021, "P": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0029, "Y": 0.0058 },
    "87": { "C": 0.0063, "D": 0.0001, "F": 0.9738, "G": 0.0003, "I": 0.0015, "L": 0.0036, "S": 0.0018, "T": 0.0002, "V": 0.011, "Y": 0.0014 },
    "88": { "A": 0.0048, "D": 0.0001, "F": 0.0003, "H": 0.0002, "I": 0.0023, "K": 0.0001, "L": 0.0002, "N": 0.0038, "P": 0.0044, "S": 0.0023, "T": 0.9814 },
    "89": { "F": 0.001, "H": 0.0016, "I": 0.0032, "L": 0.9857, "N": 0.0001, "P": 0.0028, "R": 0.005, "S": 0.0001, "T": 0.0001, "V": 0.0004 },
    "90": { "A": 0.0025, "C": 0.0001, "D": 0.0001, "I": 0.0033, "L": 0.0002, "N": 0.0046, "P": 0.0049, "R": 0.0001, "S": 0.0033, "T": 0.9809 },
    "91": { "C": 0.0001, "D": 0.0001, "F": 0.0005, "I": 0.9834, "L": 0.0014, "M": 0.0005, "N": 0.0007, "S": 0.0088, "T": 0.0027, "V": 0.0017 },
    "92": { "D": 0.0135, "E": 0.0004, "H": 0.0067, "I": 0.0012, "K": 0.0212, "N": 0.9374, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0137, "T": 0.0044, "Y": 0.001 },
    "93": { "A": 0.0004, "C": 0.0031, "D": 0.0002, "F": 0.0001, "G": 0.0256, "I": 0.0056, "K": 0.0001, "L": 0.0003, "M": 0.0004, "N": 0.0009, "P": 0.0001, "R": 0.0178, "S": 0.9411, "T": 0.0039, "V": 0.0003 },
    "94": { "A": 0.0001, "L": 0.9835, "M": 0.0019, "P": 0.0023, "Q": 0.0018, "R": 0.0083, "V": 0.002 },
    "95": { "A": 0.0018, "D": 0.0031, "E": 0.9881, "G": 0.0023, "I": 0.0002, "K": 0.0017, "N": 0.0002, "Q": 0.0011, "V": 0.0015 },
    "96": { "A": 0.9656, "D": 0.0034, "E": 0.0001, "F": 0.0001, "G": 0.0044, "L": 0.0002, "P": 0.0111, "S": 0.0023, "T": 0.0092, "V": 0.0036 },
    "97": { "A": 0.0015, "D": 0.0042, "E": 0.9903, "G": 0.0013, "K": 0.0012, "P": 0.0002, "Q": 0.0009, "V": 0.0005 },
    "98": { "A": 0.0014, "D": 0.9868, "E": 0.0047, "G": 0.0012, "H": 0.0007, "N": 0.001, "Q": 0.0001, "S": 0.0001, "V": 0.0009, "Y": 0.0032 },
    "99": { "A": 0.9847, "D": 0.0019, "G": 0.0011, "P": 0.0005, "S": 0.0054, "T": 0.0015, "V": 0.0049 },
    "100": { "A": 0.9914, "E": 0.0021, "G": 0.0007, "P": 0.0005, "S": 0.0017, "T": 0.0018, "V": 0.0017 },
    "101": { "A": 1.0 },
    "102": { "C": 0.001, "D": 0.0032, "F": 0.0015, "H": 0.0005, "N": 0.0005, "S": 0.0013, "Y": 0.992 },
    "103": { "C": 0.0004, "D": 0.0022, "E": 0.0001, "F": 0.0113, "H": 0.0004, "N": 0.0001, "S": 0.0013, "T": 0.0001, "V": 0.0001, "Y": 0.9839 },
    "104": { "C": 0.9832, "F": 0.0022, "G": 0.0037, "R": 0.0014, "S": 0.0016, "V": 0.0001, "W": 0.006, "Y": 0.0018 },
    "105": { "D": 0.0001, "E": 0.0001, "H": 0.9199, "I": 0.0001, "K": 0.0001, "L": 0.0094, "N": 0.0017, "P": 0.0003, "Q": 0.0542, "R": 0.0025, "S": 0.0003, "Y": 0.0113 },
    "106": { "D": 0.002, "E": 0.0007, "G": 0.0003, "H": 0.0061, "K": 0.0013, "L": 0.0012, "P": 0.0018, "Q": 0.983, "R": 0.0031, "S": 0.0002, "V": 0.0003 },
    "107": { "A": 0.0001, "C": 0.0037, "F": 0.0003, "G": 0.0042, "I": 0.0005, "N": 0.0028, "Q": 0.0004, "R": 0.0154, "S": 0.9512, "T": 0.0203, "V": 0.0005, "W": 0.0001, "Y": 0.0004 },
    "108": { "A": 0.0001, "C": 0.0024, "D": 0.0023, "E": 0.0005, "G": 0.004, "H": 0.0001, "I": 0.0014, "K": 0.0017, "L": 0.0003, "M": 0.0002, "N": 0.0161, "R": 0.0127, "S": 0.9476, "T": 0.0084, "V": 0.0001, "W": 0.0003, "Y": 0.0019 },
    "109": { "A": 0.0039, "C": 0.001, "D": 0.0007, "G": 0.0083, "I": 0.0026, "K": 0.0001, "L": 0.0003, "N": 0.0153, "R": 0.0075, "S": 0.951, "T": 0.0091, "V": 0.0001 },
    "113": { "F": 0.109, "I": 0.0082, "L": 0.7221, "P": 0.1362, "Q": 0.0027, "V": 0.0054, "Y": 0.0163 },
    "114": { "A": 0.0003, "C": 0.0011, "E": 0.0001, "F": 0.0785, "G": 0.0004, "H": 0.0035, "I": 0.0047, "K": 0.0005, "L": 0.8084, "N": 0.0003, "P": 0.0278, "Q": 0.0053, "R": 0.0022, "S": 0.0496, "T": 0.0042, "V": 0.0081, "W": 0.0004, "Y": 0.0046 },
    "115": { "A": 0.0028, "C": 0.0001, "D": 0.0001, "E": 0.002, "F": 0.0058, "G": 0.0071, "H": 0.0104, "I": 0.0004, "K": 0.0009, "L": 0.0434, "M": 0.0012, "P": 0.8685, "Q": 0.0057, "R": 0.0134, "S": 0.0247, "T": 0.0103, "V": 0.0012, "W": 0.0007, "Y": 0.0014 },
    "116": { "A": 0.0029, "C": 0.0074, "D": 0.0049, "E": 0.0121, "F": 0.0726, "G": 0.0385, "H": 0.0645, "I": 0.0636, "K": 0.0079, "L": 0.186, "M": 0.0017, "N": 0.0109, "P": 0.0566, "Q": 0.0632, "R": 0.0741, "S": 0.0131, "T": 0.0021, "V": 0.0183, "W": 0.1733, "Y": 0.1265 },
    "117": { "A": 0.0083, "D": 0.0001, "F": 0.0007, "G": 0.0046, "H": 0.0013, "I": 0.0006, "K": 0.0004, "L": 0.002, "M": 0.0006, "N": 0.0004, "P": 0.0046, "Q": 0.0001, "R": 0.0003, "S": 0.0292, "T": 0.9459, "V": 0.0003, "Y": 0.0004 },
    "118": { "A": 0.0001, "C": 0.0047, "F": 0.985, "G": 0.0001, "I": 0.0005, "L": 0.0036, "P": 0.0009, "Q": 0.0001, "R": 0.0006, "S": 0.0011, "V": 0.0028, "Y": 0.0005 },
    "119": { "A": 0.0002, "C": 0.0002, "D": 0.0009, "G": 0.9961, "L": 0.0001, "P": 0.0002, "Q": 0.0002, "R": 0.0001, "S": 0.0011, "V": 0.001 },
    "120": { "A": 0.0006, "E": 0.0019, "G": 0.2397, "H": 0.0065, "K": 0.001, "L": 0.0003, "P": 0.0671, "Q": 0.6775, "R": 0.0045, "S": 0.0002, "T": 0.0004, "V": 0.0001, "Y": 0.0001 },
    "121": { "A": 0.0002, "E": 0.002, "G": 0.9924, "R": 0.0019, "V": 0.0007, "W": 0.0027 },
    "122": { "A": 0.0013, "I": 0.0019, "K": 0.0003, "N": 0.0009, "P": 0.0007, "S": 0.0017, "T": 0.9932 },
    "123": { "E": 0.0007, "K": 0.8763, "L": 0.001, "N": 0.0006, "P": 0.0003, "Q": 0.0028, "R": 0.1174, "T": 0.0007, "W": 0.0001 },
    "124": { "A": 0.0013, "E": 0.0003, "G": 0.0007, "I": 0.0001, "L": 0.3477, "M": 0.0046, "P": 0.0002, "Q": 0.0003, "R": 0.0002, "V": 0.6445 },
    "125": { "A": 0.0003, "D": 0.0703, "E": 0.9239, "G": 0.0025, "K": 0.0012, "N": 0.0001, "Q": 0.001, "R": 0.0002, "V": 0.0004, "W": 0.0001 },
    "126": { "F": 0.0026, "I": 0.9904, "L": 0.0015, "M": 0.0009, "N": 0.0001, "S": 0.0017, "T": 0.0015, "V": 0.0014 },
    "127": { "E": 0.0005, "K": 0.9978, "Q": 0.0017 }
  },
  "IGLV1": {
    "1": { "H": 0.0086, "Q": 0.9914 },
    "2": { "A": 0.0057, "C": 0.0002, "F": 0.0022, "H": 0.0002, "I": 0.0001, "L": 0.0005, "P": 0.0154, "S": 0.9724, "T": 0.0013, "V": 0.0001, "Y": 0.0018 },
    "3": { "A": 0.0087, "D": 0.0001, "E": 0.001, "F": 0.0007, "G": 0.0049, "I": 0.002, "K": 0.0001, "L": 0.0128, "M": 0.003, "P": 0.0003, "Q": 0.0001, "R": 0.0002, "S": 0.0001, "T": 0.0005, "V": 0.9654 },
    "4": { "A": 0.0001, "C": 0.0001, "F": 0.0034, "H": 0.0001, "L": 0.9794, "M": 0.0023, "P": 0.0045, "Q": 0.0006, "R": 0.0026, "S": 0.0009, "T": 0.0001, "V": 0.0055, "W": 0.0004 },
    "5": { "A": 0.0048, "F": 0.0001, "H": 0.0003, "I": 0.0016, "K": 0.0008, "L": 0.0003, "M": 0.0007, "N": 0.0016, "P": 0.0339, "Q": 0.0001, "R": 0.0003, "S": 0.0036, "T": 0.9517, "V": 0.0001 },
    "6": { "D": 0.0001, "E": 0.0003, "H": 0.015, "K": 0.0024, "L": 0.0014, "P": 0.0393, "Q": 0.9392, "R": 0.0016, "S": 0.0002, "T": 0.0004, "Y": 0.0001 },
    "7": { "A": 0.0035, "E": 0.0002, "H": 0.0005, "K": 0.0001, "L": 0.0021, "P": 0.9767, "Q": 0.0048, "R": 0.0011, "S": 0.0074, "T": 0.0034, "V": 0.0002 },
    "8": { "A": 0.0019, "D": 0.0001, "H": 0.0078, "I": 0.0002, "L": 0.0038, "N": 0.0001, "P": 0.9777, "Q": 0.0001, "R": 0.0013, "S": 0.0023, "T": 0.0047, "V": 0.0001 },
    "9": { "A": 0.0068, "C": 0.0001, "D": 0.0001, "E": 0.0001, "F": 0.0005, "H": 0.0001, "I": 0.0001, "L": 0.0014, "P": 0.0195, "Q": 0.0005, "R": 0.0001, "S": 0.9686, "T": 0.0016, "V": 0.0001, "Y": 0.0005 },
    "11": { "A": 0.4177, "C": 0.0001, "D": 0.0001, "E": 0.0037, "F": 0.0003, "G": 0.0055, "H": 0.0001, "I": 0.0009, "L": 0.0099, "M": 0.0041, "P": 0.0062, "Q": 0.0003, "R": 0.0001, "S": 0.0032, "T": 0.0053, "V": 0.5424, "W": 0.0002 },
    "12": { "A": 0.0066, "C": 0.0004, "D": 0.0002, "E": 0.0001, "F": 0.0022, "G": 0.0001, "H": 0.0002, "I": 0.0001, "L": 0.0001, "P": 0.0105, "Q": 0.0001, "S": 0.9736, "T": 0.0016, "V": 0.0001, "Y": 0.0041 },
    "13": { "A": 0.1276, "C": 0.0002, "D": 0.0017, "E": 0.0322, "G": 0.8093, "K": 0.0002, "L": 0.0007, "N": 0.0001, "P": 0.0014, "R": 0.0099, "S": 0.0009, "T": 0.0003, "V": 0.0103, "W": 0.0053 },
    "14": { "A": 0.5436, "C": 0.0001, "D": 0.0029, "G": 0.0014, "H": 0.0006, "I": 0.0011, "K": 0.0001, "L": 0.0002, "N": 0.0026, "P": 0.0314, "Q": 0.0001, "R": 0.0001, "S": 0.007, "T": 0.4068, "V": 0.002, "Y": 0.0001 },
    "15": { "A": 0.0028, "G": 0.0001, "H": 0.0041, "L": 0.0029, "N": 0.0001, "P": 0.976, "Q": 0.002, "R": 0.0008, "S": 0.0034, "T": 0.0073, "V": 0.0006 },
    "16": { "A": 0.0089, "C": 0.0005, "D": 0.0001, "E": 0.0014, "F": 0.0001, "G": 0.9418, "K": 0.0001, "L": 0.0003, "M": 0.0001, "N": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0356, "S": 0.0004, "T": 0.0005, "V": 0.0068, "W": 0.0031 },
    "17": { "A": 0.0001, "E": 0.001, "G": 0.0001, "H": 0.0134, "K": 0.0045, "L": 0.0026, "M": 0.0001, "N": 0.0001, "P": 0.0465, "Q": 0.9248, "R": 0.0051, "S": 0.0002, "T": 0.0014 },
    "18": { "A": 0.0002, "C": 0.0001, "D": 0.0001, "E": 0.0014, "G": 0.0212, "H": 0.0001, "I": 0.0005, "K": 0.1007, "L": 0.0009, "M": 0.0049, "N": 0.0019, "P": 0.0014, "Q": 0.0026, "R": 0.8217, "S": 0.0169, "T": 0.0205, "V": 0.0002, "W": 0.0047 },
    "19": { "A": 0.0091, "C": 0.0003, "D": 0.0015, "F": 0.0051, "G": 0.0095, "H": 0.0001, "I": 0.0085, "L": 0.0054, "P": 0.0003, "S": 0.0003, "T": 0.0003, "V": 0.9594, "Y": 0.0001 },
    "20": { "A": 0.0057, "D": 0.0001, "F": 0.0002, "G": 0.0001, "H": 0.0006, "I": 0.0076, "K": 0.0001, "L": 0.0006, "M": 0.0001, "N": 0.0043, "P": 0.0287, "R": 0.0002, "S": 0.0094, "T": 0.9415, "V": 0.0008, "Y": 0.0001 },
    "21": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.003, "H": 0.0002, "I": 0.9475, "L": 0.0226, "M": 0.0028, "N": 0.0027, "P": 0.0003, "R": 0.0004, "S": 0.0062, "T": 0.0081, "V": 0.0059 },
    "22": { "A": 0.0149, "C": 0.0003, "D": 0.0001, "F": 0.0025, "G": 0.0001, "H": 0.0009, "I": 0.0001, "L": 0.0006, "N": 0.0001, "P": 0.0623, "R": 0.0001, "S": 0.9083, "T": 0.0068, "Y": 0.003 },
    "23": { "A": 0.0001, "C": 0.9726, "F": 0.0057, "G": 0.0061, "I": 0.0001, "L": 0.0003, "N": 0.0003, "P": 0.0001, "R": 0.0051, "S": 0.0057, "T": 0.0001, "V": 0.0003, "W": 0.0023, "Y": 0.0013 },
    "24": { "A": 0.0063, "C": 0.0001, "D": 0.0005, "F": 0.0017, "G": 0.0001, "H": 0.0004, "I": 0.0014, "K": 0.0001, "L": 0.0007, "M": 0.0001, "N": 0.0011, "P": 0.0217, "S": 0.5639, "T": 0.4001, "V": 0.0001, "Y": 0.0018 },
    "25": { "A": 0.0041, "C": 0.0008, "D": 0.0001, "E": 0.0022, "F": 0.0001, "G": 0.9784, "I": 0.0001, "K": 0.0001, "L": 0.0002, "R": 0.0049, "S": 0.0001, "T": 0.0001, "V": 0.0051, "W": 0.0037 },
    "26": { "A": 0.0016, "C": 0.0032, "D": 0.0021, "E": 0.0001, "F": 0.0002, "G": 0.0199, "H": 0.0011, "I": 0.0055, "K": 0.0003, "L": 0.0005, "N": 0.0091, "P": 0.0009, "Q": 0.0001, "R": 0.0375, "S": 0.8984, "T": 0.0187, "V": 0.0002, "W": 0.0001, "Y": 0.0003 },
    "27": { "A": 0.0022, "C": 0.0007, "D": 0.0026, "E": 0.0001, "F": 0.0004, "G": 0.0097, "H": 0.0022, "I": 0.0064, "K": 0.0008, "L": 0.0008, "M": 0.0003, "N": 0.0216, "P": 0.002, "Q": 0.0002, "R": 0.0347, "S": 0.8921, "T": 0.0195, "V": 0.0004, "W": 0.0006, "Y": 0.0026 },
    "28": { "A": 0.0118, "C": 0.0009, "D": 0.0001, "F": 0.0045, "G": 0.0003, "H": 0.0006, "I": 0.0003, "L": 0.0005, "N": 0.0004, "P": 0.0448, "R": 0.0003, "S": 0.9238, "T": 0.007, "V": 0.0001, "W": 0.0001, "Y": 0.0046 },
    "29": { "A": 0.0005, "D": 0.0066, "G": 0.0005, "H": 0.021, "I": 0.0012, "K": 0.0044, "L": 0.0001, "N": 0.9375, "P": 0.0003, "Q": 0.0006, "R": 0.0003, "S": 0.0039, "T": 0.022, "V": 0.0001, "Y": 0.0011 },
    "30": { "A": 0.0005, "C": 0.0001, "F": 0.0051, "G": 0.0003, "H": 0.0007, "I": 0.9136, "L": 0.0318, "M": 0.0028, "N": 0.0051, "P": 0.0009, "R": 0.0009, "S": 0.0116, "T": 0.0158, "V": 0.0106, "Y": 0.0001 },
    "31": { "A": 0.004, "C": 0.0003, "D": 0.0005, "E": 0.0037, "G": 0.9641, "I": 0.0005, "L": 0.0003, "P": 0.0003, "R": 0.0149, "S": 0.0003, "V": 0.0059, "W": 0.0051 },
    "35": { "A": 0.4123, "C": 0.0001, "D": 0.0012, "E": 0.0103, "F": 0.0001, "G": 0.5542, "H": 0.0002, "I": 0.0003, "K": 0.0002, "L": 0.0001, "M": 0.0001, "P": 0.0018, "Q": 0.0003, "R": 0.0056, "S": 0.0044, "T": 0.0037, "V": 0.0041, "W": 0.001 },
    "36": { "A": 0.006, "C": 0.0021, "D": 0.0092, "E": 0.0009, "F": 0.0009, "G": 0.4032, "H": 0.0046, "I": 0.0068, "K": 0.0057, "L": 0.0009, "M": 0.0004, "N": 0.1288, "P": 0.0025, "Q": 0.0007, "R": 0.0221, "S": 0.3817, "T": 0.0164, "V": 0.005, "W": 0.0001, "Y": 0.0021 },
    "37": { "A": 0.0007, "C": 0.0009, "D": 0.0063, "E": 0.0011, "F": 0.0097, "G": 0.0017, "H": 0.0193, "I": 0.001, "K": 0.0079, "L": 0.001, "N": 0.5205, "P": 0.0008, "Q": 0.0016, "R": 0.0014, "S": 0.0156, "T": 0.0137, "V": 0.0003, "Y": 0.3964 },
    "38": { "A": 0.046, "C": 0.0006, "D": 0.3919, "E": 0.0063, "F": 0.0125, "G": 0.005, "H": 0.0091, "I": 0.0036, "K": 0.0003, "L": 0.0021, "M": 0.0001, "N": 0.0076, "P": 0.0145, "Q": 0.0003, "R": 0.0008, "S": 0.017, "T": 0.2128, "V": 0.0046, "W": 0.0002, "Y": 0.2646 },
    "39": { "A": 0.0068, "C": 0.0003, "D": 0.0003, "E": 0.0026, "F": 0.0014, "G": 0.0058, "H": 0.0001, "I": 0.0069, "L": 0.0193, "P": 0.0003, "R": 0.0002, "S": 0.0002, "T": 0.0003, "V": 0.9554, "Y": 0.0001 },
    "40": { "A": 0.0019, "C": 0.0023, "D": 0.0041, "E": 0.0001, "F": 0.0055, "G": 0.0006, "H": 0.4003, "I": 0.0008, "K": 0.0018, "L": 0.0009, "M": 0.0001, "N": 0.2518, "P": 0.0211, "Q": 0.0034, "R": 0.0014, "S": 0.1175, "T": 0.0155, "V": 0.0002, "W": 0.0001, "Y": 0.1707 },
    "41": { "A": 0.0002, "C": 0.0131, "E": 0.0001, "G": 0.0076, "H": 0.0001, "L": 0.0072, "M": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0078, "S": 0.0022, "V": 0.0005, "W": 0.961, "Y": 0.0001 },
    "42": { "A": 0.0001, "C": 0.0007, "D": 0.0043, "E": 0.0001, "F": 0.0085, "H": 0.0041, "L": 0.0001, "N": 0.0016, "P": 0.0001, "S": 0.0116, "T": 0.0001, "Y": 0.9688 },
    "43": { "A": 0.0001, "E": 0.0003, "H": 0.0148, "I": 0.0001, "K": 0.0064, "L": 0.003, "M": 0.0001, "N": 0.0001, "P": 0.0133, "Q": 0.9509, "R": 0.0101, "S": 0.0001, "T": 0.0005, "V": 0.0002, "W": 0.0001, "Y": 0.0001 },
    "44": { "A": 0.0003, "C": 0.0001, "D": 0.0004, "E": 0.0009, "F": 0.0003, "G": 0.0001, "H": 0.0418, "K": 0.0049, "L": 0.0059, "M": 0.0001, "N": 0.0004, "P": 0.0176, "Q": 0.9199, "R": 0.0042, "S": 0.0001, "T": 0.0005, "V": 0.0017, "W": 0.0001, "Y": 0.0005 },
    "45": { "A": 0.0007, "C": 0.0003, "F": 0.0341, "G": 0.0004, "H": 0.0084, "I": 0.01, "L": 0.8789, "M": 0.0001, "N": 0.0002, "P": 0.0252, "Q": 0.0002, "R": 0.0112, "S": 0.0015, "T": 0.0004, "V": 0.0281, "Y": 0.0003 },
    "46": { "A": 0.002, "C": 0.0001, "E": 0.0001, "H": 0.0007, "K": 0.0001, "L": 0.0014, "P": 0.9834, "Q": 0.0049, "R": 0.0003, "S": 0.0024, "T": 0.0045, "V": 0.0001 },
    "47": { "A": 0.0039, "D": 0.0008, "E": 0.0055, "G": 0.9721, "H": 0.0003, "K": 0.0003, "L": 0.0003, "N": 0.0001, "P": 0.0003, "Q": 0.0003, "R": 0.0076, "S": 0.0001, "T": 0.0001, "V": 0.0078, "W": 0.0001 },
    "48": { "A": 0.0231, "E": 0.0009, "G": 0.0009, "H": 0.0001, "I": 0.0014, "K": 0.0334, "L": 0.0005, "M": 0.0054, "N": 0.0005, "P": 0.0181, "Q": 0.0033, "R": 0.0059, "S": 0.0068, "T": 0.8987, "V": 0.0008, "Y": 0.0001 },
    "49": { "A": 0.9692, "C": 0.0001, "D": 0.0028, "E": 0.0001, "G": 0.0023, "H": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.007, "Q": 0.0001, "R": 0.0001, "S": 0.0072, "T": 0.0069, "V": 0.0039 },
    "50": { "A": 0.0012, "H": 0.0051, "L": 0.0009, "N": 0.0001, "P": 0.986, "R": 0.0005, "S": 0.0014, "T": 0.0047 },
    "51": { "D": 0.0004, "E": 0.003, "G": 0.0001, "H": 0.0014, "I": 0.0015, "K": 0.9108, "L": 0.0005, "M": 0.0001, "N": 0.0349, "P": 0.0003, "Q": 0.013, "R": 0.0173, "S": 0.001, "T": 0.0156 },
    "52": { "A": 0.0004, "C": 0.0003, "E": 0.0001, "F": 0.0034, "G": 0.0001, "H": 0.0053, "I": 0.005, "L": 0.9349, "M": 0.0001, "N": 0.0001, "P": 0.0206, "Q": 0.0001, "R": 0.0117, "S": 0.0006, "T": 0.0009, "V": 0.0163 },
    "53": { "A": 0.0001, "C": 0.0001, "F": 0.0028, "G": 0.0001, "H": 0.0028, "I": 0.0076, "L": 0.95, "M": 0.0003, "P": 0.0149, "R": 0.0114, "S": 0.0006, "T": 0.0003, "V": 0.0089 },
    "54": { "A": 0.0001, "C": 0.0003, "D": 0.0001, "F": 0.0025, "H": 0.0003, "I": 0.9364, "K": 0.0001, "L": 0.0191, "M": 0.0096, "N": 0.0032, "P": 0.0003, "Q": 0.0001, "R": 0.0012, "S": 0.0119, "T": 0.0074, "V": 0.0074, "Y": 0.0001 },
    "55": { "A": 0.0002, "C": 0.0025, "D": 0.0104, "E": 0.0005, "F": 0.0239, "G": 0.0002, "H": 0.0155, "I": 0.0003, "K": 0.0003, "L": 0.0013, "M": 0.0003, "N": 0.0039, "P": 0.0007, "Q": 0.0013, "R": 0.0005, "S": 0.018, "T": 0.0005, "V": 0.001, "W": 0.0003, "Y": 0.9183 },
    "56": { "A": 0.0134, "C": 0.0038, "D": 0.0935, "E": 0.0276, "F": 0.0009, "G": 0.3858, "H": 0.0017, "I": 0.0031, "K": 0.0065, "L": 0.0015, "M": 0.0021, "N": 0.0133, "P": 0.0005, "Q": 0.0007, "R": 0.1656, "S": 0.2295, "T": 0.0125, "V": 0.0082, "W": 0.0011, "Y": 0.0287 },
    "57": { "A": 0.0009, "C": 0.0001, "D": 0.0526, "E": 0.0007, "F": 0.0007, "G": 0.001, "H": 0.0076, "I": 0.002, "K": 0.0106, "L": 0.0003, "M": 0.0001, "N": 0.8854, "P": 0.0004, "Q": 0.0001, "R": 0.0008, "S": 0.0126, "T": 0.0199, "V": 0.0012, "W": 0.0001, "Y": 0.003 },
    "65": { "A": 0.0022, "C": 0.0008, "D": 0.0579, "E": 0.0017, "F": 0.0013, "G": 0.0058, "H": 0.0082, "I": 0.0088, "K": 0.0164, "L": 0.0003, "N": 0.495, "P": 0.0009, "Q": 0.0008, "R": 0.0139, "S": 0.3443, "T": 0.0336, "V": 0.0023, "Y": 0.0057 },
    "66": { "A": 0.0006, "C": 0.0001, "D": 0.0057, "E": 0.008, "F": 0.0002, "G": 0.0009, "H": 0.016, "I": 0.0034, "K": 0.1091, "L": 0.0308, "M": 0.001, "N": 0.3825, "P": 0.0138, "Q": 0.3935, "R": 0.0165, "S": 0.0042, "T": 0.0097, "V": 0.0009, "Y": 0.003 },
    "67": { "A": 0.0002, "D": 0.0001, "F": 0.0001, "G": 0.0032, "H": 0.0001, "K": 0.0001, "L": 0.0312, "M": 0.0001, "P": 0.0048, "Q": 0.0028, "R": 0.9547, "S": 0.0001, "T": 0.0001, "V": 0.0006, "W": 0.0018 },
    "68": { "A": 0.0053, "F": 0.0006, "H": 0.0055, "L": 0.0033, "M": 0.0001, "P": 0.9739, "Q": 0.0003, "R": 0.0023, "S": 0.0044, "T": 0.0041, "V": 0.0001, "Y": 0.0001 },
    "69": { "A": 0.0228, "C": 0.0003, "E": 0.0007, "F": 0.0003, "L": 0.0044, "P": 0.0148, "Q": 0.0009, "R": 0.0001, "S": 0.9524, "T": 0.0026, "V": 0.0004, "W": 0.0001, "Y": 0.0002 },
    "70": { "A": 0.0024, "C": 0.0003, "D": 0.0006, "E": 0.0024, "F": 0.0001, "G": 0.9715, "L": 0.0001, "M": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0053, "S": 0.0002, "T": 0.0001, "V": 0.0102, "W": 0.0066 },
    "71": { "A": 0.0036, "C": 0.0009, "D": 0.0009, "F": 0.0114, "G": 0.0105, "H": 0.0001, "I": 0.1086, "L": 0.0035, "M": 0.0016, "N": 0.0003, "P": 0.0001, "R": 0.0002, "S": 0.0016, "T": 0.0012, "V": 0.8556 },
    "72": { "A": 0.0032, "F": 0.0001, "H": 0.0015, "L": 0.0025, "P": 0.954, "Q": 0.0001, "R": 0.0015, "S": 0.0339, "T": 0.0032, "Y": 0.0001 },
    "74": { "A": 0.02, "C": 0.0005, "D": 0.9437, "E": 0.0055, "F": 0.0001, "G": 0.012, "H": 0.002, "K": 0.0001, "N": 0.0026, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0011, "T": 0.0002, "V": 0.005, "Y": 0.0065 },
    "75": { "G": 0.0009, "H": 0.0001, "L": 0.0064, "P": 0.0012, "Q": 0.0013, "R": 0.9899, "T": 0.0001, "V": 0.0001 },
    "76": { "C": 0.0118, "D": 0.0001, "F": 0.9629, "G": 0.0001, "H": 0.0001, "I": 0.0045, "L": 0.0064, "N": 0.0001, "R": 0.0001, "S": 0.0026, "V": 0.0103, "W": 0.0001, "Y": 0.001 },
    "77": { "A": 0.0143, "C": 0.0005, "D": 0.0001, "F": 0.0026, "H": 0.0001, "L": 0.0003, "P": 0.0041, "R": 0.0003, "S": 0.9736, "T": 0.0023, "Y": 0.0018 },
    "78": { "A": 0.0249, "C": 0.0082, "D": 0.0019, "E": 0.0001, "F": 0.0003, "G": 0.955, "H": 0.0001, "N": 0.0001, "R": 0.0009, "S": 0.0021, "T": 0.0001, "V": 0.0063, "Y": 0.0001 },
    "79": { "A": 0.0131, "C": 0.001, "F": 0.0024, "H": 0.0001, "N": 0.0001, "P": 0.0028, "S": 0.9755, "T": 0.0023, "V": 0.0002, "Y": 0.0025 },
    "80": { "A": 0.0001, "D": 0.0002, "E": 0.0027, "G": 0.0001, "H": 0.0001, "I": 0.0003, "K": 0.9531, "L": 0.0001, "M": 0.0006, "N": 0.0074, "P": 0.0002, "Q": 0.0066, "R": 0.0186, "S": 0.0008, "T": 0.0086, "V": 0.0004, "Y": 0.0001 },
    "83": { "A": 0.0111, "C": 0.0006, "D": 0.0001, "F": 0.0014, "G": 0.0002, "L": 0.0001, "P": 0.0016, "S": 0.9809, "T": 0.0015, "V": 0.0002, "Y": 0.0024 },
    "84": { "A": 0.0058, "C": 0.0045, "D": 0.0088, "E": 0.0001, "F": 0.0001, "G": 0.9734, "I": 0.0001, "K": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.0007, "S": 0.0019, "T": 0.0001, "V": 0.0045 },
    "85": { "A": 0.0126, "C": 0.0001, "D": 0.0002, "F": 0.0002, "G": 0.0001, "H": 0.0001, "I": 0.0019, "K": 0.0002, "L": 0.0002, "M": 0.0007, "N": 0.0043, "P": 0.0076, "R": 0.0003, "S": 0.0085, "T": 0.9631, "V": 0.0001 },
    "86": { "A": 0.0111, "C": 0.0003, "F": 0.0001, "G": 0.0001, "K": 0.0001, "L": 0.0014, "N": 0.0001, "P": 0.0015, "Q": 0.0001, "S": 0.9799, "T": 0.0045, "V": 0.0001, "W": 0.0001, "Y": 0.0007 },
    "87": { "A": 0.9835, "D": 0.0025, "E": 0.0003, "G": 0.0041, "P": 0.0009, "S": 0.0041, "T": 0.0014, "V": 0.0032 },
    "88": { "A": 0.0122, "C": 0.0005, "F": 0.002, "G": 0.0001, "I": 0.0004, "K": 0.0001, "L": 0.0003, "N": 0.0006, "P": 0.0026, "S": 0.8676, "T": 0.1108, "V": 0.0001, "Y": 0.0028 },
    "89": { "A": 0.0001, "F": 0.0001, "I": 0.0001, "L": 0.9849, "M": 0.003, "P": 0.0033, "Q": 0.0003, "R": 0.0068, "S": 0.0001, "T": 0.0002, "V": 0.0011, "W": 0.0001 },
    "90": { "A": 0.8634, "C": 0.0009, "D": 0.0106, "E": 0.0007, "F": 0.0001, "G": 0.0953, "H": 0.0001, "I": 0.0007, "L": 0.0001, "P": 0.0007, "Q": 0.0001, "R": 0.0003, "S": 0.0066, "T": 0.0097, "V": 0.0107, "Y": 0.0001 },
    "91": { "F": 0.0009, "G": 0.0001, "I": 0.9786, "L": 0.0026, "M": 0.0007, "N": 0.0006, "P": 0.0001, "R": 0.0003, "S": 0.0101, "T": 0.0022, "V": 0.0037 },
    "92": { "A": 0.0064, "C": 0.0005, "D": 0.0007, "F": 0.0003, "G": 0.0032, "H": 0.0001, "I": 0.003, "K": 0.0003, "L": 0.0003, "N": 0.005, "P": 0.0032, "R": 0.0064, "S": 0.4599, "T": 0.5104, "V": 0.0002, "W": 0.0001, "Y": 0.0001 },
    "93": { "A": 0.003, "C": 0.0002, "D": 0.0014, "E": 0.002, "G": 0.9788, "N": 0.0001, "R": 0.0033, "S": 0.0003, "V": 0.0068, "W": 0.0039, "Y": 0.0001 },
    "94": { "A": 0.0003, "C": 0.0001, "F": 0.0006, "H": 0.0005, "I": 0.0029, "L": 0.9801, "P": 0.0025, "R": 0.0096, "S": 0.0001, "T": 0.0001, "V": 0.0032 },
    "95": { "E": 0.0011, "G": 0.0001, "H": 0.0039, "K": 0.0014, "L": 0.0026, "M": 0.0001, "P": 0.0016, "Q": 0.7951, "R": 0.1932, "S": 0.0001, "T": 0.0001, "V": 0.0001, "W": 0.0006, "Y": 0.0001 },
    "96": { "A": 0.4163, "C": 0.0002, "D": 0.0008, "E": 0.0001, "F": 0.0011, "G": 0.0013, "H": 0.0001, "I": 0.0015, "L": 0.0003, "M": 0.0001, "N": 0.0004, "P": 0.0114, "Q": 0.0001, "R": 0.0002, "S": 0.4539, "T": 0.1086, "V": 0.0025, "Y": 0.0011 },
    "97": { "A": 0.0038, "D": 0.0197, "E": 0.8615, "G": 0.1122, "K": 0.0005, "L": 0.0001, "Q": 0.0003, "S": 0.0003, "T": 0.0001, "V": 0.0011, "W": 0.0004, "Y": 0.0001 },
    "98": { "A": 0.0015, "D": 0.9851, "E": 0.0085, "G": 0.0008, "H": 0.0003, "N": 0.0007, "S": 0.0001, "V": 0.0009, "Y": 0.002 },
    "99": { "A": 0.003, "D": 0.005, "E": 0.9869, "G": 0.0028, "K": 0.0007, "L": 0.0001, "Q": 0.0004, "S": 0.0001, "V": 0.001, "Y": 0.0001 },
    "100": { "A": 0.9782, "C": 0.0001, "D": 0.002, "G": 0.0107, "P": 0.0001, "S": 0.0046, "T": 0.003, "V": 0.0013 },
    "101": { "A": 0.0034, "C": 0.0001, "D": 0.9518, "E": 0.0142, "F": 0.0002, "G": 0.0023, "H": 0.0059, "I": 0.0011, "K": 0.0001, "L": 0.0009, "N": 0.003, "Q": 0.0001, "R": 0.0002, "S": 0.0009, "T": 0.0026, "V": 0.0078, "Y": 0.0055 },
    "102": { "A": 0.0001, "C": 0.0005, "D": 0.0071, "E": 0.0003, "F": 0.002, "H": 0.001, "I": 0.0001, "L": 0.0002, "N": 0.0001, "S": 0.0002, "V": 0.0003, "Y": 0.9881 },
    "103": { "C": 0.0028, "D": 0.0059, "E": 0.0001, "F": 0.0155, "H": 0.006, "I": 0.0003, "L": 0.0001, "N": 0.0003, "Q": 0.0001, "S": 0.0036, "T": 0.0001, "V": 0.0003, "Y": 0.965 },
    "104": { "C": 0.9858, "D": 0.0001, "F": 0.0013, "G": 0.0063, "R": 0.0007, "S": 0.001, "V": 0.0002, "W": 0.0036, "Y": 0.001 },
    "105": { "A": 0.4539, "C": 0.0011, "D": 0.0001, "E": 0.0034, "F": 0.0001, "G": 0.1041, "H": 0.0024, "K": 0.0004, "L": 0.0013, "N": 0.0001, "P": 0.0016, "Q": 0.4156, "R": 0.0006, "S": 0.0087, "T": 0.0018, "V": 0.0045, "Y": 0.0004 },
    "106": { "A": 0.4249, "C": 0.0014, "E": 0.0011, "F": 0.0002, "G": 0.0025, "I": 0.0009, "K": 0.0002, "L": 0.0006, "M": 0.0001, "N": 0.0003, "P": 0.0014, "Q": 0.0001, "R": 0.0001, "S": 0.4286, "T": 0.128, "V": 0.0078, "Y": 0.0018 },
    "107": { "A": 0.0005, "C": 0.0054, "D": 0.002, "E": 0.0002, "F": 0.0083, "G": 0.0041, "H": 0.0022, "I": 0.0001, "L": 0.0017, "N": 0.0003, "Q": 0.0002, "R": 0.0018, "S": 0.0032, "T": 0.0001, "V": 0.0001, "W": 0.5632, "Y": 0.4065 },
    "108": { "A": 0.0048, "C": 0.0001, "D": 0.9765, "E": 0.0071, "F": 0.0001, "G": 0.0025, "H": 0.0014, "K": 0.0001, "L": 0.0001, "N": 0.0014, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0001, "T": 0.0014, "V": 0.0018, "Y": 0.0025 },
    "109": { "A": 0.006, "C": 0.0001, "D": 0.4465, "E": 0.0026, "F": 0.0005, "G": 0.0128, "H": 0.0009, "I": 0.0043, "K": 0.0012, "L": 0.0003, "M": 0.0001, "N": 0.0204, "P": 0.0005, "Q": 0.0001, "R": 0.0092, "S": 0.4759, "T": 0.0137, "V": 0.002, "W": 0.0001, "Y": 0.0028 },
    "110": { "A": 0.0018, "C": 0.0008, "D": 0.0015, "F": 0.0001, "G": 0.0125, "H": 0.0003, "I": 0.0041, "K": 0.0004, "L": 0.0004, "M": 0.0002, "N": 0.01, "P": 0.0002, "Q": 0.0001, "R": 0.0231, "S": 0.9291, "T": 0.0136, "V": 0.001, "W": 0.0001, "Y": 0.0007 },
    "111": { "D": 0.0015, "F": 0.0015, "L": 0.9602, "N": 0.0015, "P": 0.0074, "Q": 0.0059, "R": 0.0088, "S": 0.0015, "V": 0.0103, "Y": 0.0015 },
    "112": { "A": 0.0016, "C": 0.0005, "D": 0.0002, "E": 0.0002, "F": 0.0002, "G": 0.0089, "I": 0.0016, "K": 0.0002, "L": 0.8092, "M": 0.0072, "N": 0.0308, "P": 0.0049, "Q": 0.0028, "R": 0.014, "S": 0.1115, "T": 0.0012, "V": 0.0044, "W": 0.0002, "Y": 0.0002 },
    "113": { "A": 0.0055, "C": 0.0003, "D": 0.0017, "E": 0.0002, "F": 0.001, "G": 0.0414, "H": 0.0009, "I": 0.001, "K": 0.0003, "L": 0.6059, "M": 0.0026, "N": 0.0537, "P": 0.0023, "Q": 0.0005, "R": 0.011, "S": 0.2601, "T": 0.0056, "V": 0.0057, "W": 0.0001, "Y": 0.0004 },
    "114": { "A": 0.0362, "C": 0.0012, "D": 0.0066, "E": 0.0014, "F": 0.0017, "G": 0.2035, "H": 0.0028, "I": 0.0027, "K": 0.005, "L": 0.0731, "M": 0.0016, "N": 0.1914, "P": 0.0208, "Q": 0.0014, "R": 0.0204, "S": 0.4006, "T": 0.0145, "V": 0.0122, "W": 0.0015, "Y": 0.0015 },
    "115": { "A": 0.1013, "C": 0.0054, "D": 0.0187, "E": 0.0067, "F": 0.0233, "G": 0.4623, "H": 0.0432, "I": 0.0037, "K": 0.0038, "L": 0.0402, "M": 0.0005, "N": 0.0298, "P": 0.0328, "Q": 0.0064, "R": 0.0285, "S": 0.1203, "T": 0.0089, "V": 0.0316, "W": 0.0035, "Y": 0.0292 },
    "116": { "A": 0.0217, "C": 0.0032, "D": 0.0028, "E": 0.0145, "F": 0.0042, "G": 0.0773, "H": 0.0062, "I": 0.0011, "K": 0.0094, "L": 0.0276, "M": 0.003, "N": 0.0029, "P": 0.0654, "Q": 0.0054, "R": 0.0548, "S": 0.06, "T": 0.0024, "V": 0.2974, "W": 0.2058, "Y": 0.1348 },
    "117": { "A": 0.0039, "C": 0.0001, "D": 0.0003, "E": 0.0021, "F": 0.0025, "G": 0.0042, "H": 0.0001, "I": 0.0131, "K": 0.0002, "L": 0.0383, "M": 0.0022, "P": 0.0007, "Q": 0.0005, "R": 0.0014, "S": 0.0012, "T": 0.0009, "V": 0.9281, "W": 0.0002 },
    "118": { "C": 0.0026, "D": 0.0001, "F": 0.9848, "I": 0.0015, "L": 0.0045, "M": 0.0001, "R": 0.0001, "S": 0.0006, "V": 0.005, "W": 0.0001, "Y": 0.0007 },
    "119": { "A": 0.0011, "C": 0.0003, "D": 0.001, "E": 0.0001, "F": 0.0001, "G": 0.9941, "L": 0.0001, "M": 0.0001, "R": 0.0009, "S": 0.0009, "V": 0.0013 },
    "120": { "A": 0.0028, "C": 0.0001, "E": 0.0017, "G": 0.8043, "I": 0.0005, "L": 0.0001, "M": 0.0001, "N": 0.0001, "P": 0.0035, "R": 0.001, "S": 0.0081, "T": 0.1771, "V": 0.0004 },
    "121": { "A": 0.0008, "E": 0.0005, "G": 0.9945, "K": 0.0001, "L": 0.0001, "M": 0.0001, "P": 0.0001, "R": 0.0011, "V": 0.0011, "W": 0.0016 },
    "122": { "A": 0.0012, "C": 0.0001, "I": 0.0006, "N": 0.0007, "P": 0.0019, "S": 0.0014, "T": 0.9942 },
    "123": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "E": 0.0037, "G": 0.0003, "H": 0.0009, "I": 0.0001, "K": 0.9525, "L": 0.0001, "M": 0.0008, "N": 0.0038, "P": 0.0001, "Q": 0.0112, "R": 0.0167, "S": 0.0011, "T": 0.008, "W": 0.0003, "Y": 0.0002 },
    "124": { "A": 0.0002, "F": 0.0003, "G": 0.0003, "I": 0.0022, "L": 0.7792, "M": 0.0009, "P": 0.0005, "Q": 0.0001, "R": 0.0014, "S": 0.0001, "V": 0.2149 },
    "125": { "A": 0.0033, "C": 0.0001, "D": 0.0002, "F": 0.0001, "H": 0.0001, "I": 0.0025, "K": 0.0001, "L": 0.0001, "N": 0.0015, "P": 0.0022, "Q": 0.0001, "S": 0.0052, "T": 0.9845, "V": 0.0002 },
    "126": { "A": 0.0036, "D": 0.0004, "F": 0.0005, "G": 0.0048, "I": 0.001, "L": 0.0007, "R": 0.0003, "V": 0.9887 },
    "127": { "I": 0.0016, "L": 0.9957, "P": 0.0003, "V": 0.0024 }
  },
  "IGLV10": {
    "1": { "H": 0.0084, "Q": 0.9916 },
    "2": { "A": 0.9755, "E": 0.0014, "G": 0.0007, "P": 0.0093, "S": 0.0077, "T": 0.0039, "V": 0.0014 },
    "3": { "A": 0.0084, "C": 0.0003, "D": 0.0001, "E": 0.0037, "F": 0.0001, "G": 0.9656, "L": 0.0004, "Q": 0.0001, "R": 0.0084, "S": 0.0003, "T": 0.0001, "V": 0.0072, "W": 0.0054 },
    "4": { "F": 0.0001, "L": 0.9884, "M": 0.0018, "P": 0.0056, "Q": 0.0004, "R": 0.0031, "T": 0.0001, "V": 0.0003, "W": 0.0001 },
    "5": { "A": 0.0025, "H": 0.0001, "I": 0.0009, "L": 0.0001, "M": 0.0001, "N": 0.001, "P": 0.0301, "R": 0.0001, "S": 0.0055, "T": 0.9594 },
    "6": { "E": 0.0001, "H": 0.0105, "K": 0.0022, "L": 0.0005, "P": 0.0136, "Q": 0.9723, "R": 0.0007, "T": 0.0001 },
    "7": { "A": 0.0007, "H": 0.0004, "L": 0.0021, "P": 0.9861, "Q": 0.0043, "R": 0.001, "S": 0.0026, "T": 0.0027 },
    "8": { "A": 0.0008, "H": 0.0071, "L": 0.0025, "P": 0.9814, "R": 0.001, "S": 0.0024, "T": 0.0046, "V": 0.0003 },
    "9": { "A": 0.0052, "C": 0.0001, "L": 0.0007, "P": 0.0211, "Q": 0.0005, "S": 0.9685, "T": 0.0027, "W": 0.0005, "Y": 0.0007 },
    "11": { "A": 0.0038, "E": 0.0013, "F": 0.0003, "G": 0.0048, "I": 0.0001, "L": 0.0178, "M": 0.0058, "Q": 0.0003, "R": 0.0003, "S": 0.0004, "T": 0.0003, "V": 0.9648, "W": 0.0001 },
    "12": { "A": 0.0062, "C": 0.0012, "D": 0.0001, "F": 0.0013, "H": 0.0004, "L": 0.0001, "P": 0.011, "S": 0.9721, "T": 0.0021, "V": 0.0001, "Y": 0.0054 },
    "13": { "A": 0.0013, "D": 0.0004, "E": 0.0097, "G": 0.0004, "H": 0.0003, "I": 0.0003, "K": 0.8783, "L": 0.0003, "M": 0.0031, "N": 0.0086, "P": 0.0014, "Q": 0.0266, "R": 0.0101, "S": 0.0008, "T": 0.0577, "V": 0.0007, "W": 0.0001 },
    "14": { "A": 0.0174, "C": 0.0063, "D": 0.0272, "E": 0.0004, "G": 0.9306, "H": 0.0004, "N": 0.0001, "P": 0.0003, "R": 0.0055, "S": 0.0033, "V": 0.0084, "Y": 0.0001 },
    "15": { "A": 0.0003, "C": 0.0004, "E": 0.0003, "F": 0.0073, "G": 0.0003, "L": 0.9567, "M": 0.0029, "P": 0.0004, "Q": 0.0001, "R": 0.0003, "S": 0.0103, "V": 0.0086, "W": 0.0122 },
    "16": { "A": 0.0001, "D": 0.0003, "E": 0.0001, "G": 0.027, "I": 0.0027, "K": 0.0035, "L": 0.0005, "N": 0.0004, "P": 0.0003, "Q": 0.0004, "R": 0.9274, "S": 0.034, "T": 0.0031, "V": 0.0001 },
    "17": { "E": 0.0024, "H": 0.0075, "K": 0.0065, "L": 0.0004, "P": 0.0207, "Q": 0.9585, "R": 0.0035, "S": 0.0001, "T": 0.0004 },
    "18": { "A": 0.0016, "D": 0.0001, "F": 0.0001, "H": 0.0012, "I": 0.001, "L": 0.0001, "N": 0.0038, "P": 0.0301, "R": 0.0001, "S": 0.003, "T": 0.9588 },
    "19": { "A": 0.9847, "D": 0.0048, "G": 0.0001, "I": 0.0003, "P": 0.0041, "S": 0.0018, "T": 0.0026, "V": 0.0016 },
    "20": { "A": 0.0018, "I": 0.0012, "K": 0.0046, "N": 0.0005, "P": 0.0152, "Q": 0.0004, "R": 0.0007, "S": 0.0018, "T": 0.9738 },
    "21": { "F": 0.0018, "H": 0.0046, "I": 0.0077, "L": 0.9369, "N": 0.0001, "P": 0.0355, "Q": 0.0001, "R": 0.0098, "S": 0.0005, "T": 0.0009, "V": 0.002 },
    "22": { "A": 0.0021, "D": 0.0001, "H": 0.0007, "I": 0.0009, "K": 0.0001, "N": 0.0024, "P": 0.0291, "S": 0.0039, "T": 0.9607 },
    "23": { "C": 0.9767, "F": 0.0042, "G": 0.0046, "R": 0.0048, "S": 0.0059, "T": 0.0003, "V": 0.0001, "W": 0.0003, "Y": 0.0031 },
    "24": { "A": 0.0041, "D": 0.0003, "H": 0.0005, "I": 0.0018, "K": 0.0003, "N": 0.0016, "P": 0.0284, "R": 0.0008, "S": 0.0052, "T": 0.9563, "V": 0.0007, "W": 0.0001 },
    "25": { "A": 0.0045, "C": 0.0005, "E": 0.0014, "G": 0.9745, "L": 0.0001, "N": 0.0003, "Q": 0.0001, "R": 0.0063, "V": 0.0038, "W": 0.0085 },
    "26": { "A": 0.0005, "C": 0.0001, "D": 0.0139, "E": 0.0001, "F": 0.0008, "G": 0.0009, "H": 0.023, "I": 0.0001, "K": 0.0051, "N": 0.9353, "Q": 0.0001, "S": 0.006, "T": 0.0118, "V": 0.0004, "Y": 0.0017 },
    "27": { "A": 0.0001, "C": 0.0003, "D": 0.0025, "G": 0.0021, "H": 0.0003, "I": 0.0052, "K": 0.0017, "L": 0.0003, "N": 0.014, "P": 0.0003, "Q": 0.0001, "R": 0.0224, "S": 0.9365, "T": 0.0137, "V": 0.0003, "Y": 0.0003 },
    "28": { "A": 0.0001, "D": 0.0113, "E": 0.0001, "F": 0.0001, "H": 0.0207, "I": 0.001, "K": 0.0084, "N": 0.9352, "Q": 0.0001, "R": 0.0004, "S": 0.0071, "T": 0.0119, "Y": 0.0035 },
    "29": { "A": 0.0001, "D": 0.0031, "F": 0.0004, "H": 0.0204, "I": 0.0795, "K": 0.0156, "L": 0.0022, "M": 0.0008, "N": 0.8556, "Q": 0.0005, "S": 0.0016, "T": 0.0187, "V": 0.0004, "Y": 0.001 },
    "30": { "A": 0.0029, "D": 0.0017, "F": 0.0041, "G": 0.0065, "I": 0.0086, "L": 0.0016, "R": 0.0004, "S": 0.0003, "T": 0.0001, "V": 0.9738 },
    "35": { "A": 0.0103, "C": 0.0102, "D": 0.0077, "G": 0.957, "H": 0.0003, "P": 0.0004, "R": 0.0051, "S": 0.0038, "T": 0.0003, "V": 0.0048 },
    "36": { "A": 0.0008, "C": 0.0001, "D": 0.0192, "E": 0.0001, "F": 0.0027, "G": 0.0009, "H": 0.0164, "I": 0.0008, "K": 0.0085, "L": 0.0007, "N": 0.9031, "P": 0.0004, "Q": 0.0001, "R": 0.0007, "S": 0.0077, "T": 0.0076, "V": 0.0003, "Y": 0.0299 },
    "37": { "A": 0.001, "D": 0.0005, "E": 0.039, "G": 0.0005, "H": 0.0154, "K": 0.0076, "L": 0.0029, "N": 0.0003, "P": 0.014, "Q": 0.9092, "R": 0.008, "S": 0.0005, "T": 0.0007, "V": 0.0001, "Y": 0.0003 },
    "38": { "A": 0.0089, "C": 0.0001, "D": 0.0003, "E": 0.008, "G": 0.9653, "L": 0.0001, "Q": 0.0001, "R": 0.0082, "S": 0.0004, "T": 0.0001, "V": 0.0084 },
    "39": { "A": 0.9626, "D": 0.0005, "E": 0.0056, "G": 0.0009, "H": 0.0005, "K": 0.0001, "L": 0.0001, "P": 0.0107, "Q": 0.0003, "S": 0.0042, "T": 0.0068, "V": 0.0076 },
    "40": { "A": 0.9283, "C": 0.0005, "D": 0.0055, "F": 0.0001, "G": 0.0014, "H": 0.0003, "I": 0.0017, "L": 0.0042, "P": 0.0063, "R": 0.0003, "S": 0.0081, "T": 0.0204, "V": 0.0226, "Y": 0.0003 },
    "41": { "A": 0.0001, "C": 0.0148, "E": 0.0001, "F": 0.0004, "G": 0.005, "L": 0.0054, "P": 0.0001, "R": 0.0079, "S": 0.0056, "V": 0.0001, "W": 0.9601, "Y": 0.0004 },
    "42": { "A": 0.0001, "C": 0.0003, "H": 0.0001, "I": 0.0003, "L": 0.964, "M": 0.0033, "P": 0.0123, "Q": 0.0035, "R": 0.0094, "S": 0.0001, "T": 0.0003, "V": 0.0062, "W": 0.0001 },
    "43": { "A": 0.0001, "C": 0.0001, "E": 0.0007, "H": 0.0058, "K": 0.0037, "L": 0.0026, "P": 0.0096, "Q": 0.97, "R": 0.0071, "T": 0.0004 },
    "44": { "E": 0.0007, "H": 0.0175, "K": 0.0047, "L": 0.0048, "N": 0.0001, "P": 0.0114, "Q": 0.9582, "R": 0.0021, "S": 0.0003, "T": 0.0001 },
    "45": { "A": 0.0001, "C": 0.0001, "D": 0.0021, "F": 0.0041, "H": 0.9338, "K": 0.0005, "L": 0.0021, "N": 0.006, "P": 0.0168, "Q": 0.0096, "R": 0.0143, "S": 0.0001, "T": 0.001, "Y": 0.0094 },
    "46": { "A": 0.0001, "D": 0.0001, "E": 0.0008, "H": 0.0099, "K": 0.0033, "L": 0.001, "P": 0.0216, "Q": 0.9585, "R": 0.0038, "S": 0.0001, "T": 0.0007 },
    "47": { "A": 0.0077, "C": 0.0067, "D": 0.0035, "E": 0.0001, "G": 0.9643, "H": 0.0001, "I": 0.0001, "R": 0.0075, "S": 0.0047, "T": 0.0001, "V": 0.005, "Y": 0.0001 },
    "48": { "D": 0.0012, "F": 0.0001, "G": 0.0001, "H": 0.9558, "L": 0.0025, "N": 0.0065, "P": 0.0118, "Q": 0.011, "R": 0.0081, "S": 0.0003, "T": 0.0013, "Y": 0.0013 },
    "49": { "A": 0.001, "F": 0.0001, "H": 0.0052, "L": 0.0005, "P": 0.9856, "Q": 0.0004, "R": 0.0012, "S": 0.0013, "T": 0.0046 },
    "50": { "A": 0.0005, "H": 0.0018, "K": 0.0001, "L": 0.0005, "P": 0.9918, "R": 0.0003, "S": 0.0012, "T": 0.0037, "Y": 0.0001 },
    "51": { "A": 0.0001, "E": 0.0021, "H": 0.0007, "I": 0.002, "K": 0.9334, "N": 0.0279, "P": 0.0001, "Q": 0.0127, "R": 0.0067, "T": 0.0141, "Y": 0.0003 },
    "52": { "A": 0.0003, "C": 0.0001, "F": 0.0012, "G": 0.0003, "H": 0.0077, "I": 0.0029, "L": 0.939, "N": 0.0001, "P": 0.0247, "Q": 0.0001, "R": 0.0161, "S": 0.0004, "T": 0.0001, "V": 0.0069 },
    "53": { "A": 0.0001, "G": 0.0001, "H": 0.0004, "I": 0.0043, "K": 0.0001, "L": 0.9686, "M": 0.0001, "P": 0.0058, "Q": 0.0029, "R": 0.0092, "T": 0.0001, "V": 0.0082 },
    "54": { "A": 0.0157, "C": 0.0009, "D": 0.0001, "F": 0.0308, "H": 0.0005, "I": 0.0003, "L": 0.0013, "M": 0.0001, "N": 0.0001, "P": 0.0106, "R": 0.0001, "S": 0.9144, "T": 0.0181, "V": 0.002, "Y": 0.005 },
    "55": { "A": 0.0001, "C": 0.0027, "D": 0.0191, "F": 0.0038, "H": 0.013, "I": 0.0001, "L": 0.0001, "N": 0.0076, "P": 0.0017, "Q": 0.0005, "R": 0.0001, "S": 0.0198, "T": 0.0005, "V": 0.0009, "Y": 0.9298 },
    "56": { "C": 0.0001, "E": 0.0004, "G": 0.0031, "I": 0.0008, "K": 0.0106, "M": 0.0045, "P": 0.0001, "Q": 0.0001, "R": 0.9581, "S": 0.0186, "T": 0.0021, "W": 0.0014 },
    "57": { "A": 0.0001, "D": 0.0132, "E": 0.0007, "F": 0.0001, "G": 0.0008, "H": 0.0035, "I": 0.0025, "K": 0.0092, "N": 0.9457, "P": 0.0001, "Q": 0.0009, "R": 0.0004, "S": 0.0154, "T": 0.0063, "V": 0.0001, "Y": 0.0009 },
    "65": { "D": 0.0063, "G": 0.0017, "H": 0.0093, "I": 0.0004, "K": 0.0073, "L": 0.0007, "N": 0.9597, "R": 0.0003, "S": 0.0025, "T": 0.0113, "V": 0.0001, "Y": 0.0005 },
    "66": { "A": 0.0007, "D": 0.0156, "F": 0.0003, "G": 0.0003, "H": 0.0171, "I": 0.0037, "K": 0.0081, "L": 0.0003, "M": 0.0001, "N": 0.9196, "P": 0.0005, "Q": 0.0004, "R": 0.0016, "S": 0.0102, "T": 0.0202, "V": 0.0003, "Y": 0.0012 },
    "67": { "C": 0.0001, "G": 0.0004, "K": 0.0001, "L": 0.0042, "P": 0.0068, "Q": 0.0045, "R": 0.9827, "S": 0.0001, "T": 0.0003, "W": 0.0008 },
    "68": { "A": 0.0048, "H": 0.0059, "L": 0.001, "N": 0.0001, "P": 0.9798, "R": 0.0017, "S": 0.002, "T": 0.0046 },
    "69": { "A": 0.0173, "D": 0.0004, "F": 0.0001, "H": 0.0001, "L": 0.001, "P": 0.0255, "Q": 0.0012, "R": 0.0004, "S": 0.9479, "T": 0.0047, "Y": 0.0013 },
    "70": { "A": 0.0043, "C": 0.0001, "D": 0.0029, "E": 0.0026, "G": 0.9736, "R": 0.0041, "S": 0.0005, "T": 0.0001, "V": 0.0072, "W": 0.0043, "Y": 0.0003 },
    "71": { "A": 0.0001, "C": 0.0003, "D": 0.0001, "F": 0.0027, "G": 0.0001, "I": 0.9562, "K": 0.0001, "L": 0.0082, "M": 0.0031, "N": 0.0042, "P": 0.0004, "R": 0.0005, "S": 0.0088, "T": 0.0052, "V": 0.0098 },
    "72": { "A": 0.0234, "E": 0.0004, "F": 0.0001, "I": 0.0003, "L": 0.0008, "P": 0.0126, "Q": 0.0003, "R": 0.0003, "S": 0.9589, "T": 0.0022, "V": 0.0004, "Y": 0.0004 },
    "74": { "A": 0.0092, "C": 0.0001, "D": 0.0137, "E": 0.9525, "G": 0.0134, "H": 0.0004, "K": 0.0022, "L": 0.0001, "M": 0.0004, "N": 0.0001, "Q": 0.0027, "R": 0.0003, "S": 0.0003, "V": 0.0042, "W": 0.0001, "Y": 0.0003 },
    "75": { "A": 0.0001, "E": 0.0004, "G": 0.0225, "I": 0.005, "K": 0.0047, "L": 0.0009, "P": 0.0001, "Q": 0.0001, "R": 0.9605, "S": 0.0034, "T": 0.0014, "V": 0.0008 },
    "76": { "C": 0.0008, "F": 0.0822, "I": 0.0004, "L": 0.9154, "V": 0.001, "Y": 0.0001 },
    "77": { "A": 0.0127, "C": 0.0009, "D": 0.0003, "F": 0.0051, "L": 0.0004, "P": 0.0031, "S": 0.9675, "T": 0.0034, "V": 0.0003, "Y": 0.0063 },
    "78": { "A": 0.9758, "D": 0.0003, "E": 0.0018, "G": 0.0026, "L": 0.0034, "P": 0.0014, "R": 0.0003, "S": 0.0071, "T": 0.005, "V": 0.0024 },
    "79": { "A": 0.0186, "C": 0.0009, "D": 0.0001, "F": 0.0034, "G": 0.0001, "I": 0.0001, "L": 0.0001, "P": 0.0016, "S": 0.966, "T": 0.0033, "V": 0.0003, "Y": 0.0055 },
    "80": { "E": 0.0003, "G": 0.0027, "K": 0.0033, "L": 0.0004, "M": 0.0043, "P": 0.0001, "Q": 0.0003, "R": 0.9795, "S": 0.005, "T": 0.0034, "V": 0.0003, "W": 0.0005 },
    "83": { "A": 0.0132, "C": 0.0001, "E": 0.0003, "L": 0.0004, "P": 0.0016, "S": 0.9809, "T": 0.0034, "V": 0.0001 },
    "84": { "A": 0.0013, "D": 0.0004, "E": 0.0046, "G": 0.9831, "R": 0.0052, "V": 0.0054 },
    "85": { "D": 0.0109, "E": 0.0004, "F": 0.0003, "G": 0.0001, "H": 0.0026, "I": 0.0008, "K": 0.0042, "L": 0.0001, "N": 0.9618, "P": 0.0003, "S": 0.0118, "T": 0.0065, "Y": 0.0003 },
    "86": { "A": 0.0052, "H": 0.0001, "I": 0.0064, "K": 0.0026, "N": 0.0012, "P": 0.0098, "Q": 0.0004, "S": 0.0021, "T": 0.9719, "V": 0.0003 },
    "87": { "A": 0.9796, "C": 0.0001, "D": 0.0043, "E": 0.0005, "G": 0.0017, "I": 0.0001, "R": 0.0001, "S": 0.0059, "T": 0.0047, "V": 0.0027, "W": 0.0001 },
    "88": { "A": 0.0198, "F": 0.0025, "G": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.0001, "P": 0.0035, "Q": 0.0001, "S": 0.9657, "T": 0.0052, "V": 0.0003, "Y": 0.0024 },
    "89": { "F": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.9844, "M": 0.0017, "P": 0.0045, "Q": 0.0005, "R": 0.0069, "T": 0.0001, "V": 0.0014 },
    "90": { "A": 0.0029, "G": 0.0001, "H": 0.0003, "I": 0.0009, "K": 0.0014, "L": 0.0003, "N": 0.0037, "P": 0.0034, "Q": 0.0004, "R": 0.0003, "S": 0.0016, "T": 0.9848 },
    "91": { "C": 0.0001, "F": 0.0035, "I": 0.9755, "L": 0.0014, "M": 0.0062, "N": 0.0004, "R": 0.0004, "S": 0.0073, "T": 0.0027, "V": 0.0024 },
    "92": { "A": 0.0094, "D": 0.0005, "F": 0.0001, "G": 0.0001, "I": 0.0026, "K": 0.0004, "L": 0.0001, "M": 0.0001, "N": 0.0022, "P": 0.0063, "R": 0.0003, "S": 0.0113, "T": 0.9662, "V": 0.0001, "W": 0.0001 },
    "93": { "A": 0.002, "C": 0.0004, "D": 0.0012, "E": 0.005, "G": 0.9785, "R": 0.0069, "S": 0.0008, "V": 0.0052 },
    "94": { "A": 0.0003, "F": 0.0007, "G": 0.0001, "H": 0.0012, "I": 0.0024, "L": 0.9725, "M": 0.0001, "P": 0.0021, "Q": 0.0003, "R": 0.0181, "S": 0.0003, "V": 0.0021 },
    "95": { "D": 0.0001, "E": 0.0007, "H": 0.0041, "K": 0.0025, "L": 0.0017, "P": 0.0038, "Q": 0.9793, "R": 0.0076, "S": 0.0003 },
    "96": { "A": 0.005, "C": 0.0001, "D": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0024, "L": 0.0008, "N": 0.0003, "P": 0.9831, "Q": 0.0001, "R": 0.0008, "S": 0.0029, "T": 0.0041, "V": 0.0001 },
    "97": { "A": 0.0034, "D": 0.0093, "E": 0.9817, "G": 0.0031, "K": 0.0004, "Q": 0.0004, "R": 0.0003, "V": 0.0009, "W": 0.0003, "Y": 0.0003 },
    "98": { "A": 0.0014, "D": 0.9825, "E": 0.0099, "F": 0.0001, "G": 0.0014, "H": 0.0001, "N": 0.0007, "T": 0.0003, "V": 0.001, "Y": 0.0025 },
    "99": { "A": 0.0033, "D": 0.0065, "E": 0.9813, "G": 0.005, "K": 0.0008, "M": 0.0001, "Q": 0.0004, "R": 0.0003, "V": 0.0024 },
    "100": { "A": 0.9763, "D": 0.0025, "E": 0.0004, "G": 0.0109, "L": 0.0003, "P": 0.0005, "S": 0.0052, "T": 0.0026, "V": 0.0012, "Y": 0.0001 },
    "101": { "A": 0.0009, "C": 0.0001, "D": 0.9835, "E": 0.0041, "G": 0.001, "H": 0.0003, "N": 0.0008, "Q": 0.0003, "S": 0.0001, "T": 0.0003, "V": 0.0017, "Y": 0.0069 },
    "102": { "A": 0.0003, "C": 0.0008, "D": 0.0106, "E": 0.0005, "F": 0.0008, "H": 0.0009, "I": 0.0003, "L": 0.0001, "N": 0.0005, "S": 0.0012, "T": 0.0001, "Y": 0.9839 },
    "103": { "A": 0.0001, "C": 0.0016, "D": 0.0098, "E": 0.0001, "F": 0.0041, "H": 0.0012, "L": 0.0004, "N": 0.0004, "S": 0.0013, "T": 0.0007, "V": 0.0003, "Y": 0.9801 },
    "104": { "A": 0.0007, "C": 0.9859, "F": 0.0021, "G": 0.0065, "R": 0.0008, "S": 0.0009, "W": 0.0012, "Y": 0.002 },
    "105": { "A": 0.0223, "C": 0.0004, "F": 0.0004, "G": 0.0001, "K": 0.0001, "L": 0.0033, "P": 0.0027, "Q": 0.0007, "S": 0.9651, "T": 0.0046, "V": 0.0001, "Y": 0.0003 },
    "106": { "A": 0.9712, "D": 0.0003, "E": 0.003, "G": 0.0013, "H": 0.0007, "I": 0.0001, "P": 0.002, "Q": 0.0003, "S": 0.0107, "T": 0.0088, "V": 0.0017 },
    "107": { "C": 0.005, "E": 0.0005, "F": 0.0007, "G": 0.0073, "L": 0.0821, "Q": 0.0005, "R": 0.0013, "S": 0.002, "T": 0.0001, "V": 0.002, "W": 0.8981, "Y": 0.0004 },
    "108": { "A": 0.0035, "D": 0.9819, "E": 0.0028, "G": 0.0031, "H": 0.0004, "N": 0.0024, "P": 0.0001, "S": 0.0003, "T": 0.0013, "V": 0.0003, "Y": 0.0039 },
    "109": { "A": 0.0011, "C": 0.0011, "D": 0.0037, "E": 0.0003, "G": 0.0041, "H": 0.0001, "I": 0.0101, "K": 0.0011, "L": 0.0001, "M": 0.0001, "N": 0.0053, "P": 0.0001, "R": 0.0169, "S": 0.9382, "T": 0.0163, "V": 0.0005, "Y": 0.0011 },
    "110": { "A": 0.0012, "C": 0.0017, "D": 0.0019, "F": 0.0003, "G": 0.0142, "I": 0.0041, "K": 0.0001, "L": 0.0003, "N": 0.0039, "P": 0.0001, "R": 0.0143, "S": 0.9536, "T": 0.0043 },
    "112": { "A": 0.0316, "F": 0.0115, "H": 0.0029, "I": 0.0057, "L": 0.8649, "R": 0.0144, "S": 0.0546, "T": 0.0029, "V": 0.0115 },
    "113": { "A": 0.0024, "F": 0.002, "G": 0.0005, "H": 0.0022, "I": 0.0051, "L": 0.9144, "M": 0.0001, "N": 0.0003, "P": 0.0009, "R": 0.0128, "S": 0.0577, "V": 0.0014, "Y": 0.0001 },
    "114": { "A": 0.0378, "C": 0.0007, "D": 0.0026, "E": 0.0016, "F": 0.0001, "G": 0.0101, "H": 0.0026, "I": 0.0034, "K": 0.0069, "L": 0.0156, "M": 0.0009, "N": 0.0283, "P": 0.0016, "Q": 0.0001, "R": 0.0275, "S": 0.8472, "T": 0.0089, "V": 0.0034, "W": 0.0003, "Y": 0.0003 },
    "115": { "A": 0.7767, "C": 0.0016, "D": 0.0063, "E": 0.0068, "F": 0.005, "G": 0.0221, "H": 0.0225, "I": 0.0018, "L": 0.014, "M": 0.0001, "N": 0.0048, "P": 0.009, "Q": 0.0039, "R": 0.0033, "S": 0.0334, "T": 0.0072, "V": 0.0781, "W": 0.0008, "Y": 0.0025 },
    "116": { "A": 0.0077, "C": 0.0025, "D": 0.0012, "E": 0.0069, "F": 0.0004, "G": 0.0171, "H": 0.0185, "I": 0.0003, "K": 0.0025, "L": 0.0258, "M": 0.0007, "N": 0.0025, "P": 0.0123, "Q": 0.0376, "R": 0.0376, "S": 0.0021, "V": 0.2068, "W": 0.5853, "Y": 0.0323 },
    "117": { "A": 0.0022, "D": 0.0018, "E": 0.0021, "F": 0.0007, "G": 0.0042, "I": 0.0016, "L": 0.0174, "M": 0.0009, "P": 0.0029, "Q": 0.0007, "R": 0.0003, "S": 0.001, "V": 0.9643 },
    "118": { "A": 0.0007, "C": 0.0035, "D": 0.0001, "F": 0.9821, "G": 0.0001, "I": 0.0026, "L": 0.005, "N": 0.0001, "P": 0.0003, "R": 0.0001, "S": 0.001, "V": 0.0041, "Y": 0.0003 },
    "119": { "A": 0.0001, "C": 0.001, "D": 0.0007, "G": 0.9948, "R": 0.0001, "S": 0.002, "V": 0.0013 },
    "120": { "A": 0.0014, "E": 0.001, "G": 0.9452, "P": 0.0003, "R": 0.0013, "S": 0.0031, "T": 0.0466, "V": 0.0009, "W": 0.0001 },
    "121": { "A": 0.0008, "D": 0.0004, "E": 0.0007, "G": 0.9949, "L": 0.0001, "R": 0.0005, "V": 0.0012, "W": 0.0014 },
    "122": { "A": 0.0005, "I": 0.0007, "N": 0.0008, "P": 0.0016, "S": 0.0008, "T": 0.9957 },
    "123": { "E": 0.0013, "H": 0.0005, "K": 0.9696, "M": 0.0018, "N": 0.0045, "Q": 0.0038, "R": 0.0134, "S": 0.0005, "T": 0.0046 },
    "124": { "E": 0.0001, "F": 0.0003, "I": 0.0013, "L": 0.9377, "M": 0.0017, "P": 0.0009, "R": 0.0009, "V": 0.0571 },
    "125": { "A": 0.0014, "I": 0.0013, "K": 0.0003, "N": 0.001, "P": 0.0035, "S": 0.0018, "T": 0.9906 },
    "126": { "A": 0.0007, "D": 0.0005, "F": 0.0005, "G": 0.0071, "I": 0.001, "L": 0.0001, "V": 0.9901 },
    "127": { "I": 0.0024, "L": 0.9933, "V": 0.0043 }
  },
  "IGLV2": {
    "1": { "H": 0.0106, "Q": 0.9894 },
    "2": { "A": 0.0054, "C": 0.0005, "D": 0.0001, "F": 0.001, "H": 0.0001, "L": 0.0001, "P": 0.011, "S": 0.9779, "T": 0.0016, "Y": 0.0024 },
    "3": { "A": 0.9771, "C": 0.0001, "D": 0.0043, "G": 0.0007, "I": 0.0001, "L": 0.0002, "P": 0.0059, "S": 0.0038, "T": 0.0023, "V": 0.0056 },
    "4": { "F": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.9863, "M": 0.003, "P": 0.0049, "Q": 0.0004, "R": 0.0039, "T": 0.0001, "V": 0.0009, "W": 0.0002 },
    "5": { "A": 0.0036, "F": 0.0001, "G": 0.0001, "H": 0.0003, "I": 0.0024, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0018, "P": 0.0332, "Q": 0.0001, "S": 0.004, "T": 0.9542 },
    "6": { "A": 0.0001, "D": 0.0001, "E": 0.0005, "H": 0.0162, "K": 0.004, "L": 0.0009, "M": 0.0001, "N": 0.0001, "P": 0.0277, "Q": 0.9484, "R": 0.0017, "T": 0.0003, "Y": 0.0001 },
    "7": { "A": 0.0015, "E": 0.0001, "F": 0.0003, "H": 0.0047, "L": 0.0018, "N": 0.0001, "P": 0.9836, "Q": 0.0001, "R": 0.0005, "S": 0.0032, "T": 0.004, "V": 0.0002 },
    "8": { "A": 0.6975, "C": 0.0003, "D": 0.0037, "G": 0.0007, "H": 0.0019, "I": 0.0001, "L": 0.0014, "P": 0.1474, "R": 0.136, "S": 0.0046, "T": 0.0024, "V": 0.0041 },
    "9": { "A": 0.0099, "C": 0.0005, "D": 0.0003, "E": 0.0001, "F": 0.0018, "G": 0.0001, "H": 0.0005, "L": 0.0005, "P": 0.0327, "Q": 0.0002, "R": 0.0001, "S": 0.9459, "T": 0.0034, "V": 0.0001, "Y": 0.0037 },
    "11": { "A": 0.1381, "D": 0.0003, "E": 0.0056, "F": 0.0006, "G": 0.0059, "I": 0.0003, "L": 0.0085, "M": 0.0028, "P": 0.0014, "Q": 0.0002, "R": 0.0001, "S": 0.0014, "T": 0.0004, "V": 0.8344, "W": 0.0001 },
    "12": { "A": 0.0091, "C": 0.0007, "D": 0.0002, "F": 0.0014, "G": 0.0001, "H": 0.0003, "L": 0.0003, "P": 0.0108, "S": 0.9709, "T": 0.0025, "V": 0.0001, "W": 0.0001, "Y": 0.0034 },
    "13": { "A": 0.0145, "C": 0.0005, "D": 0.0001, "E": 0.0043, "F": 0.0001, "G": 0.9556, "H": 0.0001, "L": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0083, "S": 0.0001, "T": 0.0003, "V": 0.0108, "W": 0.005 },
    "14": { "A": 0.0092, "C": 0.0012, "D": 0.0002, "F": 0.0043, "G": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.0001, "P": 0.0117, "Q": 0.0001, "S": 0.9641, "T": 0.0047, "V": 0.0001, "W": 0.0001, "Y": 0.0039 },
    "15": { "A": 0.0016, "E": 0.0001, "F": 0.0001, "H": 0.0029, "L": 0.0079, "N": 0.0001, "P": 0.9773, "Q": 0.0002, "R": 0.0018, "S": 0.0023, "T": 0.0053, "V": 0.0003, "Y": 0.0001 },
    "16": { "A": 0.0077, "C": 0.0007, "D": 0.0005, "E": 0.0019, "F": 0.0001, "G": 0.9752, "I": 0.0001, "K": 0.0001, "L": 0.0001, "P": 0.0003, "Q": 0.0001, "R": 0.0061, "S": 0.0005, "T": 0.0001, "V": 0.0066, "W": 0.0001 },
    "17": { "A": 0.0003, "D": 0.0001, "E": 0.0018, "G": 0.0002, "H": 0.01, "K": 0.0055, "L": 0.0024, "M": 0.0001, "N": 0.0001, "P": 0.0493, "Q": 0.9231, "R": 0.0061, "S": 0.0001, "T": 0.0011 },
    "18": { "A": 0.0173, "C": 0.0001, "D": 0.0001, "E": 0.0003, "G": 0.0001, "H": 0.0001, "K": 0.0001, "L": 0.0009, "P": 0.0215, "Q": 0.0003, "R": 0.0002, "S": 0.9539, "T": 0.0044, "W": 0.0007, "Y": 0.0002 },
    "19": { "A": 0.0022, "C": 0.0002, "D": 0.0003, "F": 0.0041, "G": 0.0014, "H": 0.0007, "I": 0.6572, "K": 0.0001, "L": 0.0299, "M": 0.0003, "N": 0.0028, "P": 0.0007, "R": 0.0004, "S": 0.0064, "T": 0.009, "V": 0.2843 },
    "20": { "A": 0.0064, "D": 0.0001, "H": 0.0007, "I": 0.0023, "L": 0.0001, "N": 0.0046, "P": 0.0283, "R": 0.0001, "S": 0.0046, "T": 0.9525, "V": 0.0003 },
    "21": { "F": 0.0023, "H": 0.0003, "I": 0.9511, "L": 0.0271, "M": 0.0016, "N": 0.0018, "P": 0.0001, "R": 0.0001, "S": 0.0044, "T": 0.007, "V": 0.0044 },
    "22": { "A": 0.0089, "C": 0.0004, "D": 0.0001, "F": 0.0017, "G": 0.0001, "H": 0.0004, "L": 0.0003, "N": 0.0001, "P": 0.0375, "Q": 0.0001, "S": 0.943, "T": 0.0045, "Y": 0.0029 },
    "23": { "A": 0.0001, "C": 0.9773, "F": 0.005, "G": 0.0045, "L": 0.0003, "P": 0.0001, "R": 0.0053, "S": 0.0045, "T": 0.0001, "V": 0.0003, "W": 0.0009, "Y": 0.0016 },
    "24": { "A": 0.0052, "C": 0.0001, "D": 0.0001, "F": 0.0001, "G": 0.0002, "H": 0.0004, "I": 0.0033, "K": 0.0001, "L": 0.0002, "M": 0.0001, "N": 0.0026, "P": 0.032, "Q": 0.0001, "R": 0.0004, "S": 0.0133, "T": 0.9412, "V": 0.0003, "Y": 0.0003 },
    "25": { "A": 0.0069, "C": 0.0003, "D": 0.0006, "E": 0.0023, "G": 0.9773, "L": 0.0003, "P": 0.0001, "Q": 0.0001, "R": 0.0058, "S": 0.0001, "T": 0.0003, "V": 0.0058, "W": 0.0001 },
    "26": { "A": 0.007, "D": 0.0005, "E": 0.0001, "F": 0.0004, "G": 0.0012, "H": 0.0006, "I": 0.0045, "K": 0.0001, "L": 0.0012, "M": 0.0001, "N": 0.0037, "P": 0.0169, "Q": 0.0001, "R": 0.0006, "S": 0.0205, "T": 0.9411, "V": 0.0014, "Y": 0.0001 },
    "27": { "A": 0.0009, "C": 0.0008, "D": 0.0023, "E": 0.0001, "F": 0.0003, "G": 0.0079, "H": 0.0005, "I": 0.0062, "K": 0.0006, "L": 0.0004, "M": 0.0001, "N": 0.0147, "P": 0.0007, "Q": 0.0002, "R": 0.0223, "S": 0.9254, "T": 0.0154, "V": 0.0003, "W": 0.0001, "Y": 0.0006 },
    "28": { "A": 0.0007, "C": 0.0009, "D": 0.0018, "E": 0.0002, "F": 0.0003, "G": 0.0086, "H": 0.0009, "I": 0.0045, "K": 0.0002, "L": 0.0009, "M": 0.0002, "N": 0.0156, "P": 0.0008, "R": 0.0388, "S": 0.915, "T": 0.0094, "V": 0.0009, "Y": 0.0003 },
    "29": { "A": 0.0195, "C": 0.0001, "D": 0.9435, "E": 0.0048, "F": 0.0003, "G": 0.0028, "H": 0.0064, "I": 0.0001, "K": 0.0001, "L": 0.0001, "N": 0.0057, "P": 0.0004, "Q": 0.0002, "S": 0.0014, "T": 0.0009, "V": 0.0018, "Y": 0.012 },
    "30": { "A": 0.0098, "C": 0.0005, "D": 0.0014, "E": 0.0001, "F": 0.0073, "G": 0.0087, "H": 0.0001, "I": 0.046, "L": 0.0091, "M": 0.0003, "P": 0.0004, "R": 0.0002, "S": 0.001, "T": 0.0005, "V": 0.9143, "Y": 0.0002 },
    "31": { "A": 0.0057, "C": 0.0041, "D": 0.0043, "E": 0.0021, "F": 0.0001, "G": 0.965, "H": 0.0002, "I": 0.0002, "L": 0.0005, "N": 0.0005, "P": 0.0003, "Q": 0.0001, "R": 0.0036, "S": 0.0022, "T": 0.0005, "V": 0.0079, "W": 0.0025, "Y": 0.0003 },
    "35": { "A": 0.0339, "C": 0.0095, "D": 0.0134, "E": 0.0007, "F": 0.0041, "G": 0.6826, "H": 0.0021, "I": 0.0039, "K": 0.0012, "L": 0.0016, "M": 0.0005, "N": 0.0125, "P": 0.0018, "Q": 0.0002, "R": 0.0147, "S": 0.1905, "T": 0.0135, "V": 0.0088, "W": 0.0002, "Y": 0.0042 },
    "36": { "A": 0.0003, "C": 0.002, "D": 0.0085, "E": 0.0003, "F": 0.0115, "G": 0.0005, "H": 0.0094, "I": 0.0003, "K": 0.0003, "L": 0.0013, "M": 0.0001, "N": 0.003, "P": 0.0004, "R": 0.0001, "S": 0.0207, "T": 0.001, "V": 0.0004, "W": 0.0005, "Y": 0.9395 },
    "37": { "A": 0.001, "C": 0.0002, "D": 0.0355, "E": 0.0022, "F": 0.0003, "G": 0.002, "H": 0.0219, "I": 0.0021, "K": 0.0202, "L": 0.0001, "M": 0.0002, "N": 0.8846, "P": 0.0005, "Q": 0.0014, "R": 0.0007, "S": 0.0052, "T": 0.0152, "V": 0.0008, "Y": 0.0058 },
    "38": { "A": 0.0009, "C": 0.0024, "D": 0.0059, "E": 0.0002, "F": 0.0329, "G": 0.0001, "H": 0.0124, "I": 0.0023, "L": 0.1911, "M": 0.0001, "N": 0.0019, "P": 0.0023, "Q": 0.0003, "R": 0.0163, "S": 0.0227, "T": 0.0008, "V": 0.0019, "W": 0.0005, "Y": 0.705 },
    "39": { "A": 0.0083, "C": 0.0006, "D": 0.0013, "E": 0.0001, "F": 0.015, "G": 0.0044, "I": 0.0048, "L": 0.0061, "M": 0.0001, "N": 0.0001, "P": 0.0003, "R": 0.0002, "S": 0.0009, "T": 0.0003, "V": 0.9572, "Y": 0.0004 },
    "40": { "A": 0.0113, "C": 0.002, "D": 0.0002, "F": 0.002, "G": 0.0002, "H": 0.0009, "I": 0.0001, "K": 0.0001, "L": 0.0005, "N": 0.0001, "P": 0.0353, "R": 0.0001, "S": 0.9361, "T": 0.005, "V": 0.0001, "W": 0.0001, "Y": 0.006 },
    "41": { "C": 0.0137, "F": 0.0001, "G": 0.0045, "L": 0.0058, "M": 0.0001, "Q": 0.0001, "R": 0.0041, "S": 0.0018, "V": 0.0001, "W": 0.9697, "Y": 0.0001 },
    "42": { "A": 0.0001, "C": 0.0014, "D": 0.0035, "F": 0.0085, "H": 0.0049, "L": 0.0003, "N": 0.0009, "P": 0.0001, "S": 0.0086, "T": 0.0001, "Y": 0.9714 },
    "43": { "A": 0.0001, "E": 0.0011, "F": 0.0001, "H": 0.0162, "I": 0.0001, "K": 0.0043, "L": 0.0013, "N": 0.0003, "P": 0.0099, "Q": 0.9604, "R": 0.006, "T": 0.0003, "V": 0.0001 },
    "44": { "A": 0.0001, "C": 0.0001, "D": 0.0013, "E": 0.0014, "F": 0.0003, "G": 0.0001, "H": 0.0398, "I": 0.0001, "K": 0.0068, "L": 0.0039, "N": 0.0002, "P": 0.0142, "Q": 0.9261, "R": 0.0041, "T": 0.0004, "V": 0.0007, "W": 0.0001, "Y": 0.0003 },
    "45": { "A": 0.0004, "C": 0.0005, "D": 0.0017, "E": 0.0002, "F": 0.0022, "H": 0.8951, "I": 0.0004, "K": 0.0005, "L": 0.0056, "N": 0.007, "P": 0.0412, "Q": 0.0099, "R": 0.0081, "S": 0.0035, "T": 0.0018, "V": 0.0005, "Y": 0.0214 },
    "46": { "A": 0.0025, "H": 0.0005, "L": 0.0015, "P": 0.9799, "Q": 0.006, "R": 0.0006, "S": 0.0027, "T": 0.0062, "V": 0.0001 },
    "47": { "A": 0.0064, "C": 0.0044, "D": 0.0165, "E": 0.0003, "G": 0.9598, "H": 0.0009, "K": 0.0001, "M": 0.0001, "N": 0.0016, "P": 0.0001, "Q": 0.0001, "R": 0.0009, "S": 0.0026, "T": 0.0004, "V": 0.0056, "W": 0.0001, "Y": 0.0001 },
    "48": { "A": 0.0005, "D": 0.0004, "E": 0.0067, "G": 0.0007, "H": 0.0006, "I": 0.0009, "K": 0.92, "L": 0.0001, "M": 0.0003, "N": 0.0196, "P": 0.0005, "Q": 0.012, "R": 0.0109, "S": 0.0017, "T": 0.0247, "V": 0.0001, "Y": 0.0001 },
    "49": { "A": 0.9566, "C": 0.0001, "D": 0.0032, "F": 0.0001, "G": 0.0027, "H": 0.0001, "I": 0.0005, "L": 0.0002, "N": 0.0004, "P": 0.007, "S": 0.0062, "T": 0.003, "V": 0.02 },
    "50": { "A": 0.0009, "H": 0.0073, "L": 0.0021, "N": 0.0001, "P": 0.9815, "Q": 0.0001, "R": 0.0007, "S": 0.0012, "T": 0.0062 },
    "51": { "A": 0.0001, "D": 0.0001, "E": 0.0037, "F": 0.0001, "G": 0.0001, "H": 0.0018, "I": 0.004, "K": 0.9137, "L": 0.0003, "M": 0.0001, "N": 0.0296, "P": 0.0004, "Q": 0.0178, "R": 0.0133, "S": 0.0003, "T": 0.0145, "W": 0.0001, "Y": 0.0001 },
    "52": { "A": 0.0005, "C": 0.0001, "D": 0.0001, "F": 0.0055, "G": 0.0002, "H": 0.0022, "I": 0.0047, "K": 0.0001, "L": 0.9353, "M": 0.0005, "N": 0.0001, "P": 0.0095, "Q": 0.0001, "R": 0.0138, "S": 0.0004, "T": 0.0007, "V": 0.026 },
    "53": { "F": 0.0007, "G": 0.0003, "H": 0.0003, "I": 0.0846, "K": 0.0014, "L": 0.0424, "M": 0.8465, "N": 0.0002, "P": 0.0003, "Q": 0.0002, "R": 0.009, "S": 0.0012, "T": 0.0022, "V": 0.0106, "W": 0.0002 },
    "54": { "A": 0.0001, "C": 0.0002, "D": 0.0002, "F": 0.0038, "G": 0.0002, "I": 0.9428, "K": 0.0001, "L": 0.0217, "M": 0.0099, "N": 0.0012, "P": 0.0002, "R": 0.001, "S": 0.0088, "T": 0.0027, "V": 0.0071 },
    "55": { "A": 0.0003, "C": 0.0029, "D": 0.0079, "E": 0.0003, "F": 0.0282, "G": 0.0001, "H": 0.0078, "I": 0.0001, "K": 0.0001, "L": 0.0009, "N": 0.0008, "Q": 0.0001, "R": 0.0002, "S": 0.0154, "T": 0.0005, "V": 0.0005, "Y": 0.9339 },
    "56": { "A": 0.0127, "C": 0.0001, "D": 0.4406, "E": 0.5212, "F": 0.0001, "G": 0.0087, "H": 0.0016, "K": 0.0014, "L": 0.0003, "M": 0.0002, "N": 0.003, "P": 0.0002, "Q": 0.0018, "R": 0.0003, "S": 0.0005, "T": 0.0003, "V": 0.0029, "W": 0.0003, "Y": 0.0039 },
    "57": { "A": 0.0049, "C": 0.0014, "D": 0.0061, "F": 0.0072, "G": 0.1109, "H": 0.0001, "I": 0.0044, "K": 0.0001, "L": 0.0023, "M": 0.0001, "N": 0.0004, "P": 0.0001, "R": 0.0003, "S": 0.0005, "T": 0.0001, "V": 0.8606, "Y": 0.0004 },
    "65": { "A": 0.0019, "C": 0.0006, "D": 0.0043, "F": 0.0016, "G": 0.003, "H": 0.0009, "I": 0.0089, "K": 0.0014, "L": 0.0005, "M": 0.0005, "N": 0.0408, "P": 0.0008, "Q": 0.0001, "R": 0.0172, "S": 0.8449, "T": 0.0683, "V": 0.0014, "W": 0.0001, "Y": 0.0027 },
    "66": { "A": 0.0012, "D": 0.0072, "E": 0.01, "F": 0.0004, "G": 0.0012, "H": 0.0096, "I": 0.0028, "K": 0.4582, "L": 0.0006, "M": 0.0007, "N": 0.4503, "P": 0.0005, "Q": 0.0129, "R": 0.0104, "S": 0.0079, "T": 0.0167, "V": 0.0014, "W": 0.0001, "Y": 0.0077 },
    "67": { "A": 0.0001, "G": 0.0014, "L": 0.0043, "M": 0.0001, "N": 0.0001, "P": 0.0039, "Q": 0.0031, "R": 0.9821, "S": 0.0004, "V": 0.0001, "W": 0.0046 },
    "68": { "A": 0.0042, "D": 0.0001, "G": 0.0001, "H": 0.0039, "L": 0.0018, "P": 0.9806, "R": 0.0017, "S": 0.0032, "T": 0.0044, "V": 0.0001, "Y": 0.0001 },
    "69": { "A": 0.0231, "C": 0.0002, "E": 0.0002, "F": 0.0002, "G": 0.0001, "I": 0.0001, "L": 0.0036, "P": 0.0124, "Q": 0.0005, "R": 0.0001, "S": 0.9557, "T": 0.0029, "V": 0.0004, "W": 0.0001, "Y": 0.0006 },
    "70": { "A": 0.0023, "C": 0.0005, "D": 0.0005, "E": 0.0024, "G": 0.9731, "L": 0.0001, "N": 0.0002, "P": 0.0001, "R": 0.0039, "S": 0.0002, "T": 0.0003, "V": 0.011, "W": 0.0051, "Y": 0.0001 },
    "71": { "A": 0.0038, "C": 0.0007, "D": 0.0007, "F": 0.0099, "G": 0.0095, "I": 0.0213, "L": 0.0039, "M": 0.0002, "N": 0.0001, "P": 0.0001, "R": 0.0001, "S": 0.0009, "T": 0.0007, "V": 0.948 },
    "72": { "A": 0.0129, "C": 0.0011, "D": 0.0003, "F": 0.0016, "G": 0.0001, "H": 0.0007, "L": 0.001, "N": 0.0001, "P": 0.2914, "R": 0.0004, "S": 0.6867, "T": 0.002, "V": 0.0001, "Y": 0.0015 },
    "74": { "A": 0.0064, "C": 0.0003, "D": 0.2936, "E": 0.0053, "F": 0.0008, "G": 0.005, "H": 0.0141, "I": 0.0035, "K": 0.0108, "L": 0.0006, "M": 0.0002, "N": 0.6119, "P": 0.0015, "Q": 0.0003, "R": 0.0016, "S": 0.0141, "T": 0.0178, "V": 0.0038, "Y": 0.0083 },
    "75": { "A": 0.0001, "C": 0.0014, "G": 0.0013, "H": 0.002, "I": 0.0001, "L": 0.0083, "N": 0.0001, "P": 0.002, "R": 0.9822, "S": 0.0024, "T": 0.0001, "W": 0.0001 },
    "76": { "A": 0.0001, "C": 0.0107, "F": 0.9648, "G": 0.0003, "I": 0.0012, "L": 0.0068, "M": 0.0001, "R": 0.0001, "S": 0.003, "V": 0.0119, "Y": 0.001 },
    "77": { "A": 0.0164, "C": 0.0009, "F": 0.0035, "G": 0.0003, "H": 0.0001, "L": 0.0001, "P": 0.0043, "S": 0.9704, "T": 0.0027, "V": 0.0001, "W": 0.0001, "Y": 0.0012 },
    "78": { "A": 0.0154, "C": 0.0082, "D": 0.0014, "E": 0.0002, "F": 0.0002, "G": 0.9668, "L": 0.0001, "R": 0.0009, "S": 0.0016, "V": 0.0051, "W": 0.0002, "Y": 0.0001 },
    "79": { "A": 0.0135, "C": 0.0018, "D": 0.0003, "F": 0.0014, "G": 0.0001, "H": 0.0001, "L": 0.0001, "P": 0.0029, "R": 0.0001, "S": 0.9752, "T": 0.0018, "V": 0.0001, "Y": 0.0026 },
    "80": { "A": 0.0003, "E": 0.0023, "G": 0.0002, "H": 0.0005, "I": 0.0001, "K": 0.9668, "M": 0.0012, "N": 0.0053, "P": 0.0002, "Q": 0.0047, "R": 0.0101, "S": 0.0005, "T": 0.0075, "Y": 0.0001 },
    "83": { "A": 0.0142, "C": 0.0007, "D": 0.0001, "F": 0.0017, "H": 0.0001, "L": 0.0002, "N": 0.0001, "P": 0.0021, "R": 0.0001, "S": 0.9756, "T": 0.0022, "V": 0.0001, "Y": 0.0028 },
    "84": { "A": 0.0102, "C": 0.0055, "D": 0.01, "F": 0.0001, "G": 0.9641, "H": 0.0001, "R": 0.0016, "S": 0.0026, "T": 0.0001, "V": 0.0053, "W": 0.0001, "Y": 0.0003 },
    "85": { "A": 0.0009, "D": 0.0086, "E": 0.0003, "G": 0.0002, "H": 0.0036, "I": 0.0019, "K": 0.0043, "L": 0.0001, "M": 0.0001, "N": 0.956, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0087, "T": 0.012, "V": 0.0001, "W": 0.0001, "Y": 0.0026 },
    "86": { "A": 0.0077, "E": 0.0002, "I": 0.0001, "K": 0.002, "L": 0.0001, "M": 0.0041, "N": 0.0005, "P": 0.0075, "R": 0.001, "S": 0.0058, "T": 0.9704, "V": 0.0005, "W": 0.0001 },
    "87": { "A": 0.9877, "D": 0.0026, "E": 0.0002, "F": 0.0001, "G": 0.0009, "L": 0.0001, "P": 0.0007, "S": 0.0043, "T": 0.002, "V": 0.0014 },
    "88": { "A": 0.0104, "C": 0.0008, "D": 0.0001, "F": 0.0045, "G": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.0003, "N": 0.0001, "P": 0.0016, "Q": 0.0001, "S": 0.9748, "T": 0.0038, "V": 0.0003, "W": 0.0001, "Y": 0.003 },
    "89": { "F": 0.0001, "I": 0.0002, "L": 0.9867, "M": 0.0024, "P": 0.0028, "Q": 0.0005, "R": 0.0061, "S": 0.0001, "T": 0.0004, "V": 0.0007, "W": 0.0001 },
    "90": { "A": 0.0045, "D": 0.0003, "F": 0.0003, "G": 0.0002, "H": 0.0001, "I": 0.0033, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0017, "P": 0.0033, "R": 0.0009, "S": 0.004, "T": 0.9804, "V": 0.0004 },
    "91": { "A": 0.0004, "C": 0.0001, "F": 0.0014, "G": 0.0009, "I": 0.8503, "L": 0.0034, "M": 0.0006, "N": 0.0005, "S": 0.0114, "T": 0.0028, "V": 0.1281 },
    "92": { "A": 0.0158, "C": 0.0012, "D": 0.0001, "F": 0.0016, "H": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0025, "R": 0.0001, "S": 0.973, "T": 0.0039, "V": 0.0002, "W": 0.0001, "Y": 0.0012 },
    "93": { "A": 0.0028, "C": 0.0001, "D": 0.0012, "E": 0.0035, "G": 0.9727, "H": 0.0001, "K": 0.0001, "N": 0.0003, "Q": 0.0001, "R": 0.003, "S": 0.0005, "T": 0.0001, "V": 0.0102, "W": 0.0053, "Y": 0.0001 },
    "94": { "C": 0.0001, "F": 0.001, "G": 0.0001, "H": 0.0012, "I": 0.0026, "L": 0.9768, "P": 0.0027, "Q": 0.0002, "R": 0.0114, "S": 0.0003, "T": 0.0001, "V": 0.0035 },
    "95": { "D": 0.0001, "E": 0.0019, "H": 0.0054, "K": 0.0021, "L": 0.0049, "P": 0.0017, "Q": 0.9758, "R": 0.008, "V": 0.0001, "Y": 0.0001 },
    "96": { "A": 0.9411, "D": 0.0049, "E": 0.0009, "F": 0.0009, "G": 0.004, "I": 0.0004, "L": 0.0003, "N": 0.0002, "P": 0.0165, "R": 0.0002, "S": 0.0099, "T": 0.0164, "V": 0.0041, "Y": 0.0001 },
    "97": { "A": 0.0031, "D": 0.0255, "E": 0.9639, "G": 0.0049, "K": 0.0006, "M": 0.0001, "N": 0.0001, "Q": 0.0003, "R": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.001, "Y": 0.0002 },
    "98": { "A": 0.0014, "D": 0.9868, "E": 0.0056, "G": 0.0007, "H": 0.0007, "I": 0.0001, "K": 0.0001, "N": 0.0007, "T": 0.0001, "V": 0.0006, "Y": 0.0032 },
    "99": { "A": 0.0029, "D": 0.012, "E": 0.9799, "G": 0.0024, "K": 0.0005, "M": 0.0001, "N": 0.0001, "Q": 0.0004, "R": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.0012, "Y": 0.0001 },
    "100": { "A": 0.9727, "C": 0.0003, "D": 0.0038, "G": 0.0101, "L": 0.0001, "P": 0.0004, "S": 0.0091, "T": 0.0023, "V": 0.001, "Y": 0.0001 },
    "101": { "A": 0.0033, "D": 0.9512, "E": 0.0128, "F": 0.0008, "G": 0.0014, "H": 0.0078, "I": 0.0014, "K": 0.0004, "L": 0.0009, "N": 0.0039, "Q": 0.0003, "R": 0.0001, "S": 0.0004, "T": 0.0024, "V": 0.0035, "Y": 0.0093 },
    "102": { "C": 0.0006, "D": 0.0058, "E": 0.0006, "F": 0.0026, "H": 0.0003, "I": 0.0001, "L": 0.0001, "S": 0.0012, "Y": 0.9887 },
    "103": { "A": 0.0001, "C": 0.0032, "D": 0.0042, "E": 0.0001, "F": 0.0164, "H": 0.0074, "I": 0.0002, "L": 0.0001, "N": 0.0003, "Q": 0.0001, "S": 0.005, "T": 0.0001, "V": 0.0001, "W": 0.0003, "Y": 0.9622 },
    "104": { "A": 0.0001, "C": 0.9862, "F": 0.0016, "G": 0.0077, "H": 0.0001, "R": 0.0018, "S": 0.0013, "W": 0.0002, "Y": 0.001 },
    "105": { "A": 0.0064, "C": 0.3391, "D": 0.0005, "F": 0.0026, "G": 0.0078, "H": 0.0008, "I": 0.0041, "K": 0.0003, "L": 0.001, "M": 0.0001, "N": 0.0099, "Q": 0.0017, "R": 0.0035, "S": 0.6031, "T": 0.0127, "V": 0.0004, "W": 0.0005, "Y": 0.0055 },
    "106": { "A": 0.0127, "C": 0.0002, "E": 0.0001, "F": 0.0003, "H": 0.0001, "I": 0.0001, "L": 0.006, "P": 0.0014, "Q": 0.0001, "R": 0.0001, "S": 0.9755, "T": 0.0027, "V": 0.0004, "Y": 0.0004 },
    "107": { "A": 0.0003, "C": 0.0035, "D": 0.0099, "E": 0.0006, "F": 0.0176, "G": 0.0002, "H": 0.0074, "I": 0.0003, "K": 0.001, "L": 0.0011, "M": 0.0001, "N": 0.0039, "P": 0.0007, "Q": 0.0035, "R": 0.0007, "S": 0.005, "T": 0.0003, "V": 0.0005, "W": 0.0032, "Y": 0.9402 },
    "108": { "A": 0.4477, "D": 0.0037, "E": 0.0016, "F": 0.0002, "G": 0.0091, "I": 0.0042, "K": 0.0013, "L": 0.001, "M": 0.0005, "N": 0.0001, "P": 0.0012, "Q": 0.0001, "R": 0.0084, "S": 0.0136, "T": 0.4985, "V": 0.0085, "Y": 0.0003 },
    "109": { "A": 0.0053, "C": 0.0017, "D": 0.0099, "E": 0.0007, "F": 0.0008, "G": 0.4438, "H": 0.0016, "I": 0.0049, "K": 0.0013, "L": 0.002, "M": 0.0003, "N": 0.0078, "P": 0.0021, "Q": 0.0003, "R": 0.0114, "S": 0.4747, "T": 0.0262, "V": 0.0041, "W": 0.0001, "Y": 0.001 },
    "110": { "A": 0.0022, "C": 0.0009, "D": 0.0042, "E": 0.0006, "F": 0.0015, "G": 0.0147, "H": 0.0006, "I": 0.0079, "K": 0.0013, "L": 0.0011, "M": 0.0006, "N": 0.0099, "P": 0.0018, "R": 0.0147, "S": 0.8996, "T": 0.0341, "V": 0.0018, "W": 0.0006, "Y": 0.002 },
    "111": { "C": 0.0069, "G": 0.0069, "H": 0.0069, "I": 0.0069, "K": 0.0138, "N": 0.0552, "Q": 0.0069, "R": 0.0069, "S": 0.6897, "T": 0.0414, "Y": 0.1586 },
    "112": { "A": 0.0043, "C": 0.0007, "D": 0.01, "F": 0.0043, "G": 0.0157, "H": 0.0014, "I": 0.0086, "K": 0.0057, "L": 0.0136, "N": 0.0751, "P": 0.0057, "Q": 0.0007, "R": 0.005, "S": 0.6447, "T": 0.1008, "V": 0.0029, "Y": 0.1008 },
    "113": { "A": 0.0048, "C": 0.0005, "D": 0.0076, "E": 0.0007, "F": 0.0045, "G": 0.017, "H": 0.0019, "I": 0.0092, "K": 0.0022, "L": 0.0071, "M": 0.0002, "N": 0.0613, "P": 0.002, "Q": 0.0002, "R": 0.0169, "S": 0.7185, "T": 0.1092, "V": 0.0017, "W": 0.0003, "Y": 0.0343 },
    "114": { "A": 0.0093, "C": 0.001, "D": 0.0112, "E": 0.0022, "F": 0.0221, "G": 0.0372, "H": 0.0035, "I": 0.0154, "K": 0.0043, "L": 0.0375, "M": 0.0007, "N": 0.1073, "P": 0.027, "Q": 0.0012, "R": 0.0142, "S": 0.4013, "T": 0.2319, "V": 0.0041, "W": 0.0004, "Y": 0.0682 },
    "115": { "A": 0.0137, "C": 0.0028, "D": 0.02, "E": 0.01, "F": 0.0536, "G": 0.0241, "H": 0.0347, "I": 0.02, "K": 0.0074, "L": 0.1249, "M": 0.0009, "N": 0.0774, "P": 0.0705, "Q": 0.0033, "R": 0.0235, "S": 0.1048, "T": 0.3518, "V": 0.0118, "W": 0.0028, "Y": 0.0417 },
    "116": { "A": 0.0089, "C": 0.005, "D": 0.0089, "E": 0.0198, "F": 0.0251, "G": 0.0438, "H": 0.0068, "I": 0.0064, "K": 0.0099, "L": 0.1419, "M": 0.0037, "N": 0.0049, "P": 0.0112, "Q": 0.0075, "R": 0.0527, "S": 0.0111, "T": 0.0234, "V": 0.2836, "W": 0.1392, "Y": 0.1864 },
    "117": { "A": 0.0053, "C": 0.0002, "D": 0.0011, "E": 0.0058, "F": 0.0026, "G": 0.0066, "H": 0.0003, "I": 0.0246, "K": 0.0014, "L": 0.0468, "M": 0.0054, "N": 0.0001, "P": 0.0036, "Q": 0.0004, "R": 0.0039, "S": 0.0035, "T": 0.0039, "V": 0.884, "W": 0.0003, "Y": 0.0003 },
    "118": { "A": 0.0001, "C": 0.0037, "F": 0.982, "G": 0.0002, "I": 0.0019, "L": 0.0053, "N": 0.0001, "R": 0.0001, "S": 0.0008, "V": 0.0054, "Y": 0.0004 },
    "119": { "A": 0.0014, "C": 0.0008, "D": 0.0009, "E": 0.0005, "G": 0.9942, "L": 0.0001, "R": 0.0002, "S": 0.0009, "V": 0.0011 },
    "120": { "A": 0.004, "C": 0.0001, "D": 0.0001, "E": 0.0014, "G": 0.6926, "H": 0.0002, "I": 0.0012, "L": 0.0002, "N": 0.0007, "P": 0.0022, "R": 0.0024, "S": 0.0103, "T": 0.2827, "V": 0.0018, "Y": 0.0001 },
    "121": { "A": 0.001, "D": 0.0001, "E": 0.0008, "G": 0.9922, "M": 0.0001, "R": 0.0008, "S": 0.0001, "V": 0.0026, "W": 0.0024 },
    "122": { "A": 0.0014, "I": 0.0008, "N": 0.0006, "P": 0.0023, "S": 0.0011, "T": 0.9938 },
    "123": { "A": 0.0005, "D": 0.0005, "E": 0.0048, "G": 0.0001, "H": 0.0006, "I": 0.0001, "K": 0.9449, "M": 0.0009, "N": 0.0049, "P": 0.0003, "Q": 0.0083, "R": 0.0185, "S": 0.0027, "T": 0.0122, "V": 0.0001, "W": 0.0002, "Y": 0.0004 },
    "124": { "A": 0.0001, "D": 0.0001, "F": 0.0007, "G": 0.0006, "H": 0.0001, "I": 0.0027, "L": 0.6696, "M": 0.0009, "P": 0.0005, "Q": 0.0001, "R": 0.0018, "S": 0.0001, "V": 0.3226 },
    "125": { "A": 0.0032, "D": 0.0001, "F": 0.0001, "I": 0.0021, "K": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0014, "P": 0.0031, "S": 0.0044, "T": 0.9851, "V": 0.0002, "Y": 0.0001 },
    "126": { "A": 0.0025, "D": 0.0005, "F": 0.0007, "G": 0.0056, "I": 0.0012, "L": 0.0003, "R": 0.0002, "V": 0.989 },
    "127": { "I": 0.0014, "L": 0.9955, "P": 0.0002, "V": 0.0029 }
  },
  "IGLV3": {
    "1": { "Q": 0.0001, "S": 0.9999 },
    "2": { "A": 0.001, "C": 0.0022, "D": 0.0023, "E": 0.0001, "F": 0.011, "G": 0.0001, "H": 0.0067, "I": 0.0001, "L": 0.0011, "N": 0.0017, "P": 0.0013, "Q": 0.0001, "R": 0.0002, "S": 0.2241, "T": 0.0002, "V": 0.0003, "Y": 0.7475 },
    "3": { "A": 0.024, "D": 0.0223, "E": 0.7104, "F": 0.0002, "G": 0.0036, "H": 0.0003, "I": 0.001, "K": 0.0022, "L": 0.0046, "M": 0.0009, "P": 0.0003, "Q": 0.0039, "R": 0.0002, "S": 0.0003, "T": 0.0003, "V": 0.2248, "W": 0.0001, "Y": 0.0005 },
    "4": { "A": 0.0001, "F": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.9768, "M": 0.0038, "P": 0.005, "Q": 0.0005, "R": 0.0034, "S": 0.0001, "T": 0.0001, "V": 0.0099, "W": 0.0001 },
    "5": { "A": 0.0045, "D": 0.0001, "H": 0.0008, "I": 0.0034, "K": 0.0006, "L": 0.0003, "N": 0.0025, "P": 0.0312, "Q": 0.0002, "R": 0.0004, "S": 0.0065, "T": 0.9494, "V": 0.0001 },
    "6": { "E": 0.0008, "H": 0.0166, "K": 0.0032, "L": 0.0017, "M": 0.0001, "N": 0.0001, "P": 0.0304, "Q": 0.9452, "R": 0.0011, "S": 0.0003, "T": 0.0003, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "7": { "A": 0.0125, "D": 0.1738, "E": 0.0014, "G": 0.0008, "H": 0.0019, "I": 0.0001, "L": 0.0039, "N": 0.0002, "P": 0.7787, "Q": 0.0049, "R": 0.0014, "S": 0.0092, "T": 0.0081, "V": 0.001, "Y": 0.0021 },
    "8": { "A": 0.0031, "D": 0.0001, "E": 0.0001, "F": 0.0003, "G": 0.0001, "H": 0.0088, "I": 0.0001, "L": 0.0271, "N": 0.0001, "P": 0.9305, "R": 0.0021, "S": 0.0192, "T": 0.008, "V": 0.0003, "Y": 0.0001 },
    "9": { "A": 0.1858, "C": 0.0001, "D": 0.0014, "E": 0.0002, "F": 0.0002, "G": 0.0005, "H": 0.0001, "I": 0.0001, "L": 0.0014, "P": 0.0282, "Q": 0.0007, "S": 0.7757, "T": 0.0036, "V": 0.0014, "W": 0.0002, "Y": 0.0004 },
    "11": { "A": 0.0137, "C": 0.0001, "D": 0.0002, "E": 0.004, "F": 0.0006, "G": 0.0072, "I": 0.0005, "K": 0.0001, "L": 0.0206, "M": 0.0044, "P": 0.0002, "Q": 0.0001, "R": 0.0003, "S": 0.0003, "T": 0.0003, "V": 0.9472, "W": 0.0002 },
    "12": { "A": 0.0112, "C": 0.0008, "D": 0.0001, "E": 0.0001, "F": 0.0013, "G": 0.0001, "H": 0.0003, "I": 0.0001, "L": 0.0004, "P": 0.0147, "Q": 0.0002, "R": 0.0001, "S": 0.9618, "T": 0.0034, "V": 0.0003, "Y": 0.0052 },
    "13": { "A": 0.009, "C": 0.0001, "D": 0.0002, "E": 0.0017, "F": 0.001, "G": 0.0082, "I": 0.0002, "L": 0.0127, "M": 0.0025, "P": 0.0002, "Q": 0.0001, "R": 0.0001, "S": 0.0004, "T": 0.0001, "V": 0.9632, "W": 0.0002 },
    "14": { "A": 0.4427, "C": 0.0005, "D": 0.0034, "F": 0.0017, "G": 0.0007, "H": 0.0001, "L": 0.0003, "P": 0.0149, "S": 0.5271, "T": 0.0033, "V": 0.002, "W": 0.0001, "Y": 0.0032 },
    "15": { "A": 0.002, "F": 0.0051, "H": 0.0011, "L": 0.2095, "M": 0.001, "P": 0.762, "Q": 0.0051, "R": 0.0016, "S": 0.004, "T": 0.0061, "V": 0.0013, "W": 0.0012 },
    "16": { "A": 0.0053, "C": 0.0004, "D": 0.0003, "E": 0.0018, "G": 0.9803, "P": 0.0001, "Q": 0.0001, "R": 0.0068, "S": 0.0001, "V": 0.0046, "W": 0.0002 },
    "17": { "A": 0.0005, "D": 0.0001, "E": 0.0097, "G": 0.0005, "H": 0.0052, "K": 0.0962, "L": 0.0014, "M": 0.0008, "N": 0.0014, "P": 0.0173, "Q": 0.8592, "R": 0.0043, "S": 0.0004, "T": 0.0029, "W": 0.0001 },
    "18": { "A": 0.005, "C": 0.0001, "D": 0.0001, "E": 0.0002, "G": 0.0001, "I": 0.0007, "K": 0.0038, "L": 0.0002, "M": 0.0079, "N": 0.0003, "P": 0.0206, "Q": 0.0005, "R": 0.0012, "S": 0.0025, "T": 0.9566, "V": 0.0001 },
    "19": { "A": 0.7971, "D": 0.0049, "F": 0.0005, "G": 0.0029, "H": 0.0002, "I": 0.0005, "L": 0.0004, "P": 0.0026, "S": 0.0026, "T": 0.0027, "V": 0.1856 },
    "20": { "A": 0.0004, "C": 0.0003, "D": 0.0002, "E": 0.001, "F": 0.0001, "G": 0.0019, "H": 0.0004, "I": 0.0054, "K": 0.0075, "L": 0.0004, "M": 0.0028, "N": 0.004, "P": 0.0008, "Q": 0.0005, "R": 0.6168, "S": 0.3215, "T": 0.0347, "V": 0.0008, "W": 0.0005, "Y": 0.0001 },
    "21": { "D": 0.0001, "F": 0.0023, "H": 0.0008, "I": 0.9427, "L": 0.0208, "M": 0.0034, "N": 0.0047, "P": 0.0003, "R": 0.0003, "S": 0.0087, "T": 0.0095, "V": 0.0062 },
    "22": { "A": 0.0058, "D": 0.0003, "F": 0.0002, "H": 0.0006, "I": 0.0038, "K": 0.0012, "L": 0.0002, "M": 0.0004, "N": 0.0045, "P": 0.0328, "Q": 0.0001, "R": 0.0005, "S": 0.0113, "T": 0.9375, "V": 0.0005, "Y": 0.0002 },
    "23": { "A": 0.0001, "C": 0.9612, "D": 0.0001, "F": 0.0045, "G": 0.0085, "H": 0.0001, "L": 0.0004, "R": 0.0097, "S": 0.0061, "T": 0.0001, "V": 0.0007, "W": 0.0072, "Y": 0.0012 },
    "24": { "A": 0.0108, "C": 0.0003, "D": 0.0008, "E": 0.0054, "F": 0.0023, "G": 0.2483, "H": 0.0037, "I": 0.0001, "K": 0.0021, "L": 0.0007, "N": 0.0001, "P": 0.037, "Q": 0.1748, "R": 0.0047, "S": 0.4961, "T": 0.0045, "V": 0.0019, "W": 0.0045, "Y": 0.0018 },
    "25": { "A": 0.0143, "C": 0.0006, "D": 0.0002, "E": 0.0036, "F": 0.0001, "G": 0.9688, "H": 0.0001, "I": 0.0002, "K": 0.0001, "L": 0.0001, "Q": 0.0001, "R": 0.0048, "S": 0.0005, "T": 0.0001, "V": 0.0061, "W": 0.0001 },
    "26": { "A": 0.0214, "C": 0.0001, "D": 0.6867, "E": 0.0184, "F": 0.0002, "G": 0.0055, "H": 0.0108, "I": 0.001, "K": 0.0053, "L": 0.0001, "M": 0.0001, "N": 0.2264, "P": 0.0017, "Q": 0.001, "R": 0.0013, "S": 0.0051, "T": 0.0065, "V": 0.0029, "W": 0.0001, "Y": 0.0051 },
    "27": { "A": 0.1784, "C": 0.0003, "D": 0.0083, "E": 0.0073, "F": 0.0004, "G": 0.0048, "H": 0.006, "I": 0.0055, "K": 0.2965, "L": 0.0012, "M": 0.0003, "N": 0.2559, "P": 0.0021, "Q": 0.0073, "R": 0.009, "S": 0.1823, "T": 0.0176, "V": 0.0156, "Y": 0.0012 },
    "28": { "A": 0.0002, "C": 0.0005, "D": 0.0001, "E": 0.0001, "F": 0.0093, "G": 0.0005, "H": 0.0007, "I": 0.2383, "K": 0.0003, "L": 0.708, "M": 0.0064, "N": 0.0011, "P": 0.0034, "R": 0.0017, "S": 0.009, "T": 0.0023, "V": 0.0132, "W": 0.0048, "Y": 0.0001 },
    "29": { "A": 0.0221, "C": 0.0004, "D": 0.0022, "E": 0.0108, "G": 0.5746, "H": 0.0004, "I": 0.0011, "K": 0.0019, "L": 0.0004, "M": 0.0001, "N": 0.0001, "P": 0.1772, "Q": 0.0038, "R": 0.1833, "S": 0.0086, "T": 0.0048, "V": 0.0064, "W": 0.0017, "Y": 0.0001 },
    "36": { "A": 0.0085, "C": 0.0006, "D": 0.3019, "E": 0.0119, "F": 0.0005, "G": 0.0101, "H": 0.0031, "I": 0.0042, "K": 0.184, "L": 0.001, "M": 0.0013, "N": 0.0255, "P": 0.001, "Q": 0.0047, "R": 0.0252, "S": 0.3833, "T": 0.0242, "V": 0.0035, "W": 0.0001, "Y": 0.0055 },
    "37": { "A": 0.0009, "C": 0.0005, "D": 0.0034, "E": 0.0075, "F": 0.0042, "G": 0.0005, "H": 0.0086, "I": 0.0024, "K": 0.6048, "L": 0.0011, "M": 0.0005, "N": 0.0184, "P": 0.0022, "Q": 0.1421, "R": 0.0125, "S": 0.0057, "T": 0.0123, "V": 0.0004, "W": 0.0001, "Y": 0.1717 },
    "38": { "A": 0.0026, "C": 0.0021, "D": 0.0097, "E": 0.0005, "F": 0.0156, "G": 0.0037, "H": 0.0062, "I": 0.0025, "K": 0.0008, "L": 0.0014, "M": 0.0005, "N": 0.0309, "P": 0.0008, "Q": 0.0005, "R": 0.013, "S": 0.2212, "T": 0.0079, "V": 0.0007, "W": 0.0003, "Y": 0.6794 },
    "39": { "A": 0.6665, "C": 0.0004, "D": 0.0041, "E": 0.0032, "F": 0.0006, "G": 0.0116, "H": 0.0001, "I": 0.0029, "L": 0.0036, "M": 0.0005, "N": 0.0002, "P": 0.0044, "Q": 0.0001, "R": 0.0001, "S": 0.0089, "T": 0.0109, "V": 0.2815, "W": 0.0001, "Y": 0.0002 },
    "40": { "A": 0.0018, "C": 0.3035, "D": 0.0023, "E": 0.0001, "F": 0.0082, "G": 0.0047, "H": 0.25, "I": 0.0013, "K": 0.0001, "L": 0.001, "N": 0.0113, "P": 0.0057, "Q": 0.0047, "R": 0.0208, "S": 0.1961, "T": 0.0042, "V": 0.0003, "W": 0.0007, "Y": 0.1832 },
    "41": { "C": 0.0086, "E": 0.0002, "F": 0.0001, "G": 0.0108, "L": 0.0058, "Q": 0.0001, "R": 0.0078, "S": 0.001, "T": 0.0001, "V": 0.0003, "W": 0.965, "Y": 0.0002 },
    "42": { "A": 0.0003, "C": 0.0032, "D": 0.0073, "E": 0.0001, "F": 0.0211, "H": 0.0036, "I": 0.0001, "L": 0.0011, "M": 0.0001, "N": 0.0016, "Q": 0.0002, "R": 0.0001, "S": 0.0086, "T": 0.0001, "V": 0.0002, "Y": 0.9523 },
    "43": { "E": 0.0007, "F": 0.0001, "H": 0.0116, "K": 0.0054, "L": 0.0017, "N": 0.0003, "P": 0.0108, "Q": 0.959, "R": 0.0088, "S": 0.0003, "T": 0.0005, "V": 0.0002, "Y": 0.0007 },
    "44": { "A": 0.0001, "C": 0.0001, "E": 0.0022, "H": 0.0112, "K": 0.0084, "L": 0.0045, "N": 0.0002, "P": 0.0073, "Q": 0.959, "R": 0.0051, "S": 0.0001, "T": 0.0007, "V": 0.0011, "W": 0.0001 },
    "45": { "D": 0.0001, "E": 0.0028, "F": 0.0001, "G": 0.0001, "H": 0.0004, "I": 0.0002, "K": 0.9322, "M": 0.0018, "N": 0.007, "P": 0.0001, "Q": 0.0094, "R": 0.0339, "S": 0.0007, "T": 0.0112, "Y": 0.0001 },
    "46": { "A": 0.0152, "D": 0.0001, "E": 0.0001, "H": 0.0005, "K": 0.0001, "L": 0.002, "P": 0.9114, "Q": 0.0101, "R": 0.0005, "S": 0.051, "T": 0.0088, "V": 0.0002 },
    "47": { "A": 0.0038, "C": 0.0027, "D": 0.0025, "E": 0.0005, "G": 0.9809, "N": 0.0002, "Q": 0.0001, "R": 0.0029, "S": 0.002, "V": 0.0044 },
    "48": { "A": 0.0001, "E": 0.001, "H": 0.0073, "K": 0.0081, "L": 0.0018, "N": 0.0003, "P": 0.0133, "Q": 0.9619, "R": 0.0053, "S": 0.0001, "T": 0.0008, "Y": 0.0001 },
    "49": { "A": 0.6416, "C": 0.0004, "D": 0.0042, "E": 0.0001, "F": 0.0072, "G": 0.0014, "H": 0.0002, "I": 0.0001, "P": 0.0069, "S": 0.3299, "T": 0.0038, "V": 0.0025, "Y": 0.0017 },
    "50": { "A": 0.001, "H": 0.0045, "L": 0.0027, "P": 0.9834, "Q": 0.0003, "R": 0.0007, "S": 0.0015, "T": 0.006 },
    "51": { "A": 0.0052, "C": 0.0001, "D": 0.0001, "E": 0.0034, "F": 0.0016, "G": 0.0112, "H": 0.0001, "I": 0.0147, "K": 0.0008, "L": 0.0304, "M": 0.0047, "N": 0.0001, "P": 0.0003, "Q": 0.0005, "R": 0.0019, "S": 0.0007, "T": 0.0016, "V": 0.922, "W": 0.0006, "Y": 0.0001 },
    "52": { "A": 0.001, "D": 0.0001, "E": 0.0001, "F": 0.0018, "G": 0.0006, "H": 0.0005, "I": 0.0023, "K": 0.0005, "L": 0.9473, "M": 0.0116, "P": 0.0034, "Q": 0.0017, "R": 0.006, "S": 0.001, "T": 0.0002, "V": 0.0214, "W": 0.0005, "Y": 0.0001 },
    "53": { "A": 0.0029, "C": 0.0008, "D": 0.0011, "E": 0.0003, "F": 0.0083, "G": 0.0086, "I": 0.0097, "L": 0.0147, "M": 0.0015, "N": 0.0001, "P": 0.0003, "R": 0.0003, "S": 0.0005, "T": 0.0002, "V": 0.9506, "W": 0.0001, "Y": 0.0001 },
    "54": { "A": 0.0013, "C": 0.0001, "D": 0.0006, "F": 0.0038, "G": 0.0018, "H": 0.0003, "I": 0.7821, "K": 0.0003, "L": 0.016, "M": 0.0129, "N": 0.0023, "P": 0.0001, "R": 0.0018, "S": 0.0062, "T": 0.0035, "V": 0.1667, "Y": 0.0001 },
    "55": { "A": 0.0005, "C": 0.0027, "D": 0.0098, "E": 0.0009, "F": 0.0217, "G": 0.0001, "H": 0.0086, "I": 0.0002, "K": 0.0005, "L": 0.0008, "N": 0.0033, "P": 0.0003, "Q": 0.0007, "R": 0.001, "S": 0.0149, "T": 0.0006, "V": 0.0004, "W": 0.0002, "Y": 0.9327 },
    "56": { "A": 0.0056, "C": 0.0014, "D": 0.1764, "E": 0.0512, "F": 0.0012, "G": 0.1763, "H": 0.0053, "I": 0.0003, "K": 0.1595, "L": 0.0008, "M": 0.0001, "N": 0.0062, "P": 0.0019, "Q": 0.3212, "R": 0.0291, "S": 0.0037, "T": 0.0019, "V": 0.0024, "W": 0.0002, "Y": 0.0553 },
    "57": { "A": 0.0086, "C": 0.0001, "D": 0.7666, "E": 0.0123, "F": 0.0001, "G": 0.0049, "H": 0.0023, "I": 0.0003, "K": 0.1726, "L": 0.0001, "M": 0.0001, "N": 0.0167, "P": 0.0003, "Q": 0.0023, "R": 0.0022, "S": 0.0013, "T": 0.0027, "V": 0.0027, "Y": 0.0038 },
    "65": { "A": 0.003, "C": 0.001, "D": 0.0132, "E": 0.0005, "F": 0.001, "G": 0.0069, "H": 0.0035, "I": 0.0086, "K": 0.0073, "L": 0.0001, "M": 0.0005, "N": 0.2085, "P": 0.0009, "Q": 0.0004, "R": 0.0225, "S": 0.6605, "T": 0.0559, "V": 0.0021, "W": 0.0004, "Y": 0.0031 },
    "66": { "A": 0.0093, "D": 0.2215, "E": 0.1625, "F": 0.0005, "G": 0.0073, "H": 0.004, "I": 0.0028, "K": 0.3648, "L": 0.001, "M": 0.0006, "N": 0.1992, "P": 0.0002, "Q": 0.0051, "R": 0.0073, "S": 0.0027, "T": 0.0069, "V": 0.0023, "Y": 0.0018 },
    "67": { "C": 0.0001, "G": 0.002, "I": 0.0001, "K": 0.0003, "L": 0.0034, "M": 0.0003, "P": 0.0027, "Q": 0.002, "R": 0.9855, "S": 0.0008, "T": 0.0001, "W": 0.0027 },
    "68": { "A": 0.0043, "D": 0.0001, "F": 0.0001, "H": 0.0065, "I": 0.0001, "L": 0.0016, "N": 0.0001, "P": 0.9729, "Q": 0.0001, "R": 0.0021, "S": 0.004, "T": 0.0071, "V": 0.0003, "Y": 0.0007 },
    "69": { "A": 0.0188, "C": 0.0002, "E": 0.0001, "F": 0.0003, "G": 0.0001, "L": 0.0029, "N": 0.0001, "P": 0.0181, "R": 0.0001, "S": 0.9553, "T": 0.0034, "V": 0.0001, "W": 0.0001, "Y": 0.0003 },
    "70": { "A": 0.0016, "C": 0.0003, "D": 0.0007, "E": 0.0055, "G": 0.9744, "K": 0.0002, "L": 0.0002, "M": 0.0001, "N": 0.0003, "R": 0.0038, "S": 0.0001, "T": 0.0005, "V": 0.0071, "W": 0.0049, "Y": 0.0001 },
    "71": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.0034, "G": 0.0001, "H": 0.0003, "I": 0.9422, "L": 0.0124, "M": 0.0022, "N": 0.0054, "P": 0.0003, "R": 0.0005, "S": 0.0115, "T": 0.0105, "V": 0.011 },
    "72": { "A": 0.0029, "C": 0.0002, "F": 0.0001, "G": 0.0001, "H": 0.0025, "L": 0.0018, "P": 0.9803, "Q": 0.0012, "R": 0.0014, "S": 0.0052, "T": 0.0041 },
    "74": { "A": 0.0192, "D": 0.193, "E": 0.7465, "G": 0.0249, "H": 0.001, "K": 0.0018, "L": 0.0001, "M": 0.0001, "N": 0.0012, "P": 0.0005, "Q": 0.0021, "R": 0.0005, "S": 0.0008, "T": 0.0001, "V": 0.0054, "W": 0.0008, "Y": 0.0018 },
    "75": { "A": 0.0001, "C": 0.0005, "E": 0.0001, "G": 0.0036, "H": 0.0001, "I": 0.0003, "K": 0.0002, "L": 0.0056, "P": 0.0016, "Q": 0.0014, "R": 0.9858, "S": 0.0005, "T": 0.0001, "V": 0.0001 },
    "76": { "A": 0.0001, "C": 0.0095, "F": 0.9579, "G": 0.0003, "I": 0.0084, "L": 0.0097, "M": 0.0002, "R": 0.0001, "S": 0.0013, "T": 0.0001, "V": 0.011, "W": 0.0001, "Y": 0.0015 },
    "77": { "A": 0.0173, "C": 0.0008, "D": 0.0003, "F": 0.0021, "I": 0.0001, "L": 0.0003, "P": 0.0035, "S": 0.9705, "T": 0.0025, "V": 0.0002, "W": 0.0001, "Y": 0.0025 },
    "78": { "A": 0.0124, "C": 0.0071, "D": 0.0021, "F": 0.0001, "G": 0.9701, "R": 0.0006, "S": 0.0021, "V": 0.0055 },
    "79": { "A": 0.0169, "C": 0.0018, "D": 0.0001, "F": 0.0031, "G": 0.0003, "H": 0.0003, "L": 0.0002, "N": 0.0003, "P": 0.0034, "R": 0.0001, "S": 0.9646, "T": 0.0034, "V": 0.0001, "W": 0.0001, "Y": 0.0054 },
    "80": { "A": 0.0012, "C": 0.0003, "D": 0.0025, "E": 0.0001, "F": 0.0002, "G": 0.0023, "H": 0.004, "I": 0.0035, "K": 0.0142, "L": 0.0003, "M": 0.0001, "N": 0.5717, "P": 0.0005, "Q": 0.0001, "R": 0.0095, "S": 0.3672, "T": 0.0195, "V": 0.0005, "W": 0.0001, "Y": 0.0021 },
    "83": { "A": 0.021, "C": 0.0008, "D": 0.0001, "E": 0.0001, "F": 0.002, "G": 0.0003, "H": 0.0001, "L": 0.001, "N": 0.0002, "P": 0.004, "Q": 0.0001, "R": 0.0001, "S": 0.9642, "T": 0.0044, "V": 0.0001, "W": 0.0001, "Y": 0.0016 },
    "84": { "A": 0.0058, "C": 0.0002, "D": 0.0013, "E": 0.0082, "G": 0.9675, "H": 0.0003, "I": 0.0001, "K": 0.0001, "L": 0.0001, "M": 0.0001, "Q": 0.0003, "R": 0.0047, "S": 0.0005, "T": 0.0003, "V": 0.006, "W": 0.0045 },
    "85": { "A": 0.0021, "D": 0.0123, "E": 0.0005, "F": 0.0001, "G": 0.0007, "H": 0.0032, "I": 0.0023, "K": 0.0092, "L": 0.0001, "N": 0.7478, "P": 0.0005, "Q": 0.0005, "R": 0.0008, "S": 0.0077, "T": 0.2111, "Y": 0.001 },
    "86": { "A": 0.0055, "E": 0.0001, "G": 0.0004, "I": 0.0108, "K": 0.0043, "L": 0.0017, "M": 0.0366, "N": 0.0005, "P": 0.0032, "Q": 0.0001, "R": 0.0024, "S": 0.004, "T": 0.9272, "V": 0.003 },
    "87": { "A": 0.8224, "C": 0.0002, "D": 0.0024, "E": 0.0001, "F": 0.0007, "G": 0.004, "I": 0.0007, "L": 0.0004, "N": 0.0002, "P": 0.0004, "R": 0.0001, "S": 0.0033, "T": 0.0036, "V": 0.1615 },
    "88": { "A": 0.0063, "C": 0.0004, "F": 0.0005, "H": 0.0001, "I": 0.0018, "K": 0.0005, "L": 0.0001, "M": 0.0003, "N": 0.0021, "P": 0.0027, "R": 0.0001, "S": 0.1826, "T": 0.8011, "V": 0.0001, "Y": 0.0014 },
    "89": { "C": 0.0003, "F": 0.0042, "I": 0.0003, "K": 0.0001, "L": 0.9825, "M": 0.0027, "P": 0.0014, "Q": 0.0001, "R": 0.0029, "S": 0.0008, "T": 0.0001, "V": 0.0035, "W": 0.0012 },
    "90": { "A": 0.0055, "C": 0.0001, "F": 0.0003, "G": 0.0003, "H": 0.0001, "I": 0.0082, "K": 0.0005, "L": 0.0004, "M": 0.0003, "N": 0.0027, "P": 0.0042, "Q": 0.0001, "R": 0.0001, "S": 0.0053, "T": 0.9713, "V": 0.0008 },
    "91": { "C": 0.0002, "F": 0.0009, "G": 0.0001, "I": 0.9667, "K": 0.0001, "L": 0.0051, "M": 0.0008, "N": 0.0007, "P": 0.0001, "R": 0.0001, "S": 0.0147, "T": 0.0036, "V": 0.0069 },
    "92": { "A": 0.0027, "C": 0.001, "D": 0.001, "F": 0.0003, "G": 0.007, "H": 0.0001, "I": 0.004, "K": 0.0003, "L": 0.0001, "M": 0.0001, "N": 0.0084, "P": 0.0017, "R": 0.0147, "S": 0.7633, "T": 0.1949, "V": 0.0002, "Y": 0.0003 },
    "93": { "A": 0.0029, "C": 0.0003, "D": 0.0012, "E": 0.0068, "G": 0.721, "I": 0.0001, "K": 0.0011, "L": 0.0001, "M": 0.0004, "N": 0.0006, "P": 0.0001, "Q": 0.0001, "R": 0.2528, "S": 0.0039, "T": 0.0013, "V": 0.0039, "W": 0.0034 },
    "94": { "A": 0.2644, "D": 0.0012, "F": 0.0008, "G": 0.004, "H": 0.0001, "I": 0.0013, "K": 0.0002, "L": 0.0019, "N": 0.0017, "P": 0.0012, "S": 0.0029, "T": 0.3331, "V": 0.3872 },
    "95": { "A": 0.0016, "D": 0.0043, "E": 0.2353, "G": 0.0016, "H": 0.0034, "I": 0.0001, "K": 0.0027, "L": 0.003, "M": 0.0001, "P": 0.0018, "Q": 0.7417, "R": 0.004, "T": 0.0002, "V": 0.0001, "Y": 0.0001 },
    "96": { "A": 0.902, "D": 0.0048, "E": 0.0023, "F": 0.0003, "G": 0.0028, "H": 0.0001, "I": 0.0004, "K": 0.0001, "L": 0.0008, "M": 0.0001, "N": 0.0003, "P": 0.0084, "S": 0.0058, "T": 0.0087, "V": 0.0632 },
    "97": { "A": 0.0028, "D": 0.0067, "E": 0.3894, "F": 0.0003, "G": 0.2587, "I": 0.0044, "K": 0.0008, "L": 0.0054, "M": 0.3164, "N": 0.0001, "P": 0.0001, "Q": 0.0005, "R": 0.0032, "S": 0.0001, "T": 0.0018, "V": 0.0088, "W": 0.0005 },
    "98": { "A": 0.0009, "D": 0.9842, "E": 0.0089, "F": 0.0001, "G": 0.001, "H": 0.0006, "K": 0.0002, "N": 0.0006, "S": 0.0001, "V": 0.0009, "Y": 0.0025 },
    "99": { "A": 0.0043, "D": 0.0098, "E": 0.9801, "G": 0.0038, "K": 0.0002, "Q": 0.0003, "R": 0.0001, "V": 0.0013, "Y": 0.0001 },
    "100": { "A": 0.9721, "C": 0.0001, "D": 0.0069, "E": 0.0001, "G": 0.0123, "N": 0.0001, "P": 0.0005, "S": 0.0053, "T": 0.0012, "V": 0.0012, "Y": 0.0003 },
    "101": { "A": 0.0056, "D": 0.9683, "E": 0.0064, "G": 0.0025, "H": 0.0014, "I": 0.0001, "K": 0.0002, "L": 0.0001, "N": 0.0013, "Q": 0.0001, "R": 0.0008, "S": 0.0008, "T": 0.0005, "V": 0.0046, "W": 0.0001, "Y": 0.0073 },
    "102": { "C": 0.001, "D": 0.0092, "E": 0.0005, "F": 0.0056, "G": 0.0001, "H": 0.0006, "I": 0.0001, "L": 0.0001, "N": 0.0004, "S": 0.0012, "V": 0.0001, "Y": 0.9812 },
    "103": { "A": 0.0003, "C": 0.0039, "D": 0.0073, "E": 0.0001, "F": 0.0169, "H": 0.004, "I": 0.0003, "L": 0.0003, "N": 0.0006, "R": 0.0001, "S": 0.0048, "V": 0.0002, "W": 0.0001, "Y": 0.9612 },
    "104": { "A": 0.0001, "C": 0.9775, "D": 0.0001, "F": 0.0018, "G": 0.0077, "H": 0.0001, "P": 0.0001, "R": 0.0012, "S": 0.0019, "V": 0.0001, "W": 0.0082, "Y": 0.0013 },
    "105": { "A": 0.001, "C": 0.0009, "D": 0.0014, "E": 0.001, "F": 0.002, "G": 0.001, "H": 0.0122, "I": 0.0004, "K": 0.0036, "L": 0.0118, "M": 0.0002, "N": 0.1704, "P": 0.0023, "Q": 0.7296, "R": 0.0027, "S": 0.005, "T": 0.0012, "V": 0.0001, "Y": 0.0533 },
    "106": { "A": 0.3328, "C": 0.0018, "D": 0.0005, "E": 0.0021, "F": 0.0008, "G": 0.002, "I": 0.001, "K": 0.0002, "L": 0.004, "M": 0.0014, "P": 0.0014, "Q": 0.0001, "R": 0.0001, "S": 0.3828, "T": 0.01, "V": 0.2578, "Y": 0.0011 },
    "107": { "A": 0.1516, "C": 0.0041, "E": 0.0009, "F": 0.0008, "G": 0.0073, "H": 0.0001, "I": 0.001, "K": 0.0001, "L": 0.004, "M": 0.0003, "P": 0.0036, "Q": 0.0011, "R": 0.1843, "S": 0.0049, "T": 0.0424, "V": 0.0031, "W": 0.5885, "Y": 0.002 },
    "108": { "A": 0.0172, "D": 0.9472, "E": 0.0086, "F": 0.0001, "G": 0.0057, "H": 0.0063, "I": 0.0006, "K": 0.0003, "L": 0.0003, "N": 0.0043, "P": 0.0004, "Q": 0.0002, "S": 0.001, "T": 0.0016, "V": 0.0027, "Y": 0.0034 },
    "109": { "A": 0.0035, "C": 0.0003, "D": 0.0177, "E": 0.0008, "F": 0.0015, "G": 0.0112, "H": 0.0009, "I": 0.0113, "K": 0.0014, "L": 0.0013, "M": 0.0006, "N": 0.0183, "P": 0.0021, "Q": 0.0003, "R": 0.0149, "S": 0.8869, "T": 0.0231, "V": 0.0018, "W": 0.0002, "Y": 0.0018 },
    "110": { "A": 0.0034, "C": 0.0005, "D": 0.0076, "E": 0.0004, "F": 0.0015, "G": 0.0148, "H": 0.0021, "I": 0.0059, "K": 0.0012, "L": 0.0017, "M": 0.0001, "N": 0.0174, "P": 0.0021, "Q": 0.0001, "R": 0.016, "S": 0.891, "T": 0.0291, "V": 0.0027, "Y": 0.0023 },
    "111": { "A": 0.008, "C": 0.002, "D": 0.0119, "G": 0.5845, "H": 0.002, "I": 0.006, "L": 0.002, "N": 0.004, "P": 0.006, "R": 0.002, "S": 0.3499, "T": 0.0179, "V": 0.004 },
    "112": { "A": 0.0046, "C": 0.0027, "D": 0.0979, "E": 0.0046, "G": 0.4362, "H": 0.0096, "I": 0.0046, "K": 0.0014, "L": 0.0027, "N": 0.0901, "P": 0.0041, "R": 0.0069, "S": 0.2561, "T": 0.0709, "V": 0.0023, "Y": 0.0055 },
    "113": { "A": 0.011, "C": 0.0016, "D": 0.0644, "E": 0.0019, "F": 0.001, "G": 0.2937, "H": 0.0333, "I": 0.0042, "K": 0.0013, "L": 0.0039, "M": 0.0002, "N": 0.0806, "P": 0.0031, "Q": 0.0013, "R": 0.0124, "S": 0.3907, "T": 0.0808, "V": 0.0038, "W": 0.0001, "Y": 0.0106 },
    "114": { "A": 0.0278, "C": 0.0006, "D": 0.1539, "E": 0.005, "F": 0.0028, "G": 0.0642, "H": 0.0757, "I": 0.0063, "K": 0.0051, "L": 0.0147, "M": 0.0003, "N": 0.1467, "P": 0.0314, "Q": 0.0039, "R": 0.0145, "S": 0.2671, "T": 0.154, "V": 0.0041, "W": 0.0012, "Y": 0.0207 },
    "115": { "A": 0.0324, "C": 0.0031, "D": 0.0333, "E": 0.0098, "F": 0.0187, "G": 0.0374, "H": 0.2539, "I": 0.0149, "K": 0.0067, "L": 0.0612, "M": 0.002, "N": 0.0465, "P": 0.0494, "Q": 0.023, "R": 0.0385, "S": 0.0656, "T": 0.2058, "V": 0.0077, "W": 0.0018, "Y": 0.0884 },
    "116": { "A": 0.0503, "C": 0.0036, "D": 0.0038, "E": 0.0208, "F": 0.0055, "G": 0.0836, "H": 0.0108, "I": 0.0033, "K": 0.0077, "L": 0.0443, "M": 0.0056, "N": 0.0048, "P": 0.0294, "Q": 0.0103, "R": 0.0537, "S": 0.0064, "T": 0.0118, "V": 0.4412, "W": 0.1054, "Y": 0.0977 },
    "117": { "A": 0.0032, "C": 0.0001, "D": 0.0005, "E": 0.0036, "F": 0.0052, "G": 0.0054, "H": 0.0003, "I": 0.0214, "K": 0.0005, "L": 0.0273, "M": 0.0027, "N": 0.0003, "P": 0.0045, "Q": 0.0006, "R": 0.0029, "S": 0.0038, "T": 0.0021, "V": 0.9149, "W": 0.0002, "Y": 0.0004 },
    "118": { "C": 0.0044, "F": 0.982, "G": 0.0001, "H": 0.0001, "I": 0.0016, "L": 0.0036, "M": 0.0001, "P": 0.0001, "R": 0.0005, "S": 0.0007, "V": 0.0061, "Y": 0.0008 },
    "119": { "A": 0.0028, "C": 0.0006, "D": 0.0005, "E": 0.0002, "G": 0.9935, "R": 0.0006, "S": 0.001, "V": 0.0007, "Y": 0.0001 },
    "120": { "A": 0.0045, "C": 0.0001, "D": 0.0001, "E": 0.0009, "F": 0.0001, "G": 0.8259, "I": 0.0007, "L": 0.0001, "N": 0.0004, "P": 0.0064, "R": 0.0021, "S": 0.0055, "T": 0.1517, "V": 0.0014, "W": 0.0001 },
    "121": { "A": 0.0012, "E": 0.0012, "F": 0.0001, "G": 0.991, "L": 0.0001, "Q": 0.0001, "R": 0.0012, "T": 0.0001, "V": 0.0023, "W": 0.0029 },
    "122": { "A": 0.0012, "D": 0.0001, "I": 0.0008, "N": 0.0008, "P": 0.0014, "R": 0.0001, "S": 0.0017, "T": 0.9938, "Y": 0.0001 },
    "123": { "A": 0.0002, "E": 0.0044, "F": 0.0001, "G": 0.0002, "H": 0.0005, "I": 0.0001, "K": 0.9552, "M": 0.0015, "N": 0.0038, "P": 0.0001, "Q": 0.0084, "R": 0.016, "S": 0.0008, "T": 0.0086, "Y": 0.0001 },
    "124": { "A": 0.0001, "F": 0.0003, "G": 0.0002, "I": 0.0014, "L": 0.8194, "M": 0.0008, "P": 0.0012, "Q": 0.0001, "R": 0.0014, "V": 0.175 },
    "125": { "A": 0.0037, "D": 0.0001, "F": 0.0001, "H": 0.0001, "I": 0.0021, "K": 0.0001, "L": 0.0001, "N": 0.0021, "P": 0.0036, "R": 0.0001, "S": 0.006, "T": 0.9814, "V": 0.0005 },
    "126": { "A": 0.0026, "C": 0.0001, "D": 0.0003, "F": 0.0008, "G": 0.0072, "I": 0.0021, "L": 0.0003, "R": 0.0004, "V": 0.9862 },
    "127": { "I": 0.0008, "L": 0.9959, "P": 0.0005, "V": 0.0028 }
  },
  "IGLV4": {
    "1": { "H": 0.0175, "L": 0.0001, "Q": 0.9824 },
    "2": { "A": 0.0007, "D": 0.0001, "F": 0.0029, "H": 0.0011, "I": 0.0037, "L": 0.7653, "M": 0.0001, "P": 0.1977, "R": 0.0025, "S": 0.0023, "T": 0.0032, "V": 0.0203 },
    "3": { "A": 0.0107, "D": 0.0001, "E": 0.0027, "F": 0.0004, "G": 0.0034, "I": 0.0018, "K": 0.0002, "L": 0.0138, "M": 0.0041, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0003, "T": 0.0004, "V": 0.9619 },
    "4": { "A": 0.0012, "F": 0.0003, "G": 0.0001, "H": 0.0001, "L": 0.9633, "M": 0.0025, "P": 0.0037, "Q": 0.0006, "R": 0.0041, "S": 0.0001, "T": 0.0001, "V": 0.0239 },
    "5": { "A": 0.0019, "D": 0.0001, "H": 0.0004, "I": 0.0008, "L": 0.0001, "N": 0.0013, "P": 0.04, "S": 0.0041, "T": 0.9513 },
    "6": { "E": 0.0002, "H": 0.0332, "I": 0.0001, "K": 0.0043, "L": 0.0013, "N": 0.0003, "P": 0.0277, "Q": 0.9305, "R": 0.0015, "S": 0.0004, "T": 0.0004, "Y": 0.0001 },
    "7": { "A": 0.0086, "E": 0.0001, "L": 0.0013, "P": 0.0197, "Q": 0.0001, "S": 0.9677, "T": 0.0018, "V": 0.0001, "W": 0.0004, "Y": 0.0001 },
    "8": { "A": 0.0006, "C": 0.0002, "F": 0.0003, "H": 0.0072, "L": 0.0037, "N": 0.0001, "P": 0.8039, "Q": 0.0001, "R": 0.0008, "S": 0.1781, "T": 0.0043, "Y": 0.0008 },
    "9": { "A": 0.004, "C": 0.0003, "D": 0.0001, "F": 0.0011, "H": 0.0001, "L": 0.0001, "P": 0.0126, "S": 0.977, "T": 0.0015, "Y": 0.0034 },
    "11": { "A": 0.9732, "D": 0.0035, "E": 0.0001, "F": 0.0006, "G": 0.0004, "H": 0.0001, "I": 0.0001, "L": 0.0001, "P": 0.0086, "R": 0.0001, "S": 0.0046, "T": 0.003, "V": 0.0057 },
    "12": { "A": 0.0077, "C": 0.0004, "D": 0.0001, "F": 0.001, "H": 0.0002, "L": 0.0004, "N": 0.0001, "P": 0.0126, "S": 0.973, "T": 0.0021, "V": 0.0001, "Y": 0.0025 },
    "13": { "A": 0.9758, "D": 0.0034, "E": 0.0001, "G": 0.0023, "H": 0.0001, "L": 0.0001, "P": 0.0088, "S": 0.0065, "T": 0.0013, "V": 0.0018 },
    "14": { "A": 0.0045, "C": 0.0002, "F": 0.0032, "H": 0.0002, "L": 0.0001, "N": 0.0002, "P": 0.0183, "S": 0.9671, "T": 0.0018, "Y": 0.0044 },
    "15": { "F": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.978, "M": 0.0026, "P": 0.0106, "Q": 0.0012, "R": 0.0055, "S": 0.0003, "T": 0.0002, "V": 0.0014 },
    "16": { "A": 0.0203, "C": 0.0016, "D": 0.0008, "E": 0.0018, "G": 0.943, "H": 0.0001, "K": 0.0001, "L": 0.0005, "P": 0.0006, "Q": 0.0001, "R": 0.0166, "S": 0.0005, "T": 0.0002, "V": 0.0139, "W": 0.0001 },
    "17": { "A": 0.7884, "C": 0.0001, "D": 0.0065, "E": 0.0001, "F": 0.0003, "G": 0.0019, "H": 0.0003, "L": 0.0002, "M": 0.0001, "N": 0.0001, "P": 0.015, "S": 0.1805, "T": 0.0039, "V": 0.0016, "Y": 0.0008 },
    "18": { "A": 0.01, "C": 0.0001, "E": 0.0001, "F": 0.0001, "G": 0.0001, "L": 0.002, "P": 0.0192, "Q": 0.0004, "S": 0.9629, "T": 0.0034, "V": 0.0001, "W": 0.0014, "Y": 0.0001 },
    "19": { "A": 0.0093, "C": 0.0001, "D": 0.0014, "E": 0.0001, "F": 0.0044, "G": 0.0066, "I": 0.0037, "L": 0.0048, "P": 0.0002, "S": 0.0004, "T": 0.0001, "V": 0.969 },
    "20": { "D": 0.0001, "E": 0.0013, "H": 0.0004, "I": 0.0001, "K": 0.9119, "M": 0.001, "N": 0.0204, "P": 0.0006, "Q": 0.0134, "R": 0.0153, "S": 0.0004, "T": 0.035, "Y": 0.0001 },
    "21": { "F": 0.0067, "H": 0.0027, "I": 0.0106, "L": 0.9419, "P": 0.0238, "R": 0.0074, "S": 0.0003, "T": 0.0004, "V": 0.0062 },
    "22": { "A": 0.0011, "H": 0.0003, "I": 0.0012, "K": 0.0001, "L": 0.0001, "N": 0.0028, "P": 0.016, "S": 0.0027, "T": 0.9757 },
    "23": { "A": 0.0002, "C": 0.9783, "D": 0.0001, "F": 0.0029, "G": 0.0043, "H": 0.0001, "L": 0.0002, "R": 0.0051, "S": 0.0055, "T": 0.0003, "V": 0.0001, "W": 0.0003, "Y": 0.0027 },
    "24": { "A": 0.0023, "C": 0.0001, "F": 0.0002, "H": 0.0003, "I": 0.0037, "L": 0.0003, "M": 0.0001, "N": 0.0018, "P": 0.0287, "Q": 0.0001, "R": 0.0001, "S": 0.0086, "T": 0.9533, "V": 0.0003, "Y": 0.0001 },
    "25": { "A": 0.0001, "C": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0002, "L": 0.9779, "M": 0.0025, "P": 0.007, "Q": 0.0009, "R": 0.0082, "S": 0.0001, "T": 0.0001, "V": 0.0027 },
    "26": { "A": 0.0006, "C": 0.0014, "D": 0.0056, "G": 0.0066, "H": 0.0003, "I": 0.0036, "K": 0.0001, "L": 0.0004, "N": 0.0116, "P": 0.0008, "R": 0.0322, "S": 0.9225, "T": 0.0144, "Y": 0.0001 },
    "27": { "A": 0.0011, "C": 0.0013, "D": 0.0004, "F": 0.0006, "G": 0.0085, "H": 0.0001, "I": 0.0026, "L": 0.0001, "M": 0.0001, "N": 0.0029, "P": 0.0046, "R": 0.0263, "S": 0.9475, "T": 0.0037, "V": 0.0001, "W": 0.0001 },
    "28": { "A": 0.0079, "C": 0.0006, "D": 0.0039, "E": 0.0098, "G": 0.9453, "H": 0.0001, "I": 0.0001, "L": 0.0002, "N": 0.0001, "P": 0.0002, "Q": 0.0004, "R": 0.0136, "S": 0.0011, "V": 0.0061, "W": 0.0107 },
    "29": { "A": 0.0003, "D": 0.0009, "E": 0.0006, "F": 0.0016, "H": 0.9494, "I": 0.0001, "K": 0.0001, "L": 0.0034, "N": 0.0057, "P": 0.0189, "Q": 0.0046, "R": 0.0031, "S": 0.001, "T": 0.0004, "Y": 0.0098 },
    "30": { "A": 0.0009, "C": 0.0006, "D": 0.0011, "E": 0.0001, "F": 0.0004, "G": 0.0049, "H": 0.0004, "I": 0.0086, "K": 0.0009, "L": 0.0006, "M": 0.0002, "N": 0.0153, "P": 0.0011, "Q": 0.0006, "R": 0.0249, "S": 0.9169, "T": 0.0209, "V": 0.0004, "Y": 0.0011 },
    "36": { "A": 0.0031, "C": 0.0013, "D": 0.0075, "E": 0.0005, "F": 0.0006, "G": 0.015, "H": 0.0051, "I": 0.0028, "K": 0.0011, "L": 0.0014, "M": 0.0002, "N": 0.0525, "P": 0.0027, "R": 0.0309, "S": 0.8272, "T": 0.0443, "V": 0.001, "W": 0.0001, "Y": 0.0026 },
    "37": { "A": 0.0004, "C": 0.0014, "D": 0.0092, "E": 0.0044, "F": 0.0077, "H": 0.0158, "I": 0.0003, "K": 0.001, "L": 0.0008, "N": 0.0241, "P": 0.0006, "Q": 0.0004, "R": 0.0005, "S": 0.0206, "T": 0.0012, "V": 0.0001, "Y": 0.9117 },
    "38": { "A": 0.7571, "C": 0.0001, "D": 0.0084, "F": 0.0022, "G": 0.0064, "H": 0.0012, "I": 0.1706, "K": 0.0025, "L": 0.0029, "M": 0.0006, "N": 0.0006, "P": 0.0129, "Q": 0.0001, "R": 0.0001, "S": 0.009, "T": 0.0112, "V": 0.0137, "Y": 0.0002 },
    "39": { "A": 0.0005, "C": 0.0001, "D": 0.0001, "F": 0.001, "G": 0.0001, "H": 0.0003, "I": 0.9414, "L": 0.0181, "M": 0.0004, "N": 0.0027, "P": 0.0001, "R": 0.0005, "S": 0.0082, "T": 0.0091, "V": 0.0176, "Y": 0.0001 },
    "40": { "A": 0.9566, "D": 0.0041, "E": 0.0065, "G": 0.0022, "H": 0.0004, "K": 0.0001, "P": 0.0119, "Q": 0.0004, "R": 0.0001, "S": 0.0056, "T": 0.0077, "V": 0.0044 },
    "41": { "A": 0.0002, "C": 0.0139, "F": 0.0001, "G": 0.0069, "L": 0.0066, "M": 0.0001, "P": 0.0001, "R": 0.0084, "S": 0.0053, "V": 0.0005, "W": 0.9579 },
    "42": { "A": 0.0001, "D": 0.0006, "F": 0.0001, "H": 0.9398, "I": 0.0001, "K": 0.0001, "L": 0.0131, "N": 0.003, "P": 0.0223, "Q": 0.0103, "R": 0.0018, "S": 0.0003, "T": 0.0008, "Y": 0.0076 },
    "43": { "A": 0.0001, "E": 0.0004, "F": 0.0001, "H": 0.0108, "K": 0.0037, "L": 0.0015, "M": 0.0001, "N": 0.0002, "P": 0.0249, "Q": 0.9532, "R": 0.0035, "S": 0.0003, "T": 0.0011, "V": 0.0001, "Y": 0.0001 },
    "44": { "E": 0.0006, "G": 0.0001, "H": 0.0177, "K": 0.0081, "L": 0.0052, "M": 0.0001, "N": 0.0003, "P": 0.0107, "Q": 0.9494, "R": 0.0058, "S": 0.0001, "T": 0.0004, "V": 0.0015, "Y": 0.0001 },
    "45": { "A": 0.0004, "C": 0.0001, "E": 0.0022, "G": 0.0046, "H": 0.0136, "K": 0.0044, "L": 0.0029, "N": 0.0001, "P": 0.0126, "Q": 0.9391, "R": 0.0192, "S": 0.0004, "T": 0.0001, "V": 0.0003, "Y": 0.0002 },
    "46": { "A": 0.0057, "H": 0.0008, "K": 0.0001, "L": 0.0008, "P": 0.9725, "Q": 0.0058, "R": 0.0013, "S": 0.0079, "T": 0.0048, "V": 0.0001 },
    "47": { "A": 0.0167, "D": 0.0161, "E": 0.7389, "G": 0.2019, "H": 0.0006, "K": 0.0038, "L": 0.0006, "N": 0.0006, "P": 0.0005, "Q": 0.015, "R": 0.0008, "S": 0.0002, "T": 0.0004, "V": 0.003, "W": 0.0006, "Y": 0.0004 },
    "48": { "E": 0.0037, "G": 0.0003, "H": 0.0002, "I": 0.0001, "K": 0.9459, "M": 0.0013, "N": 0.0079, "P": 0.0004, "Q": 0.0119, "R": 0.0132, "S": 0.0003, "T": 0.015 },
    "49": { "A": 0.1816, "C": 0.0037, "D": 0.0038, "E": 0.0001, "G": 0.798, "H": 0.0001, "I": 0.0001, "N": 0.0002, "P": 0.0006, "R": 0.0039, "S": 0.0029, "T": 0.0006, "V": 0.0045, "Y": 0.0001 },
    "50": { "A": 0.0014, "F": 0.0001, "H": 0.0065, "L": 0.0029, "N": 0.0001, "P": 0.9787, "Q": 0.0004, "R": 0.0013, "S": 0.0011, "T": 0.0077 },
    "51": { "A": 0.0001, "C": 0.0001, "G": 0.0014, "H": 0.0013, "K": 0.0001, "L": 0.0043, "P": 0.002, "Q": 0.0045, "R": 0.9845, "S": 0.0001, "T": 0.0001, "V": 0.0001, "W": 0.0015 },
    "52": { "A": 0.0001, "C": 0.0029, "D": 0.0054, "E": 0.0001, "F": 0.0636, "H": 0.0045, "I": 0.0002, "K": 0.0001, "L": 0.0015, "N": 0.0013, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0068, "T": 0.0001, "V": 0.0011, "Y": 0.9119 },
    "53": { "A": 0.0001, "C": 0.0004, "F": 0.0049, "G": 0.0001, "I": 0.0011, "L": 0.968, "M": 0.0011, "P": 0.0004, "Q": 0.0001, "R": 0.0007, "S": 0.0024, "T": 0.0001, "V": 0.0129, "W": 0.0077, "Y": 0.0001 },
    "54": { "F": 0.0001, "G": 0.0001, "I": 0.0067, "K": 0.0009, "L": 0.0187, "M": 0.9618, "N": 0.0001, "P": 0.0001, "R": 0.0058, "S": 0.0002, "T": 0.0025, "V": 0.0031 },
    "55": { "A": 0.0001, "C": 0.0001, "D": 0.0002, "E": 0.0046, "F": 0.0001, "G": 0.0008, "H": 0.0007, "I": 0.0063, "K": 0.9087, "L": 0.0006, "M": 0.0069, "N": 0.0145, "P": 0.0001, "Q": 0.0097, "R": 0.0252, "S": 0.0034, "T": 0.0172, "V": 0.0005, "W": 0.0001, "Y": 0.0003 },
    "56": { "A": 0.0004, "D": 0.0001, "F": 0.0039, "G": 0.0013, "H": 0.0014, "I": 0.0202, "L": 0.8626, "M": 0.0001, "P": 0.0013, "Q": 0.0001, "R": 0.0075, "S": 0.0004, "V": 0.1008 },
    "57": { "A": 0.0011, "D": 0.0117, "E": 0.1746, "F": 0.0003, "G": 0.0044, "H": 0.0067, "I": 0.0014, "K": 0.0153, "M": 0.0001, "N": 0.7613, "P": 0.0001, "Q": 0.0006, "R": 0.0008, "S": 0.0049, "T": 0.0093, "V": 0.0006, "Y": 0.0069 },
    "58": { "A": 0.0028, "C": 0.0008, "D": 0.0034, "E": 0.0001, "F": 0.0004, "G": 0.1767, "H": 0.0006, "I": 0.0034, "K": 0.0006, "L": 0.0006, "M": 0.0007, "N": 0.0117, "P": 0.002, "R": 0.0225, "S": 0.7604, "T": 0.0105, "V": 0.0022, "Y": 0.0006 },
    "59": { "A": 0.0067, "C": 0.0004, "D": 0.7878, "E": 0.0069, "F": 0.0001, "G": 0.0115, "H": 0.0015, "I": 0.0005, "K": 0.0001, "N": 0.0048, "R": 0.0021, "S": 0.1705, "T": 0.0007, "V": 0.0009, "W": 0.0001, "Y": 0.0053 },
    "63": { "A": 0.0028, "C": 0.0035, "D": 0.0032, "E": 0.0002, "F": 0.0001, "G": 0.9794, "H": 0.0001, "I": 0.0025, "L": 0.0001, "N": 0.0001, "R": 0.0017, "S": 0.0041, "V": 0.002, "Y": 0.0001 },
    "64": { "A": 0.0008, "C": 0.001, "D": 0.0028, "E": 0.0025, "F": 0.0004, "G": 0.0079, "H": 0.0001, "I": 0.0036, "K": 0.0007, "L": 0.0001, "M": 0.0005, "N": 0.0058, "P": 0.0004, "R": 0.0272, "S": 0.9211, "T": 0.0215, "V": 0.0031, "Y": 0.0004 },
    "65": { "C": 0.0006, "D": 0.0023, "E": 0.0002, "F": 0.0017, "G": 0.0029, "H": 0.7752, "I": 0.0004, "K": 0.0003, "L": 0.0034, "N": 0.0095, "P": 0.0064, "Q": 0.0103, "R": 0.003, "S": 0.0016, "T": 0.0004, "V": 0.0004, "Y": 0.1814 },
    "66": { "A": 0.0017, "C": 0.001, "D": 0.0033, "E": 0.0006, "F": 0.0036, "G": 0.0027, "H": 0.0014, "I": 0.0272, "K": 0.008, "L": 0.0006, "M": 0.0008, "N": 0.2122, "P": 0.0012, "Q": 0.0004, "R": 0.0209, "S": 0.6474, "T": 0.0624, "V": 0.0017, "W": 0.0001, "Y": 0.0029 },
    "67": { "A": 0.0001, "D": 0.0002, "E": 0.004, "I": 0.0001, "K": 0.9425, "M": 0.0007, "N": 0.0048, "Q": 0.0105, "R": 0.0259, "S": 0.0004, "T": 0.0107, "W": 0.0001 },
    "68": { "A": 0.0053, "C": 0.0002, "D": 0.0003, "E": 0.0027, "G": 0.979, "I": 0.0004, "K": 0.0001, "L": 0.0001, "R": 0.0032, "S": 0.0001, "V": 0.0039, "W": 0.0048 },
    "69": { "A": 0.0125, "C": 0.0006, "D": 0.7854, "E": 0.0033, "F": 0.0001, "G": 0.0099, "H": 0.0015, "I": 0.0011, "K": 0.0001, "L": 0.0001, "N": 0.0032, "P": 0.0001, "R": 0.0022, "S": 0.1703, "T": 0.0013, "V": 0.0046, "W": 0.0001, "Y": 0.0037 },
    "70": { "A": 0.0037, "C": 0.0002, "D": 0.0026, "E": 0.0126, "G": 0.9645, "H": 0.0001, "K": 0.0008, "L": 0.0001, "P": 0.0001, "Q": 0.0002, "R": 0.0027, "S": 0.0003, "T": 0.0001, "V": 0.0087, "W": 0.0033 },
    "71": { "A": 0.0002, "C": 0.0003, "D": 0.0002, "F": 0.0041, "G": 0.0014, "H": 0.0001, "I": 0.7778, "K": 0.0001, "L": 0.011, "M": 0.0016, "N": 0.002, "P": 0.0001, "R": 0.0004, "S": 0.0082, "T": 0.005, "V": 0.1875, "Y": 0.0001 },
    "72": { "A": 0.0017, "H": 0.0027, "L": 0.0015, "P": 0.9872, "Q": 0.0001, "R": 0.0006, "S": 0.0048, "T": 0.0015 },
    "74": { "A": 0.0135, "C": 0.0005, "D": 0.9392, "E": 0.0151, "F": 0.0003, "G": 0.0078, "H": 0.0037, "I": 0.0001, "L": 0.0002, "N": 0.0027, "P": 0.0001, "Q": 0.0003, "R": 0.0001, "S": 0.0006, "T": 0.0002, "V": 0.0091, "Y": 0.0065 },
    "75": { "C": 0.0026, "G": 0.001, "H": 0.0013, "L": 0.0053, "P": 0.0019, "R": 0.9866, "S": 0.0013 },
    "76": { "C": 0.0086, "F": 0.9733, "G": 0.0001, "I": 0.0002, "L": 0.0042, "M": 0.0001, "S": 0.0026, "T": 0.0001, "V": 0.0101, "W": 0.0001, "Y": 0.0006 },
    "77": { "A": 0.012, "F": 0.0001, "I": 0.0001, "L": 0.0026, "P": 0.0032, "Q": 0.0001, "S": 0.9792, "T": 0.0022, "V": 0.0001, "W": 0.0001, "Y": 0.0004 },
    "78": { "A": 0.0094, "C": 0.0055, "D": 0.0013, "G": 0.973, "R": 0.0012, "S": 0.0015, "V": 0.0081 },
    "79": { "A": 0.0098, "C": 0.0015, "D": 0.0001, "F": 0.0026, "H": 0.0001, "L": 0.0028, "N": 0.0001, "P": 0.0026, "S": 0.9759, "T": 0.001, "V": 0.0001, "W": 0.0001, "Y": 0.0033 },
    "80": { "A": 0.0022, "C": 0.0009, "D": 0.0001, "F": 0.0001, "G": 0.0083, "I": 0.0036, "K": 0.0005, "L": 0.0002, "M": 0.0001, "N": 0.003, "P": 0.0001, "R": 0.0154, "S": 0.9535, "T": 0.0112, "V": 0.0005, "W": 0.0001 },
    "83": { "A": 0.0169, "C": 0.0006, "F": 0.0015, "L": 0.0003, "P": 0.0018, "R": 0.0001, "S": 0.9741, "T": 0.0012, "V": 0.0002, "W": 0.0002, "Y": 0.0031 },
    "84": { "A": 0.0054, "C": 0.0004, "D": 0.0001, "E": 0.0017, "G": 0.9793, "K": 0.0001, "L": 0.0001, "M": 0.0001, "R": 0.0018, "S": 0.0002, "T": 0.0001, "V": 0.0066, "W": 0.0042 },
    "85": { "A": 0.9572, "C": 0.0001, "D": 0.0028, "F": 0.0001, "G": 0.0028, "H": 0.0001, "L": 0.0029, "N": 0.0006, "P": 0.0032, "S": 0.0087, "T": 0.0202, "V": 0.0013 },
    "86": { "A": 0.0208, "C": 0.0001, "D": 0.1844, "E": 0.7684, "G": 0.0159, "H": 0.0005, "I": 0.0001, "K": 0.0007, "N": 0.0031, "Q": 0.0018, "S": 0.0006, "T": 0.0006, "V": 0.0016, "W": 0.0006, "Y": 0.0008 },
    "87": { "A": 0.0006, "C": 0.0033, "F": 0.0004, "G": 0.0012, "H": 0.0059, "L": 0.0053, "N": 0.0001, "P": 0.0023, "Q": 0.0001, "R": 0.9774, "S": 0.0016, "T": 0.0001, "W": 0.0006, "Y": 0.0011 },
    "88": { "A": 0.0001, "C": 0.002, "D": 0.0048, "F": 0.0079, "H": 0.0092, "I": 0.0004, "L": 0.0002, "N": 0.0017, "P": 0.0002, "Q": 0.0001, "R": 0.0001, "S": 0.0207, "T": 0.0013, "V": 0.0001, "W": 0.0002, "Y": 0.9508 },
    "89": { "F": 0.0011, "H": 0.0011, "I": 0.0031, "L": 0.9793, "P": 0.0043, "Q": 0.0001, "R": 0.0098, "S": 0.0001, "T": 0.0005, "V": 0.0008 },
    "90": { "A": 0.0063, "D": 0.0001, "F": 0.0011, "G": 0.0001, "H": 0.0006, "I": 0.0199, "K": 0.0001, "L": 0.0009, "M": 0.0001, "N": 0.0018, "P": 0.0106, "R": 0.0002, "S": 0.0118, "T": 0.9465 },
    "91": { "A": 0.0001, "C": 0.0001, "F": 0.0011, "G": 0.0001, "H": 0.0001, "I": 0.9654, "L": 0.0105, "M": 0.0011, "N": 0.0011, "P": 0.0001, "R": 0.0005, "S": 0.0128, "T": 0.0029, "V": 0.0043 },
    "92": { "A": 0.0239, "C": 0.0004, "F": 0.0015, "G": 0.0001, "H": 0.0001, "K": 0.0027, "L": 0.0001, "N": 0.0001, "P": 0.005, "Q": 0.0001, "R": 0.0001, "S": 0.9572, "T": 0.0045, "V": 0.0001, "Y": 0.0041 },
    "93": { "A": 0.0001, "C": 0.0018, "D": 0.0018, "E": 0.0001, "G": 0.008, "H": 0.0017, "I": 0.0044, "K": 0.0006, "L": 0.0003, "N": 0.1787, "P": 0.0001, "R": 0.0143, "S": 0.7832, "T": 0.0046, "V": 0.0001, "Y": 0.0002 },
    "94": { "A": 0.0001, "C": 0.0001, "F": 0.0011, "G": 0.0002, "H": 0.0025, "I": 0.0048, "L": 0.9622, "N": 0.0001, "P": 0.0027, "Q": 0.0001, "R": 0.0166, "S": 0.0003, "T": 0.0002, "V": 0.0092 },
    "95": { "A": 0.0001, "E": 0.002, "H": 0.0058, "K": 0.0024, "L": 0.0008, "P": 0.003, "Q": 0.9809, "R": 0.0048, "S": 0.0001, "V": 0.0001, "Y": 0.0001 },
    "96": { "A": 0.0293, "C": 0.0011, "D": 0.0007, "E": 0.0028, "F": 0.0493, "G": 0.0001, "H": 0.0001, "L": 0.0015, "N": 0.0001, "P": 0.0047, "S": 0.9038, "T": 0.001, "V": 0.0015, "Y": 0.0041 },
    "97": { "A": 0.0029, "D": 0.0337, "E": 0.9581, "G": 0.0025, "H": 0.0001, "K": 0.0005, "M": 0.0001, "Q": 0.0004, "R": 0.0001, "V": 0.0014, "W": 0.0001, "Y": 0.0001 },
    "98": { "A": 0.0024, "D": 0.9867, "E": 0.0061, "G": 0.0011, "H": 0.0001, "L": 0.0001, "N": 0.0006, "V": 0.001, "Y": 0.002 },
    "99": { "A": 0.0018, "D": 0.0092, "E": 0.9842, "G": 0.0027, "K": 0.0003, "N": 0.0001, "Q": 0.0004, "R": 0.0001, "V": 0.0007, "W": 0.0001, "Y": 0.0004 },
    "100": { "A": 0.9644, "C": 0.0001, "D": 0.0059, "E": 0.0001, "G": 0.0147, "I": 0.0001, "N": 0.0001, "P": 0.0008, "R": 0.0001, "S": 0.0098, "T": 0.0022, "V": 0.0012, "Y": 0.0006 },
    "101": { "A": 0.0034, "D": 0.9645, "E": 0.0069, "F": 0.0001, "G": 0.0012, "H": 0.0008, "I": 0.0004, "K": 0.0001, "L": 0.0001, "N": 0.0016, "R": 0.0001, "S": 0.0002, "T": 0.0008, "V": 0.0047, "Y": 0.0151 },
    "102": { "A": 0.0001, "C": 0.0013, "D": 0.0044, "E": 0.0008, "F": 0.0018, "H": 0.0003, "N": 0.0004, "P": 0.0001, "Q": 0.0001, "S": 0.0022, "V": 0.0001, "W": 0.0001, "Y": 0.9883 },
    "103": { "C": 0.0038, "D": 0.0079, "F": 0.0103, "G": 0.0001, "H": 0.0057, "I": 0.0008, "L": 0.0003, "N": 0.0007, "P": 0.0001, "R": 0.0001, "S": 0.0038, "T": 0.0001, "V": 0.0003, "W": 0.0002, "Y": 0.9661 },
    "104": { "C": 0.9727, "F": 0.0024, "G": 0.0081, "L": 0.0002, "R": 0.0009, "S": 0.0029, "V": 0.0002, "W": 0.0114, "Y": 0.0011 },
    "105": { "A": 0.0011, "C": 0.0002, "D": 0.002, "E": 0.1747, "G": 0.0038, "H": 0.0048, "K": 0.0018, "L": 0.0006, "M": 0.0001, "P": 0.0029, "Q": 0.8056, "R": 0.0015, "S": 0.0004, "T": 0.0001, "V": 0.0001, "W": 0.0001, "Y": 0.0001 },
    "106": { "A": 0.0127, "C": 0.0001, "D": 0.0004, "E": 0.0001, "F": 0.0001, "G": 0.0003, "H": 0.0001, "I": 0.0015, "K": 0.0003, "L": 0.0002, "M": 0.0031, "N": 0.0057, "P": 0.0009, "Q": 0.0001, "R": 0.0001, "S": 0.0081, "T": 0.9662, "V": 0.0001 },
    "107": { "A": 0.0002, "C": 0.0042, "D": 0.0027, "E": 0.0002, "F": 0.0002, "G": 0.0194, "H": 0.0001, "K": 0.0001, "L": 0.0037, "R": 0.0027, "S": 0.0016, "T": 0.0001, "V": 0.0002, "W": 0.9637, "Y": 0.0011 },
    "108": { "A": 0.0118, "C": 0.0035, "D": 0.2052, "E": 0.0018, "F": 0.0002, "G": 0.7514, "H": 0.0031, "I": 0.0046, "L": 0.0002, "N": 0.0019, "Q": 0.0001, "R": 0.0015, "S": 0.0039, "T": 0.0024, "V": 0.0072, "W": 0.0001, "Y": 0.0011 },
    "109": { "A": 0.0123, "D": 0.0014, "F": 0.0017, "G": 0.0098, "H": 0.0115, "I": 0.0058, "K": 0.0013, "L": 0.0013, "M": 0.0014, "N": 0.0044, "P": 0.0095, "Q": 0.0005, "R": 0.0035, "S": 0.1902, "T": 0.7447, "V": 0.0005, "Y": 0.0001 },
    "110": { "C": 0.0025, "D": 0.015, "G": 0.59, "I": 0.0075, "K": 0.0025, "N": 0.2275, "R": 0.005, "S": 0.1225, "T": 0.015, "V": 0.0025, "Y": 0.01 },
    "113": { "A": 0.0176, "C": 0.0012, "D": 0.026, "E": 0.0003, "F": 0.0022, "G": 0.6818, "H": 0.0043, "I": 0.0545, "K": 0.0012, "L": 0.0012, "M": 0.0012, "N": 0.1495, "P": 0.0028, "Q": 0.0006, "R": 0.0022, "S": 0.0102, "T": 0.0337, "V": 0.0056, "W": 0.0006, "Y": 0.0031 },
    "114": { "A": 0.0094, "C": 0.0017, "D": 0.0106, "E": 0.0011, "F": 0.0075, "G": 0.5411, "H": 0.0086, "I": 0.1211, "K": 0.0026, "L": 0.0086, "M": 0.0031, "N": 0.1224, "P": 0.0155, "Q": 0.0049, "R": 0.0076, "S": 0.0207, "T": 0.1037, "V": 0.0077, "W": 0.0005, "Y": 0.0015 },
    "115": { "A": 0.0101, "C": 0.0013, "D": 0.0106, "E": 0.0044, "F": 0.0334, "G": 0.0665, "H": 0.0778, "I": 0.3565, "K": 0.0073, "L": 0.0409, "M": 0.015, "N": 0.0321, "P": 0.0412, "Q": 0.0242, "R": 0.0303, "S": 0.0539, "T": 0.1539, "V": 0.0188, "W": 0.0009, "Y": 0.0211 },
    "116": { "A": 0.0149, "C": 0.0016, "D": 0.0015, "E": 0.0117, "F": 0.0008, "G": 0.0888, "H": 0.0282, "I": 0.0096, "K": 0.0101, "L": 0.0259, "M": 0.0143, "N": 0.0028, "P": 0.015, "Q": 0.106, "R": 0.1758, "S": 0.0037, "T": 0.006, "V": 0.2412, "W": 0.218, "Y": 0.0242 },
    "117": { "A": 0.0034, "C": 0.0001, "D": 0.0001, "E": 0.0025, "F": 0.0019, "G": 0.0035, "H": 0.0006, "I": 0.0139, "K": 0.0006, "L": 0.0183, "M": 0.0035, "P": 0.0008, "Q": 0.0003, "R": 0.0022, "S": 0.0011, "T": 0.0032, "V": 0.9438, "W": 0.0002 },
    "118": { "C": 0.0027, "F": 0.9848, "G": 0.0003, "I": 0.0024, "K": 0.0001, "L": 0.003, "P": 0.0001, "R": 0.0004, "S": 0.0011, "V": 0.0046, "W": 0.0001, "Y": 0.0004 },
    "119": { "A": 0.0025, "C": 0.0012, "D": 0.0007, "E": 0.0001, "G": 0.993, "K": 0.0001, "P": 0.0001, "R": 0.0006, "S": 0.0008, "V": 0.001 },
    "120": { "A": 0.002, "D": 0.0002, "E": 0.0014, "F": 0.0001, "G": 0.9157, "I": 0.0002, "N": 0.0002, "P": 0.002, "R": 0.0043, "S": 0.0059, "T": 0.067, "V": 0.0008, "W": 0.0001 },
    "121": { "A": 0.0004, "E": 0.0004, "G": 0.9946, "R": 0.0006, "V": 0.0019, "W": 0.0021 },
    "122": { "A": 0.001, "C": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0009, "N": 0.0006, "P": 0.0013, "R": 0.0001, "S": 0.0016, "T": 0.9944 },
    "123": { "D": 0.0001, "E": 0.0014, "H": 0.0003, "I": 0.0001, "K": 0.9573, "L": 0.0001, "M": 0.0004, "N": 0.0028, "P": 0.0002, "Q": 0.0166, "R": 0.014, "S": 0.0008, "T": 0.0059, "W": 0.0001, "Y": 0.0001 },
    "124": { "A": 0.0001, "F": 0.0004, "G": 0.0001, "I": 0.0007, "L": 0.9044, "M": 0.0005, "P": 0.0006, "Q": 0.0004, "R": 0.0019, "S": 0.0001, "V": 0.0909 },
    "125": { "A": 0.0017, "D": 0.0001, "I": 0.0019, "K": 0.0002, "N": 0.0009, "P": 0.0026, "S": 0.0062, "T": 0.9861, "V": 0.0002 },
    "126": { "A": 0.001, "C": 0.0001, "D": 0.0006, "F": 0.0003, "G": 0.0071, "I": 0.0011, "L": 0.0006, "R": 0.0003, "V": 0.989 },
    "127": { "I": 0.0012, "L": 0.9956, "P": 0.0003, "V": 0.0029 }
  },
  "IGLV5": {
    "1": { "H": 0.0103, "Q": 0.9897 },
    "2": { "A": 0.6908, "C": 0.0001, "D": 0.0011, "G": 0.0011, "H": 0.001, "L": 0.0049, "P": 0.2884, "R": 0.0001, "S": 0.0065, "T": 0.0054, "V": 0.0006 },
    "3": { "A": 0.0075, "C": 0.0001, "D": 0.0001, "E": 0.001, "F": 0.0002, "G": 0.0033, "I": 0.0009, "L": 0.0086, "M": 0.0009, "P": 0.0001, "S": 0.0002, "V": 0.9769 },
    "4": { "F": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.979, "M": 0.0023, "P": 0.0026, "Q": 0.0004, "R": 0.0048, "S": 0.0001, "V": 0.0105, "W": 0.0001 },
    "5": { "A": 0.0023, "I": 0.0018, "K": 0.0001, "L": 0.0003, "N": 0.0019, "P": 0.0295, "R": 0.0005, "S": 0.0056, "T": 0.9567, "V": 0.0014 },
    "6": { "D": 0.0001, "E": 0.0006, "H": 0.0103, "K": 0.0027, "L": 0.0025, "N": 0.0001, "P": 0.0237, "Q": 0.9578, "R": 0.0018, "T": 0.0003, "W": 0.0001 },
    "7": { "A": 0.0001, "H": 0.0001, "L": 0.0038, "P": 0.9863, "Q": 0.0028, "R": 0.0011, "S": 0.0026, "T": 0.0031 },
    "8": { "A": 0.1465, "C": 0.0001, "D": 0.0002, "E": 0.0001, "F": 0.0013, "H": 0.0004, "I": 0.0001, "L": 0.0006, "N": 0.0004, "P": 0.1623, "R": 0.0006, "S": 0.6093, "T": 0.0754, "V": 0.0009, "Y": 0.0018 },
    "9": { "A": 0.0035, "C": 0.0008, "F": 0.003, "H": 0.0003, "L": 0.0003, "P": 0.0074, "S": 0.978, "T": 0.0008, "Y": 0.0059 },
    "11": { "A": 0.0006, "C": 0.0001, "D": 0.0001, "F": 0.0019, "G": 0.0001, "H": 0.0568, "I": 0.0043, "L": 0.7679, "M": 0.0001, "P": 0.013, "Q": 0.0002, "R": 0.0043, "S": 0.148, "T": 0.0002, "V": 0.0017, "Y": 0.0009 },
    "12": { "A": 0.0057, "C": 0.0002, "F": 0.0013, "H": 0.0006, "L": 0.0006, "N": 0.0001, "P": 0.0227, "R": 0.0001, "S": 0.9625, "T": 0.003, "Y": 0.0032 },
    "13": { "A": 0.9752, "D": 0.0002, "E": 0.0022, "G": 0.0009, "L": 0.0001, "P": 0.0094, "Q": 0.0001, "S": 0.009, "T": 0.0018, "V": 0.0011 },
    "14": { "A": 0.0038, "C": 0.0004, "F": 0.0019, "G": 0.0001, "H": 0.0001, "L": 0.0002, "P": 0.0122, "Q": 0.0001, "R": 0.0001, "S": 0.9771, "T": 0.0014, "Y": 0.0026 },
    "15": { "A": 0.0014, "H": 0.0026, "L": 0.0042, "P": 0.9272, "Q": 0.0001, "R": 0.0009, "S": 0.0585, "T": 0.005, "Y": 0.0001 },
    "16": { "A": 0.0065, "C": 0.0004, "D": 0.0006, "E": 0.0014, "G": 0.9795, "L": 0.0001, "P": 0.0001, "Q": 0.0003, "R": 0.0033, "S": 0.0001, "V": 0.0078 },
    "17": { "A": 0.8243, "D": 0.0106, "E": 0.1344, "G": 0.0009, "H": 0.0003, "K": 0.0004, "L": 0.0001, "N": 0.0001, "P": 0.0081, "Q": 0.0009, "R": 0.0009, "S": 0.0085, "T": 0.007, "V": 0.0033, "Y": 0.0001 },
    "18": { "A": 0.0086, "D": 0.0001, "E": 0.0001, "F": 0.0001, "G": 0.0001, "I": 0.0001, "L": 0.0017, "P": 0.0046, "Q": 0.0001, "S": 0.981, "T": 0.0026, "V": 0.0001, "Y": 0.001 },
    "19": { "A": 0.9157, "D": 0.0038, "F": 0.0004, "G": 0.0011, "H": 0.0001, "I": 0.0001, "L": 0.0003, "P": 0.0072, "S": 0.0056, "T": 0.0031, "V": 0.0627 },
    "20": { "C": 0.0009, "G": 0.0023, "H": 0.0001, "I": 0.0049, "K": 0.0005, "L": 0.0005, "M": 0.0001, "N": 0.0012, "P": 0.0001, "Q": 0.0001, "R": 0.299, "S": 0.683, "T": 0.0072, "V": 0.0001 },
    "21": { "A": 0.0001, "C": 0.0006, "F": 0.0712, "H": 0.0044, "I": 0.0043, "L": 0.8586, "M": 0.0001, "P": 0.0483, "R": 0.0091, "S": 0.001, "T": 0.0009, "V": 0.0013, "Y": 0.0002 },
    "22": { "A": 0.0035, "D": 0.0001, "G": 0.0001, "H": 0.0016, "I": 0.0012, "L": 0.0002, "N": 0.0038, "P": 0.0678, "Q": 0.0001, "R": 0.0001, "S": 0.0033, "T": 0.9183 },
    "23": { "A": 0.0002, "C": 0.962, "D": 0.0001, "F": 0.0061, "G": 0.0101, "H": 0.0001, "I": 0.0001, "L": 0.0002, "R": 0.0089, "S": 0.0093, "V": 0.0002, "W": 0.0003, "Y": 0.0025 },
    "24": { "A": 0.0063, "H": 0.0003, "I": 0.0028, "K": 0.0001, "L": 0.0022, "M": 0.0533, "N": 0.0016, "P": 0.03, "R": 0.0004, "S": 0.0036, "T": 0.8994 },
    "25": { "A": 0.0001, "C": 0.0004, "F": 0.0226, "I": 0.0007, "L": 0.9634, "M": 0.0009, "P": 0.0004, "Q": 0.0001, "R": 0.0001, "S": 0.004, "V": 0.0045, "W": 0.0025, "Y": 0.0001 },
    "26": { "A": 0.0001, "C": 0.0041, "D": 0.0004, "G": 0.0049, "H": 0.0134, "I": 0.0004, "L": 0.004, "N": 0.0007, "P": 0.1528, "Q": 0.0004, "R": 0.7596, "S": 0.0585, "T": 0.0007, "V": 0.0001, "Y": 0.0001 },
    "27": { "A": 0.0001, "C": 0.0015, "D": 0.0001, "G": 0.0035, "H": 0.0003, "I": 0.0052, "K": 0.0001, "L": 0.0006, "M": 0.0001, "N": 0.0015, "P": 0.0005, "R": 0.0293, "S": 0.9508, "T": 0.0061, "V": 0.0003, "W": 0.0001 },
    "28": { "A": 0.0103, "C": 0.0079, "D": 0.1716, "E": 0.0006, "G": 0.7954, "H": 0.0009, "N": 0.0011, "P": 0.0001, "R": 0.0034, "S": 0.0021, "T": 0.0001, "V": 0.0052, "Y": 0.0011 },
    "29": { "A": 0.0004, "C": 0.0004, "D": 0.0001, "F": 0.0711, "G": 0.0002, "H": 0.0001, "I": 0.8652, "L": 0.0326, "M": 0.0028, "N": 0.0008, "P": 0.0003, "R": 0.0004, "S": 0.0052, "T": 0.0064, "V": 0.0139, "Y": 0.0001 },
    "30": { "A": 0.0009, "C": 0.0001, "D": 0.0106, "E": 0.0001, "G": 0.0006, "H": 0.016, "I": 0.0026, "K": 0.0084, "L": 0.0004, "M": 0.0001, "N": 0.8735, "P": 0.0008, "Q": 0.0001, "R": 0.0025, "S": 0.0661, "T": 0.0133, "V": 0.0001, "Y": 0.0039 },
    "31": { "A": 0.0035, "C": 0.0003, "D": 0.0009, "F": 0.0067, "G": 0.0035, "H": 0.0001, "I": 0.0099, "L": 0.0068, "P": 0.0001, "S": 0.0004, "T": 0.0001, "V": 0.9677 },
    "35": { "A": 0.0315, "C": 0.0078, "D": 0.0229, "E": 0.0017, "F": 0.0001, "G": 0.8917, "H": 0.0009, "I": 0.0007, "L": 0.0007, "N": 0.0013, "P": 0.0006, "Q": 0.0001, "R": 0.0071, "S": 0.022, "T": 0.001, "V": 0.0084, "W": 0.0013, "Y": 0.0002 },
    "36": { "A": 0.015, "C": 0.0002, "D": 0.0638, "E": 0.0004, "F": 0.0006, "G": 0.0018, "H": 0.0013, "I": 0.0052, "K": 0.0011, "L": 0.001, "M": 0.0032, "N": 0.0133, "P": 0.0169, "Q": 0.0001, "R": 0.0048, "S": 0.1596, "T": 0.7093, "V": 0.0013, "Y": 0.0012 },
    "37": { "A": 0.0001, "C": 0.005, "D": 0.0031, "E": 0.0001, "F": 0.0661, "G": 0.0001, "H": 0.0058, "I": 0.0004, "K": 0.0009, "L": 0.0004, "N": 0.004, "P": 0.0001, "Q": 0.0012, "R": 0.0014, "S": 0.023, "T": 0.0006, "V": 0.0002, "Y": 0.8875 },
    "38": { "A": 0.0002, "C": 0.0006, "D": 0.0002, "E": 0.0003, "G": 0.0055, "H": 0.0019, "I": 0.0028, "K": 0.0145, "L": 0.0009, "M": 0.0084, "N": 0.149, "P": 0.0006, "Q": 0.0006, "R": 0.7247, "S": 0.0118, "T": 0.0123, "V": 0.0004, "W": 0.064, "Y": 0.0014 },
    "39": { "A": 0.0001, "F": 0.0002, "G": 0.0002, "H": 0.0001, "I": 0.9437, "K": 0.0028, "L": 0.0281, "M": 0.006, "N": 0.0004, "Q": 0.0004, "R": 0.0069, "S": 0.0004, "T": 0.0041, "V": 0.0067 },
    "40": { "A": 0.0001, "C": 0.0009, "D": 0.0046, "F": 0.0084, "H": 0.0097, "I": 0.0001, "L": 0.0016, "M": 0.0003, "N": 0.0038, "P": 0.0004, "R": 0.0558, "S": 0.0175, "T": 0.0007, "W": 0.0009, "Y": 0.8952 },
    "41": { "A": 0.0001, "C": 0.0114, "D": 0.0001, "F": 0.0001, "G": 0.0038, "L": 0.0093, "R": 0.0031, "S": 0.0029, "V": 0.0001, "W": 0.9691 },
    "42": { "A": 0.0001, "C": 0.0014, "D": 0.0033, "E": 0.0001, "F": 0.0119, "H": 0.005, "L": 0.0004, "N": 0.0015, "R": 0.0001, "S": 0.0112, "T": 0.0004, "V": 0.0001, "Y": 0.9646 },
    "43": { "A": 0.0001, "E": 0.0004, "G": 0.0001, "H": 0.0105, "K": 0.0067, "L": 0.0017, "M": 0.0001, "N": 0.0001, "P": 0.0111, "Q": 0.9639, "R": 0.0035, "S": 0.0001, "T": 0.0009, "V": 0.0006, "W": 0.0001, "Y": 0.0001 },
    "44": { "A": 0.0002, "E": 0.0006, "H": 0.0079, "I": 0.0001, "K": 0.0094, "L": 0.0014, "M": 0.0001, "N": 0.0003, "P": 0.0116, "Q": 0.9636, "R": 0.0035, "T": 0.0008, "V": 0.0004, "W": 0.0001, "Y": 0.0001 },
    "45": { "D": 0.0001, "E": 0.0016, "G": 0.0001, "H": 0.0007, "K": 0.9422, "M": 0.0016, "N": 0.0094, "P": 0.0002, "Q": 0.016, "R": 0.0128, "S": 0.0006, "T": 0.0146, "Y": 0.0001 },
    "46": { "A": 0.0091, "E": 0.0001, "H": 0.0012, "K": 0.0001, "L": 0.0018, "P": 0.9669, "Q": 0.0091, "R": 0.0005, "S": 0.0056, "T": 0.0054 },
    "47": { "A": 0.0018, "C": 0.0002, "D": 0.0006, "E": 0.003, "G": 0.9799, "H": 0.0001, "R": 0.0047, "V": 0.0052, "W": 0.0045 },
    "48": { "A": 0.0004, "C": 0.0012, "D": 0.0001, "E": 0.0001, "F": 0.0009, "G": 0.0041, "H": 0.0006, "I": 0.0038, "K": 0.0006, "L": 0.0001, "M": 0.0001, "N": 0.0592, "P": 0.0001, "Q": 0.0001, "R": 0.0167, "S": 0.9066, "T": 0.0052, "V": 0.0001 },
    "49": { "A": 0.0004, "F": 0.0001, "H": 0.0055, "I": 0.0001, "L": 0.0723, "P": 0.9148, "Q": 0.0003, "R": 0.0018, "S": 0.0006, "T": 0.0037, "V": 0.0003 },
    "50": { "A": 0.0006, "F": 0.0009, "H": 0.0053, "L": 0.0009, "P": 0.9858, "R": 0.0009, "S": 0.0011, "T": 0.0047 },
    "51": { "C": 0.0001, "E": 0.0006, "G": 0.0003, "H": 0.0133, "K": 0.0047, "L": 0.0036, "M": 0.0004, "N": 0.0002, "P": 0.0051, "Q": 0.6765, "R": 0.2921, "S": 0.0021, "T": 0.0008, "W": 0.0002 },
    "52": { "A": 0.0003, "C": 0.002, "D": 0.0067, "E": 0.0006, "F": 0.0415, "H": 0.0043, "I": 0.0001, "K": 0.0002, "L": 0.002, "N": 0.0011, "P": 0.0001, "Q": 0.0004, "R": 0.0001, "S": 0.0148, "T": 0.0001, "V": 0.0006, "W": 0.0001, "Y": 0.925 },
    "53": { "D": 0.0001, "F": 0.0009, "H": 0.0055, "I": 0.0031, "L": 0.9523, "P": 0.0208, "R": 0.0125, "S": 0.0004, "T": 0.0002, "V": 0.0043 },
    "54": { "A": 0.0001, "F": 0.0001, "I": 0.0001, "L": 0.9829, "M": 0.0028, "P": 0.0028, "Q": 0.0013, "R": 0.007, "S": 0.0001, "V": 0.0028 },
    "55": { "A": 0.0001, "C": 0.0005, "D": 0.0006, "E": 0.0001, "F": 0.0015, "G": 0.0074, "H": 0.0008, "I": 0.0001, "K": 0.0181, "L": 0.0001, "M": 0.0039, "N": 0.0078, "P": 0.0001, "Q": 0.0003, "R": 0.7336, "S": 0.0141, "T": 0.0085, "V": 0.0004, "W": 0.0035, "Y": 0.1986 },
    "56": { "C": 0.0015, "D": 0.0043, "E": 0.0001, "F": 0.011, "H": 0.0033, "I": 0.0001, "L": 0.0002, "N": 0.0015, "P": 0.0001, "S": 0.006, "T": 0.0001, "V": 0.0002, "Y": 0.9714 },
    "57": { "A": 0.0001, "C": 0.0002, "D": 0.0018, "E": 0.0007, "F": 0.0038, "G": 0.0002, "H": 0.0575, "I": 0.0022, "K": 0.7411, "L": 0.0001, "M": 0.0006, "N": 0.0223, "P": 0.0007, "Q": 0.005, "R": 0.0171, "S": 0.0043, "T": 0.0063, "V": 0.0002, "W": 0.0001, "Y": 0.1357 },
    "58": { "A": 0.0108, "D": 0.0009, "F": 0.0009, "H": 0.0001, "I": 0.0001, "L": 0.0039, "P": 0.0057, "Q": 0.0001, "R": 0.0001, "S": 0.9714, "T": 0.0045, "V": 0.0011, "Y": 0.0004 },
    "59": { "A": 0.0081, "C": 0.0001, "D": 0.9643, "E": 0.0089, "G": 0.0032, "H": 0.0012, "I": 0.0001, "K": 0.0001, "N": 0.0022, "P": 0.0003, "Q": 0.0001, "R": 0.0002, "S": 0.0003, "T": 0.0001, "V": 0.0048, "Y": 0.006 },
    "63": { "A": 0.0169, "C": 0.0001, "D": 0.0004, "E": 0.0001, "F": 0.0011, "H": 0.0003, "K": 0.0004, "L": 0.0047, "P": 0.008, "Q": 0.0002, "S": 0.9623, "T": 0.0044, "W": 0.0001, "Y": 0.001 },
    "64": { "A": 0.0081, "D": 0.8973, "E": 0.0128, "F": 0.0004, "G": 0.0033, "H": 0.0039, "I": 0.0008, "K": 0.0009, "N": 0.0613, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0012, "T": 0.0021, "V": 0.0045, "Y": 0.0033 },
    "65": { "A": 0.0002, "D": 0.0008, "E": 0.0023, "F": 0.0002, "G": 0.0001, "H": 0.0013, "I": 0.0016, "K": 0.9248, "L": 0.0003, "M": 0.0023, "N": 0.0291, "P": 0.0001, "Q": 0.0112, "R": 0.0058, "S": 0.0005, "T": 0.0149, "V": 0.0013, "W": 0.0003, "Y": 0.0028 },
    "66": { "A": 0.0053, "C": 0.0009, "D": 0.0022, "E": 0.0072, "F": 0.0001, "G": 0.1972, "H": 0.0269, "K": 0.0093, "L": 0.0032, "N": 0.0008, "P": 0.0054, "Q": 0.7288, "R": 0.0096, "S": 0.0006, "T": 0.0004, "V": 0.0007, "Y": 0.0012 },
    "67": { "A": 0.0001, "D": 0.0001, "E": 0.0013, "G": 0.0001, "H": 0.0089, "K": 0.0129, "L": 0.0086, "M": 0.0003, "N": 0.0003, "P": 0.0078, "Q": 0.9365, "R": 0.02, "S": 0.0001, "T": 0.001, "V": 0.0012, "W": 0.0008 },
    "68": { "A": 0.0122, "C": 0.0073, "D": 0.0092, "E": 0.0001, "G": 0.9606, "H": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.003, "S": 0.0015, "T": 0.0001, "V": 0.0057, "W": 0.0001 },
    "69": { "A": 0.013, "C": 0.0013, "D": 0.0001, "F": 0.0033, "G": 0.0002, "H": 0.0002, "L": 0.0004, "N": 0.0001, "P": 0.0101, "S": 0.9664, "T": 0.0014, "V": 0.0004, "Y": 0.0032 },
    "70": { "A": 0.003, "C": 0.0003, "E": 0.002, "G": 0.9785, "I": 0.0002, "L": 0.0001, "P": 0.0001, "R": 0.0074, "S": 0.0002, "T": 0.0001, "V": 0.008, "W": 0.0001 },
    "71": { "A": 0.005, "C": 0.001, "D": 0.0013, "F": 0.0072, "G": 0.0112, "I": 0.0013, "L": 0.0038, "P": 0.0001, "R": 0.0001, "S": 0.0005, "V": 0.9685, "Y": 0.0001 },
    "72": { "A": 0.0013, "H": 0.0027, "L": 0.0015, "P": 0.9843, "R": 0.0009, "S": 0.0051, "T": 0.0043 },
    "74": { "C": 0.0041, "D": 0.0001, "F": 0.0001, "G": 0.011, "H": 0.0001, "I": 0.0092, "L": 0.0003, "M": 0.0003, "N": 0.0011, "P": 0.0001, "R": 0.025, "S": 0.9472, "T": 0.0011, "V": 0.0003 },
    "75": { "C": 0.0022, "F": 0.0001, "G": 0.0013, "H": 0.0026, "L": 0.0072, "N": 0.0001, "P": 0.0018, "R": 0.9826, "S": 0.0021 },
    "76": { "C": 0.0101, "F": 0.9753, "G": 0.0003, "I": 0.0013, "L": 0.0043, "S": 0.0013, "V": 0.0065, "Y": 0.0009 },
    "77": { "A": 0.0105, "C": 0.0007, "F": 0.0011, "I": 0.0001, "P": 0.0014, "S": 0.9832, "T": 0.0011, "V": 0.0001, "Y": 0.0018 },
    "78": { "A": 0.0069, "C": 0.0005, "D": 0.0001, "E": 0.0018, "F": 0.0001, "G": 0.9802, "I": 0.0001, "L": 0.0001, "R": 0.0022, "S": 0.0004, "V": 0.0077, "W": 0.0001 },
    "79": { "A": 0.009, "C": 0.0008, "D": 0.0004, "F": 0.0067, "G": 0.0011, "I": 0.0001, "L": 0.0004, "P": 0.0012, "S": 0.9753, "T": 0.0013, "V": 0.0003, "Y": 0.0036 },
    "80": { "D": 0.0003, "E": 0.0012, "H": 0.0008, "I": 0.0013, "K": 0.9112, "N": 0.066, "P": 0.0001, "Q": 0.0035, "R": 0.0059, "S": 0.0015, "T": 0.0081, "Y": 0.0002 },
    "81": { "A": 0.0096, "C": 0.0003, "D": 0.9629, "E": 0.0098, "G": 0.0035, "H": 0.0009, "N": 0.0028, "P": 0.0001, "R": 0.0001, "S": 0.0004, "V": 0.0026, "Y": 0.0072 },
    "82": { "A": 0.9491, "D": 0.003, "E": 0.0003, "F": 0.0011, "G": 0.0067, "H": 0.0001, "I": 0.0011, "L": 0.001, "P": 0.0062, "R": 0.0001, "S": 0.0079, "T": 0.009, "V": 0.0136, "W": 0.0001, "Y": 0.0006 },
    "83": { "A": 0.0069, "C": 0.0001, "E": 0.0001, "G": 0.0001, "K": 0.0001, "L": 0.0008, "P": 0.0006, "S": 0.9891, "T": 0.0019, "W": 0.0003, "Y": 0.0001 },
    "84": { "A": 0.9082, "D": 0.0037, "E": 0.0002, "G": 0.0016, "I": 0.0002, "K": 0.0001, "N": 0.0005, "P": 0.0008, "S": 0.006, "T": 0.0762, "V": 0.0024 },
    "85": { "C": 0.0001, "D": 0.0022, "E": 0.0001, "H": 0.002, "I": 0.0011, "K": 0.0121, "M": 0.0001, "N": 0.9667, "Q": 0.0001, "S": 0.0032, "T": 0.012, "Y": 0.0004 },
    "86": { "A": 0.8356, "D": 0.0001, "E": 0.0038, "G": 0.0007, "I": 0.0007, "K": 0.0004, "M": 0.0003, "P": 0.0011, "R": 0.0012, "S": 0.0084, "T": 0.1422, "V": 0.0055 },
    "87": { "A": 0.0112, "C": 0.0005, "D": 0.0002, "E": 0.002, "F": 0.0008, "G": 0.9646, "H": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0001, "R": 0.0012, "S": 0.0002, "T": 0.0001, "V": 0.0157, "W": 0.0033 },
    "88": { "A": 0.0001, "C": 0.0001, "F": 0.0022, "I": 0.8941, "L": 0.08, "M": 0.0058, "N": 0.0011, "P": 0.0001, "R": 0.0005, "S": 0.005, "T": 0.0069, "V": 0.0041 },
    "89": { "A": 0.0001, "C": 0.0001, "F": 0.0134, "G": 0.0001, "I": 0.0015, "L": 0.9734, "M": 0.0004, "P": 0.0002, "R": 0.0005, "S": 0.0018, "T": 0.0001, "V": 0.0085, "W": 0.0001 },
    "90": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "F": 0.0119, "G": 0.0001, "H": 0.0019, "I": 0.0052, "L": 0.898, "P": 0.0026, "R": 0.0646, "S": 0.0002, "T": 0.0004, "V": 0.0148 },
    "91": { "D": 0.0001, "F": 0.0017, "I": 0.969, "L": 0.0063, "M": 0.0011, "N": 0.0005, "S": 0.0144, "T": 0.002, "V": 0.005, "Y": 0.0001 },
    "92": { "A": 0.0208, "C": 0.0009, "F": 0.0007, "G": 0.0001, "L": 0.0001, "P": 0.0032, "S": 0.972, "T": 0.0012, "W": 0.0001, "Y": 0.0011 },
    "93": { "A": 0.0012, "C": 0.0001, "D": 0.0003, "E": 0.0023, "G": 0.9809, "R": 0.0019, "S": 0.0001, "T": 0.0001, "V": 0.0084, "W": 0.0047 },
    "94": { "D": 0.0001, "F": 0.0012, "G": 0.0003, "H": 0.0013, "I": 0.0026, "L": 0.9756, "P": 0.0028, "R": 0.0091, "S": 0.0001, "V": 0.0069 },
    "95": { "E": 0.0014, "H": 0.0044, "K": 0.0018, "L": 0.001, "N": 0.0001, "P": 0.0018, "Q": 0.9846, "R": 0.0047, "S": 0.0001, "T": 0.0001 },
    "96": { "A": 0.0322, "C": 0.0004, "D": 0.0007, "E": 0.0001, "F": 0.0024, "P": 0.0624, "S": 0.8977, "T": 0.0016, "V": 0.0001, "W": 0.0001, "Y": 0.0024 },
    "97": { "A": 0.003, "D": 0.0196, "E": 0.9714, "G": 0.0031, "K": 0.0004, "Q": 0.0001, "S": 0.0001, "T": 0.0001, "V": 0.0021 },
    "98": { "A": 0.0009, "D": 0.986, "E": 0.0074, "G": 0.0011, "H": 0.0004, "I": 0.0001, "K": 0.0001, "N": 0.0008, "V": 0.0009, "Y": 0.0023 },
    "99": { "A": 0.003, "D": 0.0106, "E": 0.9818, "G": 0.0022, "K": 0.0005, "Q": 0.0003, "V": 0.0013, "Y": 0.0003 },
    "100": { "A": 0.9771, "C": 0.0001, "D": 0.0083, "E": 0.0001, "G": 0.0051, "S": 0.0072, "T": 0.0013, "V": 0.0006, "Y": 0.0002 },
    "101": { "A": 0.0037, "D": 0.9695, "E": 0.0043, "F": 0.0001, "G": 0.0013, "H": 0.0008, "N": 0.0013, "S": 0.0002, "T": 0.0004, "V": 0.0023, "W": 0.0001, "Y": 0.0161 },
    "102": { "A": 0.0001, "C": 0.0009, "D": 0.0064, "E": 0.0007, "F": 0.0009, "H": 0.0007, "L": 0.0001, "N": 0.0005, "S": 0.0015, "Y": 0.9883 },
    "103": { "A": 0.0001, "C": 0.0047, "D": 0.008, "F": 0.0038, "H": 0.007, "L": 0.0007, "N": 0.0008, "P": 0.0001, "S": 0.0041, "T": 0.0001, "V": 0.0001, "Y": 0.9705 },
    "104": { "A": 0.0001, "C": 0.9736, "F": 0.0027, "G": 0.0076, "L": 0.0001, "P": 0.0001, "R": 0.001, "S": 0.0034, "V": 0.0002, "W": 0.0098, "Y": 0.0013 },
    "105": { "A": 0.0758, "C": 0.0002, "D": 0.0008, "F": 0.0001, "G": 0.0565, "H": 0.0001, "I": 0.0033, "K": 0.0026, "L": 0.0237, "M": 0.8202, "N": 0.0001, "P": 0.0001, "Q": 0.0006, "R": 0.0055, "S": 0.0011, "T": 0.0019, "V": 0.0072, "W": 0.0002 },
    "106": { "A": 0.0004, "C": 0.0001, "D": 0.0002, "F": 0.0052, "G": 0.0003, "I": 0.8919, "K": 0.0006, "L": 0.0019, "M": 0.0072, "N": 0.0047, "P": 0.0001, "R": 0.0002, "S": 0.0076, "T": 0.0659, "V": 0.0135 },
    "107": { "C": 0.0034, "E": 0.0001, "F": 0.0014, "G": 0.0068, "L": 0.0067, "R": 0.003, "S": 0.0018, "V": 0.0001, "W": 0.9752, "Y": 0.0016 },
    "108": { "A": 0.0016, "C": 0.0002, "D": 0.0032, "F": 0.0002, "G": 0.0001, "H": 0.7337, "I": 0.0001, "K": 0.0004, "L": 0.0021, "N": 0.0132, "P": 0.1475, "Q": 0.0028, "R": 0.0056, "S": 0.0011, "T": 0.0018, "V": 0.0001, "Y": 0.0864 },
    "109": { "A": 0.0008, "C": 0.0002, "D": 0.007, "G": 0.0083, "H": 0.0002, "I": 0.0087, "K": 0.0004, "L": 0.0001, "M": 0.0001, "N": 0.0476, "P": 0.0007, "Q": 0.0005, "R": 0.007, "S": 0.9089, "T": 0.0088, "V": 0.0003, "Y": 0.0002 },
    "110": { "D": 0.013, "I": 0.0012, "N": 0.8268, "Q": 0.0024, "R": 0.0012, "S": 0.1507, "T": 0.0024, "W": 0.0024 },
    "112": { "A": 0.0248, "I": 0.0041, "K": 0.1281, "L": 0.0124, "N": 0.0041, "P": 0.0041, "R": 0.0041, "S": 0.8058, "T": 0.0124 },
    "113": { "A": 0.066, "D": 0.0139, "G": 0.0041, "H": 0.0029, "I": 0.0033, "K": 0.0828, "L": 0.0008, "N": 0.3629, "P": 0.0033, "Q": 0.0029, "R": 0.0012, "S": 0.4285, "T": 0.0267, "V": 0.0004, "Y": 0.0004 },
    "114": { "A": 0.0911, "C": 0.0006, "D": 0.0066, "E": 0.0001, "F": 0.0014, "G": 0.0175, "H": 0.0014, "I": 0.0034, "K": 0.0296, "L": 0.0014, "M": 0.0005, "N": 0.0955, "P": 0.0028, "Q": 0.0004, "R": 0.0113, "S": 0.7035, "T": 0.0292, "V": 0.0009, "W": 0.0011, "Y": 0.0016 },
    "115": { "A": 0.6304, "C": 0.0005, "D": 0.014, "E": 0.0046, "F": 0.0099, "G": 0.0206, "H": 0.0169, "I": 0.0059, "K": 0.0029, "L": 0.0176, "M": 0.0005, "N": 0.0166, "P": 0.0208, "Q": 0.0029, "R": 0.0078, "S": 0.0755, "T": 0.0895, "V": 0.0357, "W": 0.0021, "Y": 0.0255 },
    "116": { "A": 0.0214, "C": 0.0087, "D": 0.0005, "E": 0.003, "F": 0.0031, "G": 0.0581, "H": 0.003, "I": 0.0002, "K": 0.0075, "L": 0.0159, "M": 0.0031, "N": 0.0029, "P": 0.0072, "Q": 0.0057, "R": 0.0247, "S": 0.0556, "T": 0.0019, "V": 0.2682, "W": 0.4647, "Y": 0.0445 },
    "117": { "A": 0.0029, "C": 0.0004, "D": 0.0001, "E": 0.0033, "F": 0.0017, "G": 0.003, "H": 0.0004, "I": 0.0077, "K": 0.0005, "L": 0.0113, "M": 0.0029, "N": 0.0001, "P": 0.0005, "Q": 0.0001, "R": 0.0006, "S": 0.0023, "T": 0.0014, "V": 0.9601, "W": 0.0008 },
    "118": { "A": 0.0005, "C": 0.0023, "F": 0.9841, "G": 0.0001, "H": 0.0001, "I": 0.0025, "L": 0.0032, "M": 0.0001, "Q": 0.0001, "S": 0.0023, "T": 0.0001, "V": 0.004, "W": 0.0001, "Y": 0.0006 },
    "119": { "A": 0.0002, "C": 0.0015, "D": 0.0009, "G": 0.9955, "P": 0.0001, "R": 0.0004, "S": 0.0006, "V": 0.0007, "Y": 0.0001 },
    "120": { "A": 0.0021, "E": 0.0004, "F": 0.0001, "G": 0.9291, "N": 0.0001, "P": 0.0028, "R": 0.0014, "S": 0.0062, "T": 0.0568, "V": 0.0009, "W": 0.0001, "Y": 0.0001 },
    "121": { "A": 0.0006, "C": 0.0001, "E": 0.0007, "G": 0.9952, "R": 0.0007, "V": 0.0007, "W": 0.0019 },
    "122": { "A": 0.0007, "I": 0.0008, "N": 0.0006, "P": 0.0018, "S": 0.0014, "T": 0.9946 },
    "123": { "A": 0.0001, "E": 0.0025, "F": 0.0001, "G": 0.0001, "H": 0.0004, "K": 0.9604, "L": 0.0001, "M": 0.0004, "N": 0.0035, "P": 0.0001, "Q": 0.0111, "R": 0.0124, "S": 0.0002, "T": 0.0086 },
    "124": { "G": 0.0001, "I": 0.0011, "L": 0.9235, "M": 0.0009, "P": 0.0005, "Q": 0.0003, "R": 0.0014, "V": 0.0722 },
    "125": { "A": 0.0013, "D": 0.0001, "I": 0.0018, "K": 0.0001, "L": 0.0001, "N": 0.0011, "P": 0.0034, "S": 0.0037, "T": 0.9885 },
    "126": { "A": 0.0016, "D": 0.0004, "F": 0.0004, "G": 0.0056, "I": 0.0009, "L": 0.0006, "R": 0.0001, "V": 0.9906 },
    "127": { "I": 0.0011, "L": 0.9982, "P": 0.0001, "V": 0.0006 }
  },
  "IGLV6": {
    "1": { "K": 0.0074, "N": 0.9926 },
    "2": { "C": 0.0018, "F": 0.9724, "G": 0.0001, "I": 0.0009, "L": 0.0117, "R": 0.0001, "S": 0.0062, "T": 0.0001, "V": 0.0048, "W": 0.0001, "Y": 0.0017 },
    "3": { "A": 0.0001, "F": 0.0004, "G": 0.0001, "I": 0.03, "K": 0.0027, "L": 0.0273, "M": 0.9219, "Q": 0.0001, "R": 0.0013, "T": 0.0013, "V": 0.0143, "W": 0.0005 },
    "4": { "F": 0.0001, "I": 0.0001, "L": 0.9929, "M": 0.0026, "P": 0.0024, "Q": 0.0005, "R": 0.0012, "T": 0.0001, "V": 0.0001 },
    "5": { "A": 0.0056, "F": 0.0001, "H": 0.0004, "I": 0.0031, "K": 0.0006, "L": 0.0003, "N": 0.0032, "P": 0.0264, "R": 0.0001, "S": 0.0069, "T": 0.9525, "V": 0.0007, "Y": 0.0001 },
    "6": { "A": 0.0001, "E": 0.0005, "H": 0.0133, "K": 0.0037, "L": 0.0011, "N": 0.0002, "P": 0.0216, "Q": 0.957, "R": 0.0017, "S": 0.0002, "T": 0.0006 },
    "7": { "A": 0.001, "H": 0.0074, "I": 0.0001, "L": 0.0032, "P": 0.9742, "Q": 0.0015, "R": 0.0015, "S": 0.0061, "T": 0.005, "Y": 0.0001 },
    "8": { "D": 0.0007, "H": 0.938, "K": 0.0003, "L": 0.0037, "N": 0.0045, "P": 0.0396, "Q": 0.0065, "R": 0.0044, "S": 0.0001, "T": 0.0006, "Y": 0.0015 },
    "9": { "A": 0.0069, "C": 0.0001, "D": 0.0001, "F": 0.001, "G": 0.0004, "H": 0.0006, "I": 0.0001, "P": 0.0321, "Q": 0.0001, "R": 0.0001, "S": 0.9514, "T": 0.0035, "V": 0.0001, "Y": 0.0036 },
    "11": { "A": 0.0108, "C": 0.0001, "D": 0.0001, "E": 0.0025, "F": 0.0006, "G": 0.0057, "H": 0.0001, "I": 0.0004, "L": 0.0126, "M": 0.0027, "P": 0.0002, "Q": 0.0001, "R": 0.0001, "S": 0.0003, "T": 0.0001, "V": 0.9633, "W": 0.0003 },
    "12": { "A": 0.0043, "C": 0.0001, "D": 0.0001, "E": 0.0001, "H": 0.0001, "L": 0.002, "P": 0.0087, "Q": 0.0004, "R": 0.0001, "S": 0.9811, "T": 0.0022, "W": 0.0006, "Y": 0.0001 },
    "13": { "A": 0.029, "C": 0.0001, "D": 0.0198, "E": 0.9204, "G": 0.0121, "H": 0.0006, "K": 0.0015, "L": 0.0001, "M": 0.0001, "N": 0.0002, "P": 0.0001, "Q": 0.0107, "R": 0.0003, "S": 0.001, "T": 0.0004, "V": 0.0029, "Y": 0.0007 },
    "14": { "A": 0.0094, "C": 0.0009, "D": 0.0001, "F": 0.0027, "H": 0.0004, "I": 0.0001, "L": 0.0003, "P": 0.0091, "R": 0.0001, "S": 0.9703, "T": 0.0019, "V": 0.0001, "Y": 0.0045 },
    "15": { "A": 0.0026, "E": 0.0001, "H": 0.0001, "L": 0.0037, "P": 0.9759, "Q": 0.0084, "R": 0.0009, "S": 0.0024, "T": 0.0059 },
    "16": { "A": 0.0069, "C": 0.0006, "D": 0.0001, "E": 0.0028, "G": 0.9596, "L": 0.0006, "M": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0098, "S": 0.0002, "T": 0.0001, "V": 0.0099, "W": 0.0091 },
    "17": { "A": 0.0008, "D": 0.0003, "E": 0.0079, "G": 0.0001, "H": 0.0017, "I": 0.0001, "K": 0.8791, "L": 0.0001, "M": 0.0014, "N": 0.0153, "P": 0.0009, "Q": 0.0488, "R": 0.0081, "S": 0.0004, "T": 0.0344, "V": 0.0001, "Y": 0.0003 },
    "18": { "A": 0.0042, "D": 0.0001, "E": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0001, "K": 0.004, "L": 0.0001, "M": 0.0009, "N": 0.0001, "P": 0.0288, "Q": 0.0008, "R": 0.001, "S": 0.0022, "T": 0.9575 },
    "19": { "A": 0.0035, "D": 0.0001, "E": 0.0009, "F": 0.0001, "G": 0.0034, "I": 0.0124, "K": 0.0001, "L": 0.006, "M": 0.0002, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0001, "T": 0.0005, "V": 0.9725 },
    "20": { "A": 0.0022, "H": 0.0004, "I": 0.0127, "K": 0.0007, "L": 0.0001, "N": 0.0035, "P": 0.0087, "R": 0.0006, "S": 0.0047, "T": 0.9652, "V": 0.0011 },
    "21": { "C": 0.0001, "D": 0.0001, "F": 0.0024, "G": 0.0001, "H": 0.0003, "I": 0.9612, "K": 0.0001, "L": 0.0148, "M": 0.0009, "N": 0.0029, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0065, "T": 0.0081, "V": 0.0019, "Y": 0.0001 },
    "22": { "A": 0.0096, "C": 0.0003, "D": 0.0001, "F": 0.0013, "H": 0.0004, "I": 0.0001, "L": 0.0003, "N": 0.0001, "P": 0.0274, "Q": 0.0001, "S": 0.9519, "T": 0.0056, "V": 0.0001, "Y": 0.0029 },
    "23": { "A": 0.0001, "C": 0.9797, "F": 0.0047, "G": 0.0034, "H": 0.0001, "L": 0.0003, "P": 0.0001, "R": 0.0032, "S": 0.0064, "T": 0.0001, "V": 0.0002, "W": 0.0003, "Y": 0.0015 },
    "24": { "A": 0.0043, "C": 0.0001, "F": 0.0001, "G": 0.0001, "H": 0.0006, "I": 0.0021, "L": 0.0002, "N": 0.0019, "P": 0.0204, "R": 0.0001, "S": 0.0059, "T": 0.9642, "V": 0.0001 },
    "25": { "A": 0.005, "C": 0.0042, "D": 0.001, "F": 0.0001, "G": 0.2987, "H": 0.002, "L": 0.0018, "P": 0.0034, "R": 0.6795, "S": 0.0027, "V": 0.0016 },
    "26": { "A": 0.0006, "C": 0.0012, "D": 0.0016, "F": 0.0003, "G": 0.0031, "H": 0.0008, "I": 0.0043, "K": 0.0001, "L": 0.0003, "M": 0.0001, "N": 0.0111, "P": 0.0008, "R": 0.0249, "S": 0.9368, "T": 0.0138, "V": 0.0001, "Y": 0.0001 },
    "27": { "A": 0.0002, "C": 0.001, "D": 0.0003, "F": 0.0001, "G": 0.0265, "H": 0.0003, "I": 0.0032, "L": 0.0003, "M": 0.0003, "N": 0.0027, "P": 0.0002, "R": 0.0298, "S": 0.9291, "T": 0.0057, "V": 0.0004 },
    "28": { "A": 0.0058, "C": 0.0099, "D": 0.005, "G": 0.9666, "H": 0.0002, "N": 0.0001, "P": 0.0001, "R": 0.0047, "S": 0.0038, "V": 0.0034, "W": 0.0001, "Y": 0.0001 },
    "29": { "A": 0.0004, "C": 0.0016, "D": 0.0024, "E": 0.0003, "F": 0.0003, "G": 0.0059, "H": 0.0012, "I": 0.0031, "K": 0.0007, "L": 0.0005, "N": 0.0239, "P": 0.0005, "Q": 0.0001, "R": 0.0314, "S": 0.9171, "T": 0.0088, "V": 0.0006, "Y": 0.0011 },
    "30": { "A": 0.0002, "C": 0.0001, "D": 0.0001, "F": 0.0068, "G": 0.0003, "H": 0.0001, "I": 0.9387, "L": 0.0231, "M": 0.007, "N": 0.0007, "P": 0.0001, "R": 0.0003, "S": 0.0077, "T": 0.0028, "V": 0.0121 },
    "35": { "A": 0.9417, "C": 0.0001, "D": 0.0131, "E": 0.0001, "F": 0.0001, "G": 0.0158, "H": 0.0003, "I": 0.0001, "L": 0.0001, "P": 0.0043, "R": 0.0003, "S": 0.0104, "T": 0.0051, "V": 0.0083, "Y": 0.0001 },
    "36": { "A": 0.0048, "C": 0.0003, "D": 0.0047, "F": 0.0003, "G": 0.0112, "H": 0.0006, "I": 0.0041, "L": 0.0015, "N": 0.0175, "P": 0.0006, "R": 0.0233, "S": 0.9088, "T": 0.0214, "V": 0.0001, "Y": 0.0007 },
    "37": { "A": 0.0028, "C": 0.0001, "D": 0.0079, "E": 0.0003, "F": 0.0001, "G": 0.0006, "H": 0.0217, "I": 0.0004, "K": 0.0091, "M": 0.0003, "N": 0.914, "P": 0.0006, "Q": 0.0008, "R": 0.0008, "S": 0.0139, "T": 0.0186, "V": 0.0001, "Y": 0.0078 },
    "38": { "A": 0.0003, "C": 0.0015, "D": 0.013, "E": 0.0001, "F": 0.0274, "G": 0.0001, "H": 0.0124, "I": 0.0003, "K": 0.0001, "L": 0.0013, "M": 0.0001, "N": 0.0028, "P": 0.0011, "Q": 0.0007, "R": 0.0003, "S": 0.0243, "T": 0.0005, "V": 0.0004, "W": 0.0002, "Y": 0.9131 },
    "39": { "A": 0.0047, "C": 0.0007, "D": 0.0004, "E": 0.0027, "F": 0.001, "G": 0.008, "H": 0.0001, "I": 0.0017, "L": 0.0115, "M": 0.0019, "Q": 0.0001, "R": 0.0002, "S": 0.0001, "T": 0.0001, "V": 0.9665, "W": 0.0001 },
    "40": { "A": 0.0001, "E": 0.0011, "H": 0.0205, "K": 0.0041, "L": 0.0014, "M": 0.0001, "N": 0.0001, "P": 0.0228, "Q": 0.9439, "R": 0.0035, "S": 0.0006, "T": 0.0014, "V": 0.0001, "W": 0.0001, "Y": 0.0002 },
    "41": { "A": 0.0001, "C": 0.0087, "E": 0.0001, "G": 0.0098, "L": 0.0087, "Q": 0.0001, "R": 0.0087, "S": 0.0036, "T": 0.0001, "V": 0.0006, "W": 0.9596 },
    "42": { "A": 0.0001, "C": 0.0038, "D": 0.0056, "E": 0.0001, "F": 0.0177, "G": 0.0001, "H": 0.0064, "I": 0.0011, "L": 0.0012, "N": 0.0016, "Q": 0.0001, "R": 0.0001, "S": 0.0154, "T": 0.0001, "V": 0.0001, "Y": 0.9463 },
    "43": { "A": 0.0001, "C": 0.0001, "E": 0.0001, "H": 0.0091, "K": 0.0029, "L": 0.0018, "N": 0.0001, "P": 0.0149, "Q": 0.9567, "R": 0.0136, "S": 0.0001, "T": 0.0006, "W": 0.0001 },
    "44": { "A": 0.0001, "E": 0.0005, "H": 0.0142, "K": 0.005, "L": 0.0054, "M": 0.0001, "N": 0.0001, "P": 0.0155, "Q": 0.9511, "R": 0.0067, "S": 0.0001, "T": 0.0002, "V": 0.0003, "W": 0.0001, "Y": 0.0008 },
    "45": { "A": 0.0002, "C": 0.0015, "F": 0.0001, "G": 0.0008, "H": 0.0031, "L": 0.0033, "P": 0.0127, "R": 0.9753, "S": 0.0027, "T": 0.0001, "V": 0.0002 },
    "46": { "A": 0.001, "H": 0.0003, "L": 0.001, "P": 0.9878, "Q": 0.0028, "R": 0.0009, "S": 0.001, "T": 0.0051 },
    "47": { "A": 0.0035, "C": 0.007, "D": 0.0066, "E": 0.0001, "G": 0.9726, "H": 0.0002, "N": 0.0001, "P": 0.0001, "R": 0.0035, "S": 0.0017, "T": 0.0001, "V": 0.0046 },
    "48": { "A": 0.0006, "C": 0.0007, "D": 0.0006, "G": 0.0048, "H": 0.0003, "I": 0.0035, "K": 0.0002, "L": 0.0002, "M": 0.0005, "N": 0.0043, "P": 0.0001, "Q": 0.0003, "R": 0.0301, "S": 0.947, "T": 0.0061, "V": 0.0005, "W": 0.0001 },
    "49": { "A": 0.6628, "D": 0.002, "E": 0.0002, "F": 0.0017, "G": 0.0029, "H": 0.0002, "I": 0.0001, "L": 0.0001, "P": 0.009, "Q": 0.0003, "R": 0.0006, "S": 0.3087, "T": 0.0029, "V": 0.007, "Y": 0.0015 },
    "50": { "A": 0.0014, "H": 0.0038, "L": 0.0013, "P": 0.9857, "Q": 0.0002, "R": 0.0006, "S": 0.0014, "T": 0.0056, "V": 0.0001 },
    "51": { "A": 0.0052, "D": 0.0006, "E": 0.0001, "F": 0.0003, "H": 0.0006, "I": 0.0112, "K": 0.0011, "L": 0.0007, "M": 0.0013, "N": 0.0052, "P": 0.0163, "Q": 0.0001, "R": 0.0025, "S": 0.0113, "T": 0.9411, "V": 0.0021, "W": 0.0001 },
    "52": { "A": 0.0066, "C": 0.0008, "D": 0.0001, "F": 0.0005, "G": 0.0001, "H": 0.0003, "I": 0.023, "K": 0.0001, "L": 0.005, "M": 0.0019, "N": 0.0057, "P": 0.0223, "Q": 0.0001, "R": 0.0003, "S": 0.0066, "T": 0.9235, "V": 0.0029, "Y": 0.0002 },
    "53": { "A": 0.0014, "D": 0.0001, "E": 0.001, "F": 0.0008, "G": 0.0063, "I": 0.005, "L": 0.0249, "M": 0.0053, "P": 0.0002, "R": 0.0005, "S": 0.0001, "T": 0.0002, "V": 0.9536, "W": 0.0007 },
    "54": { "C": 0.0001, "F": 0.0039, "G": 0.0001, "H": 0.0001, "I": 0.9629, "K": 0.0001, "L": 0.0111, "M": 0.0043, "N": 0.0022, "P": 0.0001, "R": 0.0004, "S": 0.0075, "T": 0.0036, "V": 0.0036, "Y": 0.0001 },
    "55": { "A": 0.0002, "C": 0.002, "D": 0.0082, "E": 0.0013, "F": 0.024, "G": 0.0002, "H": 0.0041, "I": 0.0003, "L": 0.001, "N": 0.0022, "Q": 0.0006, "R": 0.0001, "S": 0.0092, "T": 0.0001, "V": 0.0011, "W": 0.0001, "Y": 0.9453 },
    "56": { "A": 0.0146, "C": 0.0001, "D": 0.024, "E": 0.9331, "G": 0.0089, "H": 0.0032, "I": 0.0001, "K": 0.005, "L": 0.0003, "M": 0.0001, "N": 0.001, "Q": 0.0045, "R": 0.0006, "T": 0.0003, "V": 0.003, "Y": 0.0013 },
    "57": { "A": 0.0075, "D": 0.9316, "E": 0.0073, "F": 0.0001, "G": 0.0041, "H": 0.0056, "K": 0.0003, "L": 0.0001, "N": 0.0281, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0025, "T": 0.0007, "V": 0.0031, "Y": 0.0089 },
    "65": { "A": 0.0013, "C": 0.0001, "D": 0.0406, "E": 0.0022, "F": 0.0012, "G": 0.001, "H": 0.0085, "I": 0.0017, "K": 0.0167, "L": 0.0002, "N": 0.8809, "Q": 0.0005, "R": 0.002, "S": 0.0128, "T": 0.0174, "V": 0.0017, "W": 0.0003, "Y": 0.0109 },
    "66": { "A": 0.0003, "D": 0.0006, "E": 0.0105, "F": 0.0001, "G": 0.0003, "H": 0.02, "I": 0.0001, "K": 0.0154, "L": 0.0059, "M": 0.0001, "N": 0.0003, "P": 0.0072, "Q": 0.9095, "R": 0.0261, "S": 0.0001, "T": 0.0006, "V": 0.0016, "Y": 0.001 },
    "67": { "C": 0.0001, "G": 0.003, "I": 0.0033, "K": 0.0036, "L": 0.0003, "M": 0.0001, "N": 0.0003, "P": 0.0003, "Q": 0.0003, "R": 0.9625, "S": 0.0224, "T": 0.0034, "V": 0.0003 },
    "68": { "A": 0.0041, "C": 0.0001, "E": 0.0001, "F": 0.0006, "H": 0.0042, "I": 0.0001, "L": 0.0044, "M": 0.0002, "N": 0.0001, "P": 0.976, "Q": 0.0001, "R": 0.0027, "S": 0.0041, "T": 0.003, "V": 0.0003, "Y": 0.0001 },
    "69": { "A": 0.021, "C": 0.0008, "D": 0.0001, "F": 0.0036, "H": 0.0003, "I": 0.0002, "L": 0.0004, "N": 0.0001, "P": 0.0261, "Q": 0.0001, "S": 0.9423, "T": 0.0031, "Y": 0.002 },
    "70": { "A": 0.0044, "C": 0.0007, "D": 0.0006, "E": 0.0044, "G": 0.9686, "H": 0.0001, "I": 0.0001, "L": 0.0002, "M": 0.0001, "P": 0.0002, "R": 0.0041, "S": 0.0008, "T": 0.0001, "V": 0.007, "W": 0.0087 },
    "71": { "A": 0.0041, "C": 0.0006, "D": 0.0004, "E": 0.0006, "F": 0.0091, "G": 0.0101, "I": 0.0039, "L": 0.0021, "N": 0.0001, "P": 0.0001, "R": 0.0001, "S": 0.0004, "V": 0.9684, "Y": 0.0001 },
    "72": { "A": 0.0011, "C": 0.0001, "H": 0.0019, "L": 0.0017, "N": 0.0001, "P": 0.9851, "R": 0.0007, "S": 0.0071, "T": 0.002, "V": 0.0001, "Y": 0.0001 },
    "74": { "A": 0.024, "C": 0.0006, "D": 0.9202, "E": 0.0109, "F": 0.0001, "G": 0.0122, "H": 0.0046, "I": 0.0001, "K": 0.0004, "L": 0.0001, "N": 0.0051, "P": 0.0003, "R": 0.0003, "S": 0.0016, "T": 0.0004, "V": 0.0115, "W": 0.0001, "Y": 0.0075 },
    "75": { "E": 0.0001, "G": 0.0003, "H": 0.0002, "L": 0.0073, "P": 0.0027, "Q": 0.0015, "R": 0.9873, "S": 0.0002, "W": 0.0005 },
    "76": { "C": 0.0085, "F": 0.9766, "G": 0.0001, "I": 0.0016, "L": 0.0041, "R": 0.0001, "S": 0.0024, "V": 0.0057, "Y": 0.0009 },
    "77": { "A": 0.0076, "C": 0.0004, "D": 0.0001, "F": 0.0015, "H": 0.0001, "L": 0.0002, "P": 0.0033, "S": 0.9852, "T": 0.0007, "Y": 0.0009 },
    "78": { "A": 0.0105, "C": 0.012, "D": 0.0017, "F": 0.0004, "G": 0.9652, "L": 0.0001, "Q": 0.0002, "R": 0.0017, "S": 0.0024, "V": 0.0057 },
    "79": { "A": 0.0092, "C": 0.0008, "D": 0.0002, "F": 0.0021, "G": 0.0001, "H": 0.0002, "I": 0.0001, "L": 0.0001, "N": 0.0001, "P": 0.0025, "S": 0.9799, "T": 0.0022, "Y": 0.0025 },
    "80": { "A": 0.0001, "C": 0.0001, "F": 0.0031, "G": 0.0003, "H": 0.0001, "I": 0.9572, "K": 0.0007, "L": 0.0071, "M": 0.0006, "N": 0.0017, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0113, "T": 0.0061, "V": 0.0112 },
    "81": { "A": 0.0069, "C": 0.0001, "D": 0.9804, "E": 0.0017, "F": 0.0001, "G": 0.0025, "H": 0.0004, "N": 0.0021, "R": 0.0001, "S": 0.0002, "T": 0.0001, "V": 0.0011, "Y": 0.0043 },
    "82": { "A": 0.0013, "C": 0.0013, "D": 0.0008, "E": 0.0001, "F": 0.0008, "G": 0.0043, "H": 0.0001, "I": 0.0068, "K": 0.0015, "L": 0.001, "M": 0.0003, "N": 0.0014, "P": 0.0015, "R": 0.0319, "S": 0.9258, "T": 0.019, "V": 0.0007, "Y": 0.0013 },
    "83": { "A": 0.0105, "C": 0.0009, "D": 0.0001, "F": 0.002, "G": 0.0001, "L": 0.0001, "N": 0.0002, "P": 0.0038, "S": 0.9757, "T": 0.0038, "V": 0.0001, "Y": 0.0028 },
    "84": { "A": 0.0102, "C": 0.0001, "D": 0.0001, "E": 0.0001, "F": 0.0009, "G": 0.0009, "L": 0.0001, "N": 0.0001, "P": 0.0027, "R": 0.0001, "S": 0.9779, "T": 0.0045, "Y": 0.0023 },
    "85": { "A": 0.0002, "D": 0.0026, "E": 0.0001, "G": 0.0001, "H": 0.0031, "I": 0.0022, "K": 0.0038, "N": 0.9652, "P": 0.0002, "R": 0.0006, "S": 0.0056, "T": 0.0155, "V": 0.0001, "Y": 0.0006 },
    "86": { "A": 0.024, "C": 0.0001, "D": 0.0001, "F": 0.0006, "I": 0.0001, "K": 0.0001, "L": 0.0001, "P": 0.0029, "S": 0.9614, "T": 0.0083, "V": 0.0005, "Y": 0.0018 },
    "87": { "A": 0.9898, "D": 0.0016, "E": 0.0007, "F": 0.0001, "G": 0.0005, "P": 0.0008, "S": 0.0031, "T": 0.0015, "V": 0.002, "Y": 0.0001 },
    "88": { "A": 0.0106, "C": 0.0002, "D": 0.0001, "F": 0.0039, "I": 0.0001, "L": 0.0007, "P": 0.0034, "Q": 0.0001, "S": 0.9698, "T": 0.0051, "V": 0.0002, "Y": 0.0058 },
    "89": { "F": 0.0009, "H": 0.0013, "I": 0.0022, "L": 0.982, "P": 0.0041, "Q": 0.0001, "R": 0.0082, "S": 0.0002, "T": 0.0005, "V": 0.0007 },
    "90": { "A": 0.004, "F": 0.0005, "H": 0.0001, "I": 0.0087, "K": 0.0001, "L": 0.0015, "M": 0.0001, "N": 0.002, "P": 0.0071, "Q": 0.0001, "R": 0.0003, "S": 0.011, "T": 0.9646, "V": 0.0001 },
    "91": { "D": 0.0001, "F": 0.0016, "G": 0.0001, "H": 0.0002, "I": 0.9721, "L": 0.0049, "M": 0.0006, "N": 0.001, "P": 0.0001, "R": 0.0001, "S": 0.0121, "T": 0.003, "V": 0.0042 },
    "92": { "A": 0.0244, "C": 0.0006, "D": 0.0001, "F": 0.0013, "G": 0.0001, "N": 0.0001, "P": 0.0024, "S": 0.9642, "T": 0.0048, "V": 0.0002, "Y": 0.0018 },
    "93": { "A": 0.0024, "D": 0.0018, "E": 0.0036, "G": 0.9804, "L": 0.0002, "P": 0.0002, "Q": 0.0001, "R": 0.0046, "S": 0.0001, "T": 0.0009, "V": 0.0055 },
    "94": { "A": 0.0001, "G": 0.0001, "H": 0.0001, "I": 0.0001, "L": 0.9893, "M": 0.0013, "P": 0.001, "Q": 0.0006, "R": 0.0061, "S": 0.0001, "T": 0.0001, "V": 0.0012, "W": 0.0001, "Y": 0.0001 },
    "95": { "A": 0.0001, "E": 0.0094, "G": 0.0002, "H": 0.0001, "K": 0.9336, "L": 0.0001, "M": 0.0025, "N": 0.0042, "P": 0.0001, "Q": 0.0168, "R": 0.0247, "S": 0.0004, "T": 0.0068, "V": 0.001, "Y": 0.0001 },
    "96": { "A": 0.0136, "D": 0.0003, "F": 0.0001, "I": 0.0014, "K": 0.0005, "L": 0.0001, "M": 0.0001, "N": 0.0021, "P": 0.0105, "R": 0.0003, "S": 0.0053, "T": 0.9647, "V": 0.001 },
    "97": { "A": 0.0027, "D": 0.0144, "E": 0.979, "G": 0.0018, "K": 0.0009, "M": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0001, "V": 0.0007, "Y": 0.0001 },
    "98": { "A": 0.0008, "C": 0.0001, "D": 0.9925, "E": 0.0027, "G": 0.0004, "H": 0.0002, "N": 0.0004, "S": 0.0001, "T": 0.0001, "V": 0.0008, "Y": 0.002 },
    "99": { "A": 0.0034, "D": 0.0085, "E": 0.9839, "G": 0.002, "K": 0.0007, "Q": 0.0002, "V": 0.0011, "Y": 0.0002 },
    "100": { "A": 0.9802, "C": 0.0002, "D": 0.0038, "E": 0.0001, "G": 0.0064, "P": 0.0001, "S": 0.0054, "T": 0.0024, "V": 0.0008, "Y": 0.0006 },
    "101": { "A": 0.0024, "C": 0.0001, "D": 0.9643, "E": 0.0064, "F": 0.0015, "G": 0.0017, "H": 0.003, "I": 0.0003, "L": 0.0006, "N": 0.0027, "S": 0.0006, "T": 0.001, "V": 0.0038, "Y": 0.0118 },
    "102": { "C": 0.0023, "D": 0.0082, "F": 0.0026, "H": 0.0008, "I": 0.0001, "K": 0.0001, "M": 0.0001, "N": 0.0002, "P": 0.0002, "S": 0.0023, "V": 0.0001, "Y": 0.983 },
    "103": { "C": 0.0047, "D": 0.0056, "E": 0.0001, "F": 0.0154, "G": 0.0001, "H": 0.0031, "I": 0.0003, "L": 0.0002, "N": 0.0006, "R": 0.0001, "S": 0.0036, "V": 0.0003, "Y": 0.9657 },
    "104": { "A": 0.0001, "C": 0.9778, "F": 0.0015, "G": 0.0082, "L": 0.0001, "N": 0.0001, "R": 0.001, "S": 0.0022, "T": 0.0001, "V": 0.0002, "W": 0.008, "Y": 0.0007 },
    "105": { "A": 0.0001, "C": 0.0003, "E": 0.0006, "H": 0.009, "I": 0.0001, "K": 0.0014, "L": 0.0015, "M": 0.0001, "N": 0.0001, "P": 0.0022, "Q": 0.982, "R": 0.0015, "S": 0.0006, "T": 0.0003, "V": 0.0001, "Y": 0.0001 },
    "106": { "A": 0.0058, "C": 0.0023, "D": 0.0001, "F": 0.001, "G": 0.0001, "H": 0.0001, "I": 0.0001, "K": 0.0001, "N": 0.0002, "P": 0.0014, "Q": 0.0001, "R": 0.0001, "S": 0.9792, "T": 0.0041, "Y": 0.0054 },
    "107": { "A": 0.0001, "C": 0.0021, "D": 0.007, "E": 0.0002, "F": 0.0149, "G": 0.0003, "H": 0.005, "I": 0.0001, "L": 0.0014, "M": 0.0001, "N": 0.0009, "P": 0.0001, "Q": 0.0003, "R": 0.0003, "S": 0.0086, "T": 0.0001, "V": 0.002, "W": 0.0005, "Y": 0.9559 },
    "108": { "A": 0.0043, "C": 0.0001, "D": 0.9594, "E": 0.0074, "F": 0.0006, "G": 0.0069, "H": 0.0039, "I": 0.0001, "L": 0.0001, "M": 0.0001, "N": 0.0059, "Q": 0.0011, "S": 0.001, "T": 0.0011, "V": 0.0016, "Y": 0.0065 },
    "109": { "A": 0.0061, "C": 0.0004, "D": 0.017, "E": 0.0012, "F": 0.0018, "G": 0.0206, "H": 0.0017, "I": 0.0078, "K": 0.0014, "L": 0.0029, "M": 0.0003, "N": 0.024, "P": 0.0021, "Q": 0.0003, "R": 0.0131, "S": 0.8704, "T": 0.0234, "V": 0.0039, "W": 0.0002, "Y": 0.0014 },
    "110": { "A": 0.0045, "C": 0.0011, "D": 0.0096, "F": 0.0045, "G": 0.0107, "I": 0.0158, "K": 0.0006, "L": 0.0011, "M": 0.0006, "N": 0.0242, "P": 0.0011, "Q": 0.0017, "R": 0.0039, "S": 0.8816, "T": 0.0248, "V": 0.0073, "W": 0.0039, "Y": 0.0028 },
    "112": { "A": 0.0129, "D": 0.0452, "H": 0.0387, "I": 0.0065, "K": 0.0065, "L": 0.0323, "N": 0.7419, "P": 0.0065, "Q": 0.0516, "R": 0.0065, "S": 0.0194, "T": 0.0129, "Y": 0.0194 },
    "113": { "A": 0.0034, "C": 0.0005, "D": 0.009, "E": 0.0004, "F": 0.0011, "G": 0.0102, "H": 0.01, "I": 0.0094, "K": 0.0053, "L": 0.0029, "M": 0.0001, "N": 0.2074, "P": 0.0043, "Q": 0.0029, "R": 0.0124, "S": 0.6788, "T": 0.039, "V": 0.0007, "W": 0.0001, "Y": 0.0021 },
    "114": { "A": 0.006, "C": 0.0002, "D": 0.0124, "E": 0.0025, "F": 0.0034, "G": 0.0122, "H": 0.0361, "I": 0.0279, "K": 0.0133, "L": 0.0157, "M": 0.001, "N": 0.3639, "P": 0.029, "Q": 0.0246, "R": 0.0197, "S": 0.3741, "T": 0.0517, "V": 0.0026, "W": 0.0004, "Y": 0.0033 },
    "115": { "A": 0.0133, "C": 0.0022, "D": 0.035, "E": 0.0109, "F": 0.0187, "G": 0.0258, "H": 0.1976, "I": 0.0318, "K": 0.0161, "L": 0.0595, "M": 0.0027, "N": 0.2779, "P": 0.0614, "Q": 0.0514, "R": 0.0595, "S": 0.0591, "T": 0.0338, "V": 0.0104, "W": 0.0015, "Y": 0.0314 },
    "116": { "A": 0.0352, "C": 0.0034, "D": 0.0017, "E": 0.0185, "F": 0.0013, "G": 0.0911, "H": 0.0082, "I": 0.0038, "K": 0.0085, "L": 0.0228, "M": 0.0048, "N": 0.0064, "P": 0.0041, "Q": 0.0534, "R": 0.0602, "S": 0.0043, "T": 0.0041, "V": 0.3492, "W": 0.288, "Y": 0.0312 },
    "117": { "A": 0.0019, "C": 0.0001, "D": 0.0001, "E": 0.002, "F": 0.0016, "G": 0.0027, "I": 0.0217, "K": 0.0008, "L": 0.0111, "M": 0.0024, "N": 0.0006, "P": 0.0008, "R": 0.0025, "S": 0.0012, "T": 0.0003, "V": 0.95, "W": 0.0001 },
    "118": { "C": 0.0026, "D": 0.0001, "F": 0.9823, "G": 0.0005, "H": 0.0001, "I": 0.0048, "K": 0.0001, "L": 0.0029, "M": 0.0001, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0003, "S": 0.0007, "V": 0.0045, "W": 0.0003, "Y": 0.0005 },
    "119": { "A": 0.0004, "C": 0.0008, "D": 0.0009, "G": 0.995, "H": 0.0001, "L": 0.0001, "P": 0.0001, "R": 0.0005, "S": 0.0007, "V": 0.0015 },
    "120": { "A": 0.0011, "C": 0.0001, "D": 0.0003, "E": 0.0014, "G": 0.9348, "I": 0.0001, "P": 0.0002, "Q": 0.0001, "R": 0.0011, "S": 0.0105, "T": 0.0492, "V": 0.0009, "W": 0.0001 },
    "121": { "A": 0.0009, "C": 0.0001, "D": 0.0001, "E": 0.0005, "G": 0.9946, "M": 0.0001, "R": 0.0007, "V": 0.0017, "W": 0.0013 },
    "122": { "A": 0.0006, "K": 0.0001, "N": 0.0005, "P": 0.0015, "S": 0.001, "T": 0.9962 },
    "123": { "A": 0.0005, "E": 0.0018, "F": 0.0001, "H": 0.0004, "I": 0.0001, "K": 0.9334, "L": 0.0003, "M": 0.0012, "N": 0.0022, "P": 0.0002, "Q": 0.0333, "R": 0.0189, "S": 0.0003, "T": 0.0067, "V": 0.0001, "Y": 0.0005 },
    "124": { "F": 0.0001, "G": 0.0001, "I": 0.0003, "L": 0.9265, "M": 0.001, "P": 0.0008, "Q": 0.0001, "R": 0.0017, "S": 0.0001, "V": 0.0694 },
    "125": { "A": 0.0031, "G": 0.0001, "H": 0.0001, "I": 0.0014, "N": 0.0017, "P": 0.002, "S": 0.002, "T": 0.9897 },
    "126": { "A": 0.0013, "D": 0.0001, "F": 0.0001, "G": 0.0059, "I": 0.0011, "L": 0.0001, "R": 0.0001, "V": 0.9913 },
    "127": { "F": 0.0001, "I": 0.0017, "L": 0.9973, "P": 0.0001, "V": 0.0009 }
  },
  "IGLV7": {
    "1": { "H": 0.0088, "Q": 0.9912 },
    "2": { "A": 0.559, "D": 0.0036, "E": 0.0001, "G": 0.0006, "H": 0.0002, "I": 0.001, "L": 0.0002, "N": 0.0008, "P": 0.0244, "S": 0.0077, "T": 0.4002, "V": 0.002 },
    "3": { "A": 0.0056, "E": 0.0021, "F": 0.0002, "G": 0.0031, "I": 0.0001, "K": 0.0001, "L": 0.0061, "M": 0.0012, "S": 0.0002, "V": 0.9813, "W": 0.0001 },
    "4": { "A": 0.0062, "C": 0.0001, "D": 0.0001, "E": 0.0015, "F": 0.0009, "G": 0.0043, "I": 0.0001, "L": 0.0144, "M": 0.0013, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0003, "T": 0.0001, "V": 0.9703, "W": 0.0001 },
    "5": { "A": 0.0016, "D": 0.0001, "F": 0.0003, "H": 0.0002, "I": 0.0015, "N": 0.0024, "P": 0.0281, "R": 0.0001, "S": 0.0024, "T": 0.9633 },
    "6": { "D": 0.0001, "E": 0.0003, "H": 0.0108, "K": 0.0038, "L": 0.0025, "N": 0.0001, "P": 0.0283, "Q": 0.952, "R": 0.0014, "T": 0.0005, "Y": 0.0001 },
    "7": { "A": 0.0371, "C": 0.0001, "D": 0.0103, "E": 0.929, "G": 0.0022, "H": 0.0003, "K": 0.0027, "L": 0.0003, "N": 0.0001, "P": 0.0008, "Q": 0.0107, "R": 0.0002, "S": 0.0015, "T": 0.0006, "V": 0.0037, "W": 0.0001, "Y": 0.0005 },
    "8": { "A": 0.0043, "C": 0.0001, "H": 0.0096, "L": 0.0046, "P": 0.9657, "Q": 0.0001, "R": 0.002, "S": 0.0082, "T": 0.0055 },
    "9": { "A": 0.0082, "E": 0.0001, "H": 0.0001, "L": 0.0013, "P": 0.0107, "Q": 0.0001, "S": 0.9774, "T": 0.0011, "V": 0.0001, "Y": 0.0009 },
    "11": { "A": 0.0001, "F": 0.0005, "H": 0.0001, "I": 0.0001, "L": 0.9783, "M": 0.0044, "P": 0.0034, "Q": 0.0006, "R": 0.0031, "V": 0.0094, "W": 0.0001 },
    "12": { "A": 0.0021, "D": 0.0001, "H": 0.0002, "I": 0.0016, "K": 0.0001, "L": 0.0002, "N": 0.0021, "P": 0.0473, "Q": 0.0001, "R": 0.0001, "S": 0.0043, "T": 0.9418, "Y": 0.0001 },
    "13": { "A": 0.0048, "C": 0.0001, "E": 0.0012, "F": 0.0005, "G": 0.0031, "I": 0.0001, "L": 0.017, "M": 0.0016, "P": 0.0001, "Q": 0.0001, "R": 0.0001, "S": 0.0002, "V": 0.9711, "W": 0.0001 },
    "14": { "A": 0.0043, "C": 0.0012, "D": 0.0001, "F": 0.0025, "H": 0.0007, "L": 0.0003, "P": 0.0084, "S": 0.9737, "T": 0.0021, "V": 0.0001, "Y": 0.0067 },
    "15": { "A": 0.0013, "E": 0.0001, "H": 0.001, "I": 0.0001, "L": 0.0029, "P": 0.9754, "Q": 0.0073, "R": 0.0011, "S": 0.003, "T": 0.0077 },
    "16": { "A": 0.0094, "C": 0.0003, "D": 0.0013, "E": 0.0029, "F": 0.0001, "G": 0.9721, "H": 0.0001, "L": 0.0002, "R": 0.007, "S": 0.0006, "V": 0.006, "W": 0.0001 },
    "17": { "A": 0.0043, "C": 0.0006, "D": 0.0003, "E": 0.0065, "G": 0.9726, "L": 0.0001, "N": 0.0001, "Q": 0.0002, "R": 0.0055, "V": 0.0041, "W": 0.0056 },
    "18": { "A": 0.0044, "E": 0.0001, "G": 0.0001, "I": 0.0009, "K": 0.0038, "L": 0.0001, "N": 0.0001, "P": 0.0323, "Q": 0.0006, "R": 0.0005, "S": 0.002, "T": 0.9551 },
    "19": { "A": 0.007, "C": 0.0001, "D": 0.0015, "F": 0.0025, "G": 0.007, "I": 0.0034, "L": 0.0017, "P": 0.0002, "R": 0.0001, "S": 0.0001, "V": 0.9765 },
    "20": { "A": 0.0014, "D": 0.0001, "H": 0.0006, "I": 0.0038, "K": 0.0002, "L": 0.0003, "M": 0.0001, "N": 0.0016, "P": 0.0338, "S": 0.0024, "T": 0.9557, "V": 0.0001 },
    "21": { "A": 0.0001, "C": 0.0001, "F": 0.0012, "H": 0.003, "I": 0.0041, "L": 0.9475, "M": 0.0001, "P": 0.0327, "R": 0.0081, "S": 0.0006, "T": 0.0008, "V": 0.0018 },
    "22": { "A": 0.0017, "H": 0.0002, "I": 0.0006, "K": 0.0001, "L": 0.0001, "N": 0.0024, "P": 0.0342, "R": 0.0001, "S": 0.0021, "T": 0.9584, "V": 0.0001, "Y": 0.0001 },
    "23": { "C": 0.9685, "F": 0.0073, "G": 0.0057, "L": 0.001, "N": 0.0001, "P": 0.0003, "R": 0.0057, "S": 0.0059, "T": 0.0001, "V": 0.0006, "W": 0.0036, "Y": 0.0013 },
    "24": { "A": 0.4115, "C": 0.0089, "D": 0.0098, "E": 0.0006, "F": 0.0002, "G": 0.5347, "H": 0.0001, "I": 0.0001, "L": 0.0003, "N": 0.0006, "P": 0.0053, "Q": 0.0001, "R": 0.0045, "S": 0.013, "T": 0.0045, "V": 0.0057, "Y": 0.0002 },
    "25": { "A": 0.0076, "C": 0.0031, "D": 0.0001, "F": 0.0044, "G": 0.0001, "H": 0.0004, "I": 0.0001, "L": 0.0013, "P": 0.023, "R": 0.0001, "S": 0.9537, "T": 0.0037, "W": 0.0001, "Y": 0.0024 },
    "26": { "A": 0.0031, "C": 0.0006, "D": 0.0026, "E": 0.0002, "F": 0.0003, "G": 0.0022, "H": 0.0006, "I": 0.0043, "L": 0.0002, "M": 0.0001, "N": 0.0191, "P": 0.0003, "R": 0.0187, "S": 0.9351, "T": 0.012, "V": 0.0007, "Y": 0.0001 },
    "27": { "A": 0.0105, "C": 0.0001, "D": 0.0003, "F": 0.0006, "G": 0.0005, "H": 0.0009, "I": 0.0034, "L": 0.0005, "M": 0.0001, "N": 0.0029, "P": 0.0309, "R": 0.0003, "S": 0.0117, "T": 0.9373, "V": 0.0002 },
    "28": { "A": 0.0123, "C": 0.0005, "D": 0.0008, "E": 0.0059, "G": 0.9651, "I": 0.0001, "L": 0.0001, "P": 0.0001, "Q": 0.0001, "R": 0.0059, "S": 0.0003, "T": 0.0001, "V": 0.0087, "W": 0.0001 },
    "29": { "A": 0.9085, "D": 0.0052, "E": 0.0024, "G": 0.0047, "H": 0.0004, "I": 0.0011, "L": 0.0008, "M": 0.0002, "N": 0.0001, "P": 0.0403, "Q": 0.0007, "R": 0.0002, "S": 0.0076, "T": 0.0208, "V": 0.0069 },
    "30": { "A": 0.0079, "C": 0.0005, "D": 0.0008, "F": 0.0046, "G": 0.0061, "H": 0.0001, "I": 0.002, "L": 0.0025, "P": 0.0001, "R": 0.0003, "S": 0.0002, "T": 0.0004, "V": 0.9745 },
    "31": { "A": 0.0045, "D": 0.0001, "E": 0.0001, "F": 0.0001, "G": 0.0013, "H": 0.0008, "I": 0.0051, "K": 0.0001, "L": 0.0001, "N": 0.0061, "P": 0.0308, "R": 0.0001, "S": 0.008, "T": 0.9427, "V": 0.0001, "Y": 0.0003 },
    "35": { "A": 0.0007, "C": 0.0008, "D": 0.0013, "E": 0.0002, "F": 0.0008, "G": 0.0082, "H": 0.0005, "I": 0.0059, "K": 0.0005, "L": 0.0006, "M": 0.0003, "N": 0.0093, "P": 0.0031, "Q": 0.0001, "R": 0.0286, "S": 0.9265, "T": 0.0112, "V": 0.0003, "W": 0.0001, "Y": 0.001 },
    "36": { "A": 0.0117, "C": 0.0059, "D": 0.0272, "E": 0.0006, "F": 0.0001, "G": 0.9118, "H": 0.002, "I": 0.0006, "K": 0.0004, "L": 0.0008, "M": 0.0001, "N": 0.0057, "P": 0.0004, "Q": 0.0001, "R": 0.0052, "S": 0.0128, "T": 0.0057, "V": 0.0071, "W": 0.0003, "Y": 0.0013 },
    "37": { "A": 0.0003, "C": 0.0018, "D": 0.0029, "F": 0.0077, "H": 0.5661, "K": 0.0001, "L": 0.0056, "M": 0.0002, "N": 0.0158, "P": 0.0147, "Q": 0.0073, "R": 0.0026, "S": 0.0133, "T": 0.0015, "W": 0.0001, "Y": 0.3599 },
    "38": { "A": 0.0004, "C": 0.0047, "D": 0.0061, "E": 0.0004, "F": 0.0283, "G": 0.0001, "H": 0.0158, "I": 0.0005, "K": 0.0001, "L": 0.0017, "M": 0.0003, "N": 0.0029, "P": 0.0008, "Q": 0.0008, "R": 0.0008, "S": 0.0237, "T": 0.0008, "V": 0.0002, "W": 0.0065, "Y": 0.9053 },
    "39": { "A": 0.0141, "C": 0.0003, "D": 0.0001, "E": 0.0001, "H": 0.0034, "I": 0.0001, "L": 0.0015, "P": 0.9649, "Q": 0.0025, "R": 0.0006, "S": 0.0036, "T": 0.0065, "V": 0.0023 },
    "40": { "A": 0.0003, "C": 0.0034, "D": 0.0037, "E": 0.0001, "F": 0.0047, "G": 0.0006, "H": 0.0135, "I": 0.0015, "K": 0.0008, "L": 0.0008, "M": 0.0001, "N": 0.3868, "P": 0.0006, "Q": 0.0004, "R": 0.0005, "S": 0.0239, "T": 0.0128, "V": 0.0001, "Y": 0.5453 },
    "41": { "A": 0.0001, "C": 0.0098, "F": 0.0001, "G": 0.0066, "L": 0.0038, "R": 0.0062, "S": 0.0016, "W": 0.9717, "Y": 0.0001 },
    "42": { "A": 0.0001, "C": 0.0045, "F": 0.9396, "H": 0.0001, "I": 0.0087, "L": 0.0268, "M": 0.0003, "N": 0.0001, "P": 0.0003, "S": 0.0076, "T": 0.0001, "V": 0.0087, "W": 0.0001, "Y": 0.0029 },
    "43": { "E": 0.0002, "H": 0.0097, "K": 0.0041, "L": 0.0015, "N": 0.0001, "P": 0.0091, "Q": 0.9728, "R": 0.002, "S": 0.0001, "T": 0.0003, "Y": 0.0001 },
    "44": { "A": 0.0001, "D": 0.0001, "E": 0.0006, "H": 0.0084, "K": 0.0068, "L": 0.0034, "M": 0.0001, "N": 0.0002, "P": 0.0078, "Q": 0.9664, "R": 0.0048, "S": 0.0001, "T": 0.0003, "V": 0.0006, "W": 0.0001, "Y": 0.0001 },
    "45": { "A": 0.0001, "E": 0.0018, "G": 0.0001, "H": 0.0002, "I": 0.0004, "K": 0.9406, "L": 0.0001, "M": 0.0012, "N": 0.0136, "Q": 0.0079, "R": 0.0211, "S": 0.0003, "T": 0.0122, "W": 0.0001, "Y": 0.0002 },
    "46": { "A": 0.0034, "D": 0.0001, "G": 0.0001, "H": 0.0037, "L": 0.001, "P": 0.9798, "R": 0.001, "S": 0.0056, "T": 0.0046, "V": 0.0008 },
    "47": { "A": 0.0039, "C": 0.0029, "D": 0.0015, "E": 0.0013, "G": 0.9825, "K": 0.0001, "P": 0.0001, "R": 0.0017, "S": 0.0015, "T": 0.0001, "V": 0.0044 },
    "48": { "E": 0.0023, "H": 0.0141, "K": 0.0041, "L": 0.0008, "N": 0.0003, "P": 0.0082, "Q": 0.9678, "R": 0.0022, "T": 0.0002 },
    "49": { "A": 0.943, "D": 0.0026, "E": 0.002, "G": 0.0035, "H": 0.0002, "K": 0.0004, "L": 0.0003, "N": 0.0001, "P": 0.0208, "Q": 0.0001, "R": 0.0002, "S": 0.0047, "T": 0.0073, "V": 0.0147, "Y": 0.0001 },
    "50": { "A": 0.0004, "H": 0.005, "I": 0.0001, "L": 0.0022, "P": 0.9849, "Q": 0.0001, "R": 0.0009, "S": 0.0013, "T": 0.0051, "V": 0.0001 },
    "51": { "A": 0.0001, "C": 0.0001, "D": 0.0001, "E": 0.0001, "G": 0.0022, "I": 0.0005, "K": 0.0158, "L": 0.0013, "M": 0.0082, "N": 0.0003, "P": 0.0006, "Q": 0.0008, "R": 0.9491, "S": 0.0076, "T": 0.0109, "V": 0.0008, "W": 0.0015 },
    "52": { "A": 0.3684, "D": 0.0005, "E": 0.0027, "F": 0.0001, "G": 0.0021, "H": 0.0002, "I": 0.0033, "K": 0.0043, "L": 0.001, "N": 0.0006, "P": 0.0215, "Q": 0.0004, "R": 0.0006, "S": 0.0192, "T": 0.5733, "V": 0.0018, "Y": 0.0001 },
    "53": { "C": 0.0001, "F": 0.0003, "G": 0.0001, "H": 0.0001, "I": 0.0003, "L": 0.9793, "M": 0.0038, "P": 0.0022, "Q": 0.0012, "R": 0.0086, "V": 0.004, "W": 0.0001 },
    "54": { "F": 0.004, "G": 0.0002, "I": 0.9608, "L": 0.0105, "M": 0.0088, "N": 0.0022, "P": 0.0001, "R": 0.0003, "S": 0.0054, "T": 0.0016, "V": 0.0059, "Y": 0.0001 },
    "55": { "A": 0.0003, "C": 0.002, "D": 0.0052, "E": 0.0005, "F": 0.0209, "G": 0.0001, "H": 0.0084, "I": 0.0001, "L": 0.0003, "N": 0.0012, "P": 0.0001, "Q": 0.0002, "R": 0.0001, "S": 0.0132, "T": 0.0003, "V": 0.0003, "W": 0.0001, "Y": 0.9466 },
    "56": { "A": 0.0108, "C": 0.0004, "D": 0.5533, "E": 0.0082, "G": 0.0098, "H": 0.0031, "I": 0.0016, "K": 0.0004, "L": 0.0001, "M": 0.0002, "N": 0.0165, "P": 0.0003, "Q": 0.0002, "R": 0.0094, "S": 0.374, "T": 0.0054, "V": 0.0026, "Y": 0.0035 },
    "57": { "A": 0.0121, "E": 0.0001, "F": 0.0001, "G": 0.0046, "H": 0.0001, "I": 0.0103, "K": 0.0054, "L": 0.0003, "M": 0.0003, "N": 0.0003, "P": 0.0108, "Q": 0.0005, "R": 0.0007, "S": 0.0098, "T": 0.9431, "V": 0.0015, "Y": 0.0001 },
    "65": { "A": 0.002, "C": 0.0005, "D": 0.0144, "E": 0.001, "F": 0.0024, "G": 0.0077, "H": 0.0005, "I": 0.0068, "K": 0.0015, "L": 0.0006, "N": 0.0507, "P": 0.0007, "Q": 0.0003, "R": 0.0218, "S": 0.8329, "T": 0.0507, "V": 0.0017, "W": 0.0001, "Y": 0.0036 },
    "66": { "A": 0.0003, "D": 0.0076, "E": 0.0005, "F": 0.0003, "G": 0.0006, "H": 0.0105, "I": 0.0074, "K": 0.0312, "L": 0.0009, "M": 0.0009, "N": 0.9009, "P": 0.0002, "Q": 0.0011, "R": 0.0079, "S": 0.0089, "T": 0.0117, "V": 0.0015, "Y": 0.0075 },
    "67": { "A": 0.0001, "D": 0.0001, "E": 0.0022, "G": 0.0008, "H": 0.001, "I": 0.0017, "K": 0.8897, "L": 0.0001, "M": 0.0001, "N": 0.0237, "P": 0.0003, "Q": 0.0133, "R": 0.0475, "S": 0.0012, "T": 0.0179, "V": 0.0003, "Y": 0.0001 },
    "68": { "A": 0.0005, "C": 0.0003, "D": 0.0028, "E": 0.0015, "F": 0.0017, "G": 0.0003, "H": 0.8848, "I": 0.0001, "K": 0.0009, "L": 0.0077, "M": 0.0006, "N": 0.009, "P": 0.042, "Q": 0.0172, "R": 0.0077, "S": 0.0007, "T": 0.0017, "V": 0.0002, "W": 0.0001, "Y": 0.02 },
    "69": { "A": 0.0342, "C": 0.0003, "D": 0.0001, "F": 0.003, "G": 0.0002, "H": 0.0001, "I": 0.0001, "K": 0.0005, "L": 0.0006, "P": 0.0581, "Q": 0.0001, "R": 0.0003, "S": 0.8869, "T": 0.0126, "V": 0.0001, "W": 0.0002, "Y": 0.0024 },
    "70": { "A": 0.0008, "C": 0.0107, "D": 0.0001, "E": 0.0001, "F": 0.0002, "G": 0.021, "L": 0.0094, "M": 0.0001, "P": 0.0005, "R": 0.0103, "S": 0.0048, "T": 0.0002, "V": 0.0011, "W": 0.9405, "Y": 0.0001 },
    "71": { "A": 0.007, "E": 0.0001, "H": 0.0003, "I": 0.0021, "K": 0.0017, "N": 0.0024, "P": 0.0116, "Q": 0.0004, "R": 0.0006, "S": 0.0018, "T": 0.9718 },
    "72": { "A": 0.001, "F": 0.0002, "H": 0.0021, "L": 0.0017, "P": 0.9909, "Q": 0.0002, "R": 0.0006, "S": 0.0019, "T": 0.0015 },
    "74": { "A": 0.9627, "C": 0.0002, "D": 0.0096, "E": 0.0019, "G": 0.0048, "H": 0.0001, "L": 0.0001, "P": 0.0047, "R": 0.0002, "S": 0.0071, "T": 0.0026, "V": 0.0059, "Y": 0.0001 },
    "75": { "A": 0.0001, "C": 0.0002, "E": 0.0001, "G": 0.0014, "H": 0.0008, "L": 0.0066, "P": 0.002, "Q": 0.0016, "R": 0.9853, "S": 0.0001, "W": 0.0018 },
    "76": { "C": 0.0096, "F": 0.9754, "I": 0.0008, "L": 0.0055, "S": 0.0017, "V": 0.0063, "Y": 0.0008 },
    "77": { "A": 0.0118, "C": 0.0001, "E": 0.0001, "F": 0.0001, "H": 0.0001, "I": 0.0001, "K": 0.0001, "L": 0.001, "P": 0.0019, "Q": 0.0001, "R": 0.0001, "S": 0.9819, "T": 0.0026, "V": 0.0001, "Y": 0.0001 },
    "78": { "A": 0.0095, "C": 0.0041, "D": 0.0013, "G": 0.9789, "I": 0.0001, "P": 0.0002, "R": 0.0008, "S": 0.0015, "V": 0.0037 },
    "79": { "A": 0.0147, "C": 0.001, "D": 0.0001, "E": 0.0003, "F": 0.005, "G": 0.0001, "H": 0.0001, "L": 0.0003, "N": 0.0001, "P": 0.0037, "S": 0.9703, "T": 0.0014, "V": 0.0001, "Y": 0.0029 },
    "80": { "C": 0.0002, "F": 0.0036, "G": 0.0002, "H": 0.0017, "I": 0.0049, "K": 0.0001, "L": 0.9563, "P": 0.0053, "R": 0.0217, "S": 0.0003, "T": 0.0001, "V": 0.0055 },
    "83": { "A": 0.0002, "C": 0.0002, "F": 0.0038, "G": 0.0006, "H": 0.0013, "I": 0.0056, "L": 0.9486, "M": 0.0003, "P": 0.0026, "Q": 0.0001, "R": 0.0236, "S": 0.0007, "T": 0.0001, "V": 0.0122, "Y": 0.0001 },
    "84": { "A": 0.0024, "C": 0.0001, "D": 0.0015, "E": 0.0029, "G": 0.9846, "L": 0.0001, "N": 0.0001, "R": 0.0017, "S": 0.0001, "V": 0.0031, "W": 0.0034 },
    "85": { "A": 0.004, "C": 0.0039, "D": 0.0303, "E": 0.0003, "G": 0.955, "H": 0.0001, "R": 0.0004, "S": 0.0027, "T": 0.0001, "V": 0.0029, "Y": 0.0002 },
    "86": { "D": 0.0003, "E": 0.0029, "G": 0.0001, "H": 0.0001, "I": 0.0008, "K": 0.9674, "M": 0.0002, "N": 0.0051, "Q": 0.0029, "R": 0.0131, "S": 0.0005, "T": 0.0066 },
    "87": { "A": 0.9835, "D": 0.0015, "E": 0.0001, "F": 0.0001, "G": 0.0034, "L": 0.0001, "P": 0.0031, "R": 0.0001, "S": 0.0028, "T": 0.001, "V": 0.0044 },
    "88": { "A": 0.9772, "C": 0.0001, "D": 0.0024, "E": 0.0005, "G": 0.0008, "I": 0.0002, "P": 0.0004, "S": 0.0037, "T": 0.0055, "V": 0.0087, "Y": 0.0004 },
    "89": { "A": 0.0001, "F": 0.0002, "H": 0.0001, "I": 0.0002, "L": 0.9906, "M": 0.0022, "P": 0.0015, "Q": 0.0003, "R": 0.0038, "T": 0.0001, "V": 0.0009 },
    "90": { "A": 0.0022, "C": 0.0001, "I": 0.0019, "K": 0.0006, "N": 0.0013, "P": 0.0026, "Q": 0.0001, "R": 0.0003, "S": 0.0019, "T": 0.9889, "V": 0.0002 },
    "91": { "F": 0.001, "G": 0.0001, "H": 0.0001, "I": 0.0018, "L": 0.9785, "M": 0.0005, "P": 0.0011, "Q": 0.0002, "R": 0.0067, "S": 0.0001, "T": 0.0001, "V": 0.0097 },
    "92": { "A": 0.0134, "G": 0.0001, "L": 0.015, "P": 0.0011, "S": 0.9645, "T": 0.0044, "V": 0.0003, "W": 0.0012 },
    "93": { "A": 0.004, "C": 0.0033, "D": 0.0075, "E": 0.0004, "F": 0.0001, "G": 0.9677, "H": 0.0003, "K": 0.0001, "L": 0.0001, "N": 0.0037, "P": 0.0006, "Q": 0.0001, "R": 0.0016, "S": 0.0058, "T": 0.0012, "V": 0.0035, "Y": 0.0001 },
    "94": { "A": 0.5865, "E": 0.0012, "F": 0.0001, "G": 0.002, "I": 0.0001, "L": 0.0012, "M": 0.0001, "P": 0.0002, "R": 0.0001, "S": 0.0004, "T": 0.0004, "V": 0.4076, "W": 0.0001 },
    "95": { "D": 0.0001, "E": 0.0015, "H": 0.0037, "K": 0.0014, "L": 0.006, "M": 0.0001, "P": 0.0018, "Q": 0.9743, "R": 0.0082, "S": 0.0001, "T": 0.0001, "W": 0.0027, "Y": 0.0001 },
    "96": { "A": 0.0103, "C": 0.0001, "D": 0.0002, "E": 0.0001, "F": 0.0008, "G": 0.0001, "H": 0.0038, "I": 0.0008, "L": 0.0025, "P": 0.9713, "Q": 0.0003, "R": 0.0014, "S": 0.0039, "T": 0.0041, "V": 0.0003 },
    "97": { "A": 0.0029, "D": 0.0158, "E": 0.9758, "G": 0.0033, "K": 0.0006, "M": 0.0001, "Q": 0.0004, "R": 0.0001, "S": 0.0001, "V": 0.0008 },
    "98": { "A": 0.0016, "D": 0.99, "E": 0.0043, "G": 0.0013, "H": 0.0001, "M": 0.0001, "N": 0.0006, "S": 0.0001, "V": 0.0006, "Y": 0.0013 },
    "99": { "A": 0.002, "D": 0.007, "E": 0.9847, "G": 0.0033, "H": 0.0001, "K": 0.0004, "Q": 0.0017, "R": 0.0001, "S": 0.0001, "V": 0.0004, "Y": 0.0001 },
    "100": { "A": 0.9809, "D": 0.0062, "E": 0.0001, "G": 0.0047, "L": 0.0001, "P": 0.0003, "S": 0.0041, "T": 0.0022, "V": 0.0008, "Y": 0.0004 },
    "101": { "A": 0.0034, "C": 0.0001, "D": 0.0998, "E": 0.8737, "F": 0.0003, "G": 0.002, "H": 0.0002, "I": 0.0003, "K": 0.0024, "L": 0.001, "M": 0.0004, "N": 0.0017, "Q": 0.0013, "R": 0.0001, "S": 0.0002, "T": 0.0011, "V": 0.0085, "W": 0.0001, "Y": 0.0033 },
    "102": { "C": 0.0006, "D": 0.0034, "E": 0.0003, "F": 0.0015, "H": 0.0008, "I": 0.0001, "L": 0.0001, "N": 0.0002, "R": 0.0001, "S": 0.0008, "V": 0.0001, "Y": 0.992 },
    "103": { "A": 0.0001, "C": 0.002, "D": 0.0045, "F": 0.0156, "G": 0.0001, "H": 0.0043, "I": 0.0002, "L": 0.0003, "N": 0.0001, "P": 0.0001, "Q": 0.0001, "S": 0.0022, "T": 0.0001, "V": 0.0003, "Y": 0.9701 },
    "104": { "A": 0.0001, "C": 0.9902, "F": 0.0012, "G": 0.0041, "R": 0.0008, "S": 0.0015, "W": 0.0008, "Y": 0.0013 },
    "105": { "A": 0.0008, "C": 0.0014, "F": 0.0128, "G": 0.0001, "H": 0.0002, "I": 0.0012, "K": 0.0001, "L": 0.9547, "M": 0.007, "P": 0.0008, "Q": 0.0024, "R": 0.0026, "S": 0.0076, "T": 0.0002, "V": 0.0052, "W": 0.0025, "Y": 0.0002 },
    "106": { "A": 0.0002, "C": 0.0001, "F": 0.005, "G": 0.0001, "H": 0.0024, "I": 0.0066, "L": 0.9641, "M": 0.0001, "P": 0.0017, "R": 0.0054, "S": 0.0012, "T": 0.0001, "V": 0.0126, "W": 0.0001, "Y": 0.0001 },
    "107": { "A": 0.0082, "C": 0.0032, "D": 0.0025, "E": 0.0006, "F": 0.0219, "G": 0.0005, "H": 0.0073, "I": 0.0002, "K": 0.0002, "L": 0.0027, "M": 0.0013, "N": 0.0013, "P": 0.001, "Q": 0.0016, "R": 0.0003, "S": 0.5366, "T": 0.0067, "V": 0.002, "W": 0.0057, "Y": 0.3964 },
    "108": { "A": 0.0005, "C": 0.0068, "D": 0.0121, "E": 0.0007, "F": 0.0278, "G": 0.0002, "H": 0.008, "I": 0.0008, "K": 0.0001, "L": 0.0015, "N": 0.0048, "P": 0.0014, "Q": 0.0018, "R": 0.0019, "S": 0.0097, "T": 0.0006, "V": 0.0015, "W": 0.0006, "Y": 0.9194 },
    "109": { "A": 0.0086, "C": 0.0015, "D": 0.0099, "E": 0.0001, "F": 0.0005, "G": 0.41, "H": 0.0006, "I": 0.0029, "K": 0.0008, "L": 0.0004, "M": 0.0003, "N": 0.0121, "P": 0.002, "Q": 0.0001, "R": 0.0139, "S": 0.5219, "T": 0.0112, "V": 0.002, "W": 0.0001, "Y": 0.0009 },
    "110": { "A": 0.0088, "C": 0.0045, "D": 0.0204, "E": 0.0006, "G": 0.9215, "H": 0.0003, "I": 0.0036, "L": 0.0016, "N": 0.0039, "P": 0.0016, "R": 0.001, "S": 0.014, "T": 0.0013, "V": 0.0162, "Y": 0.0006 },
    "112": { "A": 0.7276, "D": 0.0037, "E": 0.0018, "F": 0.0219, "G": 0.075, "I": 0.0037, "M": 0.0037, "N": 0.0055, "P": 0.0366, "Q": 0.0055, "R": 0.0384, "S": 0.011, "T": 0.0055, "V": 0.0475, "W": 0.0055, "Y": 0.0073 },
    "113": { "A": 0.3337, "C": 0.001, "D": 0.0156, "E": 0.001, "F": 0.0004, "G": 0.5371, "H": 0.0026, "I": 0.0009, "L": 0.0018, "N": 0.0026, "P": 0.0091, "Q": 0.0275, "R": 0.0335, "S": 0.0157, "T": 0.0099, "V": 0.0064, "W": 0.0001, "Y": 0.0011 },
    "114": { "A": 0.2449, "C": 0.0008, "D": 0.0144, "E": 0.0036, "F": 0.001, "G": 0.4749, "H": 0.0062, "I": 0.0013, "K": 0.0008, "L": 0.0092, "N": 0.0032, "P": 0.03, "Q": 0.0778, "R": 0.0876, "S": 0.0171, "T": 0.0103, "V": 0.0157, "W": 0.0007, "Y": 0.0005 },
    "115": { "A": 0.368, "C": 0.0007, "D": 0.0214, "E": 0.0069, "F": 0.0206, "G": 0.0362, "H": 0.0361, "I": 0.0059, "K": 0.0025, "L": 0.0737, "M": 0.0008, "N": 0.0268, "P": 0.0886, "Q": 0.0609, "R": 0.1165, "S": 0.0419, "T": 0.0173, "V": 0.0435, "W": 0.0028, "Y": 0.0289 },
    "116": { "A": 0.0158, "C": 0.0023, "D": 0.0007, "E": 0.011, "F": 0.0029, "G": 0.1035, "H": 0.008, "I": 0.0003, "K": 0.0097, "L": 0.0489, "M": 0.0033, "N": 0.0014, "P": 0.0415, "Q": 0.044, "R": 0.1951, "S": 0.0052, "T": 0.0006, "V": 0.1733, "W": 0.2698, "Y": 0.0626 },
    "117": { "A": 0.0041, "C": 0.0006, "D": 0.0002, "E": 0.0023, "F": 0.002, "G": 0.0034, "H": 0.0003, "I": 0.0181, "K": 0.0008, "L": 0.0171, "M": 0.0038, "P": 0.0006, "Q": 0.0006, "R": 0.0012, "S": 0.0015, "T": 0.001, "V": 0.9421, "W": 0.0002 },
    "118": { "C": 0.0031, "F": 0.9865, "G": 0.0001, "I": 0.0008, "K": 0.0001, "L": 0.0031, "Q": 0.0001, "S": 0.0008, "V": 0.0043, "Y": 0.001 },
    "119": { "A": 0.0003, "C": 0.0009, "D": 0.0018, "E": 0.0001, "G": 0.995, "L": 0.0001, "P": 0.0001, "R": 0.0004, "S": 0.0006, "V": 0.0008 },
    "120": { "A": 0.0018, "D": 0.0003, "E": 0.0021, "G": 0.8876, "L": 0.0001, "N": 0.0001, "P": 0.0031, "R": 0.0026, "S": 0.0027, "T": 0.0988, "V": 0.0008 },
    "121": { "A": 0.0002, "E": 0.0007, "G": 0.9953, "R": 0.0007, "V": 0.0011, "W": 0.002 },
    "122": { "A": 0.0012, "I": 0.0011, "N": 0.0005, "P": 0.0017, "S": 0.001, "T": 0.9946 },
    "123": { "A": 0.0001, "D": 0.0001, "E": 0.0019, "F": 0.0004, "H": 0.0005, "I": 0.0003, "K": 0.952, "L": 0.0003, "M": 0.0013, "N": 0.0046, "Q": 0.0158, "R": 0.0162, "S": 0.0015, "T": 0.0047, "V": 0.0001, "W": 0.0001 },
    "124": { "D": 0.0001, "F": 0.0006, "G": 0.0001, "I": 0.0013, "L": 0.8802, "M": 0.0006, "P": 0.0005, "Q": 0.0002, "R": 0.0009, "V": 0.1154, "W": 0.0001 },
    "125": { "A": 0.0036, "D": 0.0001, "F": 0.0003, "H": 0.0001, "I": 0.002, "K": 0.0002, "L": 0.0001, "N": 0.0006, "P": 0.0026, "R": 0.0001, "S": 0.007, "T": 0.9834 },
    "126": { "A": 0.0015, "D": 0.0004, "F": 0.0003, "G": 0.0054, "I": 0.0012, "L": 0.0003, "R": 0.0001, "V": 0.9908 },
    "127": { "I": 0.0017, "L": 0.9968, "P": 0.0001, "V": 0.0013 }
  },
  "IGLV8": {
    "1": { "H": 0.0076, "Q": 0.9924 },
    "2": { "A": 0.0069, "C": 0.0001, "D": 0.0003, "H": 0.0003, "I": 0.0025, "K": 0.0002, "L": 0.0002, "N": 0.003, "P": 0.0303, "R": 0.0001, "S": 0.0039, "T": 0.9521 },
    "3": { "A": 0.0047, "E": 0.0032, "F": 0.0002, "G": 0.0036, "I": 0.0002, "K": 0.0001, "L": 0.0112, "M": 0.0034, "P": 0.0001, "V": 0.9733, "W": 0.0001 },
    "4": { "A": 0.0043, "C": 0.0001, "E": 0.0006, "F": 0.0009, "G": 0.0036, "I": 0.0002, "L": 0.0072, "M": 0.0015, "P": 0.0001, "Q": 0.0002, "R": 0.0001, "S": 0.0001, "V": 0.9812 },
    "5": { "A": 0.0017, "D": 0.0002, "H": 0.0006, "I": 0.0021, "K": 0.0002, "L": 0.0002, "M": 0.0001, "N": 0.0058, "P": 0.0156, "R": 0.0002, "S": 0.0032, "T": 0.9703 },
    "6": { "C": 0.0001, "E": 0.001, "H": 0.013, "K": 0.0044, "L": 0.0016, "P": 0.0215, "Q": 0.9556, "R": 0.0019, "S": 0.0002, "T": 0.0006, "Y": 0.0001 },
    "7": { "A": 0.036, "C": 0.0001, "D": 0.0115, "E": 0.932, "G": 0.0027, "H": 0.0002, "K": 0.0015, "L": 0.0002, "N": 0.0001, "P": 0.0003, "Q": 0.0074, "R": 0.0002, "S": 0.0017, "V": 0.0051, "W": 0.0002, "Y": 0.0009 },
    "8": { "A": 0.0039, "H": 0.0008, "L": 0.0033, "P": 0.9645, "Q": 0.0063, "R": 0.0016, "S": 0.0104, "T": 0.0092 },
    "9": { "A": 0.009, "E": 0.0002, "K": 0.0002, "L": 0.0031, "P": 0.0042, "Q": 0.0001, "R": 0.0002, "S": 0.9788, "T": 0.0027, "W": 0.0017 },
    "11": { "A": 0.0005, "C": 0.0093, "D": 0.0002, "F": 0.8792, "G": 0.0005, "H": 0.0001, "I": 0.0084, "L": 0.0704, "M": 0.0003, "P": 0.0006, "R": 0.0006, "S": 0.0103, "T": 0.0002, "V": 0.0171, "W": 0.0002, "Y": 0.0021 },
    "12": { "A": 0.0134, "E": 0.0006, "I": 0.0002, "L": 0.0019, "P": 0.0122, "Q": 0.0002, "S": 0.9682, "T": 0.0025, "Y": 0.0006 },
    "13": { "A": 0.0036, "C": 0.0002, "D": 0.0001, "E": 0.0009, "F": 0.0003, "G": 0.0047, "I": 0.0003, "L": 0.0072, "M": 0.0005, "Q": 0.0001, "S": 0.0001, "T": 0.0002, "V": 0.9817, "W": 0.0002 },
    "14": { "A": 0.0054, "C": 0.0012, "D": 0.0002, "F": 0.0037, "G": 0.0002, "H": 0.0002, "L": 0.0002, "P": 0.0071, "S": 0.9709, "T": 0.0021, "V": 0.0002, "W": 0.0001, "Y": 0.0087 },
    "15": { "A": 0.0023, "C": 0.0001, "F": 0.0001, "H": 0.0058, "L": 0.0031, "N": 0.0001, "P": 0.9717, "Q": 0.0002, "R": 0.0009, "S": 0.0051, "T": 0.0105, "V": 0.0001 },
    "16": { "A": 0.0068, "C": 0.0005, "D": 0.0006, "E": 0.0024, "G": 0.9743, "I": 0.0001, "L": 0.0001, "P": 0.0002, "R": 0.0058, "S": 0.0002, "V": 0.0088, "W": 0.0002 },
    "17": { "A": 0.0059, "C": 0.0006, "D": 0.0001, "E": 0.0062, "G": 0.9589, "I": 0.0001, "K": 0.0028, "L": 0.0003, "P": 0.0001, "Q": 0.0004, "R": 0.0096, "S": 0.0002, "T": 0.0002, "V": 0.0066, "W": 0.0081 },
    "18": { "A": 0.0051, "I": 0.0017, "K": 0.0021, "L": 0.0001, "N": 0.0003, "P": 0.0255, "Q": 0.0008, "R": 0.0009, "S": 0.004, "T": 0.9595, "V": 0.0001 },
    "19": { "A": 0.0049, "C": 0.0002, "D": 0.0015, "F": 0.0043, "G": 0.0045, "I": 0.0107, "L": 0.0039, "P": 0.0002, "R": 0.0001, "S": 0.0006, "T": 0.0002, "V": 0.969 },
    "20": { "A": 0.0012, "E": 0.0001, "H": 0.0003, "I": 0.0028, "K": 0.0028, "N": 0.0006, "P": 0.0261, "Q": 0.0006, "R": 0.0003, "S": 0.0019, "T": 0.9631, "V": 0.0002 },
    "21": { "A": 0.0002, "F": 0.0026, "H": 0.0025, "I": 0.0042, "L": 0.9606, "M": 0.0001, "P": 0.0174, "R": 0.0081, "S": 0.0001, "T": 0.0005, "V": 0.0037 },
    "22": { "A": 0.0017, "D": 0.0001, "G": 0.0001, "H": 0.0008, "I": 0.001, "L": 0.0002, "M": 0.0001, "N": 0.0027, "P": 0.0213, "R": 0.0002, "S": 0.0021, "T": 0.9698 },
    "23": { "A": 0.0002, "C": 0.9694, "F": 0.0111, "G": 0.0036, "I": 0.0001, "L": 0.0005, "P": 0.0002, "R": 0.0046, "S": 0.0058, "V": 0.0002, "W": 0.0032, "Y": 0.0011 },
    "24": { "A": 0.0361, "C": 0.0144, "D": 0.0039, "E": 0.0004, "G": 0.9284, "H": 0.0002, "L": 0.0002, "P": 0.0006, "R": 0.0043, "S": 0.0037, "T": 0.0006, "V": 0.0068, "W": 0.0002, "Y": 0.0002 },
    "25": { "A": 0.0001, "C": 0.0005, "F": 0.0252, "L": 0.9611, "M": 0.0008, "Q": 0.0001, "R": 0.0001, "S": 0.0051, "T": 0.0001, "V": 0.0042, "W": 0.0029 },
    "26": { "A": 0.0025, "C": 0.002, "D": 0.0017, "E": 0.0001, "F": 0.0002, "G": 0.0042, "H": 0.0008, "I": 0.0081, "K": 0.0036, "L": 0.0003, "N": 0.0317, "P": 0.0017, "Q": 0.0001, "R": 0.0364, "S": 0.8654, "T": 0.0408, "V": 0.0001, "Y": 0.0006 },
    "27": { "A": 0.0089, "C": 0.0009, "D": 0.0001, "F": 0.004, "G": 0.0002, "H": 0.0003, "I": 0.0001, "L": 0.0009, "P": 0.0201, "R": 0.0002, "S": 0.9566, "T": 0.0042, "V": 0.0002, "Y": 0.0034 },
    "28": { "A": 0.0115, "C": 0.0182, "D": 0.0089, "E": 0.0001, "F": 0.0001, "G": 0.9455, "H": 0.0002, "L": 0.0002, "P": 0.0002, "R": 0.0052, "S": 0.002, "T": 0.0002, "V": 0.0073, "W": 0.0002, "Y": 0.0002 },
    "29": { "A": 0.0096, "C": 0.0003, "E": 0.0001, "F": 0.0002, "H": 0.0004, "L": 0.0027, "P": 0.0312, "Q": 0.0006, "R": 0.0001, "S": 0.9504, "T": 0.0035, "V": 0.0002, "Y": 0.0009 },
    "30": { "A": 0.0092, "C": 0.0002, "D": 0.0021, "F": 0.0045, "G": 0.008, "H": 0.0002, "I": 0.0029, "L": 0.0108, "R": 0.0002, "S": 0.0005, "T": 0.0001, "V": 0.9612, "Y": 0.0002 },
    "31": { "A": 0.0095, "C": 0.0009, "D": 0.0001, "F": 0.0019, "H": 0.0008, "L": 0.0013, "P": 0.0275, "R": 0.0002, "S": 0.9493, "T": 0.0048, "V": 0.0004, "W": 0.0001, "Y": 0.0033 },
    "35": { "A": 0.015, "C": 0.0002, "D": 0.0006, "F": 0.002, "G": 0.0009, "H": 0.0013, "I": 0.0077, "K": 0.0012, "L": 0.0004, "M": 0.0012, "N": 0.0055, "P": 0.0247, "Q": 0.0013, "R": 0.0027, "S": 0.0405, "T": 0.8918, "V": 0.0017, "Y": 0.0013 },
    "36": { "A": 0.0081, "C": 0.0024, "D": 0.0101, "E": 0.0003, "F": 0.0005, "G": 0.0156, "H": 0.0036, "I": 0.0055, "K": 0.0015, "L": 0.0019, "M": 0.0002, "N": 0.0547, "P": 0.0037, "Q": 0.0002, "R": 0.0296, "S": 0.8022, "T": 0.0574, "V": 0.0011, "Y": 0.0014 },
    "37": { "A": 0.0004, "C": 0.0015, "D": 0.0063, "E": 0.0002, "F": 0.0155, "H": 0.0495, "I": 0.0005, "K": 0.0004, "L": 0.0019, "N": 0.0269, "P": 0.0028, "Q": 0.0009, "R": 0.0002, "S": 0.0498, "T": 0.0008, "V": 0.0003, "W": 0.0004, "Y": 0.8417 },
    "38": { "A": 0.0002, "C": 0.004, "D": 0.0063, "E": 0.0001, "F": 0.0199, "H": 0.0222, "I": 0.0001, "K": 0.0002, "L": 0.0018, "N": 0.0055, "P": 0.0011, "Q": 0.0011, "R": 0.0002, "S": 0.0248, "T": 0.0004, "V": 0.0002, "W": 0.0008, "Y": 0.9111 },
    "39": { "A": 0.0246, "C": 0.0004, "D": 0.0002, "F": 0.0006, "H": 0.0059, "I": 0.0002, "L": 0.0024, "N": 0.0002, "P": 0.9475, "R": 0.0004, "S": 0.0084, "T": 0.007, "V": 0.0021, "Y": 0.0001 },
    "40": { "A": 0.0019, "C": 0.0031, "D": 0.0002, "F": 0.0002, "G": 0.0106, "H": 0.0015, "I": 0.0091, "L": 0.0009, "M": 0.0002, "N": 0.0076, "P": 0.0025, "R": 0.0279, "S": 0.9024, "T": 0.0317, "V": 0.0002, "Y": 0.0002 },
    "41": { "A": 0.0001, "C": 0.0101, "E": 0.0001, "F": 0.0001, "G": 0.0056, "L": 0.0055, "M": 0.0001, "P": 0.0001, "R": 0.007, "S": 0.0026, "V": 0.0007, "W": 0.9677, "Y": 0.0003 },
    "42": { "A": 0.0002, "C": 0.0031, "D": 0.0048, "E": 0.0001, "F": 0.0131, "G": 0.0001, "H": 0.0047, "I": 0.0007, "L": 0.0003, "N": 0.0015, "P": 0.0002, "S": 0.0112, "T": 0.0001, "V": 0.0007, "Y": 0.9592 },
    "43": { "E": 0.0001, "F": 0.0001, "H": 0.0107, "K": 0.0033, "L": 0.0012, "N": 0.0002, "P": 0.0124, "Q": 0.9558, "R": 0.0156, "T": 0.0004, "V": 0.0001, "Y": 0.0001 },
    "44": { "A": 0.0002, "C": 0.0001, "E": 0.0004, "H": 0.0073, "K": 0.0042, "L": 0.0055, "P": 0.0159, "Q": 0.9623, "R": 0.0028, "T": 0.0003, "V": 0.0006, "W": 0.0003 },
    "45": { "A": 0.0081, "F": 0.0001, "H": 0.001, "I": 0.0052, "K": 0.0002, "L": 0.0002, "M": 0.0003, "N": 0.004, "P": 0.0174, "Q": 0.0001, "R": 0.0013, "S": 0.0049, "T": 0.957, "V": 0.0001 },
    "46": { "A": 0.0013, "H": 0.0006, "L": 0.0006, "P": 0.9867, "Q": 0.0037, "R": 0.0006, "S": 0.0017, "T": 0.0047, "V": 0.0002 },
    "47": { "A": 0.0057, "C": 0.0047, "D": 0.0039, "G": 0.9743, "L": 0.0001, "P": 0.0003, "R": 0.0034, "S": 0.0024, "T": 0.0002, "V": 0.005 },
    "48": { "A": 0.0002, "D": 0.0001, "E": 0.0008, "H": 0.0096, "K": 0.0028, "L": 0.0035, "M": 0.0001, "P": 0.0172, "Q": 0.9576, "R": 0.0074, "T": 0.0007, "W": 0.0001 },
    "49": { "A": 0.9192, "D": 0.0032, "E": 0.0002, "F": 0.0001, "G": 0.0013, "H": 0.0004, "I": 0.0001, "M": 0.0006, "N": 0.0002, "P": 0.0415, "S": 0.0152, "T": 0.0164, "V": 0.0017, "Y": 0.0001 },
    "50": { "A": 0.0005, "H": 0.0011, "K": 0.0001, "L": 0.0009, "P": 0.9864, "Q": 0.0052, "R": 0.0008, "S": 0.0017, "T": 0.0033 },
    "51": { "A": 0.0002, "C": 0.001, "D": 0.0001, "G": 0.0009, "H": 0.0029, "K": 0.0001, "L": 0.0087, "M": 0.0001, "P": 0.0074, "Q": 0.0001, "R": 0.9757, "S": 0.0024, "T": 0.0003 },
    "52": { "A": 0.007, "C": 0.0002, "G": 0.0002, "H": 0.0002, "I": 0.0041, "K": 0.0025, "L": 0.0006, "M": 0.0075, "N": 0.0006, "P": 0.0251, "Q": 0.0006, "R": 0.0007, "S": 0.0054, "T": 0.9451, "V": 0.0002, "W": 0.0001 },
    "53": { "A": 0.0002, "C": 0.0004, "D": 0.0001, "F": 0.0013, "G": 0.0001, "H": 0.0017, "I": 0.0061, "L": 0.9594, "M": 0.0001, "N": 0.0002, "P": 0.015, "R": 0.0088, "S": 0.0002, "T": 0.0003, "V": 0.0063 },
    "54": { "A": 0.0002, "C": 0.0004, "D": 0.0002, "F": 0.0031, "G": 0.0009, "H": 0.0002, "I": 0.9219, "L": 0.0208, "M": 0.013, "N": 0.0028, "P": 0.0006, "Q": 0.0001, "R": 0.0006, "S": 0.0102, "T": 0.0046, "V": 0.0205, "W": 0.0001 },
    "55": { "A": 0.0006, "C": 0.0028, "D": 0.0077, "E": 0.0002, "F": 0.0124, "H": 0.0119, "I": 0.0001, "L": 0.0004, "M": 0.0001, "N": 0.0015, "P": 0.0002, "Q": 0.0002, "R": 0.0002, "S": 0.0173, "T": 0.0004, "V": 0.0002, "W": 0.0002, "Y": 0.9438 },
    "56": { "A": 0.0023, "C": 0.0013, "D": 0.0063, "E": 0.0002, "F": 0.0008, "G": 0.0064, "H": 0.0019, "I": 0.0043, "K": 0.0148, "L": 0.0006, "M": 0.0013, "N": 0.0976, "P": 0.0003, "Q": 0.0002, "R": 0.022, "S": 0.8204, "T": 0.0167, "V": 0.0004, "Y": 0.0022 },
    "57": { "A": 0.0107, "C": 0.0001, "D": 0.0002, "E": 0.0001, "G": 0.0038, "H": 0.0002, "I": 0.0034, "K": 0.0035, "L": 0.0002, "M": 0.0001, "N": 0.0002, "P": 0.013, "Q": 0.0008, "R": 0.0007, "S": 0.0024, "T": 0.9602, "V": 0.0004 },
    "65": { "A": 0.0008, "C": 0.0002, "D": 0.0265, "E": 0.0032, "F": 0.0017, "G": 0.001, "H": 0.0106, "I": 0.003, "K": 0.0124, "L": 0.0001, "M": 0.0002, "N": 0.8982, "P": 0.0002, "Q": 0.0006, "R": 0.0014, "S": 0.0161, "T": 0.0179, "V": 0.0004, "W": 0.0006, "Y": 0.0049 },
    "66": { "A": 0.0142, "C": 0.0002, "F": 0.0013, "G": 0.0012, "H": 0.0007, "I": 0.0451, "K": 0.0013, "L": 0.0024, "M": 0.0012, "N": 0.0035, "P": 0.0368, "R": 0.0047, "S": 0.037, "T": 0.8424, "V": 0.0072, "Y": 0.0008 },
    "67": { "A": 0.0002, "C": 0.0013, "G": 0.0013, "H": 0.0028, "I": 0.0001, "L": 0.0087, "P": 0.01, "R": 0.9726, "S": 0.0027, "T": 0.0002, "V": 0.0001, "W": 0.0001 },
    "68": { "A": 0.0145, "C": 0.003, "D": 0.0003, "E": 0.0002, "F": 0.004, "G": 0.0002, "H": 0.0002, "I": 0.0001, "L": 0.0025, "P": 0.0166, "R": 0.0001, "S": 0.9536, "T": 0.0027, "W": 0.0002, "Y": 0.0018 },
    "69": { "A": 0.0225, "C": 0.0009, "D": 0.0002, "F": 0.0102, "G": 0.0001, "H": 0.0001, "L": 0.0006, "N": 0.0001, "P": 0.0056, "Q": 0.0001, "S": 0.9532, "T": 0.0012, "V": 0.0007, "Y": 0.0044 },
    "70": { "A": 0.0035, "C": 0.0007, "D": 0.0001, "E": 0.0029, "G": 0.9722, "L": 0.0003, "P": 0.0001, "Q": 0.0001, "R": 0.0029, "S": 0.0004, "V": 0.0108, "W": 0.006 },
    "71": { "A": 0.0051, "C": 0.0009, "D": 0.0006, "F": 0.0062, "G": 0.0074, "I": 0.0016, "L": 0.002, "S": 0.0002, "V": 0.976 },
    "72": { "A": 0.0009, "G": 0.0001, "H": 0.0017, "L": 0.0017, "N": 0.0001, "P": 0.9849, "Q": 0.0001, "R": 0.0007, "S": 0.0061, "T": 0.0037, "V": 0.0001, "Y": 0.0001 },
    "74": { "A": 0.0206, "C": 0.0009, "D": 0.9292, "E": 0.011, "F": 0.0003, "G": 0.0103, "H": 0.0036, "I": 0.0001, "K": 0.0001, "L": 0.0003, "N": 0.0067, "P": 0.0002, "R": 0.0002, "S": 0.0013, "T": 0.0001, "V": 0.0077, "Y": 0.0076 },
    "75": { "C": 0.019, "F": 0.0002, "G": 0.0007, "H": 0.0012, "I": 0.0002, "L": 0.0065, "P": 0.0018, "R": 0.9663, "S": 0.004, "W": 0.0002 },
    "76": { "C": 0.0115, "F": 0.9677, "G": 0.0006, "I": 0.0026, "L": 0.0045, "S": 0.002, "V": 0.0103, "Y": 0.0008 },
    "77": { "A": 0.0143, "C": 0.0009, "F": 0.0017, "I": 0.0002, "L": 0.0002, "P": 0.0032, "S": 0.9768, "T": 0.0014, "W": 0.0001, "Y": 0.0014 },
    "78": { "A": 0.0039, "C": 0.0073, "D": 0.001, "F": 0.0001, "G": 0.9825, "P": 0.0002, "R": 0.0007, "S": 0.0013, "V": 0.0028, "W": 0.0001 },
    "79": { "A": 0.009, "C": 0.0008, "F": 0.0033, "H": 0.0001, "L": 0.0001, "P": 0.0017, "R": 0.0001, "S": 0.979, "T": 0.0024, "V": 0.0001, "Y": 0.0035 },
    "80": { "A": 0.0002, "C": 0.0001, "D": 0.0001, "E": 0.0001, "F": 0.0027, "H": 0.0001, "I": 0.9591, "K": 0.0013, "L": 0.0056, "M": 0.0071, "N": 0.0022, "R": 0.0009, "S": 0.0123, "T": 0.0028, "V": 0.0055 },
    "83": { "A": 0.0002, "C": 0.0001, "D": 0.0002, "F": 0.0067, "G": 0.0006, "H": 0.0017, "I": 0.0052, "L": 0.9422, "M": 0.0002, "P": 0.002, "Q": 0.0002, "R": 0.0188, "S": 0.0021, "V": 0.0199 },
    "84": { "A": 0.0026, "C": 0.0002, "D": 0.0019, "E": 0.0064, "G": 0.9771, "L": 0.0001, "R": 0.0025, "S": 0.0008, "T": 0.0002, "V": 0.0039, "W": 0.0043 },
    "85": { "A": 0.0001, "D": 0.0135, "E": 0.0007, "G": 0.001, "H": 0.0028, "I": 0.0015, "K": 0.0089, "N": 0.9493, "P": 0.0001, "Q": 0.0001, "R": 0.0009, "S": 0.01, "T": 0.0097, "V": 0.0001, "Y": 0.0013 },
    "86": { "A": 0.0002, "D": 0.0001, "E": 0.0017, "H": 0.0002, "I": 0.0008, "K": 0.9476, "L": 0.0002, "M": 0.0019, "N": 0.0098, "P": 0.0001, "Q": 0.0024, "R": 0.0125, "S": 0.0013, "T": 0.0212, "Y": 0.0002 },
    "87": { "A": 0.988, "D": 0.003, "E": 0.0001, "G": 0.0014, "P": 0.0006, "S": 0.0039, "T": 0.0014, "V": 0.0016 },
    "88": { "A": 0.9836, "C": 0.0002, "D": 0.0016, "E": 0.0003, "F": 0.0002, "G": 0.0006, "L": 0.0002, "N": 0.0001, "P": 0.0006, "R": 0.0001, "S": 0.0058, "T": 0.0015, "V": 0.0054 },
    "89": { "F": 0.0005, "H": 0.0006, "I": 0.0032, "L": 0.9821, "N": 0.0001, "P": 0.0042, "R": 0.0077, "S": 0.0002, "T": 0.0008, "V": 0.0007 },
    "90": { "A": 0.0039, "E": 0.0001, "F": 0.0002, "H": 0.0006, "I": 0.0051, "K": 0.0001, "L": 0.0002, "N": 0.0029, "P": 0.0046, "S": 0.0027, "T": 0.9796, "V": 0.0001 },
    "91": { "D": 0.0001, "F": 0.0007, "G": 0.0001, "H": 0.0001, "I": 0.9693, "L": 0.0051, "M": 0.0004, "N": 0.001, "R": 0.0005, "S": 0.0138, "T": 0.0033, "V": 0.0056 },
    "92": { "A": 0.0179, "D": 0.0001, "G": 0.0003, "K": 0.0015, "M": 0.0009, "P": 0.0058, "Q": 0.0002, "R": 0.0023, "S": 0.0056, "T": 0.9655 },
    "93": { "A": 0.0014, "C": 0.0002, "E": 0.002, "G": 0.9815, "L": 0.0001, "P": 0.0002, "R": 0.0031, "S": 0.0006, "V": 0.008, "W": 0.003 },
    "94": { "A": 0.9855, "D": 0.0032, "E": 0.0005, "F": 0.0001, "G": 0.0013, "I": 0.0002, "L": 0.0018, "P": 0.0009, "R": 0.0002, "S": 0.0026, "T": 0.0009, "V": 0.0028 },
    "95": { "A": 0.0001, "D": 0.0002, "E": 0.0043, "H": 0.0036, "K": 0.0045, "L": 0.001, "N": 0.0001, "P": 0.0015, "Q": 0.9811, "R": 0.0034, "T": 0.0001, "W": 0.0001 },
    "96": { "A": 0.9701, "D": 0.0002, "E": 0.0031, "G": 0.0016, "I": 0.0001, "L": 0.0002, "M": 0.0002, "P": 0.0077, "Q": 0.0002, "S": 0.0074, "T": 0.0064, "V": 0.0028 },
    "97": { "A": 0.0025, "D": 0.9479, "E": 0.0366, "G": 0.0024, "H": 0.0012, "I": 0.0001, "K": 0.0002, "L": 0.0004, "M": 0.0004, "N": 0.004, "R": 0.0001, "S": 0.0003, "T": 0.0002, "V": 0.0009, "Y": 0.0028 },
    "98": { "A": 0.0014, "D": 0.9863, "E": 0.0062, "G": 0.0009, "H": 0.0006, "M": 0.0001, "N": 0.001, "P": 0.0001, "V": 0.0007, "Y": 0.0025 },
    "99": { "A": 0.0035, "D": 0.0158, "E": 0.9758, "G": 0.002, "K": 0.0008, "L": 0.0002, "N": 0.0001, "Q": 0.0002, "S": 0.0002, "T": 0.0001, "V": 0.0014 },
    "100": { "A": 0.024, "C": 0.0361, "D": 0.0008, "E": 0.0002, "F": 0.001, "G": 0.0036, "I": 0.0002, "L": 0.0001, "M": 0.0001, "N": 0.0002, "P": 0.0004, "R": 0.0001, "S": 0.9297, "T": 0.0006, "W": 0.0002, "Y": 0.0029 },
    "101": { "A": 0.0044, "D": 0.9321, "E": 0.0086, "F": 0.002, "G": 0.0015, "H": 0.0168, "I": 0.002, "K": 0.0001, "L": 0.003, "M": 0.0001, "N": 0.0066, "Q": 0.0001, "R": 0.0002, "S": 0.0005, "T": 0.0022, "V": 0.0096, "Y": 0.0103 },
    "102": { "C": 0.0005, "D": 0.0066, "E": 0.0009, "F": 0.0013, "H": 0.0005, "I": 0.0002, "L": 0.0002, "N": 0.0004, "S": 0.0003, "T": 0.0002, "V": 0.0001, "Y": 0.989 },
    "103": { "C": 0.004, "D": 0.0069, "E": 0.0001, "F": 0.0078, "H": 0.0081, "I": 0.0002, "L": 0.0006, "M": 0.0001, "N": 0.0003, "P": 0.0001, "Q": 0.0001, "S": 0.0038, "T": 0.0001, "V": 0.0006, "W": 0.0001, "Y": 0.9673 },
    "104": { "C": 0.9827, "D": 0.0001, "F": 0.0024, "G": 0.0069, "L": 0.0001, "R": 0.0007, "S": 0.0004, "T": 0.0001, "V": 0.0005, "W": 0.0051, "Y": 0.0011 },
    "105": { "A": 0.0348, "C": 0.0007, "D": 0.0004, "E": 0.0031, "F": 0.0018, "G": 0.0077, "I": 0.0019, "K": 0.0001, "L": 0.0415, "M": 0.0191, "N": 0.0001, "Q": 0.0016, "R": 0.0002, "S": 0.006, "T": 0.0007, "V": 0.8803 },
    "106": { "A": 0.0007, "C": 0.0002, "E": 0.0001, "F": 0.0017, "G": 0.0002, "H": 0.0003, "I": 0.0026, "K": 0.0001, "L": 0.9635, "M": 0.007, "P": 0.0018, "Q": 0.003, "R": 0.005, "S": 0.0024, "T": 0.0008, "V": 0.0104, "W": 0.0002, "Y": 0.0001 },
    "107": { "A": 0.0002, "C": 0.0043, "D": 0.0107, "E": 0.0005, "F": 0.0163, "G": 0.0003, "H": 0.0138, "I": 0.0002, "L": 0.0018, "N": 0.0029, "P": 0.0003, "Q": 0.0009, "R": 0.0006, "S": 0.0076, "T": 0.0003, "V": 0.0007, "W": 0.002, "Y": 0.9366 },
    "108": { "A": 0.001, "D": 0.0023, "F": 0.0036, "G": 0.0012, "I": 0.0122, "K": 0.001, "L": 0.0331, "M": 0.899, "P": 0.0005, "R": 0.0059, "S": 0.0006, "T": 0.0017, "V": 0.0371, "W": 0.0005, "Y": 0.0004 },
    "109": { "A": 0.0096, "C": 0.0037, "D": 0.0107, "E": 0.001, "F": 0.0006, "G": 0.9272, "H": 0.0014, "I": 0.0005, "K": 0.0001, "L": 0.0016, "M": 0.0007, "N": 0.0031, "P": 0.0031, "Q": 0.0005, "R": 0.0105, "S": 0.0164, "T": 0.0046, "V": 0.0039, "Y": 0.0007 },
    "110": { "D": 0.0043, "E": 0.0011, "G": 0.0322, "H": 0.0011, "I": 0.0021, "M": 0.0011, "N": 0.0064, "R": 0.0139, "S": 0.9238, "T": 0.0107, "V": 0.0021, "Y": 0.0011 },
    "112": { "F": 0.0067, "G": 0.8267, "I": 0.0467, "L": 0.0133, "N": 0.0133, "S": 0.04, "T": 0.0067, "V": 0.0333, "Y": 0.0133 },
    "113": { "A": 0.0012, "C": 0.0007, "D": 0.0065, "E": 0.0001, "F": 0.0011, "G": 0.1121, "H": 0.0012, "I": 0.0107, "K": 0.0005, "L": 0.0019, "N": 0.0218, "P": 0.0017, "Q": 0.0003, "R": 0.0209, "S": 0.8113, "T": 0.0063, "V": 0.001, "W": 0.0001, "Y": 0.0005 },
    "114": { "A": 0.0073, "C": 0.0025, "D": 0.0152, "E": 0.0006, "F": 0.001, "G": 0.8138, "H": 0.0004, "I": 0.0359, "K": 0.0002, "L": 0.0032, "M": 0.0013, "N": 0.0029, "P": 0.0067, "Q": 0.0001, "R": 0.0045, "S": 0.0872, "T": 0.0104, "V": 0.0051, "W": 0.0008, "Y": 0.001 },
    "115": { "A": 0.0141, "C": 0.0009, "D": 0.0123, "E": 0.0012, "F": 0.0164, "G": 0.0662, "H": 0.014, "I": 0.6079, "K": 0.003, "L": 0.0364, "M": 0.016, "N": 0.0339, "P": 0.0149, "Q": 0.0018, "R": 0.007, "S": 0.0652, "T": 0.047, "V": 0.0304, "W": 0.0006, "Y": 0.0108 },
    "116": { "A": 0.0097, "C": 0.0043, "D": 0.0007, "E": 0.0024, "F": 0.0051, "G": 0.0284, "H": 0.0015, "I": 0.0141, "K": 0.0089, "L": 0.0214, "M": 0.0195, "N": 0.0018, "P": 0.0077, "Q": 0.0031, "R": 0.0393, "S": 0.1733, "T": 0.003, "V": 0.1044, "W": 0.4897, "Y": 0.0618 },
    "117": { "A": 0.0066, "C": 0.0002, "D": 0.0001, "E": 0.007, "F": 0.0013, "G": 0.0048, "H": 0.0001, "I": 0.0141, "K": 0.0003, "L": 0.023, "M": 0.0167, "P": 0.0014, "Q": 0.0036, "R": 0.002, "S": 0.0042, "T": 0.0009, "V": 0.9135, "W": 0.0003 },
    "118": { "C": 0.0032, "F": 0.9784, "G": 0.0002, "I": 0.0028, "L": 0.0032, "R": 0.0001, "S": 0.0021, "V": 0.0088, "Y": 0.0012 },
    "119": { "A": 0.0002, "C": 0.0008, "D": 0.0009, "G": 0.9954, "P": 0.0004, "R": 0.0006, "S": 0.0007, "V": 0.0011 },
    "120": { "A": 0.0009, "D": 0.0001, "E": 0.0009, "G": 0.9146, "L": 0.0001, "N": 0.0001, "P": 0.0021, "R": 0.0016, "S": 0.0036, "T": 0.0748, "V": 0.0013 },
    "121": { "A": 0.0003, "C": 0.0001, "E": 0.0005, "G": 0.9953, "R": 0.0002, "S": 0.0001, "V": 0.0015, "W": 0.0021 },
    "122": { "A": 0.0017, "I": 0.0007, "N": 0.0006, "P": 0.0007, "S": 0.0012, "T": 0.9951 },
    "123": { "A": 0.0008, "E": 0.0024, "H": 0.0006, "I": 0.0004, "K": 0.9428, "L": 0.0001, "M": 0.0001, "N": 0.0026, "P": 0.0001, "Q": 0.0146, "R": 0.0286, "S": 0.001, "T": 0.0032, "V": 0.0005, "Y": 0.0021 },
    "124": { "A": 0.0002, "C": 0.0001, "F": 0.0002, "G": 0.0003, "I": 0.0008, "L": 0.8968, "M": 0.0013, "P": 0.0005, "Q": 0.0002, "R": 0.0013, "S": 0.0001, "V": 0.0982, "W": 0.0001 },
    "125": { "A": 0.0024, "D": 0.0003, "H": 0.0001, "I": 0.0006, "K": 0.0002, "N": 0.0013, "P": 0.0014, "S": 0.0019, "T": 0.9917 },
    "126": { "A": 0.001, "D": 0.0005, "F": 0.0009, "G": 0.0067, "I": 0.0006, "L": 0.0001, "R": 0.0004, "V": 0.9898 },
    "127": { "I": 0.0021, "L": 0.9945, "P": 0.0004, "V": 0.0031 }
  }
}


# Dict of Subject percentage -> IMGT mAb DB percentiles for humanness score values 1-100 (inclusive)
HEAVY_PERCENTILES = {
    1: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.005, 0.005, 0.007, 0.009, 0.011, 0.011, 0.011, 0.013, 0.014, 0.020, 0.036, 0.041, 0.045, 0.045, 0.048, 0.054, 0.061, 0.064, 0.064, 0.068, 0.081, 0.088, 0.093, 0.102, 0.106, 0.132, 0.138, 0.168, 0.215, 0.243, 0.275, 0.304, 0.336, 0.369, 0.397, 0.419, 0.453, 0.481, 0.501, 0.542, 0.572, 0.608, 0.646, 0.680, 0.703, 0.726, 0.771, 0.801, 0.852, 0.891, 0.930, 0.961, 0.975, 1.000],
    2: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.005, 0.009, 0.009, 0.011, 0.018, 0.021, 0.021, 0.030, 0.034, 0.041, 0.048, 0.050, 0.054, 0.054, 0.059, 0.059, 0.063, 0.064, 0.068, 0.075, 0.081, 0.086, 0.097, 0.104, 0.113, 0.132, 0.147, 0.168, 0.190, 0.224, 0.263, 0.301, 0.331, 0.358, 0.383, 0.420, 0.433, 0.483, 0.506, 0.521, 0.544, 0.585, 0.628, 0.664, 0.696, 0.714, 0.728, 0.760, 0.814, 0.850, 0.889, 0.930, 0.952, 0.970, 0.980, 1.000],
    3: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.005, 0.011, 0.014, 0.020, 0.025, 0.027, 0.030, 0.036, 0.038, 0.043, 0.045, 0.050, 0.055, 0.057, 0.061, 0.064, 0.066, 0.068, 0.070, 0.073, 0.077, 0.089, 0.097, 0.106, 0.114, 0.125, 0.134, 0.152, 0.177, 0.191, 0.236, 0.267, 0.315, 0.338, 0.356, 0.390, 0.424, 0.445, 0.474, 0.508, 0.531, 0.547, 0.576, 0.626, 0.660, 0.685, 0.707, 0.737, 0.751, 0.789, 0.843, 0.873, 0.919, 0.939, 0.966, 0.979, 0.987, 1.000],
    4: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.005, 0.009, 0.014, 0.018, 0.020, 0.021, 0.025, 0.027, 0.036, 0.039, 0.047, 0.050, 0.050, 0.054, 0.057, 0.064, 0.066, 0.066, 0.072, 0.072, 0.077, 0.091, 0.095, 0.104, 0.109, 0.113, 0.132, 0.150, 0.157, 0.186, 0.200, 0.236, 0.293, 0.313, 0.349, 0.367, 0.381, 0.419, 0.445, 0.474, 0.501, 0.526, 0.547, 0.571, 0.605, 0.644, 0.674, 0.701, 0.733, 0.748, 0.773, 0.809, 0.860, 0.891, 0.923, 0.946, 0.968, 0.986, 0.989, 1.000],
    5: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.007, 0.014, 0.016, 0.018, 0.023, 0.027, 0.029, 0.030, 0.036, 0.041, 0.045, 0.050, 0.052, 0.059, 0.064, 0.066, 0.068, 0.073, 0.073, 0.082, 0.086, 0.089, 0.100, 0.102, 0.111, 0.122, 0.129, 0.157, 0.177, 0.195, 0.216, 0.247, 0.283, 0.329, 0.347, 0.379, 0.392, 0.420, 0.445, 0.465, 0.499, 0.522, 0.535, 0.567, 0.601, 0.644, 0.671, 0.701, 0.726, 0.744, 0.778, 0.801, 0.835, 0.873, 0.905, 0.934, 0.953, 0.979, 0.989, 0.991, 1.000],
    6: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.007, 0.007, 0.014, 0.020, 0.021, 0.021, 0.027, 0.029, 0.032, 0.036, 0.038, 0.045, 0.052, 0.054, 0.054, 0.063, 0.066, 0.073, 0.081, 0.081, 0.081, 0.084, 0.091, 0.097, 0.104, 0.106, 0.122, 0.141, 0.154, 0.179, 0.199, 0.209, 0.238, 0.270, 0.306, 0.343, 0.358, 0.386, 0.410, 0.442, 0.463, 0.485, 0.513, 0.535, 0.553, 0.581, 0.626, 0.664, 0.689, 0.712, 0.733, 0.753, 0.782, 0.810, 0.844, 0.878, 0.911, 0.936, 0.962, 0.979, 0.989, 0.993, 1.000],
    7: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.005, 0.005, 0.005, 0.007, 0.007, 0.009, 0.016, 0.023, 0.027, 0.029, 0.029, 0.030, 0.032, 0.036, 0.043, 0.052, 0.054, 0.057, 0.061, 0.070, 0.072, 0.077, 0.081, 0.081, 0.084, 0.088, 0.095, 0.097, 0.107, 0.111, 0.131, 0.152, 0.165, 0.188, 0.204, 0.225, 0.263, 0.292, 0.324, 0.349, 0.374, 0.404, 0.422, 0.454, 0.478, 0.497, 0.524, 0.542, 0.565, 0.603, 0.644, 0.674, 0.699, 0.725, 0.737, 0.760, 0.792, 0.816, 0.850, 0.891, 0.914, 0.941, 0.973, 0.980, 0.989, 0.993, 1.000],
    8: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.005, 0.005, 0.005, 0.007, 0.007, 0.013, 0.014, 0.027, 0.027, 0.029, 0.030, 0.032, 0.032, 0.032, 0.039, 0.052, 0.054, 0.055, 0.064, 0.073, 0.077, 0.079, 0.079, 0.081, 0.081, 0.088, 0.089, 0.095, 0.102, 0.109, 0.120, 0.140, 0.157, 0.170, 0.199, 0.220, 0.247, 0.274, 0.299, 0.336, 0.369, 0.395, 0.415, 0.445, 0.470, 0.485, 0.504, 0.524, 0.553, 0.592, 0.623, 0.653, 0.685, 0.712, 0.732, 0.750, 0.771, 0.800, 0.826, 0.862, 0.893, 0.916, 0.943, 0.975, 0.986, 0.989, 0.993, 1.000],
    9: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.013, 0.013, 0.013, 0.021, 0.023, 0.027, 0.029, 0.032, 0.036, 0.036, 0.039, 0.057, 0.059, 0.068, 0.068, 0.070, 0.077, 0.081, 0.081, 0.082, 0.082, 0.084, 0.089, 0.093, 0.097, 0.104, 0.109, 0.120, 0.134, 0.159, 0.168, 0.179, 0.208, 0.236, 0.261, 0.283, 0.309, 0.358, 0.376, 0.406, 0.429, 0.449, 0.476, 0.501, 0.513, 0.549, 0.569, 0.606, 0.633, 0.665, 0.696, 0.725, 0.746, 0.762, 0.785, 0.816, 0.841, 0.871, 0.898, 0.925, 0.946, 0.979, 0.987, 0.991, 0.993, 1.000],
    10: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.009, 0.011, 0.013, 0.013, 0.014, 0.020, 0.023, 0.025, 0.032, 0.036, 0.036, 0.036, 0.039, 0.047, 0.059, 0.063, 0.068, 0.068, 0.073, 0.077, 0.082, 0.082, 0.082, 0.088, 0.091, 0.093, 0.097, 0.100, 0.106, 0.113, 0.132, 0.143, 0.165, 0.172, 0.190, 0.222, 0.252, 0.284, 0.306, 0.327, 0.367, 0.390, 0.417, 0.445, 0.454, 0.483, 0.503, 0.526, 0.569, 0.585, 0.614, 0.637, 0.671, 0.703, 0.730, 0.755, 0.769, 0.792, 0.821, 0.848, 0.878, 0.905, 0.925, 0.952, 0.980, 0.987, 0.991, 0.993, 1.000],
    11: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.007, 0.009, 0.009, 0.011, 0.013, 0.013, 0.014, 0.020, 0.023, 0.029, 0.036, 0.038, 0.038, 0.041, 0.052, 0.055, 0.059, 0.064, 0.068, 0.070, 0.075, 0.082, 0.082, 0.082, 0.084, 0.091, 0.093, 0.097, 0.097, 0.102, 0.116, 0.131, 0.138, 0.152, 0.172, 0.181, 0.191, 0.229, 0.265, 0.295, 0.309, 0.336, 0.372, 0.401, 0.420, 0.445, 0.460, 0.494, 0.513, 0.535, 0.578, 0.592, 0.623, 0.648, 0.685, 0.710, 0.735, 0.760, 0.776, 0.800, 0.823, 0.850, 0.882, 0.909, 0.934, 0.959, 0.982, 0.989, 0.991, 0.993, 1.000],
    12: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.009, 0.009, 0.009, 0.011, 0.013, 0.014, 0.018, 0.027, 0.030, 0.032, 0.036, 0.038, 0.038, 0.045, 0.052, 0.055, 0.064, 0.068, 0.068, 0.070, 0.075, 0.082, 0.082, 0.084, 0.084, 0.091, 0.098, 0.098, 0.102, 0.104, 0.118, 0.132, 0.143, 0.157, 0.177, 0.188, 0.202, 0.238, 0.277, 0.302, 0.318, 0.347, 0.383, 0.410, 0.428, 0.453, 0.470, 0.497, 0.522, 0.546, 0.589, 0.606, 0.633, 0.662, 0.685, 0.714, 0.739, 0.760, 0.782, 0.801, 0.830, 0.853, 0.886, 0.916, 0.941, 0.961, 0.982, 0.991, 0.991, 0.993, 1.000],
    13: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.009, 0.011, 0.011, 0.011, 0.013, 0.018, 0.021, 0.027, 0.032, 0.034, 0.038, 0.038, 0.045, 0.055, 0.057, 0.063, 0.064, 0.068, 0.068, 0.073, 0.081, 0.082, 0.084, 0.086, 0.088, 0.095, 0.098, 0.100, 0.106, 0.109, 0.123, 0.141, 0.154, 0.166, 0.179, 0.197, 0.222, 0.250, 0.288, 0.313, 0.335, 0.370, 0.392, 0.411, 0.431, 0.463, 0.483, 0.506, 0.535, 0.565, 0.599, 0.619, 0.648, 0.674, 0.691, 0.719, 0.748, 0.769, 0.791, 0.809, 0.835, 0.864, 0.891, 0.919, 0.945, 0.970, 0.984, 0.991, 0.993, 0.993, 1.000],
    14: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.007, 0.011, 0.011, 0.011, 0.011, 0.014, 0.020, 0.021, 0.030, 0.034, 0.036, 0.041, 0.043, 0.054, 0.059, 0.063, 0.064, 0.064, 0.070, 0.073, 0.077, 0.081, 0.084, 0.084, 0.086, 0.089, 0.095, 0.102, 0.102, 0.107, 0.113, 0.129, 0.143, 0.154, 0.168, 0.184, 0.202, 0.225, 0.258, 0.297, 0.315, 0.336, 0.376, 0.395, 0.419, 0.436, 0.470, 0.488, 0.517, 0.547, 0.571, 0.605, 0.633, 0.662, 0.674, 0.699, 0.723, 0.755, 0.771, 0.796, 0.812, 0.843, 0.871, 0.902, 0.923, 0.948, 0.973, 0.984, 0.991, 0.993, 0.993, 1.000],
    15: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.005, 0.007, 0.011, 0.011, 0.011, 0.011, 0.016, 0.023, 0.032, 0.034, 0.036, 0.036, 0.043, 0.050, 0.061, 0.064, 0.064, 0.064, 0.072, 0.073, 0.077, 0.079, 0.084, 0.084, 0.084, 0.091, 0.097, 0.097, 0.102, 0.106, 0.109, 0.120, 0.136, 0.150, 0.161, 0.170, 0.188, 0.211, 0.233, 0.277, 0.299, 0.318, 0.343, 0.383, 0.404, 0.428, 0.447, 0.476, 0.492, 0.528, 0.553, 0.583, 0.615, 0.635, 0.662, 0.674, 0.708, 0.730, 0.762, 0.776, 0.796, 0.818, 0.850, 0.875, 0.903, 0.927, 0.948, 0.973, 0.986, 0.991, 0.993, 0.993, 1.000],
    16: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.007, 0.007, 0.011, 0.011, 0.011, 0.013, 0.016, 0.023, 0.032, 0.036, 0.038, 0.038, 0.045, 0.057, 0.061, 0.064, 0.066, 0.068, 0.075, 0.075, 0.077, 0.081, 0.084, 0.084, 0.084, 0.093, 0.097, 0.098, 0.104, 0.106, 0.109, 0.125, 0.136, 0.156, 0.163, 0.177, 0.193, 0.218, 0.245, 0.281, 0.304, 0.324, 0.354, 0.386, 0.406, 0.436, 0.454, 0.483, 0.504, 0.528, 0.562, 0.587, 0.631, 0.642, 0.667, 0.687, 0.717, 0.737, 0.769, 0.780, 0.798, 0.826, 0.850, 0.877, 0.909, 0.930, 0.950, 0.975, 0.986, 0.991, 0.993, 0.993, 1.000],
    17: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.007, 0.007, 0.011, 0.013, 0.013, 0.013, 0.021, 0.025, 0.034, 0.036, 0.043, 0.043, 0.048, 0.059, 0.061, 0.068, 0.072, 0.072, 0.075, 0.075, 0.079, 0.081, 0.084, 0.084, 0.088, 0.093, 0.098, 0.102, 0.106, 0.107, 0.118, 0.132, 0.148, 0.159, 0.172, 0.188, 0.209, 0.231, 0.252, 0.290, 0.313, 0.333, 0.367, 0.395, 0.420, 0.444, 0.463, 0.490, 0.515, 0.544, 0.571, 0.596, 0.635, 0.648, 0.676, 0.694, 0.719, 0.744, 0.775, 0.785, 0.800, 0.832, 0.857, 0.880, 0.911, 0.934, 0.959, 0.979, 0.986, 0.991, 0.993, 0.993, 1.000],
    18: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.007, 0.007, 0.011, 0.013, 0.014, 0.016, 0.021, 0.027, 0.034, 0.041, 0.043, 0.045, 0.055, 0.059, 0.061, 0.068, 0.072, 0.072, 0.075, 0.077, 0.079, 0.081, 0.084, 0.084, 0.088, 0.095, 0.100, 0.104, 0.107, 0.111, 0.122, 0.132, 0.156, 0.159, 0.177, 0.191, 0.215, 0.242, 0.256, 0.301, 0.315, 0.342, 0.376, 0.399, 0.424, 0.447, 0.465, 0.494, 0.519, 0.551, 0.572, 0.599, 0.637, 0.653, 0.678, 0.699, 0.723, 0.746, 0.776, 0.792, 0.801, 0.835, 0.860, 0.884, 0.912, 0.939, 0.959, 0.980, 0.987, 0.991, 0.993, 0.993, 1.000],
    19: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.007, 0.007, 0.013, 0.014, 0.014, 0.016, 0.027, 0.029, 0.036, 0.043, 0.043, 0.047, 0.059, 0.066, 0.066, 0.068, 0.072, 0.077, 0.077, 0.077, 0.081, 0.081, 0.084, 0.088, 0.089, 0.095, 0.102, 0.104, 0.107, 0.113, 0.125, 0.140, 0.159, 0.166, 0.182, 0.195, 0.220, 0.242, 0.258, 0.304, 0.318, 0.352, 0.383, 0.406, 0.429, 0.449, 0.472, 0.499, 0.521, 0.555, 0.576, 0.606, 0.639, 0.657, 0.689, 0.703, 0.732, 0.753, 0.780, 0.792, 0.803, 0.839, 0.862, 0.891, 0.916, 0.939, 0.962, 0.982, 0.987, 0.991, 0.993, 0.993, 1.000],
    20: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.007, 0.013, 0.013, 0.016, 0.016, 0.020, 0.027, 0.029, 0.036, 0.047, 0.048, 0.054, 0.061, 0.066, 0.068, 0.068, 0.072, 0.077, 0.079, 0.079, 0.081, 0.081, 0.084, 0.091, 0.091, 0.095, 0.104, 0.104, 0.109, 0.114, 0.131, 0.143, 0.165, 0.174, 0.191, 0.204, 0.225, 0.243, 0.275, 0.311, 0.335, 0.356, 0.386, 0.413, 0.431, 0.453, 0.476, 0.504, 0.530, 0.562, 0.578, 0.610, 0.646, 0.664, 0.691, 0.705, 0.737, 0.758, 0.782, 0.794, 0.807, 0.848, 0.864, 0.893, 0.919, 0.943, 0.964, 0.982, 0.987, 0.991, 0.993, 0.993, 1.000],
    21: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.007, 0.013, 0.014, 0.016, 0.018, 0.025, 0.027, 0.036, 0.041, 0.048, 0.054, 0.055, 0.063, 0.066, 0.068, 0.070, 0.072, 0.077, 0.079, 0.079, 0.081, 0.082, 0.089, 0.093, 0.093, 0.098, 0.106, 0.106, 0.109, 0.114, 0.132, 0.156, 0.172, 0.182, 0.200, 0.218, 0.238, 0.265, 0.292, 0.322, 0.343, 0.361, 0.401, 0.428, 0.442, 0.462, 0.485, 0.526, 0.544, 0.569, 0.589, 0.617, 0.651, 0.673, 0.698, 0.712, 0.746, 0.766, 0.789, 0.800, 0.816, 0.852, 0.866, 0.900, 0.919, 0.946, 0.970, 0.982, 0.989, 0.991, 0.993, 0.995, 1.000],
    22: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.013, 0.016, 0.016, 0.018, 0.018, 0.027, 0.030, 0.039, 0.043, 0.054, 0.055, 0.064, 0.066, 0.068, 0.072, 0.072, 0.075, 0.079, 0.079, 0.081, 0.081, 0.086, 0.091, 0.093, 0.095, 0.098, 0.106, 0.106, 0.113, 0.118, 0.134, 0.161, 0.174, 0.182, 0.206, 0.220, 0.238, 0.272, 0.301, 0.327, 0.349, 0.370, 0.406, 0.428, 0.445, 0.462, 0.490, 0.528, 0.544, 0.574, 0.599, 0.623, 0.653, 0.674, 0.701, 0.716, 0.751, 0.773, 0.792, 0.801, 0.819, 0.852, 0.868, 0.903, 0.925, 0.946, 0.970, 0.982, 0.989, 0.993, 0.993, 0.995, 1.000],
    23: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.005, 0.013, 0.016, 0.018, 0.018, 0.020, 0.030, 0.034, 0.039, 0.043, 0.054, 0.057, 0.064, 0.066, 0.068, 0.072, 0.075, 0.075, 0.079, 0.079, 0.081, 0.082, 0.088, 0.093, 0.093, 0.095, 0.098, 0.107, 0.107, 0.113, 0.120, 0.141, 0.165, 0.175, 0.186, 0.209, 0.224, 0.240, 0.284, 0.302, 0.335, 0.361, 0.376, 0.411, 0.435, 0.456, 0.465, 0.494, 0.528, 0.553, 0.576, 0.603, 0.630, 0.657, 0.689, 0.707, 0.719, 0.757, 0.776, 0.796, 0.803, 0.821, 0.852, 0.869, 0.909, 0.927, 0.948, 0.970, 0.982, 0.989, 0.993, 0.993, 0.995, 1.000],
    24: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.007, 0.014, 0.018, 0.018, 0.021, 0.025, 0.030, 0.038, 0.039, 0.047, 0.055, 0.061, 0.064, 0.066, 0.070, 0.072, 0.075, 0.075, 0.079, 0.079, 0.081, 0.082, 0.089, 0.093, 0.093, 0.097, 0.104, 0.107, 0.109, 0.118, 0.129, 0.143, 0.175, 0.184, 0.195, 0.213, 0.231, 0.267, 0.284, 0.309, 0.340, 0.370, 0.383, 0.413, 0.440, 0.458, 0.470, 0.496, 0.537, 0.558, 0.580, 0.610, 0.633, 0.665, 0.689, 0.708, 0.725, 0.758, 0.776, 0.798, 0.803, 0.821, 0.857, 0.877, 0.911, 0.928, 0.952, 0.970, 0.984, 0.989, 0.993, 0.993, 0.995, 1.000],
    25: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.013, 0.016, 0.020, 0.020, 0.021, 0.025, 0.034, 0.038, 0.041, 0.052, 0.055, 0.061, 0.066, 0.068, 0.073, 0.075, 0.075, 0.077, 0.079, 0.081, 0.081, 0.084, 0.089, 0.095, 0.095, 0.102, 0.106, 0.109, 0.111, 0.122, 0.136, 0.157, 0.182, 0.197, 0.208, 0.216, 0.258, 0.270, 0.295, 0.320, 0.351, 0.377, 0.394, 0.420, 0.445, 0.467, 0.483, 0.503, 0.542, 0.565, 0.587, 0.612, 0.633, 0.665, 0.692, 0.714, 0.735, 0.766, 0.780, 0.798, 0.803, 0.823, 0.860, 0.878, 0.912, 0.930, 0.953, 0.970, 0.986, 0.991, 0.993, 0.995, 0.995, 1.000],
    26: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.005, 0.007, 0.014, 0.018, 0.020, 0.020, 0.023, 0.027, 0.036, 0.038, 0.045, 0.054, 0.057, 0.063, 0.066, 0.073, 0.073, 0.075, 0.075, 0.079, 0.079, 0.082, 0.082, 0.086, 0.093, 0.097, 0.097, 0.104, 0.107, 0.111, 0.113, 0.129, 0.141, 0.161, 0.182, 0.200, 0.215, 0.224, 0.265, 0.279, 0.306, 0.324, 0.352, 0.385, 0.401, 0.420, 0.445, 0.467, 0.485, 0.503, 0.547, 0.567, 0.590, 0.612, 0.633, 0.671, 0.692, 0.717, 0.741, 0.771, 0.782, 0.801, 0.807, 0.828, 0.864, 0.884, 0.916, 0.930, 0.953, 0.971, 0.987, 0.991, 0.993, 0.995, 0.996, 1.000],
    27: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.007, 0.013, 0.014, 0.018, 0.021, 0.023, 0.023, 0.029, 0.036, 0.038, 0.047, 0.054, 0.061, 0.063, 0.070, 0.073, 0.075, 0.075, 0.077, 0.079, 0.081, 0.084, 0.086, 0.086, 0.095, 0.097, 0.100, 0.104, 0.109, 0.114, 0.118, 0.140, 0.147, 0.166, 0.188, 0.202, 0.218, 0.227, 0.267, 0.290, 0.315, 0.324, 0.356, 0.388, 0.408, 0.426, 0.449, 0.467, 0.488, 0.510, 0.553, 0.572, 0.592, 0.612, 0.639, 0.673, 0.699, 0.719, 0.742, 0.771, 0.789, 0.805, 0.812, 0.834, 0.868, 0.884, 0.916, 0.932, 0.953, 0.973, 0.987, 0.991, 0.993, 0.995, 0.996, 1.000],
    28: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.005, 0.009, 0.013, 0.014, 0.018, 0.021, 0.023, 0.023, 0.029, 0.036, 0.038, 0.047, 0.057, 0.061, 0.063, 0.070, 0.073, 0.075, 0.075, 0.077, 0.081, 0.081, 0.084, 0.086, 0.088, 0.095, 0.098, 0.106, 0.113, 0.113, 0.114, 0.122, 0.141, 0.148, 0.175, 0.197, 0.206, 0.218, 0.234, 0.270, 0.297, 0.317, 0.331, 0.363, 0.390, 0.410, 0.428, 0.454, 0.469, 0.492, 0.521, 0.553, 0.574, 0.599, 0.617, 0.642, 0.680, 0.703, 0.721, 0.742, 0.773, 0.796, 0.807, 0.816, 0.835, 0.873, 0.886, 0.919, 0.934, 0.955, 0.975, 0.987, 0.993, 0.993, 0.995, 0.996, 1.000],
    29: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.007, 0.009, 0.013, 0.016, 0.020, 0.021, 0.025, 0.025, 0.029, 0.036, 0.039, 0.052, 0.059, 0.063, 0.064, 0.070, 0.075, 0.075, 0.079, 0.081, 0.084, 0.084, 0.084, 0.088, 0.091, 0.098, 0.106, 0.109, 0.113, 0.113, 0.118, 0.131, 0.147, 0.156, 0.181, 0.206, 0.211, 0.218, 0.254, 0.274, 0.306, 0.320, 0.336, 0.372, 0.392, 0.413, 0.438, 0.458, 0.474, 0.494, 0.531, 0.562, 0.576, 0.601, 0.630, 0.651, 0.687, 0.703, 0.723, 0.750, 0.782, 0.798, 0.809, 0.821, 0.846, 0.877, 0.889, 0.923, 0.936, 0.959, 0.979, 0.987, 0.993, 0.993, 0.995, 0.996, 1.000],
    30: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.007, 0.011, 0.014, 0.016, 0.021, 0.021, 0.025, 0.029, 0.029, 0.039, 0.045, 0.054, 0.059, 0.063, 0.068, 0.070, 0.079, 0.079, 0.079, 0.081, 0.084, 0.084, 0.084, 0.088, 0.093, 0.100, 0.107, 0.111, 0.113, 0.116, 0.123, 0.131, 0.148, 0.159, 0.184, 0.206, 0.213, 0.222, 0.259, 0.283, 0.309, 0.326, 0.338, 0.376, 0.394, 0.417, 0.445, 0.458, 0.481, 0.497, 0.535, 0.567, 0.578, 0.608, 0.633, 0.655, 0.694, 0.708, 0.726, 0.755, 0.782, 0.798, 0.812, 0.823, 0.850, 0.880, 0.891, 0.923, 0.939, 0.962, 0.979, 0.987, 0.993, 0.993, 0.995, 0.996, 1.000],
    31: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.009, 0.014, 0.014, 0.016, 0.021, 0.023, 0.025, 0.029, 0.029, 0.041, 0.050, 0.059, 0.063, 0.068, 0.072, 0.072, 0.079, 0.079, 0.081, 0.082, 0.084, 0.084, 0.086, 0.089, 0.093, 0.102, 0.111, 0.111, 0.113, 0.120, 0.127, 0.143, 0.154, 0.165, 0.188, 0.208, 0.216, 0.229, 0.261, 0.288, 0.318, 0.327, 0.340, 0.377, 0.395, 0.420, 0.449, 0.463, 0.488, 0.504, 0.540, 0.569, 0.580, 0.612, 0.644, 0.665, 0.694, 0.710, 0.733, 0.757, 0.789, 0.803, 0.812, 0.823, 0.853, 0.880, 0.894, 0.925, 0.946, 0.964, 0.979, 0.987, 0.993, 0.993, 0.995, 0.996, 1.000],
    32: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.005, 0.009, 0.014, 0.014, 0.016, 0.021, 0.023, 0.025, 0.029, 0.030, 0.041, 0.052, 0.059, 0.064, 0.068, 0.072, 0.073, 0.079, 0.079, 0.081, 0.082, 0.084, 0.086, 0.088, 0.089, 0.093, 0.104, 0.111, 0.111, 0.116, 0.125, 0.132, 0.143, 0.159, 0.168, 0.190, 0.209, 0.220, 0.231, 0.272, 0.293, 0.320, 0.327, 0.342, 0.377, 0.397, 0.426, 0.451, 0.469, 0.494, 0.519, 0.544, 0.571, 0.581, 0.617, 0.649, 0.667, 0.696, 0.714, 0.735, 0.764, 0.791, 0.805, 0.812, 0.823, 0.857, 0.880, 0.894, 0.927, 0.952, 0.964, 0.979, 0.987, 0.993, 0.993, 0.995, 0.996, 1.000],
    33: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.005, 0.009, 0.014, 0.014, 0.016, 0.018, 0.023, 0.025, 0.032, 0.032, 0.047, 0.050, 0.055, 0.059, 0.066, 0.068, 0.073, 0.077, 0.079, 0.079, 0.082, 0.082, 0.086, 0.088, 0.089, 0.091, 0.097, 0.106, 0.111, 0.116, 0.123, 0.131, 0.141, 0.157, 0.170, 0.177, 0.200, 0.218, 0.225, 0.247, 0.277, 0.304, 0.324, 0.331, 0.349, 0.385, 0.415, 0.435, 0.458, 0.476, 0.499, 0.524, 0.547, 0.576, 0.590, 0.621, 0.651, 0.673, 0.701, 0.719, 0.744, 0.771, 0.791, 0.810, 0.826, 0.844, 0.866, 0.882, 0.903, 0.930, 0.952, 0.964, 0.979, 0.987, 0.993, 0.993, 0.995, 0.998, 1.000],
    34: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.005, 0.009, 0.014, 0.014, 0.016, 0.020, 0.027, 0.030, 0.032, 0.038, 0.055, 0.057, 0.061, 0.063, 0.068, 0.068, 0.073, 0.077, 0.079, 0.079, 0.082, 0.084, 0.089, 0.089, 0.093, 0.093, 0.102, 0.106, 0.116, 0.122, 0.132, 0.141, 0.148, 0.161, 0.170, 0.184, 0.204, 0.224, 0.229, 0.249, 0.283, 0.311, 0.326, 0.335, 0.351, 0.394, 0.419, 0.436, 0.460, 0.479, 0.503, 0.531, 0.551, 0.581, 0.592, 0.628, 0.653, 0.674, 0.705, 0.725, 0.744, 0.773, 0.792, 0.819, 0.826, 0.848, 0.869, 0.886, 0.907, 0.932, 0.955, 0.966, 0.979, 0.987, 0.993, 0.993, 0.995, 0.998, 1.000],
    35: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.005, 0.011, 0.014, 0.014, 0.020, 0.023, 0.030, 0.032, 0.034, 0.047, 0.057, 0.061, 0.063, 0.064, 0.068, 0.072, 0.073, 0.077, 0.079, 0.079, 0.082, 0.086, 0.091, 0.091, 0.093, 0.098, 0.111, 0.111, 0.127, 0.131, 0.138, 0.145, 0.152, 0.163, 0.174, 0.186, 0.215, 0.231, 0.249, 0.267, 0.292, 0.313, 0.327, 0.338, 0.360, 0.394, 0.419, 0.442, 0.463, 0.490, 0.512, 0.535, 0.551, 0.585, 0.596, 0.635, 0.655, 0.680, 0.710, 0.732, 0.751, 0.780, 0.796, 0.825, 0.832, 0.853, 0.869, 0.889, 0.914, 0.937, 0.957, 0.966, 0.979, 0.987, 0.993, 0.993, 0.995, 0.998, 1.000],
    36: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.009, 0.011, 0.014, 0.014, 0.020, 0.023, 0.030, 0.032, 0.038, 0.054, 0.057, 0.061, 0.063, 0.064, 0.068, 0.072, 0.073, 0.077, 0.079, 0.081, 0.084, 0.088, 0.091, 0.091, 0.097, 0.106, 0.111, 0.113, 0.129, 0.131, 0.138, 0.147, 0.156, 0.168, 0.175, 0.202, 0.218, 0.231, 0.250, 0.277, 0.297, 0.317, 0.327, 0.343, 0.363, 0.394, 0.422, 0.444, 0.472, 0.497, 0.521, 0.535, 0.560, 0.585, 0.601, 0.637, 0.657, 0.682, 0.710, 0.737, 0.755, 0.780, 0.803, 0.825, 0.841, 0.855, 0.871, 0.893, 0.919, 0.939, 0.957, 0.966, 0.980, 0.987, 0.993, 0.993, 0.995, 0.998, 1.000],
    37: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.007, 0.011, 0.018, 0.018, 0.023, 0.023, 0.032, 0.034, 0.043, 0.043, 0.057, 0.059, 0.063, 0.064, 0.066, 0.068, 0.072, 0.073, 0.077, 0.082, 0.084, 0.084, 0.089, 0.091, 0.091, 0.097, 0.106, 0.113, 0.122, 0.131, 0.132, 0.140, 0.150, 0.161, 0.177, 0.181, 0.202, 0.222, 0.247, 0.259, 0.286, 0.302, 0.320, 0.329, 0.343, 0.370, 0.401, 0.435, 0.456, 0.479, 0.503, 0.522, 0.549, 0.565, 0.592, 0.610, 0.637, 0.658, 0.685, 0.716, 0.739, 0.764, 0.785, 0.809, 0.830, 0.841, 0.860, 0.873, 0.900, 0.925, 0.939, 0.957, 0.966, 0.980, 0.987, 0.993, 0.993, 0.996, 0.998, 1.000],
    38: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.005, 0.009, 0.009, 0.014, 0.018, 0.021, 0.023, 0.027, 0.032, 0.039, 0.043, 0.043, 0.059, 0.061, 0.064, 0.064, 0.066, 0.070, 0.072, 0.075, 0.081, 0.082, 0.084, 0.088, 0.089, 0.091, 0.097, 0.097, 0.106, 0.114, 0.123, 0.131, 0.134, 0.143, 0.152, 0.161, 0.177, 0.186, 0.204, 0.225, 0.247, 0.261, 0.288, 0.304, 0.322, 0.335, 0.347, 0.376, 0.404, 0.436, 0.463, 0.481, 0.510, 0.530, 0.549, 0.572, 0.596, 0.612, 0.640, 0.664, 0.692, 0.717, 0.744, 0.764, 0.787, 0.809, 0.832, 0.843, 0.862, 0.878, 0.902, 0.925, 0.939, 0.957, 0.968, 0.980, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    39: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.005, 0.009, 0.009, 0.011, 0.016, 0.018, 0.025, 0.027, 0.029, 0.038, 0.043, 0.045, 0.047, 0.061, 0.064, 0.064, 0.066, 0.068, 0.070, 0.073, 0.075, 0.081, 0.084, 0.084, 0.088, 0.089, 0.093, 0.097, 0.100, 0.107, 0.120, 0.125, 0.131, 0.136, 0.147, 0.154, 0.168, 0.177, 0.186, 0.208, 0.225, 0.247, 0.263, 0.290, 0.311, 0.322, 0.335, 0.354, 0.379, 0.411, 0.442, 0.465, 0.488, 0.517, 0.538, 0.551, 0.572, 0.596, 0.615, 0.644, 0.665, 0.692, 0.723, 0.744, 0.767, 0.791, 0.819, 0.834, 0.843, 0.862, 0.882, 0.909, 0.930, 0.939, 0.959, 0.968, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    40: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.009, 0.009, 0.014, 0.016, 0.018, 0.025, 0.027, 0.036, 0.038, 0.045, 0.047, 0.057, 0.063, 0.064, 0.064, 0.066, 0.068, 0.072, 0.075, 0.077, 0.084, 0.084, 0.084, 0.089, 0.091, 0.093, 0.098, 0.106, 0.116, 0.123, 0.125, 0.131, 0.141, 0.150, 0.156, 0.172, 0.186, 0.191, 0.211, 0.229, 0.247, 0.267, 0.295, 0.313, 0.324, 0.340, 0.363, 0.383, 0.415, 0.445, 0.467, 0.490, 0.521, 0.540, 0.551, 0.572, 0.597, 0.615, 0.649, 0.671, 0.699, 0.726, 0.748, 0.769, 0.791, 0.819, 0.835, 0.846, 0.866, 0.884, 0.911, 0.936, 0.943, 0.959, 0.968, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    41: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.007, 0.011, 0.013, 0.014, 0.016, 0.021, 0.027, 0.029, 0.032, 0.043, 0.045, 0.050, 0.052, 0.061, 0.063, 0.064, 0.066, 0.068, 0.070, 0.075, 0.075, 0.081, 0.084, 0.084, 0.084, 0.091, 0.093, 0.093, 0.098, 0.109, 0.118, 0.123, 0.127, 0.131, 0.147, 0.152, 0.165, 0.175, 0.186, 0.197, 0.213, 0.229, 0.252, 0.270, 0.295, 0.313, 0.327, 0.342, 0.374, 0.392, 0.426, 0.453, 0.470, 0.497, 0.522, 0.542, 0.562, 0.572, 0.597, 0.623, 0.651, 0.676, 0.701, 0.728, 0.755, 0.773, 0.794, 0.821, 0.835, 0.853, 0.869, 0.889, 0.916, 0.939, 0.946, 0.959, 0.968, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    42: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.007, 0.011, 0.013, 0.014, 0.016, 0.021, 0.027, 0.034, 0.036, 0.043, 0.047, 0.050, 0.052, 0.061, 0.063, 0.064, 0.066, 0.070, 0.072, 0.075, 0.075, 0.081, 0.084, 0.088, 0.091, 0.093, 0.093, 0.097, 0.107, 0.109, 0.123, 0.125, 0.129, 0.138, 0.156, 0.159, 0.166, 0.175, 0.190, 0.204, 0.218, 0.233, 0.258, 0.277, 0.299, 0.317, 0.336, 0.352, 0.381, 0.399, 0.433, 0.454, 0.474, 0.501, 0.524, 0.544, 0.562, 0.581, 0.605, 0.630, 0.657, 0.678, 0.708, 0.733, 0.755, 0.775, 0.796, 0.821, 0.837, 0.855, 0.871, 0.889, 0.925, 0.943, 0.946, 0.959, 0.968, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    43: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.009, 0.011, 0.014, 0.016, 0.020, 0.023, 0.036, 0.039, 0.043, 0.047, 0.048, 0.050, 0.052, 0.061, 0.064, 0.066, 0.068, 0.070, 0.073, 0.075, 0.075, 0.081, 0.086, 0.088, 0.091, 0.095, 0.095, 0.102, 0.107, 0.109, 0.125, 0.127, 0.131, 0.143, 0.157, 0.159, 0.166, 0.179, 0.191, 0.209, 0.220, 0.247, 0.263, 0.283, 0.304, 0.322, 0.338, 0.356, 0.386, 0.404, 0.442, 0.463, 0.481, 0.508, 0.526, 0.546, 0.567, 0.589, 0.606, 0.635, 0.658, 0.685, 0.714, 0.737, 0.760, 0.778, 0.800, 0.826, 0.843, 0.860, 0.875, 0.891, 0.932, 0.943, 0.948, 0.959, 0.970, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    44: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.011, 0.013, 0.014, 0.016, 0.020, 0.023, 0.038, 0.039, 0.043, 0.047, 0.048, 0.050, 0.052, 0.061, 0.064, 0.066, 0.068, 0.070, 0.075, 0.075, 0.079, 0.082, 0.086, 0.088, 0.093, 0.095, 0.095, 0.106, 0.107, 0.111, 0.125, 0.127, 0.131, 0.145, 0.157, 0.163, 0.168, 0.181, 0.193, 0.211, 0.236, 0.254, 0.272, 0.288, 0.304, 0.326, 0.340, 0.361, 0.390, 0.406, 0.451, 0.465, 0.485, 0.512, 0.526, 0.546, 0.574, 0.589, 0.612, 0.639, 0.664, 0.692, 0.716, 0.742, 0.760, 0.780, 0.800, 0.832, 0.844, 0.864, 0.877, 0.898, 0.932, 0.943, 0.948, 0.959, 0.971, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    45: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.011, 0.014, 0.014, 0.020, 0.020, 0.023, 0.038, 0.043, 0.045, 0.047, 0.048, 0.050, 0.052, 0.063, 0.064, 0.070, 0.072, 0.073, 0.075, 0.079, 0.082, 0.084, 0.088, 0.091, 0.095, 0.095, 0.095, 0.107, 0.109, 0.116, 0.127, 0.131, 0.134, 0.147, 0.163, 0.166, 0.174, 0.188, 0.199, 0.220, 0.240, 0.258, 0.272, 0.290, 0.315, 0.333, 0.343, 0.367, 0.390, 0.417, 0.453, 0.476, 0.496, 0.519, 0.528, 0.558, 0.583, 0.599, 0.617, 0.646, 0.671, 0.698, 0.721, 0.746, 0.766, 0.784, 0.807, 0.837, 0.846, 0.864, 0.877, 0.903, 0.932, 0.943, 0.948, 0.961, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    46: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.007, 0.011, 0.011, 0.014, 0.020, 0.020, 0.021, 0.023, 0.039, 0.045, 0.045, 0.047, 0.048, 0.050, 0.059, 0.064, 0.066, 0.072, 0.072, 0.073, 0.077, 0.081, 0.082, 0.084, 0.091, 0.093, 0.095, 0.095, 0.097, 0.107, 0.109, 0.122, 0.129, 0.132, 0.138, 0.150, 0.163, 0.170, 0.179, 0.190, 0.204, 0.222, 0.254, 0.265, 0.275, 0.299, 0.324, 0.333, 0.347, 0.372, 0.395, 0.420, 0.458, 0.479, 0.496, 0.526, 0.531, 0.564, 0.583, 0.603, 0.623, 0.646, 0.680, 0.701, 0.725, 0.750, 0.767, 0.789, 0.814, 0.837, 0.846, 0.866, 0.878, 0.905, 0.932, 0.943, 0.948, 0.962, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    47: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.007, 0.011, 0.011, 0.014, 0.020, 0.020, 0.021, 0.027, 0.039, 0.045, 0.045, 0.047, 0.048, 0.052, 0.059, 0.064, 0.068, 0.072, 0.072, 0.073, 0.077, 0.081, 0.084, 0.089, 0.091, 0.093, 0.095, 0.095, 0.097, 0.107, 0.114, 0.122, 0.129, 0.132, 0.140, 0.150, 0.163, 0.174, 0.179, 0.191, 0.211, 0.231, 0.256, 0.272, 0.284, 0.301, 0.324, 0.335, 0.349, 0.376, 0.395, 0.424, 0.462, 0.481, 0.499, 0.530, 0.538, 0.564, 0.585, 0.605, 0.624, 0.649, 0.680, 0.705, 0.726, 0.751, 0.771, 0.789, 0.818, 0.839, 0.852, 0.868, 0.887, 0.909, 0.932, 0.943, 0.948, 0.962, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    48: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.007, 0.011, 0.013, 0.016, 0.020, 0.020, 0.021, 0.027, 0.039, 0.045, 0.045, 0.047, 0.048, 0.052, 0.059, 0.064, 0.068, 0.072, 0.072, 0.073, 0.077, 0.081, 0.084, 0.089, 0.093, 0.093, 0.095, 0.095, 0.100, 0.111, 0.114, 0.123, 0.129, 0.136, 0.145, 0.150, 0.165, 0.174, 0.188, 0.199, 0.211, 0.234, 0.263, 0.274, 0.284, 0.306, 0.326, 0.338, 0.354, 0.376, 0.399, 0.431, 0.463, 0.485, 0.499, 0.533, 0.542, 0.565, 0.589, 0.606, 0.626, 0.655, 0.682, 0.707, 0.730, 0.755, 0.778, 0.794, 0.819, 0.841, 0.853, 0.869, 0.889, 0.909, 0.932, 0.943, 0.948, 0.964, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    49: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.007, 0.011, 0.013, 0.016, 0.020, 0.020, 0.021, 0.029, 0.039, 0.045, 0.045, 0.047, 0.050, 0.059, 0.061, 0.066, 0.068, 0.072, 0.072, 0.073, 0.077, 0.082, 0.084, 0.091, 0.093, 0.093, 0.095, 0.098, 0.104, 0.111, 0.114, 0.125, 0.131, 0.141, 0.145, 0.154, 0.166, 0.184, 0.191, 0.199, 0.215, 0.238, 0.267, 0.274, 0.293, 0.309, 0.326, 0.342, 0.361, 0.377, 0.406, 0.438, 0.469, 0.490, 0.503, 0.537, 0.551, 0.578, 0.594, 0.608, 0.630, 0.660, 0.687, 0.710, 0.735, 0.762, 0.780, 0.801, 0.821, 0.846, 0.857, 0.878, 0.894, 0.916, 0.932, 0.945, 0.948, 0.964, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    50: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.013, 0.016, 0.016, 0.020, 0.020, 0.021, 0.029, 0.039, 0.045, 0.045, 0.048, 0.050, 0.059, 0.063, 0.068, 0.070, 0.072, 0.072, 0.075, 0.077, 0.082, 0.086, 0.091, 0.093, 0.093, 0.095, 0.102, 0.104, 0.111, 0.118, 0.127, 0.136, 0.145, 0.147, 0.156, 0.170, 0.186, 0.191, 0.200, 0.216, 0.238, 0.268, 0.279, 0.295, 0.311, 0.327, 0.349, 0.363, 0.379, 0.410, 0.444, 0.469, 0.492, 0.512, 0.538, 0.551, 0.580, 0.594, 0.608, 0.631, 0.660, 0.687, 0.716, 0.737, 0.767, 0.785, 0.803, 0.825, 0.846, 0.859, 0.878, 0.896, 0.918, 0.936, 0.945, 0.950, 0.964, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    51: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.005, 0.009, 0.013, 0.016, 0.016, 0.020, 0.020, 0.021, 0.029, 0.039, 0.045, 0.045, 0.048, 0.050, 0.059, 0.064, 0.068, 0.070, 0.072, 0.073, 0.077, 0.077, 0.082, 0.086, 0.091, 0.095, 0.095, 0.098, 0.102, 0.104, 0.113, 0.122, 0.129, 0.138, 0.147, 0.148, 0.165, 0.174, 0.186, 0.191, 0.202, 0.229, 0.238, 0.270, 0.281, 0.297, 0.317, 0.329, 0.351, 0.367, 0.392, 0.415, 0.447, 0.472, 0.492, 0.519, 0.540, 0.558, 0.581, 0.594, 0.608, 0.633, 0.662, 0.692, 0.717, 0.741, 0.775, 0.789, 0.809, 0.826, 0.848, 0.862, 0.878, 0.898, 0.923, 0.936, 0.945, 0.950, 0.964, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    52: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.005, 0.009, 0.013, 0.016, 0.016, 0.021, 0.023, 0.023, 0.029, 0.039, 0.045, 0.045, 0.048, 0.059, 0.061, 0.064, 0.068, 0.070, 0.073, 0.073, 0.077, 0.079, 0.082, 0.086, 0.091, 0.095, 0.095, 0.098, 0.102, 0.104, 0.120, 0.122, 0.129, 0.141, 0.147, 0.150, 0.165, 0.175, 0.186, 0.193, 0.204, 0.233, 0.245, 0.270, 0.283, 0.301, 0.317, 0.335, 0.354, 0.374, 0.397, 0.420, 0.449, 0.474, 0.494, 0.522, 0.546, 0.562, 0.581, 0.597, 0.615, 0.640, 0.662, 0.694, 0.719, 0.746, 0.775, 0.794, 0.810, 0.826, 0.848, 0.864, 0.878, 0.900, 0.923, 0.936, 0.945, 0.950, 0.964, 0.973, 0.982, 0.989, 0.993, 0.993, 0.996, 0.998, 1.000],
    53: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.005, 0.009, 0.014, 0.018, 0.020, 0.023, 0.023, 0.023, 0.030, 0.039, 0.045, 0.045, 0.048, 0.059, 0.063, 0.064, 0.068, 0.072, 0.073, 0.073, 0.079, 0.079, 0.084, 0.089, 0.093, 0.100, 0.100, 0.100, 0.111, 0.113, 0.125, 0.125, 0.129, 0.141, 0.147, 0.152, 0.170, 0.177, 0.193, 0.199, 0.206, 0.238, 0.249, 0.277, 0.288, 0.306, 0.318, 0.338, 0.369, 0.385, 0.410, 0.431, 0.453, 0.479, 0.501, 0.522, 0.547, 0.567, 0.583, 0.603, 0.621, 0.644, 0.664, 0.694, 0.728, 0.748, 0.778, 0.800, 0.810, 0.832, 0.848, 0.868, 0.887, 0.905, 0.925, 0.936, 0.948, 0.950, 0.964, 0.973, 0.982, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    54: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.005, 0.013, 0.014, 0.018, 0.020, 0.023, 0.023, 0.023, 0.030, 0.043, 0.045, 0.047, 0.050, 0.061, 0.063, 0.064, 0.068, 0.072, 0.073, 0.075, 0.079, 0.081, 0.086, 0.089, 0.093, 0.100, 0.100, 0.106, 0.111, 0.120, 0.125, 0.127, 0.129, 0.141, 0.148, 0.152, 0.172, 0.179, 0.195, 0.200, 0.213, 0.245, 0.250, 0.277, 0.288, 0.308, 0.324, 0.345, 0.372, 0.390, 0.413, 0.431, 0.456, 0.485, 0.506, 0.526, 0.547, 0.567, 0.587, 0.603, 0.626, 0.644, 0.664, 0.699, 0.730, 0.748, 0.780, 0.800, 0.814, 0.834, 0.853, 0.871, 0.889, 0.905, 0.927, 0.936, 0.948, 0.952, 0.964, 0.973, 0.982, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    55: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.005, 0.013, 0.016, 0.018, 0.020, 0.023, 0.023, 0.023, 0.030, 0.043, 0.045, 0.047, 0.052, 0.061, 0.063, 0.064, 0.068, 0.072, 0.073, 0.075, 0.081, 0.081, 0.086, 0.089, 0.093, 0.100, 0.100, 0.106, 0.113, 0.122, 0.127, 0.129, 0.136, 0.141, 0.148, 0.154, 0.172, 0.181, 0.195, 0.206, 0.216, 0.247, 0.252, 0.277, 0.288, 0.311, 0.329, 0.351, 0.377, 0.395, 0.413, 0.433, 0.456, 0.488, 0.506, 0.526, 0.551, 0.572, 0.587, 0.605, 0.630, 0.646, 0.665, 0.699, 0.732, 0.753, 0.785, 0.805, 0.816, 0.841, 0.857, 0.871, 0.893, 0.909, 0.927, 0.936, 0.948, 0.952, 0.970, 0.977, 0.986, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    56: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.007, 0.013, 0.016, 0.018, 0.021, 0.023, 0.023, 0.023, 0.030, 0.043, 0.045, 0.047, 0.057, 0.061, 0.063, 0.064, 0.072, 0.073, 0.073, 0.075, 0.081, 0.081, 0.088, 0.089, 0.093, 0.106, 0.106, 0.107, 0.114, 0.125, 0.127, 0.129, 0.136, 0.141, 0.148, 0.156, 0.172, 0.181, 0.199, 0.206, 0.222, 0.247, 0.252, 0.277, 0.288, 0.315, 0.333, 0.354, 0.383, 0.399, 0.417, 0.438, 0.463, 0.492, 0.508, 0.531, 0.555, 0.574, 0.592, 0.615, 0.633, 0.649, 0.667, 0.705, 0.735, 0.757, 0.791, 0.807, 0.819, 0.843, 0.860, 0.873, 0.893, 0.909, 0.927, 0.936, 0.948, 0.955, 0.970, 0.977, 0.986, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    57: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.005, 0.007, 0.013, 0.020, 0.020, 0.021, 0.023, 0.023, 0.025, 0.036, 0.043, 0.050, 0.052, 0.057, 0.061, 0.063, 0.064, 0.072, 0.073, 0.075, 0.075, 0.082, 0.086, 0.088, 0.089, 0.098, 0.106, 0.106, 0.109, 0.116, 0.125, 0.127, 0.131, 0.136, 0.143, 0.152, 0.161, 0.175, 0.191, 0.199, 0.213, 0.227, 0.247, 0.268, 0.279, 0.302, 0.318, 0.338, 0.358, 0.383, 0.399, 0.428, 0.444, 0.469, 0.497, 0.510, 0.535, 0.564, 0.581, 0.596, 0.621, 0.635, 0.655, 0.673, 0.708, 0.737, 0.762, 0.794, 0.809, 0.828, 0.846, 0.864, 0.873, 0.894, 0.911, 0.928, 0.936, 0.948, 0.959, 0.970, 0.977, 0.986, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    58: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.005, 0.007, 0.013, 0.020, 0.020, 0.021, 0.023, 0.023, 0.025, 0.048, 0.048, 0.050, 0.052, 0.061, 0.063, 0.064, 0.064, 0.072, 0.073, 0.077, 0.079, 0.084, 0.088, 0.088, 0.098, 0.102, 0.106, 0.107, 0.113, 0.120, 0.125, 0.129, 0.131, 0.138, 0.145, 0.154, 0.161, 0.179, 0.195, 0.200, 0.213, 0.227, 0.261, 0.268, 0.281, 0.302, 0.320, 0.338, 0.358, 0.383, 0.406, 0.428, 0.447, 0.470, 0.499, 0.512, 0.537, 0.565, 0.583, 0.597, 0.624, 0.639, 0.657, 0.676, 0.710, 0.741, 0.764, 0.796, 0.812, 0.828, 0.846, 0.864, 0.882, 0.898, 0.912, 0.928, 0.937, 0.950, 0.961, 0.970, 0.977, 0.987, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    59: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.005, 0.007, 0.013, 0.020, 0.020, 0.021, 0.023, 0.023, 0.029, 0.048, 0.048, 0.050, 0.052, 0.061, 0.063, 0.064, 0.064, 0.072, 0.073, 0.077, 0.081, 0.084, 0.088, 0.089, 0.098, 0.102, 0.107, 0.107, 0.113, 0.122, 0.125, 0.129, 0.131, 0.138, 0.145, 0.156, 0.166, 0.181, 0.195, 0.202, 0.215, 0.233, 0.261, 0.274, 0.286, 0.306, 0.324, 0.342, 0.361, 0.388, 0.410, 0.428, 0.451, 0.472, 0.504, 0.517, 0.538, 0.569, 0.585, 0.606, 0.630, 0.642, 0.660, 0.685, 0.716, 0.748, 0.769, 0.803, 0.816, 0.830, 0.848, 0.864, 0.887, 0.898, 0.914, 0.930, 0.941, 0.952, 0.962, 0.971, 0.979, 0.987, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    60: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.005, 0.007, 0.016, 0.021, 0.021, 0.023, 0.023, 0.023, 0.034, 0.048, 0.048, 0.050, 0.054, 0.061, 0.063, 0.066, 0.070, 0.073, 0.075, 0.079, 0.081, 0.086, 0.088, 0.089, 0.098, 0.104, 0.107, 0.109, 0.113, 0.123, 0.125, 0.129, 0.131, 0.140, 0.145, 0.165, 0.168, 0.181, 0.197, 0.206, 0.216, 0.234, 0.261, 0.275, 0.288, 0.306, 0.324, 0.343, 0.361, 0.390, 0.413, 0.429, 0.454, 0.474, 0.508, 0.517, 0.542, 0.574, 0.589, 0.614, 0.633, 0.644, 0.662, 0.687, 0.719, 0.748, 0.776, 0.805, 0.819, 0.835, 0.848, 0.869, 0.887, 0.900, 0.918, 0.932, 0.941, 0.952, 0.962, 0.971, 0.982, 0.987, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    61: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.005, 0.007, 0.016, 0.021, 0.021, 0.023, 0.023, 0.023, 0.034, 0.048, 0.048, 0.050, 0.059, 0.063, 0.063, 0.066, 0.070, 0.075, 0.075, 0.081, 0.084, 0.088, 0.089, 0.089, 0.098, 0.104, 0.113, 0.116, 0.118, 0.123, 0.125, 0.129, 0.131, 0.141, 0.147, 0.165, 0.170, 0.191, 0.202, 0.213, 0.220, 0.236, 0.265, 0.275, 0.295, 0.311, 0.327, 0.347, 0.367, 0.390, 0.415, 0.429, 0.456, 0.481, 0.508, 0.522, 0.551, 0.576, 0.594, 0.617, 0.635, 0.644, 0.669, 0.692, 0.726, 0.755, 0.782, 0.809, 0.823, 0.839, 0.850, 0.877, 0.887, 0.903, 0.919, 0.934, 0.941, 0.955, 0.962, 0.971, 0.982, 0.987, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    62: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.005, 0.011, 0.018, 0.021, 0.021, 0.023, 0.023, 0.027, 0.034, 0.048, 0.048, 0.055, 0.061, 0.063, 0.063, 0.066, 0.070, 0.075, 0.077, 0.081, 0.086, 0.088, 0.091, 0.091, 0.100, 0.106, 0.113, 0.116, 0.118, 0.123, 0.125, 0.129, 0.138, 0.145, 0.152, 0.166, 0.175, 0.193, 0.202, 0.215, 0.220, 0.238, 0.265, 0.275, 0.299, 0.313, 0.329, 0.349, 0.376, 0.395, 0.419, 0.433, 0.463, 0.483, 0.513, 0.528, 0.556, 0.578, 0.599, 0.621, 0.639, 0.649, 0.673, 0.694, 0.726, 0.758, 0.792, 0.810, 0.825, 0.839, 0.850, 0.877, 0.887, 0.905, 0.923, 0.936, 0.941, 0.955, 0.962, 0.971, 0.982, 0.987, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    63: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.014, 0.020, 0.021, 0.021, 0.023, 0.023, 0.029, 0.034, 0.048, 0.048, 0.057, 0.061, 0.063, 0.064, 0.070, 0.072, 0.077, 0.077, 0.081, 0.086, 0.088, 0.091, 0.091, 0.102, 0.106, 0.116, 0.116, 0.118, 0.123, 0.125, 0.138, 0.140, 0.147, 0.157, 0.168, 0.188, 0.197, 0.209, 0.218, 0.227, 0.240, 0.274, 0.284, 0.301, 0.313, 0.329, 0.356, 0.379, 0.399, 0.419, 0.435, 0.463, 0.490, 0.515, 0.530, 0.556, 0.578, 0.603, 0.624, 0.639, 0.653, 0.678, 0.701, 0.730, 0.766, 0.794, 0.810, 0.826, 0.839, 0.852, 0.877, 0.887, 0.907, 0.927, 0.936, 0.943, 0.955, 0.964, 0.971, 0.984, 0.987, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    64: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.014, 0.020, 0.021, 0.021, 0.023, 0.023, 0.029, 0.034, 0.048, 0.048, 0.059, 0.063, 0.063, 0.064, 0.070, 0.075, 0.077, 0.077, 0.082, 0.088, 0.089, 0.091, 0.091, 0.102, 0.109, 0.116, 0.116, 0.120, 0.123, 0.127, 0.138, 0.140, 0.148, 0.159, 0.175, 0.190, 0.199, 0.213, 0.218, 0.229, 0.242, 0.275, 0.288, 0.308, 0.317, 0.331, 0.358, 0.379, 0.401, 0.419, 0.436, 0.470, 0.490, 0.519, 0.535, 0.562, 0.583, 0.605, 0.633, 0.642, 0.655, 0.682, 0.703, 0.735, 0.773, 0.794, 0.810, 0.830, 0.839, 0.852, 0.882, 0.889, 0.907, 0.928, 0.939, 0.945, 0.955, 0.964, 0.971, 0.984, 0.987, 0.991, 0.993, 0.993, 0.996, 0.998, 1.000],
    65: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.014, 0.020, 0.021, 0.021, 0.023, 0.023, 0.029, 0.034, 0.048, 0.048, 0.059, 0.063, 0.063, 0.066, 0.070, 0.075, 0.077, 0.077, 0.084, 0.088, 0.091, 0.093, 0.095, 0.104, 0.109, 0.116, 0.116, 0.120, 0.123, 0.132, 0.138, 0.140, 0.150, 0.166, 0.175, 0.193, 0.200, 0.216, 0.218, 0.231, 0.254, 0.277, 0.290, 0.309, 0.320, 0.338, 0.363, 0.383, 0.401, 0.422, 0.440, 0.472, 0.496, 0.522, 0.538, 0.564, 0.589, 0.612, 0.635, 0.648, 0.657, 0.689, 0.712, 0.739, 0.776, 0.796, 0.818, 0.830, 0.843, 0.862, 0.884, 0.893, 0.909, 0.930, 0.939, 0.945, 0.955, 0.964, 0.971, 0.984, 0.989, 0.991, 0.993, 0.993, 0.998, 0.998, 1.000],
    66: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.020, 0.021, 0.021, 0.023, 0.023, 0.029, 0.034, 0.048, 0.052, 0.059, 0.063, 0.063, 0.066, 0.070, 0.075, 0.077, 0.079, 0.084, 0.088, 0.093, 0.093, 0.097, 0.104, 0.109, 0.116, 0.116, 0.120, 0.123, 0.132, 0.138, 0.140, 0.150, 0.166, 0.175, 0.193, 0.204, 0.216, 0.218, 0.234, 0.256, 0.281, 0.293, 0.309, 0.320, 0.343, 0.365, 0.388, 0.403, 0.422, 0.440, 0.476, 0.497, 0.524, 0.540, 0.567, 0.596, 0.619, 0.637, 0.648, 0.660, 0.692, 0.714, 0.742, 0.780, 0.798, 0.818, 0.830, 0.844, 0.866, 0.884, 0.893, 0.909, 0.932, 0.939, 0.945, 0.955, 0.964, 0.971, 0.986, 0.989, 0.991, 0.993, 0.995, 0.998, 0.998, 1.000],
    67: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.020, 0.021, 0.021, 0.023, 0.023, 0.030, 0.034, 0.052, 0.054, 0.059, 0.063, 0.064, 0.066, 0.072, 0.077, 0.079, 0.081, 0.084, 0.088, 0.093, 0.093, 0.097, 0.104, 0.113, 0.116, 0.120, 0.122, 0.123, 0.134, 0.140, 0.143, 0.150, 0.166, 0.179, 0.197, 0.208, 0.218, 0.220, 0.236, 0.256, 0.283, 0.295, 0.311, 0.327, 0.347, 0.372, 0.388, 0.406, 0.426, 0.444, 0.481, 0.497, 0.528, 0.542, 0.567, 0.605, 0.621, 0.639, 0.648, 0.664, 0.698, 0.714, 0.751, 0.782, 0.800, 0.818, 0.830, 0.846, 0.866, 0.884, 0.896, 0.912, 0.934, 0.939, 0.948, 0.955, 0.966, 0.971, 0.986, 0.989, 0.991, 0.995, 0.995, 0.998, 0.998, 1.000],
    68: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.020, 0.021, 0.021, 0.023, 0.023, 0.030, 0.034, 0.052, 0.054, 0.059, 0.063, 0.064, 0.068, 0.073, 0.077, 0.081, 0.081, 0.084, 0.089, 0.093, 0.100, 0.102, 0.104, 0.113, 0.116, 0.122, 0.123, 0.125, 0.134, 0.140, 0.145, 0.154, 0.170, 0.179, 0.199, 0.209, 0.218, 0.222, 0.236, 0.256, 0.283, 0.304, 0.320, 0.329, 0.354, 0.376, 0.392, 0.408, 0.426, 0.447, 0.481, 0.499, 0.530, 0.546, 0.572, 0.605, 0.623, 0.639, 0.648, 0.667, 0.701, 0.719, 0.753, 0.785, 0.800, 0.818, 0.830, 0.852, 0.868, 0.887, 0.903, 0.918, 0.934, 0.939, 0.948, 0.955, 0.966, 0.973, 0.986, 0.989, 0.991, 0.995, 0.995, 0.998, 0.998, 1.000],
    69: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.020, 0.021, 0.021, 0.023, 0.025, 0.034, 0.034, 0.052, 0.054, 0.061, 0.066, 0.066, 0.070, 0.077, 0.079, 0.081, 0.081, 0.084, 0.091, 0.095, 0.100, 0.102, 0.106, 0.120, 0.120, 0.123, 0.123, 0.125, 0.136, 0.143, 0.145, 0.161, 0.177, 0.186, 0.202, 0.215, 0.218, 0.227, 0.242, 0.263, 0.293, 0.317, 0.327, 0.329, 0.358, 0.381, 0.392, 0.411, 0.428, 0.458, 0.488, 0.506, 0.535, 0.551, 0.581, 0.612, 0.626, 0.640, 0.651, 0.673, 0.710, 0.723, 0.758, 0.787, 0.801, 0.826, 0.843, 0.862, 0.880, 0.894, 0.905, 0.918, 0.934, 0.939, 0.948, 0.955, 0.966, 0.973, 0.986, 0.989, 0.993, 0.995, 0.996, 0.998, 1.000, 1.000],
    70: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.020, 0.021, 0.021, 0.023, 0.025, 0.034, 0.036, 0.052, 0.054, 0.061, 0.066, 0.068, 0.072, 0.077, 0.079, 0.081, 0.082, 0.084, 0.091, 0.095, 0.100, 0.104, 0.109, 0.120, 0.120, 0.123, 0.123, 0.129, 0.138, 0.147, 0.150, 0.161, 0.177, 0.186, 0.202, 0.216, 0.218, 0.227, 0.243, 0.268, 0.299, 0.317, 0.327, 0.329, 0.360, 0.383, 0.395, 0.413, 0.428, 0.463, 0.492, 0.510, 0.537, 0.551, 0.587, 0.612, 0.628, 0.640, 0.653, 0.678, 0.716, 0.726, 0.762, 0.787, 0.801, 0.826, 0.843, 0.864, 0.880, 0.898, 0.907, 0.919, 0.936, 0.939, 0.948, 0.955, 0.966, 0.973, 0.986, 0.989, 0.993, 0.995, 0.998, 1.000, 1.000, 1.000],
    71: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.021, 0.021, 0.021, 0.023, 0.025, 0.036, 0.041, 0.052, 0.055, 0.063, 0.066, 0.068, 0.072, 0.077, 0.079, 0.081, 0.084, 0.084, 0.091, 0.095, 0.100, 0.106, 0.114, 0.120, 0.120, 0.123, 0.123, 0.131, 0.140, 0.147, 0.150, 0.161, 0.177, 0.188, 0.202, 0.216, 0.218, 0.227, 0.245, 0.272, 0.304, 0.320, 0.329, 0.338, 0.360, 0.385, 0.395, 0.415, 0.435, 0.463, 0.497, 0.515, 0.537, 0.553, 0.589, 0.612, 0.628, 0.642, 0.658, 0.691, 0.717, 0.726, 0.764, 0.787, 0.803, 0.828, 0.843, 0.868, 0.880, 0.898, 0.909, 0.919, 0.936, 0.945, 0.948, 0.957, 0.968, 0.973, 0.986, 0.989, 0.993, 0.995, 1.000, 1.000, 1.000, 1.000],
    72: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.005, 0.007, 0.009, 0.018, 0.021, 0.023, 0.023, 0.023, 0.030, 0.039, 0.045, 0.054, 0.057, 0.063, 0.066, 0.070, 0.072, 0.079, 0.079, 0.081, 0.084, 0.084, 0.091, 0.095, 0.100, 0.106, 0.114, 0.120, 0.120, 0.123, 0.123, 0.134, 0.141, 0.148, 0.154, 0.161, 0.177, 0.193, 0.204, 0.216, 0.222, 0.229, 0.258, 0.275, 0.309, 0.322, 0.335, 0.343, 0.361, 0.386, 0.397, 0.417, 0.440, 0.467, 0.501, 0.519, 0.540, 0.558, 0.597, 0.615, 0.635, 0.651, 0.667, 0.692, 0.717, 0.732, 0.767, 0.787, 0.803, 0.830, 0.848, 0.868, 0.882, 0.902, 0.912, 0.919, 0.937, 0.945, 0.950, 0.959, 0.968, 0.975, 0.986, 0.989, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000],
    73: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.007, 0.009, 0.018, 0.021, 0.023, 0.023, 0.023, 0.030, 0.039, 0.045, 0.054, 0.057, 0.066, 0.070, 0.072, 0.077, 0.079, 0.082, 0.084, 0.086, 0.086, 0.091, 0.095, 0.102, 0.109, 0.118, 0.120, 0.122, 0.123, 0.127, 0.136, 0.141, 0.150, 0.157, 0.161, 0.179, 0.193, 0.206, 0.216, 0.225, 0.233, 0.259, 0.275, 0.309, 0.324, 0.338, 0.343, 0.363, 0.390, 0.404, 0.422, 0.445, 0.472, 0.508, 0.522, 0.544, 0.567, 0.599, 0.621, 0.639, 0.653, 0.671, 0.696, 0.721, 0.733, 0.771, 0.789, 0.805, 0.832, 0.852, 0.871, 0.882, 0.905, 0.912, 0.923, 0.939, 0.948, 0.952, 0.959, 0.968, 0.975, 0.986, 0.989, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000],
    74: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.007, 0.009, 0.018, 0.021, 0.023, 0.023, 0.023, 0.030, 0.039, 0.045, 0.054, 0.057, 0.066, 0.070, 0.072, 0.077, 0.079, 0.082, 0.084, 0.086, 0.086, 0.091, 0.095, 0.104, 0.109, 0.118, 0.120, 0.122, 0.123, 0.129, 0.136, 0.143, 0.150, 0.157, 0.165, 0.182, 0.195, 0.206, 0.216, 0.225, 0.238, 0.259, 0.275, 0.309, 0.327, 0.338, 0.343, 0.365, 0.394, 0.404, 0.426, 0.449, 0.472, 0.512, 0.522, 0.546, 0.571, 0.601, 0.623, 0.640, 0.657, 0.674, 0.699, 0.723, 0.737, 0.775, 0.794, 0.812, 0.832, 0.853, 0.871, 0.886, 0.907, 0.916, 0.927, 0.941, 0.950, 0.953, 0.962, 0.968, 0.975, 0.986, 0.989, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000],
    75: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.007, 0.007, 0.007, 0.009, 0.020, 0.023, 0.023, 0.023, 0.027, 0.030, 0.041, 0.045, 0.054, 0.059, 0.066, 0.070, 0.072, 0.079, 0.079, 0.084, 0.086, 0.086, 0.086, 0.091, 0.106, 0.107, 0.109, 0.120, 0.120, 0.123, 0.125, 0.129, 0.140, 0.143, 0.150, 0.157, 0.165, 0.186, 0.195, 0.209, 0.216, 0.225, 0.242, 0.261, 0.275, 0.309, 0.329, 0.340, 0.352, 0.374, 0.395, 0.410, 0.428, 0.453, 0.476, 0.513, 0.526, 0.551, 0.578, 0.601, 0.623, 0.642, 0.662, 0.676, 0.703, 0.723, 0.739, 0.775, 0.794, 0.814, 0.835, 0.853, 0.871, 0.891, 0.909, 0.918, 0.928, 0.943, 0.950, 0.953, 0.962, 0.968, 0.975, 0.986, 0.991, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000],
    76: [0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.007, 0.007, 0.007, 0.007, 0.013, 0.021, 0.023, 0.023, 0.027, 0.027, 0.032, 0.041, 0.047, 0.055, 0.059, 0.068, 0.070, 0.073, 0.079, 0.084, 0.084, 0.086, 0.086, 0.086, 0.095, 0.106, 0.107, 0.109, 0.120, 0.120, 0.123, 0.125, 0.131, 0.140, 0.143, 0.150, 0.157, 0.165, 0.186, 0.195, 0.209, 0.216, 0.229, 0.243, 0.263, 0.281, 0.313, 0.333, 0.342, 0.360, 0.376, 0.397, 0.415, 0.429, 0.454, 0.478, 0.517, 0.533, 0.560, 0.578, 0.603, 0.626, 0.644, 0.665, 0.685, 0.703, 0.725, 0.742, 0.776, 0.796, 0.816, 0.839, 0.853, 0.873, 0.893, 0.909, 0.918, 0.928, 0.943, 0.950, 0.953, 0.962, 0.970, 0.977, 0.986, 0.991, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000],
    77: [0, 0, 0, 0, 0, 0, 0, 0, 0.005, 0.007, 0.007, 0.007, 0.011, 0.014, 0.021, 0.023, 0.027, 0.027, 0.029, 0.032, 0.045, 0.052, 0.057, 0.061, 0.068, 0.073, 0.077, 0.082, 0.084, 0.084, 0.086, 0.086, 0.088, 0.097, 0.107, 0.111, 0.113, 0.120, 0.122, 0.125, 0.127, 0.132, 0.140, 0.147, 0.150, 0.161, 0.166, 0.188, 0.195, 0.209, 0.218, 0.231, 0.254, 0.270, 0.283, 0.318, 0.333, 0.345, 0.367, 0.386, 0.399, 0.415, 0.431, 0.462, 0.481, 0.519, 0.533, 0.569, 0.587, 0.612, 0.633, 0.651, 0.673, 0.689, 0.712, 0.732, 0.753, 0.780, 0.801, 0.818, 0.844, 0.857, 0.878, 0.894, 0.909, 0.921, 0.932, 0.943, 0.950, 0.955, 0.964, 0.970, 0.977, 0.987, 0.991, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000],
    78: [0, 0, 0, 0, 0, 0, 0, 0, 0.005, 0.007, 0.007, 0.007, 0.011, 0.014, 0.021, 0.023, 0.027, 0.027, 0.029, 0.032, 0.045, 0.054, 0.057, 0.061, 0.070, 0.073, 0.077, 0.082, 0.084, 0.084, 0.086, 0.086, 0.091, 0.098, 0.107, 0.111, 0.113, 0.120, 0.122, 0.125, 0.127, 0.132, 0.141, 0.147, 0.154, 0.161, 0.166, 0.188, 0.195, 0.209, 0.220, 0.233, 0.254, 0.270, 0.284, 0.320, 0.335, 0.347, 0.369, 0.386, 0.401, 0.415, 0.435, 0.465, 0.485, 0.522, 0.535, 0.571, 0.587, 0.614, 0.635, 0.653, 0.674, 0.692, 0.712, 0.737, 0.762, 0.785, 0.803, 0.821, 0.844, 0.857, 0.878, 0.894, 0.909, 0.923, 0.932, 0.945, 0.952, 0.957, 0.964, 0.970, 0.977, 0.987, 0.991, 0.993, 0.998, 1.000, 1.000, 1.000, 1.000],
    79: [0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.007, 0.007, 0.009, 0.014, 0.020, 0.021, 0.029, 0.029, 0.029, 0.030, 0.038, 0.048, 0.057, 0.061, 0.068, 0.072, 0.077, 0.082, 0.082, 0.084, 0.084, 0.086, 0.089, 0.093, 0.102, 0.113, 0.113, 0.116, 0.122, 0.122, 0.125, 0.127, 0.134, 0.145, 0.150, 0.157, 0.166, 0.179, 0.193, 0.200, 0.215, 0.224, 0.238, 0.259, 0.275, 0.293, 0.329, 0.338, 0.349, 0.374, 0.388, 0.408, 0.424, 0.449, 0.472, 0.490, 0.522, 0.544, 0.578, 0.592, 0.615, 0.637, 0.653, 0.682, 0.694, 0.714, 0.742, 0.766, 0.789, 0.803, 0.821, 0.850, 0.859, 0.878, 0.896, 0.911, 0.923, 0.932, 0.945, 0.952, 0.959, 0.964, 0.970, 0.977, 0.987, 0.991, 0.993, 0.998, 1.000, 1.000, 1.000, 1.000],
    80: [0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.007, 0.007, 0.009, 0.014, 0.021, 0.023, 0.029, 0.029, 0.029, 0.030, 0.043, 0.050, 0.057, 0.063, 0.068, 0.075, 0.077, 0.082, 0.082, 0.084, 0.086, 0.086, 0.095, 0.097, 0.102, 0.114, 0.114, 0.120, 0.122, 0.122, 0.125, 0.129, 0.138, 0.150, 0.159, 0.166, 0.172, 0.182, 0.197, 0.202, 0.218, 0.231, 0.250, 0.267, 0.283, 0.304, 0.331, 0.340, 0.354, 0.374, 0.390, 0.413, 0.426, 0.454, 0.474, 0.496, 0.524, 0.551, 0.578, 0.596, 0.621, 0.640, 0.660, 0.683, 0.694, 0.719, 0.744, 0.771, 0.792, 0.805, 0.825, 0.855, 0.860, 0.880, 0.898, 0.912, 0.925, 0.936, 0.945, 0.952, 0.959, 0.964, 0.970, 0.977, 0.989, 0.991, 0.993, 0.998, 1.000, 1.000, 1.000, 1.000],
    81: [0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.007, 0.007, 0.011, 0.018, 0.023, 0.025, 0.029, 0.029, 0.029, 0.030, 0.043, 0.050, 0.057, 0.063, 0.070, 0.075, 0.082, 0.082, 0.082, 0.084, 0.088, 0.091, 0.095, 0.097, 0.102, 0.114, 0.116, 0.120, 0.122, 0.123, 0.125, 0.132, 0.140, 0.154, 0.163, 0.174, 0.182, 0.191, 0.200, 0.204, 0.224, 0.236, 0.254, 0.268, 0.290, 0.306, 0.333, 0.342, 0.358, 0.376, 0.394, 0.415, 0.440, 0.456, 0.479, 0.503, 0.535, 0.556, 0.583, 0.603, 0.633, 0.648, 0.664, 0.685, 0.699, 0.728, 0.753, 0.775, 0.794, 0.812, 0.828, 0.855, 0.860, 0.886, 0.898, 0.914, 0.925, 0.936, 0.948, 0.952, 0.959, 0.964, 0.970, 0.979, 0.989, 0.991, 0.995, 0.998, 1.000, 1.000, 1.000, 1.000],
    82: [0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.011, 0.020, 0.023, 0.025, 0.029, 0.029, 0.029, 0.030, 0.048, 0.057, 0.059, 0.063, 0.070, 0.079, 0.082, 0.082, 0.082, 0.088, 0.091, 0.091, 0.095, 0.098, 0.104, 0.114, 0.116, 0.120, 0.122, 0.123, 0.129, 0.134, 0.141, 0.157, 0.163, 0.175, 0.184, 0.191, 0.200, 0.211, 0.224, 0.238, 0.256, 0.272, 0.293, 0.306, 0.333, 0.343, 0.360, 0.383, 0.394, 0.417, 0.444, 0.460, 0.483, 0.508, 0.540, 0.556, 0.587, 0.612, 0.633, 0.648, 0.669, 0.691, 0.708, 0.730, 0.755, 0.775, 0.798, 0.814, 0.830, 0.857, 0.864, 0.889, 0.898, 0.914, 0.925, 0.937, 0.948, 0.953, 0.959, 0.964, 0.970, 0.982, 0.989, 0.991, 0.996, 0.998, 1.000, 1.000, 1.000, 1.000],
    83: [0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.007, 0.007, 0.014, 0.020, 0.023, 0.029, 0.029, 0.029, 0.029, 0.036, 0.050, 0.061, 0.063, 0.072, 0.075, 0.081, 0.082, 0.082, 0.082, 0.088, 0.091, 0.095, 0.097, 0.104, 0.107, 0.114, 0.116, 0.120, 0.123, 0.125, 0.131, 0.141, 0.145, 0.157, 0.163, 0.175, 0.186, 0.195, 0.204, 0.213, 0.227, 0.243, 0.259, 0.277, 0.295, 0.306, 0.333, 0.345, 0.363, 0.383, 0.397, 0.429, 0.445, 0.462, 0.487, 0.517, 0.546, 0.560, 0.590, 0.617, 0.633, 0.655, 0.674, 0.696, 0.717, 0.739, 0.755, 0.780, 0.800, 0.816, 0.834, 0.859, 0.864, 0.889, 0.900, 0.914, 0.928, 0.939, 0.948, 0.953, 0.959, 0.964, 0.970, 0.984, 0.989, 0.991, 0.996, 0.998, 1.000, 1.000, 1.000, 1.000],
    84: [0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.007, 0.009, 0.016, 0.020, 0.027, 0.029, 0.029, 0.029, 0.032, 0.041, 0.050, 0.061, 0.063, 0.073, 0.077, 0.081, 0.082, 0.082, 0.082, 0.089, 0.091, 0.097, 0.098, 0.106, 0.107, 0.114, 0.116, 0.120, 0.123, 0.127, 0.134, 0.141, 0.148, 0.157, 0.170, 0.179, 0.186, 0.195, 0.204, 0.215, 0.227, 0.245, 0.263, 0.281, 0.295, 0.308, 0.335, 0.349, 0.367, 0.392, 0.410, 0.433, 0.447, 0.467, 0.494, 0.526, 0.549, 0.567, 0.596, 0.621, 0.639, 0.660, 0.678, 0.703, 0.721, 0.741, 0.758, 0.780, 0.801, 0.816, 0.835, 0.860, 0.868, 0.891, 0.902, 0.921, 0.934, 0.939, 0.948, 0.953, 0.959, 0.964, 0.973, 0.984, 0.991, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000, 1.000],
    85: [0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.009, 0.011, 0.018, 0.020, 0.029, 0.029, 0.029, 0.029, 0.032, 0.048, 0.052, 0.061, 0.068, 0.073, 0.077, 0.084, 0.086, 0.086, 0.086, 0.089, 0.097, 0.097, 0.102, 0.106, 0.107, 0.114, 0.116, 0.122, 0.125, 0.132, 0.140, 0.147, 0.156, 0.168, 0.175, 0.182, 0.191, 0.202, 0.211, 0.218, 0.236, 0.254, 0.274, 0.283, 0.295, 0.311, 0.343, 0.354, 0.379, 0.408, 0.420, 0.445, 0.458, 0.479, 0.501, 0.533, 0.553, 0.574, 0.608, 0.637, 0.644, 0.665, 0.687, 0.710, 0.730, 0.742, 0.766, 0.782, 0.805, 0.821, 0.839, 0.864, 0.873, 0.898, 0.903, 0.923, 0.937, 0.946, 0.952, 0.957, 0.962, 0.968, 0.973, 0.986, 0.993, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000, 1.000],
    86: [0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.009, 0.011, 0.013, 0.018, 0.021, 0.029, 0.029, 0.029, 0.029, 0.039, 0.050, 0.055, 0.061, 0.070, 0.073, 0.084, 0.084, 0.086, 0.086, 0.086, 0.089, 0.097, 0.098, 0.104, 0.107, 0.109, 0.114, 0.116, 0.122, 0.125, 0.134, 0.141, 0.152, 0.165, 0.175, 0.175, 0.186, 0.195, 0.206, 0.216, 0.225, 0.245, 0.267, 0.281, 0.284, 0.297, 0.313, 0.343, 0.361, 0.390, 0.411, 0.426, 0.451, 0.463, 0.479, 0.510, 0.538, 0.558, 0.581, 0.617, 0.640, 0.648, 0.673, 0.692, 0.714, 0.730, 0.744, 0.767, 0.789, 0.807, 0.823, 0.841, 0.868, 0.875, 0.903, 0.911, 0.925, 0.943, 0.946, 0.952, 0.959, 0.964, 0.970, 0.975, 0.986, 0.993, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000, 1.000],
    87: [0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.013, 0.020, 0.027, 0.029, 0.029, 0.029, 0.030, 0.039, 0.052, 0.055, 0.061, 0.070, 0.073, 0.084, 0.086, 0.086, 0.086, 0.088, 0.089, 0.098, 0.100, 0.106, 0.107, 0.111, 0.116, 0.122, 0.125, 0.131, 0.134, 0.141, 0.159, 0.170, 0.177, 0.184, 0.190, 0.195, 0.206, 0.220, 0.229, 0.254, 0.268, 0.283, 0.292, 0.299, 0.315, 0.347, 0.369, 0.392, 0.417, 0.431, 0.454, 0.469, 0.481, 0.512, 0.540, 0.562, 0.589, 0.617, 0.642, 0.651, 0.676, 0.694, 0.719, 0.733, 0.744, 0.771, 0.794, 0.810, 0.830, 0.848, 0.873, 0.884, 0.907, 0.911, 0.932, 0.945, 0.950, 0.953, 0.959, 0.966, 0.973, 0.975, 0.986, 0.993, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000, 1.000],
    88: [0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.013, 0.021, 0.027, 0.029, 0.029, 0.029, 0.030, 0.039, 0.052, 0.055, 0.061, 0.072, 0.075, 0.086, 0.086, 0.086, 0.088, 0.088, 0.093, 0.098, 0.102, 0.106, 0.109, 0.111, 0.120, 0.122, 0.127, 0.132, 0.134, 0.141, 0.159, 0.170, 0.177, 0.190, 0.193, 0.199, 0.213, 0.224, 0.234, 0.259, 0.272, 0.286, 0.297, 0.306, 0.322, 0.352, 0.381, 0.399, 0.419, 0.436, 0.456, 0.476, 0.494, 0.522, 0.542, 0.569, 0.592, 0.624, 0.646, 0.655, 0.680, 0.696, 0.721, 0.737, 0.751, 0.775, 0.801, 0.823, 0.839, 0.857, 0.877, 0.893, 0.907, 0.914, 0.932, 0.946, 0.950, 0.953, 0.959, 0.966, 0.973, 0.977, 0.989, 0.993, 0.993, 0.996, 1.000, 1.000, 1.000, 1.000, 1.000],
    89: [0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.014, 0.023, 0.027, 0.029, 0.029, 0.029, 0.038, 0.041, 0.052, 0.059, 0.064, 0.077, 0.079, 0.086, 0.086, 0.086, 0.088, 0.093, 0.097, 0.102, 0.104, 0.106, 0.109, 0.114, 0.120, 0.123, 0.131, 0.132, 0.134, 0.154, 0.170, 0.174, 0.181, 0.200, 0.202, 0.209, 0.216, 0.240, 0.249, 0.268, 0.281, 0.290, 0.304, 0.315, 0.338, 0.372, 0.388, 0.406, 0.426, 0.445, 0.467, 0.487, 0.515, 0.537, 0.562, 0.574, 0.605, 0.633, 0.649, 0.660, 0.696, 0.707, 0.732, 0.742, 0.767, 0.794, 0.819, 0.839, 0.850, 0.866, 0.886, 0.894, 0.911, 0.919, 0.939, 0.952, 0.953, 0.957, 0.962, 0.966, 0.975, 0.979, 0.989, 0.995, 0.996, 0.998, 1.000, 1.000, 1.000, 1.000, 1.000],
    90: [0, 0, 0, 0, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.018, 0.023, 0.027, 0.029, 0.029, 0.030, 0.039, 0.048, 0.057, 0.063, 0.075, 0.077, 0.081, 0.086, 0.088, 0.091, 0.095, 0.095, 0.098, 0.102, 0.107, 0.107, 0.109, 0.120, 0.125, 0.131, 0.132, 0.140, 0.143, 0.157, 0.172, 0.181, 0.190, 0.208, 0.208, 0.215, 0.225, 0.250, 0.263, 0.272, 0.286, 0.301, 0.311, 0.322, 0.356, 0.376, 0.392, 0.415, 0.438, 0.462, 0.479, 0.497, 0.524, 0.540, 0.569, 0.585, 0.612, 0.640, 0.655, 0.676, 0.701, 0.712, 0.735, 0.748, 0.773, 0.801, 0.828, 0.846, 0.857, 0.868, 0.889, 0.898, 0.916, 0.928, 0.941, 0.953, 0.955, 0.961, 0.968, 0.970, 0.977, 0.980, 0.991, 0.996, 0.998, 0.998, 1.000, 1.000, 1.000, 1.000, 1.000]
}


LIGHT_PERCENTILES = {
    1: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.007, 0.007, 0.007, 0.009, 0.009, 0.011, 0.023, 0.027, 0.027, 0.028, 0.037, 0.037, 0.044, 0.048, 0.066, 0.069, 0.076, 0.087, 0.094, 0.117, 0.144, 0.160, 0.185, 0.215, 0.247, 0.284, 0.302, 0.323, 0.355, 0.382, 0.410, 0.464, 0.490, 0.529, 0.556, 0.584, 0.609, 0.650, 0.677, 0.687, 0.710, 0.726, 0.734, 1.000],
    2: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.007, 0.007, 0.009, 0.009, 0.012, 0.012, 0.014, 0.016, 0.021, 0.023, 0.027, 0.030, 0.046, 0.046, 0.057, 0.060, 0.064, 0.067, 0.073, 0.083, 0.091, 0.107, 0.115, 0.146, 0.160, 0.187, 0.206, 0.229, 0.256, 0.282, 0.316, 0.339, 0.371, 0.400, 0.442, 0.480, 0.499, 0.510, 0.531, 0.575, 0.611, 0.647, 0.673, 0.696, 0.710, 0.723, 0.741, 0.750, 0.755, 1.000],
    3: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.007, 0.007, 0.007, 0.007, 0.012, 0.014, 0.014, 0.020, 0.021, 0.021, 0.023, 0.025, 0.032, 0.032, 0.036, 0.048, 0.057, 0.059, 0.062, 0.071, 0.073, 0.089, 0.094, 0.103, 0.115, 0.140, 0.163, 0.171, 0.197, 0.238, 0.249, 0.277, 0.304, 0.345, 0.361, 0.401, 0.432, 0.458, 0.504, 0.524, 0.538, 0.568, 0.607, 0.629, 0.655, 0.680, 0.698, 0.714, 0.725, 0.744, 0.755, 0.764, 1.000],
    4: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.011, 0.012, 0.016, 0.020, 0.021, 0.021, 0.023, 0.025, 0.025, 0.027, 0.028, 0.030, 0.034, 0.036, 0.044, 0.055, 0.059, 0.064, 0.066, 0.071, 0.075, 0.083, 0.089, 0.105, 0.119, 0.126, 0.137, 0.162, 0.178, 0.211, 0.231, 0.261, 0.275, 0.300, 0.337, 0.369, 0.405, 0.440, 0.467, 0.494, 0.524, 0.536, 0.581, 0.597, 0.629, 0.648, 0.673, 0.691, 0.707, 0.725, 0.742, 0.755, 0.762, 0.773, 1.000],
    5: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.009, 0.012, 0.016, 0.020, 0.020, 0.023, 0.023, 0.025, 0.027, 0.027, 0.027, 0.030, 0.034, 0.034, 0.037, 0.046, 0.059, 0.060, 0.067, 0.071, 0.073, 0.082, 0.089, 0.096, 0.112, 0.123, 0.133, 0.151, 0.176, 0.197, 0.222, 0.242, 0.281, 0.293, 0.321, 0.353, 0.387, 0.417, 0.458, 0.485, 0.504, 0.535, 0.558, 0.591, 0.609, 0.636, 0.655, 0.675, 0.694, 0.710, 0.730, 0.744, 0.758, 0.769, 0.776, 1.000],
    6: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.009, 0.011, 0.012, 0.014, 0.016, 0.021, 0.021, 0.023, 0.023, 0.025, 0.027, 0.030, 0.030, 0.032, 0.036, 0.036, 0.037, 0.046, 0.064, 0.071, 0.076, 0.080, 0.085, 0.092, 0.099, 0.114, 0.126, 0.142, 0.153, 0.176, 0.197, 0.213, 0.249, 0.277, 0.297, 0.336, 0.357, 0.387, 0.412, 0.440, 0.471, 0.499, 0.520, 0.545, 0.574, 0.600, 0.627, 0.650, 0.664, 0.684, 0.698, 0.714, 0.734, 0.748, 0.760, 0.776, 0.785, 1.000],
    7: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.009, 0.011, 0.012, 0.016, 0.018, 0.018, 0.021, 0.023, 0.025, 0.025, 0.027, 0.028, 0.030, 0.032, 0.034, 0.037, 0.037, 0.050, 0.057, 0.071, 0.076, 0.080, 0.082, 0.089, 0.094, 0.105, 0.124, 0.133, 0.149, 0.165, 0.188, 0.210, 0.226, 0.261, 0.286, 0.314, 0.345, 0.369, 0.393, 0.423, 0.442, 0.471, 0.504, 0.536, 0.549, 0.581, 0.611, 0.632, 0.655, 0.670, 0.687, 0.702, 0.719, 0.739, 0.753, 0.766, 0.778, 0.789, 1.000],
    8: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.009, 0.011, 0.014, 0.016, 0.018, 0.018, 0.021, 0.023, 0.025, 0.027, 0.028, 0.030, 0.030, 0.032, 0.034, 0.037, 0.046, 0.055, 0.062, 0.064, 0.075, 0.080, 0.082, 0.085, 0.094, 0.101, 0.123, 0.131, 0.151, 0.169, 0.187, 0.206, 0.220, 0.236, 0.281, 0.311, 0.336, 0.366, 0.384, 0.410, 0.442, 0.469, 0.499, 0.544, 0.556, 0.572, 0.609, 0.629, 0.648, 0.663, 0.677, 0.696, 0.710, 0.728, 0.744, 0.757, 0.767, 0.783, 0.794, 1.000],
    9: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.009, 0.011, 0.012, 0.016, 0.018, 0.018, 0.018, 0.021, 0.025, 0.025, 0.027, 0.028, 0.032, 0.034, 0.034, 0.041, 0.046, 0.057, 0.060, 0.062, 0.066, 0.078, 0.080, 0.087, 0.094, 0.099, 0.110, 0.126, 0.140, 0.162, 0.176, 0.201, 0.211, 0.224, 0.247, 0.286, 0.320, 0.339, 0.368, 0.393, 0.426, 0.449, 0.487, 0.517, 0.549, 0.561, 0.577, 0.613, 0.632, 0.650, 0.664, 0.684, 0.698, 0.710, 0.728, 0.744, 0.757, 0.773, 0.787, 0.796, 1.000],
    10: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.009, 0.012, 0.014, 0.016, 0.018, 0.018, 0.018, 0.023, 0.025, 0.025, 0.027, 0.030, 0.036, 0.039, 0.043, 0.043, 0.052, 0.059, 0.060, 0.062, 0.064, 0.069, 0.083, 0.091, 0.094, 0.103, 0.114, 0.126, 0.133, 0.147, 0.169, 0.188, 0.206, 0.220, 0.238, 0.263, 0.298, 0.332, 0.353, 0.380, 0.421, 0.448, 0.469, 0.497, 0.531, 0.560, 0.584, 0.593, 0.623, 0.639, 0.661, 0.671, 0.687, 0.703, 0.712, 0.728, 0.751, 0.764, 0.778, 0.790, 0.796, 1.000],
    11: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.011, 0.012, 0.014, 0.016, 0.018, 0.018, 0.018, 0.023, 0.025, 0.025, 0.028, 0.036, 0.036, 0.041, 0.046, 0.046, 0.052, 0.059, 0.060, 0.062, 0.066, 0.080, 0.085, 0.091, 0.098, 0.112, 0.115, 0.128, 0.137, 0.151, 0.172, 0.188, 0.211, 0.224, 0.245, 0.275, 0.304, 0.332, 0.359, 0.398, 0.428, 0.456, 0.472, 0.503, 0.536, 0.561, 0.584, 0.600, 0.625, 0.650, 0.666, 0.679, 0.687, 0.703, 0.718, 0.732, 0.753, 0.766, 0.778, 0.790, 0.796, 1.000],
    12: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.011, 0.011, 0.011, 0.014, 0.014, 0.021, 0.021, 0.023, 0.023, 0.025, 0.025, 0.025, 0.027, 0.030, 0.036, 0.036, 0.039, 0.043, 0.046, 0.050, 0.055, 0.062, 0.062, 0.066, 0.080, 0.083, 0.092, 0.098, 0.108, 0.117, 0.128, 0.133, 0.146, 0.171, 0.181, 0.199, 0.218, 0.234, 0.265, 0.293, 0.325, 0.353, 0.373, 0.403, 0.432, 0.465, 0.496, 0.526, 0.549, 0.574, 0.606, 0.615, 0.641, 0.663, 0.666, 0.682, 0.694, 0.707, 0.719, 0.735, 0.758, 0.771, 0.780, 0.794, 0.801, 1.000],
    13: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.011, 0.011, 0.011, 0.014, 0.020, 0.021, 0.021, 0.023, 0.025, 0.025, 0.032, 0.034, 0.036, 0.036, 0.036, 0.036, 0.044, 0.046, 0.052, 0.052, 0.057, 0.064, 0.066, 0.071, 0.085, 0.089, 0.099, 0.108, 0.112, 0.121, 0.131, 0.139, 0.158, 0.178, 0.187, 0.208, 0.222, 0.243, 0.275, 0.298, 0.343, 0.366, 0.387, 0.421, 0.467, 0.481, 0.506, 0.535, 0.556, 0.590, 0.615, 0.629, 0.654, 0.664, 0.671, 0.687, 0.696, 0.709, 0.725, 0.741, 0.760, 0.780, 0.789, 0.796, 0.803, 1.000],
    14: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.011, 0.011, 0.012, 0.014, 0.020, 0.021, 0.021, 0.023, 0.025, 0.030, 0.034, 0.034, 0.036, 0.036, 0.036, 0.037, 0.044, 0.046, 0.052, 0.055, 0.059, 0.066, 0.071, 0.075, 0.085, 0.091, 0.101, 0.112, 0.112, 0.121, 0.133, 0.140, 0.163, 0.179, 0.187, 0.208, 0.226, 0.252, 0.288, 0.307, 0.348, 0.371, 0.394, 0.428, 0.467, 0.488, 0.512, 0.540, 0.563, 0.593, 0.620, 0.636, 0.654, 0.666, 0.675, 0.689, 0.700, 0.710, 0.730, 0.744, 0.762, 0.782, 0.790, 0.798, 0.805, 1.000],
    15: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.007, 0.012, 0.014, 0.020, 0.021, 0.021, 0.021, 0.023, 0.025, 0.030, 0.032, 0.034, 0.034, 0.036, 0.036, 0.039, 0.041, 0.048, 0.050, 0.059, 0.060, 0.060, 0.069, 0.073, 0.080, 0.094, 0.098, 0.108, 0.114, 0.114, 0.123, 0.139, 0.147, 0.172, 0.187, 0.195, 0.211, 0.236, 0.265, 0.293, 0.318, 0.366, 0.382, 0.407, 0.446, 0.483, 0.499, 0.528, 0.549, 0.574, 0.606, 0.622, 0.639, 0.659, 0.673, 0.679, 0.694, 0.703, 0.716, 0.734, 0.744, 0.767, 0.783, 0.792, 0.801, 0.810, 1.000],
    16: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.007, 0.012, 0.014, 0.014, 0.020, 0.021, 0.021, 0.021, 0.030, 0.030, 0.030, 0.032, 0.034, 0.034, 0.036, 0.036, 0.039, 0.041, 0.050, 0.052, 0.059, 0.062, 0.064, 0.071, 0.076, 0.082, 0.096, 0.098, 0.112, 0.114, 0.114, 0.131, 0.144, 0.153, 0.172, 0.187, 0.197, 0.215, 0.245, 0.268, 0.298, 0.321, 0.369, 0.385, 0.412, 0.451, 0.490, 0.503, 0.531, 0.552, 0.577, 0.609, 0.623, 0.647, 0.663, 0.675, 0.680, 0.698, 0.705, 0.718, 0.734, 0.746, 0.774, 0.785, 0.792, 0.805, 0.812, 1.000],
    17: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.012, 0.012, 0.018, 0.020, 0.021, 0.021, 0.028, 0.030, 0.030, 0.032, 0.034, 0.034, 0.036, 0.037, 0.039, 0.039, 0.043, 0.048, 0.057, 0.059, 0.066, 0.071, 0.071, 0.078, 0.082, 0.096, 0.099, 0.107, 0.115, 0.119, 0.121, 0.139, 0.151, 0.162, 0.183, 0.195, 0.206, 0.234, 0.256, 0.288, 0.304, 0.341, 0.377, 0.401, 0.426, 0.469, 0.496, 0.512, 0.540, 0.556, 0.584, 0.615, 0.634, 0.652, 0.668, 0.680, 0.682, 0.698, 0.710, 0.723, 0.734, 0.748, 0.778, 0.790, 0.806, 0.819, 0.822, 1.000],
    18: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.012, 0.012, 0.018, 0.020, 0.021, 0.021, 0.028, 0.030, 0.034, 0.036, 0.037, 0.037, 0.037, 0.039, 0.039, 0.039, 0.046, 0.057, 0.060, 0.066, 0.069, 0.071, 0.083, 0.092, 0.094, 0.096, 0.105, 0.107, 0.115, 0.119, 0.131, 0.144, 0.155, 0.165, 0.187, 0.197, 0.213, 0.242, 0.261, 0.290, 0.311, 0.346, 0.385, 0.409, 0.435, 0.478, 0.503, 0.513, 0.542, 0.561, 0.588, 0.618, 0.636, 0.654, 0.670, 0.680, 0.684, 0.703, 0.710, 0.723, 0.737, 0.751, 0.783, 0.801, 0.813, 0.819, 0.824, 1.000],
    19: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.012, 0.018, 0.020, 0.021, 0.021, 0.028, 0.028, 0.032, 0.036, 0.037, 0.037, 0.037, 0.039, 0.039, 0.046, 0.052, 0.052, 0.062, 0.066, 0.069, 0.069, 0.083, 0.087, 0.094, 0.098, 0.103, 0.108, 0.108, 0.119, 0.130, 0.140, 0.144, 0.162, 0.171, 0.188, 0.206, 0.226, 0.252, 0.274, 0.298, 0.330, 0.371, 0.403, 0.435, 0.460, 0.490, 0.508, 0.522, 0.551, 0.577, 0.600, 0.620, 0.641, 0.659, 0.675, 0.682, 0.691, 0.705, 0.710, 0.726, 0.741, 0.760, 0.798, 0.808, 0.815, 0.821, 0.826, 1.000],
    20: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.012, 0.018, 0.020, 0.021, 0.021, 0.028, 0.028, 0.034, 0.036, 0.037, 0.037, 0.039, 0.039, 0.043, 0.048, 0.053, 0.055, 0.066, 0.066, 0.069, 0.069, 0.083, 0.091, 0.096, 0.103, 0.107, 0.108, 0.117, 0.130, 0.131, 0.142, 0.153, 0.163, 0.174, 0.201, 0.210, 0.240, 0.252, 0.277, 0.311, 0.337, 0.378, 0.407, 0.439, 0.465, 0.490, 0.510, 0.524, 0.556, 0.579, 0.607, 0.625, 0.645, 0.659, 0.677, 0.684, 0.696, 0.705, 0.714, 0.726, 0.742, 0.762, 0.798, 0.808, 0.815, 0.822, 0.826, 1.000],
    21: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.007, 0.009, 0.014, 0.020, 0.020, 0.020, 0.021, 0.021, 0.030, 0.036, 0.036, 0.036, 0.037, 0.037, 0.039, 0.044, 0.046, 0.052, 0.059, 0.066, 0.069, 0.069, 0.071, 0.071, 0.085, 0.094, 0.099, 0.107, 0.108, 0.112, 0.123, 0.130, 0.139, 0.147, 0.160, 0.172, 0.187, 0.202, 0.215, 0.242, 0.259, 0.284, 0.314, 0.343, 0.394, 0.417, 0.455, 0.471, 0.494, 0.512, 0.528, 0.567, 0.586, 0.613, 0.629, 0.652, 0.668, 0.682, 0.686, 0.696, 0.705, 0.714, 0.730, 0.744, 0.766, 0.805, 0.812, 0.819, 0.824, 0.828, 1.000],
    22: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.007, 0.009, 0.014, 0.020, 0.020, 0.020, 0.021, 0.023, 0.032, 0.036, 0.036, 0.036, 0.037, 0.037, 0.039, 0.044, 0.052, 0.053, 0.062, 0.069, 0.069, 0.069, 0.071, 0.071, 0.085, 0.094, 0.101, 0.107, 0.108, 0.112, 0.123, 0.135, 0.139, 0.147, 0.165, 0.176, 0.188, 0.202, 0.217, 0.242, 0.259, 0.293, 0.318, 0.357, 0.401, 0.423, 0.458, 0.472, 0.494, 0.517, 0.529, 0.567, 0.590, 0.616, 0.636, 0.657, 0.670, 0.682, 0.686, 0.698, 0.707, 0.718, 0.732, 0.750, 0.769, 0.806, 0.817, 0.819, 0.824, 0.828, 1.000],
    23: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.005, 0.009, 0.011, 0.016, 0.020, 0.020, 0.021, 0.028, 0.028, 0.034, 0.036, 0.036, 0.037, 0.037, 0.037, 0.039, 0.048, 0.059, 0.059, 0.066, 0.069, 0.071, 0.073, 0.083, 0.083, 0.089, 0.098, 0.103, 0.107, 0.108, 0.112, 0.123, 0.135, 0.146, 0.155, 0.167, 0.181, 0.188, 0.206, 0.222, 0.247, 0.263, 0.302, 0.323, 0.357, 0.407, 0.437, 0.464, 0.472, 0.499, 0.520, 0.533, 0.575, 0.600, 0.629, 0.639, 0.661, 0.671, 0.682, 0.687, 0.700, 0.709, 0.726, 0.739, 0.751, 0.776, 0.812, 0.817, 0.819, 0.824, 0.828, 1.000],
    24: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.005, 0.007, 0.011, 0.012, 0.020, 0.020, 0.020, 0.023, 0.028, 0.030, 0.034, 0.036, 0.036, 0.037, 0.037, 0.039, 0.044, 0.055, 0.059, 0.060, 0.066, 0.069, 0.071, 0.073, 0.083, 0.083, 0.091, 0.098, 0.103, 0.107, 0.110, 0.114, 0.124, 0.140, 0.149, 0.158, 0.172, 0.181, 0.190, 0.206, 0.226, 0.247, 0.266, 0.304, 0.332, 0.361, 0.412, 0.439, 0.465, 0.488, 0.503, 0.520, 0.544, 0.579, 0.604, 0.631, 0.643, 0.663, 0.671, 0.682, 0.687, 0.700, 0.712, 0.728, 0.739, 0.755, 0.782, 0.813, 0.817, 0.819, 0.824, 0.828, 1.000],
    25: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.007, 0.007, 0.007, 0.007, 0.011, 0.016, 0.020, 0.020, 0.021, 0.023, 0.030, 0.034, 0.036, 0.036, 0.036, 0.037, 0.039, 0.043, 0.052, 0.057, 0.059, 0.060, 0.069, 0.069, 0.083, 0.083, 0.083, 0.087, 0.091, 0.099, 0.105, 0.108, 0.110, 0.121, 0.130, 0.142, 0.151, 0.162, 0.174, 0.183, 0.195, 0.210, 0.229, 0.249, 0.279, 0.313, 0.334, 0.373, 0.419, 0.446, 0.467, 0.490, 0.504, 0.524, 0.549, 0.584, 0.615, 0.638, 0.643, 0.663, 0.677, 0.682, 0.689, 0.700, 0.714, 0.734, 0.744, 0.760, 0.785, 0.813, 0.817, 0.819, 0.824, 0.828, 1.000],
    26: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.007, 0.007, 0.009, 0.012, 0.020, 0.020, 0.021, 0.023, 0.025, 0.032, 0.034, 0.036, 0.036, 0.037, 0.043, 0.048, 0.052, 0.053, 0.059, 0.060, 0.064, 0.069, 0.082, 0.083, 0.083, 0.087, 0.087, 0.091, 0.105, 0.105, 0.110, 0.114, 0.128, 0.137, 0.144, 0.153, 0.165, 0.181, 0.190, 0.197, 0.215, 0.233, 0.256, 0.295, 0.320, 0.345, 0.387, 0.428, 0.455, 0.485, 0.494, 0.506, 0.536, 0.556, 0.590, 0.620, 0.639, 0.648, 0.670, 0.680, 0.682, 0.693, 0.707, 0.718, 0.735, 0.744, 0.764, 0.785, 0.813, 0.817, 0.819, 0.824, 0.838, 1.000],
    27: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.007, 0.007, 0.009, 0.012, 0.020, 0.020, 0.021, 0.023, 0.025, 0.032, 0.034, 0.037, 0.037, 0.039, 0.043, 0.048, 0.052, 0.053, 0.060, 0.060, 0.064, 0.069, 0.082, 0.083, 0.083, 0.087, 0.087, 0.098, 0.105, 0.105, 0.110, 0.117, 0.130, 0.137, 0.147, 0.153, 0.165, 0.181, 0.192, 0.199, 0.215, 0.238, 0.263, 0.295, 0.320, 0.355, 0.393, 0.435, 0.456, 0.488, 0.497, 0.510, 0.540, 0.561, 0.590, 0.620, 0.639, 0.650, 0.675, 0.680, 0.682, 0.696, 0.707, 0.721, 0.737, 0.746, 0.766, 0.787, 0.815, 0.817, 0.819, 0.829, 0.838, 1.000],
    28: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.007, 0.007, 0.007, 0.011, 0.012, 0.020, 0.021, 0.021, 0.023, 0.030, 0.032, 0.034, 0.037, 0.037, 0.039, 0.043, 0.048, 0.052, 0.060, 0.064, 0.064, 0.066, 0.071, 0.082, 0.083, 0.085, 0.087, 0.092, 0.105, 0.107, 0.108, 0.114, 0.121, 0.131, 0.140, 0.147, 0.153, 0.167, 0.183, 0.194, 0.201, 0.220, 0.245, 0.275, 0.304, 0.329, 0.359, 0.409, 0.442, 0.464, 0.488, 0.504, 0.517, 0.551, 0.570, 0.595, 0.625, 0.650, 0.659, 0.677, 0.680, 0.684, 0.698, 0.710, 0.723, 0.737, 0.748, 0.773, 0.789, 0.815, 0.817, 0.822, 0.835, 0.840, 1.000],
    29: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.007, 0.009, 0.009, 0.014, 0.016, 0.020, 0.023, 0.023, 0.025, 0.032, 0.032, 0.036, 0.037, 0.037, 0.039, 0.046, 0.052, 0.052, 0.060, 0.064, 0.067, 0.069, 0.071, 0.083, 0.083, 0.085, 0.087, 0.096, 0.105, 0.107, 0.108, 0.115, 0.123, 0.131, 0.142, 0.149, 0.153, 0.167, 0.185, 0.195, 0.201, 0.222, 0.249, 0.279, 0.307, 0.330, 0.359, 0.409, 0.442, 0.464, 0.494, 0.504, 0.520, 0.552, 0.572, 0.599, 0.625, 0.652, 0.668, 0.677, 0.680, 0.687, 0.700, 0.712, 0.726, 0.741, 0.748, 0.774, 0.789, 0.815, 0.821, 0.824, 0.835, 0.840, 1.000],
    30: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.009, 0.009, 0.014, 0.014, 0.016, 0.021, 0.023, 0.025, 0.025, 0.032, 0.034, 0.037, 0.039, 0.046, 0.048, 0.052, 0.052, 0.057, 0.066, 0.069, 0.080, 0.080, 0.082, 0.085, 0.085, 0.089, 0.092, 0.098, 0.105, 0.107, 0.110, 0.117, 0.126, 0.133, 0.144, 0.153, 0.156, 0.169, 0.187, 0.195, 0.208, 0.226, 0.250, 0.281, 0.313, 0.337, 0.364, 0.412, 0.444, 0.464, 0.494, 0.504, 0.524, 0.560, 0.581, 0.609, 0.627, 0.657, 0.675, 0.677, 0.687, 0.689, 0.707, 0.712, 0.726, 0.742, 0.750, 0.774, 0.790, 0.817, 0.821, 0.828, 0.837, 0.844, 1.000],
    31: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.005, 0.005, 0.005, 0.005, 0.009, 0.009, 0.009, 0.014, 0.014, 0.020, 0.023, 0.023, 0.025, 0.030, 0.032, 0.036, 0.037, 0.039, 0.046, 0.048, 0.052, 0.052, 0.059, 0.066, 0.080, 0.080, 0.082, 0.083, 0.085, 0.085, 0.091, 0.094, 0.101, 0.107, 0.108, 0.112, 0.123, 0.130, 0.133, 0.146, 0.153, 0.156, 0.176, 0.187, 0.197, 0.215, 0.233, 0.250, 0.282, 0.314, 0.337, 0.369, 0.414, 0.444, 0.464, 0.496, 0.506, 0.524, 0.560, 0.581, 0.609, 0.627, 0.661, 0.675, 0.680, 0.689, 0.691, 0.707, 0.712, 0.726, 0.742, 0.751, 0.774, 0.792, 0.817, 0.822, 0.831, 0.838, 0.847, 1.000],
    32: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.005, 0.007, 0.007, 0.009, 0.011, 0.014, 0.016, 0.016, 0.020, 0.021, 0.025, 0.030, 0.030, 0.032, 0.032, 0.036, 0.037, 0.043, 0.046, 0.050, 0.052, 0.059, 0.060, 0.076, 0.080, 0.082, 0.082, 0.083, 0.087, 0.091, 0.096, 0.098, 0.103, 0.108, 0.108, 0.114, 0.126, 0.130, 0.139, 0.149, 0.156, 0.162, 0.176, 0.190, 0.202, 0.217, 0.240, 0.258, 0.288, 0.327, 0.350, 0.380, 0.423, 0.453, 0.483, 0.496, 0.515, 0.529, 0.560, 0.588, 0.611, 0.634, 0.664, 0.677, 0.680, 0.691, 0.691, 0.707, 0.712, 0.728, 0.742, 0.755, 0.778, 0.794, 0.817, 0.828, 0.831, 0.838, 0.847, 1.000],
    33: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.007, 0.009, 0.009, 0.011, 0.011, 0.016, 0.016, 0.020, 0.020, 0.021, 0.030, 0.030, 0.030, 0.032, 0.032, 0.036, 0.037, 0.043, 0.046, 0.050, 0.053, 0.059, 0.073, 0.076, 0.080, 0.082, 0.082, 0.083, 0.087, 0.092, 0.098, 0.098, 0.103, 0.108, 0.110, 0.117, 0.128, 0.131, 0.139, 0.149, 0.160, 0.165, 0.179, 0.190, 0.208, 0.218, 0.240, 0.263, 0.293, 0.330, 0.352, 0.387, 0.426, 0.453, 0.483, 0.499, 0.515, 0.529, 0.560, 0.588, 0.620, 0.636, 0.664, 0.679, 0.680, 0.691, 0.691, 0.709, 0.716, 0.728, 0.742, 0.757, 0.780, 0.794, 0.817, 0.828, 0.831, 0.840, 0.849, 1.000],
    34: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.009, 0.009, 0.009, 0.011, 0.011, 0.016, 0.021, 0.025, 0.027, 0.027, 0.030, 0.032, 0.032, 0.032, 0.034, 0.036, 0.037, 0.043, 0.046, 0.052, 0.057, 0.062, 0.078, 0.080, 0.082, 0.082, 0.082, 0.085, 0.091, 0.092, 0.098, 0.098, 0.103, 0.108, 0.112, 0.117, 0.130, 0.139, 0.146, 0.151, 0.160, 0.167, 0.183, 0.195, 0.210, 0.220, 0.243, 0.281, 0.304, 0.339, 0.364, 0.407, 0.432, 0.467, 0.487, 0.501, 0.517, 0.538, 0.568, 0.591, 0.627, 0.641, 0.670, 0.679, 0.680, 0.691, 0.696, 0.712, 0.719, 0.732, 0.748, 0.758, 0.782, 0.798, 0.821, 0.828, 0.835, 0.842, 0.854, 1.000],
    35: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.009, 0.009, 0.009, 0.011, 0.011, 0.021, 0.021, 0.025, 0.027, 0.027, 0.030, 0.032, 0.032, 0.032, 0.034, 0.036, 0.037, 0.043, 0.048, 0.052, 0.062, 0.064, 0.078, 0.080, 0.082, 0.082, 0.082, 0.085, 0.092, 0.092, 0.098, 0.098, 0.105, 0.110, 0.112, 0.121, 0.131, 0.142, 0.147, 0.151, 0.160, 0.169, 0.183, 0.201, 0.211, 0.227, 0.250, 0.290, 0.313, 0.346, 0.375, 0.412, 0.444, 0.469, 0.488, 0.501, 0.519, 0.540, 0.570, 0.604, 0.631, 0.648, 0.670, 0.679, 0.680, 0.693, 0.700, 0.712, 0.723, 0.732, 0.750, 0.762, 0.782, 0.799, 0.821, 0.828, 0.837, 0.842, 0.854, 1.000],
    36: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.005, 0.009, 0.009, 0.009, 0.011, 0.020, 0.023, 0.023, 0.025, 0.027, 0.027, 0.032, 0.032, 0.034, 0.034, 0.036, 0.036, 0.037, 0.043, 0.052, 0.057, 0.062, 0.067, 0.080, 0.080, 0.082, 0.083, 0.087, 0.092, 0.092, 0.092, 0.098, 0.098, 0.105, 0.110, 0.114, 0.128, 0.140, 0.144, 0.147, 0.151, 0.165, 0.169, 0.185, 0.204, 0.211, 0.231, 0.261, 0.295, 0.318, 0.355, 0.382, 0.414, 0.446, 0.471, 0.492, 0.508, 0.520, 0.547, 0.581, 0.615, 0.636, 0.652, 0.671, 0.679, 0.680, 0.694, 0.702, 0.716, 0.725, 0.734, 0.751, 0.762, 0.782, 0.805, 0.822, 0.829, 0.837, 0.842, 0.854, 1.000],
    37: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.005, 0.009, 0.009, 0.009, 0.011, 0.020, 0.023, 0.023, 0.025, 0.027, 0.027, 0.032, 0.034, 0.034, 0.036, 0.036, 0.036, 0.037, 0.044, 0.053, 0.057, 0.064, 0.067, 0.080, 0.080, 0.082, 0.085, 0.087, 0.092, 0.092, 0.094, 0.098, 0.101, 0.107, 0.110, 0.119, 0.128, 0.142, 0.144, 0.147, 0.151, 0.165, 0.171, 0.187, 0.204, 0.213, 0.233, 0.265, 0.298, 0.320, 0.357, 0.382, 0.417, 0.448, 0.472, 0.499, 0.512, 0.520, 0.547, 0.584, 0.618, 0.638, 0.655, 0.673, 0.679, 0.680, 0.694, 0.702, 0.716, 0.725, 0.734, 0.753, 0.764, 0.782, 0.805, 0.824, 0.831, 0.838, 0.845, 0.854, 1.000],
    38: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.007, 0.009, 0.009, 0.009, 0.011, 0.020, 0.023, 0.025, 0.025, 0.027, 0.027, 0.034, 0.034, 0.034, 0.036, 0.036, 0.036, 0.044, 0.050, 0.055, 0.060, 0.067, 0.067, 0.080, 0.080, 0.083, 0.085, 0.089, 0.092, 0.094, 0.094, 0.101, 0.103, 0.107, 0.115, 0.121, 0.130, 0.142, 0.144, 0.147, 0.151, 0.165, 0.172, 0.188, 0.204, 0.215, 0.238, 0.272, 0.304, 0.330, 0.369, 0.391, 0.430, 0.449, 0.474, 0.503, 0.517, 0.528, 0.554, 0.590, 0.622, 0.641, 0.657, 0.673, 0.682, 0.684, 0.694, 0.702, 0.716, 0.726, 0.734, 0.753, 0.766, 0.785, 0.808, 0.824, 0.833, 0.840, 0.845, 0.856, 1.000],
    39: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.009, 0.009, 0.009, 0.011, 0.020, 0.023, 0.025, 0.025, 0.028, 0.030, 0.034, 0.034, 0.034, 0.036, 0.039, 0.044, 0.048, 0.050, 0.057, 0.064, 0.067, 0.067, 0.080, 0.087, 0.087, 0.087, 0.089, 0.092, 0.094, 0.099, 0.101, 0.103, 0.108, 0.119, 0.123, 0.135, 0.142, 0.144, 0.149, 0.155, 0.174, 0.181, 0.195, 0.208, 0.222, 0.247, 0.279, 0.309, 0.336, 0.375, 0.401, 0.435, 0.453, 0.483, 0.508, 0.524, 0.535, 0.558, 0.593, 0.623, 0.647, 0.663, 0.677, 0.682, 0.686, 0.696, 0.703, 0.718, 0.726, 0.734, 0.757, 0.769, 0.789, 0.810, 0.824, 0.835, 0.844, 0.849, 0.858, 1.000],
    40: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.009, 0.009, 0.009, 0.011, 0.014, 0.021, 0.023, 0.025, 0.027, 0.028, 0.030, 0.034, 0.034, 0.034, 0.036, 0.039, 0.044, 0.048, 0.052, 0.060, 0.066, 0.069, 0.069, 0.080, 0.087, 0.087, 0.089, 0.089, 0.092, 0.096, 0.101, 0.101, 0.105, 0.112, 0.123, 0.123, 0.139, 0.144, 0.146, 0.149, 0.160, 0.176, 0.187, 0.197, 0.208, 0.224, 0.249, 0.291, 0.311, 0.337, 0.375, 0.407, 0.437, 0.455, 0.487, 0.512, 0.524, 0.538, 0.565, 0.597, 0.629, 0.652, 0.666, 0.680, 0.682, 0.686, 0.696, 0.703, 0.721, 0.726, 0.734, 0.757, 0.771, 0.789, 0.813, 0.824, 0.835, 0.844, 0.851, 0.858, 1.000],
    41: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.009, 0.009, 0.009, 0.011, 0.014, 0.021, 0.025, 0.027, 0.028, 0.028, 0.034, 0.034, 0.034, 0.034, 0.036, 0.039, 0.046, 0.052, 0.059, 0.062, 0.069, 0.069, 0.082, 0.083, 0.087, 0.087, 0.089, 0.089, 0.098, 0.101, 0.103, 0.103, 0.108, 0.112, 0.126, 0.131, 0.139, 0.144, 0.146, 0.151, 0.167, 0.185, 0.195, 0.201, 0.217, 0.233, 0.254, 0.300, 0.321, 0.345, 0.382, 0.414, 0.440, 0.462, 0.488, 0.513, 0.526, 0.545, 0.572, 0.599, 0.632, 0.657, 0.668, 0.682, 0.686, 0.687, 0.700, 0.707, 0.721, 0.726, 0.734, 0.758, 0.771, 0.789, 0.813, 0.826, 0.835, 0.845, 0.853, 0.860, 1.000],
    42: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.009, 0.009, 0.009, 0.011, 0.014, 0.021, 0.027, 0.028, 0.030, 0.030, 0.034, 0.034, 0.034, 0.036, 0.039, 0.039, 0.046, 0.052, 0.059, 0.062, 0.071, 0.082, 0.083, 0.085, 0.087, 0.087, 0.089, 0.091, 0.099, 0.103, 0.103, 0.103, 0.110, 0.112, 0.126, 0.133, 0.139, 0.144, 0.146, 0.151, 0.171, 0.187, 0.195, 0.201, 0.222, 0.234, 0.256, 0.306, 0.325, 0.357, 0.384, 0.419, 0.446, 0.467, 0.488, 0.515, 0.526, 0.547, 0.575, 0.604, 0.638, 0.659, 0.668, 0.682, 0.687, 0.687, 0.703, 0.707, 0.721, 0.726, 0.734, 0.758, 0.771, 0.790, 0.815, 0.826, 0.835, 0.845, 0.853, 0.860, 1.000],
    43: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.009, 0.009, 0.009, 0.011, 0.014, 0.023, 0.027, 0.030, 0.030, 0.030, 0.034, 0.034, 0.034, 0.037, 0.041, 0.048, 0.052, 0.059, 0.059, 0.062, 0.071, 0.082, 0.083, 0.085, 0.089, 0.089, 0.089, 0.092, 0.099, 0.103, 0.103, 0.105, 0.110, 0.117, 0.126, 0.133, 0.139, 0.144, 0.151, 0.155, 0.174, 0.187, 0.195, 0.201, 0.224, 0.240, 0.268, 0.307, 0.332, 0.361, 0.389, 0.423, 0.449, 0.471, 0.497, 0.520, 0.533, 0.551, 0.581, 0.607, 0.647, 0.663, 0.673, 0.684, 0.689, 0.691, 0.703, 0.707, 0.721, 0.726, 0.735, 0.762, 0.774, 0.790, 0.817, 0.829, 0.840, 0.847, 0.856, 0.863, 1.000],
    44: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.009, 0.009, 0.009, 0.011, 0.020, 0.025, 0.027, 0.030, 0.030, 0.030, 0.034, 0.034, 0.034, 0.037, 0.041, 0.048, 0.053, 0.059, 0.059, 0.066, 0.071, 0.083, 0.083, 0.087, 0.089, 0.089, 0.089, 0.094, 0.103, 0.103, 0.103, 0.107, 0.110, 0.117, 0.126, 0.135, 0.140, 0.147, 0.153, 0.158, 0.174, 0.187, 0.195, 0.202, 0.224, 0.243, 0.274, 0.307, 0.337, 0.364, 0.393, 0.425, 0.451, 0.478, 0.501, 0.520, 0.538, 0.551, 0.584, 0.611, 0.648, 0.663, 0.673, 0.684, 0.691, 0.691, 0.703, 0.709, 0.721, 0.726, 0.735, 0.762, 0.774, 0.799, 0.819, 0.829, 0.840, 0.851, 0.856, 0.863, 1.000],
    45: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.005, 0.009, 0.009, 0.011, 0.011, 0.021, 0.025, 0.028, 0.030, 0.030, 0.030, 0.034, 0.034, 0.037, 0.043, 0.048, 0.053, 0.057, 0.064, 0.066, 0.066, 0.071, 0.083, 0.083, 0.089, 0.089, 0.089, 0.091, 0.096, 0.103, 0.103, 0.103, 0.108, 0.112, 0.119, 0.126, 0.135, 0.140, 0.149, 0.155, 0.165, 0.178, 0.187, 0.199, 0.208, 0.229, 0.250, 0.281, 0.316, 0.345, 0.371, 0.396, 0.437, 0.462, 0.481, 0.506, 0.531, 0.545, 0.560, 0.593, 0.613, 0.650, 0.664, 0.675, 0.684, 0.691, 0.696, 0.703, 0.710, 0.723, 0.728, 0.737, 0.764, 0.778, 0.801, 0.819, 0.831, 0.842, 0.851, 0.858, 0.865, 1.000],
    46: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.005, 0.009, 0.011, 0.011, 0.012, 0.021, 0.027, 0.028, 0.030, 0.030, 0.032, 0.036, 0.036, 0.043, 0.044, 0.053, 0.053, 0.059, 0.064, 0.066, 0.066, 0.071, 0.083, 0.083, 0.089, 0.089, 0.089, 0.091, 0.096, 0.103, 0.103, 0.103, 0.108, 0.112, 0.119, 0.126, 0.135, 0.142, 0.149, 0.158, 0.165, 0.179, 0.187, 0.202, 0.210, 0.231, 0.252, 0.281, 0.323, 0.348, 0.371, 0.401, 0.444, 0.467, 0.481, 0.510, 0.531, 0.549, 0.565, 0.593, 0.613, 0.650, 0.664, 0.675, 0.684, 0.691, 0.696, 0.705, 0.712, 0.725, 0.730, 0.742, 0.766, 0.778, 0.805, 0.821, 0.831, 0.842, 0.851, 0.858, 0.865, 1.000],
    47: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.005, 0.009, 0.011, 0.011, 0.012, 0.021, 0.027, 0.030, 0.032, 0.032, 0.032, 0.041, 0.041, 0.044, 0.046, 0.059, 0.060, 0.062, 0.066, 0.066, 0.066, 0.073, 0.083, 0.083, 0.089, 0.089, 0.089, 0.092, 0.099, 0.103, 0.103, 0.105, 0.108, 0.114, 0.119, 0.131, 0.139, 0.144, 0.153, 0.158, 0.169, 0.181, 0.194, 0.204, 0.210, 0.236, 0.258, 0.297, 0.332, 0.355, 0.380, 0.407, 0.446, 0.469, 0.481, 0.515, 0.536, 0.549, 0.572, 0.593, 0.620, 0.650, 0.671, 0.677, 0.684, 0.693, 0.698, 0.707, 0.716, 0.728, 0.735, 0.746, 0.773, 0.787, 0.806, 0.821, 0.831, 0.845, 0.853, 0.858, 0.865, 1.000],
    48: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.005, 0.009, 0.011, 0.011, 0.012, 0.021, 0.028, 0.030, 0.032, 0.032, 0.032, 0.041, 0.041, 0.044, 0.048, 0.059, 0.060, 0.062, 0.066, 0.066, 0.066, 0.073, 0.083, 0.083, 0.089, 0.089, 0.089, 0.094, 0.099, 0.103, 0.103, 0.107, 0.108, 0.114, 0.119, 0.131, 0.139, 0.147, 0.153, 0.158, 0.171, 0.185, 0.195, 0.204, 0.210, 0.238, 0.265, 0.298, 0.332, 0.361, 0.384, 0.409, 0.446, 0.469, 0.485, 0.517, 0.538, 0.551, 0.572, 0.593, 0.622, 0.650, 0.671, 0.677, 0.684, 0.693, 0.698, 0.709, 0.718, 0.730, 0.735, 0.746, 0.773, 0.789, 0.806, 0.822, 0.831, 0.847, 0.853, 0.860, 0.865, 1.000],
    49: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.011, 0.012, 0.023, 0.028, 0.030, 0.032, 0.032, 0.032, 0.041, 0.043, 0.046, 0.048, 0.059, 0.060, 0.064, 0.066, 0.067, 0.069, 0.073, 0.083, 0.083, 0.089, 0.089, 0.091, 0.098, 0.101, 0.105, 0.105, 0.107, 0.108, 0.114, 0.124, 0.131, 0.139, 0.151, 0.158, 0.165, 0.176, 0.194, 0.197, 0.208, 0.218, 0.245, 0.272, 0.304, 0.339, 0.366, 0.385, 0.412, 0.451, 0.474, 0.492, 0.522, 0.538, 0.551, 0.577, 0.597, 0.625, 0.650, 0.673, 0.679, 0.686, 0.694, 0.698, 0.712, 0.718, 0.730, 0.739, 0.746, 0.773, 0.789, 0.806, 0.822, 0.833, 0.849, 0.853, 0.860, 0.865, 1.000],
    50: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.011, 0.012, 0.025, 0.028, 0.030, 0.032, 0.032, 0.032, 0.043, 0.043, 0.046, 0.048, 0.059, 0.060, 0.064, 0.066, 0.067, 0.069, 0.073, 0.083, 0.083, 0.089, 0.092, 0.094, 0.099, 0.101, 0.105, 0.107, 0.107, 0.108, 0.119, 0.124, 0.133, 0.142, 0.151, 0.158, 0.167, 0.179, 0.194, 0.197, 0.208, 0.218, 0.250, 0.279, 0.306, 0.339, 0.368, 0.389, 0.414, 0.458, 0.478, 0.494, 0.524, 0.540, 0.558, 0.577, 0.602, 0.629, 0.659, 0.673, 0.679, 0.689, 0.694, 0.700, 0.712, 0.719, 0.732, 0.739, 0.746, 0.773, 0.789, 0.806, 0.824, 0.833, 0.849, 0.854, 0.860, 0.865, 1.000],
    51: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.011, 0.020, 0.027, 0.028, 0.030, 0.032, 0.032, 0.032, 0.046, 0.046, 0.050, 0.052, 0.059, 0.064, 0.064, 0.067, 0.069, 0.069, 0.073, 0.083, 0.083, 0.089, 0.092, 0.096, 0.101, 0.101, 0.105, 0.107, 0.107, 0.110, 0.121, 0.124, 0.139, 0.146, 0.151, 0.160, 0.172, 0.179, 0.194, 0.199, 0.210, 0.222, 0.259, 0.282, 0.309, 0.339, 0.373, 0.391, 0.423, 0.462, 0.481, 0.496, 0.526, 0.545, 0.558, 0.581, 0.607, 0.639, 0.663, 0.677, 0.680, 0.691, 0.696, 0.702, 0.712, 0.719, 0.734, 0.741, 0.748, 0.773, 0.789, 0.806, 0.824, 0.833, 0.849, 0.854, 0.860, 0.865, 1.000],
    52: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.007, 0.009, 0.011, 0.014, 0.021, 0.027, 0.028, 0.030, 0.036, 0.036, 0.037, 0.046, 0.046, 0.050, 0.055, 0.059, 0.064, 0.064, 0.069, 0.071, 0.071, 0.073, 0.083, 0.083, 0.089, 0.092, 0.098, 0.101, 0.101, 0.107, 0.107, 0.107, 0.110, 0.121, 0.126, 0.139, 0.146, 0.151, 0.163, 0.176, 0.183, 0.195, 0.199, 0.210, 0.227, 0.259, 0.286, 0.311, 0.339, 0.375, 0.393, 0.433, 0.469, 0.485, 0.508, 0.526, 0.547, 0.558, 0.584, 0.613, 0.648, 0.668, 0.679, 0.680, 0.691, 0.696, 0.705, 0.712, 0.721, 0.734, 0.741, 0.748, 0.774, 0.790, 0.806, 0.824, 0.837, 0.849, 0.854, 0.860, 0.865, 1.000],
    53: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.011, 0.014, 0.021, 0.028, 0.028, 0.030, 0.036, 0.036, 0.037, 0.046, 0.048, 0.050, 0.055, 0.059, 0.064, 0.064, 0.069, 0.071, 0.073, 0.073, 0.083, 0.083, 0.089, 0.092, 0.098, 0.101, 0.101, 0.107, 0.107, 0.107, 0.110, 0.121, 0.126, 0.142, 0.147, 0.155, 0.165, 0.176, 0.187, 0.197, 0.199, 0.210, 0.227, 0.263, 0.286, 0.314, 0.341, 0.380, 0.403, 0.437, 0.469, 0.485, 0.512, 0.531, 0.551, 0.560, 0.591, 0.615, 0.652, 0.668, 0.679, 0.680, 0.691, 0.698, 0.707, 0.714, 0.721, 0.734, 0.742, 0.748, 0.778, 0.792, 0.806, 0.824, 0.837, 0.849, 0.854, 0.861, 0.865, 1.000],
    54: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.011, 0.020, 0.025, 0.030, 0.030, 0.032, 0.036, 0.036, 0.039, 0.048, 0.050, 0.052, 0.055, 0.059, 0.066, 0.066, 0.071, 0.071, 0.073, 0.073, 0.083, 0.085, 0.089, 0.096, 0.101, 0.101, 0.103, 0.107, 0.107, 0.108, 0.114, 0.121, 0.126, 0.142, 0.149, 0.156, 0.167, 0.176, 0.187, 0.197, 0.199, 0.211, 0.233, 0.268, 0.288, 0.316, 0.345, 0.380, 0.405, 0.444, 0.469, 0.485, 0.515, 0.533, 0.552, 0.561, 0.597, 0.618, 0.652, 0.668, 0.679, 0.682, 0.694, 0.700, 0.707, 0.714, 0.721, 0.734, 0.744, 0.751, 0.780, 0.794, 0.806, 0.824, 0.837, 0.849, 0.854, 0.861, 0.867, 1.000],
    55: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.011, 0.020, 0.025, 0.030, 0.030, 0.036, 0.036, 0.036, 0.039, 0.048, 0.050, 0.052, 0.055, 0.062, 0.066, 0.066, 0.071, 0.071, 0.073, 0.083, 0.083, 0.085, 0.089, 0.096, 0.101, 0.101, 0.103, 0.107, 0.107, 0.110, 0.115, 0.121, 0.130, 0.144, 0.151, 0.156, 0.167, 0.176, 0.187, 0.199, 0.201, 0.213, 0.233, 0.268, 0.288, 0.316, 0.346, 0.380, 0.410, 0.444, 0.469, 0.488, 0.515, 0.535, 0.552, 0.561, 0.600, 0.622, 0.654, 0.668, 0.679, 0.682, 0.694, 0.702, 0.707, 0.714, 0.721, 0.735, 0.746, 0.755, 0.782, 0.794, 0.806, 0.824, 0.837, 0.849, 0.856, 0.863, 0.867, 1.000],
    56: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.011, 0.020, 0.025, 0.032, 0.032, 0.036, 0.036, 0.037, 0.039, 0.048, 0.050, 0.052, 0.055, 0.062, 0.066, 0.066, 0.071, 0.071, 0.083, 0.083, 0.085, 0.087, 0.092, 0.098, 0.101, 0.101, 0.103, 0.107, 0.107, 0.112, 0.115, 0.123, 0.133, 0.146, 0.153, 0.158, 0.167, 0.176, 0.187, 0.201, 0.202, 0.217, 0.238, 0.268, 0.293, 0.318, 0.350, 0.380, 0.416, 0.446, 0.472, 0.501, 0.520, 0.535, 0.554, 0.565, 0.600, 0.622, 0.654, 0.668, 0.679, 0.684, 0.696, 0.703, 0.710, 0.714, 0.723, 0.737, 0.746, 0.755, 0.783, 0.794, 0.808, 0.826, 0.837, 0.851, 0.856, 0.863, 0.867, 1.000],
    57: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.011, 0.020, 0.025, 0.032, 0.032, 0.036, 0.036, 0.037, 0.039, 0.048, 0.050, 0.052, 0.055, 0.062, 0.066, 0.067, 0.071, 0.071, 0.083, 0.083, 0.085, 0.087, 0.094, 0.098, 0.101, 0.101, 0.103, 0.107, 0.107, 0.112, 0.115, 0.126, 0.133, 0.147, 0.153, 0.158, 0.167, 0.176, 0.188, 0.201, 0.204, 0.220, 0.238, 0.268, 0.293, 0.321, 0.352, 0.382, 0.416, 0.448, 0.478, 0.503, 0.520, 0.536, 0.554, 0.565, 0.602, 0.623, 0.654, 0.668, 0.679, 0.686, 0.696, 0.703, 0.710, 0.716, 0.723, 0.739, 0.746, 0.757, 0.783, 0.794, 0.808, 0.826, 0.837, 0.851, 0.856, 0.863, 0.867, 1.000],
    58: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.018, 0.020, 0.027, 0.032, 0.032, 0.036, 0.036, 0.043, 0.043, 0.048, 0.052, 0.052, 0.057, 0.062, 0.067, 0.069, 0.071, 0.071, 0.083, 0.085, 0.087, 0.089, 0.096, 0.098, 0.103, 0.103, 0.107, 0.108, 0.110, 0.115, 0.117, 0.126, 0.133, 0.147, 0.155, 0.160, 0.172, 0.179, 0.192, 0.202, 0.206, 0.220, 0.245, 0.274, 0.302, 0.325, 0.352, 0.384, 0.416, 0.448, 0.478, 0.504, 0.522, 0.538, 0.556, 0.568, 0.604, 0.629, 0.654, 0.668, 0.680, 0.687, 0.698, 0.709, 0.712, 0.716, 0.728, 0.742, 0.748, 0.757, 0.787, 0.794, 0.808, 0.826, 0.837, 0.851, 0.856, 0.863, 0.867, 1.000],
    59: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.018, 0.021, 0.027, 0.032, 0.032, 0.036, 0.037, 0.043, 0.043, 0.048, 0.052, 0.052, 0.057, 0.062, 0.069, 0.069, 0.071, 0.071, 0.083, 0.085, 0.087, 0.089, 0.098, 0.099, 0.103, 0.105, 0.108, 0.110, 0.110, 0.115, 0.121, 0.130, 0.133, 0.147, 0.155, 0.160, 0.172, 0.179, 0.192, 0.202, 0.206, 0.220, 0.249, 0.274, 0.302, 0.325, 0.352, 0.384, 0.416, 0.448, 0.478, 0.504, 0.522, 0.538, 0.556, 0.570, 0.604, 0.629, 0.654, 0.670, 0.680, 0.687, 0.698, 0.709, 0.712, 0.716, 0.730, 0.742, 0.748, 0.757, 0.789, 0.794, 0.808, 0.826, 0.837, 0.851, 0.856, 0.863, 0.867, 1.000],
    60: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.018, 0.021, 0.028, 0.032, 0.032, 0.036, 0.039, 0.043, 0.043, 0.048, 0.052, 0.052, 0.059, 0.062, 0.069, 0.069, 0.071, 0.073, 0.085, 0.087, 0.092, 0.098, 0.099, 0.103, 0.105, 0.108, 0.110, 0.114, 0.115, 0.121, 0.124, 0.130, 0.135, 0.147, 0.155, 0.163, 0.176, 0.185, 0.194, 0.210, 0.211, 0.222, 0.250, 0.282, 0.304, 0.330, 0.353, 0.384, 0.419, 0.456, 0.480, 0.506, 0.522, 0.542, 0.561, 0.575, 0.606, 0.632, 0.655, 0.673, 0.682, 0.689, 0.703, 0.709, 0.714, 0.721, 0.730, 0.744, 0.750, 0.757, 0.789, 0.796, 0.808, 0.828, 0.838, 0.851, 0.856, 0.865, 0.869, 1.000],
    61: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.005, 0.009, 0.011, 0.018, 0.023, 0.028, 0.032, 0.036, 0.036, 0.039, 0.043, 0.043, 0.048, 0.052, 0.052, 0.059, 0.062, 0.069, 0.069, 0.073, 0.073, 0.087, 0.087, 0.094, 0.098, 0.101, 0.105, 0.108, 0.108, 0.114, 0.114, 0.115, 0.123, 0.124, 0.130, 0.135, 0.147, 0.160, 0.169, 0.176, 0.185, 0.194, 0.210, 0.211, 0.226, 0.252, 0.284, 0.306, 0.332, 0.355, 0.385, 0.419, 0.456, 0.480, 0.508, 0.526, 0.544, 0.561, 0.575, 0.606, 0.634, 0.655, 0.675, 0.682, 0.689, 0.703, 0.709, 0.714, 0.721, 0.732, 0.744, 0.750, 0.757, 0.790, 0.796, 0.808, 0.828, 0.838, 0.853, 0.858, 0.865, 0.870, 1.000],
    62: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.005, 0.005, 0.009, 0.011, 0.018, 0.023, 0.028, 0.032, 0.036, 0.036, 0.039, 0.043, 0.043, 0.048, 0.052, 0.053, 0.059, 0.062, 0.069, 0.071, 0.073, 0.073, 0.087, 0.089, 0.094, 0.099, 0.103, 0.105, 0.108, 0.108, 0.114, 0.114, 0.119, 0.124, 0.124, 0.130, 0.139, 0.151, 0.160, 0.171, 0.176, 0.185, 0.194, 0.210, 0.215, 0.229, 0.258, 0.284, 0.311, 0.332, 0.357, 0.387, 0.433, 0.458, 0.481, 0.508, 0.526, 0.547, 0.561, 0.581, 0.611, 0.634, 0.657, 0.677, 0.682, 0.689, 0.703, 0.709, 0.716, 0.723, 0.732, 0.746, 0.751, 0.760, 0.790, 0.796, 0.810, 0.831, 0.840, 0.853, 0.860, 0.869, 0.874, 1.000],
    63: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.005, 0.007, 0.009, 0.012, 0.018, 0.023, 0.028, 0.032, 0.036, 0.036, 0.039, 0.043, 0.043, 0.048, 0.052, 0.053, 0.059, 0.062, 0.069, 0.071, 0.073, 0.075, 0.087, 0.089, 0.096, 0.101, 0.103, 0.105, 0.108, 0.108, 0.114, 0.114, 0.121, 0.124, 0.128, 0.137, 0.144, 0.153, 0.162, 0.172, 0.178, 0.187, 0.195, 0.211, 0.217, 0.238, 0.261, 0.288, 0.313, 0.336, 0.359, 0.396, 0.435, 0.458, 0.481, 0.510, 0.526, 0.549, 0.565, 0.590, 0.620, 0.636, 0.657, 0.677, 0.684, 0.689, 0.705, 0.710, 0.716, 0.723, 0.732, 0.746, 0.751, 0.760, 0.790, 0.799, 0.812, 0.833, 0.844, 0.854, 0.861, 0.870, 0.874, 1.000],
    64: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.005, 0.007, 0.009, 0.018, 0.018, 0.023, 0.028, 0.032, 0.036, 0.036, 0.043, 0.043, 0.043, 0.048, 0.052, 0.053, 0.059, 0.064, 0.069, 0.071, 0.073, 0.075, 0.087, 0.089, 0.096, 0.101, 0.103, 0.105, 0.110, 0.110, 0.115, 0.115, 0.123, 0.124, 0.130, 0.139, 0.144, 0.153, 0.162, 0.172, 0.178, 0.187, 0.197, 0.213, 0.217, 0.242, 0.263, 0.291, 0.313, 0.336, 0.364, 0.401, 0.439, 0.462, 0.485, 0.512, 0.526, 0.549, 0.565, 0.590, 0.620, 0.636, 0.661, 0.679, 0.684, 0.689, 0.705, 0.710, 0.716, 0.723, 0.734, 0.746, 0.751, 0.760, 0.792, 0.799, 0.812, 0.835, 0.845, 0.856, 0.863, 0.870, 0.874, 1.000],
    65: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.005, 0.007, 0.009, 0.018, 0.018, 0.023, 0.028, 0.032, 0.036, 0.036, 0.043, 0.043, 0.043, 0.048, 0.052, 0.053, 0.059, 0.067, 0.071, 0.071, 0.073, 0.075, 0.087, 0.089, 0.098, 0.103, 0.105, 0.107, 0.110, 0.114, 0.117, 0.117, 0.123, 0.124, 0.130, 0.139, 0.144, 0.155, 0.163, 0.174, 0.179, 0.190, 0.199, 0.213, 0.226, 0.245, 0.265, 0.291, 0.313, 0.339, 0.368, 0.405, 0.444, 0.465, 0.488, 0.513, 0.533, 0.554, 0.574, 0.597, 0.627, 0.639, 0.663, 0.682, 0.687, 0.689, 0.705, 0.712, 0.718, 0.723, 0.734, 0.746, 0.751, 0.760, 0.792, 0.799, 0.813, 0.835, 0.845, 0.860, 0.865, 0.872, 0.876, 1.000],
    66: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.005, 0.005, 0.005, 0.007, 0.009, 0.018, 0.018, 0.023, 0.028, 0.032, 0.036, 0.036, 0.043, 0.043, 0.046, 0.050, 0.052, 0.053, 0.064, 0.067, 0.071, 0.073, 0.075, 0.085, 0.087, 0.092, 0.098, 0.103, 0.107, 0.107, 0.110, 0.117, 0.117, 0.119, 0.123, 0.124, 0.130, 0.142, 0.144, 0.158, 0.169, 0.176, 0.179, 0.190, 0.199, 0.215, 0.229, 0.247, 0.268, 0.300, 0.313, 0.345, 0.371, 0.405, 0.444, 0.467, 0.490, 0.515, 0.533, 0.556, 0.579, 0.600, 0.627, 0.639, 0.663, 0.682, 0.687, 0.689, 0.705, 0.712, 0.718, 0.723, 0.734, 0.746, 0.751, 0.760, 0.792, 0.803, 0.813, 0.837, 0.845, 0.860, 0.865, 0.872, 0.876, 1.000],
    67: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.005, 0.007, 0.009, 0.018, 0.020, 0.023, 0.028, 0.032, 0.036, 0.037, 0.043, 0.043, 0.048, 0.052, 0.052, 0.057, 0.066, 0.067, 0.071, 0.073, 0.075, 0.089, 0.091, 0.092, 0.101, 0.103, 0.107, 0.107, 0.110, 0.117, 0.117, 0.121, 0.123, 0.126, 0.131, 0.144, 0.144, 0.162, 0.171, 0.176, 0.181, 0.192, 0.202, 0.215, 0.238, 0.252, 0.274, 0.302, 0.316, 0.346, 0.371, 0.407, 0.449, 0.474, 0.490, 0.515, 0.535, 0.556, 0.583, 0.604, 0.631, 0.641, 0.664, 0.684, 0.687, 0.691, 0.705, 0.714, 0.718, 0.723, 0.734, 0.746, 0.753, 0.764, 0.794, 0.803, 0.813, 0.837, 0.847, 0.860, 0.865, 0.874, 0.879, 1.000],
    68: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.007, 0.007, 0.009, 0.020, 0.020, 0.023, 0.028, 0.032, 0.037, 0.037, 0.043, 0.043, 0.048, 0.052, 0.052, 0.059, 0.066, 0.067, 0.071, 0.075, 0.076, 0.091, 0.091, 0.092, 0.101, 0.105, 0.107, 0.107, 0.114, 0.117, 0.119, 0.121, 0.123, 0.131, 0.131, 0.144, 0.146, 0.162, 0.172, 0.178, 0.181, 0.192, 0.202, 0.218, 0.238, 0.252, 0.274, 0.302, 0.316, 0.346, 0.371, 0.409, 0.449, 0.474, 0.503, 0.515, 0.535, 0.558, 0.584, 0.604, 0.636, 0.643, 0.664, 0.684, 0.687, 0.694, 0.705, 0.714, 0.719, 0.725, 0.735, 0.746, 0.753, 0.767, 0.794, 0.803, 0.813, 0.837, 0.849, 0.861, 0.867, 0.874, 0.879, 1.000],
    69: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.005, 0.007, 0.007, 0.011, 0.020, 0.020, 0.023, 0.028, 0.034, 0.037, 0.039, 0.044, 0.050, 0.050, 0.052, 0.053, 0.059, 0.066, 0.067, 0.073, 0.076, 0.078, 0.091, 0.091, 0.092, 0.103, 0.105, 0.107, 0.107, 0.114, 0.119, 0.121, 0.121, 0.123, 0.131, 0.131, 0.147, 0.153, 0.165, 0.172, 0.178, 0.181, 0.197, 0.204, 0.222, 0.240, 0.256, 0.274, 0.302, 0.318, 0.348, 0.380, 0.409, 0.453, 0.483, 0.510, 0.520, 0.542, 0.558, 0.584, 0.613, 0.639, 0.643, 0.666, 0.684, 0.689, 0.694, 0.707, 0.714, 0.719, 0.728, 0.737, 0.748, 0.757, 0.769, 0.794, 0.806, 0.817, 0.842, 0.851, 0.863, 0.867, 0.874, 0.881, 1.000],
    70: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.005, 0.007, 0.007, 0.011, 0.020, 0.020, 0.023, 0.028, 0.034, 0.041, 0.041, 0.048, 0.050, 0.050, 0.053, 0.053, 0.059, 0.066, 0.067, 0.073, 0.076, 0.078, 0.091, 0.091, 0.092, 0.103, 0.105, 0.107, 0.107, 0.114, 0.119, 0.121, 0.121, 0.123, 0.131, 0.133, 0.147, 0.156, 0.165, 0.172, 0.178, 0.181, 0.197, 0.204, 0.224, 0.240, 0.256, 0.274, 0.302, 0.318, 0.348, 0.385, 0.417, 0.460, 0.496, 0.510, 0.520, 0.542, 0.558, 0.586, 0.616, 0.639, 0.643, 0.670, 0.686, 0.689, 0.694, 0.707, 0.714, 0.719, 0.728, 0.737, 0.748, 0.757, 0.769, 0.794, 0.806, 0.817, 0.842, 0.853, 0.863, 0.867, 0.876, 0.881, 1.000],
    71: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.005, 0.009, 0.009, 0.016, 0.020, 0.020, 0.027, 0.030, 0.036, 0.041, 0.044, 0.048, 0.052, 0.053, 0.053, 0.057, 0.059, 0.067, 0.071, 0.076, 0.078, 0.080, 0.091, 0.092, 0.096, 0.105, 0.105, 0.107, 0.112, 0.119, 0.121, 0.121, 0.121, 0.124, 0.133, 0.133, 0.147, 0.156, 0.167, 0.172, 0.178, 0.181, 0.197, 0.204, 0.224, 0.242, 0.258, 0.277, 0.302, 0.329, 0.361, 0.394, 0.435, 0.472, 0.501, 0.510, 0.528, 0.547, 0.565, 0.588, 0.618, 0.639, 0.647, 0.670, 0.686, 0.689, 0.694, 0.707, 0.714, 0.723, 0.728, 0.737, 0.750, 0.758, 0.769, 0.796, 0.808, 0.819, 0.842, 0.854, 0.863, 0.867, 0.876, 0.881, 1.000],
    72: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.009, 0.009, 0.009, 0.016, 0.020, 0.021, 0.027, 0.030, 0.036, 0.041, 0.044, 0.048, 0.052, 0.053, 0.053, 0.059, 0.059, 0.067, 0.071, 0.076, 0.078, 0.080, 0.091, 0.092, 0.096, 0.105, 0.105, 0.107, 0.114, 0.119, 0.121, 0.121, 0.121, 0.124, 0.133, 0.133, 0.147, 0.156, 0.167, 0.172, 0.178, 0.183, 0.197, 0.204, 0.224, 0.243, 0.258, 0.277, 0.302, 0.329, 0.362, 0.394, 0.448, 0.472, 0.501, 0.510, 0.528, 0.549, 0.567, 0.590, 0.618, 0.639, 0.647, 0.670, 0.686, 0.689, 0.694, 0.707, 0.714, 0.723, 0.730, 0.737, 0.750, 0.760, 0.771, 0.798, 0.810, 0.819, 0.842, 0.854, 0.863, 0.869, 0.876, 0.881, 1.000],
    73: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.005, 0.009, 0.009, 0.009, 0.016, 0.020, 0.021, 0.028, 0.030, 0.037, 0.041, 0.044, 0.050, 0.052, 0.053, 0.053, 0.059, 0.062, 0.069, 0.071, 0.076, 0.078, 0.080, 0.092, 0.094, 0.096, 0.105, 0.107, 0.107, 0.114, 0.119, 0.121, 0.121, 0.121, 0.126, 0.133, 0.135, 0.149, 0.160, 0.171, 0.176, 0.179, 0.183, 0.197, 0.210, 0.224, 0.243, 0.259, 0.282, 0.306, 0.336, 0.368, 0.394, 0.449, 0.478, 0.501, 0.510, 0.528, 0.551, 0.568, 0.595, 0.623, 0.639, 0.648, 0.670, 0.687, 0.689, 0.696, 0.709, 0.714, 0.725, 0.730, 0.742, 0.755, 0.762, 0.778, 0.801, 0.810, 0.821, 0.844, 0.854, 0.863, 0.869, 0.876, 0.883, 1.000],
    74: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.007, 0.009, 0.009, 0.009, 0.016, 0.020, 0.021, 0.028, 0.030, 0.037, 0.041, 0.044, 0.050, 0.052, 0.053, 0.053, 0.059, 0.062, 0.069, 0.071, 0.076, 0.078, 0.080, 0.092, 0.094, 0.098, 0.105, 0.107, 0.107, 0.114, 0.119, 0.121, 0.121, 0.121, 0.126, 0.135, 0.135, 0.149, 0.160, 0.171, 0.176, 0.181, 0.183, 0.199, 0.210, 0.226, 0.245, 0.259, 0.288, 0.311, 0.337, 0.368, 0.394, 0.449, 0.478, 0.501, 0.512, 0.531, 0.552, 0.568, 0.597, 0.623, 0.639, 0.652, 0.670, 0.687, 0.691, 0.698, 0.709, 0.716, 0.725, 0.730, 0.742, 0.755, 0.762, 0.778, 0.801, 0.810, 0.824, 0.844, 0.854, 0.863, 0.869, 0.881, 0.885, 1.000],
    75: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.005, 0.007, 0.009, 0.009, 0.009, 0.016, 0.020, 0.021, 0.030, 0.030, 0.037, 0.044, 0.048, 0.052, 0.052, 0.053, 0.053, 0.059, 0.062, 0.069, 0.071, 0.076, 0.080, 0.080, 0.092, 0.098, 0.098, 0.105, 0.107, 0.107, 0.115, 0.119, 0.121, 0.121, 0.121, 0.126, 0.135, 0.135, 0.151, 0.160, 0.172, 0.176, 0.181, 0.188, 0.199, 0.210, 0.231, 0.247, 0.261, 0.293, 0.316, 0.341, 0.371, 0.401, 0.449, 0.480, 0.503, 0.513, 0.535, 0.556, 0.570, 0.597, 0.623, 0.641, 0.652, 0.670, 0.687, 0.693, 0.700, 0.709, 0.718, 0.725, 0.732, 0.742, 0.755, 0.762, 0.778, 0.801, 0.810, 0.824, 0.844, 0.856, 0.863, 0.874, 0.883, 0.888, 1.000],
    76: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.005, 0.009, 0.009, 0.009, 0.009, 0.018, 0.020, 0.023, 0.030, 0.030, 0.037, 0.044, 0.048, 0.052, 0.052, 0.053, 0.053, 0.059, 0.062, 0.069, 0.071, 0.078, 0.080, 0.080, 0.094, 0.098, 0.099, 0.105, 0.107, 0.108, 0.115, 0.119, 0.121, 0.121, 0.123, 0.128, 0.135, 0.135, 0.151, 0.160, 0.172, 0.178, 0.183, 0.188, 0.199, 0.217, 0.233, 0.247, 0.261, 0.295, 0.320, 0.343, 0.375, 0.401, 0.451, 0.480, 0.503, 0.517, 0.538, 0.556, 0.570, 0.597, 0.625, 0.645, 0.657, 0.671, 0.689, 0.694, 0.700, 0.709, 0.718, 0.726, 0.732, 0.744, 0.758, 0.764, 0.778, 0.801, 0.813, 0.824, 0.845, 0.860, 0.869, 0.877, 0.886, 0.888, 1.000],
    77: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.005, 0.009, 0.009, 0.009, 0.014, 0.018, 0.020, 0.023, 0.030, 0.030, 0.037, 0.044, 0.048, 0.052, 0.052, 0.053, 0.053, 0.059, 0.062, 0.069, 0.073, 0.078, 0.080, 0.080, 0.094, 0.098, 0.099, 0.105, 0.107, 0.108, 0.115, 0.119, 0.121, 0.121, 0.123, 0.128, 0.135, 0.135, 0.151, 0.162, 0.172, 0.179, 0.183, 0.192, 0.206, 0.217, 0.233, 0.247, 0.261, 0.295, 0.325, 0.346, 0.382, 0.410, 0.455, 0.481, 0.504, 0.519, 0.542, 0.556, 0.570, 0.600, 0.629, 0.645, 0.659, 0.671, 0.689, 0.698, 0.700, 0.710, 0.718, 0.726, 0.732, 0.744, 0.758, 0.766, 0.782, 0.805, 0.813, 0.824, 0.845, 0.860, 0.869, 0.877, 0.888, 0.890, 1.000],
    78: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.009, 0.009, 0.009, 0.016, 0.021, 0.023, 0.025, 0.030, 0.030, 0.041, 0.046, 0.052, 0.052, 0.052, 0.053, 0.057, 0.059, 0.066, 0.071, 0.075, 0.078, 0.080, 0.080, 0.094, 0.098, 0.099, 0.105, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.123, 0.131, 0.137, 0.140, 0.151, 0.165, 0.178, 0.185, 0.185, 0.192, 0.206, 0.218, 0.233, 0.249, 0.263, 0.300, 0.327, 0.355, 0.394, 0.417, 0.458, 0.485, 0.506, 0.522, 0.542, 0.558, 0.574, 0.606, 0.632, 0.645, 0.663, 0.671, 0.693, 0.698, 0.702, 0.710, 0.719, 0.726, 0.734, 0.744, 0.758, 0.766, 0.782, 0.805, 0.813, 0.826, 0.853, 0.860, 0.869, 0.879, 0.888, 0.890, 1.000],
    79: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.009, 0.009, 0.009, 0.018, 0.021, 0.023, 0.025, 0.030, 0.030, 0.041, 0.046, 0.052, 0.052, 0.052, 0.053, 0.057, 0.059, 0.066, 0.071, 0.075, 0.078, 0.080, 0.080, 0.094, 0.098, 0.099, 0.105, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.124, 0.131, 0.139, 0.140, 0.153, 0.165, 0.178, 0.185, 0.185, 0.192, 0.206, 0.220, 0.234, 0.249, 0.265, 0.302, 0.329, 0.355, 0.394, 0.417, 0.458, 0.485, 0.506, 0.522, 0.542, 0.558, 0.574, 0.607, 0.634, 0.645, 0.663, 0.671, 0.693, 0.698, 0.702, 0.710, 0.719, 0.726, 0.735, 0.744, 0.760, 0.769, 0.782, 0.805, 0.813, 0.831, 0.853, 0.860, 0.869, 0.879, 0.888, 0.892, 1.000],
    80: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.009, 0.009, 0.009, 0.018, 0.021, 0.023, 0.025, 0.030, 0.030, 0.041, 0.048, 0.052, 0.052, 0.052, 0.053, 0.057, 0.059, 0.066, 0.071, 0.075, 0.078, 0.080, 0.080, 0.094, 0.098, 0.099, 0.105, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.124, 0.133, 0.140, 0.142, 0.156, 0.169, 0.178, 0.185, 0.185, 0.194, 0.208, 0.224, 0.234, 0.254, 0.265, 0.306, 0.330, 0.355, 0.394, 0.417, 0.458, 0.488, 0.510, 0.524, 0.544, 0.558, 0.574, 0.607, 0.636, 0.647, 0.664, 0.677, 0.694, 0.700, 0.703, 0.714, 0.723, 0.730, 0.735, 0.748, 0.760, 0.771, 0.782, 0.805, 0.817, 0.831, 0.854, 0.860, 0.869, 0.879, 0.888, 0.893, 1.000],
    81: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.009, 0.009, 0.014, 0.018, 0.021, 0.023, 0.025, 0.030, 0.030, 0.041, 0.048, 0.052, 0.052, 0.052, 0.053, 0.057, 0.059, 0.066, 0.071, 0.075, 0.078, 0.080, 0.080, 0.094, 0.098, 0.099, 0.107, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.124, 0.133, 0.140, 0.147, 0.158, 0.169, 0.179, 0.185, 0.187, 0.194, 0.210, 0.224, 0.240, 0.254, 0.266, 0.306, 0.330, 0.357, 0.396, 0.417, 0.460, 0.490, 0.513, 0.526, 0.545, 0.561, 0.577, 0.613, 0.636, 0.647, 0.666, 0.677, 0.694, 0.700, 0.705, 0.718, 0.723, 0.730, 0.735, 0.750, 0.762, 0.771, 0.782, 0.805, 0.819, 0.831, 0.854, 0.860, 0.869, 0.879, 0.890, 0.895, 1.000],
    82: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.009, 0.009, 0.014, 0.018, 0.023, 0.023, 0.025, 0.030, 0.030, 0.044, 0.048, 0.052, 0.052, 0.052, 0.053, 0.057, 0.059, 0.066, 0.071, 0.076, 0.078, 0.080, 0.082, 0.096, 0.099, 0.099, 0.107, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.124, 0.137, 0.144, 0.147, 0.162, 0.169, 0.179, 0.187, 0.187, 0.194, 0.210, 0.224, 0.240, 0.258, 0.270, 0.313, 0.336, 0.357, 0.405, 0.417, 0.462, 0.494, 0.515, 0.528, 0.547, 0.567, 0.583, 0.615, 0.639, 0.648, 0.668, 0.680, 0.694, 0.705, 0.705, 0.718, 0.725, 0.734, 0.744, 0.755, 0.762, 0.771, 0.782, 0.808, 0.821, 0.833, 0.854, 0.860, 0.869, 0.883, 0.893, 0.895, 1.000],
    83: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.009, 0.009, 0.014, 0.018, 0.023, 0.023, 0.025, 0.030, 0.030, 0.044, 0.048, 0.052, 0.052, 0.052, 0.053, 0.057, 0.059, 0.066, 0.071, 0.076, 0.078, 0.080, 0.082, 0.096, 0.099, 0.099, 0.107, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.124, 0.137, 0.144, 0.147, 0.162, 0.169, 0.179, 0.187, 0.187, 0.195, 0.210, 0.227, 0.242, 0.258, 0.270, 0.314, 0.341, 0.361, 0.405, 0.419, 0.464, 0.494, 0.515, 0.528, 0.551, 0.567, 0.586, 0.615, 0.641, 0.650, 0.670, 0.684, 0.698, 0.705, 0.707, 0.718, 0.728, 0.735, 0.746, 0.757, 0.762, 0.771, 0.782, 0.810, 0.821, 0.833, 0.854, 0.860, 0.870, 0.883, 0.893, 0.895, 1.000],
    84: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.005, 0.007, 0.009, 0.009, 0.014, 0.020, 0.023, 0.023, 0.027, 0.030, 0.036, 0.044, 0.052, 0.052, 0.052, 0.052, 0.053, 0.057, 0.059, 0.067, 0.071, 0.076, 0.078, 0.080, 0.082, 0.096, 0.099, 0.099, 0.107, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.128, 0.137, 0.147, 0.149, 0.162, 0.171, 0.181, 0.187, 0.190, 0.195, 0.215, 0.229, 0.242, 0.261, 0.274, 0.314, 0.343, 0.373, 0.409, 0.423, 0.465, 0.494, 0.519, 0.529, 0.558, 0.574, 0.593, 0.622, 0.647, 0.655, 0.671, 0.687, 0.698, 0.705, 0.709, 0.719, 0.730, 0.737, 0.748, 0.757, 0.764, 0.771, 0.783, 0.810, 0.821, 0.833, 0.856, 0.863, 0.876, 0.886, 0.899, 0.899, 1.000],
    85: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.005, 0.007, 0.009, 0.009, 0.014, 0.020, 0.023, 0.023, 0.027, 0.030, 0.036, 0.044, 0.052, 0.052, 0.052, 0.052, 0.053, 0.057, 0.062, 0.069, 0.075, 0.076, 0.080, 0.080, 0.083, 0.098, 0.099, 0.099, 0.107, 0.108, 0.108, 0.117, 0.121, 0.121, 0.121, 0.128, 0.137, 0.147, 0.151, 0.162, 0.171, 0.181, 0.187, 0.190, 0.195, 0.217, 0.231, 0.243, 0.261, 0.274, 0.314, 0.345, 0.375, 0.410, 0.423, 0.472, 0.496, 0.522, 0.531, 0.558, 0.574, 0.595, 0.623, 0.647, 0.657, 0.671, 0.687, 0.703, 0.707, 0.709, 0.723, 0.732, 0.739, 0.748, 0.757, 0.764, 0.773, 0.785, 0.810, 0.822, 0.833, 0.856, 0.867, 0.879, 0.888, 0.902, 0.904, 1.000],
    86: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.007, 0.007, 0.009, 0.009, 0.016, 0.020, 0.023, 0.023, 0.027, 0.030, 0.043, 0.046, 0.052, 0.052, 0.052, 0.052, 0.053, 0.060, 0.062, 0.073, 0.075, 0.076, 0.082, 0.083, 0.085, 0.098, 0.099, 0.103, 0.108, 0.108, 0.110, 0.117, 0.121, 0.121, 0.123, 0.130, 0.139, 0.151, 0.153, 0.165, 0.172, 0.181, 0.188, 0.190, 0.201, 0.220, 0.234, 0.243, 0.261, 0.277, 0.318, 0.350, 0.378, 0.410, 0.425, 0.474, 0.497, 0.524, 0.535, 0.560, 0.575, 0.597, 0.627, 0.648, 0.657, 0.677, 0.691, 0.705, 0.709, 0.714, 0.723, 0.732, 0.741, 0.753, 0.758, 0.767, 0.774, 0.787, 0.817, 0.824, 0.840, 0.858, 0.872, 0.885, 0.895, 0.904, 0.908, 1.000],
    87: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.007, 0.009, 0.009, 0.016, 0.020, 0.023, 0.023, 0.030, 0.039, 0.044, 0.046, 0.052, 0.052, 0.052, 0.052, 0.053, 0.060, 0.064, 0.073, 0.075, 0.078, 0.082, 0.083, 0.085, 0.098, 0.099, 0.103, 0.108, 0.112, 0.114, 0.119, 0.121, 0.121, 0.123, 0.130, 0.139, 0.151, 0.153, 0.165, 0.174, 0.181, 0.188, 0.190, 0.202, 0.226, 0.242, 0.245, 0.263, 0.282, 0.320, 0.352, 0.380, 0.417, 0.435, 0.487, 0.503, 0.531, 0.549, 0.575, 0.593, 0.613, 0.643, 0.654, 0.661, 0.680, 0.691, 0.705, 0.710, 0.714, 0.725, 0.732, 0.742, 0.753, 0.758, 0.767, 0.776, 0.789, 0.821, 0.824, 0.842, 0.860, 0.874, 0.885, 0.897, 0.906, 0.908, 1.000],
    88: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.007, 0.009, 0.009, 0.018, 0.021, 0.023, 0.023, 0.030, 0.043, 0.044, 0.046, 0.052, 0.052, 0.053, 0.053, 0.055, 0.062, 0.064, 0.075, 0.076, 0.078, 0.082, 0.083, 0.085, 0.098, 0.099, 0.103, 0.108, 0.112, 0.114, 0.121, 0.121, 0.121, 0.126, 0.130, 0.139, 0.151, 0.153, 0.165, 0.176, 0.185, 0.190, 0.192, 0.215, 0.229, 0.242, 0.250, 0.266, 0.282, 0.334, 0.361, 0.393, 0.432, 0.448, 0.492, 0.519, 0.544, 0.554, 0.579, 0.604, 0.622, 0.647, 0.655, 0.666, 0.680, 0.698, 0.705, 0.710, 0.719, 0.726, 0.737, 0.744, 0.755, 0.762, 0.769, 0.782, 0.798, 0.821, 0.826, 0.845, 0.861, 0.877, 0.892, 0.902, 0.908, 0.909, 1.000],
    89: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.007, 0.007, 0.009, 0.009, 0.018, 0.021, 0.023, 0.028, 0.032, 0.043, 0.046, 0.052, 0.053, 0.053, 0.057, 0.057, 0.059, 0.062, 0.067, 0.075, 0.076, 0.080, 0.082, 0.085, 0.085, 0.098, 0.099, 0.103, 0.108, 0.114, 0.114, 0.121, 0.121, 0.123, 0.128, 0.133, 0.140, 0.153, 0.155, 0.171, 0.183, 0.194, 0.197, 0.201, 0.224, 0.238, 0.247, 0.254, 0.275, 0.295, 0.343, 0.373, 0.400, 0.437, 0.451, 0.497, 0.526, 0.545, 0.560, 0.590, 0.611, 0.625, 0.647, 0.657, 0.670, 0.686, 0.702, 0.709, 0.716, 0.725, 0.732, 0.742, 0.750, 0.760, 0.766, 0.771, 0.787, 0.803, 0.824, 0.831, 0.853, 0.869, 0.885, 0.899, 0.904, 0.909, 0.909, 1.000],
    90: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.007, 0.007, 0.009, 0.009, 0.018, 0.021, 0.025, 0.028, 0.032, 0.043, 0.046, 0.052, 0.053, 0.053, 0.057, 0.059, 0.060, 0.066, 0.067, 0.075, 0.076, 0.080, 0.083, 0.085, 0.085, 0.098, 0.099, 0.103, 0.108, 0.114, 0.114, 0.121, 0.121, 0.124, 0.133, 0.135, 0.142, 0.155, 0.160, 0.174, 0.185, 0.194, 0.199, 0.211, 0.224, 0.238, 0.249, 0.256, 0.281, 0.302, 0.348, 0.377, 0.400, 0.437, 0.455, 0.506, 0.531, 0.549, 0.561, 0.591, 0.613, 0.631, 0.650, 0.666, 0.673, 0.689, 0.707, 0.712, 0.718, 0.725, 0.732, 0.742, 0.753, 0.762, 0.767, 0.774, 0.789, 0.805, 0.829, 0.840, 0.853, 0.869, 0.885, 0.899, 0.904, 0.911, 0.911, 1.000]
}

MEAN_PERCENTILES = {
    1: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.009, 0.021, 0.025, 0.030, 0.032, 0.035, 0.044, 0.050, 0.055, 0.059, 0.066, 0.089, 0.099, 0.110, 0.122, 0.133, 0.163, 0.191, 0.213, 0.243, 0.287, 0.340, 0.388, 0.452, 0.473, 0.498, 0.528, 0.559, 0.587, 0.606, 0.638, 0.668, 0.695, 0.725, 0.768, 0.807, 0.848, 0.920, 0.968, 1.000],
    2: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.014, 0.021, 0.028, 0.032, 0.039, 0.043, 0.048, 0.053, 0.059, 0.062, 0.066, 0.074, 0.083, 0.094, 0.108, 0.122, 0.137, 0.154, 0.168, 0.195, 0.223, 0.257, 0.298, 0.342, 0.387, 0.431, 0.477, 0.495, 0.532, 0.550, 0.574, 0.603, 0.622, 0.651, 0.679, 0.702, 0.732, 0.770, 0.798, 0.830, 0.888, 0.949, 0.975, 1.000],
    3: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.009, 0.018, 0.020, 0.023, 0.028, 0.034, 0.041, 0.046, 0.053, 0.062, 0.064, 0.071, 0.076, 0.087, 0.092, 0.096, 0.101, 0.122, 0.137, 0.156, 0.168, 0.195, 0.232, 0.257, 0.298, 0.360, 0.401, 0.434, 0.479, 0.495, 0.530, 0.553, 0.583, 0.596, 0.615, 0.638, 0.661, 0.697, 0.718, 0.752, 0.784, 0.812, 0.851, 0.906, 0.956, 0.980, 1.000],
    4: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.009, 0.011, 0.016, 0.020, 0.020, 0.021, 0.030, 0.032, 0.039, 0.050, 0.059, 0.064, 0.067, 0.073, 0.078, 0.080, 0.089, 0.096, 0.108, 0.113, 0.128, 0.149, 0.161, 0.172, 0.206, 0.234, 0.271, 0.293, 0.333, 0.394, 0.441, 0.475, 0.496, 0.523, 0.548, 0.574, 0.590, 0.610, 0.626, 0.658, 0.690, 0.711, 0.738, 0.770, 0.791, 0.821, 0.862, 0.920, 0.965, 0.984, 1.000],
    5: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.011, 0.011, 0.018, 0.018, 0.021, 0.021, 0.025, 0.041, 0.048, 0.051, 0.059, 0.069, 0.076, 0.078, 0.080, 0.082, 0.087, 0.094, 0.110, 0.112, 0.122, 0.138, 0.160, 0.174, 0.195, 0.229, 0.264, 0.298, 0.346, 0.385, 0.438, 0.466, 0.496, 0.520, 0.546, 0.560, 0.585, 0.601, 0.621, 0.640, 0.676, 0.706, 0.725, 0.750, 0.784, 0.805, 0.846, 0.879, 0.933, 0.970, 0.988, 1.000],
    6: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.012, 0.014, 0.020, 0.020, 0.027, 0.030, 0.043, 0.048, 0.055, 0.060, 0.071, 0.076, 0.078, 0.083, 0.083, 0.087, 0.096, 0.110, 0.112, 0.122, 0.135, 0.158, 0.176, 0.200, 0.223, 0.250, 0.291, 0.337, 0.381, 0.415, 0.461, 0.491, 0.511, 0.537, 0.553, 0.576, 0.592, 0.613, 0.633, 0.654, 0.690, 0.711, 0.734, 0.759, 0.789, 0.807, 0.849, 0.890, 0.940, 0.973, 0.988, 1.000],
    7: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.014, 0.014, 0.016, 0.021, 0.027, 0.030, 0.037, 0.050, 0.057, 0.066, 0.067, 0.078, 0.078, 0.078, 0.083, 0.089, 0.096, 0.101, 0.112, 0.119, 0.128, 0.140, 0.165, 0.181, 0.215, 0.236, 0.273, 0.321, 0.358, 0.397, 0.445, 0.475, 0.495, 0.512, 0.541, 0.564, 0.582, 0.598, 0.621, 0.633, 0.665, 0.697, 0.722, 0.741, 0.764, 0.794, 0.814, 0.860, 0.902, 0.945, 0.977, 0.989, 1.000],
    8: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.012, 0.014, 0.014, 0.021, 0.027, 0.035, 0.037, 0.046, 0.064, 0.067, 0.071, 0.073, 0.078, 0.082, 0.083, 0.087, 0.096, 0.101, 0.108, 0.122, 0.128, 0.140, 0.151, 0.176, 0.202, 0.225, 0.248, 0.301, 0.346, 0.381, 0.429, 0.457, 0.486, 0.507, 0.528, 0.555, 0.574, 0.583, 0.613, 0.633, 0.644, 0.681, 0.709, 0.723, 0.754, 0.777, 0.801, 0.828, 0.867, 0.913, 0.947, 0.979, 0.989, 1.000],
    9: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.011, 0.012, 0.014, 0.018, 0.023, 0.039, 0.041, 0.046, 0.053, 0.064, 0.067, 0.071, 0.076, 0.083, 0.083, 0.087, 0.089, 0.090, 0.099, 0.105, 0.121, 0.129, 0.133, 0.147, 0.163, 0.186, 0.215, 0.241, 0.261, 0.328, 0.360, 0.397, 0.445, 0.472, 0.489, 0.512, 0.541, 0.560, 0.582, 0.590, 0.621, 0.635, 0.651, 0.690, 0.718, 0.729, 0.762, 0.782, 0.810, 0.837, 0.876, 0.926, 0.950, 0.979, 0.991, 1.000],
    10: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.011, 0.012, 0.012, 0.018, 0.032, 0.035, 0.039, 0.044, 0.051, 0.066, 0.069, 0.074, 0.078, 0.083, 0.085, 0.087, 0.090, 0.092, 0.094, 0.103, 0.113, 0.122, 0.133, 0.138, 0.160, 0.174, 0.193, 0.227, 0.259, 0.293, 0.351, 0.372, 0.429, 0.456, 0.486, 0.505, 0.528, 0.551, 0.569, 0.587, 0.603, 0.629, 0.640, 0.667, 0.695, 0.722, 0.736, 0.775, 0.784, 0.812, 0.842, 0.887, 0.927, 0.950, 0.979, 0.991, 1.000],
    11: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.012, 0.012, 0.018, 0.027, 0.034, 0.041, 0.043, 0.053, 0.059, 0.066, 0.074, 0.074, 0.078, 0.085, 0.087, 0.090, 0.092, 0.092, 0.096, 0.105, 0.121, 0.128, 0.137, 0.138, 0.165, 0.177, 0.204, 0.238, 0.264, 0.303, 0.356, 0.379, 0.441, 0.466, 0.488, 0.512, 0.537, 0.557, 0.571, 0.589, 0.610, 0.633, 0.649, 0.676, 0.699, 0.722, 0.738, 0.778, 0.791, 0.819, 0.849, 0.887, 0.931, 0.961, 0.982, 0.991, 1.000],
    12: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.011, 0.016, 0.021, 0.028, 0.035, 0.043, 0.043, 0.046, 0.060, 0.067, 0.069, 0.074, 0.078, 0.087, 0.090, 0.090, 0.090, 0.094, 0.094, 0.098, 0.113, 0.126, 0.137, 0.138, 0.145, 0.176, 0.191, 0.218, 0.250, 0.273, 0.328, 0.369, 0.424, 0.454, 0.475, 0.504, 0.527, 0.544, 0.559, 0.580, 0.598, 0.619, 0.640, 0.658, 0.677, 0.706, 0.722, 0.746, 0.784, 0.794, 0.826, 0.856, 0.897, 0.933, 0.963, 0.984, 0.993, 1.000],
    13: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.007, 0.014, 0.028, 0.030, 0.037, 0.037, 0.046, 0.048, 0.055, 0.060, 0.071, 0.076, 0.076, 0.083, 0.090, 0.092, 0.092, 0.094, 0.096, 0.099, 0.101, 0.115, 0.128, 0.137, 0.138, 0.154, 0.188, 0.200, 0.229, 0.262, 0.287, 0.348, 0.401, 0.441, 0.468, 0.491, 0.509, 0.535, 0.544, 0.564, 0.587, 0.606, 0.631, 0.651, 0.661, 0.690, 0.709, 0.736, 0.752, 0.787, 0.807, 0.833, 0.863, 0.897, 0.938, 0.970, 0.988, 0.993, 1.000],
    14: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.021, 0.030, 0.030, 0.037, 0.041, 0.046, 0.053, 0.057, 0.062, 0.073, 0.078, 0.080, 0.090, 0.092, 0.092, 0.092, 0.094, 0.098, 0.101, 0.103, 0.115, 0.128, 0.138, 0.140, 0.160, 0.193, 0.207, 0.236, 0.268, 0.309, 0.356, 0.406, 0.447, 0.475, 0.495, 0.514, 0.541, 0.550, 0.571, 0.589, 0.612, 0.640, 0.656, 0.672, 0.693, 0.715, 0.738, 0.754, 0.789, 0.816, 0.837, 0.869, 0.906, 0.941, 0.970, 0.988, 0.993, 1.000],
    15: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.014, 0.021, 0.030, 0.032, 0.035, 0.041, 0.046, 0.051, 0.057, 0.059, 0.073, 0.076, 0.082, 0.089, 0.090, 0.092, 0.094, 0.094, 0.096, 0.101, 0.103, 0.106, 0.119, 0.131, 0.140, 0.152, 0.177, 0.202, 0.220, 0.250, 0.282, 0.321, 0.379, 0.415, 0.457, 0.488, 0.502, 0.518, 0.541, 0.557, 0.574, 0.590, 0.617, 0.645, 0.658, 0.676, 0.695, 0.723, 0.739, 0.761, 0.794, 0.826, 0.842, 0.872, 0.908, 0.943, 0.970, 0.988, 0.993, 1.000],
    16: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.009, 0.020, 0.025, 0.032, 0.034, 0.035, 0.046, 0.051, 0.055, 0.057, 0.066, 0.073, 0.078, 0.083, 0.090, 0.090, 0.092, 0.094, 0.094, 0.096, 0.101, 0.106, 0.115, 0.121, 0.131, 0.140, 0.160, 0.179, 0.204, 0.232, 0.259, 0.293, 0.335, 0.387, 0.427, 0.470, 0.489, 0.505, 0.520, 0.544, 0.562, 0.578, 0.598, 0.622, 0.647, 0.661, 0.679, 0.700, 0.725, 0.745, 0.768, 0.800, 0.826, 0.848, 0.876, 0.908, 0.945, 0.972, 0.988, 0.993, 1.000],
    17: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.007, 0.023, 0.027, 0.028, 0.034, 0.043, 0.046, 0.051, 0.057, 0.057, 0.060, 0.069, 0.074, 0.085, 0.089, 0.090, 0.090, 0.094, 0.094, 0.094, 0.103, 0.105, 0.112, 0.117, 0.129, 0.138, 0.147, 0.174, 0.195, 0.211, 0.245, 0.273, 0.303, 0.362, 0.402, 0.441, 0.482, 0.498, 0.509, 0.525, 0.553, 0.571, 0.592, 0.603, 0.629, 0.652, 0.667, 0.684, 0.709, 0.730, 0.748, 0.775, 0.801, 0.828, 0.860, 0.887, 0.917, 0.949, 0.973, 0.988, 0.993, 1.000],
    18: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.009, 0.023, 0.028, 0.030, 0.041, 0.044, 0.046, 0.051, 0.057, 0.057, 0.064, 0.069, 0.076, 0.087, 0.090, 0.090, 0.092, 0.094, 0.098, 0.098, 0.103, 0.105, 0.119, 0.124, 0.135, 0.140, 0.154, 0.179, 0.204, 0.213, 0.254, 0.280, 0.309, 0.369, 0.415, 0.449, 0.489, 0.500, 0.512, 0.527, 0.555, 0.580, 0.596, 0.610, 0.633, 0.652, 0.667, 0.690, 0.716, 0.732, 0.750, 0.777, 0.803, 0.833, 0.869, 0.892, 0.917, 0.949, 0.973, 0.989, 0.993, 1.000],
    19: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.025, 0.028, 0.028, 0.037, 0.043, 0.044, 0.046, 0.051, 0.059, 0.060, 0.064, 0.076, 0.082, 0.089, 0.090, 0.092, 0.094, 0.098, 0.098, 0.099, 0.103, 0.108, 0.121, 0.129, 0.142, 0.145, 0.168, 0.186, 0.206, 0.229, 0.268, 0.289, 0.326, 0.381, 0.433, 0.456, 0.495, 0.505, 0.528, 0.541, 0.562, 0.587, 0.603, 0.617, 0.640, 0.654, 0.668, 0.702, 0.718, 0.739, 0.752, 0.782, 0.816, 0.846, 0.874, 0.895, 0.917, 0.949, 0.973, 0.991, 0.993, 1.000],
    20: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.007, 0.009, 0.027, 0.028, 0.032, 0.039, 0.043, 0.044, 0.046, 0.053, 0.060, 0.064, 0.069, 0.078, 0.087, 0.089, 0.092, 0.096, 0.098, 0.099, 0.099, 0.101, 0.106, 0.108, 0.124, 0.133, 0.142, 0.149, 0.177, 0.191, 0.211, 0.232, 0.273, 0.300, 0.335, 0.390, 0.436, 0.457, 0.500, 0.505, 0.530, 0.543, 0.566, 0.590, 0.605, 0.621, 0.642, 0.654, 0.681, 0.702, 0.722, 0.739, 0.759, 0.791, 0.816, 0.848, 0.876, 0.899, 0.918, 0.950, 0.975, 0.991, 0.993, 1.000],
    21: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.011, 0.020, 0.027, 0.032, 0.035, 0.043, 0.044, 0.046, 0.050, 0.059, 0.062, 0.066, 0.076, 0.085, 0.090, 0.090, 0.092, 0.096, 0.098, 0.099, 0.099, 0.103, 0.108, 0.124, 0.131, 0.140, 0.147, 0.158, 0.183, 0.202, 0.220, 0.243, 0.278, 0.312, 0.353, 0.404, 0.441, 0.472, 0.502, 0.512, 0.539, 0.550, 0.576, 0.594, 0.613, 0.626, 0.647, 0.660, 0.691, 0.707, 0.723, 0.745, 0.770, 0.800, 0.817, 0.855, 0.885, 0.899, 0.922, 0.952, 0.979, 0.991, 0.993, 1.000],
    22: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.023, 0.028, 0.032, 0.037, 0.043, 0.044, 0.048, 0.053, 0.059, 0.064, 0.069, 0.080, 0.089, 0.090, 0.092, 0.092, 0.096, 0.098, 0.101, 0.103, 0.103, 0.112, 0.124, 0.131, 0.140, 0.151, 0.165, 0.184, 0.204, 0.222, 0.250, 0.282, 0.323, 0.367, 0.410, 0.447, 0.477, 0.502, 0.518, 0.541, 0.551, 0.580, 0.596, 0.615, 0.629, 0.647, 0.661, 0.695, 0.707, 0.723, 0.748, 0.771, 0.803, 0.817, 0.856, 0.885, 0.901, 0.924, 0.954, 0.979, 0.991, 0.993, 1.000],
    23: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.007, 0.021, 0.027, 0.032, 0.035, 0.041, 0.043, 0.048, 0.050, 0.055, 0.060, 0.064, 0.071, 0.085, 0.089, 0.092, 0.092, 0.096, 0.096, 0.099, 0.101, 0.103, 0.103, 0.119, 0.128, 0.135, 0.142, 0.158, 0.170, 0.199, 0.209, 0.229, 0.257, 0.287, 0.326, 0.378, 0.429, 0.457, 0.482, 0.504, 0.527, 0.546, 0.559, 0.583, 0.598, 0.617, 0.633, 0.652, 0.676, 0.695, 0.709, 0.734, 0.755, 0.771, 0.805, 0.817, 0.858, 0.890, 0.904, 0.924, 0.956, 0.979, 0.991, 0.993, 1.000],
    24: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.005, 0.007, 0.014, 0.021, 0.027, 0.035, 0.039, 0.041, 0.044, 0.048, 0.050, 0.059, 0.062, 0.066, 0.076, 0.085, 0.090, 0.092, 0.096, 0.096, 0.099, 0.101, 0.101, 0.103, 0.105, 0.124, 0.128, 0.137, 0.152, 0.158, 0.174, 0.200, 0.220, 0.234, 0.262, 0.293, 0.333, 0.399, 0.431, 0.463, 0.486, 0.507, 0.530, 0.548, 0.562, 0.583, 0.599, 0.619, 0.638, 0.660, 0.676, 0.695, 0.713, 0.734, 0.757, 0.777, 0.807, 0.823, 0.862, 0.892, 0.906, 0.927, 0.956, 0.980, 0.991, 0.993, 1.000],
    25: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.007, 0.009, 0.009, 0.016, 0.025, 0.027, 0.035, 0.039, 0.043, 0.046, 0.048, 0.051, 0.059, 0.062, 0.069, 0.082, 0.089, 0.092, 0.092, 0.096, 0.096, 0.099, 0.103, 0.103, 0.105, 0.119, 0.128, 0.135, 0.142, 0.156, 0.160, 0.181, 0.204, 0.225, 0.245, 0.270, 0.309, 0.363, 0.406, 0.440, 0.468, 0.495, 0.518, 0.534, 0.548, 0.571, 0.589, 0.605, 0.619, 0.644, 0.663, 0.676, 0.702, 0.720, 0.739, 0.757, 0.782, 0.807, 0.824, 0.865, 0.894, 0.908, 0.929, 0.956, 0.980, 0.991, 0.993, 1.000],
    26: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.007, 0.007, 0.009, 0.009, 0.009, 0.025, 0.030, 0.032, 0.037, 0.041, 0.044, 0.046, 0.051, 0.051, 0.060, 0.066, 0.078, 0.085, 0.090, 0.092, 0.096, 0.096, 0.099, 0.101, 0.105, 0.105, 0.112, 0.122, 0.133, 0.137, 0.145, 0.156, 0.165, 0.186, 0.209, 0.232, 0.254, 0.277, 0.328, 0.376, 0.417, 0.447, 0.475, 0.498, 0.521, 0.539, 0.548, 0.574, 0.590, 0.606, 0.624, 0.649, 0.663, 0.681, 0.702, 0.723, 0.739, 0.759, 0.785, 0.809, 0.828, 0.871, 0.897, 0.908, 0.933, 0.961, 0.982, 0.991, 0.993, 1.000],
    27: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.007, 0.009, 0.011, 0.014, 0.025, 0.030, 0.032, 0.037, 0.041, 0.044, 0.046, 0.051, 0.053, 0.062, 0.067, 0.078, 0.089, 0.090, 0.096, 0.096, 0.098, 0.101, 0.101, 0.105, 0.108, 0.115, 0.128, 0.133, 0.137, 0.149, 0.158, 0.165, 0.190, 0.216, 0.234, 0.261, 0.284, 0.335, 0.388, 0.422, 0.449, 0.479, 0.500, 0.525, 0.539, 0.551, 0.580, 0.590, 0.610, 0.624, 0.658, 0.667, 0.684, 0.706, 0.723, 0.743, 0.759, 0.787, 0.810, 0.828, 0.872, 0.899, 0.911, 0.933, 0.965, 0.982, 0.991, 0.993, 1.000],
    28: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.009, 0.009, 0.011, 0.011, 0.021, 0.027, 0.030, 0.035, 0.039, 0.043, 0.046, 0.050, 0.051, 0.055, 0.067, 0.074, 0.082, 0.089, 0.092, 0.096, 0.098, 0.101, 0.101, 0.103, 0.108, 0.108, 0.117, 0.128, 0.135, 0.138, 0.152, 0.160, 0.168, 0.195, 0.225, 0.245, 0.270, 0.296, 0.344, 0.399, 0.426, 0.452, 0.484, 0.511, 0.530, 0.539, 0.555, 0.585, 0.598, 0.612, 0.631, 0.660, 0.668, 0.688, 0.706, 0.725, 0.748, 0.761, 0.793, 0.814, 0.833, 0.876, 0.899, 0.913, 0.938, 0.966, 0.982, 0.993, 0.993, 1.000],
    29: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.009, 0.009, 0.011, 0.016, 0.025, 0.027, 0.030, 0.039, 0.043, 0.046, 0.046, 0.050, 0.051, 0.060, 0.069, 0.074, 0.085, 0.090, 0.092, 0.096, 0.099, 0.101, 0.101, 0.106, 0.110, 0.112, 0.122, 0.128, 0.137, 0.144, 0.156, 0.161, 0.179, 0.202, 0.225, 0.250, 0.271, 0.300, 0.358, 0.402, 0.429, 0.456, 0.484, 0.514, 0.532, 0.543, 0.557, 0.585, 0.599, 0.613, 0.637, 0.663, 0.672, 0.691, 0.707, 0.729, 0.754, 0.766, 0.793, 0.817, 0.837, 0.878, 0.901, 0.922, 0.941, 0.966, 0.986, 0.993, 0.993, 1.000],
    30: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.007, 0.009, 0.011, 0.016, 0.025, 0.025, 0.032, 0.034, 0.044, 0.046, 0.046, 0.048, 0.051, 0.053, 0.066, 0.071, 0.078, 0.085, 0.096, 0.096, 0.099, 0.099, 0.101, 0.101, 0.108, 0.117, 0.119, 0.126, 0.135, 0.138, 0.147, 0.158, 0.167, 0.181, 0.202, 0.229, 0.252, 0.282, 0.309, 0.360, 0.406, 0.431, 0.459, 0.488, 0.516, 0.535, 0.548, 0.560, 0.587, 0.599, 0.621, 0.647, 0.665, 0.677, 0.699, 0.709, 0.734, 0.757, 0.768, 0.794, 0.821, 0.844, 0.883, 0.902, 0.922, 0.941, 0.968, 0.986, 0.993, 0.993, 1.000],
    31: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.005, 0.005, 0.007, 0.007, 0.009, 0.016, 0.018, 0.025, 0.027, 0.032, 0.035, 0.044, 0.046, 0.046, 0.050, 0.051, 0.057, 0.066, 0.076, 0.082, 0.087, 0.098, 0.098, 0.099, 0.101, 0.101, 0.106, 0.108, 0.117, 0.121, 0.126, 0.137, 0.140, 0.147, 0.163, 0.170, 0.184, 0.206, 0.234, 0.261, 0.282, 0.312, 0.363, 0.406, 0.436, 0.461, 0.488, 0.518, 0.539, 0.550, 0.560, 0.592, 0.603, 0.624, 0.651, 0.665, 0.683, 0.702, 0.713, 0.734, 0.762, 0.768, 0.794, 0.821, 0.848, 0.885, 0.902, 0.924, 0.945, 0.972, 0.986, 0.993, 0.993, 1.000],
    32: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.005, 0.007, 0.009, 0.009, 0.016, 0.018, 0.023, 0.025, 0.027, 0.035, 0.041, 0.044, 0.046, 0.050, 0.053, 0.055, 0.062, 0.067, 0.076, 0.083, 0.090, 0.098, 0.098, 0.101, 0.101, 0.101, 0.106, 0.115, 0.119, 0.124, 0.126, 0.137, 0.144, 0.152, 0.165, 0.174, 0.188, 0.213, 0.243, 0.271, 0.284, 0.314, 0.381, 0.417, 0.441, 0.470, 0.489, 0.528, 0.539, 0.551, 0.567, 0.596, 0.605, 0.628, 0.652, 0.668, 0.683, 0.704, 0.715, 0.741, 0.762, 0.768, 0.794, 0.821, 0.851, 0.887, 0.902, 0.929, 0.945, 0.972, 0.986, 0.993, 0.993, 1.000],
    33: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.004, 0.007, 0.009, 0.014, 0.014, 0.018, 0.023, 0.023, 0.025, 0.028, 0.035, 0.044, 0.044, 0.048, 0.051, 0.053, 0.059, 0.069, 0.076, 0.080, 0.085, 0.094, 0.098, 0.099, 0.101, 0.101, 0.103, 0.115, 0.119, 0.121, 0.128, 0.129, 0.140, 0.147, 0.156, 0.168, 0.184, 0.195, 0.215, 0.245, 0.275, 0.291, 0.319, 0.387, 0.422, 0.454, 0.475, 0.489, 0.530, 0.543, 0.555, 0.574, 0.598, 0.608, 0.633, 0.656, 0.668, 0.686, 0.711, 0.722, 0.741, 0.762, 0.770, 0.800, 0.830, 0.858, 0.888, 0.906, 0.936, 0.947, 0.972, 0.986, 0.993, 0.993, 1.000],
    34: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.007, 0.009, 0.014, 0.020, 0.020, 0.023, 0.023, 0.023, 0.027, 0.030, 0.044, 0.044, 0.046, 0.048, 0.053, 0.059, 0.064, 0.071, 0.080, 0.083, 0.087, 0.094, 0.099, 0.101, 0.101, 0.101, 0.103, 0.117, 0.119, 0.124, 0.129, 0.133, 0.140, 0.152, 0.161, 0.174, 0.190, 0.200, 0.222, 0.254, 0.282, 0.300, 0.346, 0.388, 0.422, 0.456, 0.477, 0.502, 0.534, 0.543, 0.557, 0.578, 0.599, 0.612, 0.644, 0.660, 0.672, 0.695, 0.713, 0.725, 0.745, 0.768, 0.770, 0.809, 0.835, 0.863, 0.890, 0.908, 0.940, 0.949, 0.973, 0.986, 0.993, 0.993, 1.000],
    35: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.007, 0.011, 0.014, 0.020, 0.020, 0.023, 0.023, 0.027, 0.027, 0.039, 0.044, 0.046, 0.046, 0.050, 0.057, 0.059, 0.066, 0.074, 0.082, 0.085, 0.087, 0.099, 0.099, 0.101, 0.101, 0.103, 0.110, 0.117, 0.121, 0.124, 0.133, 0.137, 0.144, 0.156, 0.163, 0.181, 0.193, 0.211, 0.229, 0.257, 0.284, 0.319, 0.351, 0.394, 0.426, 0.457, 0.480, 0.511, 0.534, 0.544, 0.559, 0.583, 0.603, 0.617, 0.649, 0.663, 0.679, 0.697, 0.715, 0.725, 0.746, 0.768, 0.771, 0.814, 0.837, 0.869, 0.892, 0.908, 0.940, 0.956, 0.975, 0.986, 0.993, 0.993, 1.000],
    36: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.009, 0.014, 0.020, 0.021, 0.021, 0.023, 0.023, 0.027, 0.028, 0.044, 0.044, 0.046, 0.046, 0.050, 0.059, 0.062, 0.067, 0.080, 0.082, 0.085, 0.089, 0.099, 0.101, 0.101, 0.101, 0.103, 0.115, 0.117, 0.124, 0.128, 0.137, 0.138, 0.145, 0.160, 0.167, 0.183, 0.197, 0.211, 0.239, 0.270, 0.287, 0.319, 0.362, 0.401, 0.431, 0.461, 0.484, 0.516, 0.534, 0.550, 0.560, 0.590, 0.606, 0.626, 0.651, 0.667, 0.679, 0.700, 0.715, 0.732, 0.752, 0.768, 0.771, 0.817, 0.837, 0.871, 0.897, 0.918, 0.941, 0.961, 0.975, 0.986, 0.993, 0.993, 1.000],
    37: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.007, 0.012, 0.014, 0.021, 0.021, 0.021, 0.023, 0.025, 0.027, 0.037, 0.044, 0.044, 0.046, 0.048, 0.053, 0.062, 0.069, 0.074, 0.083, 0.083, 0.087, 0.094, 0.099, 0.101, 0.101, 0.101, 0.103, 0.115, 0.117, 0.126, 0.129, 0.137, 0.140, 0.147, 0.160, 0.167, 0.186, 0.204, 0.218, 0.241, 0.270, 0.293, 0.330, 0.374, 0.408, 0.434, 0.468, 0.491, 0.516, 0.539, 0.553, 0.566, 0.594, 0.612, 0.637, 0.654, 0.670, 0.683, 0.700, 0.716, 0.734, 0.755, 0.768, 0.773, 0.817, 0.840, 0.874, 0.899, 0.918, 0.943, 0.961, 0.975, 0.986, 0.993, 0.995, 1.000],
    38: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.007, 0.007, 0.012, 0.016, 0.021, 0.023, 0.023, 0.023, 0.025, 0.032, 0.041, 0.044, 0.046, 0.046, 0.050, 0.055, 0.064, 0.076, 0.078, 0.083, 0.083, 0.090, 0.099, 0.099, 0.101, 0.101, 0.101, 0.103, 0.115, 0.119, 0.129, 0.131, 0.138, 0.142, 0.149, 0.160, 0.172, 0.188, 0.204, 0.225, 0.248, 0.271, 0.293, 0.340, 0.378, 0.418, 0.438, 0.475, 0.496, 0.521, 0.543, 0.555, 0.582, 0.596, 0.612, 0.637, 0.656, 0.672, 0.683, 0.700, 0.718, 0.739, 0.759, 0.770, 0.775, 0.819, 0.840, 0.876, 0.901, 0.922, 0.947, 0.963, 0.975, 0.986, 0.993, 0.995, 1.000],
    39: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.007, 0.012, 0.016, 0.016, 0.021, 0.023, 0.023, 0.023, 0.030, 0.034, 0.043, 0.046, 0.048, 0.048, 0.055, 0.060, 0.074, 0.078, 0.078, 0.083, 0.087, 0.098, 0.099, 0.099, 0.101, 0.101, 0.101, 0.105, 0.115, 0.121, 0.129, 0.133, 0.138, 0.145, 0.151, 0.163, 0.172, 0.193, 0.209, 0.229, 0.248, 0.278, 0.310, 0.348, 0.379, 0.420, 0.443, 0.480, 0.502, 0.521, 0.543, 0.560, 0.587, 0.601, 0.617, 0.640, 0.660, 0.674, 0.686, 0.707, 0.720, 0.743, 0.764, 0.771, 0.777, 0.819, 0.844, 0.881, 0.902, 0.922, 0.949, 0.966, 0.975, 0.989, 0.993, 0.995, 1.000],
    40: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.005, 0.011, 0.012, 0.016, 0.021, 0.023, 0.023, 0.023, 0.023, 0.032, 0.041, 0.043, 0.046, 0.048, 0.055, 0.059, 0.064, 0.076, 0.078, 0.083, 0.083, 0.096, 0.098, 0.099, 0.099, 0.101, 0.101, 0.103, 0.105, 0.117, 0.121, 0.129, 0.135, 0.140, 0.145, 0.160, 0.163, 0.177, 0.193, 0.215, 0.232, 0.252, 0.278, 0.317, 0.348, 0.385, 0.433, 0.450, 0.480, 0.504, 0.532, 0.546, 0.560, 0.594, 0.605, 0.621, 0.647, 0.663, 0.674, 0.690, 0.711, 0.722, 0.745, 0.764, 0.771, 0.778, 0.819, 0.846, 0.885, 0.904, 0.926, 0.949, 0.966, 0.975, 0.989, 0.993, 0.995, 1.000],
    41: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.012, 0.012, 0.014, 0.016, 0.021, 0.023, 0.023, 0.023, 0.034, 0.037, 0.043, 0.048, 0.048, 0.051, 0.055, 0.064, 0.076, 0.078, 0.082, 0.083, 0.087, 0.098, 0.098, 0.099, 0.099, 0.103, 0.103, 0.105, 0.117, 0.119, 0.126, 0.133, 0.135, 0.140, 0.149, 0.161, 0.172, 0.183, 0.200, 0.216, 0.239, 0.254, 0.280, 0.317, 0.358, 0.397, 0.436, 0.457, 0.482, 0.511, 0.537, 0.550, 0.566, 0.598, 0.606, 0.622, 0.651, 0.663, 0.676, 0.693, 0.711, 0.725, 0.755, 0.766, 0.773, 0.778, 0.821, 0.848, 0.887, 0.904, 0.931, 0.950, 0.966, 0.975, 0.989, 0.993, 0.995, 1.000],
    42: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.005, 0.012, 0.012, 0.016, 0.016, 0.021, 0.023, 0.023, 0.025, 0.035, 0.041, 0.044, 0.048, 0.048, 0.053, 0.059, 0.069, 0.076, 0.078, 0.083, 0.089, 0.092, 0.098, 0.098, 0.099, 0.101, 0.103, 0.103, 0.112, 0.117, 0.122, 0.128, 0.135, 0.135, 0.140, 0.149, 0.168, 0.176, 0.186, 0.207, 0.220, 0.239, 0.259, 0.284, 0.321, 0.363, 0.413, 0.436, 0.466, 0.491, 0.514, 0.541, 0.551, 0.569, 0.601, 0.608, 0.624, 0.651, 0.665, 0.677, 0.693, 0.713, 0.727, 0.759, 0.766, 0.773, 0.778, 0.824, 0.853, 0.887, 0.906, 0.933, 0.950, 0.968, 0.975, 0.989, 0.993, 0.995, 1.000],
    43: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.007, 0.012, 0.016, 0.016, 0.016, 0.023, 0.023, 0.028, 0.030, 0.039, 0.041, 0.048, 0.050, 0.051, 0.053, 0.062, 0.069, 0.076, 0.080, 0.083, 0.089, 0.096, 0.098, 0.098, 0.099, 0.101, 0.103, 0.103, 0.115, 0.117, 0.124, 0.128, 0.135, 0.135, 0.140, 0.154, 0.170, 0.179, 0.193, 0.213, 0.223, 0.245, 0.262, 0.291, 0.330, 0.372, 0.417, 0.441, 0.473, 0.495, 0.520, 0.544, 0.557, 0.574, 0.605, 0.610, 0.633, 0.656, 0.668, 0.686, 0.702, 0.715, 0.730, 0.761, 0.766, 0.778, 0.784, 0.826, 0.856, 0.894, 0.910, 0.934, 0.952, 0.968, 0.975, 0.989, 0.993, 0.995, 1.000],
    44: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.002, 0.004, 0.007, 0.012, 0.016, 0.016, 0.016, 0.023, 0.023, 0.030, 0.032, 0.039, 0.041, 0.050, 0.050, 0.051, 0.053, 0.064, 0.071, 0.076, 0.080, 0.083, 0.092, 0.096, 0.098, 0.098, 0.101, 0.103, 0.103, 0.103, 0.115, 0.121, 0.126, 0.128, 0.135, 0.137, 0.144, 0.154, 0.172, 0.179, 0.193, 0.213, 0.227, 0.248, 0.268, 0.305, 0.335, 0.376, 0.418, 0.443, 0.479, 0.500, 0.520, 0.548, 0.557, 0.578, 0.605, 0.612, 0.633, 0.658, 0.670, 0.686, 0.702, 0.718, 0.732, 0.761, 0.766, 0.780, 0.789, 0.828, 0.863, 0.894, 0.913, 0.934, 0.954, 0.970, 0.975, 0.989, 0.993, 0.995, 1.000],
    45: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.009, 0.014, 0.016, 0.016, 0.016, 0.023, 0.028, 0.034, 0.039, 0.043, 0.046, 0.050, 0.051, 0.051, 0.055, 0.066, 0.073, 0.076, 0.080, 0.087, 0.092, 0.096, 0.098, 0.099, 0.101, 0.103, 0.103, 0.105, 0.115, 0.122, 0.126, 0.131, 0.135, 0.140, 0.151, 0.165, 0.176, 0.183, 0.199, 0.215, 0.234, 0.255, 0.278, 0.310, 0.351, 0.381, 0.422, 0.454, 0.488, 0.507, 0.527, 0.550, 0.566, 0.589, 0.606, 0.615, 0.645, 0.660, 0.676, 0.688, 0.706, 0.722, 0.739, 0.762, 0.771, 0.782, 0.791, 0.830, 0.869, 0.897, 0.915, 0.934, 0.954, 0.970, 0.980, 0.989, 0.993, 0.995, 1.000],
    46: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.009, 0.011, 0.014, 0.016, 0.016, 0.018, 0.025, 0.034, 0.034, 0.039, 0.043, 0.046, 0.050, 0.051, 0.051, 0.059, 0.071, 0.074, 0.078, 0.080, 0.090, 0.094, 0.096, 0.098, 0.099, 0.103, 0.103, 0.103, 0.112, 0.119, 0.124, 0.128, 0.131, 0.135, 0.140, 0.156, 0.167, 0.176, 0.190, 0.202, 0.218, 0.239, 0.262, 0.285, 0.317, 0.358, 0.387, 0.426, 0.463, 0.488, 0.509, 0.527, 0.550, 0.569, 0.596, 0.612, 0.617, 0.649, 0.667, 0.679, 0.688, 0.707, 0.722, 0.741, 0.764, 0.773, 0.782, 0.793, 0.832, 0.871, 0.897, 0.915, 0.936, 0.954, 0.972, 0.980, 0.989, 0.993, 0.995, 1.000],
    47: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.009, 0.011, 0.014, 0.018, 0.018, 0.020, 0.028, 0.037, 0.037, 0.043, 0.046, 0.046, 0.050, 0.051, 0.051, 0.062, 0.073, 0.074, 0.080, 0.080, 0.090, 0.096, 0.096, 0.098, 0.099, 0.103, 0.103, 0.110, 0.115, 0.119, 0.124, 0.128, 0.131, 0.135, 0.145, 0.160, 0.168, 0.177, 0.190, 0.204, 0.220, 0.243, 0.266, 0.287, 0.323, 0.362, 0.390, 0.433, 0.472, 0.489, 0.514, 0.530, 0.550, 0.571, 0.599, 0.612, 0.626, 0.656, 0.670, 0.679, 0.691, 0.709, 0.723, 0.743, 0.764, 0.773, 0.784, 0.794, 0.840, 0.878, 0.897, 0.917, 0.936, 0.956, 0.973, 0.980, 0.989, 0.993, 0.995, 1.000],
    48: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.004, 0.009, 0.011, 0.014, 0.018, 0.018, 0.020, 0.028, 0.037, 0.039, 0.043, 0.046, 0.046, 0.050, 0.051, 0.051, 0.062, 0.073, 0.074, 0.080, 0.080, 0.090, 0.096, 0.096, 0.098, 0.099, 0.103, 0.110, 0.113, 0.115, 0.121, 0.124, 0.128, 0.131, 0.140, 0.151, 0.161, 0.170, 0.179, 0.191, 0.204, 0.223, 0.245, 0.266, 0.291, 0.324, 0.362, 0.395, 0.434, 0.473, 0.496, 0.514, 0.530, 0.551, 0.574, 0.601, 0.612, 0.626, 0.663, 0.670, 0.679, 0.691, 0.711, 0.725, 0.743, 0.764, 0.773, 0.785, 0.800, 0.844, 0.878, 0.899, 0.917, 0.936, 0.956, 0.973, 0.982, 0.989, 0.993, 0.995, 1.000],
    49: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.009, 0.011, 0.014, 0.018, 0.018, 0.025, 0.028, 0.037, 0.039, 0.044, 0.046, 0.046, 0.051, 0.051, 0.053, 0.064, 0.073, 0.076, 0.080, 0.082, 0.092, 0.096, 0.096, 0.099, 0.099, 0.103, 0.110, 0.113, 0.115, 0.121, 0.126, 0.133, 0.137, 0.142, 0.160, 0.165, 0.172, 0.184, 0.193, 0.206, 0.230, 0.248, 0.271, 0.293, 0.328, 0.372, 0.401, 0.445, 0.475, 0.504, 0.523, 0.532, 0.557, 0.585, 0.605, 0.613, 0.642, 0.665, 0.670, 0.681, 0.693, 0.715, 0.725, 0.748, 0.766, 0.777, 0.787, 0.807, 0.846, 0.881, 0.901, 0.917, 0.941, 0.961, 0.973, 0.982, 0.989, 0.993, 0.995, 1.000],
    50: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.009, 0.011, 0.014, 0.018, 0.020, 0.025, 0.028, 0.037, 0.041, 0.046, 0.046, 0.046, 0.051, 0.051, 0.055, 0.067, 0.074, 0.076, 0.080, 0.083, 0.096, 0.096, 0.098, 0.099, 0.099, 0.103, 0.110, 0.115, 0.119, 0.122, 0.126, 0.133, 0.137, 0.145, 0.161, 0.165, 0.172, 0.186, 0.195, 0.211, 0.230, 0.255, 0.275, 0.296, 0.330, 0.374, 0.411, 0.449, 0.477, 0.504, 0.523, 0.537, 0.562, 0.590, 0.608, 0.619, 0.645, 0.668, 0.672, 0.681, 0.693, 0.715, 0.729, 0.748, 0.766, 0.777, 0.789, 0.810, 0.848, 0.881, 0.902, 0.917, 0.941, 0.961, 0.973, 0.982, 0.989, 0.993, 0.995, 1.000],
    51: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.004, 0.009, 0.011, 0.014, 0.018, 0.025, 0.028, 0.032, 0.037, 0.041, 0.046, 0.048, 0.050, 0.051, 0.053, 0.060, 0.067, 0.074, 0.078, 0.080, 0.083, 0.096, 0.098, 0.098, 0.099, 0.099, 0.103, 0.110, 0.115, 0.119, 0.122, 0.129, 0.133, 0.138, 0.151, 0.165, 0.165, 0.172, 0.188, 0.197, 0.213, 0.236, 0.259, 0.275, 0.301, 0.335, 0.381, 0.418, 0.456, 0.479, 0.504, 0.523, 0.539, 0.564, 0.596, 0.613, 0.619, 0.649, 0.668, 0.676, 0.688, 0.695, 0.716, 0.729, 0.750, 0.768, 0.778, 0.791, 0.812, 0.848, 0.883, 0.906, 0.917, 0.943, 0.961, 0.973, 0.982, 0.989, 0.993, 0.995, 1.000],
    52: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.011, 0.011, 0.014, 0.018, 0.028, 0.032, 0.032, 0.037, 0.041, 0.046, 0.048, 0.050, 0.051, 0.053, 0.060, 0.069, 0.074, 0.080, 0.080, 0.083, 0.096, 0.098, 0.098, 0.099, 0.101, 0.103, 0.110, 0.119, 0.121, 0.122, 0.129, 0.133, 0.140, 0.154, 0.165, 0.165, 0.174, 0.190, 0.199, 0.216, 0.236, 0.262, 0.275, 0.307, 0.340, 0.385, 0.424, 0.459, 0.482, 0.511, 0.527, 0.543, 0.571, 0.601, 0.617, 0.624, 0.651, 0.668, 0.676, 0.688, 0.695, 0.716, 0.732, 0.750, 0.768, 0.778, 0.793, 0.812, 0.848, 0.885, 0.906, 0.917, 0.943, 0.961, 0.973, 0.982, 0.989, 0.993, 0.995, 1.000],
    53: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.004, 0.005, 0.005, 0.011, 0.011, 0.016, 0.018, 0.028, 0.032, 0.032, 0.037, 0.041, 0.048, 0.048, 0.050, 0.051, 0.053, 0.066, 0.076, 0.076, 0.080, 0.080, 0.085, 0.096, 0.098, 0.098, 0.099, 0.101, 0.103, 0.110, 0.119, 0.121, 0.124, 0.133, 0.137, 0.145, 0.154, 0.165, 0.167, 0.179, 0.190, 0.200, 0.223, 0.246, 0.264, 0.275, 0.309, 0.353, 0.388, 0.434, 0.463, 0.491, 0.512, 0.528, 0.550, 0.576, 0.601, 0.621, 0.626, 0.656, 0.670, 0.677, 0.688, 0.697, 0.720, 0.738, 0.752, 0.768, 0.778, 0.794, 0.816, 0.853, 0.892, 0.908, 0.917, 0.947, 0.961, 0.973, 0.982, 0.989, 0.993, 0.995, 1.000],
    54: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.005, 0.007, 0.007, 0.011, 0.012, 0.016, 0.023, 0.028, 0.034, 0.034, 0.039, 0.043, 0.048, 0.050, 0.050, 0.051, 0.057, 0.069, 0.076, 0.076, 0.080, 0.085, 0.085, 0.096, 0.098, 0.098, 0.099, 0.103, 0.103, 0.110, 0.121, 0.121, 0.126, 0.133, 0.138, 0.152, 0.160, 0.167, 0.170, 0.183, 0.195, 0.200, 0.223, 0.250, 0.264, 0.275, 0.316, 0.358, 0.394, 0.440, 0.468, 0.491, 0.514, 0.530, 0.550, 0.576, 0.605, 0.621, 0.626, 0.658, 0.670, 0.677, 0.688, 0.697, 0.720, 0.739, 0.757, 0.770, 0.782, 0.800, 0.819, 0.856, 0.894, 0.908, 0.922, 0.949, 0.961, 0.973, 0.982, 0.989, 0.993, 0.995, 1.000],
    55: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.005, 0.007, 0.007, 0.011, 0.012, 0.016, 0.027, 0.028, 0.034, 0.034, 0.041, 0.043, 0.048, 0.050, 0.050, 0.053, 0.059, 0.069, 0.076, 0.076, 0.080, 0.085, 0.089, 0.096, 0.098, 0.098, 0.101, 0.103, 0.110, 0.115, 0.121, 0.122, 0.126, 0.133, 0.140, 0.152, 0.160, 0.167, 0.170, 0.183, 0.195, 0.204, 0.223, 0.252, 0.264, 0.275, 0.323, 0.365, 0.399, 0.441, 0.470, 0.493, 0.516, 0.530, 0.553, 0.580, 0.605, 0.621, 0.628, 0.658, 0.676, 0.677, 0.690, 0.699, 0.725, 0.743, 0.757, 0.775, 0.784, 0.801, 0.826, 0.858, 0.895, 0.910, 0.922, 0.949, 0.961, 0.973, 0.984, 0.989, 0.993, 0.995, 1.000],
    56: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.005, 0.007, 0.009, 0.011, 0.012, 0.018, 0.028, 0.030, 0.034, 0.034, 0.041, 0.043, 0.048, 0.050, 0.050, 0.053, 0.064, 0.069, 0.076, 0.076, 0.085, 0.085, 0.092, 0.098, 0.098, 0.098, 0.103, 0.103, 0.110, 0.119, 0.122, 0.122, 0.126, 0.133, 0.140, 0.161, 0.163, 0.168, 0.170, 0.184, 0.195, 0.206, 0.225, 0.252, 0.264, 0.277, 0.337, 0.369, 0.404, 0.449, 0.473, 0.495, 0.518, 0.534, 0.557, 0.583, 0.605, 0.621, 0.635, 0.660, 0.676, 0.677, 0.690, 0.700, 0.730, 0.746, 0.757, 0.777, 0.787, 0.809, 0.828, 0.858, 0.897, 0.910, 0.924, 0.949, 0.961, 0.973, 0.984, 0.989, 0.993, 0.995, 1.000],
    57: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.005, 0.007, 0.009, 0.011, 0.012, 0.023, 0.028, 0.030, 0.034, 0.039, 0.041, 0.043, 0.048, 0.050, 0.050, 0.055, 0.066, 0.069, 0.076, 0.080, 0.085, 0.087, 0.092, 0.098, 0.098, 0.099, 0.103, 0.110, 0.113, 0.121, 0.122, 0.122, 0.126, 0.138, 0.151, 0.161, 0.163, 0.168, 0.170, 0.190, 0.195, 0.206, 0.234, 0.254, 0.266, 0.284, 0.339, 0.372, 0.406, 0.452, 0.475, 0.500, 0.520, 0.537, 0.562, 0.583, 0.605, 0.624, 0.637, 0.663, 0.676, 0.677, 0.691, 0.707, 0.732, 0.748, 0.757, 0.777, 0.791, 0.810, 0.830, 0.860, 0.897, 0.915, 0.929, 0.949, 0.961, 0.973, 0.984, 0.989, 0.993, 0.995, 1.000],
    58: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.005, 0.007, 0.009, 0.016, 0.018, 0.025, 0.028, 0.032, 0.035, 0.041, 0.043, 0.043, 0.050, 0.050, 0.055, 0.060, 0.067, 0.073, 0.080, 0.082, 0.087, 0.087, 0.096, 0.098, 0.098, 0.099, 0.103, 0.110, 0.117, 0.121, 0.122, 0.126, 0.135, 0.140, 0.158, 0.163, 0.165, 0.168, 0.174, 0.191, 0.195, 0.209, 0.239, 0.254, 0.266, 0.287, 0.346, 0.376, 0.417, 0.456, 0.475, 0.504, 0.521, 0.539, 0.562, 0.585, 0.605, 0.624, 0.638, 0.665, 0.676, 0.677, 0.699, 0.707, 0.736, 0.748, 0.762, 0.777, 0.794, 0.816, 0.832, 0.862, 0.902, 0.915, 0.929, 0.949, 0.961, 0.973, 0.984, 0.991, 0.993, 0.995, 1.000],
    59: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.002, 0.004, 0.007, 0.007, 0.009, 0.016, 0.018, 0.025, 0.028, 0.032, 0.035, 0.041, 0.043, 0.043, 0.050, 0.050, 0.055, 0.060, 0.067, 0.073, 0.080, 0.083, 0.087, 0.087, 0.096, 0.098, 0.098, 0.099, 0.103, 0.110, 0.119, 0.121, 0.122, 0.126, 0.135, 0.140, 0.160, 0.163, 0.165, 0.168, 0.176, 0.193, 0.195, 0.215, 0.241, 0.257, 0.268, 0.293, 0.351, 0.376, 0.422, 0.457, 0.477, 0.511, 0.525, 0.543, 0.566, 0.589, 0.610, 0.626, 0.642, 0.665, 0.676, 0.683, 0.700, 0.709, 0.738, 0.748, 0.762, 0.780, 0.794, 0.819, 0.832, 0.865, 0.904, 0.915, 0.929, 0.949, 0.963, 0.975, 0.984, 0.991, 0.993, 0.995, 1.000],
    60: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.004, 0.009, 0.009, 0.011, 0.016, 0.020, 0.025, 0.030, 0.032, 0.035, 0.041, 0.044, 0.048, 0.050, 0.051, 0.055, 0.064, 0.071, 0.073, 0.083, 0.085, 0.087, 0.090, 0.096, 0.098, 0.098, 0.099, 0.106, 0.115, 0.119, 0.121, 0.124, 0.129, 0.135, 0.145, 0.161, 0.163, 0.165, 0.168, 0.179, 0.193, 0.199, 0.215, 0.241, 0.259, 0.268, 0.301, 0.356, 0.381, 0.429, 0.459, 0.479, 0.516, 0.528, 0.546, 0.567, 0.592, 0.613, 0.626, 0.642, 0.670, 0.676, 0.686, 0.700, 0.716, 0.741, 0.752, 0.764, 0.784, 0.798, 0.821, 0.837, 0.867, 0.906, 0.915, 0.931, 0.949, 0.965, 0.975, 0.988, 0.991, 0.993, 0.995, 1.000],
    61: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.002, 0.004, 0.005, 0.009, 0.009, 0.011, 0.016, 0.020, 0.028, 0.030, 0.032, 0.037, 0.041, 0.044, 0.048, 0.050, 0.051, 0.059, 0.067, 0.074, 0.076, 0.083, 0.085, 0.087, 0.090, 0.096, 0.098, 0.098, 0.105, 0.106, 0.115, 0.119, 0.124, 0.126, 0.129, 0.138, 0.151, 0.161, 0.163, 0.165, 0.170, 0.179, 0.193, 0.199, 0.216, 0.241, 0.261, 0.275, 0.312, 0.362, 0.387, 0.438, 0.463, 0.484, 0.518, 0.528, 0.548, 0.571, 0.596, 0.613, 0.629, 0.642, 0.670, 0.676, 0.688, 0.702, 0.720, 0.743, 0.754, 0.768, 0.785, 0.805, 0.824, 0.839, 0.869, 0.906, 0.915, 0.933, 0.949, 0.966, 0.975, 0.988, 0.991, 0.993, 0.995, 1.000],
    62: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.005, 0.009, 0.009, 0.011, 0.016, 0.020, 0.028, 0.030, 0.032, 0.037, 0.043, 0.044, 0.048, 0.050, 0.051, 0.062, 0.069, 0.074, 0.076, 0.083, 0.087, 0.089, 0.096, 0.096, 0.098, 0.098, 0.105, 0.106, 0.115, 0.121, 0.124, 0.126, 0.131, 0.138, 0.154, 0.161, 0.165, 0.165, 0.176, 0.183, 0.193, 0.202, 0.218, 0.248, 0.266, 0.275, 0.317, 0.362, 0.392, 0.441, 0.468, 0.486, 0.520, 0.528, 0.553, 0.574, 0.598, 0.615, 0.629, 0.644, 0.670, 0.677, 0.691, 0.704, 0.723, 0.745, 0.755, 0.773, 0.787, 0.807, 0.826, 0.846, 0.871, 0.906, 0.915, 0.933, 0.950, 0.966, 0.977, 0.988, 0.991, 0.993, 0.995, 1.000],
    63: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.007, 0.009, 0.009, 0.011, 0.016, 0.021, 0.028, 0.030, 0.034, 0.037, 0.043, 0.044, 0.048, 0.051, 0.051, 0.066, 0.069, 0.074, 0.082, 0.085, 0.087, 0.089, 0.096, 0.098, 0.098, 0.098, 0.105, 0.113, 0.119, 0.121, 0.124, 0.126, 0.133, 0.142, 0.158, 0.161, 0.165, 0.165, 0.177, 0.184, 0.193, 0.209, 0.223, 0.255, 0.268, 0.289, 0.319, 0.367, 0.404, 0.443, 0.472, 0.489, 0.523, 0.534, 0.557, 0.580, 0.598, 0.615, 0.635, 0.644, 0.674, 0.679, 0.691, 0.706, 0.729, 0.745, 0.757, 0.775, 0.787, 0.809, 0.826, 0.848, 0.878, 0.906, 0.915, 0.933, 0.952, 0.968, 0.979, 0.988, 0.991, 0.993, 0.995, 1.000],
    64: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.007, 0.009, 0.009, 0.011, 0.016, 0.021, 0.028, 0.030, 0.034, 0.037, 0.043, 0.044, 0.050, 0.051, 0.051, 0.069, 0.069, 0.074, 0.083, 0.085, 0.087, 0.089, 0.096, 0.098, 0.098, 0.098, 0.112, 0.117, 0.119, 0.121, 0.124, 0.128, 0.137, 0.144, 0.160, 0.165, 0.165, 0.165, 0.179, 0.188, 0.199, 0.209, 0.227, 0.257, 0.268, 0.293, 0.323, 0.371, 0.406, 0.445, 0.475, 0.495, 0.523, 0.537, 0.559, 0.583, 0.599, 0.615, 0.635, 0.644, 0.674, 0.681, 0.691, 0.707, 0.729, 0.748, 0.759, 0.775, 0.789, 0.812, 0.830, 0.848, 0.885, 0.906, 0.915, 0.934, 0.952, 0.970, 0.979, 0.988, 0.991, 0.993, 0.995, 1.000],
    65: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.007, 0.009, 0.009, 0.011, 0.016, 0.021, 0.028, 0.030, 0.034, 0.037, 0.043, 0.044, 0.050, 0.051, 0.053, 0.069, 0.071, 0.074, 0.083, 0.085, 0.087, 0.089, 0.096, 0.098, 0.098, 0.099, 0.113, 0.117, 0.121, 0.121, 0.126, 0.129, 0.138, 0.147, 0.161, 0.165, 0.165, 0.172, 0.181, 0.188, 0.199, 0.211, 0.227, 0.262, 0.278, 0.298, 0.328, 0.376, 0.413, 0.449, 0.479, 0.496, 0.523, 0.541, 0.566, 0.585, 0.605, 0.621, 0.638, 0.647, 0.674, 0.683, 0.695, 0.709, 0.734, 0.752, 0.761, 0.777, 0.791, 0.816, 0.832, 0.853, 0.887, 0.906, 0.915, 0.934, 0.952, 0.970, 0.979, 0.988, 0.993, 0.993, 0.996, 1.000],
    66: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.004, 0.007, 0.009, 0.009, 0.011, 0.016, 0.021, 0.028, 0.030, 0.034, 0.037, 0.043, 0.048, 0.050, 0.051, 0.057, 0.069, 0.071, 0.074, 0.083, 0.085, 0.087, 0.092, 0.096, 0.098, 0.098, 0.112, 0.117, 0.117, 0.121, 0.121, 0.126, 0.129, 0.138, 0.147, 0.161, 0.165, 0.168, 0.172, 0.183, 0.190, 0.202, 0.216, 0.232, 0.264, 0.280, 0.298, 0.330, 0.379, 0.420, 0.450, 0.482, 0.500, 0.527, 0.543, 0.567, 0.585, 0.605, 0.621, 0.638, 0.647, 0.677, 0.683, 0.695, 0.709, 0.736, 0.752, 0.764, 0.778, 0.793, 0.817, 0.837, 0.855, 0.888, 0.906, 0.915, 0.936, 0.954, 0.970, 0.979, 0.988, 0.993, 0.993, 0.996, 1.000],
    67: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.009, 0.009, 0.009, 0.011, 0.016, 0.021, 0.028, 0.034, 0.034, 0.039, 0.043, 0.048, 0.050, 0.051, 0.059, 0.071, 0.071, 0.074, 0.083, 0.085, 0.089, 0.094, 0.096, 0.098, 0.099, 0.112, 0.117, 0.119, 0.121, 0.121, 0.126, 0.133, 0.138, 0.147, 0.161, 0.165, 0.170, 0.174, 0.184, 0.190, 0.204, 0.220, 0.241, 0.266, 0.280, 0.301, 0.342, 0.381, 0.426, 0.454, 0.486, 0.500, 0.530, 0.550, 0.571, 0.585, 0.605, 0.621, 0.638, 0.651, 0.679, 0.684, 0.695, 0.715, 0.739, 0.754, 0.764, 0.778, 0.798, 0.821, 0.837, 0.858, 0.888, 0.906, 0.918, 0.936, 0.956, 0.970, 0.979, 0.988, 0.993, 0.995, 0.996, 1.000],
    68: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.004, 0.009, 0.009, 0.009, 0.011, 0.016, 0.021, 0.028, 0.034, 0.035, 0.039, 0.043, 0.048, 0.051, 0.057, 0.064, 0.071, 0.073, 0.078, 0.083, 0.085, 0.089, 0.094, 0.096, 0.098, 0.103, 0.113, 0.119, 0.121, 0.121, 0.121, 0.128, 0.133, 0.140, 0.152, 0.161, 0.165, 0.170, 0.176, 0.186, 0.195, 0.206, 0.222, 0.243, 0.268, 0.280, 0.307, 0.358, 0.381, 0.427, 0.454, 0.486, 0.502, 0.532, 0.551, 0.576, 0.587, 0.608, 0.622, 0.638, 0.652, 0.679, 0.684, 0.697, 0.715, 0.743, 0.755, 0.766, 0.780, 0.801, 0.821, 0.839, 0.860, 0.890, 0.906, 0.920, 0.936, 0.956, 0.972, 0.979, 0.988, 0.993, 0.995, 0.996, 1.000],
    69: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.009, 0.009, 0.009, 0.011, 0.018, 0.025, 0.032, 0.035, 0.037, 0.039, 0.044, 0.051, 0.055, 0.059, 0.064, 0.073, 0.074, 0.078, 0.083, 0.085, 0.089, 0.094, 0.096, 0.098, 0.105, 0.115, 0.121, 0.121, 0.121, 0.122, 0.128, 0.135, 0.144, 0.156, 0.161, 0.167, 0.170, 0.176, 0.188, 0.195, 0.207, 0.223, 0.252, 0.270, 0.287, 0.317, 0.367, 0.387, 0.429, 0.459, 0.488, 0.511, 0.537, 0.557, 0.580, 0.590, 0.613, 0.626, 0.640, 0.656, 0.679, 0.688, 0.706, 0.718, 0.743, 0.757, 0.768, 0.782, 0.807, 0.823, 0.842, 0.860, 0.894, 0.911, 0.924, 0.940, 0.957, 0.973, 0.979, 0.988, 0.993, 0.995, 0.996, 1.000],
    70: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.005, 0.009, 0.009, 0.009, 0.011, 0.018, 0.025, 0.035, 0.035, 0.039, 0.041, 0.048, 0.051, 0.055, 0.059, 0.064, 0.074, 0.074, 0.078, 0.083, 0.085, 0.089, 0.094, 0.096, 0.098, 0.105, 0.115, 0.121, 0.121, 0.121, 0.122, 0.128, 0.137, 0.144, 0.156, 0.161, 0.167, 0.170, 0.177, 0.191, 0.195, 0.209, 0.225, 0.252, 0.270, 0.294, 0.323, 0.371, 0.392, 0.431, 0.465, 0.488, 0.512, 0.537, 0.559, 0.582, 0.594, 0.613, 0.631, 0.644, 0.656, 0.684, 0.688, 0.707, 0.718, 0.746, 0.757, 0.768, 0.782, 0.807, 0.823, 0.846, 0.865, 0.894, 0.911, 0.924, 0.940, 0.957, 0.973, 0.979, 0.988, 0.993, 0.995, 0.996, 1.000],
    71: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.005, 0.009, 0.009, 0.009, 0.016, 0.020, 0.032, 0.035, 0.035, 0.041, 0.041, 0.048, 0.055, 0.057, 0.060, 0.067, 0.074, 0.076, 0.078, 0.083, 0.085, 0.092, 0.094, 0.096, 0.101, 0.106, 0.115, 0.121, 0.121, 0.121, 0.124, 0.128, 0.138, 0.144, 0.160, 0.161, 0.167, 0.174, 0.179, 0.191, 0.197, 0.213, 0.225, 0.254, 0.275, 0.300, 0.339, 0.374, 0.401, 0.438, 0.473, 0.489, 0.514, 0.539, 0.564, 0.583, 0.598, 0.619, 0.633, 0.644, 0.660, 0.684, 0.691, 0.709, 0.722, 0.746, 0.757, 0.770, 0.782, 0.816, 0.823, 0.848, 0.865, 0.894, 0.911, 0.927, 0.940, 0.961, 0.973, 0.979, 0.988, 0.993, 0.995, 0.996, 1.000],
    72: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.007, 0.009, 0.009, 0.011, 0.016, 0.021, 0.032, 0.035, 0.035, 0.041, 0.041, 0.048, 0.055, 0.059, 0.060, 0.067, 0.074, 0.076, 0.078, 0.083, 0.085, 0.092, 0.094, 0.096, 0.101, 0.106, 0.115, 0.121, 0.121, 0.121, 0.124, 0.128, 0.138, 0.144, 0.160, 0.161, 0.168, 0.176, 0.183, 0.191, 0.199, 0.216, 0.229, 0.255, 0.280, 0.309, 0.342, 0.378, 0.411, 0.443, 0.473, 0.491, 0.516, 0.541, 0.567, 0.587, 0.599, 0.619, 0.633, 0.647, 0.660, 0.684, 0.695, 0.709, 0.723, 0.748, 0.761, 0.771, 0.784, 0.816, 0.823, 0.848, 0.872, 0.894, 0.911, 0.929, 0.940, 0.963, 0.973, 0.979, 0.988, 0.993, 0.995, 0.996, 1.000],
    73: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.007, 0.009, 0.009, 0.011, 0.018, 0.021, 0.034, 0.035, 0.035, 0.041, 0.048, 0.050, 0.057, 0.059, 0.060, 0.073, 0.076, 0.078, 0.078, 0.083, 0.087, 0.092, 0.094, 0.098, 0.103, 0.106, 0.115, 0.121, 0.121, 0.122, 0.126, 0.129, 0.142, 0.144, 0.160, 0.165, 0.168, 0.177, 0.184, 0.191, 0.202, 0.220, 0.236, 0.259, 0.282, 0.310, 0.346, 0.385, 0.415, 0.449, 0.477, 0.495, 0.520, 0.546, 0.571, 0.589, 0.601, 0.621, 0.633, 0.651, 0.663, 0.684, 0.702, 0.709, 0.725, 0.750, 0.762, 0.773, 0.789, 0.817, 0.826, 0.851, 0.876, 0.894, 0.913, 0.929, 0.947, 0.968, 0.973, 0.979, 0.988, 0.993, 0.995, 0.996, 1.000],
    74: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.005, 0.007, 0.009, 0.009, 0.011, 0.018, 0.021, 0.035, 0.035, 0.035, 0.041, 0.048, 0.050, 0.057, 0.059, 0.060, 0.073, 0.076, 0.078, 0.078, 0.083, 0.087, 0.092, 0.094, 0.098, 0.103, 0.106, 0.115, 0.121, 0.121, 0.122, 0.126, 0.131, 0.142, 0.144, 0.161, 0.165, 0.170, 0.179, 0.186, 0.191, 0.204, 0.223, 0.238, 0.259, 0.282, 0.310, 0.353, 0.385, 0.415, 0.449, 0.479, 0.495, 0.521, 0.546, 0.573, 0.590, 0.605, 0.622, 0.635, 0.651, 0.667, 0.686, 0.702, 0.709, 0.727, 0.750, 0.764, 0.775, 0.791, 0.821, 0.832, 0.851, 0.878, 0.895, 0.915, 0.931, 0.947, 0.970, 0.975, 0.980, 0.988, 0.993, 0.995, 0.996, 1.000],
    75: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.004, 0.007, 0.007, 0.009, 0.011, 0.011, 0.018, 0.027, 0.035, 0.035, 0.037, 0.041, 0.048, 0.050, 0.057, 0.059, 0.067, 0.073, 0.076, 0.078, 0.082, 0.083, 0.087, 0.094, 0.094, 0.101, 0.105, 0.106, 0.115, 0.121, 0.121, 0.124, 0.126, 0.135, 0.142, 0.144, 0.163, 0.165, 0.176, 0.179, 0.186, 0.193, 0.207, 0.223, 0.238, 0.262, 0.285, 0.314, 0.360, 0.388, 0.417, 0.450, 0.482, 0.496, 0.527, 0.550, 0.578, 0.590, 0.608, 0.622, 0.637, 0.652, 0.668, 0.686, 0.704, 0.713, 0.727, 0.752, 0.764, 0.775, 0.791, 0.821, 0.833, 0.853, 0.879, 0.899, 0.918, 0.933, 0.949, 0.970, 0.975, 0.980, 0.988, 0.995, 0.995, 0.996, 1.000],
    76: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.007, 0.007, 0.007, 0.009, 0.011, 0.011, 0.020, 0.030, 0.035, 0.035, 0.037, 0.041, 0.048, 0.051, 0.057, 0.059, 0.067, 0.076, 0.076, 0.082, 0.082, 0.083, 0.089, 0.094, 0.094, 0.101, 0.105, 0.106, 0.115, 0.121, 0.122, 0.124, 0.126, 0.135, 0.142, 0.144, 0.163, 0.167, 0.176, 0.179, 0.186, 0.193, 0.211, 0.225, 0.238, 0.264, 0.287, 0.316, 0.362, 0.390, 0.420, 0.457, 0.486, 0.504, 0.528, 0.553, 0.578, 0.594, 0.610, 0.622, 0.640, 0.652, 0.670, 0.686, 0.704, 0.715, 0.730, 0.755, 0.766, 0.777, 0.796, 0.823, 0.835, 0.853, 0.879, 0.901, 0.918, 0.933, 0.949, 0.970, 0.979, 0.984, 0.989, 0.995, 0.995, 0.996, 1.000],
    77: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.007, 0.007, 0.007, 0.011, 0.011, 0.018, 0.023, 0.030, 0.035, 0.035, 0.037, 0.041, 0.048, 0.051, 0.057, 0.060, 0.067, 0.076, 0.076, 0.082, 0.082, 0.083, 0.089, 0.094, 0.094, 0.101, 0.105, 0.108, 0.115, 0.121, 0.122, 0.124, 0.128, 0.135, 0.142, 0.147, 0.165, 0.167, 0.177, 0.179, 0.190, 0.202, 0.215, 0.225, 0.243, 0.268, 0.293, 0.317, 0.369, 0.392, 0.424, 0.463, 0.488, 0.505, 0.539, 0.555, 0.580, 0.596, 0.617, 0.626, 0.642, 0.654, 0.672, 0.686, 0.707, 0.718, 0.730, 0.761, 0.768, 0.778, 0.800, 0.828, 0.837, 0.853, 0.885, 0.904, 0.922, 0.933, 0.949, 0.970, 0.979, 0.986, 0.989, 0.995, 0.995, 0.996, 1.000],
    78: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.007, 0.007, 0.007, 0.009, 0.011, 0.012, 0.018, 0.025, 0.030, 0.035, 0.035, 0.039, 0.041, 0.050, 0.051, 0.057, 0.064, 0.069, 0.076, 0.076, 0.082, 0.083, 0.085, 0.092, 0.094, 0.098, 0.103, 0.106, 0.108, 0.115, 0.121, 0.122, 0.124, 0.131, 0.137, 0.142, 0.154, 0.167, 0.170, 0.177, 0.181, 0.190, 0.202, 0.215, 0.229, 0.246, 0.271, 0.294, 0.323, 0.369, 0.397, 0.431, 0.466, 0.489, 0.509, 0.541, 0.559, 0.580, 0.599, 0.617, 0.629, 0.642, 0.654, 0.676, 0.690, 0.709, 0.720, 0.734, 0.762, 0.768, 0.780, 0.801, 0.828, 0.839, 0.853, 0.892, 0.904, 0.922, 0.934, 0.949, 0.973, 0.980, 0.986, 0.989, 0.995, 0.996, 0.996, 1.000],
    79: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.007, 0.007, 0.009, 0.009, 0.012, 0.012, 0.021, 0.028, 0.030, 0.035, 0.035, 0.039, 0.041, 0.051, 0.057, 0.057, 0.066, 0.073, 0.078, 0.082, 0.082, 0.083, 0.087, 0.092, 0.096, 0.099, 0.103, 0.106, 0.108, 0.115, 0.121, 0.124, 0.124, 0.135, 0.137, 0.145, 0.156, 0.167, 0.172, 0.177, 0.183, 0.191, 0.206, 0.222, 0.230, 0.250, 0.273, 0.301, 0.330, 0.371, 0.399, 0.434, 0.470, 0.491, 0.511, 0.546, 0.562, 0.583, 0.599, 0.619, 0.631, 0.642, 0.656, 0.679, 0.695, 0.713, 0.720, 0.739, 0.764, 0.771, 0.784, 0.805, 0.828, 0.839, 0.855, 0.894, 0.904, 0.926, 0.934, 0.950, 0.975, 0.980, 0.986, 0.989, 0.995, 0.996, 0.996, 1.000],
    80: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.007, 0.007, 0.009, 0.009, 0.012, 0.012, 0.025, 0.028, 0.032, 0.035, 0.035, 0.041, 0.043, 0.051, 0.057, 0.059, 0.066, 0.073, 0.078, 0.082, 0.082, 0.083, 0.087, 0.092, 0.096, 0.101, 0.103, 0.106, 0.108, 0.117, 0.122, 0.124, 0.126, 0.135, 0.144, 0.147, 0.156, 0.167, 0.172, 0.177, 0.183, 0.193, 0.207, 0.222, 0.238, 0.257, 0.277, 0.303, 0.333, 0.371, 0.399, 0.436, 0.473, 0.495, 0.516, 0.550, 0.566, 0.585, 0.601, 0.628, 0.635, 0.647, 0.660, 0.686, 0.697, 0.716, 0.725, 0.741, 0.764, 0.777, 0.785, 0.805, 0.828, 0.839, 0.862, 0.895, 0.908, 0.927, 0.936, 0.950, 0.975, 0.980, 0.986, 0.989, 0.995, 0.996, 0.996, 1.000],
    81: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.007, 0.007, 0.009, 0.011, 0.012, 0.012, 0.025, 0.030, 0.032, 0.035, 0.035, 0.041, 0.046, 0.051, 0.057, 0.059, 0.066, 0.073, 0.078, 0.082, 0.082, 0.083, 0.087, 0.094, 0.099, 0.101, 0.105, 0.108, 0.121, 0.121, 0.122, 0.124, 0.126, 0.135, 0.145, 0.147, 0.161, 0.167, 0.172, 0.177, 0.184, 0.199, 0.209, 0.222, 0.243, 0.261, 0.278, 0.307, 0.342, 0.381, 0.402, 0.441, 0.482, 0.502, 0.523, 0.553, 0.569, 0.585, 0.608, 0.629, 0.637, 0.647, 0.663, 0.688, 0.699, 0.718, 0.727, 0.745, 0.764, 0.777, 0.787, 0.812, 0.830, 0.840, 0.869, 0.897, 0.913, 0.927, 0.938, 0.950, 0.975, 0.980, 0.986, 0.991, 0.995, 0.996, 0.996, 1.000],
    82: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.007, 0.007, 0.009, 0.011, 0.012, 0.018, 0.025, 0.030, 0.032, 0.035, 0.037, 0.041, 0.046, 0.053, 0.057, 0.059, 0.066, 0.076, 0.082, 0.082, 0.082, 0.083, 0.087, 0.099, 0.099, 0.101, 0.106, 0.108, 0.121, 0.121, 0.124, 0.124, 0.128, 0.137, 0.147, 0.147, 0.161, 0.168, 0.172, 0.177, 0.188, 0.199, 0.209, 0.227, 0.245, 0.266, 0.278, 0.309, 0.346, 0.387, 0.406, 0.441, 0.484, 0.509, 0.527, 0.553, 0.571, 0.587, 0.615, 0.631, 0.638, 0.647, 0.672, 0.690, 0.700, 0.720, 0.732, 0.746, 0.766, 0.777, 0.791, 0.814, 0.830, 0.848, 0.869, 0.899, 0.913, 0.927, 0.940, 0.952, 0.975, 0.982, 0.986, 0.991, 0.995, 0.996, 0.996, 1.000],
    83: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.005, 0.007, 0.007, 0.009, 0.011, 0.012, 0.020, 0.025, 0.030, 0.032, 0.035, 0.037, 0.044, 0.050, 0.057, 0.057, 0.062, 0.066, 0.076, 0.082, 0.082, 0.082, 0.083, 0.090, 0.099, 0.099, 0.103, 0.106, 0.108, 0.121, 0.122, 0.124, 0.124, 0.128, 0.140, 0.147, 0.147, 0.163, 0.170, 0.174, 0.177, 0.190, 0.199, 0.209, 0.232, 0.250, 0.266, 0.278, 0.309, 0.346, 0.387, 0.406, 0.449, 0.495, 0.509, 0.532, 0.555, 0.574, 0.590, 0.619, 0.631, 0.640, 0.651, 0.677, 0.690, 0.700, 0.723, 0.736, 0.752, 0.768, 0.782, 0.791, 0.817, 0.832, 0.849, 0.874, 0.902, 0.913, 0.927, 0.940, 0.952, 0.975, 0.982, 0.986, 0.991, 0.995, 0.996, 0.996, 1.000],
    84: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.005, 0.007, 0.007, 0.009, 0.011, 0.012, 0.023, 0.027, 0.032, 0.035, 0.035, 0.039, 0.044, 0.050, 0.057, 0.057, 0.062, 0.066, 0.076, 0.082, 0.082, 0.082, 0.083, 0.096, 0.099, 0.099, 0.105, 0.106, 0.117, 0.121, 0.122, 0.124, 0.124, 0.129, 0.140, 0.147, 0.147, 0.163, 0.170, 0.174, 0.183, 0.190, 0.199, 0.213, 0.236, 0.250, 0.270, 0.285, 0.316, 0.346, 0.387, 0.415, 0.459, 0.498, 0.512, 0.537, 0.559, 0.578, 0.599, 0.621, 0.631, 0.642, 0.660, 0.679, 0.691, 0.702, 0.727, 0.739, 0.755, 0.773, 0.784, 0.791, 0.817, 0.835, 0.856, 0.879, 0.904, 0.915, 0.929, 0.941, 0.952, 0.975, 0.982, 0.986, 0.991, 0.996, 0.996, 0.996, 1.000],
    85: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.005, 0.007, 0.007, 0.009, 0.011, 0.018, 0.025, 0.028, 0.032, 0.035, 0.037, 0.043, 0.048, 0.050, 0.057, 0.057, 0.062, 0.069, 0.076, 0.082, 0.082, 0.082, 0.092, 0.099, 0.099, 0.101, 0.105, 0.106, 0.121, 0.122, 0.122, 0.124, 0.126, 0.133, 0.142, 0.149, 0.151, 0.167, 0.170, 0.177, 0.184, 0.191, 0.200, 0.222, 0.238, 0.255, 0.275, 0.294, 0.321, 0.351, 0.390, 0.427, 0.463, 0.500, 0.518, 0.541, 0.564, 0.587, 0.606, 0.622, 0.635, 0.645, 0.660, 0.684, 0.695, 0.716, 0.730, 0.739, 0.761, 0.777, 0.785, 0.800, 0.823, 0.839, 0.858, 0.879, 0.904, 0.918, 0.931, 0.941, 0.952, 0.977, 0.984, 0.989, 0.993, 0.996, 0.996, 0.996, 1.000],
    86: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.004, 0.005, 0.007, 0.009, 0.009, 0.011, 0.020, 0.025, 0.028, 0.032, 0.037, 0.037, 0.043, 0.048, 0.053, 0.057, 0.062, 0.069, 0.071, 0.076, 0.082, 0.082, 0.085, 0.094, 0.099, 0.099, 0.101, 0.106, 0.108, 0.121, 0.122, 0.122, 0.126, 0.126, 0.138, 0.142, 0.151, 0.152, 0.170, 0.170, 0.181, 0.184, 0.193, 0.206, 0.222, 0.243, 0.262, 0.280, 0.298, 0.328, 0.353, 0.395, 0.434, 0.468, 0.504, 0.523, 0.550, 0.569, 0.598, 0.612, 0.629, 0.635, 0.645, 0.663, 0.684, 0.699, 0.718, 0.730, 0.739, 0.762, 0.778, 0.791, 0.800, 0.828, 0.846, 0.871, 0.885, 0.908, 0.918, 0.934, 0.945, 0.952, 0.977, 0.986, 0.989, 0.993, 0.996, 0.996, 0.996, 1.000],
    87: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.007, 0.009, 0.011, 0.012, 0.023, 0.025, 0.028, 0.032, 0.037, 0.037, 0.043, 0.050, 0.055, 0.057, 0.067, 0.071, 0.071, 0.076, 0.082, 0.087, 0.092, 0.094, 0.099, 0.099, 0.101, 0.106, 0.115, 0.121, 0.122, 0.124, 0.126, 0.128, 0.138, 0.142, 0.152, 0.158, 0.170, 0.172, 0.181, 0.184, 0.195, 0.209, 0.230, 0.245, 0.264, 0.282, 0.307, 0.333, 0.356, 0.399, 0.443, 0.479, 0.512, 0.528, 0.559, 0.576, 0.599, 0.615, 0.629, 0.637, 0.654, 0.672, 0.690, 0.706, 0.722, 0.730, 0.739, 0.768, 0.784, 0.791, 0.801, 0.828, 0.846, 0.874, 0.888, 0.910, 0.920, 0.936, 0.947, 0.959, 0.979, 0.986, 0.991, 0.993, 0.996, 0.996, 0.996, 1.000],
    88: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.007, 0.009, 0.011, 0.012, 0.023, 0.025, 0.028, 0.032, 0.037, 0.039, 0.044, 0.050, 0.055, 0.059, 0.067, 0.071, 0.073, 0.080, 0.082, 0.087, 0.092, 0.094, 0.099, 0.099, 0.101, 0.106, 0.115, 0.122, 0.124, 0.124, 0.128, 0.129, 0.138, 0.149, 0.152, 0.160, 0.172, 0.176, 0.181, 0.188, 0.195, 0.215, 0.234, 0.250, 0.268, 0.287, 0.309, 0.339, 0.358, 0.402, 0.463, 0.493, 0.520, 0.537, 0.567, 0.580, 0.603, 0.619, 0.633, 0.647, 0.667, 0.679, 0.697, 0.706, 0.727, 0.730, 0.745, 0.771, 0.789, 0.796, 0.803, 0.830, 0.851, 0.883, 0.892, 0.911, 0.924, 0.936, 0.949, 0.959, 0.979, 0.988, 0.991, 0.996, 0.996, 0.996, 0.996, 1.000],
    89: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.007, 0.009, 0.011, 0.018, 0.023, 0.025, 0.030, 0.035, 0.039, 0.043, 0.048, 0.053, 0.055, 0.060, 0.069, 0.073, 0.074, 0.080, 0.085, 0.090, 0.092, 0.094, 0.099, 0.099, 0.105, 0.106, 0.119, 0.122, 0.124, 0.126, 0.129, 0.133, 0.144, 0.151, 0.156, 0.167, 0.174, 0.179, 0.186, 0.193, 0.204, 0.225, 0.245, 0.255, 0.273, 0.296, 0.326, 0.358, 0.372, 0.427, 0.475, 0.505, 0.528, 0.548, 0.573, 0.589, 0.608, 0.624, 0.640, 0.654, 0.670, 0.690, 0.706, 0.713, 0.727, 0.736, 0.748, 0.784, 0.789, 0.798, 0.812, 0.848, 0.869, 0.890, 0.899, 0.922, 0.933, 0.941, 0.952, 0.963, 0.982, 0.989, 0.991, 0.996, 0.996, 0.996, 0.996, 1.000],
    90: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.007, 0.007, 0.009, 0.012, 0.018, 0.023, 0.025, 0.030, 0.037, 0.039, 0.048, 0.050, 0.055, 0.055, 0.064, 0.071, 0.073, 0.076, 0.080, 0.087, 0.090, 0.092, 0.096, 0.101, 0.101, 0.106, 0.106, 0.122, 0.122, 0.126, 0.128, 0.131, 0.140, 0.149, 0.152, 0.161, 0.172, 0.176, 0.181, 0.190, 0.202, 0.211, 0.234, 0.245, 0.257, 0.282, 0.303, 0.330, 0.367, 0.399, 0.441, 0.482, 0.511, 0.537, 0.559, 0.580, 0.594, 0.613, 0.629, 0.644, 0.656, 0.676, 0.695, 0.706, 0.715, 0.727, 0.738, 0.754, 0.787, 0.793, 0.809, 0.832, 0.856, 0.871, 0.895, 0.908, 0.926, 0.933, 0.947, 0.954, 0.965, 0.982, 0.989, 0.991, 0.996, 0.996, 0.996, 0.996, 1.000]
}
